// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "api.h"

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <zmq.h>

#include "protos/control-rpc.pb-c.h"
#include "protos/control-scan.pb-c.h"
#include "protos/control-statistic.pb-c.h"
#include "protos/control-verbosity.pb-c.h"
#include "protos/control-version.pb-c.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "clock.h"
#include "common/statistics.h"
#include "config.h"
#include "fscan.h"

#define zerror(fmt, args...) error(fmt ": %s", ##args, zmq_strerror(zmq_errno()))

#define ZMQ_SOCKET_RCVTIMEO 200

static void *zmq_socket_create(void *context, int type)
{
	void *socket;

	if ((socket = zmq_socket(context, type)) == NULL) {
		zerror("Failed to create socket");
		return NULL;
	}

	int zero = 0;

	if (zmq_setsockopt(socket, ZMQ_LINGER, &zero, sizeof(zero)) < 0) {
		zerror("Failed to configure socket");
		return NULL;
	}

	int rcvtimeo = ZMQ_SOCKET_RCVTIMEO;

	if (zmq_setsockopt(socket, ZMQ_RCVTIMEO, &rcvtimeo, sizeof(rcvtimeo)) < 0) {
		zerror("Failed to configure socket");
		return NULL;
	}

	return socket;
}

static int zmq_socket_recv(struct core *core, zmq_msg_t *msg, void *socket)
{
	while (true) {
		if (zmq_msg_recv(msg, socket, 0) < 0) {
			int error = zmq_errno();

			if (!api_running(core) || error == ETERM) {
				return -1;
			}

			if (error == EAGAIN) {
				continue;
			}

			zerror("Failed to receive message");
			return -1;
		}

		return 0;
	}
}

struct rpc_request {
	char *service;
	uint8_t *data;
	size_t len;
};

static void rpc_request_free(struct rpc_request *request)
{
	if (request) {
		free(request->service);
		free(request->data);
		free(request);
	}
}

static struct rpc_request *rpc_recv(struct core *core)
{
	struct rpc_request *request;
	zmq_msg_t zmq_msg;
	size_t len;

	if (zmq_msg_init(&zmq_msg) < 0) {
		zerror("Failed to initialize zmq message");
		return NULL;
	}

	if (zmq_socket_recv(core, &zmq_msg, core->api->zmq_control) < 0) {
		zmq_msg_close(&zmq_msg);
		return NULL;
	}

	dump(zmq_msg_data(&zmq_msg), zmq_msg_size(&zmq_msg), "api.req");

	Proto__Control__Request *rpc = NULL;

	if ((rpc = proto__control__request__unpack(NULL, zmq_msg_size(&zmq_msg),
						   zmq_msg_data(&zmq_msg))) == NULL) {
		error("Failed to de-serialize control message");
		zmq_msg_close(&zmq_msg);
		return NULL;
	}

	zmq_msg_close(&zmq_msg);

	if ((request = zalloc(sizeof(*request))) == NULL) {
		eerror("Failed to allocate request");
		proto__control__request__free_unpacked(rpc, NULL);
		return NULL;
	}

	len = strlen(rpc->service) + 1;

	if ((request->service = malloc(len)) == NULL) {
		eerror("Failed to allocate service");
		proto__control__request__free_unpacked(rpc, NULL);
		rpc_request_free(request);
		return NULL;

	} else {
		memcpy(request->service, rpc->service, len);
	}

	if (rpc->has_protobuf) {
		request->len = rpc->protobuf.len;

		if ((request->data = malloc(request->len)) == NULL) {
			eerror("Failed to allocate data");
			proto__control__request__free_unpacked(rpc, NULL);
			rpc_request_free(request);
			return NULL;

		} else {
			memcpy(request->data, rpc->protobuf.data, request->len);
		}
	}

	proto__control__request__free_unpacked(rpc, NULL);
	return request;
}

enum rpc_status {
	RPC_OK = 0,
	RPC_FAILED = 1,
	RPC_BAD_REQUEST = 2,
	RPC_NOT_IMPLEMENTED = 3,
};

static int rpc_send(struct core *core, enum rpc_status status, uint8_t *data, size_t len)
{
	Proto__Control__Reply rpc = PROTO__CONTROL__REPLY__INIT;
	zmq_msg_t zmq_msg;
	size_t size;

	rpc.status = status;

	if (data && len) {
		rpc.has_protobuf = 1;
		rpc.protobuf.len = len;

		if ((rpc.protobuf.data = malloc(len)) == NULL) {
			eerror("Failed to allocate data");
			return -1;

		} else {
			memcpy(rpc.protobuf.data, data, len);
		}
	}

	size = proto__control__reply__get_packed_size(&rpc);

	if (zmq_msg_init_size(&zmq_msg, size) < 0) {
		zerror("Failed to initialize zmq message");
		free(rpc.protobuf.data);
		return -1;

	} else {
		proto__control__reply__pack(&rpc, zmq_msg_data(&zmq_msg));
		dump(zmq_msg_data(&zmq_msg), zmq_msg_size(&zmq_msg), "api.rep");
	}

	free(rpc.protobuf.data);

	if (zmq_msg_send(&zmq_msg, core->api->zmq_control, 0) < 0) {
		if (api_running(core)) {
			zerror("Failed to send control message");
		}

		zmq_msg_close(&zmq_msg);
		return -1;
	}

	zmq_msg_close(&zmq_msg);
	return 0;
}

static int rpc_verbosity(uint8_t *data, size_t len)
{
	Proto__Verbosity__Request *request = NULL;

	if ((request = proto__verbosity__request__unpack(NULL, len, data)) == NULL) {
		error("Failed to de-serialize verbosity message");
		return RPC_BAD_REQUEST;
	}

	setlogpriority(request->level);
	proto__verbosity__request__free_unpacked(request, NULL);
	return RPC_OK;
}

static int rpc_version(uint8_t **data, size_t *len)
{
	Proto__Version__Reply reply = PROTO__VERSION__REPLY__INIT;

	if (aprintf(&reply.version, "lorad %s", VERSION), 0) {
		error("Failed to format version string");
		return RPC_FAILED;
	}

	*len = proto__version__reply__get_packed_size(&reply);

	if ((*data = malloc(*len)) == NULL) {
		eerror("Failed to allocate memory");
		free(reply.version);
		return RPC_FAILED;
	}

	proto__version__reply__pack(&reply, *data);
	free(reply.version);
	return RPC_OK;
}

#define ALLOCATE(type, name, free)                   \
	type *name = NULL;                           \
	if ((name = malloc(sizeof(type))) == NULL) { \
		eerror("Failed to allocate " #name); \
		FREE(free);                          \
		return RPC_FAILED;                   \
	}

static void free_scan_status(Proto__Scan__StatusReply *reply)
{
	size_t n;

	if (reply) {
		for (n = 0; n < reply->n_receivers; n++) {
			free(reply->receivers[n]);
		}

		free(reply->receivers);
	}
}

static void free_fscan_receivers(struct fscan_receivers *fscan_receivers)
{
	if (fscan_receivers) {
		free(fscan_receivers);
	}
}

#undef FREE
#define FREE free_scan_status

static int rpc_scan_status(struct core *core, uint8_t **data, size_t *len)
{
	size_t n_receivers;
	struct fscan_receivers *fscan_receivers;
	size_t i = 0;

	ALLOCATE(Proto__Scan__StatusReply, reply, reply);
	proto__scan__status_reply__init(reply);

	fscan_get_scan_status(core, &n_receivers, &fscan_receivers);

	reply->n_receivers = n_receivers;
	reply->receivers = (Proto__Scan__StatusReply__Receiver **)malloc(sizeof(Proto__Scan__StatusReply__Receiver *) * n_receivers);

	for (i = 0; i < reply->n_receivers; i++) {
		reply->receivers[i] = (Proto__Scan__StatusReply__Receiver *)malloc(sizeof(Proto__Scan__StatusReply__Receiver));
		proto__scan__status_reply__receiver__init(reply->receivers[i]);
		reply->receivers[i]->id = fscan_receivers[i].id;
		reply->receivers[i]->busy = fscan_receivers[i].busy;
	}

	*len = proto__scan__status_reply__get_packed_size(reply);

	if ((*data = malloc(*len)) == NULL) {
		eerror("Failed to allocate memory");
		free_scan_status(reply);
		free_fscan_receivers(fscan_receivers);
		return RPC_FAILED;
	}

	proto__scan__status_reply__pack(reply, *data);
	free_scan_status(reply);
	free_fscan_receivers(fscan_receivers);
	return RPC_OK;
}

static int rpc_scan_start(struct core *core, uint8_t *data, size_t len)
{
	Proto__Scan__StartRequest *request = NULL;

	if ((request = proto__scan__start_request__unpack(NULL, len, data)) == NULL) {
		error("Failed to de-serialize scan start message");
		return RPC_BAD_REQUEST;
	}

	fscan_start_scan(core, request->receiver_id, request->n_frequencies, request->frequencies, request->has_samples, request->samples);
	proto__scan__start_request__free_unpacked(request, NULL);
	return RPC_OK;
}

static int rpc_scan_stop(struct core *core, uint8_t *data, size_t len)
{
	Proto__Scan__StopRequest *request = NULL;

	if ((request = proto__scan__stop_request__unpack(NULL, len, data)) == NULL) {
		error("Failed to de-serialize scan stop message");
		return RPC_BAD_REQUEST;
	}

	fscan_stop_scan(core, request->receiver_id);
	proto__scan__stop_request__free_unpacked(request, NULL);
	return RPC_OK;
}

static int add_statistic_signed_int(Proto__Statistic__Reply *reply, const char *name, int32_t value)
{
	Proto__Statistic__Reply__Statistic *statistic = NULL;

	if ((statistic = malloc(sizeof(*statistic))) == NULL) {
		eerror("Failed to allocate statistic");
		return -1;

	} else {
		proto__statistic__reply__statistic__init(statistic);
	}

	if ((reply->statistics = realloc(reply->statistics, (reply->n_statistics + 1) * sizeof(statistic))) == NULL) {
		eerror("Failed to re-allocate statistics");
		return -1;

	} else {
		reply->statistics[reply->n_statistics] = statistic;
		reply->n_statistics++;
	}

	if (aprintf(&statistic->name, "%s", name), 0) {
		error("Failed to format statistic name");
		return -1;
	}

	statistic->statistics_case = PROTO__STATISTIC__REPLY__STATISTIC__STATISTICS_SIGNED_INT;
	statistic->signed_int = value;
	return 0;
}

static int add_statistic_unsigned_int(Proto__Statistic__Reply *reply, const char *name, uint32_t value)
{
	Proto__Statistic__Reply__Statistic *statistic = NULL;

	if ((statistic = malloc(sizeof(*statistic))) == NULL) {
		eerror("Failed to allocate statistic");
		return -1;

	} else {
		proto__statistic__reply__statistic__init(statistic);
	}

	if ((reply->statistics = realloc(reply->statistics, (reply->n_statistics + 1) * sizeof(statistic))) == NULL) {
		eerror("Failed to re-allocate statistics");
		return -1;

	} else {
		reply->statistics[reply->n_statistics] = statistic;
		reply->n_statistics++;
	}

	if (aprintf(&statistic->name, "%s", name), 0) {
		error("Failed to format statistic name");
		return -1;
	}

	statistic->statistics_case = PROTO__STATISTIC__REPLY__STATISTIC__STATISTICS_UNSIGNED_INT;
	statistic->unsigned_int = value;
	return 0;
}

static int add_statistic_float(Proto__Statistic__Reply *reply, const char *name, float value)
{
	Proto__Statistic__Reply__Statistic *statistic = NULL;

	if ((statistic = malloc(sizeof(*statistic))) == NULL) {
		eerror("Failed to allocate statistic");
		return -1;

	} else {
		proto__statistic__reply__statistic__init(statistic);
	}

	if ((reply->statistics = realloc(reply->statistics, (reply->n_statistics + 1) * sizeof(statistic))) == NULL) {
		eerror("Failed to re-allocate statistics");
		return -1;

	} else {
		reply->statistics[reply->n_statistics] = statistic;
		reply->n_statistics++;
	}

	if (aprintf(&statistic->name, "%s", name), 0) {
		error("Failed to format statistic name");
		return -1;
	}

	statistic->statistics_case = PROTO__STATISTIC__REPLY__STATISTIC__STATISTICS_FLOAT;
	statistic->float_ = value;
	return 0;
}

static int add_statistic_string(Proto__Statistic__Reply *reply, const char *name, const char *value)
{
	Proto__Statistic__Reply__Statistic *statistic = NULL;

	if ((statistic = malloc(sizeof(*statistic))) == NULL) {
		eerror("Failed to allocate statistic");
		return -1;

	} else {
		proto__statistic__reply__statistic__init(statistic);
	}

	if ((reply->statistics = realloc(reply->statistics, (reply->n_statistics + 1) * sizeof(statistic))) == NULL) {
		eerror("Failed to re-allocate statistics");
		return -1;

	} else {
		reply->statistics[reply->n_statistics] = statistic;
		reply->n_statistics++;
	}

	if (aprintf(&statistic->name, "%s", name), 0) {
		error("Failed to format statistic name");
		return -1;
	}

	statistic->statistics_case = PROTO__STATISTIC__REPLY__STATISTIC__STATISTICS_STRING;
	if (aprintf(&statistic->string, "%s", value), 0) {
		error("Failed to format statistic value");
		return -1;
	}
	return 0;
}

static void free_statistics(Proto__Statistic__Reply *reply)
{
	size_t n;

	if (reply) {
		for (n = 0; n < reply->n_statistics; n++) {
			free(reply->statistics[n]->name);
			if (reply->statistics[n]->statistics_case == PROTO__STATISTIC__REPLY__STATISTIC__STATISTICS_STRING) {
				free(reply->statistics[n]->string);
			}
			free(reply->statistics[n]);
		}
		free(reply->statistics);

		free(reply->time);
		free(reply);
	}
}

#undef FREE
#define FREE free_statistics

#define ADD_STATISTIC_SIGNED_INT(reply, name, value)            \
	if (add_statistic_signed_int(reply, name, value) < 0) { \
		error("Failed to add " name " statistic");      \
		FREE(reply);                                    \
		return RPC_FAILED;                              \
	}

#define ADD_STATISTIC_UNSIGNED_INT(reply, name, value)            \
	if (add_statistic_unsigned_int(reply, name, value) < 0) { \
		error("Failed to add " name " statistic");        \
		FREE(reply);                                      \
		return RPC_FAILED;                                \
	}

#define ADD_STATISTIC_FLOAT(reply, name, value)            \
	if (add_statistic_float(reply, name, value) < 0) { \
		error("Failed to add " name " statistic"); \
		FREE(reply);                               \
		return RPC_FAILED;                         \
	}

#define ADD_STATISTIC_STRING(reply, name, value)            \
	if (add_statistic_string(reply, name, value) < 0) { \
		error("Failed to add " name " statistic");  \
		FREE(reply);                                \
		return RPC_FAILED;                          \
	}

static int rpc_statistics(struct core *core, uint8_t **data, size_t *len)
{
	struct timespec tp;

	ALLOCATE(Proto__Statistic__Reply, reply, reply);
	proto__statistic__reply__init(reply);

	ADD_STATISTIC_UNSIGNED_INT(reply, "uplink_received", core->statistics->uplink_received);
	ADD_STATISTIC_UNSIGNED_INT(reply, "uplink_invalid", core->statistics->uplink_invalid);
	ADD_STATISTIC_UNSIGNED_INT(reply, "uplink_valid", core->statistics->uplink_valid);
	ADD_STATISTIC_UNSIGNED_INT(reply, "uplink_bytes_received", core->statistics->uplink_bytes_received);
	ADD_STATISTIC_UNSIGNED_INT(reply, "downlink_requested", core->statistics->downlink_requested);
	ADD_STATISTIC_UNSIGNED_INT(reply, "downlink_scheduled", core->statistics->downlink_scheduled);
	ADD_STATISTIC_UNSIGNED_INT(reply, "downlink_transmitted", core->statistics->downlink_transmitted);
	ADD_STATISTIC_UNSIGNED_INT(reply, "downlink_bytes_transmitted", core->statistics->downlink_bytes_transmitted);

	if (core->statistics->has_radio_temperature) {
		ADD_STATISTIC_FLOAT(reply, "radio_temperature", core->statistics->radio_temperature);
	}

	if (core->statistics->has_gps_coordinates) {
		ADD_STATISTIC_FLOAT(reply, "gps_latitude", core->statistics->gps_latitude);
		ADD_STATISTIC_FLOAT(reply, "gps_longitude", core->statistics->gps_longitude);
		ADD_STATISTIC_SIGNED_INT(reply, "gps_altitude", core->statistics->gps_altitude);
	}

	if (core->statistics->has_gps_lost_pulse) {
		ADD_STATISTIC_UNSIGNED_INT(reply, "gps_lost_pulse", core->statistics->gps_lost_pulse);
	}

	fill_version_statistics(core->statistics);

	if (core->statistics->has_version_fpga) {
		ADD_STATISTIC_UNSIGNED_INT(reply, "version_fpga", core->statistics->version_fpga);
	}

	if (core->statistics->has_version_dsp) {
		ADD_STATISTIC_UNSIGNED_INT(reply, "version_dsp", core->statistics->version_dsp);
	}

	if (core->statistics->has_version_hal) {
		ADD_STATISTIC_STRING(reply, "version_hal", core->statistics->version_hal);
	}

	ALLOCATE(Proto__Time__Time, reply_time, reply);
	proto__time__time__init(reply_time);
	reply->time = reply_time;

	if (clock_get_tai_time(core->clock->tai_offset, &tp) == 0) {
		reply_time->format = PROTO__TIME__TIME__TIME_FORMAT__TAI_TIME;
		reply_time->time = tp.tv_sec;
		reply_time->has_microseconds = 1;
		reply_time->microseconds = tp.tv_nsec / 1000;

	} else if (clock_get_utc_time(&tp) == 0) {
		reply_time->format = PROTO__TIME__TIME__TIME_FORMAT__UNIX_TIME;
		reply_time->time = tp.tv_sec;
		reply_time->has_microseconds = 1;
		reply_time->microseconds = tp.tv_nsec / 1000;

	} else {
		free(reply_time);
		reply->time = NULL;
	}

	*len = proto__statistic__reply__get_packed_size(reply);

	if ((*data = malloc(*len)) == NULL) {
		eerror("Failed to allocate memory");
		free_statistics(reply);
		return RPC_FAILED;
	}

	proto__statistic__reply__pack(reply, *data);
	free_statistics(reply);
	return RPC_OK;
}

#undef FREE
#undef ALLOCATE
#undef ADD_STATISTIC_SIGNED_INT
#undef ADD_STATISTIC_UNSIGNED_INT
#undef ADD_STATISTIC_FLOAT

static enum rpc_status rpc_run(struct core *core, struct rpc_request *request, uint8_t **data, size_t *len)
{
	// Verbosity management
	if (strcmp(request->service, "set_verbosity") == 0) {
		return rpc_verbosity(request->data, request->len);

		// Request for version
	} else if (strcmp(request->service, "get_version") == 0) {
		return rpc_version(data, len);

		// Frequency scan management
	} else if (strcmp(request->service, "scan_status") == 0) {
		return rpc_scan_status(core, data, len);

	} else if (strcmp(request->service, "scan_start") == 0) {
		return rpc_scan_start(core, request->data, request->len);

	} else if (strcmp(request->service, "scan_stop") == 0) {
		return rpc_scan_stop(core, request->data, request->len);

		// Request for statistics
	} else if (strcmp(request->service, "get_statistics") == 0) {
		return rpc_statistics(core, data, len);
	}

	return RPC_NOT_IMPLEMENTED;
}

static void *rpc_thread(void *arg)
{
	struct core *core = arg;
	struct rpc_request *request;
	enum rpc_status status;
	uint8_t *data;
	size_t len;

	while (api_running(core)) {
		data = NULL;
		len = 0;

		if ((request = rpc_recv(core)) == NULL) {
			if (api_running(core)) {
				error("Failed to receive RPC request");
			}

			continue;
		}

		status = rpc_run(core, request, &data, &len);
		rpc_request_free(request);

		if (rpc_send(core, status, data, len) < 0) {
			if (api_running(core)) {
				error("Failed to send RPC reply");
			}
		}

		free(data);
	}

	pthread_exit(NULL);
}

struct api *api_init(void)
{
	struct api *api = NULL;

	if ((api = zalloc(sizeof(*api))) == NULL) {
		eerror("Failed to allocate API");
		return NULL;
	}

	if ((api->zmq_context = zmq_ctx_new()) == NULL) {
		zerror("Failed to create context");
		return NULL;
	}

	srand(time(NULL));
	api->message_id = ((uint32_t)rand() & 0xfff00000);

	return api;
}

int api_start(struct core *core)
{
	size_t nb;

	if ((core->api->zmq_uplink =
		 zmq_socket_create(core->api->zmq_context, ZMQ_PUB)) == NULL) {
		error("Failed to create uplink socket");
		api_shutdown(core);
		return -1;
	}

	if ((core->api->zmq_downlink =
		 zmq_socket_create(core->api->zmq_context, ZMQ_PULL)) == NULL) {
		error("Failed to create downlink socket");
		api_shutdown(core);
		return -1;
	}

	if ((core->api->zmq_control =
		 zmq_socket_create(core->api->zmq_context, ZMQ_REP)) == NULL) {
		error("Failed to create control socket");
		api_shutdown(core);
		return -1;
	}

	for (nb = 0; nb < core->config->api.uplinks.nb_strings; nb++) {
		if (zmq_bind(core->api->zmq_uplink,
			     core->config->api.uplinks.strings[nb]) < 0) {
			zerror("Failed to bind uplink socket %s", core->config->api.uplinks.strings[nb]);
			api_shutdown(core);
			return -1;
		}
	}

	for (nb = 0; nb < core->config->api.downlinks.nb_strings; nb++) {
		if (zmq_bind(core->api->zmq_downlink,
			     core->config->api.downlinks.strings[nb]) < 0) {
			zerror("Failed to bind downlink socket");
			api_shutdown(core);
			return -1;
		}
	}

	for (nb = 0; nb < core->config->api.controls.nb_strings; nb++) {
		if (zmq_bind(core->api->zmq_control,
			     core->config->api.controls.strings[nb]) < 0) {
			zerror("Failed to bind control socket");
			api_shutdown(core);
			return -1;
		}
	}

	if (pthread_create(&core->api->rpc_thread, NULL, rpc_thread, core) < 0) {
		error("Failed to start control thread");
		api_shutdown(core);
		return -1;
	}

	sleep_ms(500); /* 200 is enough to 100% receive the 1st message */

	return 0;
}

int api_running(struct core *core)
{
	if (core && core->api) {
		return core->api->zmq_uplink &&
		       core->api->zmq_downlink &&
		       core->api->zmq_control &&
		       core->api->zmq_context;
	}

	return 0;
}

int api_join(struct core *core)
{
	if (pthread_join(core->api->rpc_thread, NULL) < 0) {
		error("Failed to join control thread");
		return -1;
	}

	return 0;
}

void api_shutdown(struct core *core)
{
	if (api_running(core)) {
		if (core->api->zmq_uplink) {
			if (zmq_close(core->api->zmq_uplink) < 0) {
				zerror("Failed to close uplink socket");
			}

			core->api->zmq_uplink = NULL;
		}

		if (core->api->zmq_downlink) {
			if (zmq_close(core->api->zmq_downlink) < 0) {
				zerror("Failed to close downlink socket");
			}

			core->api->zmq_downlink = NULL;
		}

		if (core->api->zmq_control) {
			if (zmq_close(core->api->zmq_control) < 0) {
				zerror("Failed to close control socket");
			}

			core->api->zmq_control = NULL;
		}

#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(4, 2, 3)
		sleep_ms(ZMQ_SOCKET_RCVTIMEO);
#endif
	}
}

void api_free(struct core *core)
{
	if (core && core->api) {
		if (core->api->zmq_context) {
			if (zmq_ctx_term(core->api->zmq_context) < 0) {
				zerror("Failed to terminate context");
			}

			core->api->zmq_context = NULL;
		}

		free(core->api);
	}
}

void api_free_lora(Proto__Lora__Uplink *message)
{
	size_t n;

	if (message) {
		if (message->data) {
			free(message->data->payload.data);
			free(message->data);
		}

		for (n = 0; n < message->n_radios; n++) {
			if (message->radios[n]->demodulator) {
				free(message->radios[n]->demodulator->time);
				free(message->radios[n]->demodulator->snr);
				free(message->radios[n]->demodulator);
			}

			if (message->radios[n]->localization) {
				if (message->radios[n]->localization->nanoseconds_case == PROTO__LORA__UPLINK__RADIO__LOCALIZATION__NANOSECONDS_ENCRYPTED) {
					free(message->radios[n]->localization->encrypted.data);
				}

				free(message->radios[n]->localization);
			}

			free(message->radios[n]);
		}

		free(message->radios);
		free(message);
	}
}

void api_free_gfsk(Proto__Gfsk__Uplink *message)
{
	size_t n;

	if (message) {
		if (message->data) {
			free(message->data->payload.data);
			free(message->data);
		}

		for (n = 0; n < message->n_radios; n++) {
			if (message->radios[n]->demodulator) {
				free(message->radios[n]->demodulator->time);
				free(message->radios[n]->demodulator);
			}

			free(message->radios[n]);
		}

		free(message->radios);
		free(message);
	}
}

static int api_send_container(struct core *core, Proto__Uplink__Container *container)
{
	size_t size = proto__uplink__container__get_packed_size(container);
	zmq_msg_t zmq_msg;

	if (zmq_msg_init_size(&zmq_msg, size) < 0) {
		zerror("Failed to initialize zmq message");
		return -1;

	} else {
		proto__uplink__container__pack(container, zmq_msg_data(&zmq_msg));
		dump(zmq_msg_data(&zmq_msg), zmq_msg_size(&zmq_msg), "api.send");
	}

	if (zmq_msg_send(&zmq_msg, core->api->zmq_uplink, 0) < 0) {
		if (api_running(core)) {
			zerror("Failed to send uplink message");
		}

		zmq_msg_close(&zmq_msg);
		return -1;
	}

	zmq_msg_close(&zmq_msg);
	return 0;
}

int api_send_lora(struct core *core, Proto__Lora__Uplink *message)
{
	Proto__Uplink__Container container = PROTO__UPLINK__CONTAINER__INIT;
	container.messages_case = PROTO__UPLINK__CONTAINER__MESSAGES_LORA;
	container.lora = message;

	return api_send_container(core, &container);
}

int api_send_gfsk(struct core *core, Proto__Gfsk__Uplink *message)
{
	Proto__Uplink__Container container = PROTO__UPLINK__CONTAINER__INIT;
	container.messages_case = PROTO__UPLINK__CONTAINER__MESSAGES_GFSK;
	container.gfsk = message;

	return api_send_container(core, &container);
}

static int api_send_event_container(struct core *core, Proto__Event__Uplink *message)
{
	Proto__Uplink__Container container = PROTO__UPLINK__CONTAINER__INIT;
	container.messages_case = PROTO__UPLINK__CONTAINER__MESSAGES_EVENT;
	container.event = message;

	return api_send_container(core, &container);
}

static int api_send_scan_container(struct core *core, Proto__Scan__Uplink *message)
{
	Proto__Uplink__Container container = PROTO__UPLINK__CONTAINER__INIT;
	container.messages_case = PROTO__UPLINK__CONTAINER__MESSAGES_SCAN;
	container.scan = message;

	return api_send_container(core, &container);
}

int api_send_event_system(struct core *core, Proto__Event__System__Event event)
{
	Proto__Event__System message = PROTO__EVENT__SYSTEM__INIT;
	message.event = event;

	Proto__Event__Uplink container = PROTO__EVENT__UPLINK__INIT;
	container.events_case = PROTO__EVENT__UPLINK__EVENTS_SYSTEM;
	container.system = &message;

	return api_send_event_container(core, &container);
}

int api_send_event_transmission(struct core *core,
				uint32_t message_id, uint32_t transmission_id,
				Proto__Event__Transmission__Status status)
{
	Proto__Event__Transmission message = PROTO__EVENT__TRANSMISSION__INIT;
	message.message_id = message_id;
	message.transmission_id = transmission_id;
	message.status = status;

	Proto__Event__Uplink container = PROTO__EVENT__UPLINK__INIT;
	container.events_case = PROTO__EVENT__UPLINK__EVENTS_TRANSMISSION;
	container.transmission = &message;

	return api_send_event_container(core, &container);
}

int api_send_scan(struct core *core, uint32_t receiver_id, uint64_t sample_duration_us, uint32_t base_frequency,
		  uint32_t frequency_step, Proto__Time__Time *start_time, size_t n_single_scans,
		  struct single_scan *single_scans)
{
	unsigned int s;
	int ret;
	Proto__Scan__Uplink__SingleScan **proto_single_scans;
	Proto__Scan__Uplink__SingleScan *single_scan;
	Proto__Scan__Uplink message = PROTO__SCAN__UPLINK__INIT;

	proto_single_scans = malloc(n_single_scans * sizeof(Proto__Scan__Uplink__SingleScan *));

	for (s = 0; s < n_single_scans; s++) {
		single_scan = malloc(sizeof(Proto__Scan__Uplink__SingleScan));
		proto__scan__uplink__single_scan__init(single_scan);

		single_scan->rssi = single_scans[s].rssi;
		single_scan->has_frequency_offset = true;
		single_scan->frequency_offset = single_scans[s].frequency_offset;
		single_scan->has_microseconds = true;
		single_scan->microseconds = single_scans[s].duration;

		proto_single_scans[s] = single_scan;
	}

	message.receiver_id = receiver_id;
	message.microseconds = sample_duration_us;
	message.base_frequency = base_frequency;
	message.has_frequency_step = true;
	message.frequency_step = frequency_step;
	message.start_time = start_time;
	message.n_scans = n_single_scans;
	message.scans = proto_single_scans;

	ret = api_send_scan_container(core, &message);

	for (s = 0; s < n_single_scans; s++) {
		free(proto_single_scans[s]);
	}
	free(proto_single_scans);

	return ret;
}

Proto__Downlink__Container *api_recv(struct core *core)
{
	zmq_msg_t zmq_msg;

	if (zmq_msg_init(&zmq_msg) < 0) {
		zerror("Failed to initialize zmq message");
		return NULL;
	}

	if (zmq_socket_recv(core, &zmq_msg, core->api->zmq_downlink) < 0) {
		zmq_msg_close(&zmq_msg);
		return NULL;
	}

	dump(zmq_msg_data(&zmq_msg), zmq_msg_size(&zmq_msg), "api.recv");

	Proto__Downlink__Container *container = NULL;

	if ((container = proto__downlink__container__unpack(NULL, zmq_msg_size(&zmq_msg),
							    zmq_msg_data(&zmq_msg))) == NULL) {
		error("Failed to de-serialize downlink message");
		zmq_msg_close(&zmq_msg);
		return NULL;
	}

	zmq_msg_close(&zmq_msg);
	return container;
}
