// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "clock.h"

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "common/statistics.h"

#define PPS_SYS_ROOT "/sys/class/pps/pps%u"
#define PPS_SYS_NAME PPS_SYS_ROOT "/name"
#define PPS_SYS_ASSERT PPS_SYS_ROOT "/assert"

// Check if PPS device name start with "pps" to filter out "ptp" devices
static int pps_check_device(int number)
{
	char *path = NULL;
	char name[32];
	int fd;

	if (aprintf(&path, PPS_SYS_NAME, number) < 0) {
		error("Failed to format PPS name");
		return -1;
	}

	if (access(path, F_OK) < 0) {
		free(path);
		return -1;
	}

	if ((fd = open(path, O_RDONLY)) < 0) {
		eerror("Failed to open PPS name");
		free(path);
		return -1;
	}

	free(path);

	if (read(fd, name, sizeof(name)) < 0) {
		eerror("Failed to read PPS name");
		close(fd);
		return -1;
	}

	close(fd);

	if (strncmp(name, "pps", 3) != 0) {
		return 0;
	}

	return 1;
}

// Open the PPS assert file
static int pps_open_device(int number)
{
	char *path = NULL;
	int fd;

	if (aprintf(&path, PPS_SYS_ASSERT, number) < 0) {
		error("Failed to format PPS assert");
		return -1;
	}

	if (access(path, F_OK) < 0) {
		free(path);
		return -1;
	}

	if ((fd = open(path, O_RDONLY)) < 0) {
		eerror("Failed to open PPS assert");
		free(path);
		return -1;
	}

	free(path);

	return fd;
}

static int pps_open(void)
{
	int number;
	int fd;

	for (number = 0;; number++) {
		switch (pps_check_device(number)) {
		case 1:
			info("Using PPS device " PPS_SYS_ROOT, number);
			break;

		case 0:
			continue;

		default:
			info("No PPS device detected");
			return -1;
		}

		if ((fd = pps_open_device(number)) < 0) {
			return -1;
		}

		return fd;
	}
}

static int pps_close(int fd)
{
	if (fd >= 0 && close(fd) < 0) {
		ewarning("Failed to close PPS");
		return -1;
	}

	return 0;
}

static int pps_read(int fd, struct timespec *pps_time)
{
	char assert[64];
	ssize_t length;

	if (lseek(fd, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition PPS offset");
		return -1;
	}

	if ((length = read(fd, assert, sizeof(assert) - 1)) < 0) {
		eerror("Failed to read PPS assert");
		return -1;
	}

	assert[length] = '\0';

	if (sscanf(assert, "%li.%9li#%*i", &pps_time->tv_sec, &pps_time->tv_nsec) != 2) {
		eerror("Invalid PPS assert");
		return -1;
	}

	return 0;
}

static int check_pps(struct core *core)
{
	static uint32_t saved_delay = 0;
	static uint32_t saved_pps = 0;
	static uint32_t pps_sync = 0;
	uint32_t pps, delay;
	double pps_double;

	if (core->clock->pps_fd < 0) {
		return -1;
	}

	if (pps_read(core->clock->pps_fd, &core->clock->pps_time) < 0) {
		error("Failed to get PPS value");
		return -1;
	}

	pps_double = core->clock->pps_time.tv_sec + core->clock->pps_time.tv_nsec * 1E-9;
	pps = pps_double + 0.5;

	if (pps_sync == 0) {
		if (saved_pps == 0 || pps == 0) {
			saved_pps = pps;
			return 0;
		}

		pps_sync = 1;
	}

	delay = pps - saved_pps;
	saved_pps = pps;

	core->statistics->has_gps_lost_pulse = 1;

	if (delay == 0) {
		core->statistics->gps_lost_pulse += 1;
		saved_delay = 0;
		return 0;
	}

	if (saved_delay == 0 && delay > 1) {
		warning("%u pps pulse%s have been lost",
			delay - 1, (delay > 2) ? "s" : "");
	}

	saved_delay = delay;

	return 1;
}

static void *valid_thread(void *arg)
{
	struct timespec system, tai;
	struct core *core = arg;
	time_t tai_offset;
	int pps_enabled;

	while (core->clock->running) {
		// Extract TAI offset from system clocks
		if (clock_gettime(CLOCK_REALTIME, &system) == 0 &&
		    clock_gettime(CLOCK_TAI, &tai) == 0) {
			tai_offset = time_diff_ms(&system, &tai) / 1000;

			if (core->clock->tai_offset != tai_offset) {
				core->clock->tai_offset = tai_offset;
				info("TAI offset has been set to %lu", tai_offset);
			}
		}

		// Check if PPS is enabled on the demodulator
		if ((pps_enabled = check_pps(core)) >= 0) {
			if (core->clock->pps_enabled != pps_enabled) {
				core->clock->pps_enabled = pps_enabled;
				info("PPS is %sabled", (pps_enabled) ? "en" : "dis");
				api_send_event_system(core, (pps_enabled)
								? PROTO__EVENT__SYSTEM__EVENT__GPS_LOCKED
								: PROTO__EVENT__SYSTEM__EVENT__GPS_UNLOCKED);
			}
		}

		sleep_sec(1, core->clock->running);
	}

	pthread_exit(NULL);
}

struct clock *clock_init(void)
{
	struct clock *clock = NULL;

	if ((clock = zalloc(sizeof(*clock))) == NULL) {
		eerror("Failed to allocate clock");
		return NULL;
	}

	if ((clock->core = clock_core_init()) == NULL) {
		free(clock);
		return NULL;
	}

	clock->pps_fd = pps_open();

	return clock;
}

int clock_start(struct core *core)
{
	core->clock->running = 1;

	if (pthread_create(&core->clock->valid_thread, NULL, valid_thread, core) != 0) {
		error("Failed to start clock validation thread");
		clock_core_shutdown(core);
		return -1;
	}

	if (clock_core_start(core) < 0) {
		clock_core_shutdown(core);
		return -1;
	}

	return 0;
}

int clock_join(struct core *core)
{
	if (clock_core_join(core) < 0) {
		return -1;
	}

	if (pthread_join(core->clock->valid_thread, NULL) != 0) {
		error("Failed to join clock validation thread");
		return -1;
	}

	return 0;
}

void clock_shutdown(struct core *core)
{
	if (core) {
		clock_core_shutdown(core);

		if (core->clock) {
			core->clock->running = 0;
		}
	}
}

void clock_free(struct core *core)
{
	if (core) {
		clock_core_free(core);
		pps_close(core->clock->pps_fd);
		free(core->clock);
	}
}

// The offset between the GPS epoch and the UNIX epoch.
// This is also the UNIX time of 1980-01-06T00:00:00.
#define GPS_UNIX_EPOCH 315964800

// The offset between the TAI time and the GPS time.
// TAI time is always ahead of GPS time by 19 seconds.
#define TAI_GPS_OFFSET 19

void clock_convert_tai_to_gps(struct timespec *tai, struct timespec *gps)
{
	gps->tv_sec = tai->tv_sec - GPS_UNIX_EPOCH - TAI_GPS_OFFSET;
	gps->tv_nsec = tai->tv_nsec;
}

void clock_convert_gps_to_tai(struct timespec *gps, struct timespec *tai)
{
	tai->tv_sec = gps->tv_sec + GPS_UNIX_EPOCH + TAI_GPS_OFFSET;
	tai->tv_nsec = gps->tv_nsec;
}

int clock_convert_tai_to_utc(time_t tai_offset, struct timespec *tai, struct timespec *utc)
{
	if (tai_offset == 0) {
		return -1;
	}

	utc->tv_sec = tai->tv_sec - tai_offset;
	utc->tv_nsec = tai->tv_nsec;

	return 0;
}

int clock_get_monotonic_time(struct timespec *tp)
{
	if (clock_gettime(CLOCK_MONOTONIC, tp) < 0) {
		eerror("Failed to get monotonic clock");
		return -1;
	}

	return 0;
}

int clock_get_utc_time(struct timespec *tp)
{
	if (clock_gettime(CLOCK_REALTIME, tp) < 0) {
		eerror("Failed to get system clock");
		return -1;
	}

	return 0;
}

int clock_get_tai_time(time_t tai_offset, struct timespec *tp)
{
	if (tai_offset == 0) {
		return -1;
	}

	if (clock_gettime(CLOCK_TAI, tp) < 0) {
		eerror("Failed to get atomic clock");
		return -1;
	}

	return 0;
}

int clock_get_gps_time(time_t tai_offset, struct timespec *tp)
{
	struct timespec tai;

	if (clock_get_tai_time(tai_offset, &tai) < 0) {
		return -1;
	}

	clock_convert_tai_to_gps(&tai, tp);

	return 0;
}
