// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <stdlib.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

#ifndef RUNSTATEDIR
#define RUNSTATEDIR "/var/run/lora"
#endif

struct config *config_init(void)
{
	struct config *config = NULL;

	if ((config = zalloc(sizeof(*config))) == NULL) {
		eerror("Failed to allocate configuration");
		return NULL;
	}

	if ((config->core = config_core_init()) == NULL) {
		error("Failed to initialize core configuration");
		return NULL;
	}

	config->beacon.frequency = 869525000;
	config->beacon.frequencies = 1;
	config->beacon.bandwidth = 125000;
	config->beacon.spreading_factor = 9;
	config->beacon.power = 14;

	return config;
}

static int parse_beacon(struct core *core, JSON_Object *root)
{
	struct config_beacon *beacon = &core->config->beacon;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n, l;

	JSON_BOOLEAN(root, "beacon_enable", beacon->enable);
	JSON_NUMBER(root, "beacon_period", beacon->period);
	JSON_NUMBER(root, "beacon_freq_hz", beacon->frequency);
	JSON_NUMBER(root, "beacon_freq_nb", beacon->frequencies);
	JSON_NUMBER(root, "beacon_freq_step", beacon->frequency_step);
	JSON_NUMBER(root, "beacon_power", beacon->power);

	if (JSON_NUMBER(root, "beacon_bw_hz", beacon->bandwidth) &&
	    beacon->bandwidth != 125000 &&
	    beacon->bandwidth != 500000) {
		error("Invalid beacon bandwidth: %u", beacon->bandwidth);
		return -1;
	}

	if (JSON_NUMBER(root, "beacon_datarate", beacon->spreading_factor) &&
	    beacon->spreading_factor != 8 &&
	    beacon->spreading_factor != 9 &&
	    beacon->spreading_factor != 10 &&
	    beacon->spreading_factor != 12) {
		error("Invalid beacon spreading factor: %u", beacon->spreading_factor);
		return -1;
	}

	if ((array = json_object_get_array(root, "beacon_infodesc")) != NULL) {
		debug("Parsing array: \"gateway_conf.beacon_infodesc\"");

		for (n = 0; n < json_array_get_count(array); n++) {
			if ((object = json_array_get_object(array, n)) != NULL) {
				debug("Parsing object: \"gateway_conf.beacon_infodesc[%zu]\"", n);

				if (beacon->nb_coordinates < sizeofa(beacon->coordinates)) {
					l = JSON_NUMBER(object, "latitude", beacon->coordinates[beacon->nb_coordinates].latitude);
					l += JSON_NUMBER(object, "longitude", beacon->coordinates[beacon->nb_coordinates].longitude);

					if (l > 0) {
						if (l == 2) {
							if (beacon->coordinates[beacon->nb_coordinates].latitude > 90.0 ||
							    beacon->coordinates[beacon->nb_coordinates].latitude < -90.0) {
								error("Invalid latitude: %.10g", beacon->coordinates[beacon->nb_coordinates].latitude);
								return -1;
							}

							if (beacon->coordinates[beacon->nb_coordinates].longitude > 180.0 ||
							    beacon->coordinates[beacon->nb_coordinates].longitude < -180.0) {
								error("Invalid longitude: %.10g", beacon->coordinates[beacon->nb_coordinates].longitude);
								return -1;
							}

							beacon->nb_coordinates++;
						}

						continue;
					}
				}

				if (!beacon->has_netid) {
					beacon->has_netid = JSON_HEX(object, "netid", beacon->netid);

					if (beacon->has_netid) {
						JSON_HEX(object, "gwid", beacon->gwid);
						continue;
					}
				}
			}
		}
	}

	if (beacon->period > 0 && beacon->period < 6) {
		error("Invalid configuration for beacon period, must be >= 6s");
		return -1;
	}

	return 0;
}

static int parse_gateway(struct core *core, JSON_Object *root)
{
	JSON_Object *object = NULL;

	if ((object = json_object_get_object(root, "gateway_conf")) != NULL) {
		debug("Parsing object: \"gateway_conf\"");

		if (parse_beacon(core, object) < 0) {
			error("Failed to parse beacon");
			return -1;
		}
	}

	return 0;
}

static int parse_api(struct core *core, JSON_Object *root)
{
	const char *string = NULL;
	JSON_Array *array = NULL;
	size_t nb;

	if ((array = json_object_dotget_array(root, "api.uplink")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;
			}

			if (stringlist_stradd(&core->config->api.uplinks, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;
			}

			debug("Value \"api.uplink\" configured to: \"%s\"", string);
		}
	}

	if ((array = json_object_dotget_array(root, "api.downlink")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;
			}

			if (stringlist_stradd(&core->config->api.downlinks, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;
			}

			debug("Value \"api.downlink\" configured to: \"%s\"", string);
		}
	}

	if ((array = json_object_dotget_array(root, "api.control")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;
			}

			if (stringlist_stradd(&core->config->api.controls, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;
			}

			debug("Value \"api.control\" configured to: \"%s\"", string);
		}
	}

	return 0;
}

int config_parse_file(struct core *core, const char *filename)
{
	JSON_Value *root = NULL;

	if (access(filename, R_OK) < 0) {
		eerror("Cannot read file");
		return -1;
	}

	if ((root = json_parse_file_with_comments(filename)) == NULL) {
		error("Invalid JSON file");
		return -1;
	}

	if (config_core_parse(core, json_value_get_object(root)) < 0) {
		error("Failed to parse core configuration");
		json_value_free(root);
		return -1;
	}

	if (parse_gateway(core, json_value_get_object(root)) < 0) {
		error("Failed to parse gateway configuration");
		json_value_free(root);
		return -1;
	}

	if (parse_api(core, json_value_get_object(root)) < 0) {
		error("Failed to parse API configuration");
		json_value_free(root);
		return -1;
	}

	json_value_free(root);

	return 0;
}

int config_apply(struct core *core)
{
	size_t n;

	if (config_core_apply(core) < 0) {
		error("Failed to apply core configuration");
		return -1;
	}

	if (core->config->beacon.enable &&
	    core->config->beacon.period > 0 &&
	    core->config->beacon.frequencies > 0) {
		if (core->config->beacon.frequencies > 1) {
			info("Using beacon: enable=true period=%u freq_hz=%u:%u:%zu bw=%u sf=%u power=%i",
			     core->config->beacon.period,
			     core->config->beacon.frequency,
			     core->config->beacon.frequency_step,
			     core->config->beacon.frequencies,
			     core->config->beacon.bandwidth,
			     core->config->beacon.spreading_factor,
			     core->config->beacon.power);
		} else {
			info("Using beacon: enable=true period=%u freq_hz=%u bw=%u sf=%u power=%i",
			     core->config->beacon.period,
			     core->config->beacon.frequency,
			     core->config->beacon.bandwidth,
			     core->config->beacon.spreading_factor,
			     core->config->beacon.power);
		}

		for (n = 0; n < core->config->beacon.nb_coordinates; n++) {
			info("Using beacon: infodesc=%zu latitude=%.10g longitude=%.10g",
			     n,
			     core->config->beacon.coordinates[n].latitude,
			     core->config->beacon.coordinates[n].longitude);
		}

		if (core->config->beacon.has_netid) {
			info("Using beacon: infodesc=3 netid=%.6X gwid=%.6X",
			     core->config->beacon.netid,
			     core->config->beacon.gwid);
		}
	} else {
		info("Using beacon: enable=false");
	}

	if (core->config->api.uplinks.nb_strings == 0 &&
	    stringlist_stradd(&core->config->api.uplinks,
			      "ipc://" RUNSTATEDIR "/uplink") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;
	}

	if (core->config->api.downlinks.nb_strings == 0 &&
	    stringlist_stradd(&core->config->api.downlinks,
			      "ipc://" RUNSTATEDIR "/downlink") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;
	}

	if (core->config->api.controls.nb_strings == 0 &&
	    stringlist_stradd(&core->config->api.controls,
			      "ipc://" RUNSTATEDIR "/lorad") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;
	}

	for (n = 0; n < core->config->api.uplinks.nb_strings; n++) {
		info("Using api: uplink=\"%s\"", core->config->api.uplinks.strings[n]);
	}

	for (n = 0; n < core->config->api.downlinks.nb_strings; n++) {
		info("Using api: downlink=\"%s\"", core->config->api.downlinks.strings[n]);
	}

	for (n = 0; n < core->config->api.controls.nb_strings; n++) {
		info("Using api: control=\"%s\"", core->config->api.controls.strings[n]);
	}

	return 0;
}

void config_free(struct core *core)
{
	if (core && core->config) {
		config_core_free(core);
		stringlist_free(&core->config->api.uplinks);
		stringlist_free(&core->config->api.downlinks);
		stringlist_free(&core->config->api.controls);
		free(core->config);
	}
}
