// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "core_fscan.h"

#include <stdlib.h>
#include <time.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "clock.h"

/* TODO: MGR: function to factorize */
static Proto__Time__Time *get_start_time(struct core *core)
{
	Proto__Time__Time *start_time;
	struct timespec tp;

	start_time = malloc(sizeof(Proto__Time__Time));
	proto__time__time__init(start_time);

	if (clock_get_tai_time(core->clock->tai_offset, &tp) == 0) {
		start_time->format = PROTO__TIME__TIME__TIME_FORMAT__TAI_TIME;
		start_time->time = tp.tv_sec;
		start_time->has_microseconds = 1;
		start_time->microseconds = tp.tv_nsec / 1000;

	} else if (clock_get_utc_time(&tp) == 0) {
		start_time->format = PROTO__TIME__TIME__TIME_FORMAT__UNIX_TIME;
		start_time->time = tp.tv_sec;
		start_time->has_microseconds = 1;
		start_time->microseconds = tp.tv_nsec / 1000;

	} else {
		free(start_time);
		start_time = NULL;
	}

	return start_time;
}

static void *scan_thread(void *arg)
{
	unsigned int s, f;
	struct sniffer *sniffer = arg;
	uint32_t base_frequency, frequency_step;
	struct timespec t_sample_start, t_sample_end, t_freq_start, t_freq_end;
	uint64_t sample_duration_us;
	size_t n_single_scans;
	struct single_scan *single_scans;
	Proto__Time__Time *start_time = NULL;

	base_frequency = sniffer->parameters.frequencies[0];
	frequency_step = (sniffer->parameters.n_frequencies > 1) ? (sniffer->parameters.frequencies[1] - sniffer->parameters.frequencies[0]) : 0;
	n_single_scans = sniffer->parameters.n_frequencies;
	single_scans = (struct single_scan *)zalloc(sniffer->parameters.n_frequencies * sizeof(struct single_scan));

	api_send_event_system(sniffer->core, PROTO__EVENT__SYSTEM__EVENT__SCAN_STARTED);

	for (s = 0; ((s < sniffer->parameters.samples) || sniffer->parameters.infinite_loop) && sniffer->running;
	     s = sniffer->parameters.infinite_loop ? s : s + 1) {
		start_time = get_start_time(sniffer->core);
		clock_gettime(CLOCK_MONOTONIC_RAW, &t_sample_start);

		for (f = 0; f < sniffer->parameters.n_frequencies && sniffer->running; f++) {
			clock_gettime(CLOCK_MONOTONIC_RAW, &t_freq_start);
			if (fscan_do_scan(&single_scans[f].rssi, sniffer->core, sniffer->parameters.receiver_id, sniffer->parameters.frequencies[f]) < 0) {
				error("Unable to retrieve RSSI");
				n_single_scans = 0;
				break;
			}
			clock_gettime(CLOCK_MONOTONIC_RAW, &t_freq_end);
			single_scans[f].duration = (t_freq_end.tv_sec - t_freq_start.tv_sec) * 1000000 +
						   (t_freq_end.tv_nsec - t_freq_start.tv_nsec) / 1000;
			single_scans[f].frequency_offset =
			    (f > 1) ? (sniffer->parameters.frequencies[f] - sniffer->parameters.frequencies[f - 1] - frequency_step) : 0;
		}
		clock_gettime(CLOCK_MONOTONIC_RAW, &t_sample_end);
		sample_duration_us = (t_sample_end.tv_sec - t_sample_start.tv_sec) * 1000000 +
				     (t_sample_end.tv_nsec - t_sample_start.tv_nsec) / 1000;

		/* Send scan result now */
		api_send_scan(sniffer->core, sniffer->parameters.receiver_id, sample_duration_us, base_frequency, frequency_step,
			      start_time, n_single_scans, single_scans);

		free(start_time);
		start_time = NULL;

		/* Wait before next scan (otherwise scaning takes a significative CPU time) */
		sleep_ms(200);
	}

	if (start_time) {
		free(start_time);
		start_time = NULL;
	}
	free(single_scans);
	free(sniffer->parameters.frequencies);
	sniffer->running = 0;

	api_send_event_system(sniffer->core, PROTO__EVENT__SYSTEM__EVENT__SCAN_STOPPED);

	pthread_exit(NULL);
}

struct fscan *fscan_init(void)
{
	struct fscan *fscan = NULL;

	if ((fscan = zalloc(sizeof(*fscan))) == NULL) {
		eerror("Failed to allocate fscan");
		return NULL;

	} else if ((fscan->core = fscan_core_init()) == NULL) {
		return NULL;
	}

	return fscan;
}

int fscan_start(struct core *core)
{
	unsigned int s;

	fscan_get_capabilities(core, &core->fscan->sniffer_caps);

	core->fscan->sniffers = zalloc(core->fscan->sniffer_caps.sniffer_nb * sizeof(struct sniffer));

	for (s = 0; s < core->fscan->sniffer_caps.sniffer_nb; s++) {
		if (fscan_core_start(s) < 0) {
			core->fscan->sniffers[s].running = true;
		}
	}

	return 0;
}

int fscan_join(struct core *core)
{
	unsigned int s;

	/* TODO: MGR: sometimes, join is locked if thread has terminated by itself... */

	for (s = 0; s < core->fscan->sniffer_caps.sniffer_nb; s++) {
		if (core->fscan->sniffers[s].thread && pthread_join(core->fscan->sniffers[s].thread, NULL) != 0) {
			error("Failed to join fscan validation thread");
		}
		core->fscan->sniffers[s].thread = 0;
	}

	free(core->fscan->sniffers);

	return 0;
}

void fscan_shutdown(struct core *core)
{
	unsigned int s;

	if (core) {
		for (s = 0; s < core->fscan->sniffer_caps.sniffer_nb; s++) {
			if (core->fscan->sniffers) {
				core->fscan->sniffers[s].running = 0;
			}
			fscan_core_shutdown(s);
		}
	}
}

void fscan_free(struct core *core)
{
	if (core) {
		fscan_core_free(core);
		free(core->fscan);
	}
}

static bool is_sniffer_busy(struct core *core, uint32_t receiver_id)
{
	if (fscan_is_lbt_enabled(receiver_id) || core->fscan->sniffers[receiver_id].running) {
		return true;
	} else {
		return false;
	}
}

int fscan_get_scan_status(struct core *core, size_t *n_receivers, struct fscan_receivers **fscan_receivers)
{
	unsigned int i = 0;
	struct fscan_receivers *pfscan_receivers;

	/* TODO: MGR: maybe need a mutex to protect sniffer struct */

	*n_receivers = core->fscan->sniffer_caps.sniffer_nb;

	pfscan_receivers = (struct fscan_receivers *)malloc(*n_receivers * sizeof(struct fscan_receivers));

	for (i = 0; i < *n_receivers; i++) {
		pfscan_receivers[i].id = i;
		pfscan_receivers[i].busy = is_sniffer_busy(core, i);
	}

	*fscan_receivers = pfscan_receivers;

	return 0;
}

int fscan_start_scan(struct core *core, uint32_t receiver_id, size_t n_frequencies,
		     uint32_t *frequencies, bool has_samples, uint32_t samples)
{
	unsigned int i, j;
	uint32_t *no_dup_frequencies = NULL;
	size_t n_no_dup_frequencies = 0;
	uint32_t temp_freq;

	if (receiver_id >= core->fscan->sniffer_caps.sniffer_nb) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}

	if (is_sniffer_busy(core, receiver_id)) {
		warning("Sniffer %u busy, nothing to do", receiver_id);
		return 0;
	}

	/* Delete duplicates, consider "0" as an invalid value */
	no_dup_frequencies = (uint32_t *)zalloc(n_frequencies * sizeof(uint32_t));
	for (i = 0; i < n_frequencies; i++) {
		for (j = 0; j < n_no_dup_frequencies; j++) {
			if (frequencies[i] == no_dup_frequencies[j]) {
				break;
			}
		}
		if ((j == n_no_dup_frequencies) && (frequencies[i] != 0)) {
			no_dup_frequencies[j] = frequencies[i];
			n_no_dup_frequencies++;
		}
	}

	/* Realloc to remove all "0" invalid values */
	if (n_no_dup_frequencies == 0) {
		warning("No valid frequencies found");
		free(no_dup_frequencies);
		return -1;
	}
	no_dup_frequencies = (uint32_t *)realloc(no_dup_frequencies, n_no_dup_frequencies * sizeof(uint32_t));

	/* Bubble sort values in ascending order */
	for (i = 0; i < n_no_dup_frequencies; i++) {
		for (j = 0; j < (n_no_dup_frequencies - i - 1); j++) {
			if (no_dup_frequencies[j] > no_dup_frequencies[j + 1]) {
				temp_freq = no_dup_frequencies[j];
				no_dup_frequencies[j] = no_dup_frequencies[j + 1];
				no_dup_frequencies[j + 1] = temp_freq;
			}
		}
	}

	core->fscan->sniffers[receiver_id].core = core;
	core->fscan->sniffers[receiver_id].running = 1;
	core->fscan->sniffers[receiver_id].parameters.receiver_id = receiver_id;
	core->fscan->sniffers[receiver_id].parameters.n_frequencies = n_no_dup_frequencies;
	core->fscan->sniffers[receiver_id].parameters.frequencies = no_dup_frequencies;
	core->fscan->sniffers[receiver_id].parameters.infinite_loop = has_samples ? false : true;
	core->fscan->sniffers[receiver_id].parameters.samples = samples;

	if (pthread_create(&core->fscan->sniffers[receiver_id].thread, NULL, scan_thread, &core->fscan->sniffers[receiver_id]) != 0) {
		error("Failed to start fscan validation thread");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	return 0;
}

int fscan_stop_scan(struct core *core, uint32_t receiver_id)
{
	if (receiver_id >= core->fscan->sniffer_caps.sniffer_nb) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}

	if (!core->fscan->sniffers[receiver_id].running) {
		warning("sniffer %u is not running, nothing to do", receiver_id);
		return 0;
	}

	core->fscan->sniffers[receiver_id].running = 0;

	if (core->fscan->sniffers[receiver_id].thread && pthread_join(core->fscan->sniffers[receiver_id].thread, NULL) != 0) {
		error("Failed to join fscan validation thread");
		return -1;
	}
	core->fscan->sniffers[receiver_id].thread = 0;

	return 0;
}
