// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include <getopt.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "utils/daemon.h"
#include "utils/logger.h"
#include "utils/stringlist.h"

#include "common/core.h"
#include "common/temperature.h"

#ifndef RUNSTATEDIR
#define RUNSTATEDIR "/var/run/lorad"
#endif

#define PIDFILE RUNSTATEDIR "/lorad.pid"

#define COOLDOWN_SEC (10 * 60)

static struct core *core = NULL;

static void print_help(FILE *stream)
{
	fprintf(stream,
		"The LoRa daemon\n"
		"Usage:\n"
		"  lorad [OPTION...] FILE...\n\n"
		"  -g, --gps GPS        Use the given GPS device.\n"
		"  -s, --serial SERIAL  Use the given serial number.\n"
		"  -f, --foreground     Do not run in background.\n"
		"  -v, --verbose        Turn on verbose output.\n"
		"                       This option can be used multiple times.\n"
		"      --version        Print version string and exit.\n"
		"  -h, --help           Print this help message and exit.\n");
}

static void signal_handler(int signal)
{
	switch (signal) {
	case SIGINT:
	case SIGQUIT:
	case SIGTERM:
		notice("Stopping lorad");
		core_shutdown(core);
		break;
	}
}

int main(int argc, char **argv)
{
	int status = EXIT_FAILURE;
	int verbosity = LOG_WARNING;
	int pid = -1, pidfile = -1;
	int foreground = 0;
	int opt, opti;

	struct stringlist filenames = STRINGLIST_INIT;
	char *gps_device = NULL;
	uint32_t serial_number = 0;

	const struct option opts[] = {
	    {"gps", required_argument, NULL, 'g'},
	    {"serial", required_argument, NULL, 's'},
	    {"foreground", no_argument, NULL, 'f'},
	    {"verbose", no_argument, NULL, 'v'},
	    {"version", no_argument, NULL, 0},
	    {"help", no_argument, NULL, 'h'},
	    {NULL, 0, NULL, 0},
	};

	while ((opt = getopt_long(argc, argv, "g:s:fhv", opts, &opti)) != EOF) {
		switch (opt) {
		case 0:
			if (strcmp("version", opts[opti].name) == 0) {
				fprintf(stdout, "lorad " VERSION "\n");
				return EXIT_SUCCESS;
			}
			break;

		case 'g':
			gps_device = optarg;
			break;

		case 's':
			sscanf(optarg, "%x", &serial_number);
			break;

		case 'f':
			foreground = 1;
			break;

		case 'h':
			print_help(stdout);
			return EXIT_SUCCESS;

		case 'v':
			verbosity++;
			break;

		default:
			print_help(stderr);
			return EXIT_FAILURE;
		}
	}

	if (optind >= argc) {
		print_help(stderr);
		return EXIT_FAILURE;
	}

	while (optind < argc) {
		if (stringlist_stradd(&filenames, argv[optind++]) == NULL) {
			perror("Failed to allocate filename");
			stringlist_free(&filenames);
			return EXIT_FAILURE;
		}
	}

	// From this point, clean_exit have to be used.
	openlog("lorad", LOG_PID | LOG_PERROR, LOG_LOCAL1);

	setlogpriority(verbosity);

	// Run the process in background as soon as possible.
	const struct signal signals[] = {
	    {SIGINT, signal_handler},
	    {SIGQUIT, signal_handler},
	    {SIGTERM, signal_handler},
	    {0, NULL},
	};

	if (foreground == 0 && (pid = daemonize()) < 0) {
		critical("Failed to daemonize");
		goto clean_exit;

	} else if (foreground == 0 && pid > 0) {
		status = EXIT_SUCCESS;
		goto clean_exit;

	} else if ((pidfile = create_pidfile(PIDFILE)) < 0) {
		critical("Failed to create PID file");
		goto clean_exit;

	} else if (handle_signals(signals) < 0) {
		critical("Failed to handle signals");
		goto clean_exit;
	}

	notice("Starting lorad " VERSION);

	if ((core = core_init(&filenames, gps_device, serial_number)) == NULL) {
		critical("Failed to initialize core");
		goto clean_exit;

	} else if (core_start(core) < 0) {
		critical("Failed to start core");
		goto clean_exit;

	} else if (core_join(core) < 0) {
		critical("Failed to join core");
		goto clean_exit;
	}

	if (core->temperature && core->temperature->cooldown) {
		info("Starting maximum temperature cooldown");
		sleep(COOLDOWN_SEC);
	}

	status = EXIT_SUCCESS;

clean_exit:
	core_free(core);

	if (pidfile >= 0) {
		close(pidfile);

		if (unlink(PIDFILE) < 0) {
			ewarning("Failed to remove PID file: %s", PIDFILE);
		}
	}

	closelog();
	stringlist_free(&filenames);

	exit(status);
}
