// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "api.h"

#include <stdlib.h>
#include <string.h>

#include "utils/logger.h"

#include "clock.h"
#include "common/statistics.h"

static int convert_counter_to_time(struct core *core, uint32_t counter, Proto__Time__Time *time)
{
	struct timespec tp;

	if (clock_get_tai_time_from_counter(core, counter, &tp) == 0) {
		time->format = PROTO__TIME__TIME__TIME_FORMAT__TAI_TIME;
		time->time = tp.tv_sec;
		time->has_microseconds = 1;
		time->microseconds = tp.tv_nsec / 1000;
		return 0;

	} else if (clock_get_utc_time_from_counter(core, counter, &tp) == 0) {
		time->format = PROTO__TIME__TIME__TIME_FORMAT__UNIX_TIME;
		time->time = tp.tv_sec;
		time->has_microseconds = 1;
		time->microseconds = tp.tv_nsec / 1000;
		return 0;

	} else if (clock_get_utc_time(&tp) == 0) {
		time->format = PROTO__TIME__TIME__TIME_FORMAT__UNIX_TIME;
		time->time = tp.tv_sec;
		time->has_microseconds = 0;
		return 0;
	}

	return -1;
}

static int convert_time_to_counter(struct core *core, Proto__Time__AccurateContinuousTime *time, uint32_t *counter)
{
	struct timespec tp;

	tp.tv_sec = time->time;
	tp.tv_nsec = time->microseconds * 1000;

	if (!core->clock->pps_enabled) {
		return -PROTO__EVENT__TRANSMISSION__STATUS__GPS_UNLOCKED;
	}

	switch (time->format) {
	case PROTO__TIME__ACCURATE_CONTINUOUS_TIME__CONTINUOUS_TIME_FORMAT__TAI_TIME:
		if (clock_get_counter_from_tai_time(core, &tp, counter) < 0) {
			return -PROTO__EVENT__TRANSMISSION__STATUS__TRANSMISSION_FAILED;
		}
		break;

	case PROTO__TIME__ACCURATE_CONTINUOUS_TIME__CONTINUOUS_TIME_FORMAT__GPS_TIME:
		if (clock_get_counter_from_gps_time(core, &tp, counter) < 0) {
			return -PROTO__EVENT__TRANSMISSION__STATUS__TRANSMISSION_FAILED;
		}
		break;

	default:
		error("Unsupported time format");
		return -PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT;
	}

	return 0;
}

#define ALLOCATE(type, name, free)                   \
	type *name = NULL;                           \
	if ((name = malloc(sizeof(type))) == NULL) { \
		eerror("Failed to allocate " #name); \
		FREE(free);                          \
		return -1;                           \
	}

#define ALLOCATE_ARRAY(type, name, base, free)                                                   \
	if ((base->name = realloc(base->name, (base->n_##name + 1) * sizeof(type *))) == NULL) { \
		eerror("Failed to re-allocate " #name);                                          \
		FREE(free);                                                                      \
		return -1;                                                                       \
	}

#define ALLOCATE_BINARY(binary, size, free)          \
	binary.len = size;                           \
	if ((binary.data = malloc(size)) == NULL) {  \
		eerror("Failed to allocate memory"); \
		FREE(free);                          \
		return -1;                           \
	}

#undef FREE
#define FREE api_free_lora

static int core_send_lora(struct core *core, struct lgw_pkt_rx_s *rxpkt)
{
	core->api->message_id += 1;

	ALLOCATE(Proto__Lora__Uplink, message, message);
	proto__lora__uplink__init(message);

	message->message_id = core->api->message_id;

	ALLOCATE(Proto__Lora__Uplink__Data, data, message);
	proto__lora__uplink__data__init(data);
	message->data = data;

	ALLOCATE_BINARY(data->payload, rxpkt->size, message);
	memcpy(data->payload.data, rxpkt->payload, rxpkt->size);

	data->has_frequency = 1;
	data->frequency = rxpkt->freq_hz;

	data->has_crc = (rxpkt->status == STAT_UNDEFINED) ? 0 : 1;
	switch (rxpkt->status) {
	case STAT_CRC_OK:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_OK;
		break;
	case STAT_CRC_BAD:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_ERROR;
		break;
	case STAT_NO_CRC:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_NONE;
		break;
	default:
		data->has_crc = 0;
		break;
	}

	data->has_bandwidth = (rxpkt->bandwidth == BW_UNDEFINED) ? 0 : 1;
	switch (rxpkt->bandwidth) {
	case BW_125KHZ:
		data->bandwidth = PROTO__LORA__BANDWIDTH__BW_125K;
		break;
	case BW_250KHZ:
		data->bandwidth = PROTO__LORA__BANDWIDTH__BW_250K;
		break;
	case BW_500KHZ:
		data->bandwidth = PROTO__LORA__BANDWIDTH__BW_500K;
		break;
	default:
		data->has_bandwidth = 0;
		break;
	}

	data->has_spreading_factor =
	    (rxpkt->datarate == DR_UNDEFINED) ? 0 : 1;
	switch (rxpkt->datarate) {
	case DR_LORA_SF7:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_7;
		break;
	case DR_LORA_SF8:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_8;
		break;
	case DR_LORA_SF9:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_9;
		break;
	case DR_LORA_SF10:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_10;
		break;
	case DR_LORA_SF11:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_11;
		break;
	case DR_LORA_SF12:
		data->spreading_factor = PROTO__LORA__SPREADING_FACTOR__SF_12;
		break;
	default:
		data->has_spreading_factor = 0;
		break;
	}

	data->has_coding_rate = (rxpkt->coderate == CR_UNDEFINED) ? 0 : 1;
	switch (rxpkt->coderate) {
	case CR_LORA_4_5:
		data->coding_rate = PROTO__LORA__CODING_RATE__CR_4_5;
		break;
	case CR_LORA_4_6:
		data->coding_rate = PROTO__LORA__CODING_RATE__CR_4_6;
		break;
	case CR_LORA_4_7:
		data->coding_rate = PROTO__LORA__CODING_RATE__CR_4_7;
		break;
	case CR_LORA_4_8:
		data->coding_rate = PROTO__LORA__CODING_RATE__CR_4_8;
		break;
	default:
		data->has_coding_rate = 0;
		break;
	}

	ALLOCATE(Proto__Lora__Uplink__Radio, radio, message);
	proto__lora__uplink__radio__init(radio);

	ALLOCATE_ARRAY(Proto__Lora__Uplink__Radio, radios, message, message);
	message->radios[message->n_radios] = radio;
	message->n_radios++;

	radio->radio_id = ((rxpkt->if_chain & 0x0F) << 0) | ((rxpkt->rf_chain & 0x0F) << 8);

	ALLOCATE(Proto__Lora__Uplink__Radio__Demodulator, demodulator, message);
	proto__lora__uplink__radio__demodulator__init(demodulator);
	radio->demodulator = demodulator;

	demodulator->counter = rxpkt->count_us;
	demodulator->has_rssi = 1;
	demodulator->rssi = rxpkt->rssi;

	ALLOCATE(Proto__Lora__SNR, snr, message);
	proto__lora__snr__init(snr);
	demodulator->snr = snr;

	snr->snr = rxpkt->snr;
	snr->has_max = 1;
	snr->max = rxpkt->snr_max;
	snr->has_min = 1;
	snr->min = rxpkt->snr_min;

	ALLOCATE(Proto__Time__Time, absolute_time, message);
	proto__time__time__init(absolute_time);
	demodulator->time = absolute_time;

	if (convert_counter_to_time(core, rxpkt->count_us, absolute_time) < 0) {
		api_free_lora(message);
		return -1;

	} else if (api_send_lora(core, message) < 0) {
		api_free_lora(message);
		return -1;
	}

	api_free_lora(message);
	return 0;
}

#undef FREE
#define FREE api_free_gfsk

static int core_send_gfsk(struct core *core, struct lgw_pkt_rx_s *rxpkt)
{
	core->api->message_id += 1;

	ALLOCATE(Proto__Gfsk__Uplink, message, message);
	proto__gfsk__uplink__init(message);

	message->message_id = core->api->message_id;

	ALLOCATE(Proto__Gfsk__Uplink__Data, data, message);
	proto__gfsk__uplink__data__init(data);
	message->data = data;

	ALLOCATE_BINARY(data->payload, rxpkt->size, message);
	memcpy(data->payload.data, rxpkt->payload, rxpkt->size);

	data->has_frequency = 1;
	data->frequency = rxpkt->freq_hz;

	data->has_crc = (rxpkt->status == STAT_UNDEFINED) ? 0 : 1;
	switch (rxpkt->status) {
	case STAT_CRC_OK:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_OK;
		break;
	case STAT_CRC_BAD:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_ERROR;
		break;
	case STAT_NO_CRC:
		data->crc = PROTO__UPLINK__CRCSTATUS__CRC_NONE;
		break;
	default:
		data->has_crc = 0;
		break;
	}

	data->has_bandwidth = (rxpkt->bandwidth == BW_UNDEFINED) ? 0 : 1;
	switch (rxpkt->bandwidth) {
	case BW_7K8HZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_7K8;
		break;
	case BW_15K6HZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_15K6;
		break;
	case BW_31K2HZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_31K2;
		break;
	case BW_62K5HZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_62K5;
		break;
	case BW_125KHZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_125K;
		break;
	case BW_250KHZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_250K;
		break;
	case BW_500KHZ:
		data->bandwidth = PROTO__GFSK__BANDWIDTH__BW_500K;
		break;
	default:
		data->has_bandwidth = 0;
		break;
	}

	data->has_bitrate = (rxpkt->datarate == DR_UNDEFINED) ? 0 : 1;
	data->bitrate = rxpkt->datarate;

	// Radio parts.
	ALLOCATE(Proto__Gfsk__Uplink__Radio, radio, message);
	proto__gfsk__uplink__radio__init(radio);

	ALLOCATE_ARRAY(Proto__Gfsk__Uplink__Radio, radios, message, message);
	message->radios[message->n_radios] = radio;
	message->n_radios++;

	radio->radio_id = ((rxpkt->if_chain & 0x0F) << 0) | ((rxpkt->rf_chain & 0x0F) << 8);

	ALLOCATE(Proto__Gfsk__Uplink__Radio__Demodulator, demodulator, message);
	proto__gfsk__uplink__radio__demodulator__init(demodulator);
	radio->demodulator = demodulator;

	demodulator->counter = rxpkt->count_us;
	demodulator->has_rssi = 1;
	demodulator->rssi = rxpkt->rssi;

	ALLOCATE(Proto__Time__Time, absolute_time, message);
	proto__time__time__init(absolute_time);
	demodulator->time = absolute_time;

	if (convert_counter_to_time(core, rxpkt->count_us, absolute_time) < 0) {
		api_free_gfsk(message);
		return -1;

	} else if (api_send_gfsk(core, message) < 0) {
		api_free_gfsk(message);
		return -1;
	}

	api_free_gfsk(message);
	return 0;
}

#undef FREE
#undef ALLOCATE
#undef ALLOCATE_ARRAY
#undef ALLOCATE_BINARY

static int core_send(struct core *core, struct lgw_pkt_rx_s *rxpkt)
{
	switch (rxpkt->modulation) {
	case MOD_LORA:
		return core_send_lora(core, rxpkt);

	case MOD_FSK:
		return core_send_gfsk(core, rxpkt);

	default:
		error("Unsupported modulation");
		return -1;
	}
}

int api_core_send(struct core *core, struct lgw_pkt_rx_s *rxpkt, uint8_t rxpkt_nb)
{
	uint8_t i;

	for (i = 0; i < rxpkt_nb; i++) {
		core->statistics->uplink_received++;
		core->statistics->uplink_bytes_received += rxpkt[i].size;

		if (rxpkt[i].status == STAT_CRC_OK) {
			core->statistics->uplink_valid++;

		} else if (rxpkt[i].status == STAT_CRC_BAD) {
			core->statistics->uplink_invalid++;
		}

		if (core_send(core, &rxpkt[i]) < 0) {
			return -1;
		}
	}

	return 0;
}

int api_core_send_event(struct core *core, struct lgw_pkt_tx_s *txpkt, enum jit_error_e jit_error)
{
	Proto__Event__Transmission__Status status;

	switch (jit_error) {
	case HAL_NOT_SYNCHRONIZED:
		status = PROTO__EVENT__TRANSMISSION__STATUS__NOT_SYNCHRONIZED;
		break;
	case HAL_SEND_LBT:
		status = PROTO__EVENT__TRANSMISSION__STATUS__CHANNEL_BUSY;
		break;
	case HAL_SEND_OK:
		status = PROTO__EVENT__TRANSMISSION__STATUS__TRANSMITTED;
		break;
	case JIT_ERROR_COLLISION_BEACON:
		status = PROTO__EVENT__TRANSMISSION__STATUS__SCHEDULING_FAILED;
		break;
	case JIT_ERROR_COLLISION_PACKET:
		status = PROTO__EVENT__TRANSMISSION__STATUS__SCHEDULING_FAILED;
		break;
	case JIT_ERROR_GPS_UNLOCKED:
		status = PROTO__EVENT__TRANSMISSION__STATUS__GPS_UNLOCKED;
		break;
	case JIT_ERROR_TOO_EARLY:
		status = PROTO__EVENT__TRANSMISSION__STATUS__BAD_TIMING;
		break;
	case JIT_ERROR_TOO_LATE:
		status = PROTO__EVENT__TRANSMISSION__STATUS__BAD_TIMING;
		break;
	case JIT_ERROR_TX_FREQ:
		status = PROTO__EVENT__TRANSMISSION__STATUS__BAD_FREQUENCY;
		break;
	case JIT_ERROR_TX_POWER:
		status = PROTO__EVENT__TRANSMISSION__STATUS__BAD_POWER;
		break;
	default:
		status = PROTO__EVENT__TRANSMISSION__STATUS__TRANSMISSION_FAILED;
		break;
	}

	return api_send_event_transmission(core, txpkt->message_id, txpkt->transmission_id, status);
}

static int core_recv_lora(struct core *core, Proto__Lora__Downlink *message, struct lgw_pkt_tx_s *txpkt, enum jit_pkt_type_e *downlink_type)
{
	memset(txpkt, 0, sizeof(*txpkt));
	txpkt->modulation = MOD_LORA;

	txpkt->message_id = message->message_id;

	memcpy(txpkt->payload, message->payload.data, (message->payload.len > sizeof(txpkt->payload)) ? sizeof(txpkt->payload) : message->payload.len);
	txpkt->size = message->payload.len;

	if (message->required_transmissions != 1) {
		error("Unsupported multiple required transmissions");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (message->n_configurations == 0) {
		error("No configuration defined");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (message->n_configurations > 1) {
		error("Unsupported multiple configurations");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;
	}

	Proto__Lora__Downlink__Configuration *configuration = message->configurations[0];

	if (configuration->n_timings == 0) {
		error("No timing defined");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (configuration->n_timings > 1) {
		error("Unsupported multiple timings");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;
	}

	Proto__Downlink__Timing *timing = configuration->timings[0];

	txpkt->preamble = message->preamble;
	txpkt->no_crc = !message->crc_enabled;
	txpkt->no_header = message->header_disabled;
	txpkt->invert_pol = message->polarity_inverted;

#if 0
	txpkt->rf_chain = (configuration->radio_id & 0x00000F00) >> 8;
#else
	txpkt->rf_chain = 0; // XXX: temporary hack: forced to RF chain 0. cf LORA-531
#endif

	txpkt->freq_hz = configuration->frequency;

	switch (configuration->bandwidth) {
	case PROTO__LORA__BANDWIDTH__BW_125K:
		txpkt->bandwidth = BW_125KHZ;
		break;
	case PROTO__LORA__BANDWIDTH__BW_250K:
		txpkt->bandwidth = BW_250KHZ;
		break;
	case PROTO__LORA__BANDWIDTH__BW_500K:
		txpkt->bandwidth = BW_500KHZ;
		break;
	default:
		error("Unsupported bandwidth");
		api_send_event_transmission(core, message->message_id, timing->transmission_id,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT);
		return -1;
	}

	switch (configuration->spreading_factor) {
	case PROTO__LORA__SPREADING_FACTOR__SF_7:
		txpkt->datarate = DR_LORA_SF7;
		break;
	case PROTO__LORA__SPREADING_FACTOR__SF_8:
		txpkt->datarate = DR_LORA_SF8;
		break;
	case PROTO__LORA__SPREADING_FACTOR__SF_9:
		txpkt->datarate = DR_LORA_SF9;
		break;
	case PROTO__LORA__SPREADING_FACTOR__SF_10:
		txpkt->datarate = DR_LORA_SF10;
		break;
	case PROTO__LORA__SPREADING_FACTOR__SF_11:
		txpkt->datarate = DR_LORA_SF11;
		break;
	case PROTO__LORA__SPREADING_FACTOR__SF_12:
		txpkt->datarate = DR_LORA_SF12;
		break;
	default:
		error("Unsupported spreading factor");
		api_send_event_transmission(core, message->message_id, timing->transmission_id,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT);
		return -1;
	}

	txpkt->rf_power = configuration->power;

	switch (configuration->coding_rate) {
	case PROTO__LORA__CODING_RATE__CR_4_5:
		txpkt->coderate = CR_LORA_4_5;
		break;
	case PROTO__LORA__CODING_RATE__CR_4_6:
		txpkt->coderate = CR_LORA_4_6;
		break;
	case PROTO__LORA__CODING_RATE__CR_4_7:
		txpkt->coderate = CR_LORA_4_7;
		break;
	case PROTO__LORA__CODING_RATE__CR_4_8:
		txpkt->coderate = CR_LORA_4_8;
		break;
	default:
		error("Unsupported coding rate");
		api_send_event_transmission(core, message->message_id, timing->transmission_id,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT);
		return -1;
	}

	txpkt->transmission_id = timing->transmission_id;

	if (timing->priority == 0) {
		error("Invalid priority");
		api_send_event_transmission(core, message->message_id, timing->transmission_id,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT);
		return -1;
	}

	switch (timing->timing_case) {
	case PROTO__DOWNLINK__TIMING__TIMING_COUNTER:
		txpkt->tx_mode = TIMESTAMPED;
		txpkt->count_us = timing->counter;

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_A;
		break;

	case PROTO__DOWNLINK__TIMING__TIMING_TIME:
		txpkt->tx_mode = TIMESTAMPED;

		int status;

		if ((status = convert_time_to_counter(core, timing->time, &txpkt->count_us)) < 0) {
			api_send_event_transmission(core, message->message_id, timing->transmission_id, -status);
			return -1;
		}

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_B;
		break;

	default:
		txpkt->tx_mode = IMMEDIATE;

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_C;
		break;
	}

	return 0;
}

static int core_recv_gfsk(struct core *core, Proto__Gfsk__Downlink *message, struct lgw_pkt_tx_s *txpkt, enum jit_pkt_type_e *downlink_type)
{
	memset(txpkt, 0, sizeof(*txpkt));
	txpkt->modulation = MOD_FSK;

	txpkt->message_id = message->message_id;

	memcpy(txpkt->payload, message->payload.data, (message->payload.len > sizeof(txpkt->payload)) ? sizeof(txpkt->payload) : message->payload.len);
	txpkt->size = message->payload.len;

	if (message->required_transmissions != 1) {
		error("Unsupported multiple required transmissions");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (message->n_configurations == 0) {
		error("No configuration defined");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (message->n_configurations > 1) {
		error("Unsupported multiple configurations");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;
	}

	Proto__Gfsk__Downlink__Configuration *configuration = message->configurations[0];

	if (configuration->n_timings == 0) {
		error("No timing defined");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;

	} else if (configuration->n_timings > 1) {
		error("Unsupported multiple timings");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;
	}

	Proto__Downlink__Timing *timing = configuration->timings[0];

	txpkt->preamble = message->preamble;
	txpkt->no_header = message->length_fixed;
	txpkt->no_crc = (message->crc == PROTO__GFSK__CRCTYPE__CRC_DISABLED) ? 1 : 0;

	if (message->encoding != PROTO__GFSK__ENCODING_TYPE__ENC_WHITENING) {
		error("Unsupported encoding, only whitening encoding is supported");
		api_send_event_transmission(core, message->message_id, 0,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_MESSAGE_FORMAT);
		return -1;
	}

#if 0
	txpkt->rf_chain = (configuration->radio_id & 0x00000F00) >> 8;
#else
	txpkt->rf_chain = 0; // XXX: temporary hack: forced to RF chain 0. cf LORA-531
#endif

	txpkt->freq_hz = configuration->frequency;
	txpkt->datarate = configuration->bitrate;

	txpkt->rf_power = configuration->power;
	txpkt->f_dev = configuration->frequency_deviation;

	txpkt->transmission_id = timing->transmission_id;

	if (timing->priority == 0) {
		error("Invalid priority");
		api_send_event_transmission(core, message->message_id, timing->transmission_id,
					    PROTO__EVENT__TRANSMISSION__STATUS__BAD_TRANSMISSION_FORMAT);
		return -1;
	}

	switch (timing->timing_case) {
	case PROTO__DOWNLINK__TIMING__TIMING_COUNTER:
		txpkt->tx_mode = TIMESTAMPED;
		txpkt->count_us = timing->counter;

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_A;
		break;

	case PROTO__DOWNLINK__TIMING__TIMING_TIME:
		txpkt->tx_mode = TIMESTAMPED;

		int status;

		if ((status = convert_time_to_counter(core, timing->time, &txpkt->count_us)) < 0) {
			api_send_event_transmission(core, message->message_id, timing->transmission_id, -status);
			return -1;
		}

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_B;
		break;

	default:
		txpkt->tx_mode = IMMEDIATE;

		*downlink_type = JIT_PKT_TYPE_DOWNLINK_CLASS_C;
		break;
	}

	return 0;
}

int api_core_recv(struct core *core, struct lgw_pkt_tx_s *txpkt, enum jit_pkt_type_e *downlink_type)
{
	Proto__Downlink__Container *container = NULL;

	if ((container = api_recv(core)) == NULL) {
		return -1;
	}

	switch (container->messages_case) {
	case PROTO__DOWNLINK__CONTAINER__MESSAGES_LORA:
		if (core_recv_lora(core, container->lora, txpkt, downlink_type) < 0) {
			proto__downlink__container__free_unpacked(container, NULL);
			return -1;
		}

		break;

	case PROTO__DOWNLINK__CONTAINER__MESSAGES_GFSK:
		if (core_recv_gfsk(core, container->gfsk, txpkt, downlink_type) < 0) {
			proto__downlink__container__free_unpacked(container, NULL);
			return -1;
		}

		break;

	default:
		error("Unsupported modulation");
		proto__downlink__container__free_unpacked(container, NULL);
		return -1;
	}

	proto__downlink__container__free_unpacked(container, NULL);
	return 0;
}
