// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils/logger.h"
#include "utils/memory.h"

static int parse_sx1301_lbt(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n, m, l;

	if ((object = json_object_get_object(root, "lbt_cfg")) != NULL) {
		debug("Parsing object: \"SX1301_conf.lbt_cfg\"");

		JSON_BOOLEAN(object, "enable", config_core->lbt.enable);
		JSON_NUMBER(object, "rssi_target", config_core->lbt.rssi_target);

		if ((array = json_object_get_array(object, "chan_cfg")) != NULL) {
			debug("Parsing array: \"SX1301_conf.lbt_cfg.chan_cfg\"");

			for (n = 0, m = 0;
			     n < json_array_get_count(array) &&
			     m < sizeofa(config_core->lbt.channels);
			     n++) {

				if ((object = json_array_get_object(array, n)) != NULL) {
					debug("Parsing object: \"SX1301_conf.lbt_cfg.chan_cfg[%zu]\"", n);

					l = JSON_NUMBER(object, "freq_hz", config_core->lbt.channels[n].freq_hz);
					l += JSON_NUMBER(object, "scan_time_us", config_core->lbt.channels[n].scan_time_us);

					if (l == 2) {
						if (config_core->lbt.channels[n].scan_time_us != 128 &&
						    config_core->lbt.channels[n].scan_time_us != 5000) {
							error("Invalid scan_time_us: %u", config_core->lbt.channels[n].scan_time_us);
							return -1;
						}

						config_core->lbt.nb_channel = ++m;
					}
				}
			}
		}
	}

	return 0;
}

static int parse_sx1301_sniffer(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;

	if ((object = json_object_get_object(root, "sniffer")) != NULL) {
		debug("Parsing object: \"SX1301_conf.sniffer\"");

		JSON_NUMBER(object, "rssi_offset", config_core->sniffer.rssi_offset);
	}

	return 0;
}

static int parse_sx1301_lut(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	char name[32];
	size_t n;

	for (n = 0; n < sizeofa(config_core->txlut.lut); n++) {
		snprintf(name, sizeof(name), "tx_lut_%u", n);

		if ((object = json_object_get_object(root, name)) != NULL) {
			debug("Parsing object: \"SX1301_conf.%s\"", name);

			JSON_NUMBER(object, "dig_gain", config_core->txlut.lut[n].dig_gain);
			JSON_NUMBER(object, "pa_gain", config_core->txlut.lut[n].pa_gain);
			JSON_NUMBER(object, "mix_gain", config_core->txlut.lut[n].mix_gain);
			JSON_NUMBER(object, "rf_power", config_core->txlut.lut[n].rf_power);

			if (config_core->txlut.size < n + 1) {
				config_core->txlut.size = n + 1;
			}
		}
	}

	return 0;
}

static int parse_sx1301_rf(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	char name[32];
	size_t n;

	for (n = 0; n < sizeofa(config_core->rxrf); n++) {
		snprintf(name, sizeof(name), "radio_%u", n);

		if ((object = json_object_get_object(root, name)) != NULL) {
			debug("Parsing object: \"SX1301_conf.%s\"", name);

			JSON_BOOLEAN(object, "enable", config_core->rxrf[n].enable);
			JSON_NUMBER(object, "freq", config_core->rxrf[n].freq_hz);
			JSON_NUMBER(object, "rssi_offset", config_core->rxrf[n].rssi_offset);
			JSON_BOOLEAN(object, "tx_enable", config_core->rxrf[n].tx_enable);
			JSON_NUMBER(object, "tx_freq_min", config_core->rfchains[n].tx_freq_min);
			JSON_NUMBER(object, "tx_freq_max", config_core->rfchains[n].tx_freq_max);
			JSON_NUMBER(object, "tx_notch_freq", config_core->rxrf[n].tx_notch_freq);
		}
	}

	return 0;
}

static int parse_sx1301_if(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	uint32_t fdev = 0;
	uint32_t bw = 0;
	uint32_t sf = 0;
	char name[32];
	size_t n;

	/* Set configuration for Lora multi-SF channels */
	for (n = 0; n < LGW_MULTI_NB; ++n) {
		snprintf(name, sizeof(name), "chan_multiSF_%u", n);

		if ((object = json_object_get_object(root, name)) != NULL) {
			debug("Parsing object: \"SX1301_conf.%s\"", name);

			JSON_BOOLEAN(object, "enable", config_core->rxif[n].enable);
			JSON_NUMBER(object, "radio", config_core->rxif[n].rf_chain);
			JSON_NUMBER(object, "if", config_core->rxif[n].freq_hz);
		}
	}

	/* Set configuration for Lora standard channel */
	if ((object = json_object_get_object(root, "chan_Lora_std")) != NULL) {
		debug("Parsing object: \"SX1301_conf.chan_Lora_std\"");

		JSON_BOOLEAN(object, "enable", config_core->rxif[8].enable);
		JSON_NUMBER(object, "radio", config_core->rxif[8].rf_chain);
		JSON_NUMBER(object, "if", config_core->rxif[8].freq_hz);
		JSON_NUMBER(object, "bandwidth", bw);
		JSON_NUMBER(object, "spread_factor", sf);

		switch (bw) {
		case 500000:
			config_core->rxif[8].bandwidth = BW_500KHZ;
			break;
		case 250000:
			config_core->rxif[8].bandwidth = BW_250KHZ;
			break;
		case 125000:
			config_core->rxif[8].bandwidth = BW_125KHZ;
			break;
		default:
			config_core->rxif[8].bandwidth = BW_UNDEFINED;
			break;
		}

		switch (sf) {
		case 7:
			config_core->rxif[8].datarate = DR_LORA_SF7;
			break;
		case 8:
			config_core->rxif[8].datarate = DR_LORA_SF8;
			break;
		case 9:
			config_core->rxif[8].datarate = DR_LORA_SF9;
			break;
		case 10:
			config_core->rxif[8].datarate = DR_LORA_SF10;
			break;
		case 11:
			config_core->rxif[8].datarate = DR_LORA_SF11;
			break;
		case 12:
			config_core->rxif[8].datarate = DR_LORA_SF12;
			break;
		default:
			config_core->rxif[8].datarate = DR_UNDEFINED;
			break;
		}
	}

	/* Set configuration for FSK channel */
	if ((object = json_object_get_object(root, "chan_FSK")) != NULL) {
		debug("Parsing object: \"SX1301_conf.chan_FSK\"");

		JSON_BOOLEAN(object, "enable", config_core->rxif[9].enable);
		JSON_NUMBER(object, "radio", config_core->rxif[9].rf_chain);
		JSON_NUMBER(object, "if", config_core->rxif[9].freq_hz);
		JSON_NUMBER(object, "datarate", config_core->rxif[9].datarate);
		JSON_NUMBER(object, "bandwidth", bw);
		JSON_NUMBER(object, "freq_deviation", fdev);

		/* if "bandwidth" is set, it has priority over "freq_deviation" */
		if ((bw == 0) && (fdev != 0)) {
			bw = 2 * fdev + config_core->rxif[9].datarate;
		}

		if (bw == 0) {
			config_core->rxif[9].bandwidth = BW_UNDEFINED;
		} else if (bw <= 7800) {
			config_core->rxif[9].bandwidth = BW_7K8HZ;
		} else if (bw <= 15600) {
			config_core->rxif[9].bandwidth = BW_15K6HZ;
		} else if (bw <= 31200) {
			config_core->rxif[9].bandwidth = BW_31K2HZ;
		} else if (bw <= 62500) {
			config_core->rxif[9].bandwidth = BW_62K5HZ;
		} else if (bw <= 125000) {
			config_core->rxif[9].bandwidth = BW_125KHZ;
		} else if (bw <= 250000) {
			config_core->rxif[9].bandwidth = BW_250KHZ;
		} else if (bw <= 500000) {
			config_core->rxif[9].bandwidth = BW_500KHZ;
		} else {
			config_core->rxif[9].bandwidth = BW_UNDEFINED;
		}
	}

	return 0;
}

static int parse_sx1301(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;

	if ((object = json_object_get_object(root, "SX1301_conf")) == NULL) {
		return 0;
	}

	debug("Parsing object: \"SX1301_conf\"");

	/* Board configuration */
	JSON_BOOLEAN(object, "lorawan_public", config_core->board.lorawan_public);

	/* Gateway specificities */
	JSON_NUMBER(object, "antenna_gain", config_core->antenna_gain);
	JSON_NUMBER(object, "insertion_loss", config_core->insertion_loss);

	if (parse_sx1301_lbt(core, object) < 0) {
		error("Failed to parse LBT configuration");
		return -1;

	} else if (parse_sx1301_lut(core, object) < 0) {
		error("Failed to parse LUT configuration");
		return -1;

	} else if (parse_sx1301_rf(core, object) < 0) {
		error("Failed to parse RF configuration");
		return -1;

	} else if (parse_sx1301_sniffer(core, object) < 0) {
		error("Failed to parse sniffer configuration");
		return -1;

	} else if (parse_sx1301_if(core, object) < 0) {
		error("Failed to parse IF configuration");
		return -1;
	}

	return 0;
}

struct config_core *config_core_init(void)
{
	struct config_core *config_core = NULL;
	size_t n;

	if ((config_core = zalloc(sizeof(*config_core))) == NULL) {
		eerror("Failed to allocate core configuration");
		return NULL;
	}

	config_core->board.clksrc = 1;

	for (n = 0; n < sizeofa(config_core->rxrf); n++) {
		config_core->rxrf[n].type = LGW_RADIO_TYPE_SX1257;
		config_core->rxrf[n].tx_notch_freq = 129000;
	}

	for (n = 0; n < sizeofa(config_core->txlut.lut); n++) {
		config_core->txlut.lut[n].dac_gain = 3;
	}

	return config_core;
}

static int memcheck(uint8_t byte, const void *buf, size_t len)
{
	size_t n;

	for (n = 0; n < len; n++) {
		if (((uint8_t *)buf)[n] != byte) {
			return 0;
		}
	}

	return 1;
}

int config_core_parse_calib(struct core *core, struct calib *calib)
{
	struct config_core *config_core = core->config->core;
	uint8_t dig_gain;
	uint8_t pa_gain;
	uint8_t mix_gain;
	int8_t rf_power;
	size_t n, m;

	if (memcheck(0x00, &calib->manufacturer, CALIB_SIZE_HEADER) ||
	    memcheck(0xFF, &calib->manufacturer, CALIB_SIZE_HEADER)) {
		debug("Invalid calibration header: Skipping...");

	} else {
		debug("calib: manufacturer: \"%s\" (%u)",
		      calib_manufacturer_name(calib->manufacturer),
		      calib->manufacturer);

		debug("calib: board: \"%s\" (%u)",
		      calib_board_name(calib->board_type),
		      calib->board_type);

		debug("calib: hardware_version: %u",
		      calib->hardware_version);

		debug("calib: serial_number: %02X%02X%02X",
		      calib->serial_number[0],
		      calib->serial_number[1],
		      calib->serial_number[2]);
	}

	if (memcheck(0x00, &calib->frequency_offset, CALIB_SIZE_OFFSET) ||
	    memcheck(0xFF, &calib->frequency_offset, CALIB_SIZE_OFFSET)) {
		warning("Invalid calibration offset: Setting default value...");

		config_core->rxrf[0].rssi_offset = -166;
		config_core->rxrf[1].rssi_offset = -166;
		config_core->sniffer.rssi_offset = -4;

	} else {
		debug("calib: offset: frequency=%iHz radio=[%gdB, %gdB] sniffer=%gdB",
		      calib->frequency_offset,
		      calib->rssi_offset[CALIB_RADIO_0],
		      calib->rssi_offset[CALIB_RADIO_1],
		      calib->rssi_offset[CALIB_SNIFFER]);

		for (n = 0; n < sizeofa(calib->rssi_offset) && n < LGW_RF_CHAIN_NB; n++) {
			config_core->rxrf[n].rssi_offset = calib->rssi_offset[n];
		}

#if defined(PLATFORM_WSTA)
		/* we don't trust calibration on this platform */
		debug("calib: offset: sniffer set to default value: -4dB");
		config_core->sniffer.rssi_offset = -4;
#else
		config_core->sniffer.rssi_offset = calib->rssi_offset[CALIB_SNIFFER];
#endif
	}

	struct lgw_tx_gain_s default_lut = {
	    .dig_gain = 3,
	    .pa_gain = 1,
	    .dac_gain = 3,
	    .mix_gain = 13,
	    .rf_power = 6,
	};

	if (memcheck(0x00, &calib->tx_lut, CALIB_SIZE_TX_LUT) ||
	    memcheck(0xFF, &calib->tx_lut, CALIB_SIZE_TX_LUT)) {
		warning("Invalid calibration tx_lut: Setting default value...");

		config_core->txlut.lut[0] = default_lut;
		config_core->txlut.size = 1;

	} else {
		for (n = 0, m = 0; n < sizeofa(calib->tx_lut) && n < TX_GAIN_LUT_SIZE_MAX; n++) {
			dig_gain = CALIB_GAIN(DIG, calib->tx_lut[n].dig_pa_mix_gain);
			pa_gain = CALIB_GAIN(PA, calib->tx_lut[n].dig_pa_mix_gain);
			mix_gain = CALIB_GAIN(MIX, calib->tx_lut[n].dig_pa_mix_gain);
			rf_power = calib->tx_lut[n].rf_power;

			if (mix_gain < 8) {
				warning("Invalid mix_gain value: Skipping tx_lut[%02zu]...", n);
				continue;
			}

			debug("calib: tx_lut[%02zu]: "
			      "dig_gain=%u pa_gain=%u "
			      "mix_gain=%02u rf_power=%idBm",
			      n, dig_gain, pa_gain, mix_gain, rf_power);

			config_core->txlut.lut[m].dig_gain = dig_gain;
			config_core->txlut.lut[m].pa_gain = pa_gain;
			config_core->txlut.lut[m].mix_gain = mix_gain;
			config_core->txlut.lut[m].rf_power = rf_power;
			config_core->txlut.size = m + 1;
			m++;
		}

		if (config_core->txlut.size == 0) {
			warning("Empty calibration tx_lut: Setting default value...");

			config_core->txlut.lut[0] = default_lut;
			config_core->txlut.size = 1;
		}
	}

	return 0;
}

int config_core_parse(struct core *core, JSON_Object *root)
{
	if (parse_sx1301(core, root) < 0) {
		error("Failed to parse SX1301 configuration");
		return -1;
	}

	return 0;
}

int config_core_apply(struct core *core)
{
	struct config_core *config_core = core->config->core;
	size_t n;

	if (lgw_board_setconf(config_core->board) != LGW_HAL_SUCCESS) {
		error("Failed to configure board");
		return -1;
	}

	config_core->sniffer.rssi_offset -= config_core->antenna_gain;
	config_core->sniffer.rssi_offset += config_core->insertion_loss;

	config_core->lbt.rssi_offset = config_core->sniffer.rssi_offset;

	if (lgw_lbt_setconf(config_core->lbt) != LGW_HAL_SUCCESS) {
		error("Failed to configure LBT");
		return -1;
	}

	if (lgw_txgain_setconf(&(config_core->txlut)) != LGW_HAL_SUCCESS) {
		error("Failed to configure concentrator TX Gain LUT");
		return -1;
	}

	for (n = 0; n < sizeofa(config_core->rxrf); n++) {
		config_core->rxrf[n].rssi_offset -= config_core->antenna_gain;
		config_core->rxrf[n].rssi_offset += config_core->insertion_loss;

		if (lgw_rxrf_setconf(n, config_core->rxrf[n]) != LGW_HAL_SUCCESS) {
			error("Failed to configure radio %u", n);
			return -1;
		}
	}

	for (n = 0; n < sizeofa(config_core->rxif); n++) {
		if (lgw_rxif_setconf(n, config_core->rxif[n]) != LGW_HAL_SUCCESS) {
			error("Failed to configure channel %u", n);
			return -1;
		}
	}

	return 0;
}

void config_core_free(struct core *core)
{
	if (core && core->config) {
		free(core->config->core);
	}
}
