// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "core.h"

#include <stdbool.h>
#include <stdlib.h>

#include "lib/loragw/spi.h"

#include "lib/loragw/aux.h"
#include "lib/loragw/hal.h"
#include "lib/loragw/reg.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "beacon.h"
#include "clock.h"
#include "common/statistics.h"
#include "config.h"
#include "core.h"
#include "core_fscan.h"
#include "jit.h"

#define lock_guard(_m, _t, _f) ({ _t r; pthread_mutex_lock(_m); r = _f; pthread_mutex_unlock(_m); r; })

static pthread_mutex_t mx_concent = PTHREAD_MUTEX_INITIALIZER;

static struct jit_queue_s jit_queue;

#define FETCH_SLEEP_MS 10
#define NB_PKT_MAX 8

#define LIBRARY_VERSION "5.0.1"

static int counter_sampled_on_pps = 0;

int get_counter_sampled_on_pps(void)
{
	return counter_sampled_on_pps;
}

int set_counter_sampled_on_pps(int mode)
{
	if (counter_sampled_on_pps != mode) {
		if (pthread_mutex_lock(&mx_concent) != 0) {
			return -1;
		}

		lgw_reg_w(LGW_GPS_EN, mode);

		pthread_mutex_unlock(&mx_concent);

		counter_sampled_on_pps = mode;
	}

	return 0;
}

int get_counter(uint32_t *counter)
{
	if (pthread_mutex_lock(&mx_concent) != 0) {
		return -1;
	}

	lgw_get_trigcnt(counter);

	pthread_mutex_unlock(&mx_concent);

	return 0;
}

static void *thread_uplink(void *arg)
{
	struct core *core = arg;
	struct lgw_pkt_rx_s rxpkt[NB_PKT_MAX];
	int rxpkt_nb;

	while (core->running) {
		if ((rxpkt_nb = lock_guard(&mx_concent, int, lgw_receive(sizeofa(rxpkt), rxpkt))) < 0) {
			error("Failed to receive messages");
			continue;

		} else if (rxpkt_nb == 0) {
			wait_ms(FETCH_SLEEP_MS);
			continue;
		}

		if (api_core_send(core, rxpkt, rxpkt_nb) < 0) {
			if (api_running(core)) {
				error("Failed to send uplink message");
			}

		} else {
			info("Sent %u uplink message%s", rxpkt_nb, (rxpkt_nb > 1) ? "s" : "");
		}
	}

	pthread_exit(NULL);
}

static int check_power(struct core *core, int8_t power)
{
	struct lgw_tx_gain_lut_s *txlut = &core->config->core->txlut;
	int index;

	for (index = 0; index < txlut->size; index++) {
		if (txlut->lut[index].rf_power <= power) {
			return 1;
		}
	}

	return 0;
}

static int check_frequency(struct core *core, uint32_t frequency, uint8_t rf_chain)
{
	struct config_core *config_core = core->config->core;

	if (config_core->rfchains[rf_chain].tx_freq_min != 0 &&
	    frequency < config_core->rfchains[rf_chain].tx_freq_min) {
		return 0;
	}

	if (config_core->rfchains[rf_chain].tx_freq_max != 0 &&
	    frequency > config_core->rfchains[rf_chain].tx_freq_max) {
		return 0;
	}

	return 1;
}

static void *thread_downlink(void *arg)
{
	struct core *core = arg;
	enum jit_pkt_type_e downlink_type;
	struct lgw_pkt_tx_s txpkt;
	enum jit_error_e jit;

	while (core->running) {
		if (api_core_recv(core, &txpkt, &downlink_type) < 0) {
			if (api_running(core)) {
				error("Failed to receive downlink message");
			}

			continue;
		}

		info("Received downlink message");
		core->statistics->downlink_requested++;

		/* Adjust packet RF power with antenna gain and insertion loss */
		debug("Requested power: %i dB", txpkt.rf_power);
		txpkt.rf_power -= core->config->core->antenna_gain;
		txpkt.rf_power += core->config->core->insertion_loss;
		debug("Selected power: %i dB", txpkt.rf_power);

		if (!check_power(core, txpkt.rf_power)) {
			error("Unsupported power: %i", txpkt.rf_power);
			api_core_send_event(core, &txpkt, JIT_ERROR_TX_POWER);
			continue;
		}

		if (!check_frequency(core, txpkt.freq_hz, txpkt.rf_chain)) {
			error("Unsupported frequency: %u", txpkt.freq_hz);
			api_core_send_event(core, &txpkt, JIT_ERROR_TX_FREQ);
			continue;
		}

		if ((jit = jit_enqueue(core, &jit_queue, &txpkt, downlink_type)) != JIT_ERROR_OK) {
			error("Failed to enqueue downlink message");
			api_core_send_event(core, &txpkt, jit);
			continue;
		}

		core->statistics->downlink_scheduled++;
	}

	pthread_exit(NULL);
}

static void *thread_jit(void *arg)
{
	struct core *core = arg;
	struct lgw_pkt_tx_s txpkt;
	enum jit_pkt_type_e type;
	int result, index = -1;
	enum jit_error_e jit;
	uint8_t status;

	while (core->running) {
		wait_ms(10);

		if ((jit = jit_peek(core, &jit_queue, &index)) != JIT_ERROR_OK) {
			if (jit != JIT_ERROR_EMPTY) {
				error("Failed to peek message");
			}

			continue;

		} else if (index < 0) {
			continue;

		} else if (jit_dequeue(&jit_queue, index, &txpkt, &type) != JIT_ERROR_OK) {
			error("Failed to dequeue message");
			continue;
		}

		if (lock_guard(&mx_concent, int, lgw_status(TX_STATUS, &status)) == LGW_HAL_ERROR) {
			error("Failed to get concentrator status");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if (status == TX_EMITTING) {
			error("Concentrator is busy");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if (status == TX_SCHEDULED) {
			warning("Overwriting an already scheduled message");
		}

		if ((result = lock_guard(&mx_concent, int, lgw_send(core, txpkt))) == LGW_HAL_ERROR) {
			error("Failed to send message");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if (result == LGW_LBT_ISSUE) {
			error("Failed to send message: Not allowed (LBT)");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_LBT);
			}
			continue;

		} else if (type == JIT_PKT_TYPE_BEACON) {
			info("Sent beacon");

		} else {
			info("Sent downlink message");
			api_core_send_event(core, &txpkt, HAL_SEND_OK);
			core->statistics->downlink_transmitted++;
			core->statistics->downlink_bytes_transmitted += txpkt.size;
		}
	}

	pthread_exit(NULL);
}

static void *thread_beacon(void *arg)
{
	struct core *core = arg;
	struct config_beacon *config_beacon = &core->config->beacon;

	if (config_beacon->enable &&
	    config_beacon->period > 0 &&
	    config_beacon->frequencies > 0) {
		while (core->running) {
			if (core->clock->pps_enabled &&
			    core->clock->tai_offset != 0 &&
			    beacon_core_schedule(core, &jit_queue) < 0) {
				error("Failed to schedule the next beacon");
			}

			sleep_sec(3, core->running);
		}
	}

	pthread_exit(NULL);
}

/* Usefull when GPS device is reset by "system" */
static int gps_reset(struct core *core)
{
	if (core->gps_enabled) {
		if (lgw_gps_disable(core->gps_fd) != LGW_HAL_SUCCESS) {
			warning("Failed to close GPS device during reset");
		}
		core->gps_enabled = 0;
	}

	if (!core->gps_enabled) {
		if (lgw_gps_enable(core->gps_device, "ubx7", 0, &core->gps_fd) < 0) {
			error("Failed to initialize GPS during reset");
			return -1;
		}
		core->gps_enabled = 1;
	}

	return 0;
}

static ssize_t read_nmea_sentence(struct core *core, char *buffer, size_t size)
{
	size_t length;
	ssize_t len;
	int started;
	char byte;

	for (started = 0, length = 0; length < size;) {
		if ((len = read(core->gps_fd, &byte, sizeof(byte))) < 0) {
			eerror("Failed to read serial");
			if (gps_reset(core) < 0) {
				error("Unable to reset GPS");
			}
			sleep_sec(1, core->running);
			return -1;
		}

		if (len == 0) {
			error("Unexpected behaviour, no data read on GPS device");
			if (gps_reset(core) < 0) {
				error("Unable to reset GPS");
			}
			sleep_sec(1, core->running);
			return -1;
		}

		if (started == 1) {
			if (byte == '\r' || byte == '\n') {
				buffer[length] = '\0';
				return length;
			}

			buffer[length++] = byte;

		} else if (byte == '$') {
			started = 1;
			buffer[length++] = byte;
		}
	}

	return -1;
}

static void *thread_gps(void *arg)
{
	struct core *core = arg;
	ssize_t length;
	char buffer[256];
	struct coord_s coord;

	while (core->running) {
		if ((length = read_nmea_sentence(core, buffer, sizeof(buffer))) < 0) {
			eerror("Failed to read NMEA sentence");
			continue;
		}

		if (lgw_parse_nmea(buffer, length) != NMEA_GGA) {
			continue;
		}

		if (lgw_gps_get(NULL, NULL, &coord, NULL) < 0) {
			continue;
		}

		core->statistics->gps_latitude = coord.lat;
		core->statistics->gps_longitude = coord.lon;
		core->statistics->gps_altitude = coord.alt;
		core->statistics->has_gps_coordinates = 1;
	}

	pthread_exit(NULL);
}

void fill_version_statistics(struct statistics *statistics)
{
	lgw_get_fpga_version((uint8_t *)&statistics->version_fpga);
	if (statistics->version_fpga != 0) {
		statistics->has_version_fpga = 1;
	}
	statistics->has_version_dsp = 0;
	strncpy(statistics->version_hal, LIBRARY_VERSION, 32);
	statistics->has_version_hal = 1;

	debug("FPGA version: %d ; HAL version: %s",
	      statistics->has_version_fpga ? statistics->version_fpga : 0,
	      statistics->has_version_hal ? statistics->version_hal : "");
}

static int parse_calib(struct core *core)
{
	struct calib *calib;

	debug("Parsing calibration");

	if (lgw_connect(true, 0) < 0) {
		error("Failed to connect concentrator");
		return -1;
	}

	if ((calib = calib_read()) == NULL) {
		error("Failed to read calibration");
		lgw_disconnect();
		return -1;
	}

	if (config_core_parse_calib(core, calib) < 0) {
		error("Failed to parse calibration");
		lgw_disconnect();
		free(calib);
		return -1;
	}

	lgw_disconnect();
	free(calib);

	return 0;
}

struct core *core_init(struct stringlist *filenames, char *gps_device, uint32_t serial_number)
{
	struct core *core = NULL;
	size_t nb;

	if ((core = zalloc(sizeof(*core))) == NULL) {
		eerror("Failed to allocate core");
		return NULL;
	}

	if ((core->config = config_init()) == NULL) {
		error("Failed to initialize configuration");
		core_free(core);
		return NULL;
	}

	core->config->beacon.gwid = serial_number & 0x00FFFFFF;

	if (parse_calib(core) < 0) {
		error("Failed to parse calibration");
		core_free(core);
		return NULL;
	}

	if ((core->api = api_init()) == NULL) {
		error("Failed to initialize API");
		core_free(core);
		return NULL;
	}

	if ((core->clock = clock_init()) == NULL) {
		error("Failed to initialize clock");
		core_free(core);
		return NULL;
	}

	if ((core->fscan = fscan_init()) == NULL) {
		error("Failed to initialize fscan");
		core_free(core);
		return NULL;
	}

	if ((core->statistics = zalloc(sizeof(*core->statistics))) == NULL) {
		eerror("Failed to allocate statistic");
		core_free(core);
		return NULL;
	}

	for (nb = 0; filenames && nb < filenames->nb_strings; nb++) {
		char *filename = filenames->strings[nb];

		info("Parsing configuration file: %s", filename);

		if (config_parse_file(core, filename)) {
			error("Failed to parse configuration file %s", filename);
			core_free(core);
			return NULL;
		}
	}

	if (gps_device != NULL) {
		core->gps_enabled = 1;

		if (lgw_gps_enable(gps_device, "ubx7", 0, &core->gps_fd) < 0) {
			error("Failed to initialize GPS");
			core_free(core);
			return NULL;
		}
		strncpy(core->gps_device, gps_device, sizeof(core->gps_device));
	}

	jit_queue_init(&jit_queue);

	return core;
}

int core_start(struct core *core)
{
	if (config_apply(core)) {
		error("Failed to apply configuration");
		return -1;
	}

	if (lgw_start() < 0) {
		error("Failed to start concentrator");
		return -1;
	}

	core->started = 1;

	if (clock_start(core) < 0) {
		error("Failed to start clock");
		return -1;
	}

	if (fscan_start(core) < 0) {
		error("Failed to start fscan");
		return -1;
	}

	if (api_start(core) < 0) {
		error("Failed to start API");
		return -1;
	}

	core->running = 1;

	if (core->gps_enabled) {
		if (pthread_create(&core->gps_thread, NULL, thread_gps, core) != 0) {
			error("Failed to start gps thread");
			core_shutdown(core);
			return -1;
		}
	}

	if (pthread_create(&core->beacon_thread, NULL, thread_beacon, core) != 0) {
		error("Failed to start beacon thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->jit_thread, NULL, thread_jit, core) != 0) {
		error("Failed to start jit thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->downlink_thread, NULL, thread_downlink, core) != 0) {
		error("Failed to start downlink thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->uplink_thread, NULL, thread_uplink, core) != 0) {
		error("Failed to start uplink thread");
		core_shutdown(core);
		return -1;
	}

	api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__STARTED);

	return 0;
}

int core_join(struct core *core)
{
	if (pthread_join(core->uplink_thread, NULL) != 0) {
		error("Failed to join uplink thread");
		return -1;
	}

	if (pthread_join(core->downlink_thread, NULL) != 0) {
		error("Failed to join downlink thread");
		return -1;
	}

	if (pthread_join(core->jit_thread, NULL) != 0) {
		error("Failed to join jit thread");
		return -1;
	}

	if (pthread_join(core->beacon_thread, NULL) != 0) {
		error("Failed to join beacon thread");
		return -1;
	}

	if (core->gps_enabled) {
		if (pthread_join(core->gps_thread, NULL) != 0) {
			error("Failed to join gps thread");
			return -1;
		}
	}

	if (api_join(core) < 0) {
		error("Failed to join API threads");
		return -1;
	}

	if (fscan_join(core) < 0) {
		error("Failed to join fscan threads");
		return -1;
	}

	if (clock_join(core) < 0) {
		error("Failed to join clock threads");
		return -1;
	}

	return 0;
}

void core_shutdown(struct core *core)
{
	if (core) {
		core->running = 0;
		api_shutdown(core);
		fscan_shutdown(core);
		clock_shutdown(core);
	}
}

void core_free(struct core *core)
{
	if (core) {
		if (core->gps_enabled) {
			if (lgw_gps_disable(core->gps_fd) != LGW_HAL_SUCCESS) {
				warning("Failed to close GPS device");
			}
		}

		if (core->started &&
		    lgw_stop() != LGW_HAL_SUCCESS) {
			warning("Failed to stop concentrator");
		}

		fscan_free(core);
		clock_free(core);
		config_free(core);
		free(core->statistics);

		api_free(core);
		free(core);
	}
}
