// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "core_fscan.h"

#include <stdlib.h>
#include <sys/types.h>

#if !defined(PLATFORM_WIFC)
#include "utils/clock.h"
#endif
#include "utils/logger.h"
#include "utils/memory.h"

#include "config.h"
#include "core.h"

#if defined(PLATFORM_WIFC)
#include "drivers/sx1239.h"
#else
#include "drivers/sx1272.h"
#endif

#include "lib/loragw/lbt.h"
#if !defined(PLATFORM_WIFC)
#include "lib/loragw/aux.h"
#include "lib/loragw/fpga.h"
#include "lib/loragw/hal.h"
#include "lib/loragw/radio.h"
#include "lib/loragw/reg.h"
#endif

#if defined(PLATFORM_WIFC)
#define SNIFFER_NB_MAX 1
#define SNIFFER_SPI_PATH "/dev/spidev1.0"
#define SNIFFER_SPI_SPEED 10000000
#define SNIFFER_GPIO_EN 10
#define SNIFFER_GPIO_RST 135
#define SNIFFER_BANDWIDTH 200000
#define SNIFFER_BITRATE 600
#define SNIFFER_LNA_GAIN 1
#define SNIFFER_DCC_FREQ 1
#else
#define SNIFFER_NB_MAX 1
#define SNIFFER_BANDWIDTH 200000
#define SNIFFER_BITRATE 1000
#endif

#if defined(PLATFORM_WIFC)
static struct {
	struct spi_device *spi;
	struct gpio_device *enable;
	struct gpio_device *reset;
	struct sx1239_device *sniffer;
} sniffer_list[SNIFFER_NB_MAX] = {0};
#else
struct sx1272_device *sniffer = NULL;
#endif

struct fscan_core *fscan_core_init(void)
{
	struct fscan_core *fscan_core = NULL;

	if ((fscan_core = zalloc(sizeof(*fscan_core))) == NULL) {
		eerror("Failed to allocate fscan core");
		return NULL;
	}

	return fscan_core;
}

int fscan_core_start(uint32_t receiver_id)
{
	if (lbt_is_enabled()) {
		notice("LBT is enabled, no frequency scan available");
		return -1;
	}

#if defined(PLATFORM_WIFC)
	if ((sniffer_list[receiver_id].spi = spi_open(SNIFFER_SPI_PATH, NULL)) == NULL) {
		error("Failed to open sniffer spi");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if (sniffer_list[receiver_id].spi->set_speed(sniffer_list[receiver_id].spi, SNIFFER_SPI_SPEED) < 0) {
		error("Failed to configure sniffer spi");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if ((sniffer_list[receiver_id].enable = gpio_open(SNIFFER_GPIO_EN, GPIO_OUTPUT)) == NULL) {
		error("Failed to open sniffer gpio enable");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if ((sniffer_list[receiver_id].reset = gpio_open(SNIFFER_GPIO_RST, GPIO_OUTPUT)) == NULL) {
		error("Failed to open sniffer gpio reset");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if ((sniffer_list[receiver_id].sniffer = sx1239_open(sniffer_list[receiver_id].spi,
							     sniffer_list[receiver_id].enable,
							     sniffer_list[receiver_id].reset)) == NULL) {
		error("Failed to open sniffer");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if (sniffer_list[receiver_id].sniffer->set_bandwidth(sniffer_list[receiver_id].sniffer, SNIFFER_BANDWIDTH) < 0) {
		error("Failed to configure sniffer bandwidth");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if (sniffer_list[receiver_id].sniffer->set_bitrate(sniffer_list[receiver_id].sniffer, SNIFFER_BITRATE) < 0) {
		error("Failed to configure sniffer bit rate");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if (sniffer_list[receiver_id].sniffer->set_dcc_freq(sniffer_list[receiver_id].sniffer, SNIFFER_DCC_FREQ) < 0) {
		error("Failed to configure sniffer DCC frequency");
		fscan_core_shutdown(receiver_id);
		return -1;
	}

	if (sniffer_list[receiver_id].sniffer->set_lna_gain(sniffer_list[receiver_id].sniffer, SNIFFER_LNA_GAIN) < 0) {
		error("Failed to configure sniffer LNA gain");
		fscan_core_shutdown(receiver_id);
		return -1;
	}
#else
	(void) receiver_id;
	sniffer = sx1272_open(SNIFFER_BANDWIDTH, SNIFFER_BITRATE);
	if (sniffer == NULL) {
		error("Failed to open sx1272 device");
		return -1;
	}
#endif

	return 0;
}

void fscan_core_shutdown(uint32_t receiver_id)
{
#if defined(PLATFORM_WIFC)
	if (sniffer_list[receiver_id].sniffer) {
		sniffer_list[receiver_id].sniffer->close(sniffer_list[receiver_id].sniffer);
	}

	if (sniffer_list[receiver_id].reset) {
		sniffer_list[receiver_id].reset->close(sniffer_list[receiver_id].reset);
	}

	if (sniffer_list[receiver_id].enable) {
		sniffer_list[receiver_id].enable->close(sniffer_list[receiver_id].enable);
	}

	if (sniffer_list[receiver_id].spi) {
		sniffer_list[receiver_id].spi->close(sniffer_list[receiver_id].spi);
	}
#else
	(void) receiver_id;
	if (sniffer) {
		sniffer->close(sniffer);
		sniffer = NULL;
	}
#endif
}

void fscan_core_free(struct core *core)
{
	if (core && core->fscan && core->fscan->core) {
		free(core->fscan->core);
	}
}

int fscan_get_capabilities(struct core *core, struct sniffer_caps *caps)
{
	(void)core;

	caps->sniffer_nb = SNIFFER_NB_MAX;

	return 0;
}

int fscan_do_scan(float *rssi, struct core *core, uint32_t receiver_id, uint32_t frequency)
{
	int rssi_value;

	if (receiver_id >= SNIFFER_NB_MAX) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}

#if defined(PLATFORM_WIFC)
	/* Update frequency scan */
	if (sniffer_list[receiver_id].sniffer->start_rx(sniffer_list[receiver_id].sniffer, frequency) < 0) {
		error("Failed to update frequency");
		return -1;
	}

	/* Retrieve scan result */
	if ((rssi_value = sniffer_list[receiver_id].sniffer->rssi(sniffer_list[receiver_id].sniffer)) < 0) {
		error("Failed to get rssi");
		return -1;
	}

	*rssi = -((float)rssi_value / 2) + core->config->core->sniffer.rssi_offset;
#else
	/* set frequency */
	if (sniffer->set_freq(sniffer, frequency) < 0) {
		error("Failed to update frequency to %lu", (unsigned long)frequency);
		return -1;
	}

	/* Retrieve rssi */
	if (sniffer->rssi(sniffer, 1, &rssi_value) < 0) {
		error("Failed to get rssi at frequency %lu", (unsigned long)frequency);
		return -1;
	}
	*rssi = -((float)rssi_value / 2) + core->config->core->sniffer.rssi_offset;
#endif

	return 0;
}

bool fscan_is_lbt_enabled(uint32_t receiver_id)
{
	/* Unused parameter */
	receiver_id = receiver_id;

	if (lbt_is_enabled()) {
		return true;
	} else {
		return false;
	}
}
