// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "lib/sx1301ar/err.h"
#include "lib/sx1301ar/spi.h"

#include "drivers/gpio.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#define lerror(fmt, args...) error(fmt ": %s", ##args, sx1301ar_err_message(sx1301ar_errno))

#define SLOT_MAX 5
#define SLOT_ROOT "/dev/slot/%zu/"
#define SLOT_SPIDEV SLOT_ROOT "spidev0"
#define SLOT_TTYACM SLOT_ROOT "ttyACM"

static int gpio_power_ids[SLOT_MAX] = {198, 199, 200, 201, 193};

static int *spi_fd[SLOT_MAX] = {0};
static pthread_mutex_t *spi_mutex[SLOT_MAX] = {0};

#define _declare_spi(func, slot, const...)                                                             \
	static int spi##slot##_##func(uint8_t h, uint16_t a, const uint8_t *d, uint32_t l, uint8_t *s) \
	{                                                                                              \
		int ret;                                                                               \
		pthread_mutex_lock(spi_mutex[slot]);                                                   \
		ret = spi_linuxdev_##func(h, *spi_fd[slot], a, d, l, s);                               \
		pthread_mutex_unlock(spi_mutex[slot]);                                                 \
		return ret;                                                                            \
	}

#define declare_spi(slot) _declare_spi(read, slot) _declare_spi(write, slot, const)

declare_spi(0);
declare_spi(1);
declare_spi(2);
declare_spi(3);
declare_spi(4);

static int discover_slot(struct config_board *board, size_t nb_slot)
{
	char *ttyacm = NULL;
	char *spidev = NULL;
	int ret = -1;

	board->spi_fd = -1;

	// Checking for ttyACM (Gecko)
	if (aprintf(&ttyacm, SLOT_TTYACM, nb_slot + 1) < 0) {
		error("Failed to format ttyacm path");
		goto out;
	}

	if (access(ttyacm, F_OK) < 0) {
		debug("Cannot access ttyacm: %s", ttyacm);
		ret = 0;
		goto out;
	}

	// Checking for spidev (FPGA)
	if (aprintf(&spidev, SLOT_SPIDEV, nb_slot + 1) < 0) {
		error("Failed to format spidev path");
		goto out;
	}

	if (access(spidev, F_OK) < 0) {
		debug("Cannot access spidev: %s", spidev);
		ret = 0;
		goto out;
	}

	// Opening spidev
	if (spi_linuxdev_open(spidev, -1, &board->spi_fd) < 0) {
		error("Failed to open spidev: %s", spidev);
		goto out;
	}

	spi_fd[nb_slot] = &board->spi_fd;

	// Initializing spi mutex
	if (pthread_mutex_init(&board->spi_mutex, NULL) != 0) {
		error("Failed to initialize mutex");
		goto out;
	}

	spi_mutex[nb_slot] = &board->spi_mutex;

	switch (nb_slot) {
	case 0:
		board->board.spi_read = &spi0_read;
		board->board.spi_write = &spi0_write;
		break;
	case 1:
		board->board.spi_read = &spi1_read;
		board->board.spi_write = &spi1_write;
		break;
	case 2:
		board->board.spi_read = &spi2_read;
		board->board.spi_write = &spi2_write;
		break;
	case 3:
		board->board.spi_read = &spi3_read;
		board->board.spi_write = &spi3_write;
		break;
	case 4:
		board->board.spi_read = &spi4_read;
		board->board.spi_write = &spi4_write;
		break;
	}

	// Opening power gpio
	if ((board->power = gpio_open(gpio_power_ids[nb_slot], GPIO_OUTPUT | GPIO_HIGH)) == NULL) {
		error("Failed to open power gpio");
		goto out;
	}

	board->nb_slot = nb_slot;

	return 1;

out:
	if (board->spi_fd >= 0) {
		spi_linuxdev_close(board->spi_fd);
	}

	free(spidev);
	free(ttyacm);

	return ret;
}

static int init_slot(struct config_core *config)
{
	size_t nb_slot, nb_board;

	for (config->nb_slots = 0, nb_slot = 0;
	     nb_slot < SLOT_MAX && config->nb_slots < sizeofa(config->boards);
	     nb_slot++) {

		switch (discover_slot(&config->boards[config->nb_slots], nb_slot)) {
		case 1:
			info("Using slot #%zu for board #%zu", nb_slot + 1, config->nb_slots);
			config->nb_slots++;
			break;

		case 0:
			debug("Skipping slot #%zu", nb_slot + 1);
			break;

		default:
			error("Failed to discover slot");
			return -1;
		}
	}

	if (config->nb_slots == 0) {
		error("No slot detected");
		return -1;
	}

	debug("Power cycling slots...");

	for (nb_board = 0; nb_board < config->nb_boards; nb_board++) {
		struct config_board *board = &config->boards[nb_board];

		if (board->power->set_value(board->power, GPIO_LOW) < 0) {
			error("slot #%zu: Failed to power off", board->nb_slot);
			return -1;
		}

		sleep_ms(100);

		if (board->power->set_value(board->power, GPIO_HIGH) < 0) {
			error("slot #%zu: Failed to power on", board->nb_slot);
			return -1;
		}

		sleep_ms(100);
	}

	info("Waiting for fpga...");
	sleep_ms(10000);

	return 0;
}

static int parse_rfchain_lut(struct core *core, JSON_Object *root, size_t nb_board, size_t nb_rfchain)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	sx1301ar_tx_gain_lut_t *txlut = &board->board.rf_chain[nb_rfchain].tx_lut;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n;

	if ((array = json_object_get_array(root, "tx_lut")) != NULL) {
		debug("Parsing array: \"SX1301_array_conf[%zu].rf_chain_conf[%zu].tx_lut\"", nb_board, nb_rfchain);

		for (n = 0;
		     n < json_array_get_count(array) &&
		     n < sizeofa(txlut->lut);
		     n++) {
			if ((object = json_array_get_object(array, n)) != NULL) {
				debug("Parsing object: \"SX1301_array_conf[%zu].rf_chain_conf[%zu].tx_lut[%zu]\"", nb_board, nb_rfchain, n);

				JSON_NUMBER(object, "rf_power", txlut->lut[n].rf_power);
				JSON_NUMBER(object, "fpga_dig_gain", txlut->lut[n].fpga_dig_gain);
				JSON_NUMBER(object, "ad9361_atten", txlut->lut[n].ad9361_gain.atten);
				JSON_NUMBER(object, "ad9361_auxdac_vref", txlut->lut[n].ad9361_gain.auxdac_vref);
				JSON_NUMBER(object, "ad9361_auxdac_word", txlut->lut[n].ad9361_gain.auxdac_word);
				JSON_NUMBER(object, "ad9361_tcomp_coeff_a", txlut->lut[n].ad9361_tcomp.coeff_a);
				JSON_NUMBER(object, "ad9361_tcomp_coeff_b", txlut->lut[n].ad9361_tcomp.coeff_b);

				if (txlut->size < n + 1) {
					txlut->size = n + 1;
				}
			}
		}
	}

	return 0;
}

static int parse_rfchain(struct core *core, JSON_Object *root, size_t nb_board)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n;

	if ((array = json_object_get_array(root, "rf_chain_conf")) != NULL) {
		debug("Parsing array: \"SX1301_array_conf[%zu].rf_chain_conf\"", nb_board);

		for (n = 0;
		     n < json_array_get_count(array) &&
		     n < sizeofa(board->board.rf_chain);
		     n++) {
			sx1301ar_rfchain_t *rfchain = &board->board.rf_chain[n];

			if ((object = json_array_get_object(array, n)) != NULL) {
				debug("Parsing object: \"SX1301_array_conf[%zu].rf_chain_conf[%zu]\"", nb_board, n);

				JSON_BOOLEAN(object, "rx_enable", rfchain->rx_enable);
				JSON_BOOLEAN(object, "tx_enable", rfchain->tx_enable);
				JSON_NUMBER(object, "antenna_gain", rfchain->antenna_gain);
				JSON_NUMBER(object, "insertion_loss", rfchain->insertion_loss);
				JSON_NUMBER(object, "rssi_offset", rfchain->rssi_offset);
				JSON_NUMBER(object, "rssi_offset_coeff_a", rfchain->rssi_offset_coeff_a);
				JSON_NUMBER(object, "rssi_offset_coeff_b", rfchain->rssi_offset_coeff_b);
				JSON_NUMBER(object, "tx_freq_min", board->rfchains[n].tx_freq_min);
				JSON_NUMBER(object, "tx_freq_max", board->rfchains[n].tx_freq_max);

				if (parse_rfchain_lut(core, object, nb_board, n) < 0) {
					error("Failed to parse rfchain lut");
					return -1;
				}
			}
		}
	}

	return 0;
}

static int parse_chip_object(struct core *core, JSON_Object *root, size_t nb_board, size_t nb_chip)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	sx1301ar_chip_cfg_t *chip = &board->chips[nb_chip].chip;
	JSON_Object *object = NULL;
	char name[32];
	size_t n;

	debug("Parsing object: \"SX1301_array_conf[%zu].SX1301_conf[%zu]\"", nb_board, nb_chip);

	JSON_NUMBER(root, "chip_center_freq", chip->freq_hz);
	JSON_NUMBER(root, "chip_rf_chain", chip->rf_chain);

	for (n = 0; n < SX1301AR_CHIP_CHAN_NB; ++n) {
		sx1301ar_chan_cfg_t *chan = &board->chips[nb_chip].chans[n];

		if (n < SX1301AR_CHIP_MULTI_NB) {
			snprintf(name, sizeof(name), "chan_multiSF_%u", n);

		} else if (n == SX1301AR_CHIP_LSA_IDX) {
			snprintf(name, sizeof(name), "chan_LoRa_std");

		} else if (n == SX1301AR_CHIP_FSK_IDX) {
			snprintf(name, sizeof(name), "chan_FSK");
		}

		if ((object = json_object_get_object(root, name)) == NULL) {
			continue;
		} else {
			debug("Parsing object: \"SX1301_array_conf[%zu].SX1301_conf[%zu].%s\"", nb_board, nb_chip, name);
			chan->enable = true;
		}

		JSON_BOOLEAN(object, "enable", chan->enable);

		if (chan->enable) {
			board->enable = true;
		} else {
			continue;
		}

		JSON_NUMBER(object, "chan_rx_freq", chan->freq_hz);

		if (n < SX1301AR_CHIP_MULTI_NB) {
			JSON_Value *sf = NULL;
			const char *str = NULL;
			uint32_t one = 0;
			uint32_t two = 0;

			if ((sf = json_object_dotget_value(object, "spread_factor")) != NULL) {
				if (json_value_get_type(sf) == JSONString) {
					str = json_value_get_string(sf);

					switch (sscanf(str, "%u-%u", &one, &two)) {
					case 2:
						chan->modrate = sx1301ar_sf_range_nb2enum(one, two);
						break;
					case 1:
						chan->modrate = sx1301ar_sf_nb2enum(one);
						break;
					default:
						error("Invalid format for \"spread_factor\"");
						return -1;
					}

					debug("Value \"spread_factor\" configured to: \"%s\"", str);

				} else if (json_value_get_type(sf) == JSONNumber) {
					one = json_value_get_number(sf);
					chan->modrate = sx1301ar_sf_nb2enum(one);
					debug("Value \"spread_factor\" configured to: %u", one);

				} else {
					warning("Wrong data type for \"spread_factor\"");
				}
			}

			chan->bandwidth = BW_125K;

		} else if (n == SX1301AR_CHIP_LSA_IDX) {
			uint32_t bw = 0;
			uint32_t sf = 0;

			JSON_NUMBER(object, "bandwidth", bw);
			JSON_NUMBER(object, "spread_factor", sf);

			chan->bandwidth = sx1301ar_bw_nb2enum(bw);
			chan->modrate = sx1301ar_sf_nb2enum(sf);

		} else if (n == SX1301AR_CHIP_FSK_IDX) {
			uint32_t bw = 0;

			JSON_NUMBER(object, "bandwidth", bw);
			JSON_NUMBER(object, "bit_rate", chan->modrate);

			chan->bandwidth = sx1301ar_bw_nb2enum(bw);
		}
	}

	if (board->board.nb_chip < nb_chip + 1) {
		board->board.nb_chip = nb_chip + 1;
	}

	return 0;
}

static int parse_chip(struct core *core, JSON_Object *root, size_t nb_board)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n;

	if ((array = json_object_get_array(root, "SX1301_conf")) != NULL) {
		debug("Parsing array: \"SX1301_array_conf[%zu].SX1301_conf\"", nb_board);

		for (n = 0;
		     n < json_array_get_count(array) &&
		     n < sizeofa(board->chips);
		     n++) {
			if ((object = json_array_get_object(array, n)) != NULL) {
				if (parse_chip_object(core, object, nb_board, n) < 0) {
					error("Failed to parse chip object");
					return -1;
				}
			}
		}
	}

	return 0;
}

static int parse_lbt(struct core *core, JSON_Object *root, size_t nb_board)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t n, m, l;

	if ((object = json_object_get_object(root, "lbt_conf")) != NULL) {
		debug("Parsing object: \"SX1301_array_conf[%zu].lbt_conf\"", nb_board);

		JSON_BOOLEAN(object, "enable", board->lbt.enable);
		JSON_NUMBER(object, "rssi_target", board->lbt.rssi_target);

		if ((array = json_object_get_array(object, "chan_cfg")) != NULL) {
			debug("Parsing array: \"SX1301_array_conf[%zu].lbt_conf.chan_cfg\"", nb_board);

			for (n = 0, m = 0;
			     n < json_array_get_count(array) &&
			     n < sizeofa(board->lbt.channels);
			     n++) {
				if ((object = json_array_get_object(array, n)) != NULL) {
					debug("Parsing object: \"SX1301_array_conf[%zu].lbt_conf.chan_cfg[%zu]\"", nb_board, n);

					l = JSON_NUMBER(object, "freq_hz", board->lbt.channels[n].freq_hz);
					l += JSON_NUMBER(object, "scan_time_us", board->lbt.channels[n].scan_time_us);
					l += JSON_NUMBER(object, "transmit_time_ms", board->lbt.channels[n].transmit_time_ms);

					if (l == 3) {
						if (board->lbt.channels[n].scan_time_us != 128 &&
						    board->lbt.channels[n].scan_time_us != 5000) {
							error("Invalid scan_time_us: %u", board->lbt.channels[n].scan_time_us);
							return -1;
						}

						board->lbt.nb_channel = ++m;
					}
				}
			}
		}
	}

	return 0;
}

static int parse_board(struct core *core, JSON_Object *root, size_t nb_board)
{
	struct config_core *config_core = core->config->core;
	struct config_board *board = &config_core->boards[nb_board];
	const char *fsk_sync = NULL;
	const char *aes_key = NULL;

	debug("Parsing object: \"SX1301_array_conf[%zu]\"", nb_board);

	JSON_BOOLEAN(root, "full_duplex", board->board.full_duplex);
	JSON_BOOLEAN(root, "loramac_public", board->board.loramac_public);

	JSON_NUMBER(root, "board_rx_bw", board->board.rx_bw_hz);
	JSON_NUMBER(root, "board_rx_freq", board->board.rx_freq_hz);

	JSON_STRING(root, "FSK_sync", fsk_sync);
	if (fsk_sync) {
		size_t fsk_sync_len = strlen(fsk_sync);

		if (fsk_sync_len < 2 || fsk_sync_len > 16 || fsk_sync_len % 2) {
			error("Invalid format for value \"FSK_sync\": \"%s\"", fsk_sync);
			return -1;
		}

		if (sscanf(fsk_sync, "%llx", &board->board.fsk_sync_word) < 1) {
			error("Failed to parse value \"FSK_sync\": \"%s\"", fsk_sync);
			return -1;
		}

		board->board.fsk_sync_size = fsk_sync_len / 2;
	}

	JSON_NUMBER(root, "calibration_temperature_celsius_room", board->board.room_temp_ref);
	JSON_NUMBER(root, "calibration_temperature_code_ad9361", board->board.ad9361_temp_ref);

	JSON_NUMBER(root, "dsp_stat_interval", board->board.dsp_stat_interval);
	JSON_NUMBER(root, "nb_dsp", board->board.nb_dsp);

	JSON_STRING(root, "aes_key", aes_key);
	if (aes_key) {
		size_t aes_key_len = strlen(aes_key);

		if (aes_key_len != 32) {
			error("Invalid format for value \"aes_key\": \"%s\"", aes_key);
			return -1;
		}

		if (sscanf(aes_key, "%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx"
				    "%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx",
			   &board->board.aes_key[0],
			   &board->board.aes_key[1],
			   &board->board.aes_key[2],
			   &board->board.aes_key[3],
			   &board->board.aes_key[4],
			   &board->board.aes_key[5],
			   &board->board.aes_key[6],
			   &board->board.aes_key[7],
			   &board->board.aes_key[8],
			   &board->board.aes_key[9],
			   &board->board.aes_key[10],
			   &board->board.aes_key[11],
			   &board->board.aes_key[12],
			   &board->board.aes_key[13],
			   &board->board.aes_key[14],
			   &board->board.aes_key[15]) < 16) {
			error("Failed to parse value \"aes_key\": \"%s\"", aes_key);
			return -1;
		}
	}

	JSON_BOOLEAN(root, "fine_timestamp_decrypt_enable", board->board.fine_timestamp_decrypt_enable);
	JSON_BOOLEAN(root, "fts_debug", board->board.debug_tmst);
	JSON_BOOLEAN(root, "fts_match_crc_error", board->board.match_tmst_crc_err);
	JSON_NUMBER(root, "fts_version", board->board.main_tmst_version);

	if (parse_rfchain(core, root, nb_board) < 0) {
		error("Failed to parse rfchain");
		return -1;
	}

	if (parse_chip(core, root, nb_board) < 0) {
		error("Failed to parse chip");
		return -1;
	}

	if (parse_lbt(core, root, nb_board) < 0) {
		error("Failed to parse lbt");
		return -1;
	}

	return 0;
}

static int parse_boards(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t nb_board;

	if ((array = json_object_get_array(root, "SX1301_array_conf")) != NULL) {
		debug("Parsing array: \"SX1301_array_conf\"");
		for (nb_board = 0;
		     nb_board < json_array_get_count(array) &&
		     nb_board < config_core->nb_slots;
		     nb_board++) {

			if ((object = json_array_get_object(array, nb_board)) == NULL) {
				return -1;
			}

			if (parse_board(core, object, nb_board) < 0) {
				return -1;
			}
		}
	}

	return 0;
}

struct config_core *config_core_init(void)
{
	struct config_core *config_core = NULL;
	size_t nb_board, nb_chip, nb_chan;

	if ((config_core = zalloc(sizeof(*config_core))) == NULL) {
		eerror("Failed to allocate core configuration");
		return NULL;
	}

	for (nb_board = 0; nb_board < sizeofa(config_core->boards); nb_board++) {
		struct config_board *board = &config_core->boards[nb_board];
		board->board = sx1301ar_init_board_cfg();

		// All Kerlink boards are MASTER
		board->board.board_type = BRD_MASTER;

		// Default FSK sync word (default)
		board->board.fsk_sync_size = 3;
		board->board.fsk_sync_word = 0xC194C1;

		// All Kerlink boards have 1 DSP (default)
		board->board.nb_dsp = 1;

		// All Kerlink boards have 2 chips (default)
		// The first one is configured as "gps latched" timer chip
		// The second one is configured as "free run" timer chip
		board->board.nb_chip = 2;

		board->lbt = sx1301ar_init_lbt_cfg();

		for (nb_chip = 0; nb_chip < sizeofa(board->chips); nb_chip++) {
			struct config_chip *chip = &board->chips[nb_chip];
			chip->chip = sx1301ar_init_chip_cfg();

			// All Kerlink chips are enabled
			chip->chip.enable = true;
			chip->chip.freq_hz = SX1301AR_MIN_FREQ;

			for (nb_chan = 0; nb_chan < sizeofa(chip->chans); nb_chan++) {
				chip->chans[nb_chan] = sx1301ar_init_chan_cfg();
			}
		}
	}

	if (init_slot(config_core) < 0) {
		error("Failed to initialize slots");
		free(config_core);
		return NULL;
	}

	return config_core;
}

int config_core_parse(struct core *core, JSON_Object *root)
{
	if (parse_boards(core, root) < 0) {
		error("Failed to parse boards configuration");
		return -1;
	}

	return 0;
}

#define LBT_RSSI_SHIFT_OFFSET 336

int config_core_apply(struct core *core)
{
	struct config_core *config_core = core->config->core;
	size_t nb_board, nb_chip, nb_chan, nb_rfchain;

	for (nb_board = 0, config_core->nb_boards = 0;
	     nb_board < config_core->nb_slots;
	     nb_board++) {

		if (config_core->boards[nb_board].enable) {
			config_core->nb_boards++;
		} else {
			break;
		}
	}

	for (nb_board = 0; nb_board < config_core->nb_boards; nb_board++) {
		struct config_board *board = &config_core->boards[nb_board];

		for (nb_rfchain = 0;
		     nb_rfchain < sizeofa(board->board.rf_chain);
		     nb_rfchain++) {
			board->board.rf_chain[nb_rfchain].rssi_offset -= board->board.rf_chain[nb_rfchain].antenna_gain;
			board->board.rf_chain[nb_rfchain].rssi_offset += board->board.rf_chain[nb_rfchain].insertion_loss;
		}

		if (sx1301ar_conf_board(nb_board, &board->board) < 0) {
			lerror("board #%zu: Failed to configure board", nb_board);
			return -1;
		}

		board->lbt.rssi_shift = LBT_RSSI_SHIFT_OFFSET + board->board.rf_chain[0].rssi_offset;

		if (sx1301ar_conf_lbt(nb_board, &board->lbt) < 0) {
			lerror("board #%zu: Failed to configure LBT", nb_board);
			return -1;
		}

		for (nb_chip = 0; nb_chip < sizeofa(board->chips); nb_chip++) {
			struct config_chip *chip = &board->chips[nb_chip];

			if (sx1301ar_conf_chip(nb_board, nb_chip, &chip->chip) < 0) {
				lerror("board #%zu: Failed to configure chip %zu", nb_board, nb_chip);
				return -1;
			}

			for (nb_chan = 0; nb_chan < sizeofa(chip->chans); nb_chan++) {
				uint8_t idx_chan = (nb_chip << 4) + nb_chan;

				if (sx1301ar_conf_chan(nb_board, idx_chan, &chip->chans[nb_chan]) < 0) {
					lerror("board #%zu: Failed to configure channel %u", nb_board, idx_chan);
					return -1;
				}
			}
		}
	}

	return 0;
}

void config_core_free(struct core *core)
{
	size_t nb_board;

	if (core && core->config) {
		for (nb_board = 0; nb_board < core->config->core->nb_boards; nb_board++) {
			struct config_board *board = &core->config->core->boards[nb_board];

			spi_linuxdev_close(board->spi_fd);
			pthread_mutex_destroy(&board->spi_mutex);
		}

		free(core->config->core);
	}
}
