// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "core.h"

#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

#include "lib/sx1301ar/aux.h"
#include "lib/sx1301ar/err.h"
#include "lib/sx1301ar/gps.h"
#include "lib/sx1301ar/hal.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "beacon.h"
#include "clock.h"
#include "common/statistics.h"
#include "common/temperature.h"
#include "config.h"
#include "core_fscan.h"
#include "jit.h"

#define lwarning(fmt, args...) warning(fmt ": %s", ##args, sx1301ar_err_message(sx1301ar_errno))
#define lerror(fmt, args...) error(fmt ": %s", ##args, sx1301ar_err_message(sx1301ar_errno))

static jit_queue_t jit_queue;

#define FETCH_SLEEP_MS 10

#define LIBRARY_VERSION "5.1.0"

static void *thread_uplink(void *arg)
{
	struct core *core = arg;
	sx1301ar_rx_pkt_t rxpkt[SX1301AR_MAX_PKT_NB];
	size_t nb_board, nb_packets;
	uint8_t rxpkt_nb;

	for (nb_board = 0, nb_packets = 0; core->running; nb_board++) {
		if (nb_board >= core->config->core->nb_boards) {
			if (nb_packets == 0) {
				sx1301ar_wait_ms(FETCH_SLEEP_MS);
			}

			nb_board = 0;
			nb_packets = 0;
		}

		if (sx1301ar_fetch(nb_board, rxpkt, sizeofa(rxpkt), &rxpkt_nb) < 0) {
			lerror("Failed to receive messages");
			continue;

		} else if (rxpkt_nb == 0) {
			continue;

		} else {
			nb_packets += rxpkt_nb;
		}

		if (api_core_send(core, rxpkt, rxpkt_nb, nb_board) < 0) {
			if (api_running(core)) {
				error("Failed to send uplink message");
			}

		} else {
			info("Sent %u uplink message%s", rxpkt_nb, (rxpkt_nb > 1) ? "s" : "");
		}
	}

	pthread_exit(NULL);
}

static int check_power(struct core *core, size_t nb_board, int8_t power, uint8_t rf_chain)
{
	sx1301ar_board_cfg_t *board = &core->config->core->boards[nb_board].board;
	int index;

	for (index = 0; index < board->rf_chain[rf_chain].tx_lut.size; index++) {
		if (board->rf_chain[rf_chain].tx_lut.lut[index].rf_power <= power) {
			return 1;
		}
	}

	return 0;
}

static int check_frequency(struct core *core, size_t nb_board, uint32_t frequency, uint8_t rf_chain)
{
	struct config_board *config_board = &core->config->core->boards[nb_board];

	if (config_board->rfchains[rf_chain].tx_freq_min != 0 &&
	    frequency < config_board->rfchains[rf_chain].tx_freq_min) {
		return 0;
	}

	if (config_board->rfchains[rf_chain].tx_freq_max != 0 &&
	    frequency > config_board->rfchains[rf_chain].tx_freq_max) {
		return 0;
	}

	return 1;
}

static void *thread_downlink(void *arg)
{
	struct core *core = arg;
	jit_pkt_type_t downlink_type;
	sx1301ar_tx_pkt_t txpkt;
	size_t nb_board;
	jit_error_t jit;

	while (core->running) {
		if (api_core_recv(core, &txpkt, &downlink_type, &nb_board) < 0) {
			if (api_running(core)) {
				error("Failed to receive downlink message");
			}

			continue;
		}

		info("Received downlink message");
		core->statistics->downlink_requested++;

		/* Adjust packet RF power with antenna gain and insertion loss */
		debug("Board #%u: Requested power: %i dB", nb_board, txpkt.rf_power);
		txpkt.rf_power -= core->config->core->boards[nb_board].board.rf_chain[txpkt.rf_chain].antenna_gain;
		txpkt.rf_power += core->config->core->boards[nb_board].board.rf_chain[txpkt.rf_chain].insertion_loss;
		debug("Board #%u: Selected power: %i dB", nb_board, txpkt.rf_power);

		if (!check_power(core, nb_board, txpkt.rf_power, txpkt.rf_chain)) {
			error("Board #%u: Unsupported power: %i", nb_board, txpkt.rf_power);
			api_core_send_event(core, &txpkt, JIT_ERROR_TX_POWER);
			continue;
		}

		if (!check_frequency(core, nb_board, txpkt.freq_hz, txpkt.rf_chain)) {
			error("Board #%u: Unsupported frequency: %u", nb_board, txpkt.freq_hz);
			api_core_send_event(core, &txpkt, JIT_ERROR_TX_FREQ);
			continue;
		}

		if ((jit = jit_enqueue(core, &jit_queue, nb_board, &txpkt, downlink_type)) != JIT_ERROR_OK) {
			error("Failed to enqueue downlink message");
			api_core_send_event(core, &txpkt, jit);
			continue;
		}

		core->statistics->downlink_scheduled++;
	}

	pthread_exit(NULL);
}

static void *thread_jit(void *arg)
{
	struct core *core = arg;
	sx1301ar_tstat_t status;
	sx1301ar_tx_pkt_t txpkt;
	int result, index = -1;
	jit_pkt_type_t type;
	uint8_t nb_board;
	jit_error_t jit;

	while (core->running) {
		sx1301ar_wait_ms(10);

		if ((jit = jit_peek(core, &jit_queue, &index)) != JIT_ERROR_OK) {
			if (jit != JIT_ERROR_EMPTY) {
				error("Failed to peek message");
			}

			continue;

		} else if (index < 0) {
			continue;

		} else if (jit_dequeue(&jit_queue, index, &nb_board, &txpkt, &type) != JIT_ERROR_OK) {
			error("Failed to dequeue message");
			continue;

		} else if (sx1301ar_tx_status(nb_board, &status) < 0) {
			lerror("Failed to get concentrator status");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if (status == TX_EMITTING) {
			error("Concentrator is busy");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if (status == TX_SCHEDULED) {
			warning("Overwriting an already scheduled message");
		}

		if (clock_convert_counter(core, 0, nb_board,
					  txpkt.count_us, &txpkt.count_us) < 0) {
			error("Failed to convert counter");
			if (type != JIT_PKT_TYPE_BEACON) {
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
			}
			continue;

		} else if ((result = sx1301ar_send(nb_board, &txpkt)) < 0 &&
			   sx1301ar_errno == ERR_NO_TX_ENABLE &&
			   nb_board != 0 &&
			   type != JIT_PKT_TYPE_BEACON) {
			warning("TX disabled on board %u, trying with board 0...", nb_board);

			if (clock_convert_counter(core, nb_board, 0,
						  txpkt.count_us, &txpkt.count_us) < 0) {
				error("Failed to convert counter");
				api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
				continue;
			}

			result = sx1301ar_send(0, &txpkt);
		}

		if (result < 0) {
			lerror("Failed to send message");

			if (type != JIT_PKT_TYPE_BEACON) {
				if (sx1301ar_errno == ERR_LBT_FORBIDDEN) {
					api_core_send_event(core, &txpkt, HAL_SEND_LBT);

				} else {
					api_core_send_event(core, &txpkt, HAL_SEND_FAIL);
				}
			}

		} else if (type == JIT_PKT_TYPE_BEACON) {
			info("Sent beacon");

		} else {
			info("Sent downlink message");
			api_core_send_event(core, &txpkt, HAL_SEND_OK);
			core->statistics->downlink_transmitted++;
			core->statistics->downlink_bytes_transmitted += txpkt.size;
		}
	}

	pthread_exit(NULL);
}

static void *thread_beacon(void *arg)
{
	struct core *core = arg;
	struct config_beacon *config_beacon = &core->config->beacon;

	if (config_beacon->enable &&
	    config_beacon->period > 0 &&
	    config_beacon->frequencies > 0) {
		while (core->running) {
			if (core->clock->pps_enabled &&
			    core->clock->tai_offset != 0 &&
			    beacon_core_schedule(core, &jit_queue) < 0) {
				error("Failed to schedule the next beacon");
			}

			sleep_sec(3, core->running);
		}
	}

	pthread_exit(NULL);
}

static struct termios ttyopt_restore;

static int gps_init(const char *gsp_device)
{
	struct termios ttyopt;
	int fd;

	if ((fd = open(gsp_device, O_RDWR | O_NOCTTY)) < 0) {
		eerror("Failed to open GPS device: %s", gsp_device);
		return -1;

	} else if (tcgetattr(fd, &ttyopt_restore) < 0) {
		eerror("Failed to get GPS serial parameters");
		close(fd);
		return -1;

	} else {
		memcpy(&ttyopt, &ttyopt_restore, sizeof(ttyopt));
	}

	/* The following configuration should allow to:
	 *   - Get ASCII NMEA messages
	 *   - Get UBX binary messages
	 *   - Send UBX binary commands
	 * Note: as binary data have to be read/written, we need to disable
	 * various character processing to avoid loosing data */

	/* Control Modes */
	ttyopt.c_cflag |= CLOCAL;  /* local connection, no modem control */
	ttyopt.c_cflag |= CREAD;   /* enable receiving characters */
	ttyopt.c_cflag |= CS8;     /* 8 bit frames */
	ttyopt.c_cflag &= ~PARENB; /* no parity */
	ttyopt.c_cflag &= ~CSTOPB; /* one stop bit */
	/* Input Modes */
	ttyopt.c_iflag |= IGNPAR; /* ignore bytes with parity errors */
	ttyopt.c_iflag &= ~ICRNL; /* do not map CR to NL on input*/
	ttyopt.c_iflag &= ~IGNCR; /* do not ignore carriage return on input */
	ttyopt.c_iflag &= ~IXON;  /* disable Start/Stop output control */
	ttyopt.c_iflag &= ~IXOFF; /* do not send Start/Stop characters */
	/* Output Modes */
	ttyopt.c_oflag = 0; /* disable everything on output as we only write binary */
	/* Local Modes */
	ttyopt.c_lflag &= ~ICANON; /* disable canonical input - cannot use with binary input */
	ttyopt.c_lflag &= ~ISIG;   /* disable check for INTR, QUIT, SUSP special characters */
	ttyopt.c_lflag &= ~IEXTEN; /* disable any special control character */
	ttyopt.c_lflag &= ~ECHO;   /* do not echo back every character typed */
	ttyopt.c_lflag &= ~ECHOE;  /* does not erase the last character in current line */
	ttyopt.c_lflag &= ~ECHOK;  /* do not echo NL after KILL character */

	/* Settings for non-canonical mode:
	 * read will block for until the lesser of VMIN or requested chars have been received */
	ttyopt.c_cc[VMIN] = 1; /* byte granularity at init time to ensure we get all UBX responses */
	ttyopt.c_cc[VTIME] = 0;

	if (tcsetattr(fd, TCSANOW, &ttyopt) < 0) {
		eerror("Failed to update GPS serial parameters");
		close(fd);
		return -1;
	}

	/* Display message and flush */
	info("Using GPS device: %s", gsp_device);
	tcflush(fd, TCIOFLUSH);

	return fd;
}

static ssize_t read_nmea_sentence(int fd, char *buffer, size_t size)
{
	size_t length;
	ssize_t len;
	int started;
	char byte;

	for (started = 0, length = 0; length < size;) {
		if ((len = read(fd, &byte, sizeof(byte))) < 0) {
			eerror("Failed to read serial");
			return -1;
		}

		if (len == 0) {
			continue;
		}

		if (started == 1) {
			if (byte == '\r' || byte == '\n') {
				buffer[length] = '\0';
				return length;
			}

			buffer[length++] = byte;

		} else if (byte == '$') {
			started = 1;
			buffer[length++] = byte;
		}
	}

	return -1;
}

static void *thread_gps(void *arg)
{
	struct core *core = arg;
	ssize_t length;
	char buffer[256];
	sx1301ar_coord_t coord;
	sx1301ar_nmea_msg_t nmea;

	while (core->running) {
		if ((length = read_nmea_sentence(core->gps_fd, buffer, sizeof(buffer))) < 0) {
			eerror("Failed to read NMEA sentence");
			continue;
		}

		if (sx1301ar_get_nmea_type(buffer, length, &nmea) < 0) {
			lerror("Failed to get NMEA type");
			continue;
		}

		if (nmea == NMEA_GGA) {
			if (sx1301ar_parse_gga(buffer, length, &coord, NULL, NULL) != 0) {
				continue;
			}

			core->statistics->gps_latitude = coord.lat;
			core->statistics->gps_longitude = coord.lon;
			core->statistics->gps_altitude = coord.alt;
			core->statistics->has_gps_coordinates = 1;
		}
	}

	pthread_exit(NULL);
}

void fill_version_statistics(struct statistics *statistics)
{
	int16_t version_fpga, version_dsp;

	strncpy(statistics->version_hal, LIBRARY_VERSION, 32);
	statistics->has_version_hal = 1;
	sx1301ar_version_info(0, &version_fpga, &version_dsp);
	if (version_fpga != -1) {
		statistics->version_fpga = version_fpga;
		statistics->has_version_fpga = 1;
	}
	if (version_dsp != -1) {
		statistics->version_dsp = version_dsp;
		statistics->has_version_dsp = 1;
	}

	debug("FPGA version: %d ; DSP version: %d ; HAL version: %s",
	      statistics->has_version_fpga ? statistics->version_fpga : 0,
	      statistics->has_version_dsp ? statistics->version_dsp : 0,
	      statistics->has_version_hal ? statistics->version_hal : "");
}

struct core *core_init(struct stringlist *filenames, char *gps_device, uint32_t serial_number)
{
	struct core *core = NULL;
	size_t nb;

	if ((core = zalloc(sizeof(*core))) == NULL) {
		eerror("Failed to allocate core");
		return NULL;
	}

	if ((core->config = config_init()) == NULL) {
		error("Failed to initialize configuration");
		core_free(core);
		return NULL;
	}

	core->config->beacon.gwid = serial_number & 0x00FFFFFF;

	if ((core->api = api_init()) == NULL) {
		error("Failed to initialize API");
		core_free(core);
		return NULL;
	}

	if ((core->clock = clock_init()) == NULL) {
		error("Failed to initialize clock");
		core_free(core);
		return NULL;
	}

	if ((core->fscan = fscan_init()) == NULL) {
		error("Failed to initialize fscan");
		core_free(core);
		return NULL;
	}

	if ((core->temperature = temperature_init()) == NULL) {
		error("Failed to initialize temperature");
		core_free(core);
		return NULL;
	}

	if ((core->statistics = zalloc(sizeof(*core->statistics))) == NULL) {
		eerror("Failed to allocate statistic");
		core_free(core);
		return NULL;
	}

	for (nb = 0; filenames && nb < filenames->nb_strings; nb++) {
		char *filename = filenames->strings[nb];

		info("Parsing configuration file: %s", filename);

		if (config_parse_file(core, filename)) {
			error("Failed to parse configuration file %s", filename);
			core_free(core);
			return NULL;
		}
	}

	if (gps_device != NULL) {
		core->gps_enabled = 1;

		if ((core->gps_fd = gps_init(gps_device)) < 0) {
			error("Failed to initialize GPS");
			core_free(core);
			return NULL;
		}
	}

	jit_queue_init(&jit_queue);

	return core;
}

int core_start(struct core *core)
{
	if (config_apply(core)) {
		error("Failed to apply configuration");
		return -1;
	}

	if (sx1301ar_start(core->config->core->nb_boards) < 0) {
		lerror("Failed to start concentrator");
		return -1;
	}

	core->started = 1;

	if (clock_start(core) < 0) {
		error("Failed to start clock");
		return -1;
	}

	if (fscan_start(core) < 0) {
		error("Failed to start fscan");
		return -1;
	}

	if (temperature_start(core) < 0) {
		error("Failed to start temperature");
		core_shutdown(core);
		return -1;
	}

	if (api_start(core) < 0) {
		error("Failed to start API");
		return -1;
	}

	core->running = 1;

	if (core->gps_enabled) {
		if (pthread_create(&core->gps_thread, NULL, thread_gps, core) != 0) {
			error("Failed to start gps thread");
			core_shutdown(core);
			return -1;
		}
	}

	if (pthread_create(&core->beacon_thread, NULL, thread_beacon, core) != 0) {
		error("Failed to start beacon thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->jit_thread, NULL, thread_jit, core) != 0) {
		error("Failed to start jit thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->downlink_thread, NULL, thread_downlink, core) != 0) {
		error("Failed to start downlink thread");
		core_shutdown(core);
		return -1;
	}

	if (pthread_create(&core->uplink_thread, NULL, thread_uplink, core) != 0) {
		error("Failed to start uplink thread");
		core_shutdown(core);
		return -1;
	}

	api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__STARTED);

	return 0;
}

int core_join(struct core *core)
{
	if (pthread_join(core->uplink_thread, NULL) != 0) {
		error("Failed to join uplink thread");
		return -1;
	}

	if (pthread_join(core->downlink_thread, NULL) != 0) {
		error("Failed to join downlink thread");
		return -1;
	}

	if (pthread_join(core->jit_thread, NULL) != 0) {
		error("Failed to join jit thread");
		return -1;
	}

	if (pthread_join(core->beacon_thread, NULL) != 0) {
		error("Failed to join beacon thread");
		return -1;
	}

	if (core->gps_enabled) {
		if (pthread_join(core->gps_thread, NULL) != 0) {
			error("Failed to join gps thread");
			return -1;
		}
	}

	if (api_join(core) < 0) {
		error("Failed to join API threads");
		return -1;
	}

	if (temperature_join(core) < 0) {
		error("Failed to join temperature threads");
		return -1;
	}

	if (fscan_join(core) < 0) {
		error("Failed to join fscan threads");
		return -1;
	}

	if (clock_join(core) < 0) {
		error("Failed to join clock threads");
		return -1;
	}

	return 0;
}

void core_shutdown(struct core *core)
{
	if (core) {
		core->running = 0;
		api_shutdown(core);
		temperature_shutdown(core);
		fscan_shutdown(core);
		clock_shutdown(core);
	}
}

void core_free(struct core *core)
{
	if (core) {
		if (core->gps_enabled) {
			if (tcsetattr(core->gps_fd, TCSANOW, &ttyopt_restore) < 0) {
				ewarning("Failed to restore GPS serial parameters");
			}

			if (close(core->gps_fd) < 0) {
				ewarning("Failed to close GPS device");
			}
		}

		if (core->started &&
		    sx1301ar_stop(core->config->core->nb_boards) < 0) {
			lwarning("Failed to stop concentrator");
		}

		temperature_free(core);
		fscan_free(core);
		clock_free(core);
		config_free(core);
		free(core->statistics);

		api_free(core);
		free(core);
	}
}
