// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"
#include "core_fscan.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "lib/sx1301ar/lbt.h"
#include "lib/sx1301ar/reg.h"

#define SNIFFER_NB_MAX 5
/* TODO: MGR: sniffer number should be dynamic depending on board number */

#define RSSI_MEASURE_NB 100
#define RSSI_RANGE 256
#define DEFAULT_RSSI_RATE_DIV 9 /* RSSI sampling rate = 125kHz/(div+1) */

#define FREQ_TO_FREQ_REG(freq, rx_freq_hz) (int32_t)(-((((float)freq - (float)rx_freq_hz) / 1e6) * 512) + 0.5); // TODO: MGR: check if + 0.5 is needed
#define FREQ_REG_TO_FREQ(freq_reg, rx_freq_hz) (((((float)freq_reg - 0.5) / 512) * 1e6) + rx_freq_hz)		// TODO: MGR: check if - 0.5 is needed

struct sniffer_desc {
	const uint32_t board_id;
};
static struct sniffer_desc sniffers[SNIFFER_NB_MAX] = {
    {.board_id = 0},
    {.board_id = 1},
    {.board_id = 2},
    {.board_id = 3}};

struct fscan_core *fscan_core_init(void)
{
	struct fscan_core *fscan_core = NULL;

	if ((fscan_core = zalloc(sizeof(*fscan_core))) == NULL) {
		eerror("Failed to allocate fscan core");
		return NULL;
	}

	return fscan_core;
}

int fscan_core_start(uint32_t receiver_id)
{
	uint32_t board_id;

	if (receiver_id >= SNIFFER_NB_MAX) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}
	board_id = sniffers[receiver_id].board_id;
	if (sx1301ar_i_lbt_is_enabled(board_id)) {
		notice("LBT is enabled on board %d, no frequency scan available", board_id);
		return -1;
	}

	return 0;
}

void fscan_core_shutdown(uint32_t receiver_id)
{
	/* Unused parameter */
	receiver_id = receiver_id;

	/* Nothing to shutdown for this platform */
}

void fscan_core_free(struct core *core)
{
	if (core && core->fscan && core->fscan->core) {
		free(core->fscan->core);
	}
}

int fscan_get_capabilities(struct core *core, struct sniffer_caps *caps)
{
	caps->sniffer_nb = core->config->core->nb_boards;

	return 0;
}

int fscan_do_scan(float *rssi, struct core *core, uint32_t receiver_id, uint32_t frequency)
{
	uint32_t board_id;
	int32_t if_freq;
	int32_t r;
	int32_t reg_val;
	uint8_t read_burst[RSSI_RANGE * 2];
	uint16_t rssi_histo;
	uint16_t rssi_nb_total = 0;
	uint32_t rssi_sum = 0;
	uint32_t nb_loop = 0;
	uint32_t rx_freq_hz;
	float rssi_offset;

	if (receiver_id >= core->fscan->sniffer_caps.sniffer_nb) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}
	board_id = sniffers[receiver_id].board_id;

	/* Configure RSSI histogram */
	if (sx1301ar_reg_brd_w(board_id, BREG_RSSI_HISTO_RATE_DIV, DEFAULT_RSSI_RATE_DIV) < 0) {
		error("Board %u: failed to set BREG_RSSI_HISTO_RATE_DIV register", board_id);
		return -1;
	}
	if (sx1301ar_reg_brd_w(board_id, BREG_RSSI_HISTO_LEN, RSSI_MEASURE_NB - 1) < 0) {
		error("Board %u: failed to set BREG_RSSI_HISTO_LEN register", board_id);
		return -1;
	}

	rx_freq_hz = core->config->core->boards[board_id].board.rx_freq_hz;
	/* The IF register is 13-bits. Returns if frequency is outside the range */
	if_freq = FREQ_TO_FREQ_REG(frequency, rx_freq_hz);
	if ((if_freq < -4095) || (if_freq > 4095)) {
		error("Frequency %d (%d)is out of range [%d;%d] ([%d;%d])",
		      frequency, if_freq,
		      (int)FREQ_REG_TO_FREQ(-4095, rx_freq_hz), (int)FREQ_REG_TO_FREQ(4095, rx_freq_hz), -4095, 4095);

		return -1;
	}
	if (sx1301ar_reg_brd_w(board_id, BREG_IF_FREQ_RSSI, if_freq) < 0) {
		error("Board %u: failed to set BREG_IF_FREQ_RSSI register", board_id);
		return -1;
	}

	/* Start histogram */
	if (sx1301ar_reg_brd_w(board_id, BREG_RSSI_HISTO_START, 1) < 0) {
		error("Board %u: failed to set BREG_RSSI_HISTO_START register", board_id);
		return -1;
	}

	/* Wait until rssi_pts have been processed */
	do {
		if (sx1301ar_reg_brd_r(board_id, BREG_RSSI_HISTO_VALID, &reg_val) != 0) {
			error("Failed to read histogram value");
		}
		nb_loop++;
		sleep_ms(10);
	} while ((reg_val != 1) && (nb_loop < 1000));

	/* Unset start bit */
	if (sx1301ar_reg_brd_w(board_id, BREG_RSSI_HISTO_START, 0) < 0) {
		error("Board %u: failed to set BREG_RSSI_HISTO_START register", board_id);
		return -1;
	}

	/* Check if we could get a valid RSSI histogram */
	if (reg_val == 0) {
		error("Failed to get valid RSSI histogram for frequency %d", frequency);
		return -1;
	}

	/* Read histogram */
	if (sx1301ar_reg_brd_w(board_id, BREG_RSSI_HISTO_RAM_ADDR, 0) < 0) {
		error("Board %u: failed to set BREG_RSSI_HISTO_RAM_ADDR register", board_id);
		return -1;
	}
	if (sx1301ar_reg_brd_rb(board_id, BREG_RSSI_HISTO_RAM_DATA, read_burst, RSSI_RANGE * 2) < 0) {
		error("Board %u: failed to set  BREG_RSSI_HISTO_RAM_DATAregister", board_id);
		return -1;
	}

	rssi_offset = -139 - 4 + core->config->core->boards[board_id].board.rf_chain[0].antenna_gain -
		      core->config->core->boards[board_id].board.rf_chain[0].insertion_loss;
	for (r = 0; r < RSSI_RANGE; r++) {
		rssi_histo = (uint16_t)read_burst[2 * r] | ((uint16_t)read_burst[(2 * r) + 1] << 8);
		rssi_nb_total += rssi_histo;
		rssi_sum += (rssi_histo * r);
	}
	*rssi = (((float)rssi_sum) / rssi_nb_total) + rssi_offset;

	return 0;
}

bool fscan_is_lbt_enabled(uint32_t receiver_id)
{
	uint32_t board_id;

	if (receiver_id >= SNIFFER_NB_MAX) {
		error("Invalid receiver id %u", receiver_id);
		return true;
	}

	board_id = sniffers[receiver_id].board_id;
	if (sx1301ar_i_lbt_is_enabled(board_id)) {
		return true;
	} else {
		return false;
	}
}
