/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    LoRa concentrator : Just In Time TX scheduling queue

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include "jit.h"

#include <assert.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>  /* printf, fprintf, snprintf, fopen, fputs */
#include <stdlib.h> /* qsort_r */
#include <string.h> /* memset, memcpy */

#include "lib/sx1301ar/aux.h"

#include "utils/logger.h"

#include "clock.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */
#define TX_START_DELAY 1500						 /* microseconds */
#define TX_MARGIN_DELAY 1000						 /* Packet overlap margin in microseconds */
#define TX_JIT_DELAY 30000						 /* Pre-delay to program packet for TX in microseconds */
#define TX_MAX_ADVANCE_DELAY ((JIT_NUM_BEACON_IN_QUEUE + 1) * 128 * 1E6) /* Maximum advance delay accepted for a TX packet, compared to current time */

#define BEACON_GUARD 3000000    /* Interval where no ping slot can be placed, \
				    to ensure beacon can be sent */
#define BEACON_RESERVED 2120000 /* Time on air of the beacon, with some margin */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES (GLOBAL) ------------------------------------------- */
static pthread_mutex_t mx_jit_queue = PTHREAD_MUTEX_INITIALIZER; /* control access to JIT queue */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ----------------------------------------- */

bool jit_queue_is_full(jit_queue_t *queue)
{
	bool result;

	pthread_mutex_lock(&mx_jit_queue);

	result = (queue->num_pkt == JIT_QUEUE_MAX) ? true : false;

	pthread_mutex_unlock(&mx_jit_queue);

	return result;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool jit_queue_is_empty(jit_queue_t *queue)
{
	bool result;

	pthread_mutex_lock(&mx_jit_queue);

	result = (queue->num_pkt == 0) ? true : false;

	pthread_mutex_unlock(&mx_jit_queue);

	return result;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void jit_queue_init(jit_queue_t *queue)
{
	int i;

	pthread_mutex_lock(&mx_jit_queue);

	memset(queue, 0, sizeof(*queue));
	for (i = 0; i < JIT_QUEUE_MAX; i++) {
		queue->nodes[i].pre_delay = 0;
		queue->nodes[i].post_delay = 0;
	}

	pthread_mutex_unlock(&mx_jit_queue);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int compare(const void *a, const void *b, void *arg)
{
	jit_node_t *p = (jit_node_t *)a;
	jit_node_t *q = (jit_node_t *)b;
	int *counter = (int *)arg;
	int p_count, q_count;

	p_count = p->pkt.count_us;
	q_count = q->pkt.count_us;

	if (p_count > q_count) {
		*counter = *counter + 1;
	}

	return (p_count - q_count);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void jit_sort_queue(jit_queue_t *queue)
{
	int counter = 0;

	if (queue->num_pkt == 0) {
		return;
	}

	qsort_r(queue->nodes, queue->num_pkt, sizeof(queue->nodes[0]), compare, &counter);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool jit_collision_test(uint32_t p1_count_us, uint32_t p1_pre_delay, uint32_t p1_post_delay, uint32_t p2_count_us, uint32_t p2_pre_delay, uint32_t p2_post_delay)
{
	if (((p1_count_us - p2_count_us) <= (p1_pre_delay + p2_post_delay + TX_MARGIN_DELAY)) ||
	    ((p2_count_us - p1_count_us) <= (p2_pre_delay + p1_post_delay + TX_MARGIN_DELAY))) {
		return true;
	} else {
		return false;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

jit_error_t jit_enqueue(struct core *core, jit_queue_t *queue, uint8_t brd, sx1301ar_tx_pkt_t *packet, jit_pkt_type_t pkt_type)
{
	int i = 0;
	uint32_t time_us;
	uint32_t packet_post_delay = 0;
	uint32_t packet_pre_delay = 0;
	uint32_t target_pre_delay = 0;
	jit_error_t err_collision;
	uint32_t asap_count_us;

	if (clock_get_current_counter(core, 0, &time_us) < 0) {
		return HAL_NOT_SYNCHRONIZED;
	}

	if (packet == NULL) {
		error("Invalid parameter");
		return JIT_ERROR_INVALID;
	}

	if (jit_queue_is_full(queue)) {
		error("Cannot enqueue packet, JIT queue is full");
		return JIT_ERROR_FULL;
	}

	/* Compute packet pre/post delays depending on packet's type */
	switch (pkt_type) {
	case JIT_PKT_TYPE_DOWNLINK_CLASS_A:
	case JIT_PKT_TYPE_DOWNLINK_CLASS_B:
	case JIT_PKT_TYPE_DOWNLINK_CLASS_C:
		packet_pre_delay = TX_START_DELAY + TX_JIT_DELAY;
		packet_post_delay = (uint32_t)(sx1301ar_tx_pkt_time_on_air(0, packet, NULL, NULL, NULL)) * 1000UL; /* in us */
		break;
	case JIT_PKT_TYPE_BEACON:
		/* As defined in LoRaWAN spec */
		packet_pre_delay = TX_START_DELAY + BEACON_GUARD + TX_JIT_DELAY;
		packet_post_delay = BEACON_RESERVED;
		break;
	default:
		break;
	}

	pthread_mutex_lock(&mx_jit_queue);

	/* An immediate downlink becomes a timestamped downlink "ASAP" */
	/* Set the packet count_us to the first available slot */
	if (pkt_type == JIT_PKT_TYPE_DOWNLINK_CLASS_C) {
		/* change tx_mode to timestamped */
		packet->tx_mode = TX_TIMESTAMPED;

		/* Search for the ASAP timestamp to be given to the packet */
		asap_count_us = time_us + 1E6; /* take one second margin */
		if (queue->num_pkt == 0) {
			/* If the jit queue is empty, we can insert this packet */
		} else {
			/* Else we can try to insert it:
                - ASAP meaning NOW + MARGIN
                - at the last index of the queue
                - between 2 downlinks in the queue
            */

			/* First, try if the ASAP time collides with an already enqueued downlink */
			for (i = 0; i < queue->num_pkt; i++) {
				if (jit_collision_test(asap_count_us, packet_pre_delay, packet_post_delay, queue->nodes[i].pkt.count_us, queue->nodes[i].pre_delay, queue->nodes[i].post_delay) == true) {
					debug("Cannot insert IMMEDIATE downlink at count_us=%u, collides with %u (index=%d)", asap_count_us, queue->nodes[i].pkt.count_us, i);
					break;
				}
			}
			if (i == queue->num_pkt) {
				/* No collision with ASAP time, we can insert it */
			} else {
				/* Search for the best slot then */
				for (i = 0; i < queue->num_pkt; i++) {
					asap_count_us = queue->nodes[i].pkt.count_us + queue->nodes[i].post_delay + packet_pre_delay + TX_JIT_DELAY + TX_MARGIN_DELAY;
					if (i == (queue->num_pkt - 1)) {
						/* Last packet index, we can insert after this one */
					} else {
						/* Check if packet can be inserted between this index and the next one */
						if (jit_collision_test(asap_count_us, packet_pre_delay, packet_post_delay, queue->nodes[i + 1].pkt.count_us, queue->nodes[i + 1].pre_delay, queue->nodes[i + 1].post_delay) == true) {
							warning("Failed to insert IMMEDIATE downlink (count_us=%u), continue...", asap_count_us);
							continue;
						} else {
							break;
						}
					}
				}
			}
		}
		/* Set packet with ASAP timestamp */
		packet->count_us = asap_count_us;
	}

	/* Check criteria_1: is it already too late to send this packet ?
     *  The packet should arrive at least at (tmst - TX_START_DELAY) to be programmed into concentrator
     *  Note: - Also add some margin, to be checked how much is needed, if needed
     *        - Valid for both Downlinks and Beacon packets
     *
     *  Warning: unsigned arithmetic (handle roll-over)
     *      t_packet < t_current + TX_START_DELAY + MARGIN
     */
	if ((packet->count_us - time_us) <= (TX_START_DELAY + TX_MARGIN_DELAY + TX_JIT_DELAY)) {
		error("Packet REJECTED, already too late to send it (current=%u, packet=%u, type=%d)", time_us, packet->count_us, pkt_type);
		pthread_mutex_unlock(&mx_jit_queue);
		return JIT_ERROR_TOO_LATE;
	}

	/* Check criteria_2: Does packet timestamp seem plausible compared to current time
     *  We do not expect the server to program a downlink too early compared to current time
     *  Class A: downlink has to be sent in a 1s or 2s time window after RX
     *  Class B: downlink has to occur in a 128s time window
     *  Class C: no check needed, departure time has been calculated previously
     *  So let's define a safe delay above which we can say that the packet is out of bound: TX_MAX_ADVANCE_DELAY
     *  Note: - Valid for Downlinks only, not for Beacon packets
     *
     *  Warning: unsigned arithmetic (handle roll-over)
                t_packet > t_current + TX_MAX_ADVANCE_DELAY
     */
	if ((pkt_type == JIT_PKT_TYPE_DOWNLINK_CLASS_A) || (pkt_type == JIT_PKT_TYPE_DOWNLINK_CLASS_B)) {
		if ((packet->count_us - time_us) > TX_MAX_ADVANCE_DELAY) {
			error("Packet REJECTED, timestamp seems wrong, too much in advance (current=%u, packet=%u, type=%d)", time_us, packet->count_us, pkt_type);
			pthread_mutex_unlock(&mx_jit_queue);
			return JIT_ERROR_TOO_EARLY;
		}
	}

	/* Check criteria_3: does this new packet overlap with a packet already enqueued ?
     *  Note: - need to take into account packet's pre_delay and post_delay of each packet
     *        - Valid for both Downlinks and beacon packets
     *        - Beacon guard can be ignored if we try to queue a Class A downlink
     */
	for (i = 0; i < queue->num_pkt; i++) {
		/* We ignore Beacon Guard for Class A/C downlinks */
		if (((pkt_type == JIT_PKT_TYPE_DOWNLINK_CLASS_A) || (pkt_type == JIT_PKT_TYPE_DOWNLINK_CLASS_C)) && (queue->nodes[i].pkt_type == JIT_PKT_TYPE_BEACON)) {
			target_pre_delay = TX_START_DELAY;
		} else {
			target_pre_delay = queue->nodes[i].pre_delay;
		}

		/* Check if there is a collision
         *  Warning: unsigned arithmetic (handle roll-over)
         *      t_packet_new - pre_delay_packet_new < t_packet_prev + post_delay_packet_prev (OVERLAP on post delay)
         *      t_packet_new + post_delay_packet_new > t_packet_prev - pre_delay_packet_prev (OVERLAP on pre delay)
         */
		if (jit_collision_test(packet->count_us, packet_pre_delay, packet_post_delay, queue->nodes[i].pkt.count_us, target_pre_delay, queue->nodes[i].post_delay) == true) {
			switch (queue->nodes[i].pkt_type) {
			case JIT_PKT_TYPE_DOWNLINK_CLASS_A:
			case JIT_PKT_TYPE_DOWNLINK_CLASS_B:
			case JIT_PKT_TYPE_DOWNLINK_CLASS_C:
				error("Packet (type=%d) REJECTED, collision with packet already programmed at %u (%u)", pkt_type, queue->nodes[i].pkt.count_us, packet->count_us);
				err_collision = JIT_ERROR_COLLISION_PACKET;
				break;
			case JIT_PKT_TYPE_BEACON:
				if (pkt_type != JIT_PKT_TYPE_BEACON) {
					/* do not overload logs for beacon/beacon collision, as it is expected to happen with beacon pre-scheduling algorith used */
					error("Packet (type=%d) REJECTED, collision with beacon already programmed at %u (%u)", pkt_type, queue->nodes[i].pkt.count_us, packet->count_us);
				}
				err_collision = JIT_ERROR_COLLISION_BEACON;
				break;
			default:
				error("Unknown packet type, should not occur, BUG?");
				assert(0);
				break;
			}
			pthread_mutex_unlock(&mx_jit_queue);
			return err_collision;
		}
	}

	/* Finally enqueue it */
	/* Insert packet at the end of the queue */
	queue->nodes[queue->num_pkt].brd = brd;
	memcpy(&(queue->nodes[queue->num_pkt].pkt), packet, sizeof(sx1301ar_tx_pkt_t));
	queue->nodes[queue->num_pkt].pre_delay = packet_pre_delay;
	queue->nodes[queue->num_pkt].post_delay = packet_post_delay;
	queue->nodes[queue->num_pkt].pkt_type = pkt_type;
	if (pkt_type == JIT_PKT_TYPE_BEACON) {
		queue->num_beacon++;
	}
	queue->num_pkt++;
	/* Sort the queue in ascending order of packet timestamp */
	jit_sort_queue(queue);

	/* Done */
	pthread_mutex_unlock(&mx_jit_queue);

	return JIT_ERROR_OK;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

jit_error_t jit_dequeue(jit_queue_t *queue, int index, uint8_t *brd, sx1301ar_tx_pkt_t *packet, jit_pkt_type_t *pkt_type)
{
	if ((packet == NULL) || (brd == NULL) || (pkt_type == NULL)) {
		error("Invalid parameter");
		return JIT_ERROR_INVALID;
	}

	if ((index < 0) || (index >= JIT_QUEUE_MAX)) {
		error("Invalid parameter");
		return JIT_ERROR_INVALID;
	}

	if (jit_queue_is_empty(queue)) {
		error("Cannot dequeue packet, JIT queue is empty");
		return JIT_ERROR_EMPTY;
	}

	pthread_mutex_lock(&mx_jit_queue);

	/* Dequeue requested packet */
	*brd = queue->nodes[index].brd;
	memcpy(packet, &(queue->nodes[index].pkt), sizeof(sx1301ar_tx_pkt_t));
	queue->num_pkt--;
	*pkt_type = queue->nodes[index].pkt_type;
	if (*pkt_type == JIT_PKT_TYPE_BEACON) {
		queue->num_beacon--;
	}

	/* Replace dequeued packet with last packet of the queue */
	memcpy(&(queue->nodes[index]), &(queue->nodes[queue->num_pkt]), sizeof(jit_node_t));
	memset(&(queue->nodes[queue->num_pkt]), 0, sizeof(jit_node_t));

	/* Sort queue in ascending order of packet timestamp */
	jit_sort_queue(queue);

	/* Done */
	pthread_mutex_unlock(&mx_jit_queue);

	return JIT_ERROR_OK;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

jit_error_t jit_peek(struct core *core, jit_queue_t *queue, int *pkt_idx)
{
	/* Return index of node containing a packet inline with given time */
	int i = 0;
	int idx_highest_priority = -1;
	uint32_t time_us;

	if (pkt_idx == NULL) {
		error("Invalid parameter");
		return JIT_ERROR_INVALID;
	}

	if (jit_queue_is_empty(queue)) {
		return JIT_ERROR_EMPTY;
	}

	if (clock_get_current_counter(core, 0, &time_us) < 0) {
		return HAL_NOT_SYNCHRONIZED;
	}

	pthread_mutex_lock(&mx_jit_queue);

	/* Search for highest priority packet to be sent */
	for (i = 0; i < queue->num_pkt; i++) {
		/* First check if that packet is outdated:
         *  If a packet seems too much in advance, and was not rejected at enqueue time,
         *  it means that we missed it for peeking, we need to drop it
         *
         *  Warning: unsigned arithmetic
         *      t_packet > t_current + TX_MAX_ADVANCE_DELAY
         */
		if ((queue->nodes[i].pkt.count_us - time_us) >= TX_MAX_ADVANCE_DELAY) {
			/* We drop the packet to avoid lock-up */
			queue->num_pkt--;
			if (queue->nodes[i].pkt_type == JIT_PKT_TYPE_BEACON) {
				queue->num_beacon--;
				warning("--- Beacon dropped (current_time=%u, packet_time=%u) ---", time_us, queue->nodes[i].pkt.count_us);
			} else {
				warning("--- Packet dropped (current_time=%u, packet_time=%u) ---", time_us, queue->nodes[i].pkt.count_us);
			}

			/* Replace dropped packet with last packet of the queue */
			memcpy(&(queue->nodes[i]), &(queue->nodes[queue->num_pkt]), sizeof(jit_node_t));
			memset(&(queue->nodes[queue->num_pkt]), 0, sizeof(jit_node_t));

			/* Sort queue in ascending order of packet timestamp */
			jit_sort_queue(queue);

			/* restart loop  after purge to find packet to be sent */
			i = 0;
			continue;
		}

		/* Then look for highest priority packet to be sent:
         *  Warning: unsigned arithmetic (handle roll-over)
         *      t_packet < t_highest
         */
		if ((idx_highest_priority == -1) || (((queue->nodes[i].pkt.count_us - time_us) < (queue->nodes[idx_highest_priority].pkt.count_us - time_us)))) {
			idx_highest_priority = i;
		}
	}

	/* Peek criteria 1: look for a packet to be sent in next TX_JIT_DELAY ms timeframe
     *  Warning: unsigned arithmetic (handle roll-over)
     *      t_packet < t_current + TX_JIT_DELAY
     */
	if ((queue->nodes[idx_highest_priority].pkt.count_us - time_us) < TX_JIT_DELAY) {
		*pkt_idx = idx_highest_priority;
	} else {
		*pkt_idx = -1;
	}

	pthread_mutex_unlock(&mx_jit_queue);

	return JIT_ERROR_OK;
}
