// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "eeprom.h"

#include <assert.h>
#include <stdint.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

struct eeprom_93lc46b {
	struct eeprom_device dev;

	struct spi_device *spi;
};

#define REG_START 0x01
#define REG_ERAL 0x20
#define REG_EWDS 0x00
#define REG_EWEN 0x30
#define REG_READ 0x80
#define REG_WRITE 0x40

#define EEPROM_SIZE 128
#define REG_SIZE(type) (EEPROM_SIZE / sizeof(type))
#define REG_ADDR_MASK(type) (REG_SIZE(type) - 1)

static int command(struct eeprom_93lc46b *eeprom, uint8_t command)
{
	struct spi_msg msg[1];
	uint8_t hdr[2];

	assert(eeprom);

	memset(msg, 0, sizeof(msg));

	hdr[0] = REG_START;
	hdr[1] = command;

	msg[0].mosi = hdr;
	msg[0].len = sizeof(hdr);

	if (eeprom->spi->msg(eeprom->spi, msg, sizeofa(msg)) < 0) {
		return -1;
	}

	sleep_ns(250);

	return 0;
}

static int ready(struct eeprom_93lc46b *eeprom)
{
	struct spi_msg msg[1];
	uint8_t ready;
	int count;

	assert(eeprom);

	memset(msg, 0, sizeof(msg));

	msg[0].miso = &ready;
	msg[0].len = sizeof(ready);

	for (count = 0; count < 100; count++) {
		sleep_us(100);

		if (eeprom->spi->msg(eeprom->spi, msg, sizeofa(msg)) < 0) {
			return -1;
		}

		if (ready != 0) {
			return 0;
		}
	}

	error("Failed to check eeprom status: Timeout");

	return -1;
}

static int u16_read(struct eeprom_93lc46b *eeprom, size_t off, void *buf,
		    size_t len)
{
	struct spi_msg msg[2];
	uint8_t hdr[2];

	assert(eeprom);
	assert(off + len <= REG_SIZE(uint16_t));
	assert(buf && len);

	memset(msg, 0, sizeof(msg));

	hdr[0] = REG_START;
	hdr[1] = REG_READ | (off & REG_ADDR_MASK(uint16_t));

	msg[0].mosi = hdr;
	msg[0].len = sizeof(hdr);
	msg[0].mode = SPI_MSG_HEADER;
	msg[1].miso = buf;
	msg[1].len = len * sizeof(uint16_t);
	msg[1].mode = SPI_MODE_1;

	if (eeprom->spi->msg(eeprom->spi, msg, sizeofa(msg)) < 0) {
		return -1;
	}

	return 0;
}

static int u16_write(struct eeprom_93lc46b *eeprom, size_t off, const void *buf,
		     size_t len)
{
	struct spi_msg msg[2];
	uint8_t hdr[2];
	size_t n;

	assert(eeprom);
	assert(off + len <= REG_SIZE(uint16_t));
	assert(buf && len);

	for (n = 0; n < len; n++, off++, buf += sizeof(uint16_t)) {
		memset(msg, 0, sizeof(msg));

		hdr[0] = REG_START;
		hdr[1] = REG_WRITE | (off & REG_ADDR_MASK(uint16_t));

		msg[0].mosi = hdr;
		msg[0].len = sizeof(hdr);
		msg[0].mode = SPI_MSG_HEADER;
		msg[1].mosi = buf;
		msg[1].len = sizeof(uint16_t);

		if (eeprom->spi->msg(eeprom->spi, msg, sizeofa(msg)) < 0) {
			return -1;
		}

		sleep_ns(250);

		if (ready(eeprom) < 0) {
			return -1;
		}
	}

	return 0;
}

static int eeprom_93lc46b_erase(struct eeprom_device *dev)
{
	struct eeprom_93lc46b *eeprom;
	int ret;

	assert(dev);

	eeprom = container_of(dev, struct eeprom_93lc46b, dev);

	if (command(eeprom, REG_EWEN) < 0) {
		return -1;
	}

	if ((ret = command(eeprom, REG_ERAL)) < 0) {
		goto out;
	}

	if ((ret = ready(eeprom)) < 0) {
		goto out;
	}

out:
	command(eeprom, REG_EWDS);

	return ret;
}

// For the next functions we have:
//
//          off                         len
//  |------------------>|<----------------------------->|               buf
//  | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | A | B | C | D | E | F |
//  |   0   |   1   |   2   |   3   |   4   |   5   |   6   |   7   |
//  |-------------->|<------------------------------------->|       u16_buf
//      u16_off     |-->|           u16_len
//              u16_head_off

static int eeprom_93lc46b_read(struct eeprom_device *dev, size_t off, void *buf,
			       size_t len)
{
	struct eeprom_93lc46b *eeprom;

	uint8_t *u16_buf;
	size_t u16_off = off / sizeof(uint16_t);
	size_t u16_len = ((((off + len) - 1) / sizeof(uint16_t)) + 1) - u16_off;
	size_t u16_head_off = off - (u16_off * sizeof(uint16_t));

	assert(dev);
	assert(off + len <= REG_SIZE(*buf));
	assert(buf && len);

	eeprom = container_of(dev, struct eeprom_93lc46b, dev);

	if ((u16_buf = zalloc(u16_len, sizeof(uint16_t))) == NULL) {
		eerror("Failed to allocate memory");
		return -1;
	}

	if (u16_read(eeprom, u16_off, u16_buf, u16_len) < 0) {
		free(u16_buf);
		return -1;
	}

	memcpy(buf, u16_buf + u16_head_off, len);
	free(u16_buf);

	return 0;
}

static int eeprom_93lc46b_write(struct eeprom_device *dev, size_t off,
				const void *buf, size_t len)
{
	struct eeprom_93lc46b *eeprom;
	int ret;

	uint8_t *u16_buf;
	size_t u16_off = off / sizeof(uint16_t);
	size_t u16_len = ((((off + len) - 1) / sizeof(uint16_t)) + 1) - u16_off;
	size_t u16_head_off = off - (u16_off * sizeof(uint16_t));

	assert(dev);
	assert(off + len <= REG_SIZE(*buf));
	assert(buf && len);

	eeprom = container_of(dev, struct eeprom_93lc46b, dev);

	if ((u16_buf = zalloc(u16_len, sizeof(uint16_t))) == NULL) {
		eerror("Failed to allocate memory");
		return -1;
	}

	if (len < (u16_len * sizeof(uint16_t))) {
		if (u16_read(eeprom, u16_off, u16_buf, u16_len) < 0) {
			free(u16_buf);
			return -1;
		}
	}

	memcpy(u16_buf + u16_head_off, buf, len);

	if (command(eeprom, REG_EWEN) < 0) {
		free(u16_buf);
		return -1;
	}

	ret = u16_write(eeprom, u16_off, u16_buf, u16_len);

	command(eeprom, REG_EWDS);
	free(u16_buf);

	return ret;
}

static void eeprom_93lc46b_close(struct eeprom_device *dev)
{
	struct eeprom_93lc46b *eeprom;

	if (dev) {
		eeprom = container_of(dev, struct eeprom_93lc46b, dev);

		free(eeprom);
	}
}

struct eeprom_device *eeprom_open(struct spi_device *spi)
{
	struct eeprom_93lc46b *eeprom;

	assert(spi);

	if ((eeprom = zalloc(sizeof(*eeprom))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	eeprom->dev.close = eeprom_93lc46b_close;
	eeprom->dev.erase = eeprom_93lc46b_erase;
	eeprom->dev.read = eeprom_93lc46b_read;
	eeprom->dev.write = eeprom_93lc46b_write;

	eeprom->spi = spi;

	if (eeprom->spi->set_mode(eeprom->spi, SPI_MODE_0 | SPI_MSB_FIRST |
						   SPI_WORD_8BITS) < 0) {
		error("Failed to configure spi");
		eeprom_93lc46b_close(&eeprom->dev);
		return NULL;
	}

	return &eeprom->dev;
}
