// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "gpio.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

struct gpio_sysfs {
	struct gpio_device dev;

	int id;
	int direction;
	int value;
};

#define SYSFS_EXPORT "/sys/class/gpio/export"
#define SYSFS_UNEXPORT "/sys/class/gpio/unexport"
#define SYSFS_ROOT "/sys/class/gpio/gpio%u"
#define SYSFS_DIRECTION SYSFS_ROOT "/direction"
#define SYSFS_VALUE SYSFS_ROOT "/value"

static int gpio_get_value(struct gpio_device *dev)
{
	struct gpio_sysfs *gpio;
	char value[2];

	assert(dev);

	gpio = container_of(dev, struct gpio_sysfs, dev);

	if (lseek(gpio->value, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;
	}

	if (read(gpio->value, value, sizeof(value)) < 0) {
		eerror("Failed to read");
		return -1;
	}

	if (*value == '1') {
		return GPIO_HIGH;

	} else if (*value == '0') {
		return GPIO_LOW;
	}

	return -1;
}

static int gpio_set_value(struct gpio_device *dev, uint8_t mode)
{
	struct gpio_sysfs *gpio;

	assert(dev);

	gpio = container_of(dev, struct gpio_sysfs, dev);

	if (lseek(gpio->value, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;
	}

	if (write(gpio->value, (mode & GPIO_HIGH) ? "1" : "0", 1) < 0) {
		eerror("Failed to write");
		return -1;
	}

	return 0;
}

static int gpio_get_mode(struct gpio_device *dev)
{
	struct gpio_sysfs *gpio;
	char mode[3];

	assert(dev);

	gpio = container_of(dev, struct gpio_sysfs, dev);

	if (lseek(gpio->direction, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;
	}

	if (read(gpio->direction, mode, sizeof(mode)) < 0) {
		eerror("Failed to read");
		return -1;
	}

	if (strncmp(mode, "in", 2) == 0) {
		return GPIO_INPUT;

	} else if (strncmp(mode, "out", 3) == 0) {
		return GPIO_OUTPUT;
	}

	return -1;
}

static int gpio_set_mode(struct gpio_device *dev, uint8_t mode)
{
	struct gpio_sysfs *gpio;

	assert(dev);

	gpio = container_of(dev, struct gpio_sysfs, dev);

	if (lseek(gpio->direction, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;
	}

	if (write(gpio->direction, (mode & GPIO_OUTPUT) ? "out" : "in",
		  (mode & GPIO_OUTPUT) ? 3 : 2) < 0) {
		eerror("Failed to write");
		return -1;
	}

	if (mode & GPIO_OUTPUT && gpio_set_value(dev, mode) < 0) {
		return -1;
	}

	return 0;
}

static int is_exported(unsigned int id)
{
	char *path = NULL;
	int ret;

	if (aprintf(&path, SYSFS_ROOT, id) < 0) {
		return -1;
	}

	ret = access(path, F_OK);

	if (ret == 0) {
		free(path);
		return 1;
	}

	if (errno == ENOENT) {
		free(path);
		return 0;
	}

	eerror("Failed to access: %s", path);
	free(path);

	return -1;
}

static int export(unsigned int id)
{
	char *export_id = NULL;
	int exported, fd;
	ssize_t len;

	if ((exported = is_exported(id)) < 0) {
		return -1;
	}

	if (exported == 1) {
		return 0;
	}

	if ((fd = open(SYSFS_EXPORT, O_WRONLY)) < 0) {
		eerror("Failed to open: %s", SYSFS_EXPORT);
		return -1;
	}

	if ((len = aprintf(&export_id, "%u", id)) < 0) {
		close(fd);
		return -1;
	}

	if (write(fd, export_id, len) < 0) {
		eerror("Failed to write: %s", SYSFS_EXPORT);
		free(export_id);
		close(fd);
		return -1;
	}

	free(export_id);
	close(fd);

	return 0;
}

static int unexport(unsigned int id)
{
	char *unexport_id = NULL;
	int exported, fd;
	ssize_t len;

	if ((exported = is_exported(id)) < 0) {
		return -1;
	}

	if (exported == 0) {
		return 0;
	}

	if ((fd = open(SYSFS_UNEXPORT, O_WRONLY)) < 0) {
		eerror("Failed to open: %s", SYSFS_UNEXPORT);
		return -1;
	}

	if ((len = aprintf(&unexport_id, "%u", id)) < 0) {
		close(fd);
		return -1;
	}

	if (write(fd, unexport_id, len) < 0) {
		eerror("Failed to write: %s", SYSFS_UNEXPORT);
		free(unexport_id);
		close(fd);
		return -1;
	}

	free(unexport_id);
	close(fd);

	return 0;
}

static void gpio_close(struct gpio_device *dev)
{
	struct gpio_sysfs *gpio;

	if (dev) {
		gpio = container_of(dev, struct gpio_sysfs, dev);

		if (gpio->direction >= 0) {
			gpio_set_mode(dev, GPIO_INPUT);
			close(gpio->direction);
		}

		if (gpio->value >= 0) {
			close(gpio->value);
		}

		if (gpio->id >= 0) {
			unexport(gpio->id);
		}

		free(gpio);
	}
}

static int open_path(const char *format, unsigned int id)
{
	char *path = NULL;
	int fd;

	assert(format);

	if (aprintf(&path, format, id) < 0) {
		return -1;
	}

	if ((fd = open(path, O_RDWR)) < 0) {
		eerror("Failed to open: %s", path);
		free(path);
		return -1;
	}

	free(path);

	return fd;
}

struct gpio_device *gpio_open(unsigned int id, uint8_t mode)
{
	struct gpio_sysfs *gpio = NULL;

	if ((gpio = zalloc(sizeof(*gpio))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	gpio->dev.close = gpio_close;
	gpio->dev.get_mode = gpio_get_mode;
	gpio->dev.set_mode = gpio_set_mode;
	gpio->dev.get_value = gpio_get_value;
	gpio->dev.set_value = gpio_set_value;

	gpio->id = -1;
	gpio->direction = -1;
	gpio->value = -1;

	if (export(id) < 0) {
		gpio_close(&gpio->dev);
		return NULL;

	} else {
		gpio->id = id;
	}

	debug("Opening gpio device " SYSFS_ROOT, id);

	if ((gpio->direction = open_path(SYSFS_DIRECTION, id)) < 0) {
		gpio_close(&gpio->dev);
		return NULL;
	}

	if ((gpio->value = open_path(SYSFS_VALUE, id)) < 0) {
		gpio_close(&gpio->dev);
		return NULL;
	}

	if (gpio_set_mode(&gpio->dev, mode) < 0) {
		gpio_close(&gpio->dev);
		return NULL;
	}

	return &gpio->dev;
}
