// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "gpio.h"

#include <assert.h>
#include <fcntl.h>
#include <linux/kerlink_gpio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

struct gpio_wsta {
	struct gpio_device dev;

	int fd;
	uint8_t mode;
};

#define GPIO_PORT_0_MAX 8
#define GPIO_PORT_1_MAX GPIO_PORT_0_MAX + 24
#define GPIO_PORT_2_MAX GPIO_PORT_1_MAX + 13

#define GPIO_PORT(id)                                                          \
	(id < GPIO_PORT_0_MAX) ? 0 : (id < GPIO_PORT_1_MAX)                    \
					 ? 1                                   \
					 : (id < GPIO_PORT_2_MAX) ? 2 : 3

#define GPIO_NUMBER(id)                                                        \
	(id < GPIO_PORT_0_MAX) ? id : (id < GPIO_PORT_1_MAX)                   \
					  ? id - GPIO_PORT_0_MAX               \
					  : (id < GPIO_PORT_2_MAX)             \
						? id - GPIO_PORT_1_MAX         \
						: id - GPIO_PORT_2_MAX

#define GPIO_DEVICE(id) "/dev/p%1u_%02u", GPIO_PORT(id), GPIO_NUMBER(id)

static int gpio_get_value(struct gpio_device *dev)
{
	struct gpio_wsta *gpio;
	uint32_t value;

	assert(dev);

	gpio = container_of(dev, struct gpio_wsta, dev);

	if (ioctl(gpio->fd, GPIO_READ, &value) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	return (value) ? GPIO_HIGH : GPIO_LOW;
}

static int gpio_set_value(struct gpio_device *dev, uint8_t mode)
{
	struct gpio_wsta *gpio;

	assert(dev);

	gpio = container_of(dev, struct gpio_wsta, dev);

	if (ioctl(gpio->fd, (mode & GPIO_HIGH) ? GPIO_WRITE_HIGH_LEVEL
						: GPIO_WRITE_LOW_LEVEL,
		  NULL) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	return 0;
}

static int gpio_get_mode(struct gpio_device *dev)
{
	struct gpio_wsta *gpio;

	assert(dev);

	gpio = container_of(dev, struct gpio_wsta, dev);

	return gpio->mode & GPIO_OUTPUT;
}

static int gpio_set_mode(struct gpio_device *dev, uint8_t mode)
{
	struct gpio_wsta *gpio;

	assert(dev);

	gpio = container_of(dev, struct gpio_wsta, dev);

	if (mode & GPIO_OUTPUT) {
		if (gpio_set_value(dev, mode) < 0) {
			return -1;
		}

	} else {
		if (ioctl(gpio->fd, GPIO_SET_INPUT, NULL) < 0) {
			eerror("Failed to control device");
			return -1;
		}
	}

	gpio->mode = mode;

	return 0;
}

static void gpio_close(struct gpio_device *dev)
{
	struct gpio_wsta *gpio;

	if (dev) {
		gpio = container_of(dev, struct gpio_wsta, dev);

		if (gpio->fd >= 0) {
			gpio_set_mode(dev, GPIO_INPUT);
			close(gpio->fd);
		}

		free(gpio);
	}
}

struct gpio_device *gpio_open(unsigned int id, uint8_t mode)
{
	struct gpio_wsta *gpio = NULL;
	char *path = NULL;

	if ((gpio = zalloc(sizeof(*gpio))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	gpio->dev.close = gpio_close;
	gpio->dev.get_mode = gpio_get_mode;
	gpio->dev.set_mode = gpio_set_mode;
	gpio->dev.get_value = gpio_get_value;
	gpio->dev.set_value = gpio_set_value;

	gpio->fd = -1;

	if (aprintf(&path, GPIO_DEVICE(id)) < 0) {
		gpio_close(&gpio->dev);
		return NULL;
	}

	debug("Opening gpio device %s", path);

	if ((gpio->fd = open(path, O_RDWR | O_NONBLOCK)) < 0) {
		eerror("Failed to open: %s", path);
		gpio_close(&gpio->dev);
		return NULL;
	}

	if (gpio_set_mode(&gpio->dev, mode) < 0) {
		gpio_close(&gpio->dev);
		return NULL;
	}

	return &gpio->dev;
}
