// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "led.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

struct led_sysfs {
	struct led_device dev;

	int trigger;
	int shot;
};

#define SYSFS_ROOT "/sys/class/leds/%s"
#define SYSFS_TRIGGER SYSFS_ROOT "/trigger"
#define SYSFS_SHOT SYSFS_ROOT "/shot"

static int led_blink(struct led_device *dev)
{
	struct led_sysfs *led;

	assert(dev);

	led = container_of(dev, struct led_sysfs, dev);

	if (lseek(led->shot, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;
	}

	if (write(led->shot, "1", 1) < 0) {
		eerror("Failed to write");
		return -1;
	}

	return 0;
}

static void led_close(struct led_device *dev)
{
	struct led_sysfs *led;

	if (dev) {
		led = container_of(dev, struct led_sysfs, dev);

		if (led->trigger >= 0) {
			close(led->trigger);
		}

		if (led->shot >= 0) {
			close(led->shot);
		}

		free(led);
	}
}

static int open_path(const char *format, const char *name, int flags)
{
	char *path = NULL;
	int fd;

	assert(format);

	if (aprintf(&path, format, name) < 0) {
		return -1;
	}

	if ((fd = open(path, flags)) < 0) {
		eerror("Failed to open: %s", path);
		free(path);
		return -1;
	}

	free(path);

	return fd;
}

struct led_device *led_open(const char *name)
{
	struct led_sysfs *led = NULL;

	if ((led = zalloc(sizeof(*led))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	led->dev.close = led_close;
	led->dev.blink = led_blink;

	led->trigger = -1;
	led->shot = -1;

	debug("Opening led device " SYSFS_ROOT, name);

	if ((led->trigger = open_path(SYSFS_TRIGGER, name, O_RDWR)) < 0) {
		led_close(&led->dev);
		return NULL;
	}

	if (write(led->trigger, "oneshot", 7) < 0) {
		eerror("Failed to configure");
		led_close(&led->dev);
		return NULL;
	}

	if ((led->shot = open_path(SYSFS_SHOT, name, O_WRONLY)) < 0) {
		led_close(&led->dev);
		return NULL;
	}

	return &led->dev;
}
