// Copyright (c) 2018 Kerlink SA. All rights reserved.

// Include spidev.h first to remove name collisions
#include <linux/spi/spidev.h>
#undef SPI_CPHA
#undef SPI_CPOL
#undef SPI_MODE_0
#undef SPI_MODE_1
#undef SPI_MODE_2
#undef SPI_MODE_3
#undef SPI_LSB_FIRST

#include "spi.h"

#include <assert.h>
#include <fcntl.h>
#include <pthread.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

struct spi_spidev_imx {
	struct spi_device dev;

	pthread_mutex_t mutex;

	int fd;
	struct gpio_device *csn;

	uint32_t speed_hz;
	uint16_t mode;
};

#define SPI_BURST_SIZE 4096

static int get_speed(struct spi_spidev_imx *spi, uint32_t *speed_hz)
{
	assert(spi);
	assert(speed_hz);

	if (spi->speed_hz) {
		*speed_hz = spi->speed_hz;
		return 0;
	}

	if (ioctl(spi->fd, SPI_IOC_RD_MAX_SPEED_HZ, speed_hz) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	return 0;
}

static int spi_get_speed(struct spi_device *dev, uint32_t *speed_hz)
{
	struct spi_spidev_imx *spi;
	int ret;

	assert(dev);
	assert(speed_hz);

	spi = container_of(dev, struct spi_spidev_imx, dev);

	if (pthread_mutex_lock(&spi->mutex) != 0) {
		error("Failed to lock mutex");
		return -1;
	}

	ret = get_speed(spi, speed_hz);
	pthread_mutex_unlock(&spi->mutex);

	return ret;
}

static int set_speed(struct spi_spidev_imx *spi, uint32_t speed_hz)
{
	uint32_t dev_speed_hz;

	assert(spi);

	if (get_speed(spi, &dev_speed_hz) < 0) {
		error("Failed to get spi speed");
		return -1;
	}

	if (dev_speed_hz != speed_hz) {
		if (ioctl(spi->fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed_hz) < 0) {
			eerror("Failed to control device");
			return -1;
		}

		spi->speed_hz = speed_hz;
	}

	return 0;
}

static int spi_set_speed(struct spi_device *dev, uint32_t speed_hz)
{
	struct spi_spidev_imx *spi;
	int ret;

	assert(dev);

	spi = container_of(dev, struct spi_spidev_imx, dev);

	if (pthread_mutex_lock(&spi->mutex) != 0) {
		error("Failed to lock mutex");
		return -1;
	}

	ret = set_speed(spi, speed_hz);
	pthread_mutex_unlock(&spi->mutex);

	return ret;
}

static int get_mode(struct spi_spidev_imx *spi, uint16_t *mode)
{
	uint8_t val;

	assert(spi);
	assert(mode);

	if (spi->mode) {
		*mode = spi->mode;
		return 0;
	}

	*mode = 0;

	if (ioctl(spi->fd, SPI_IOC_RD_MODE, &val) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	if (val == 3) {
		*mode |= SPI_MODE_3;
	} else if (val == 2) {
		*mode |= SPI_MODE_2;
	} else if (val == 1) {
		*mode |= SPI_MODE_1;
	} else if (val == 0) {
		*mode |= SPI_MODE_0;
	}

	if (ioctl(spi->fd, SPI_IOC_RD_LSB_FIRST, &val) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	if (val == 1) {
		*mode |= SPI_LSB_FIRST;
	} else if (val == 0) {
		*mode |= SPI_MSB_FIRST;
	}

	if (ioctl(spi->fd, SPI_IOC_RD_BITS_PER_WORD, &val) < 0) {
		eerror("Failed to control device");
		return -1;
	}

	if (val <= 8) {
		*mode |= SPI_WORD_8BITS;
	} else if (val <= 16) {
		*mode |= SPI_WORD_16BITS;
	} else if (val <= 32) {
		*mode |= SPI_WORD_32BITS;
	}

	return 0;
}

static int spi_get_mode(struct spi_device *dev, uint16_t *mode)
{
	struct spi_spidev_imx *spi;
	int ret;

	assert(dev);
	assert(mode);

	spi = container_of(dev, struct spi_spidev_imx, dev);

	if (pthread_mutex_lock(&spi->mutex) != 0) {
		error("Failed to lock mutex");
		return -1;
	}

	ret = get_mode(spi, mode);
	pthread_mutex_unlock(&spi->mutex);

	return ret;
}

static int set_mode(struct spi_spidev_imx *spi, uint16_t mode)
{
	uint16_t dev_mode;
	uint8_t val;

	assert(spi);

	if (get_mode(spi, &dev_mode) < 0) {
		error("Failed to get spi mode");
		return -1;
	}

	if (mode & SPI_MODE_BIT &&
	    (dev_mode & SPI_MODE_MASK) != (mode & SPI_MODE_MASK)) {
		switch (mode & SPI_MODE_MASK) {
		case SPI_MODE_3:
			val = 3;
			break;
		case SPI_MODE_2:
			val = 2;
			break;
		case SPI_MODE_1:
			val = 1;
			break;
		default:
			val = 0;
			break;
		}

		if (ioctl(spi->fd, SPI_IOC_WR_MODE, &val) < 0) {
			eerror("Failed to control device");
			return -1;
		}

		spi->mode &= ~SPI_MODE_MASK;
		spi->mode |= mode & SPI_MODE_MASK;
	}

	if (mode & SPI_BYTES_ORDER_BIT &&
	    (dev_mode & SPI_BYTES_ORDER_MASK) !=
		(mode & SPI_BYTES_ORDER_MASK)) {
		switch (mode & SPI_BYTES_ORDER_MASK) {
		case SPI_LSB_FIRST:
			val = 1;
			break;
		default:
			val = 0;
			break;
		}

		if (ioctl(spi->fd, SPI_IOC_WR_LSB_FIRST, &val) < 0) {
			eerror("Failed to control device");
			return -1;
		}

		spi->mode &= ~SPI_BYTES_ORDER_MASK;
		spi->mode |= mode & SPI_BYTES_ORDER_MASK;
	}

	if (mode & SPI_WORD_BIT &&
	    (dev_mode & SPI_WORD_MASK) != (mode & SPI_WORD_MASK)) {
		switch (mode & SPI_WORD_MASK) {
		case SPI_WORD_32BITS:
			val = 32;
			break;
		case SPI_WORD_16BITS:
			val = 16;
			break;
		default:
			val = 8;
			break;
		}

		if (ioctl(spi->fd, SPI_IOC_WR_BITS_PER_WORD, &val) < 0) {
			eerror("Failed to control device");
			return -1;
		}

		spi->mode &= ~SPI_WORD_MASK;
		spi->mode |= mode & SPI_WORD_MASK;
	}

	return 0;
}

static int spi_set_mode(struct spi_device *dev, uint16_t mode)
{
	struct spi_spidev_imx *spi;
	int ret;

	assert(dev);

	spi = container_of(dev, struct spi_spidev_imx, dev);

	if (pthread_mutex_lock(&spi->mutex) != 0) {
		error("Failed to lock mutex");
		return -1;
	}

	ret = set_mode(spi, mode);
	pthread_mutex_unlock(&spi->mutex);

	return ret;
}

static int transfer(struct spi_spidev_imx *spi, struct spi_msg *msg, size_t len)
{
	struct spi_ioc_transfer xfer;
	int ret = 0;
	size_t n;

	uint16_t saved_mode, mode;
	int restore_mode;

	int keep_cs = 0;
	int delay_cs;

	assert(spi);
	assert(msg && len);

	if (spi->csn) {
		spi->csn->set_value(spi->csn, GPIO_LOW);
	}

	for (n = 1; n <= len; n++, msg++) {
		restore_mode = 0;
		delay_cs = 0;

		if (msg->len == 0) {
			continue;
		}

		memset(&xfer, 0, sizeof(xfer));

		xfer.tx_buf = (uintptr_t)msg->mosi;
		xfer.rx_buf = (uintptr_t)msg->miso;
		xfer.len = msg->len;
		xfer.speed_hz = msg->speed_hz;
		xfer.delay_usecs = msg->delay_us;

		if (msg->mode & SPI_WORD_BIT) {
			switch (msg->mode & SPI_WORD_MASK) {
			case SPI_WORD_32BITS:
				xfer.bits_per_word = 32;
				break;
			case SPI_WORD_16BITS:
				xfer.bits_per_word = 16;
				break;
			default:
				xfer.bits_per_word = 8;
				break;
			}
		}

		mode = msg->mode & (SPI_MODE_MASK | SPI_BYTES_ORDER_MASK);

		if (mode) {
			if ((ret = get_mode(spi, &saved_mode)) < 0) {
				error("Failed to backup spi mode");
				goto out;
			}

			restore_mode = 1;

			if ((ret = set_mode(spi, mode)) < 0) {
				error("Failed to set spi mode");
				goto out;
			}
		}

		if (msg->mode & SPI_CS_CHANGE) {
			if (n < len) {
				delay_cs = 10;

			} else {
				xfer.cs_change = 1;
				keep_cs = 1;
			}

		} else {
			if (n < len) {
				xfer.cs_change = 1;
			}
		}

		if (ioctl(spi->fd, SPI_IOC_MESSAGE(1), &xfer) <
		    (int)(xfer.len)) {
			eerror("Failed to control device");
			ret = -1;
			goto out;
		}

		if (delay_cs) {
			if (spi->csn) {
				spi->csn->set_value(spi->csn, GPIO_HIGH);
			}

			sleep_us(delay_cs);

			if (spi->csn) {
				spi->csn->set_value(spi->csn, GPIO_LOW);
			}
		}

		if (restore_mode) {
			if ((ret = set_mode(spi, saved_mode)) < 0) {
				error("Failed to restore spi mode");
				goto out;
			}

			restore_mode = 0;
		}
	}

out:
	if (ret < 0 || !keep_cs) {
		if (spi->csn) {
			spi->csn->set_value(spi->csn, GPIO_HIGH);
		}
	}

	if (restore_mode) {
		set_mode(spi, saved_mode);
	}

	return ret;
}

static int spi_msg(struct spi_device *dev, struct spi_msg *msg, size_t len)
{
	struct spi_spidev_imx *spi;
	struct spi_msg *smsg;
	size_t slen, sbytes;
	size_t hlen, hbytes;
	size_t dlen, dbytes;
	size_t abytes, rbytes;
	int ret;

	assert(dev);

	spi = container_of(dev, struct spi_spidev_imx, dev);

	if ((smsg = zalloc(len, sizeof(*smsg))) == NULL) {
		eerror("Failed to allocate memory");
		return -1;
	}

	for (hlen = 0, hbytes = 0; hlen < len - 1; hlen++) {
		if (msg[hlen].mode & SPI_MSG_HEADER) {
			smsg[hlen] = msg[hlen];
			hbytes += msg[hlen].len;

		} else {
			break;
		}
	}

	if (hbytes > SPI_BURST_SIZE) {
		error("Invalid spi headers: Too long");
		free(smsg);
		return -1;
	}

	if (pthread_mutex_lock(&spi->mutex) != 0) {
		error("Failed to lock mutex");
		free(smsg);
		return -1;
	}

	for (dlen = hlen, dbytes = 0, ret = 0; dlen < len && ret == 0;) {

		for (slen = hlen, sbytes = hbytes,
		    abytes = SPI_BURST_SIZE - sbytes;
		     dlen < len && abytes > 0; slen++) {

			smsg[slen] = msg[dlen];

			if (msg[dlen].mosi) {
				smsg[slen].mosi = msg[dlen].mosi + dbytes;
			}

			if (msg[dlen].miso) {
				smsg[slen].miso = msg[dlen].miso + dbytes;
			}

			rbytes = msg[dlen].len - dbytes;

			if (rbytes > abytes) {
				smsg[slen].mode &= ~SPI_CS_CHANGE;
				smsg[slen].len = abytes;
				sbytes += abytes;
				dbytes += abytes;
				abytes = 0;

			} else {
				smsg[slen].len = rbytes;
				sbytes += rbytes;
				dbytes = 0;
				abytes -= rbytes;
				dlen++;
			}
		}

		ret = transfer(spi, smsg, slen);
	}

	pthread_mutex_unlock(&spi->mutex);
	free(smsg);

	return ret;
}

static void spi_close(struct spi_device *dev)
{
	struct spi_spidev_imx *spi;

	if (dev) {
		spi = container_of(dev, struct spi_spidev_imx, dev);

		pthread_mutex_destroy(&spi->mutex);

		if (spi->csn) {
			spi->csn->set_value(spi->csn, GPIO_HIGH);
		}

		if (spi->fd >= 0) {
			close(spi->fd);
		}

		free(spi);
	}
}

struct spi_device *spi_open(const char *path, struct gpio_device *csn)
{
	struct spi_spidev_imx *spi = NULL;

	assert(path);

	debug("Opening spi device %s", path);

	if ((spi = zalloc(sizeof(*spi))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	spi->dev.close = spi_close;
	spi->dev.get_speed = spi_get_speed;
	spi->dev.set_speed = spi_set_speed;
	spi->dev.get_mode = spi_get_mode;
	spi->dev.set_mode = spi_set_mode;
	spi->dev.msg = spi_msg;

	spi->csn = csn;

	if (pthread_mutex_init(&spi->mutex, NULL) != 0) {
		error("Failed to initialize mutex");
		spi_close(&spi->dev);
		return NULL;
	}

	if ((spi->fd = open(path, O_RDWR)) < 0) {
		eerror("Failed to open: %s", path);
		spi_close(&spi->dev);
		return NULL;
	}

	if (spi_get_speed(&spi->dev, &spi->speed_hz) < 0) {
		eerror("Failed to get spi speed");
		spi_close(&spi->dev);
		return NULL;
	}

	if (spi_get_mode(&spi->dev, &spi->mode) < 0) {
		eerror("Failed to get spi mode");
		spi_close(&spi->dev);
		return NULL;
	}

	return &spi->dev;
}
