// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "spimux.h"

#include <assert.h>

#include "utils/logger.h"
#include "utils/memory.h"

struct spi_spimux {
	struct spi_device dev;

	struct spi_device *spi;
	uint8_t id;

	uint32_t speed_hz;
	uint16_t mode;
};

static int spi_spimux_get_speed(struct spi_device *dev, uint32_t *speed_hz)
{
	struct spi_spimux *spi;

	assert(dev);
	assert(speed_hz);

	spi = container_of(dev, struct spi_spimux, dev);

	if (spi->speed_hz) {
		*speed_hz = spi->speed_hz;
		return 0;
	}

	return spi->spi->get_speed(spi->spi, speed_hz);
}

static int spi_spimux_set_speed(struct spi_device *dev, uint32_t speed_hz)
{
	struct spi_spimux *spi;

	assert(dev);

	spi = container_of(dev, struct spi_spimux, dev);

	spi->speed_hz = speed_hz;

	return 0;
}

static int spi_spimux_get_mode(struct spi_device *dev, uint16_t *mode)
{
	struct spi_spimux *spi;
	uint16_t dev_mode;

	assert(dev);
	assert(mode);

	spi = container_of(dev, struct spi_spimux, dev);

	if (spi->spi->get_mode(spi->spi, &dev_mode) < 0) {
		return -1;
	}

	*mode = 0;

	if (spi->mode & SPI_MODE_BIT) {
		*mode |= spi->mode & SPI_MODE_MASK;
	} else {
		*mode |= dev_mode & SPI_MODE_MASK;
	}

	if (spi->mode & SPI_BYTES_ORDER_BIT) {
		*mode |= spi->mode & SPI_BYTES_ORDER_MASK;
	} else {
		*mode |= dev_mode & SPI_BYTES_ORDER_MASK;
	}

	if (spi->mode & SPI_WORD_BIT) {
		*mode |= spi->mode & SPI_WORD_MASK;
	} else {
		*mode |= dev_mode & SPI_WORD_MASK;
	}

	return 0;
}

static int spi_spimux_set_mode(struct spi_device *dev, uint16_t mode)
{
	struct spi_spimux *spi;

	assert(dev);

	spi = container_of(dev, struct spi_spimux, dev);

	spi->mode = mode;

	return 0;
}

static int spi_spimux_msg(struct spi_device *dev, struct spi_msg *msg,
			  size_t len)
{
	struct spi_msg *full_msg;
	struct spi_spimux *spi;
	size_t n;
	int ret;

	assert(dev);
	assert(msg && len);

	spi = container_of(dev, struct spi_spimux, dev);

	if ((full_msg = zalloc(len + 1, sizeof(*full_msg))) == NULL) {
		eerror("Failed to allocate memory");
		return -1;
	}

	full_msg[0].mosi = &spi->id;
	full_msg[0].len = sizeof(spi->id);
	full_msg[0].speed_hz = spi->speed_hz;
	full_msg[0].mode = spi->mode | SPI_MSG_HEADER;

	for (n = 0; n < len; n++) {
		full_msg[n + 1] = msg[n];

		if (msg[n].speed_hz) {
			full_msg[n + 1].speed_hz = msg[n].speed_hz;
		} else {
			full_msg[n + 1].speed_hz = spi->speed_hz;
		}

		full_msg[n + 1].mode = 0;

		if (msg[n].mode & SPI_MODE_BIT) {
			full_msg[n + 1].mode |= msg[n].mode & SPI_MODE_MASK;
		} else {
			full_msg[n + 1].mode |= spi->mode & SPI_MODE_MASK;
		}

		if (msg[n].mode & SPI_BYTES_ORDER_BIT) {
			full_msg[n + 1].mode |=
			    msg[n].mode & SPI_BYTES_ORDER_MASK;
		} else {
			full_msg[n + 1].mode |=
			    spi->mode & SPI_BYTES_ORDER_MASK;
		}

		if (msg[n].mode & SPI_WORD_BIT) {
			full_msg[n + 1].mode |= msg[n].mode & SPI_WORD_MASK;
		} else {
			full_msg[n + 1].mode |= spi->mode & SPI_WORD_MASK;
		}

		full_msg[n + 1].mode |= msg[n].mode & SPI_CS_CHANGE;
		full_msg[n + 1].mode |= msg[n].mode & SPI_MSG_HEADER;
	}

	ret = spi->spi->msg(spi->spi, full_msg, len + 1);

	free(full_msg);

	return ret;
}

static void spi_spimux_close(struct spi_device *dev)
{
	struct spi_spimux *spi;

	if (dev) {
		spi = container_of(dev, struct spi_spimux, dev);

		free(spi);
	}
}

struct spimux {
	struct spimux_device dev;

	struct spi_device *spi;
};

struct spi_device *spimux_get_spi(struct spimux_device *dev, uint8_t id)
{
	struct spimux *spimux;
	struct spi_spimux *spi = NULL;

	assert(dev);

	spimux = container_of(dev, struct spimux, dev);

	if ((spi = zalloc(sizeof(*spi))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	spi->dev.close = spi_spimux_close;
	spi->dev.set_speed = spi_spimux_set_speed;
	spi->dev.get_speed = spi_spimux_get_speed;
	spi->dev.set_mode = spi_spimux_set_mode;
	spi->dev.get_mode = spi_spimux_get_mode;
	spi->dev.msg = spi_spimux_msg;

	spi->spi = spimux->spi;
	spi->id = id;

	return &spi->dev;
}

static void spimux_close(struct spimux_device *dev)
{
	struct spimux *spimux;

	if (dev) {
		spimux = container_of(dev, struct spimux, dev);

		free(spimux);
	}
}

struct spimux_device *spimux_open(struct spi_device *spi)
{
	struct spimux *spimux;

	assert(spi);

	if ((spimux = zalloc(sizeof(*spimux))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	spimux->dev.close = spimux_close;
	spimux->dev.get_spi = spimux_get_spi;

	spimux->spi = spi;

	if (spimux->spi->set_mode(spimux->spi, SPI_MODE_0 | SPI_MSB_FIRST |
						   SPI_WORD_8BITS) < 0) {
		error("Failed to configure spi");
		spimux_close(&spimux->dev);
		return NULL;
	}

	return &spimux->dev;
}
