// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "sx1239.h"

#include <arpa/inet.h>
#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

struct sx1239 {
	struct sx1239_device dev;

	struct spi_device *spi;
	struct gpio_device *enable;
	struct gpio_device *reset;
};

#define REG_ADDR_MASK 0x7F
#define REG_READ 0x00
#define REG_WRITE 0x80

#define REG_OP_MODE 0x01
#define REG_DATA_MODUL 0x02
#define REG_BITRATE 0x03
#define REG_FRF 0x07
#define REG_LNA 0x18
#define REG_RX_BW 0x19
#define REG_RSSI_CONFIG 0x23
#define REG_RSSI_VALUE 0x24
#define REG_IRQ_FLAGS1 0x27
#define REG_TEST_LNA 0x58
#define REG_TEST_TCXO 0x59

#define REG_OP_MODE_MODE_MASK 0x1C
#define REG_OP_MODE_MODE_FS 0x08
#define REG_OP_MODE_MODE_RX 0x10

#define REG_DATA_MODUL_MODULATION_MASK 0x18
#define REG_DATA_MODUL_MODULATION_FSK 0x00
#define REG_DATA_MODUL_MODE_MASK 0x60
#define REG_DATA_MODUL_MODE_CONTINIOUS 0x60

#define REG_LNA_ZIN_MASK 0x80
#define REG_LNA_ZIN_200_OHMS 0x80
#define REG_LNA_GAIN_MASK 0x07
#define REG_LNA_GAIN_G1 0x01
#define REG_LNA_GAIN_G2 0x02
#define REG_LNA_GAIN_G3 0x03
#define REG_LNA_GAIN_G4 0x04
#define REG_LNA_GAIN_G5 0x05
#define REG_LNA_GAIN_G6 0x06

#define REG_RX_BW_DCC_FREQ_MASK 0xE0
#define REG_RX_BW_DCC_FREQ_16 0x00
#define REG_RX_BW_DCC_FREQ_8 0x20
#define REG_RX_BW_DCC_FREQ_4 0x40
#define REG_RX_BW_DCC_FREQ_2 0x60
#define REG_RX_BW_DCC_FREQ_1 0x80
#define REG_RX_BW_DCC_FREQ_1_2 0xA0
#define REG_RX_BW_DCC_FREQ_1_4 0xC0
#define REG_RX_BW_DCC_FREQ_1_8 0xE0
#define REG_RX_BW_RX_BW_MANT_MASK 0x18
#define REG_RX_BW_RX_BW_MANT_16 0x00
#define REG_RX_BW_RX_BW_MANT_20 0x08
#define REG_RX_BW_RX_BW_MANT_24 0x10
#define REG_RX_BW_RX_BW_EXP_MASK 0x07

#define REG_RSSI_CONFIG_START 0x01
#define REG_RSSI_CONFIG_DONE 0x02

#define REG_IRQ_FLAGS1_PLL_LOCK 0x10
#define REG_IRQ_FLAGS1_RX_READY 0x40

#define REG_TEST_LNA_HIGH_SENSITIVITY 0x2D

#define REG_TEST_TCXO_INPUT_MASK 0x10
#define REG_TEST_TCXO_INPUT_XTA 0x10

#define FSTEP 61.03515625
#define FXOSC 32000000

static int reg_read(struct sx1239_device *dev, uint8_t addr, uint8_t *buf,
		    size_t len)
{
	struct sx1239 *sx1239;
	struct spi_msg msg[2];
	uint8_t hdr;

	assert(dev);
	assert(!(addr & ~REG_ADDR_MASK));
	assert(buf && len);

	sx1239 = container_of(dev, struct sx1239, dev);

	memset(msg, 0, sizeof(msg));

	hdr = REG_READ | (addr & REG_ADDR_MASK);

	msg[0].mosi = &hdr;
	msg[0].len = sizeof(hdr);
	msg[0].mode = SPI_MSG_HEADER;
	msg[1].miso = buf;
	msg[1].len = len;

	return sx1239->spi->msg(sx1239->spi, msg, sizeofa(msg));
}

static int reg_write(struct sx1239_device *dev, uint8_t addr,
		     const uint8_t *buf, size_t len)
{
	struct sx1239 *sx1239;
	struct spi_msg msg[2];
	uint8_t hdr;

	assert(dev);
	assert(!(addr & ~REG_ADDR_MASK));
	assert(buf && len);

	sx1239 = container_of(dev, struct sx1239, dev);

	memset(msg, 0, sizeof(msg));

	hdr = REG_WRITE | (addr & REG_ADDR_MASK);

	msg[0].mosi = &hdr;
	msg[0].len = sizeof(hdr);
	msg[0].mode = SPI_MSG_HEADER;
	msg[1].mosi = buf;
	msg[1].len = len;

	return sx1239->spi->msg(sx1239->spi, msg, sizeofa(msg));
}

static int reg_get(struct sx1239_device *dev, uint8_t addr)
{
	uint8_t val;

	if (reg_read(dev, addr, &val, 1) < 0) {
		error("Failed to read register");
		return -1;
	}

	return val;
}

static int reg_set_full(struct sx1239_device *dev, uint8_t addr, uint8_t val)
{
	if (reg_write(dev, addr, &val, 1) < 0) {
		error("Failed to write register");
		return -1;
	}

	return 0;
}

static int reg_set_mask(struct sx1239_device *dev, uint8_t addr, uint8_t val,
			uint8_t mask)
{
	int old, new;

	if ((old = reg_get(dev, addr)) < 0) {
		error("Failed to get register");
		return -1;
	}

	new = old;
	new &= ~mask;
	new |= val &mask;

	if (new != old && reg_set_full(dev, addr, new) < 0) {
		error("Failed to set full register");
		return -1;
	}

	return 0;
}

#define _reg_set_suffix(dev, addr, val, mask, suffix, ...) suffix
#define reg_set_suffix(args...) _reg_set_suffix(args, _mask, _full)
#define __reg_set(suffix, args...) reg_set##suffix(args)
#define _reg_set(suffix, args...) __reg_set(suffix, args)
#define reg_set(args...) _reg_set(reg_set_suffix(args), args)

static int reg_check(struct sx1239_device *dev, uint8_t addr, uint8_t val,
		     uint8_t mask)
{
	int count, current;

	for (count = 0; count < 100; count++) {
		if ((current = reg_get(dev, addr)) < 0) {
			error("Failed to get register");
			return -1;
		}

		if ((current & mask) == val) {
			return 0;
		}

		sleep_us(10);
	}

	error("Failed to check register bit: Timeout");

	return -1;
}

static int set_op_mode(struct sx1239_device *dev, uint8_t mode)
{
	if (reg_set(dev, REG_OP_MODE, mode, REG_OP_MODE_MODE_MASK) < 0) {
		error("Failed to set register");
		return -1;
	}

	if (reg_check(dev, REG_OP_MODE, mode, REG_OP_MODE_MODE_MASK) < 0) {
		error("Failed to check mode");
		return -1;
	}

	return 0;
}

static int set_freq(struct sx1239_device *dev, uint32_t freq)
{
	union {
		uint32_t dword;
		uint8_t bytes[4];
	} f;

	f.dword = htonl(0.5 + freq / FSTEP);

	if (reg_write(dev, REG_FRF, &f.bytes[1], 3) < 0) {
		error("Failed to write RF carrier frequency");
		return -1;
	}

	return 0;
}

static int sx1239_set_bitrate(struct sx1239_device *dev, uint32_t bitrate)
{
	union {
		uint16_t word;
		uint8_t bytes[2];
	} br;

	br.word = htons((FXOSC + bitrate / 2) / bitrate);

	if (reg_write(dev, REG_BITRATE, br.bytes, 2) < 0) {
		error("Failed to write bit rate");
		return -1;
	}

	return 0;
}

static int sx1239_set_dcc_freq(struct sx1239_device *dev, uint8_t dcc_freq)
{
	uint8_t df;

	switch (dcc_freq) {
	case 8:
		df = REG_RX_BW_DCC_FREQ_16;
		break;
	case 7:
		df = REG_RX_BW_DCC_FREQ_8;
		break;
	case 6:
		df = REG_RX_BW_DCC_FREQ_4;
		break;
	case 5:
		df = REG_RX_BW_DCC_FREQ_2;
		break;
	case 4:
		df = REG_RX_BW_DCC_FREQ_1;
		break;
	case 3:
		df = REG_RX_BW_DCC_FREQ_1_2;
		break;
	case 2:
		df = REG_RX_BW_DCC_FREQ_1_4;
		break;
	default:
		df = REG_RX_BW_DCC_FREQ_1_8;
		break;
	}

	if (reg_set(dev, REG_RX_BW, df, REG_RX_BW_DCC_FREQ_MASK) < 0) {
		error("Failed to configure DCC cutoff frequency");
		return -1;
	}

	return 0;
}

static int sx1239_set_lna_gain(struct sx1239_device *dev, uint8_t lna_gain)
{
	uint8_t lg;

	switch (lna_gain) {
	case 6:
		lg = REG_LNA_GAIN_G6;
		break;
	case 5:
		lg = REG_LNA_GAIN_G5;
		break;
	case 4:
		lg = REG_LNA_GAIN_G4;
		break;
	case 3:
		lg = REG_LNA_GAIN_G3;
		break;
	case 2:
		lg = REG_LNA_GAIN_G2;
		break;
	default:
		lg = REG_LNA_GAIN_G1;
		break;
	}

	if (reg_set(dev, REG_LNA, lg, REG_LNA_GAIN_MASK) < 0) {
		error("Failed to configure LNA gain");
		return -1;
	}

	return 0;
}

static int dev_init(struct sx1239_device *dev)
{
	if (sx1239_set_dcc_freq(dev, 0) < 0) {
		error("Failed to configure default DCC cutoff frequency");
		return -1;
	}

	if (sx1239_set_lna_gain(dev, 0) < 0) {
		error("Failed to configure default LNA gain");
		return -1;
	}

	if (reg_set(dev, REG_TEST_LNA, REG_TEST_LNA_HIGH_SENSITIVITY) < 0) {
		error("Failed to configure sensitivity");
		return -1;
	}

	if (reg_set(dev, REG_TEST_TCXO, REG_TEST_TCXO_INPUT_XTA,
		    REG_TEST_TCXO_INPUT_MASK) < 0) {
		error("Failed to configure oscillator");
		return -1;
	}

	if (reg_set(dev, REG_LNA, REG_LNA_ZIN_200_OHMS, REG_LNA_ZIN_MASK) < 0) {
		error("Failed to configure LNA");
		return -1;
	}

	if (reg_set(dev, REG_DATA_MODUL, REG_DATA_MODUL_MODE_CONTINIOUS |
					     REG_DATA_MODUL_MODULATION_FSK,
		    REG_DATA_MODUL_MODE_MASK | REG_DATA_MODUL_MODULATION_MASK) <
	    0) {
		error("Failed to configure modulation");
		return -1;
	}

	return 0;
}

static struct {
	uint8_t mant_exp;
	uint32_t bw_fsk;
	uint32_t bw_ook;
} rxbw_regs[] = {
    {REG_RX_BW_RX_BW_MANT_24 | 7, 2600, 1300},
    {REG_RX_BW_RX_BW_MANT_20 | 7, 3100, 1600},
    {REG_RX_BW_RX_BW_MANT_16 | 7, 3900, 2000},
    {REG_RX_BW_RX_BW_MANT_24 | 6, 5200, 2600},
    {REG_RX_BW_RX_BW_MANT_20 | 6, 6300, 3100},
    {REG_RX_BW_RX_BW_MANT_16 | 6, 7800, 3900},
    {REG_RX_BW_RX_BW_MANT_24 | 5, 10400, 5200},
    {REG_RX_BW_RX_BW_MANT_20 | 5, 12500, 6300},
    {REG_RX_BW_RX_BW_MANT_16 | 5, 15600, 7800},
    {REG_RX_BW_RX_BW_MANT_24 | 4, 20800, 10400},
    {REG_RX_BW_RX_BW_MANT_20 | 4, 25000, 12500},
    {REG_RX_BW_RX_BW_MANT_16 | 4, 31300, 15600},
    {REG_RX_BW_RX_BW_MANT_24 | 3, 41700, 20800},
    {REG_RX_BW_RX_BW_MANT_20 | 3, 50000, 25000},
    {REG_RX_BW_RX_BW_MANT_16 | 3, 62500, 31300},
    {REG_RX_BW_RX_BW_MANT_24 | 2, 83300, 41700},
    {REG_RX_BW_RX_BW_MANT_20 | 2, 100000, 50000},
    {REG_RX_BW_RX_BW_MANT_16 | 2, 125000, 62500},
    {REG_RX_BW_RX_BW_MANT_24 | 1, 166700, 83300},
    {REG_RX_BW_RX_BW_MANT_20 | 1, 200000, 100000},
    {REG_RX_BW_RX_BW_MANT_16 | 1, 250000, 125000},
    {REG_RX_BW_RX_BW_MANT_24 | 0, 333300, 166700},
    {REG_RX_BW_RX_BW_MANT_20 | 0, 400000, 200000},
    {REG_RX_BW_RX_BW_MANT_16 | 0, 500000, 250000},
};

static int sx1239_set_bandwidth(struct sx1239_device *dev, uint32_t bandwidth)
{
	uint8_t mant_exp;
	int mode;
	size_t n;

	if ((mode = reg_get(dev, REG_DATA_MODUL)) < 0) {
		error("Failed to get mode");
		return -1;
	}

	for (mant_exp = 0, n = 0; n < sizeofa(rxbw_regs); n++) {
		if ((mode & REG_DATA_MODUL_MODULATION_MASK) ==
		    REG_DATA_MODUL_MODULATION_FSK) {
			if (bandwidth <= 2 * rxbw_regs[n].bw_fsk) {
				mant_exp = rxbw_regs[n].mant_exp;
				break;
			}

		} else {
			if (bandwidth <= 2 * rxbw_regs[n].bw_ook) {
				mant_exp = rxbw_regs[n].mant_exp;
				break;
			}
		}
	}

	assert(mant_exp > 0);

	if (reg_set(dev, REG_RX_BW, mant_exp,
		    REG_RX_BW_RX_BW_MANT_MASK | REG_RX_BW_RX_BW_EXP_MASK) < 0) {
		error("Failed to set bandwidth");
		return -1;
	}

	return 0;
}

static int sx1239_start_rx(struct sx1239_device *dev, uint32_t freq)
{
	if (set_freq(dev, freq) < 0) {
		error("Failed to set RF carrier frequency");
		return -1;
	}

	if (set_op_mode(dev, REG_OP_MODE_MODE_FS) < 0) {
		error("Failed to set FS mode");
		return -1;
	}

	if (set_op_mode(dev, REG_OP_MODE_MODE_RX) < 0) {
		error("Failed to set RX mode");
		return -1;
	}

	if (reg_check(dev, REG_IRQ_FLAGS1, REG_IRQ_FLAGS1_RX_READY,
		      REG_IRQ_FLAGS1_RX_READY) < 0) {
		error("Failed to wait for RX ready");
		return -1;
	}

	return 0;
}

static int sx1239_rssi(struct sx1239_device *dev)
{
	int rssi;

	if (reg_set(dev, REG_RSSI_CONFIG, REG_RSSI_CONFIG_START) < 0) {
		error("Failed to start RSSI measurement");
		return -1;
	}

	if (reg_check(dev, REG_RSSI_CONFIG, REG_RSSI_CONFIG_DONE,
		      REG_RSSI_CONFIG_DONE) < 0) {
		error("Failed to wait for RSSI done");
		return -1;
	}

	if ((rssi = reg_get(dev, REG_RSSI_VALUE)) < 0) {
		error("Failed to get RSSI");
		return -1;
	}

	return rssi;
}

static void sx1239_close(struct sx1239_device *dev)
{
	struct sx1239 *sx1239;

	if (dev) {
		sx1239 = container_of(dev, struct sx1239, dev);

		if (sx1239->reset) {
			sx1239->reset->set_value(sx1239->reset, GPIO_HIGH);
		}

		if (sx1239->enable) {
			sx1239->enable->set_value(sx1239->enable, GPIO_LOW);
		}

		free(sx1239);
	}
}

struct sx1239_device *sx1239_open(struct spi_device *spi,
				  struct gpio_device *enable,
				  struct gpio_device *reset)
{
	struct sx1239 *sx1239;

	assert(spi);

	if ((sx1239 = zalloc(sizeof(*sx1239))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	sx1239->dev.close = sx1239_close;
	sx1239->dev.set_dcc_freq = sx1239_set_dcc_freq;
	sx1239->dev.set_lna_gain = sx1239_set_lna_gain;
	sx1239->dev.set_bandwidth = sx1239_set_bandwidth;
	sx1239->dev.set_bitrate = sx1239_set_bitrate;
	sx1239->dev.start_rx = sx1239_start_rx;
	sx1239->dev.rssi = sx1239_rssi;

	sx1239->spi = spi;
	sx1239->reset = reset;
	sx1239->enable = enable;

	if (sx1239->spi->set_mode(sx1239->spi, SPI_MODE_0 | SPI_MSB_FIRST |
						   SPI_WORD_8BITS) < 0) {
		error("Failed to configure spi");
		sx1239_close(&sx1239->dev);
		return NULL;
	}

	if (enable) {
		if (enable->set_mode(enable, GPIO_OUTPUT | GPIO_LOW) < 0) {
			error("Failed to configure gpio enable");
			sx1239_close(&sx1239->dev);
			return NULL;
		}

		if (enable->set_value(enable, GPIO_HIGH) < 0) {
			error("Failed to set gpio enable");
			sx1239_close(&sx1239->dev);
			return NULL;
		}
	}

	if (reset) {
		if (reset->set_mode(reset, GPIO_OUTPUT | GPIO_HIGH) < 0) {
			error("Failed to configure gpio reset");
			sx1239_close(&sx1239->dev);
			return NULL;
		}

		sleep_ms(10);

		if (reset->set_value(reset, GPIO_HIGH) < 0) {
			error("Failed to set gpio reset");
			sx1239_close(&sx1239->dev);
			return NULL;
		}

		sleep_ms(10);

		if (reset->set_value(reset, GPIO_LOW) < 0) {
			error("Failed to reset gpio reset");
			sx1239_close(&sx1239->dev);
			return NULL;
		}
	}

	if (dev_init(&sx1239->dev) < 0) {
		error("Failed to initialize device");
		sx1239_close(&sx1239->dev);
		return NULL;
	}

	return &sx1239->dev;
}
