// Copyright (c) 2019 Kerlink SA. All rights reserved.

#include "drivers/sx1272.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "lib/loragw/hal.h"
#include "lib/loragw/radio.h"
#include "lib/loragw/spi.h"
#include "lib/loragw/sx1272_fsk.h"
#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"


#define FXOSC                           32000000

/**
@struct lgw_radio_FSK_bandwidth_s
@brief Associate a bandwidth in kHz with its corresponding register values
*/
struct lgw_sx1272_FSK_bandwidth_s {
    uint32_t    RxBwKHz;
    uint8_t     RxBwMant;
    uint8_t     RxBwExp;
};

const struct lgw_sx1272_FSK_bandwidth_s sx1272_FskBandwidths[] =
{
    { 2600  , 2, 7 },   /* LGW_SX127X_RXBW_2K6_HZ */
    { 3100  , 1, 7 },   /* LGW_SX127X_RXBW_3K1_HZ */
    { 3900  , 0, 7 },   /* ... */
    { 5200  , 2, 6 },
    { 6300  , 1, 6 },
    { 7800  , 0, 6 },
    { 10400 , 2, 5 },
    { 12500 , 1, 5 },
    { 15600 , 0, 5 },
    { 20800 , 2, 4 },
    { 25000 , 1, 4 },   /* ... */
    { 31300 , 0, 4 },
    { 41700 , 2, 3 },
    { 50000 , 1, 3 },
    { 62500 , 0, 3 },
    { 83333 , 2, 2 },
    { 100000, 1, 2 },
    { 125000, 0, 2 },
    { 166700, 2, 1 },
    { 200000, 1, 1 },   /* ... */
    { 250000, 0, 1 }    /* LGW_SX127X_RXBW_250K_HZ */
};

static int sx1272_reg_check(uint8_t addr, uint8_t val, uint8_t mask)
{
	int count;
	uint8_t reg_val;

	for (count = 0; count < 100; count++) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(addr, &reg_val)) {
			error("Failed to get register 0x%x", addr);
			return -1;
		} else if ((reg_val & mask) == val)
			return 0;
		else
			sleep_us(1000);
	}

	error("Check failed for register 0x%x: 0x%x != 0x%x", addr, reg_val, val);
	return -1;
}

static int sx1272_reg_set_mask(uint8_t addr, uint8_t val, uint8_t mask)
{
	uint8_t reg_val;
	uint8_t new_val;

	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(addr, &reg_val)) {
		error("Failed to get register");
		return -1;
	}

	new_val = reg_val & ~mask;
	new_val |= val & mask;

	if (new_val != reg_val) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(addr, new_val)) {
			error("Failed to set register");
			return -1;
		}
	}
	return 0;
}

static int sx1272_set_mode(struct sx1272_device *dev, uint8_t mode)
{
	uint8_t reg_val;

	(void) dev;
	if (sx1272_reg_set_mask(SX1272_REG_OPMODE, mode, SX1272_REG_OPMODE_MODE_MASK)) {
		error("Failed to set opmode register");
		return -1;
	}
	if (sx1272_reg_check(SX1272_REG_OPMODE, mode,
				SX1272_REG_OPMODE_MODE_MASK) < 0) {
		lgw_sx127x_reg_r(SX1272_REG_OPMODE, &reg_val);
		error("Mode in OPMODE reg is 0x%02x instead of 0x%02x", reg_val, mode);
		return -1;
	}
	/*lgw_sx127x_reg_r(SX1272_REG_OPMODE, &reg_val);
	info("sx1272_set_mode OPMODE value: 0x%x", reg_val);*/

	if (sx1272_reg_check(SX1272_REG_IRQFLAGS1, SX1272_REG_IRQFLAGS1_MODE_READY,
				SX1272_REG_IRQFLAGS1_MODE_READY) < 0) {
		error("Programmed mode not correctly set");
		return -1;
	}
	return 0;
}

static int sx1272_set_br(struct sx1272_device *dev, uint32_t br)
{
	uint16_t bitrate;

	(void) dev;
	bitrate = (FXOSC / br);

	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_BITRATEMSB, (bitrate & 0xFF00) >> 8)) {
		error("Failed to set BITRATE MSB");
		return -1;
	}
	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_BITRATELSB, (bitrate & 0x00FF))) {
		error("Failed to set BITRATE LSB");
		return -1;
	}

	return 0;
}

static int sx1272_set_bw(struct sx1272_device *dev, uint32_t bw)
{
	uint8_t rxBwExp;
	uint8_t rxBwMant;
	unsigned int i;

	(void) dev;
	/* we need to use mid bandwidth in KHz */
	for (i=0; i< sizeof(sx1272_FskBandwidths) /
			sizeof(struct lgw_sx1272_FSK_bandwidth_s); i++) {
		if (bw <= sx1272_FskBandwidths[i].RxBwKHz * 2) {
			rxBwExp = sx1272_FskBandwidths[i].RxBwExp;
			rxBwMant = sx1272_FskBandwidths[i].RxBwMant;
			if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_RXBW,
						rxBwExp | (rxBwMant << 3))) {
				error("Failed to set bandwidth");
				return -1;
			}
			info("Selected single side bandwidth: %lu", (unsigned long) sx1272_FskBandwidths[i].RxBwKHz);
			return 0;
		}
	}

	error("bandwidth (%lu) not supported", (unsigned long)bw);
	return -1;
}

static int sx1272_set_freq(struct sx1272_device *dev, uint32_t freq)
{
	uint64_t freq_reg;
	int x;

	(void) dev;
	freq_reg = ((uint64_t)freq << 19) / (uint64_t)32000000;
	x = lgw_sx127x_reg_w(SX1272_REG_FRFMSB, (freq_reg >> 16) & 0xFF);
	x = lgw_sx127x_reg_w(SX1272_REG_FRFMID, (freq_reg >> 8) & 0xFF);
	x = lgw_sx127x_reg_w(SX1272_REG_FRFLSB, (freq_reg >> 0) & 0xFF);
	if (LGW_SPI_SUCCESS != x) {
		error("Failed to set frequency");
		return -1;
	}

	return 0;
}

/* rssi in db is 0 - result /2 */
static int sx1272_rssi(struct sx1272_device *dev, uint8_t firstAtThisFreq, int* rssi)
{
	uint8_t reg_rssi;

	(void) dev;
	/* start measure */
	if (firstAtThisFreq) {
		if (sx1272_reg_set_mask(SX1272_REG_RXCONFIG, 
				SX1272_REG_RXCONFIG_RESTART_WITH_PLL_LOCK_MASK, 
				SX1272_REG_RXCONFIG_RESTART_WITH_PLL_LOCK_MASK)) {
			error("Failed to set RXCONFIG register (with pll)");
			return -1;
		}
	} else {
		if (sx1272_reg_set_mask(SX1272_REG_RXCONFIG, 
				SX1272_REG_RXCONFIG_RESTART_WITHOUT_PLL_LOCK_MASK, 
				SX1272_REG_RXCONFIG_RESTART_WITHOUT_PLL_LOCK_MASK)) {
			error("Failed to set RXCONFIG register (without pll)");
			return -1;
		}
	}

	/* wait rx ready flag */
	if (sx1272_reg_check(SX1272_REG_IRQFLAGS1, SX1272_REG_IRQFLAGS1_RX_READY,
				SX1272_REG_IRQFLAGS1_RX_READY) < 0) {
		error("SX1272 not RX ready");
		return -1;
	}

	/* read result */
	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(SX1272_REG_RSSIVALUE, &reg_rssi)) {
		error("Failed to read RSSI VALUE register");
		return -1;
	}

	/*rssidb = 0.0f - ((float)reg_rssi) / 2;*/
	*rssi = reg_rssi;
	return 0;
}

static void sx1272_dump_regs(struct sx1272_device *dev)
{
	int i;
	uint8_t reg_val;

	(void) dev;
	info("Dumping SX1272 registers");
	for(i=0; i<0x71; i++) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(i, &reg_val)) {
			info("%02x: error", i);
		} else {
			info("%02x: %02x", i, reg_val);
		}
	}
}

void sx1272_close(struct sx1272_device *dev)
{
	if(dev) {
		lgw_stop();
		free(dev);
	}
	return;
}

struct sx1272_device *sx1272_open(uint32_t bandwidth, uint32_t bitrate)
{
	struct sx1272_device *sx1272 = NULL;
	int x;
	uint8_t version;
	uint8_t initialMode;
	uint8_t AdcBwAuto = 0;
	uint8_t AdcBw = 7;
	uint8_t AdcLowPwr = 0;
	uint8_t AdcTrim = 6;
	uint8_t AdcTest = 0;
	uint8_t RssiSmoothing = 5;

	/* create returned structure */
	if ((sx1272 = zalloc(sizeof(*sx1272))) == NULL) {
		eerror("Failed to allocate memory");
		goto sx1272_open_exit;
	}

	sx1272->close = sx1272_close;
	sx1272->set_mode = sx1272_set_mode;
	sx1272->set_bandwidth = sx1272_set_bw;
	sx1272->set_bitrate = sx1272_set_br;
	sx1272->set_freq = sx1272_set_freq;
	sx1272->rssi = sx1272_rssi;
	sx1272->dump_regs = sx1272_dump_regs;

	/* Read SX1272 version register */
	x = lgw_sx127x_reg_r(0x42, &version);
	if (x != LGW_SPI_SUCCESS) {
		error("Failed to read sx127x version register");
		goto sx1272_open_exit;
	}
	/* Read SX1272 mode */
	x = lgw_sx127x_reg_r(SX1272_REG_OPMODE, &initialMode);
	if (x != LGW_SPI_SUCCESS) {
		error("Failed to read sx127x mode register");
		goto sx1272_open_exit;
	}
	info("sx1272 version: 0x%x, initial mode:0x%x", version, initialMode);

	/* go to standby mode */
	if(sx1272_set_mode(sx1272, SX1272_REG_OPMODE_MODE_STANDBY) != 0) {
		error("mode standby KO");
		goto sx1272_open_exit;
	} else {
		info("Setting mode standby OK");
	}

	/* PLL Fast Hop */
	if(sx1272_reg_set_mask(SX1272_REG_PLLHOP, SX1272_REG_PLLHOP_FAST_HOP_ON_MASK,
				SX1272_REG_PLLHOP_FAST_HOP_ON_MASK) != 0)
		goto sx1272_open_exit;

	/* LNA gain and boost */
	if(sx1272_reg_set_mask(SX1272_REG_LNA, SX1272_REG_LNA_LNA_GAIN_MASK, SX1272_REG_LNA_LNA_GAIN_G1) != 0)
		goto sx1272_open_exit;
	if(sx1272_reg_set_mask(SX1272_REG_LNA, SX1272_REG_LNA_LNA_BOOST_MASK, SX1272_REG_LNA_LNA_BOOST_IMPROVED) != 0)
		goto sx1272_open_exit;

	/* undocumented config */
	x |= lgw_sx127x_reg_w(0x68, AdcBw | (AdcBwAuto << 3));
	x |= lgw_sx127x_reg_w(0x69, AdcTest | (AdcTrim << 4) | (AdcLowPwr << 7));

	/* bitrate (was 1000) in spectral scan */
	if(sx1272_set_br(sx1272, bitrate) != 0)
		goto sx1272_open_exit;
	/* frequency deviation */
	x |= lgw_sx127x_reg_w(SX1272_REG_FDEVMSB, 2);
	x |= lgw_sx127x_reg_w(SX1272_REG_FDEVLSB, 225);
	/* Disable AGC */
	if(sx1272_reg_set_mask(SX1272_REG_RXCONFIG, SX1272_REG_RXCONFIG_AGC_AUTO_ON_MASK, 0) != 0)
		goto sx1272_open_exit;
	/* 64 samples when getting average RSSI */
	x |= lgw_sx127x_reg_w(SX1272_REG_RSSICONFIG, RssiSmoothing);
	/* bandwidth */
	if(sx1272_set_bw(sx1272, bandwidth) != 0)
		goto sx1272_open_exit;
	/* other */
	x |= lgw_sx127x_reg_w(SX1272_REG_RXDELAY, 2);
	x |= lgw_sx127x_reg_w(SX1272_REG_PLL, 0x10); /* PLL BW set to 75 KHz */
	x |= lgw_sx127x_reg_w(0x47, 1); /* optimize PLL start-up time */

	if (x != 0) {
		error("Something failed in SX1272 configuration");
		goto sx1272_open_exit;
	}
	/* no frequency set. Will be set for each scan */

	/* RX mode */
	sleep(1); /* necessary?*/
	if (sx1272_set_mode(sx1272, SX1272_REG_OPMODE_MODE_RX) != 0) {
		error("Failed to set mode RX");
		sx1272_dump_regs(NULL);
		goto sx1272_open_exit;
	}
	info("Setting mode RX OK");

	return sx1272;

sx1272_open_exit:
	if (sx1272) {
		free(sx1272);
	}
	lgw_stop();
	return NULL;
}
