//Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "calib.h"

#include <stdlib.h>

#include "spi.h"

#include "utils/logger.h"
#include "utils/memory.h"

extern void *lgw_spi_target;

#define CALIB_LORA_MODE_OFFSET -139.0f

static int calib_compare_tx_lut(const void *a, const void *b)
{
	return (((struct calib_tx_lut *)a)->rf_power - ((struct calib_tx_lut *)b)->rf_power);
}

struct calib *calib_read(void)
{
	struct calib *calib;

	if ((calib = zalloc(sizeof(*calib))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	if (lgw_spi_rb(lgw_spi_target, LGW_SPI_MUX_MODE1, LGW_SPI_MUX_TARGET_EEPROM, 0, calib, sizeof(*calib)) < 0) {
		error("Failed to read eeprom");
		free(calib);
		return NULL;
	}

	if (calib->rssi_offset[CALIB_SNIFFER] < (CALIB_LORA_MODE_OFFSET / 2)) {
		/* Calibration has been done in LORA mode, correction according to datasheet (5.5.5) */
		calib->rssi_offset[CALIB_SNIFFER] -= CALIB_LORA_MODE_OFFSET;
	}

	qsort(calib->tx_lut, sizeofa(calib->tx_lut), sizeof(*calib->tx_lut), calib_compare_tx_lut);

	return calib;
}

struct calib_manufacturer {
	uint8_t manufacturer;
	const char *manufacturer_name;
} calib_manufacturers[] = {
    {CALIB_MANUFACTURER_KERLINK, "Kerlink"},
};

const char *calib_manufacturer_name(uint8_t manufacturer)
{
	uint8_t n;

	for (n = 0; n < sizeofa(calib_manufacturers); n++) {
		if (calib_manufacturers[n].manufacturer == manufacturer) {
			return calib_manufacturers[n].manufacturer_name;
		}
	}

	return "Unknown";
}

struct calib_board {
	uint8_t board_type;
	const char *board_name;
} calib_boards[] = {
    {CALIB_BOARD_TYPE_WSTA_868MHZ, "Wirnet Station 868 MHz"},
    {CALIB_BOARD_TYPE_WSTA_868MHZ_27DBM, "Wirnet Station 868 MHz / 27 dBm"},
    {CALIB_BOARD_TYPE_WSTA_915MHZ_27DBM, "Wirnet Station 915 MHz / 27 dBm"},
    {CALIB_BOARD_TYPE_WSTA_923MHZ_27DBM, "Wirnet Station 923 MHZ / 27 dBm"},
    {CALIB_BOARD_TYPE_WSTA_433MHZ, "Wirnet Station 433 MHz"},
    {CALIB_BOARD_TYPE_WSTA_923MHZ_27DBM_REVB, "Wirnet Station 923 MHz / 27 dBm / revision B"},
    {CALIB_BOARD_TYPE_WSTA_868MHZ_27DBM_REVB, "Wirnet Station 868 MHz / 27 dBm / revision B"},
    {CALIB_BOARD_TYPE_WSTA_868MHZ_27DBM_REVC, "Wirnet Station 868 MHz / 27 dBm / revision C"},
    {CALIB_BOARD_TYPE_WSTA_915MHZ_27DBM_REVB, "Wirnet Station 915 MHz / 27 dBm / revision B"},
    {CALIB_BOARD_TYPE_WSTA_923MHZ_27DBM_REVC, "Wirnet Station 923 MHz / 27 dBm / revision C"},
    {CALIB_BOARD_TYPE_WIFC_868MHZ, "Wirnet iFemtocell 868 MHz"},
    {CALIB_BOARD_TYPE_WIFC_915MHZ, "Wirnet iFemtocell 915 MHz"},
    {CALIB_BOARD_TYPE_WIFC_915MHZ, "Wirnet iFemtocell 923 MHz"},
    {CALIB_BOARD_TYPE_WIIS_923MHZ, "Wirnet iStation 923 MHz"},
    {CALIB_BOARD_TYPE_WIIS_915MHZ, "Wirnet iStation 915 MHz"},
    {CALIB_BOARD_TYPE_WIIS_868MHZ, "Wirnet iStation 868 MHz"},
    {CALIB_BOARD_TYPE_FEVO_868MHZ, "Wirnet iFemtocell Evolution 868 MHz"},
    {CALIB_BOARD_TYPE_FEVO_915MHZ, "Wirnet iFemtocell Evolution 915 MHz"},
    {CALIB_BOARD_TYPE_FEVO_915MHZ, "Wirnet iFemtocell Evolution 923 MHz"},
};

const char *calib_board_name(uint8_t board_type)
{
	uint8_t n;

	for (n = 0; n < sizeofa(calib_boards); n++) {
		if (calib_boards[n].board_type == board_type) {
			return calib_boards[n].board_name;
		}
	}

	return "Unknown";
}
