/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Functions used to handle the Listen Before Talk feature

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <stdlib.h>  /* abs, labs, llabs */
#include <string.h>  /* memset */

#include "aux.h"
#include "fpga.h"
#include "lbt.h"
#include "radio.h"
#include "reg.h"

#include "clock.h"

#include "utils/logger.h"

#if defined(PLATFORM_WIFC)
#include "drivers/sx1239.h"

#define SNIFFER_SPI_PATH "/dev/spidev1.0"
#define SNIFFER_SPI_SPEED 10000000
#define SNIFFER_GPIO_EN 10
#define SNIFFER_GPIO_RST 135
#define SNIFFER_BANDWIDTH 200000
#define SNIFFER_BITRATE 4800

#define LBT_STARTUP_TIME 4
#define LBT_SCAN_INTERVAL 200000000
#define LBT_SUB_SCAN_INTERVAL (LBT_SCAN_INTERVAL / LBT_CHANNEL_FREQ_NB)

static struct spi_device *sniffer_spi;
static struct gpio_device *sniffer_enable;
static struct gpio_device *sniffer_reset;
static struct sx1239_device *sniffer;

static int lbt_counter[LBT_CHANNEL_FREQ_NB];
#endif

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define CHECK_NULL(a)                 \
	if (a == NULL) {              \
		return LGW_REG_ERROR; \
	}

#define LBT_TIMESTAMP_MASK 0x007FF000 /* 11-bits timestamp */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE TYPES -------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS ---------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- INTERNAL SHARED VARIABLES -------------------------------------------- */

extern void *lgw_spi_target;     /*! generic pointer to the SPI device */
extern uint8_t lgw_spi_mux_mode; /*! current SPI mux mode used */
extern uint16_t lgw_i_tx_start_delay_us;

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

static bool lbt_enable;
static uint8_t lbt_nb_active_channel;
static int8_t lbt_rssi_target_dBm;
static int8_t lbt_rssi_offset_dB;
static struct lgw_conf_lbt_chan_s lbt_channel_cfg[LBT_CHANNEL_FREQ_NB];

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS ---------------------------------------------------- */

bool is_equal_freq(uint32_t a, uint32_t b);

#if defined(PLATFORM_WIFC)
static int lbt_scan(unsigned int lbt_channel_index)
{
	int s, samples = 1;
	int free = 1;
	int rssi;

	if (lbt_channel_index >= lbt_nb_active_channel) {
		return LGW_LBT_ERROR;
	}

	if (sniffer->start_rx(sniffer, lbt_channel_cfg[lbt_channel_index].freq_hz) < 0) {
		error("Failed to update frequency");
		return LGW_LBT_ERROR;
	}

	if (lbt_channel_cfg[lbt_channel_index].scan_time_us == 5000) {
		samples = 5; /* About 1ms per get_rssi */
	}

	for (s = 0; s < samples; s++) {
		if ((rssi = sniffer->rssi(sniffer)) < 0) {
			error("Failed to get rssi");
			free = 0;
			goto update;

		} else if (lbt_rssi_target_dBm < -(rssi / 2) + lbt_rssi_offset_dB) {
			free = 0;
			goto update;
		}
	}

update:
	debug("RSSI detected: %i\n", -(rssi / 2) + lbt_rssi_offset_dB);

	if (free) {
		lbt_counter[lbt_channel_index] = 0;

	} else if (lbt_counter[lbt_channel_index] >= 30) {
		lbt_counter[lbt_channel_index] = -1;

	} else if (lbt_counter[lbt_channel_index] >= 0) {
		lbt_counter[lbt_channel_index] += 1;
	}

	return LGW_LBT_SUCCESS;
}

#endif

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int lbt_setconf(struct lgw_conf_lbt_s *conf)
{
	int i;

	/* Check input parameters */
	if (conf == NULL) {
		return LGW_LBT_ERROR;
	}

	/* Set internal LBT config according to parameters */
	lbt_enable = conf->enable;

	if (lbt_enable == false) {
		info("Using lbt: enable=false");
		return LGW_LBT_SUCCESS;
	}

	if ((conf->nb_channel < 1) || (conf->nb_channel > LBT_CHANNEL_FREQ_NB)) {
		error("Number of defined LBT channels is out of range (%u)", conf->nb_channel);
		return LGW_LBT_ERROR;
	}

	lbt_nb_active_channel = conf->nb_channel;
	lbt_rssi_target_dBm = conf->rssi_target;
	lbt_rssi_offset_dB = conf->rssi_offset;

	info("Using lbt: enable=true rssi_target=%i rssi_offset=%i",
	     lbt_rssi_target_dBm,
	     lbt_rssi_offset_dB);

	/* Initialize LBT channels configuration */
	memset(lbt_channel_cfg, 0, sizeof lbt_channel_cfg);

	for (i = 0; i < lbt_nb_active_channel; i++) {
		lbt_channel_cfg[i].freq_hz = conf->channels[i].freq_hz;
		lbt_channel_cfg[i].scan_time_us = conf->channels[i].scan_time_us;

		info("Using lbt: channel=%-2u freq_hz=%u scan_time_us=%u",
		     i,
		     lbt_channel_cfg[i].freq_hz,
		     lbt_channel_cfg[i].scan_time_us);
	}

	return LGW_LBT_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int lbt_setup(void)
{
#if defined(PLATFORM_WIFC)
	int channel;

	for (channel = 0; channel < LBT_CHANNEL_FREQ_NB; channel++) {
		lbt_counter[channel] = -1;
	}

	if ((sniffer_spi = spi_open(SNIFFER_SPI_PATH, NULL)) == NULL) {
		error("Failed to open sniffer spi");
		return LGW_LBT_ERROR;
	}

	if (sniffer_spi->set_speed(sniffer_spi, SNIFFER_SPI_SPEED) < 0) {
		error("Failed to configure sniffer spi");
		return LGW_LBT_ERROR;
	}

	if ((sniffer_enable = gpio_open(SNIFFER_GPIO_EN, GPIO_OUTPUT)) == NULL) {
		error("Failed to open sniffer gpio enable");
		return LGW_LBT_ERROR;
	}

	if ((sniffer_reset = gpio_open(SNIFFER_GPIO_RST, GPIO_OUTPUT)) == NULL) {
		error("Failed to open sniffer gpio reset");
		return LGW_LBT_ERROR;
	}

	if ((sniffer = sx1239_open(sniffer_spi, sniffer_enable, sniffer_reset)) == NULL) {
		error("Failed to open sniffer");
		return LGW_LBT_ERROR;
	}

	if (sniffer->set_bandwidth(sniffer, SNIFFER_BANDWIDTH) < 0) {
		error("Failed to configure sniffer bandwidth");
		return LGW_LBT_ERROR;
	}

	if (sniffer->set_bitrate(sniffer, SNIFFER_BITRATE) < 0) {
		error("Failed to configure sniffer bit rate");
		return LGW_LBT_ERROR;
	}

#else

	int x, i;
	int32_t val;
	uint32_t freq_offset;
	uint32_t lbt_start_freq;

	/* Check if LBT feature is supported by FPGA */
	x = lgw_fpga_reg_r(LGW_FPGA_FEATURE, &val);
	if (x != LGW_REG_SUCCESS) {
		error("Failed to read FPGA Features register");
		return LGW_LBT_ERROR;
	}
	if (TAKE_N_BITS_FROM((uint8_t)val, 2, 1) != 1) {
		error("No support for LBT in FPGA");
		return LGW_LBT_ERROR;
	}

	/* Get FPGA lowest frequency for LBT channels */
	x = lgw_fpga_reg_r(LGW_FPGA_LBT_INITIAL_FREQ, &val);
	if (x != LGW_REG_SUCCESS) {
		error("Failed to read LBT initial frequency from FPGA");
		return LGW_LBT_ERROR;
	}
	switch (val) {
	case 0:
		lbt_start_freq = 915000000;
		break;
	case 1:
		lbt_start_freq = 863000000;
		break;
	default:
		error("LBT start frequency %d is not supported", val);
		return LGW_LBT_ERROR;
	}

	/* Configure SX127x for FSK */
	x = lgw_setup_sx127x(lbt_start_freq, MOD_FSK, LGW_SX127X_RXBW_100K_HZ, lbt_rssi_offset_dB); /* 200KHz LBT channels */
	if (x != LGW_REG_SUCCESS) {
		error("Failed to configure SX127x for LBT");
		return LGW_LBT_ERROR;
	}

	/* Configure FPGA for LBT */
	val = -2 * lbt_rssi_target_dBm; /* Convert RSSI target in dBm to FPGA register format */
	x = lgw_fpga_reg_w(LGW_FPGA_RSSI_TARGET, val);
	if (x != LGW_REG_SUCCESS) {
		error("Failed to configure FPGA for LBT");
		return LGW_LBT_ERROR;
	}
	/* Set default values for non-active LBT channels */
	for (i = lbt_nb_active_channel; i < LBT_CHANNEL_FREQ_NB; i++) {
		lbt_channel_cfg[i].freq_hz = lbt_start_freq;
		lbt_channel_cfg[i].scan_time_us = 128; /* fastest scan for non-active channels */
	}
	/* Configure FPGA for both active and non-active LBT channels */
	for (i = 0; i < LBT_CHANNEL_FREQ_NB; i++) {
		/* Check input parameters */
		if (lbt_channel_cfg[i].freq_hz < lbt_start_freq) {
			error("LBT channel frequency is out of range (%u)", lbt_channel_cfg[i].freq_hz);
			return LGW_LBT_ERROR;
		}
		if ((lbt_channel_cfg[i].scan_time_us != 128) && (lbt_channel_cfg[i].scan_time_us != 5000)) {
			error("LBT channel scan time is not supported (%u)", lbt_channel_cfg[i].scan_time_us);
			return LGW_LBT_ERROR;
		}
		/* Configure */
		freq_offset = (lbt_channel_cfg[i].freq_hz - lbt_start_freq) / 100E3; /* 100kHz unit */
		x = lgw_fpga_reg_w(LGW_FPGA_LBT_CH0_FREQ_OFFSET + i, (int32_t)freq_offset);
		if (x != LGW_REG_SUCCESS) {
			error("Failed to configure FPGA for LBT channel %d (freq offset)", i);
			return LGW_LBT_ERROR;
		}
		if (lbt_channel_cfg[i].scan_time_us == 5000) { /* configured to 128 by default */
			x = lgw_fpga_reg_w(LGW_FPGA_LBT_SCAN_TIME_CH0 + i, 1);
			if (x != LGW_REG_SUCCESS) {
				error("Failed to configure FPGA for LBT channel %d (freq offset)", i);
				return LGW_LBT_ERROR;
			}
		}
	}

#endif

	return LGW_LBT_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int lbt_start(void)
{
#if !defined(PLATFORM_WIFC)
	int x;

	x = lgw_fpga_reg_w(LGW_FPGA_CTRL_FEATURE_START, 1);
	if (x != LGW_REG_SUCCESS) {
		error("Failed to start LBT FSM");
		return LGW_LBT_ERROR;
	}

#endif

	return LGW_LBT_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int lbt_is_channel_free(struct core *core, struct lgw_pkt_tx_s *pkt_data, uint16_t tx_start_delay, bool *tx_allowed)
{
	int i;
#if !defined(PLATFORM_WIFC)
	int32_t val;
#endif
	uint32_t tx_start_time = 0;
	uint32_t tx_end_time = 0;
	uint32_t delta_time = 0;
	uint32_t sx1301_time = 0;
	uint32_t sx1301_tmst = 0;
	uint32_t lbt_time = 0;
	uint32_t lbt_time1 = 0;
	uint32_t lbt_time2 = 0;
	uint32_t tx_max_time = 0;
	int lbt_channel_decod_1 = -1;
	int lbt_channel_decod_2 = -1;
	uint32_t packet_duration = 0;
	time_t current_time = time(NULL);
	uint32_t pause_tmst = 0;
	time_t pause_time = 0;
#if defined(PLATFORM_WIFC)
	struct timespec tsnow;
	uint64_t tmst_us_1 = 0, tmst_us_2 = 0, tmst_now = 0;
	uint32_t lbt_dsensing1, lbt_dsensing2;
#endif

	/* Check input parameters */
	if ((pkt_data == NULL) || (tx_allowed == NULL)) {
		return LGW_LBT_ERROR;
	}

	/* Check if TX is allowed */
	if (lbt_enable == true) {
		/* TX allowed for LoRa only */
		if (pkt_data->modulation != MOD_LORA) {
			*tx_allowed = false;
			debug("TX is not allowed for this modulation (%x)", pkt_data->modulation);
			return LGW_LBT_SUCCESS;
		}

		/* Select LBT Channel corresponding to required TX frequency */
		lbt_channel_decod_1 = -1;
		lbt_channel_decod_2 = -1;
		if (pkt_data->bandwidth == BW_125KHZ) {
			for (i = 0; i < lbt_nb_active_channel; i++) {
				if (is_equal_freq(pkt_data->freq_hz, lbt_channel_cfg[i].freq_hz) == true) {
					lbt_channel_decod_1 = i;
					lbt_channel_decod_2 = i;
					if (lbt_channel_cfg[i].scan_time_us == 5000) {
						tx_max_time = 4000000; /* 4 seconds */
					} else {		       /* scan_time_us = 128 */
						tx_max_time = 400000;  /* 400 milliseconds */
					}
					break;
				}
			}
		} else if (pkt_data->bandwidth == BW_250KHZ) {
			/* In case of 250KHz, the TX freq has to be in between 2 consecutive channels of 200KHz BW.
                The TX can only be over 2 channels, not more */
			for (i = 0; i < (lbt_nb_active_channel - 1); i++) {
				if ((is_equal_freq(pkt_data->freq_hz, (lbt_channel_cfg[i].freq_hz + lbt_channel_cfg[i + 1].freq_hz) / 2) == true) && ((lbt_channel_cfg[i + 1].freq_hz - lbt_channel_cfg[i].freq_hz) == 200E3)) {
					lbt_channel_decod_1 = i;
					lbt_channel_decod_2 = i + 1;
					if (lbt_channel_cfg[i].scan_time_us == 5000) {
						tx_max_time = 4000000; /* 4 seconds */
					} else {		       /* scan_time_us = 128 */
						tx_max_time = 200000;  /* 200 milliseconds */
					}
					break;
				}
			}
		} else {
			/* Nothing to do for now */
		}

		if (lbt_channel_decod_1 < 0 || lbt_channel_decod_2 < 0) {
			error("LBT: TX frequency is not configured");
			*tx_allowed = false;
			return LGW_LBT_SUCCESS;
		}

#if defined(PLATFORM_WIFC)
		/* Scan channels ( from 3 to 6ms / channel on high IO load ) */
		lbt_scan(lbt_channel_decod_1);
		clock_gettime(CLOCK_MONOTONIC, &tsnow);
		tmst_us_1 = tsnow.tv_sec * 1000000 + tsnow.tv_nsec / 1000;

		if (lbt_channel_decod_1 != lbt_channel_decod_2) {
			lbt_scan(lbt_channel_decod_2);
			clock_gettime(CLOCK_MONOTONIC, &tsnow);
			tmst_us_2 = tsnow.tv_sec * 1000000 + tsnow.tv_nsec / 1000;
		}
#endif

		/* Get SX1301 time at last PPS */
		clock_get_last_pps_counter(core, &sx1301_time);

		/* Get SX1301 current time  */
		clock_get_current_counter(core, &sx1301_tmst);

		switch (pkt_data->tx_mode) {
		case TIMESTAMPED:
			tx_start_time = pkt_data->count_us;
			break;
		case ON_GPS:
			tx_start_time = (sx1301_time + (uint32_t)tx_start_delay + 1000000);
			break;
		case IMMEDIATE:
			error("tx_mode IMMEDIATE is not supported when LBT is enabled");
		/* FALLTHROUGH  */
		default:
			return LGW_LBT_ERROR;
		}

		/* Compute variables without lbt mask */
		packet_duration = lgw_time_on_air(pkt_data) * 1000UL;
		tx_end_time = tx_start_time + packet_duration;

		/* Pre compute pause variables */
		if (tx_max_time == 200000) {
			/* dual channel, 128us carrier sense time */
			if (packet_duration > 3000) {
				pause_tmst = tx_end_time + 2000;
			}

		} else if (tx_max_time == 400000) {
			/* single channel, 128us carrier sense time */
			if (packet_duration > 200000) {
				pause_tmst = tx_end_time + 10 * packet_duration;

			} else if (packet_duration > 6000) {
				pause_tmst = tx_end_time + 2000;
			}

		} else if (tx_max_time == 4000000) {
			/* 5ms carrier sense time */
			pause_tmst = tx_end_time + 50000;
		}

		if (pause_tmst > 0) {
			pause_time = current_time + 60; /* 1mn absolute epoch window */
		}

		/* Update pause time validity */
		for (i = lbt_channel_decod_1; i >= 0 && i <= lbt_channel_decod_2; i++) {
			if ((lbt_channel_cfg[i].pause_time > 0 &&
			     current_time > lbt_channel_cfg[i].pause_time) ||
			    (lbt_channel_cfg[i].pause_tmst > 0 &&
			     sx1301_tmst > lbt_channel_cfg[i].pause_tmst)) {
				lbt_channel_cfg[i].pause_tmst = 0;
				lbt_channel_cfg[i].pause_time = 0;
			}
		}

		/* Manage pause rejection */
		for (i = lbt_channel_decod_1; i >= 0 && i <= lbt_channel_decod_2; i++) {
			if (lbt_channel_cfg[i].pause_time > 0) {
				error("TX request rejected (LBT pause)");
				*tx_allowed = false;
				return LGW_LBT_SUCCESS;
			}
		}

		/* Apply lbt mask */
		tx_start_time &= LBT_TIMESTAMP_MASK;
		tx_end_time &= LBT_TIMESTAMP_MASK;

/* Get last time when selected channel was free */

#if defined(PLATFORM_WIFC)
		if ((lbt_channel_decod_1 >= 0) && (lbt_channel_decod_2 >= 0)) {
			lbt_time = lbt_time1 = 0;
			if (lbt_counter[lbt_channel_decod_1] == 0) {
				clock_gettime(CLOCK_MONOTONIC, &tsnow);
				tmst_now = tsnow.tv_sec * 1000000 + tsnow.tv_nsec / 1000;
				lbt_dsensing1 = tmst_now - tmst_us_1;
				lbt_time = lbt_time1 = (sx1301_tmst - lbt_dsensing1) & LBT_TIMESTAMP_MASK;
			}

			if (lbt_channel_decod_1 != lbt_channel_decod_2) {
				lbt_time2 = 0;
				if (lbt_counter[lbt_channel_decod_2] == 0) {
					lbt_dsensing2 = tmst_now - tmst_us_2;
					lbt_time = lbt_time2 = (sx1301_tmst - lbt_dsensing2) & LBT_TIMESTAMP_MASK;

					if (lbt_time2 < lbt_time1) {
						lbt_time = lbt_time2;
					}
				}
			}
		} else {
			lbt_time = 0;
		}

#else

		if ((lbt_channel_decod_1 >= 0) && (lbt_channel_decod_2 >= 0)) {
			lgw_fpga_reg_w(LGW_FPGA_LBT_TIMESTAMP_SELECT_CH, (int32_t)lbt_channel_decod_1);
			lgw_fpga_reg_r(LGW_FPGA_LBT_TIMESTAMP_CH, &val);
			lbt_time = lbt_time1 = (uint32_t)(val & 0x0000FFFF) * 256; /* 16bits (1LSB = 256µs) */

			if (lbt_channel_decod_1 != lbt_channel_decod_2) {
				lgw_fpga_reg_w(LGW_FPGA_LBT_TIMESTAMP_SELECT_CH, (int32_t)lbt_channel_decod_2);
				lgw_fpga_reg_r(LGW_FPGA_LBT_TIMESTAMP_CH, &val);
				lbt_time2 = (uint32_t)(val & 0x0000FFFF) * 256; /* 16bits (1LSB = 256µs) */

				if (lbt_time2 < lbt_time1) {
					lbt_time = lbt_time2;
				}
			}
		} else {
			lbt_time = 0;
		}

#endif

		if (lbt_time < tx_end_time) {
			delta_time = tx_end_time - lbt_time;
		} else {
			delta_time = (LBT_TIMESTAMP_MASK - lbt_time) + tx_end_time;
		}

		debug("lbt: T0=%u Tsensing=%u Dwindow=%u Tstart=%u Dpacket=%u Delta=%i",
		      sx1301_time & LBT_TIMESTAMP_MASK,
		      lbt_time - 2048,
		      tx_max_time,
		      tx_start_time,
		      packet_duration,
		      delta_time);

		/* send data if allowed */
		/* lbt_time: last time when channel was free */
		/* tx_max_time: maximum time allowed to send packet since last free time */
		/* 2048: some margin */
		if ((delta_time < (tx_max_time - 2048)) && (lbt_time != 0)) {
			*tx_allowed = true;

			/* Update pause time validity */
			for (i = lbt_channel_decod_1; i <= lbt_channel_decod_2; i++) {
				lbt_channel_cfg[i].pause_tmst = pause_tmst;
				lbt_channel_cfg[i].pause_time = pause_time;
			}
		} else {
			error("TX request rejected (LBT)");
			*tx_allowed = false;
		}
	} else {
		/* Always allow if LBT is disabled */
		*tx_allowed = true;
	}

	return LGW_LBT_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool lbt_is_enabled(void)
{
	return lbt_enable;
}

uint32_t lbt_get_lowest_frequency(void)
{
	uint32_t freq_hz = UINT32_MAX;
	int i;

	if (lbt_enable) {
		for (i = 0; i < lbt_nb_active_channel; i++) {
			if (lbt_channel_cfg[i].freq_hz < freq_hz) {
				freq_hz = lbt_channel_cfg[i].freq_hz;
			}
		}

		return freq_hz;
	}

	return 0;
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

/* As given frequencies have been converted from float to integer, some aliasing
issues can appear, so we can't simply check for equality, but have to take some
margin */
bool is_equal_freq(uint32_t a, uint32_t b)
{
	int64_t diff;
	int64_t a64 = (int64_t)a;
	int64_t b64 = (int64_t)b;

	/* Calculate the difference */
	diff = llabs(a64 - b64);

	/* Check for acceptable diff range */
	if (diff <= 10000) {
		return true;
	}

	return false;
}

/* --- EOF ------------------------------------------------------------------ */
