/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Host specific functions to address the LoRa concentrator registers through
    a SPI interface.
    Single-byte read/write and burst read/write.
    Does not handle pagination.
    Could be used with multiple SPI ports in parallel (explicit file descriptor)

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Sylvain Miermont
*/

#include "spi.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "firmware/fpga/lbt_863.h"
#include "firmware/fpga/lbt_915.h"
#include "firmware/fpga/spectral_scan.h"
#include "lbt.h"

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "drivers/eeprom.h"
#include "drivers/gpio.h"
#include "drivers/spi.h"
#include "drivers/spimux.h"

#define REG_ADDR_MASK 0x7F
#define REG_READ 0x00
#define REG_WRITE 0x80

#define DEMOD_SPI_SPEED 2000000
#define FPGA_SPI_SPEED 2000000
#define EEPROM_SPI_SPEED 2000000

#if defined(PLATFORM_WIIS) || defined(PLATFORM_FEVO)
#define DEMOD_SPIDEV "/dev/spidev0.0"
#define FPGA_GPIO_CDONE 130  /* Was 141 on wiis v1 */
#define FPGA_GPIO_CRESET 129 /* Was 140 on wiis v1 */
#define FPGA_GPIO_CSN 96
#define FPGA_SPIDEV "/dev/spidev1.0"
#define RADIO_GPIO_ENABLE 102
#endif

#if defined(PLATFORM_WIFC)
#define DEMOD_SPIDEV "/dev/spidev0.0"
#define FPGA_GPIO_CDONE 130
#define FPGA_GPIO_CRESET 129
#define FPGA_GPIO_CSN 128
#define FPGA_SPIDEV "/dev/spidev1.1"
#define RADIO_GPIO_ENABLE 102
#endif

#if defined(PLATFORM_WSTA)
#define DEMOD_SPIDEV "/dev/spidev32766.0"
#define FPGA_GPIO_CDONE 24
#define FPGA_GPIO_CRESET 27
#define FPGA_GPIO_CSN 28
#define FPGA_SPIDEV "/dev/spidev32766.1"
#define RADIO_GPIO_CLOCK_ENABLE 7
#define RADIO_GPIO_ENABLE 6
#endif

struct lgw_spi {
	struct gpio_device *fpga_cdone;
	struct gpio_device *fpga_creset;
	struct gpio_device *fpga_csn;
	struct gpio_device *radio_clock_enable;
	struct gpio_device *radio_enable;

	struct spi_device *sx1301_spi;
	struct spi_device *fpga_mux_spi;

	struct spimux_device *fpga_mux;

	struct spi_device *fpga_spi;
	struct spi_device *eeprom_spi;
	struct spi_device *sx127x_spi;

	struct eeprom_device *eeprom;
};

static int radio_enable(struct lgw_spi *lgw_spi, int enable)
{
	assert(lgw_spi);

	debug("Powering %s the radio...", (enable) ? "on" : "off");

	if (lgw_spi->radio_enable->set_value(lgw_spi->radio_enable, enable) < 0) {
		return -1;
	}

	if (lgw_spi->radio_clock_enable) {
		if (lgw_spi->radio_clock_enable->set_value(lgw_spi->radio_clock_enable, GPIO_HIGH) < 0) {
			return -1;
		}

		if (lgw_spi->radio_clock_enable->set_value(lgw_spi->radio_clock_enable, GPIO_LOW) < 0) {
			return -1;
		}
	}

	sleep_ms(100);

	return 0;
}

static int fpga_program_firmware(struct lgw_spi *lgw_spi, const void *buf, size_t len)
{

	struct spi_msg msg[2];
	uint8_t dummy[12];

	assert(lgw_spi);

	if (lgw_spi->fpga_csn->set_value(lgw_spi->fpga_csn, GPIO_LOW) < 0) {
		error("Failed to reset fpga csn gpio");
		return -1;
	}

	if (lgw_spi->fpga_creset->set_value(lgw_spi->fpga_creset, GPIO_LOW) < 0) {
		error("Failed to reset fpga creset gpio");
		return -1;
	}

	sleep_ns(200);

	if (lgw_spi->fpga_creset->set_value(lgw_spi->fpga_creset, GPIO_HIGH) < 0) {
		error("Failed to set fpga creset gpio mode");
		return -1;
	}

	sleep_us(800);

	memset(msg, 0, sizeof(msg));
	memset(dummy, 0X0F, sizeof(dummy));

	msg[0].mosi = buf;
	msg[0].len = len;
	msg[0].mode = SPI_MODE_3;
	msg[1].mosi = dummy;
	msg[1].len = sizeof(dummy);
	msg[1].mode = SPI_MODE_3;

	if (lgw_spi->fpga_mux_spi->msg(lgw_spi->fpga_mux_spi, msg, sizeofa(msg)) < 0) {
		error("Failed to send fpga firmware");
		return -1;
	}

	if (lgw_spi->fpga_cdone->get_value(lgw_spi->fpga_cdone) != GPIO_HIGH) {
		error("Failed to get fpga cdone gpio");
		return -1;
	}

	return 0;
}

static int fpga_program(struct lgw_spi *lgw_spi)
{
	if (!lbt_is_enabled()) {
		info("Programming FPGA with spectral scan firmware");
		return fpga_program_firmware(lgw_spi, fpga_spectral_scan, sizeof(fpga_spectral_scan));

	} else if (lbt_get_lowest_frequency() > 915000000) {
		info("Programming FPGA with LBT 915 MHz firmware");
		return fpga_program_firmware(lgw_spi, fpga_lbt_915, sizeof(fpga_lbt_915));

	} else if (lbt_get_lowest_frequency() > 863000000) {
		info("Programming FPGA with LBT 863 MHz firmware");
		return fpga_program_firmware(lgw_spi, fpga_lbt_863, sizeof(fpga_lbt_863));

	} else {
		error("Failed to find a suitable fpga firmware");
		return -1;
	}
}

int lgw_spi_open(void **pdevice)
{
	struct lgw_spi *lgw_spi = NULL;

	assert(pdevice);

	if ((lgw_spi = zalloc(sizeof(*lgw_spi))) == NULL) {
		eerror("Failed to allocate memory");
		return -1;
	}

	*pdevice = lgw_spi;

	if ((lgw_spi->fpga_cdone = gpio_open(FPGA_GPIO_CDONE, GPIO_INPUT)) == NULL) {
		error("Failed to open fpga cdone gpio");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->fpga_creset = gpio_open(FPGA_GPIO_CRESET, GPIO_OUTPUT | GPIO_LOW)) == NULL) {
		error("Failed to open fpga creset gpio");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->fpga_csn = gpio_open(FPGA_GPIO_CSN, GPIO_OUTPUT | GPIO_HIGH)) == NULL) {
		error("Failed to open fpga csn gpio");
		lgw_spi_close(lgw_spi);
		return -1;
	}

#ifdef RADIO_GPIO_CLOCK_ENABLE
	if ((lgw_spi->radio_clock_enable = gpio_open(RADIO_GPIO_CLOCK_ENABLE, GPIO_OUTPUT | GPIO_LOW)) == NULL) {
		error("Failed to open radio enable clock gpio");
		lgw_spi_close(lgw_spi);
		return -1;
	}
#endif

	if ((lgw_spi->radio_enable = gpio_open(RADIO_GPIO_ENABLE, GPIO_OUTPUT | GPIO_LOW)) == NULL) {
		error("Failed to open radio enable gpio");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->sx1301_spi = spi_open(DEMOD_SPIDEV, NULL)) == NULL ||
	    lgw_spi->sx1301_spi->set_speed(lgw_spi->sx1301_spi, DEMOD_SPI_SPEED) < 0 ||
	    lgw_spi->sx1301_spi->set_mode(lgw_spi->sx1301_spi, SPI_MODE_0 | SPI_MSB_FIRST | SPI_WORD_8BITS) < 0) {
		error("Failed to open sx1301 spi");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->fpga_mux_spi = spi_open(FPGA_SPIDEV, lgw_spi->fpga_csn)) == NULL ||
	    lgw_spi->fpga_mux_spi->set_speed(lgw_spi->fpga_mux_spi, FPGA_SPI_SPEED) < 0) {
		error("Failed to open fpga mux spi");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->fpga_mux = spimux_open(lgw_spi->fpga_mux_spi)) == NULL) {
		error("Failed to open fpga mux device");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->fpga_spi = lgw_spi->fpga_mux->get_spi(lgw_spi->fpga_mux, LGW_SPI_MUX_TARGET_FPGA)) == NULL) {
		error("Failed to get fpga spi");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->eeprom_spi = lgw_spi->fpga_mux->get_spi(lgw_spi->fpga_mux, LGW_SPI_MUX_TARGET_EEPROM)) == NULL ||
	    lgw_spi->eeprom_spi->set_speed(lgw_spi->eeprom_spi, EEPROM_SPI_SPEED) < 0) {
		error("Failed to get eeprom spi");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->sx127x_spi = lgw_spi->fpga_mux->get_spi(lgw_spi->fpga_mux, LGW_SPI_MUX_TARGET_SX127X)) == NULL) {
		error("Failed to get sx127x spi");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if ((lgw_spi->eeprom = eeprom_open(lgw_spi->eeprom_spi)) == NULL) {
		error("Failed to open eeprom device");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if (radio_enable(lgw_spi, 0) < 0 ||
	    radio_enable(lgw_spi, 1) < 0) {
		error("Failed to reset radio power");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	if (fpga_program(lgw_spi) < 0) {
		error("Failed to program fpga");
		lgw_spi_close(lgw_spi);
		return -1;
	}

	return 0;
}

int lgw_spi_close(void *dev)
{
	struct lgw_spi *lgw_spi = (struct lgw_spi *)dev;

	assert(dev);

	if (lgw_spi) {
		if (lgw_spi->eeprom) {
			lgw_spi->eeprom->close(lgw_spi->eeprom);
		}

		if (lgw_spi->sx127x_spi) {
			lgw_spi->sx127x_spi->close(lgw_spi->sx127x_spi);
		}

		if (lgw_spi->eeprom_spi) {
			lgw_spi->eeprom_spi->close(lgw_spi->eeprom_spi);
		}

		if (lgw_spi->fpga_spi) {
			lgw_spi->fpga_spi->close(lgw_spi->fpga_spi);
		}

		if (lgw_spi->fpga_mux) {
			lgw_spi->fpga_mux->close(lgw_spi->fpga_mux);
		}

		if (lgw_spi->fpga_mux_spi) {
			lgw_spi->fpga_mux_spi->close(lgw_spi->fpga_mux_spi);
		}

		if (lgw_spi->sx1301_spi) {
			lgw_spi->sx1301_spi->close(lgw_spi->sx1301_spi);
		}

		if (lgw_spi->radio_enable) {
			radio_enable(dev, 0);
			lgw_spi->radio_enable->close(lgw_spi->radio_enable);
		}

		if (lgw_spi->radio_clock_enable) {
			lgw_spi->radio_clock_enable->close(lgw_spi->radio_clock_enable);
		}

		if (lgw_spi->fpga_csn) {
			lgw_spi->fpga_csn->close(lgw_spi->fpga_csn);
		}

		if (lgw_spi->fpga_creset) {
			lgw_spi->fpga_creset->close(lgw_spi->fpga_creset);
		}

		if (lgw_spi->fpga_cdone) {
			lgw_spi->fpga_cdone->close(lgw_spi->fpga_cdone);
		}

		free(lgw_spi);
	}

	return 0;
}

static int lgw_spi_msg(void *dev, uint8_t target, struct spi_msg *msg, size_t len)
{
	struct lgw_spi *lgw_spi = (struct lgw_spi *)dev;

	assert(dev);
	assert(msg && len);

	switch (target) {
	case LGW_SPI_MUX_TARGET_SX1301:
		return lgw_spi->sx1301_spi->msg(lgw_spi->sx1301_spi, msg, len);

	case LGW_SPI_MUX_TARGET_FPGA:
		return lgw_spi->fpga_spi->msg(lgw_spi->fpga_spi, msg, len);

	case LGW_SPI_MUX_TARGET_SX127X:
		return lgw_spi->sx127x_spi->msg(lgw_spi->sx127x_spi, msg, len);

	default:
		error("Invalid spi target");
		return -1;
	}
}

int lgw_spi_wb(void *dev, uint8_t mode, uint8_t target, uint8_t addr, const void *buf, size_t len)
{
	struct lgw_spi *lgw_spi = (struct lgw_spi *)dev;
	struct spi_msg msg[2];
	uint8_t hdr;

	(void)mode;

	assert(dev);
	assert(buf && len);

	if (target == LGW_SPI_MUX_TARGET_EEPROM) {
		return lgw_spi->eeprom->write(lgw_spi->eeprom, addr, buf, len);
	}

	assert(!(addr & ~REG_ADDR_MASK));

	memset(msg, 0, sizeof(msg));

	hdr = REG_WRITE | (addr & REG_ADDR_MASK);

	msg[0].mosi = &hdr;
	msg[0].len = sizeof(hdr);
	msg[0].mode = SPI_MSG_HEADER;
	msg[1].mosi = buf;
	msg[1].len = len;

	return lgw_spi_msg(dev, target, msg, sizeofa(msg));
}

int lgw_spi_rb(void *dev, uint8_t mode, uint8_t target, uint8_t addr, void *buf, size_t len)
{
	struct lgw_spi *lgw_spi = (struct lgw_spi *)dev;
	struct spi_msg msg[2];
	uint8_t hdr;

	(void)mode;

	assert(dev);
	assert(buf && len);

	if (target == LGW_SPI_MUX_TARGET_EEPROM) {
		return lgw_spi->eeprom->read(lgw_spi->eeprom, addr, buf, len);
	}

	assert(!(addr & ~REG_ADDR_MASK));

	memset(msg, 0, sizeof(msg));

	hdr = REG_READ | (addr & REG_ADDR_MASK);

	msg[0].mosi = &hdr;
	msg[0].len = sizeof(hdr);
	msg[0].mode = SPI_MSG_HEADER;
	msg[1].miso = buf;
	msg[1].len = len;

	return lgw_spi_msg(dev, target, msg, sizeofa(msg));
}

int lgw_spi_w(void *dev, uint8_t mode, uint8_t target, uint8_t addr, uint8_t val)
{
	return lgw_spi_wb(dev, mode, target, addr, &val, sizeof(val));
}

int lgw_spi_r(void *dev, uint8_t mode, uint8_t target, uint8_t addr, uint8_t *val)
{
	return lgw_spi_rb(dev, mode, target, addr, val, sizeof(*val));
}
