/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Auxiliary functions for LoRa Gateway v2

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <math.h>   /* round */
#include <stdint.h> /* C99 types */
#include <time.h>   /* clock_nanosleep */

#include "aux.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

void sx1301ar_wait_us(uint32_t delay_us)
{
	struct timespec dly;
	struct timespec rem;

	dly.tv_sec = delay_us / 1000000;
	dly.tv_nsec = (delay_us % 1000000) * 1000;

	while ((dly.tv_sec > 0) || (dly.tv_nsec > 1000)) {
		/*
        rem is set ONLY if clock_nanosleep is interrupted (eg. by a signal).
        Must be zeroed each time or will get into an infinite loop after an IT.
        */
		rem.tv_sec = 0;
		rem.tv_nsec = 0;
		clock_nanosleep(CLOCK_MONOTONIC, 0, &dly, &rem);
		dly = rem;
	}

	return;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void sx1301ar_wait_ms(uint32_t delay_ms)
{
	sx1301ar_wait_us(delay_ms * 1000);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

unsigned long long sx1301ar_timestamp(void)
{
	struct timeval tm;
	unsigned long long time_ms;

	gettimeofday(&tm, NULL);

	time_ms = tm.tv_sec * 1E6 + tm.tv_usec;

	return time_ms;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void sx1301ar_calculate_fletcher_checksum(uint8_t *data, size_t size, uint8_t *crc_a, uint8_t *crc_b)
{
	unsigned int i;
	uint8_t ck_a, ck_b;

	ck_a = 0;
	ck_b = 0;
	for (i = 0; i < size; i++) {
		ck_a = ck_a + data[i];
		ck_b = ck_b + ck_a;
	}

	*crc_a = ck_a;
	*crc_b = ck_b;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

uint32_t bytes_be_to_uint32_le(uint8_t *bytes)
{
	uint32_t val = 0;

	if (bytes != NULL) {
		/* Big endian to Little Endian */
		val = bytes[0] << 24;
		val |= bytes[1] << 16;
		val |= bytes[2] << 8;
		val |= bytes[3];
	}

	return val;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

uint32_t bytes_le_to_uint32_le(uint8_t *bytes)
{
	uint32_t val = 0;

	if (bytes != NULL) {
		/* Big endian to Little Endian */
		val = bytes[0];
		val |= bytes[1] << 8;
		val |= bytes[2] << 16;
		val |= bytes[3] << 24;
	}

	return val;
}
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

uint16_t bytes_le_to_uint16_le(uint8_t *bytes)
{
	uint16_t val = 0;

	if (bytes != NULL) {
		/* Little endian to Little Endian */
		val = bytes[0];
		val |= bytes[1] << 8;
	}

	return val;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

uint16_t bytes_be_to_uint16_le(uint8_t *bytes)
{
	uint16_t val = 0;

	if (bytes != NULL) {
		/* Big endian to Little Endian */
		val = bytes[0] << 8;
		val |= bytes[1];
	}

	return val;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int16_t bytes_be_to_int16_le(uint8_t *bytes)
{
	int16_t val = 0;

	if (bytes != NULL) {
		/* Big endian to Little Endian */
		val = bytes[0] << 8;
		val |= bytes[1];
	}

	return val;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

uint64_t bytes_be_to_uint64_le(uint8_t *bytes)
{
	int i;
	union u_convert_uint64_t {
		uint64_t u64;
		uint8_t b[8];
	};
	union u_convert_uint64_t val;

	if (bytes != NULL) {
		/* Convert 8-bytes byte array to uint64_t */
		for (i = 0; i < 8; i++) {
			/* Big endian to Little Endian */
			val.b[i] = bytes[7 - i];
		}
	} else {
		val.u64 = 0;
	}

	return val.u64;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

float bytes_to_float(uint8_t *bytes)
{
	union u_convert_float {
		float f;
		uint8_t b[4];
	};
	union u_convert_float val;

	if (bytes != NULL) {
		val.b[0] = bytes[0];
		val.b[1] = bytes[1];
		val.b[2] = bytes[2];
		val.b[3] = bytes[3];
	} else {
		val.f = 0.0;
	}

	return val.f;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int16_t chip_if_freq2reg(int32_t if_freq, int16_t *p_freq_bias)
{
	/*
    13 bits signed reg: -4096 < reg value < 4095
    register value = 512 -> -1 MHz IF

    reg = Freq_Hz * 512 / 1e6
        = Freq_Hz * 2**3 / 15625
    */
	double if_freq_reg, if_freq_reg_rd;

	if ((if_freq <= 16000000) && (if_freq >= -15996094)) {
		if_freq_reg = (double)(if_freq << 3) / 15625;
		if_freq_reg_rd = round(if_freq_reg);
		if (p_freq_bias != NULL) {
			*p_freq_bias = (int32_t)round(((if_freq_reg_rd - if_freq_reg) * 15625) / 32.0);
		}
		return (int16_t)if_freq_reg_rd;
	} else {
		if (p_freq_bias != NULL) {
			*p_freq_bias = 0;
		}
		return 0x8000; /* error */
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int32_t chip_if_reg2freq(int16_t if_reg)
{
	/*
    13 bits signed reg: -4096 < reg value < 4095
    register value = 512 -> -1 MHz IF

    Freq_Hz = -reg * 1e6 / 512
            = -reg * 15625 / 2**3
    */
	if ((if_reg <= 4095) && (if_reg >= -4096)) {
		return (int32_t)((-if_reg * 15625) >> 3);
	} else {
		return 0x80000000; /* error */
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int16_t chan_if_freq2reg(int32_t if_freq, int16_t *p_freq_bias)
{
	/*
    13 bits signed reg: -4096 < reg value < 4095
    register value = 256 -> -125 kHz IF

    reg = -Freq_Hz * 256 / 125e3
        = -Freq_Hz * 2**5 / 15625
    */
	double if_freq_reg, if_freq_reg_rd;

	if ((if_freq <= 2000000) && (if_freq >= -1999512)) {
		if_freq_reg = (double)(-if_freq << 5) / 15625;
		if_freq_reg_rd = round(if_freq_reg);
		if (p_freq_bias != NULL) {
			*p_freq_bias = (int32_t)round(((if_freq_reg_rd - if_freq_reg) * 15625) / 32.0);
		}
		return (int16_t)if_freq_reg_rd;
	} else {
		if (p_freq_bias != NULL) {
			*p_freq_bias = 0;
		}
		return 0x8000; /* error */
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#if 0
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
void _meas_time_start(struct timeval *tm)
{
#if (DEBUG_PERF > 0) && (DEBUG_PERF <= 4)
	gettimeofday(tm, NULL);
#endif
}

void _meas_time_stop(int debug_level, struct timeval start_time, const char *str)
{
#if (DEBUG_PERF > 0) && (DEBUG_PERF <= 4)
	struct timeval tm;
	double time_ms;
	char *indent[] = {"", " ..", " ....", " ......"};

	gettimeofday(&tm, NULL);

	time_ms = (tm.tv_sec - start_time.tv_sec) * 1000.0 + (tm.tv_usec - start_time.tv_usec) / 1000.0;
	if ((debug_level > 0) && (debug_level <= DEBUG_PERF)) {
		debug("PERF:%s %s %f ms\n", indent[debug_level - 1], str, time_ms);
	}
#endif
}
#pragma GCC diagnostic pop
#endif

/* --- EOF ------------------------------------------------------------------ */
