/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    DSP functions for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <inttypes.h> /* PRIu64 */
#include <math.h>     /* sin, cos, sqrt */
#include <stdbool.h>  /* bool type */
#include <stdint.h>   /* C99 types */
#include <stdio.h>    /* NULL */
#include <string.h>   /* memcpy */

#include "aes.h"
#include "aux.h"
#include "dsp.h"
#include "err.h"
#include "hal.h"
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#define GWLOC_MATCHING_MARGIN_US 128      /* Margin for timestamps matching, in microseconds */
#define GWLOC_PURGE_THRESHOLD_SF7 400E3   /* Number of µs after which a SF7 pkt's fine timestamp is obsolete */
#define GWLOC_PURGE_THRESHOLD_SF8 650E3   /* Number of µs after which a SF8 pkt's fine timestamp is obsolete */
#define GWLOC_PURGE_THRESHOLD_SF9 700E3   /* Number of µs after which a SF9 pkt's fine timestamp is obsolete */
#define GWLOC_PURGE_THRESHOLD_SF10 700E3  /* Number of µs after which a SF10 pkt's fine timestamp is obsolete */
#define GWLOC_PURGE_THRESHOLD_SF11 1500E3 /* Number of µs after which a SF11 pkt's fine timestamp is obsolete */
#define GWLOC_PURGE_THRESHOLD_SF12 3000E3 /* Number of µs after which a SF12 pkt's fine timestamp is obsolete */
#define GWLOC_DECRYPT_FINE_TIMESTAMPS 1

static uint32_t nb_match = 0;

typedef struct {
	uint32_t hs;
	uint32_t hspps;
	uint32_t hsapps;
	uint32_t sapps;
	uint32_t sx1301pps;
	bool isvalid;
} gwloc_hstime_ref_t;

/* AES variables */
#if GWLOC_DECRYPT_FINE_TIMESTAMPS
extern uint8_t brd_aes_key[SX1301AR_MAX_BOARD_NB][SX1301AR_BOARD_AES_KEY_SIZE / 8];
unsigned char aes_dec_buff[SX1301AR_BOARD_AES_DATA_SIZE];
aes_context aes_ctx;
#endif /* GWLOC_DECRYPT_FINE_TIMESTAMPS */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

static uint8_t gwloc_dsp_fifo[GWLOC_BOARD_DSP_FIFO_SIZE];	/* raw data buffer containing all messages data from DSP */
static uint8_t gwloc_dsp_fifo_crypto[GWLOC_BOARD_DSP_FIFO_SIZE]; /* raw data buffer containing all messages data from DSP, with encrypted fine timestamps */
static gwloc_hstime_ref_t gwloc_hstime_ref = {0, 0, 0, 0, 0, false};

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

#if DEBUG_DSP
static void gwloc_tmst_print(gwloc_tmst_queue_t *queue);
#endif
static int gwloc_read_counter(uint8_t brd, gwloc_hstime_ref_t *counters);
static int gwloc_cnt2ls(uint32_t cnt, uint32_t *ls);
static int dsp_poll_reg_timeout(uint8_t brd, uint8_t dsp, sx1301ar_breg_t idx, int32_t expected_val, int32_t timeout_ms);
static int dsp_boot(uint8_t brd, uint8_t dsp);

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int gwloc_dsp_boot(uint8_t brd, uint8_t dsp, uint8_t retry)
{
	uint8_t nb_retry = 0;
	int x;

	do {
		if (nb_retry > 0) {
			debug("[BRD%u] DSP%u boot retry %u", brd, dsp, nb_retry);
		}
		x = dsp_boot(brd, dsp);
		nb_retry += 1;
	} while ((nb_retry <= retry) && (x != 0));

	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_stop(uint8_t brd, uint8_t dsp, uint8_t nb_dsp)
{
	int x, i;
	int32_t val;
	uint8_t rst_nb = 0;
	sx1301ar_breg_t dsp_gpio_out, dsp_spi_reg_en, dsp_rstn, dsp_boot_en;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((dsp >= GWLOC_BOARD_DSP_NB) || (nb_dsp > GWLOC_BOARD_DSP_NB)) {
		SX1301AR_ERR_RETURN(ERR_BAD_DSP_NB);
	}

	switch (dsp) {
	case 0:
		dsp_gpio_out = BREG_DSP1_GPIO_OUT;
		dsp_spi_reg_en = BREG_DSP1_SPI_REG_EN;
		dsp_rstn = BREG_DSP1_RSTN;
		dsp_boot_en = BREG_DSP1_BOOT_EN;
		break;
	case 1:
		dsp_gpio_out = BREG_DSP2_GPIO_OUT;
		dsp_spi_reg_en = BREG_DSP2_SPI_REG_EN;
		dsp_rstn = BREG_DSP2_RSTN;
		dsp_boot_en = BREG_DSP2_BOOT_EN;
		break;
	case 2:
		dsp_gpio_out = BREG_DSP3_GPIO_OUT;
		dsp_spi_reg_en = BREG_DSP3_SPI_REG_EN;
		dsp_rstn = BREG_DSP3_RSTN;
		dsp_boot_en = BREG_DSP3_BOOT_EN;
		break;
	}

	/* Put the DSP in a known state */
	x = sx1301ar_reg_brd_w(brd, dsp_boot_en, 0);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_gpio_out, 0);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_spi_reg_en, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Put the DSP in RESET */
	x = sx1301ar_reg_brd_w(brd, dsp_rstn, 0);
	SX1301AR_ERR_PROPAGATE(x);

	debug("[BRD%u] DSP%d STOPPED", brd, dsp);

	/* If all DSPs are in RESET, stop the DSP PLL */
	for (i = 0; i < nb_dsp; i++) {
		switch (i) {
		case 0:
			dsp_rstn = BREG_DSP1_RSTN;
			break;
		case 1:
			dsp_rstn = BREG_DSP2_RSTN;
			break;
		case 2:
			dsp_rstn = BREG_DSP3_RSTN;
			break;
		}
		x = sx1301ar_reg_brd_r(brd, dsp_rstn, &val);
		if (sx1301ar_errno != ERR_BAD_BOARD_REG) /* ignore error if DSP does not exist on this board */
		{
			SX1301AR_ERR_PROPAGATE(x);
		}
		if (x == 0) {
			rst_nb += !val;
		}
	}
	if (rst_nb >= nb_dsp) {
		x = sx1301ar_reg_brd_w(brd, BREG_DSP_PLL_RESET, 1);
		SX1301AR_ERR_PROPAGATE(x);
		debug("[BRD%u] DSP PLL STOPPED", brd);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_fetch(uint8_t brd, uint8_t dsp, gwloc_dsp_msg_t *msg, uint8_t max_nb, uint8_t *nb_msg)
{
	uint16_t sz = 0; /* size of the payload */
	int res;
	sx1301ar_breg_t idx = -1;
	/* parsing variables */
	gwloc_dsp_msg_type_t msg_type;
	uint16_t buffer_index;
	uint16_t nb_msg_found = 0;
	uint16_t payload_length;
	uint8_t ck_a, ck_b;
	uint8_t ck_a_rcv, ck_b_rcv;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (dsp >= GWLOC_BOARD_DSP_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_DSP_NB);
	}
	if ((msg == NULL) || (nb_msg == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Check DSP FIFO fill counter -> number of bytes available */
	switch (dsp) {
	case 0:
		idx = BREG_DSP_1_FIFO_RD_NB_BYTE;
		break;
	case 1:
		idx = BREG_DSP_2_FIFO_RD_NB_BYTE;
		break;
	case 2:
		idx = BREG_DSP_3_FIFO_RD_NB_BYTE;
		break;
	default:
		/* Should not happen */
		break;
	}
	res = sx1301ar_reg_brd_rb(brd, idx, gwloc_dsp_fifo, 2);
	if (res != 0) {
		error("[BRD%u][*DSP%u*] failed to read status (%s)", brd, dsp, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	sz = bytes_le_to_uint16_le(gwloc_dsp_fifo + 0);
	if (sz == 0) {
		return 0;
	}

	/* Get data from FIFO */
	switch (dsp) {
	case 0:
		idx = BREG_DSP_1_FIFO_RD_DATA;
		break;
	case 1:
		idx = BREG_DSP_2_FIFO_RD_DATA;
		break;
	case 2:
		idx = BREG_DSP_3_FIFO_RD_DATA;
		break;
	default:
		/* Should not happen */
		break;
	}
	res = sx1301ar_reg_brd_rb(brd, idx, gwloc_dsp_fifo, sz);
	if (res != 0) {
		error("[BRD%u][*DSP%u*] failed to read data", brd, dsp);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Parse raw data and fill messages array */
	buffer_index = 0;
	while ((nb_msg_found <= max_nb) && (buffer_index < sz)) {
		/* Get message sync words */
		if ((gwloc_dsp_fifo[buffer_index] != GWLOC_MSG_SYNCWORD_1) ||
		    (gwloc_dsp_fifo[buffer_index + 1] != GWLOC_MSG_SYNCWORD_2)) {
			buffer_index++;
			continue;
		}

		/* Get message type */
		msg_type = (gwloc_dsp_msg_type_t)gwloc_dsp_fifo[buffer_index + 2];
		if ((msg_type == 0) || (msg_type >= GWLOC_DSP_MSG_TYPE_NB)) {
			error("[BRD%u][*DSP%d*] wrong message type, continue...", brd, dsp);
			buffer_index++;
			continue;
		} else {
			msg[nb_msg_found].id = msg_type;
		}

		/* Get message size - assuming big endian format */
		payload_length = bytes_be_to_uint16_le(gwloc_dsp_fifo + buffer_index + 3);
		if ((buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length + GWLOC_DSP_MSG_CRC_SIZE) > sz) {
			warning("[BRD%u][*DSP%d*] aborting truncated message (size=%u), got %u messages", brd, dsp, sz, nb_msg_found);
			break;
		} else {
			msg[nb_msg_found].payload_length = payload_length;
		}

		/* CRC check */
		/* Get CRC from received message */
		ck_a_rcv = gwloc_dsp_fifo[buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length];
		ck_b_rcv = gwloc_dsp_fifo[buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length + 1];
		/* Calculate theoritical CRC of actual payload */
		sx1301ar_calculate_fletcher_checksum(gwloc_dsp_fifo + buffer_index + GWLOC_DSP_MSG_HEADER_SIZE, payload_length, &ck_a, &ck_b);

		/* Compare checksums and parse if OK */
		if ((ck_a != ck_a_rcv) || (ck_b != ck_b_rcv)) {
			error("[BRD%u][*DSP%d*] wrong CRC, continue...", brd, dsp);
			buffer_index++;
			/* continue parsing ... */
			continue;
		} else {
			/* store message info */
			msg[nb_msg_found].crc_a = ck_a_rcv;
			msg[nb_msg_found].crc_b = ck_b_rcv;
			msg[nb_msg_found].payload = gwloc_dsp_fifo + buffer_index + GWLOC_DSP_MSG_HEADER_SIZE;
			nb_msg_found += 1;
			/* move buffer index toward next message */
			buffer_index += (GWLOC_DSP_MSG_HEADER_SIZE + payload_length + GWLOC_DSP_MSG_CRC_SIZE);
		}
	}

	/* If we have at least one message, copy the complete buffer content */
	if (nb_msg_found > 0) {
		*nb_msg = nb_msg_found;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_fpga_fetch(uint8_t brd, uint8_t dsp, gwloc_dsp_msg_t *msg, uint8_t max_nb, uint8_t *nb_msg)
{
	sx1301ar_breg_t idx = -1;
	uint16_t sz = 0; /* size of the payload */
	int res;
	/* parsing variables */
	gwloc_dsp_msg_type_t msg_type;
	uint16_t buffer_index;
	uint16_t nb_msg_found = 0;
	uint16_t payload_length;
	uint8_t ck_a, ck_b;
	uint8_t ck_a_rcv, ck_b_rcv;
	/* performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (dsp >= GWLOC_BOARD_DSP_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_DSP_NB);
	}
	if ((msg == NULL) || (nb_msg == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Check DSP FIFO fill counter -> number of bytes available */
	switch (dsp) {
	case 0:
		idx = BREG_DSP_1_FIFO_RD_NB_BYTE_CRYPTO;
		break;
	case 1:
		idx = BREG_DSP_2_FIFO_RD_NB_BYTE_CRYPTO;
		break;
	case 2:
		idx = BREG_DSP_3_FIFO_RD_NB_BYTE_CRYPTO;
		break;
	default:
		/* Should not happen */
		break;
	}
	res = sx1301ar_reg_brd_rb(brd, idx, gwloc_dsp_fifo_crypto, 2);
	if (res != 0) {
		error("[BRD%u] failed to read FPGA FIFO status", brd);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	sz = bytes_le_to_uint16_le(gwloc_dsp_fifo_crypto + 0);
	if (sz == 0) {
		return 0;
	}

	/* Get data from FIFO */
	switch (dsp) {
	case 0:
		idx = BREG_DSP_1_FIFO_RD_DATA_CRYPTO;
		break;
	case 1:
		idx = BREG_DSP_2_FIFO_RD_DATA_CRYPTO;
		break;
	case 2:
		idx = BREG_DSP_3_FIFO_RD_DATA_CRYPTO;
		break;
	default:
		/* Should not happen */
		break;
	}
	res = sx1301ar_reg_brd_rb(brd, idx, gwloc_dsp_fifo_crypto, sz);
	if (res != 0) {
		error("[BRD%u][*DSP%u*] failed to read FPGA FIFO data (%s)", brd, dsp, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Parse raw data and fill messages array */
	buffer_index = 0;
	while ((nb_msg_found <= max_nb) && (buffer_index < sz)) {
		/* Get message sync words */
		if ((gwloc_dsp_fifo_crypto[buffer_index] != GWLOC_MSG_SYNCWORD_1) ||
		    (gwloc_dsp_fifo_crypto[buffer_index + 1] != GWLOC_MSG_SYNCWORD_2)) {
			buffer_index++;
			continue;
		}

		/* Get message type */
		msg_type = (gwloc_dsp_msg_type_t)gwloc_dsp_fifo_crypto[buffer_index + 2];
		if ((msg_type == 0) || (msg_type >= GWLOC_DSP_MSG_TYPE_NB)) {
			error("[BRD%u] [*FPGA*] wrong message type, continue...", brd);
			buffer_index++;
			continue;
		} else {
			msg[nb_msg_found].id = msg_type;
		}

		/* Get message size - assuming big endian format */
		payload_length = bytes_be_to_uint16_le(gwloc_dsp_fifo_crypto + buffer_index + 3);
		if ((buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length + GWLOC_DSP_MSG_CRC_SIZE) > sz) {
			warning("[BRD%u][*FPGA*] aborting truncated message (size=%u), got %u messages", brd, sz, nb_msg_found);
			break;
		} else {
			msg[nb_msg_found].payload_length = payload_length;
		}

		/* CRC check */
		/* Get CRC from received message */
		ck_a_rcv = gwloc_dsp_fifo_crypto[buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length];
		ck_b_rcv = gwloc_dsp_fifo_crypto[buffer_index + GWLOC_DSP_MSG_HEADER_SIZE + payload_length + 1];
		/* Calculate theoritical CRC of actual payload */
		sx1301ar_calculate_fletcher_checksum(gwloc_dsp_fifo_crypto + buffer_index + GWLOC_DSP_MSG_HEADER_SIZE, payload_length, &ck_a, &ck_b);

		/* Compare checksums and parse if OK */
		if ((ck_a != ck_a_rcv) || (ck_b != ck_b_rcv)) {
			error("[BRD%u][*FPGA*] wrong CRC, continue...", brd);
			buffer_index++;
			/* continue parsing ... */
			continue;
		} else {
			/* store message info */
			msg[nb_msg_found].crc_a = ck_a_rcv;
			msg[nb_msg_found].crc_b = ck_b_rcv;
			msg[nb_msg_found].payload = gwloc_dsp_fifo_crypto + buffer_index + GWLOC_DSP_MSG_HEADER_SIZE;
			nb_msg_found += 1;
			/* move buffer index toward next message */
			buffer_index += (GWLOC_DSP_MSG_HEADER_SIZE + payload_length + GWLOC_DSP_MSG_CRC_SIZE);
		}
	}

	/* If we have at least one message, copy the complete buffer content */
	if (nb_msg_found > 0) {
		*nb_msg = nb_msg_found;
	}

	/* Compute time spent in this function */
	_meas_time_stop(2, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_parse_timestamp(uint8_t brd, sx1301ar_btype_t brd_type, uint8_t dsp, gwloc_dsp_msg_t *tmst_msg, gwloc_dsp_tmst_t *tmst)
{
	int i;
	uint16_t buffer_index = 0;
	uint8_t lora_sf;
	/* performances variables */
	struct timeval tm;
	sx1301ar_board_cfg_t brd_cfg;

	/* Check input parameters */
	if ((tmst_msg == NULL) || (tmst == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Record board number */
	tmst->brd = brd;

	/* Get packet modulation rate (SF7...) */
	lora_sf = tmst_msg->payload[buffer_index++];
	switch (lora_sf) {
	case 7:
		tmst->modrate = MR_SF7;
		break;
	case 8:
		tmst->modrate = MR_SF8;
		break;
	case 9:
		tmst->modrate = MR_SF9;
		break;
	case 10:
		tmst->modrate = MR_SF10;
		break;
	case 11:
		tmst->modrate = MR_SF11;
		break;
	case 12:
		tmst->modrate = MR_SF12;
		break;
	default:
		tmst->modrate = MR_UNDEFINED;
		error("unsupported modrate - %d", lora_sf);
		break;
	}

	/* Get channel on which the packet with this fine timestamp has been received */
	tmst->chan = tmst_msg->payload[buffer_index++] + (dsp * 16 * SX1301AR_BOARD_NB_CHIP_PER_DSP);
	tmst->chan += ((brd_type == BRD_SLAVE) ? 16 * 2 : 0); /* Shift channel number for slave board (16 * chip_nb) */ /* FIXME: get nb chips from board context */

	/* Get the fine timestamp at which packet has been received */
	tmst->ts_lsp_ls_int = tmst_msg->payload[buffer_index++] << 16;
	tmst->ts_lsp_ls_int |= tmst_msg->payload[buffer_index++] << 8;
	tmst->ts_lsp_ls_int |= tmst_msg->payload[buffer_index++];

	for (i = 0; i < SX1301AR_BOARD_AES_DATA_SIZE; i++) {
		tmst->ts_lsp_frac_enc[i] = tmst_msg->payload[buffer_index + i];
	}
	buffer_index += 16;

#if GWLOC_DECRYPT_FINE_TIMESTAMPS
	if (sx1301ar_get_conf_board(brd, &brd_cfg)) {
		error("Unable to retrieve board configuration (board #%u)", brd);
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	if (brd_cfg.fine_timestamp_decrypt_enable) {
		uint64_t ts64;

		/* Decrypt fine timestamp */
		aes_set_key(&aes_ctx, brd_aes_key[brd], SX1301AR_BOARD_AES_KEY_SIZE);
		aes_decrypt(&aes_ctx, tmst->ts_lsp_frac_enc, aes_dec_buff);

		ts64 = bytes_be_to_uint64_le(aes_dec_buff + 8);
		tmst->ts_lsp_frac_dec = (uint32_t)(ts64 / 32);

		if (tmst->ts_lsp_frac_dec >= 1E9) {
			tmst->ts_lsp_frac_dec = 0;
		}

	} else {
		tmst->ts_lsp_frac_dec = 0;
	}
#else
	tmst->ts_lsp_frac_dec = 0;
#endif

	/* Get the RSSI of the received packet */
	tmst->rssi_chan = tmst_msg->payload[buffer_index++];

	/* Get the SNR of the received packet */
	tmst->snr = tmst_msg->payload[buffer_index++];

	/* Get the frequency offset of received packet */
	tmst->freq_offset = bytes_to_float(tmst_msg->payload + buffer_index); /* 4 bytes */
	buffer_index += 4;

	/* Get the signal RSSI of the received packet */
	tmst->rssi_sig = tmst_msg->payload[buffer_index++];

	/* Get the signal preambule amplitude standard deviation of the received packet */
	tmst->rssi_sig_std = tmst_msg->payload[buffer_index++];

	/* Get fine timestamp status */
	tmst->status = tmst_msg->payload[buffer_index++];

	/* Get sub band fine timestamp */
	tmst->ts2_lsp_delta = bytes_be_to_int16_le(tmst_msg->payload + buffer_index); /* 2 bytes */
	buffer_index += 2;

	/* Get fine timestamp debug information 1 */
	tmst->ts_debug1 = bytes_be_to_uint16_le(tmst_msg->payload + buffer_index); /* 2 bytes */
	buffer_index += 2;

	/* Get fine timestamp debug information 2 */
	tmst->ts_debug2 = bytes_be_to_uint16_le(tmst_msg->payload + buffer_index); /* 2 bytes */
	buffer_index += 2;

	/* Compute time spent in this function */
	_meas_time_stop(2, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_parse_timestamp_status(uint8_t status, bool *isvalid)
{
	uint8_t ts_validity;
	uint8_t ts_sign;
	uint8_t ts_nb_sec;
	/* performances variables */
	struct timeval tm;

	if (isvalid == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* This status indicates if fine timestamp was positive or negative when calculated,
        and how many seconds have been added/removed to the calculated value to have a fts
        in the range of 0->1second */
	/* The format of the status is as follows:
            7:6 - 00 indicates that FTS was in proper range
                  01 indicates that FTS was greater than 8seconds, which is an error. FTS is set to 0.
                  10 indicates that FTS was lower than -7seconds, which is an error, FTS is set to 0.
            5   - 1 indicates that FTS was negative
                  0 indicates that FTS was positive
            4:0 - Number of seconds increase/decrease to have a FTS in proper range (this reflects how long PPS has been lost */
	ts_validity = TAKE_N_BITS_FROM(status, 6, 2); /* bits 7:6 - fts valid or not */
	switch (ts_validity) {
	case 0:
		*isvalid = true;
		break;
	case 1:
		debug("fine timestamp is not valid (positive)"); /* greater than 8 seconds */
		*isvalid = false;
		break;
	case 2:
		debug("fine timestamp is not valid (negative)"); /* lower than -7 seconds */
		*isvalid = false;
		break;
	default:
		debug("fine timestamp validity unknown (%u)", ts_validity); /* lower than -7 seconds */
		*isvalid = false;
		break;
	}

	if (*isvalid == true) {
		ts_sign = TAKE_N_BITS_FROM(status, 5, 1);   /* bit 5 - fts was positive or negative */
		ts_nb_sec = TAKE_N_BITS_FROM(status, 0, 5); /* bits 4:0 - number of seconds added/removed */
		int adjustment = (ts_sign == 1) ? (-1 * ts_nb_sec) : ts_nb_sec;
		if (adjustment) {
			debug("fine timestamp adjustment: %d", adjustment);
		}
	}

	/* Compute time spent in this function */
	_meas_time_stop(2, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_parse_config(gwloc_dsp_msg_t *msg, gwloc_dsp_cfg_t *cfg)
{
	uint16_t buffer_index = 0;
	int i;

	/* Check input parameters */
	if ((msg == NULL) || (cfg == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Get DSP version */
	cfg->dsp_version = msg->payload[buffer_index++];

	/* Get LoRa syncword (public or private) */
	cfg->lorawan_public = (bool)msg->payload[buffer_index++];

	/* Get channels setting (diversity or 16 channels) */
	cfg->diversity = (bool)msg->payload[buffer_index++];

	/* Fine timestamp debug mode on/off (1 byte) */
	cfg->debug_tmst = msg->payload[buffer_index++];

	/* main fine timestamp version selected (1 byte) */
	cfg->main_tmst_version = msg->payload[buffer_index++];

	/* Get DSP PLL0 frequency */
	cfg->pll0 = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get DSP PLL1 frequency */
	cfg->pll1 = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of channels configured */
	cfg->chan_nb = msg->payload[buffer_index++];

	/* Get channels frequencies */
	for (i = 0; i < (SX1301AR_BOARD_NB_CHIP_PER_DSP * SX1301AR_CHIP_MULTI_NB); i++) {
		cfg->chan_freq[i] = bytes_to_float(msg->payload + buffer_index);
		buffer_index += 4;
	}

	/* Get DSP report period (in minutes) */
	cfg->report_period = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_dsp_parse_status(gwloc_dsp_msg_t *msg, gwloc_dsp_status_t *stat)
{
	uint16_t buffer_index = 0;

	/* Check input parameters */
	if ((msg == NULL) || (stat == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Get log counter */
	stat->log_counter = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of jobs added */
	stat->nb_jobs_added = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of jobs executed */
	stat->nb_jobs_executed = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of valid timestamps */
	stat->nb_valid_timestamps = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of jobs in queue */
	stat->nb_jobs_in_queue = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get maximum number of jobs in queue since last report */
	stat->nb_jobs_max_in_queue = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of sync lost since last report */
	stat->nb_lost_sync = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get accumulated number of sync lost */
	stat->nb_lost_sync_acc = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get number of jobs dropped since last report */
	stat->nb_jobs_dropped = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get accumulated number of jobs dropped */
	stat->nb_jobs_dropped_acc = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get maximum size of SX1301_0 FIFO since last report */
	stat->sx1301_0_max_fifo = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get maximum size of SX1301_1 FIFO since last report */
	stat->sx1301_1_max_fifo = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get maximum size of DSP FIFO since last report */
	stat->dsp_max_fifo = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	/* Get maximum size of DSP Encrypted FIFO since last report */
	stat->dsp_max_enc_fifo = bytes_be_to_uint32_le(msg->payload + buffer_index);
	buffer_index += 4;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool gwloc_tmst_queue_is_full(const gwloc_tmst_queue_t *queue)
{
	return ((queue->num_tmst == GWLOC_TMST_QUEUE_MAX) ? true : false);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool gwloc_tmst_queue_is_empty(const gwloc_tmst_queue_t *queue)
{
	if (queue == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	return ((queue->num_tmst == 0) ? true : false);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void gwloc_tmst_queue_init(gwloc_tmst_queue_t *queue)
{
	int i;

	if (queue != NULL) {
		memset(queue, 0, sizeof(*queue));
		for (i = 0; i < GWLOC_TMST_QUEUE_MAX; i++) {
			queue->nodes[i].is_sent = true; /* indicates that slot is available */
		}
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_tmst_enqueue(gwloc_tmst_queue_t *queue, const gwloc_dsp_tmst_t *fine_tmst)
{
	int i;

	if ((queue == NULL) || (fine_tmst == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if (gwloc_tmst_queue_is_full(queue)) {
		error("%s: queue is full", __FUNCTION__);
		return -1;
	}

	/* enqueue timestamp */
	for (i = 0; i < GWLOC_TMST_QUEUE_MAX; i++) {
		if (queue->nodes[i].is_sent) { /* slot is available */
			memcpy(&(queue->nodes[i].tmst), fine_tmst, sizeof(gwloc_dsp_tmst_t));
			queue->nodes[i].systime = sx1301ar_timestamp();
			queue->nodes[i].age_us = 0;
			queue->nodes[i].is_sent = false;
			queue->num_tmst++;
			break;
		}
	}

#if DEBUG_DSP && 0
	gwloc_tmst_print(queue);
#endif

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_tmst_dequeue(gwloc_tmst_queue_t *queue, int index, gwloc_dsp_tmst_t *fine_tmst)
{
	if ((queue == NULL) || (fine_tmst == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if ((index < 0) || (index >= GWLOC_TMST_QUEUE_MAX)) {
		error("%s: invalid parameter", __FUNCTION__);
		return -1;
	}

	if (gwloc_tmst_queue_is_empty(queue)) {
		error("%s: cannot dequeue packet, queue is empty", __FUNCTION__);
		return -1;
	}

	/* dequeue timestamp */
	memcpy(fine_tmst, &(queue->nodes[index].tmst), sizeof(gwloc_dsp_tmst_t));
	queue->nodes[index].is_sent = true;
	queue->nodes[index].systime = 0;
	queue->nodes[index].age_us = 0;
	queue->num_tmst--;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool gwloc_match_timestamps(const sx1301ar_rx_pkt_t *pkt, gwloc_dsp_tmst_t *fine_tmst, bool diversity)
{
	uint32_t count_diff;
	uint32_t pkt_count_us_corrected;
	uint32_t pkt_count_ls_corrected = 0;
	double pkt_t_payload = 0.0;
	double pkt_t_syncword = 0.0;
	bool ret = false;
	uint32_t idx;

	if ((pkt == NULL) || (fine_tmst == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	for (idx = 0; idx < SX1301AR_BOARD_RFCHAIN_NB; idx++) {
		if (pkt->rsig[idx].is_valid == true) {
			/* we can use this index for matching */
			break;
		}
	}

	/* Try to match SX1301 packet with available fine timestamps in queue based on Channel and modulation rate */
	/* Note: the diversity test below is just for testing purpose, when we want to ba able to have several
             sx1301 configured with same channels with no diversity (ei. full capacity check) */
	if (((pkt->rsig[idx].chan == fine_tmst->chan) || ((abs(pkt->rsig[idx].chan - fine_tmst->chan) == 16) && (diversity == true))) && (pkt->modrate == fine_tmst->modrate)) {
		/* Substract packet header+payload durations to coarse timestamp in order to compare with last preambule symbol fine timestamp */
		sx1301ar_rx_pkt_time_on_air(pkt, NULL, &pkt_t_syncword, &pkt_t_payload);
		pkt_count_us_corrected = pkt->count_us - (uint32_t)((pkt_t_payload + pkt_t_syncword) * 1000.0) + 114; /* no need to cast to int32_t first here, as it can't be negative */
		/* Convert 1MHz/32bits counter to 125KHz/24bits counter */
		gwloc_cnt2ls(pkt_count_us_corrected, &pkt_count_ls_corrected);
		/* Check if it matches */
		count_diff = abs((int32_t)pkt_count_ls_corrected - (int32_t)fine_tmst->ts_lsp_ls_int);
		if (count_diff < (GWLOC_MATCHING_MARGIN_US / 8)) {
			nb_match++;
			ret = true;
		}
	}

	return ret;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_tmst_peek(gwloc_tmst_queue_t *queue, const sx1301ar_rx_pkt_t *pkt, bool diversity, int *index, uint8_t *nb_match_tmst)
{
	int i;

	if ((queue == NULL) || (pkt == NULL) || (index == NULL) || (nb_match_tmst == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Search for coarse/fine timestamps correlation */
	*index = -1;
	*nb_match_tmst = 0;
	if (gwloc_tmst_queue_is_empty(queue)) {
		debug("%s: cannot peek packet, queue is empty", __FUNCTION__);
		return 0;
	}

	for (i = 0; i < GWLOC_TMST_QUEUE_MAX; i++) {
		if (!queue->nodes[i].is_sent) {
			if (gwloc_match_timestamps(pkt, &queue->nodes[i].tmst, diversity)) {
				if (*nb_match_tmst == SX1301AR_BOARD_RFCHAIN_NB) {
					warning("more than %d fine timestamps have been found", SX1301AR_BOARD_RFCHAIN_NB);
					break;
				}
				*index++ = i;
				(*nb_match_tmst)++;
			}
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_tmst_purge(gwloc_tmst_queue_t *queue)
{
	int i;
	int num_tmst_purged = 0;
	uint32_t purge_threshold;
	unsigned long long tmst_age;

	if (queue == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if (gwloc_tmst_queue_is_empty(queue)) {
		return 0;
	}

	/* Search for coarse/fine timestamps correlation */
	for (i = 0; i < GWLOC_TMST_QUEUE_MAX; i++) {
		if (!queue->nodes[i].is_sent) {
			/* Get the purge thrshold time based on SF */
			switch (queue->nodes[i].tmst.modrate) {
			case MR_SF7:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF7;
				break;
			case MR_SF8:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF8;
				break;
			case MR_SF9:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF9;
				break;
			case MR_SF10:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF10;
				break;
			case MR_SF11:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF11;
				break;
			case MR_SF12:
				purge_threshold = GWLOC_PURGE_THRESHOLD_SF12;
				break;
			default:
				error("wrong modulation rate %d in %s", queue->nodes[i].tmst.modrate, __FUNCTION__);
				purge_threshold = 0; /* purge immediatly */
				break;
			}

			/* check if fine timestamp is too old */
			tmst_age = sx1301ar_timestamp() - queue->nodes[i].systime;
			if (tmst_age > (unsigned long long)purge_threshold) {
				queue->nodes[i].is_sent = true;
				queue->nodes[i].systime = 0;
				queue->nodes[i].age_us = 0;
				queue->num_tmst--;
				num_tmst_purged++;
			} else {
				queue->nodes[i].age_us = tmst_age;
			}
		}
	}

	debug("purged %d timestamps from queue, remains %d", num_tmst_purged, queue->num_tmst);

	return num_tmst_purged;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_trighs(uint8_t brd, uint32_t *cnt_hs)
{
	int x;
	int32_t val;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cnt_hs == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Get high speed counter value at last PPS (HSPPS) */
	x = sx1301ar_reg_brd_r(brd, BREG_HIGH_SPEED_TIMESTAMP_GPS, &val);
	SX1301AR_ERR_PROPAGATE(x);

	*cnt_hs = (uint32_t)val;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_read_counter(uint8_t brd, gwloc_hstime_ref_t *counters)
{
	int x;
	int32_t val;
	uint8_t val_b[16]; /* We want to read 4 32-bits registers */
	/* performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (counters == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Read FPGA counters */
	_meas_time_start(&tm);
	x = sx1301ar_reg_brd_rb(brd, BREG_HIGH_SPEED_TIMESTAMP, &val_b[0], sizeof val_b);
	_meas_time_stop(3, tm, "read BREG_HIGH_SPEED_TIMESTAMP");
	SX1301AR_ERR_PROPAGATE(x);

	(*counters).hs = bytes_le_to_uint32_le(val_b + 0);

	(*counters).hspps = bytes_le_to_uint32_le(val_b + 4);

	(*counters).hsapps = bytes_le_to_uint32_le(val_b + 8);

	(*counters).sapps = bytes_le_to_uint32_le(val_b + 12);
	(*counters).sapps &= 0x00FFFFFF; /* 24 bits counter */

	/* Read counter from SX1301 (sampled on GPS PPS) */
	_meas_time_start(&tm);
	x = sx1301ar_reg_chip_r(brd, SX1301AR_GPS_TIMER_CHIP, CREG_TIMESTAMP, &val);
	SX1301AR_ERR_PROPAGATE(x);
	_meas_time_stop(3, tm, "read CREG_TIMESTAMP");
	(*counters).sx1301pps = (uint32_t)val;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_update_ref_counter(uint8_t brd)
{
	int x;
	gwloc_hstime_ref_t ref1 = {0, 0, 0, 0, 0, false};
	gwloc_hstime_ref_t ref2 = {0, 0, 0, 0, 0, false};
	/* performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* invalidate current reference */
	gwloc_hstime_ref.isvalid = false;

	/* Get all high speed counters in one burst, to be sure they are all based on same PPS */
	/*  BREG_HIGH_SPEED_TIMESTAMP
        BREG_HIGH_SPEED_TIMESTAMP_GPS
        BREG_HIGH_SPEED_TIMESTAMP_SAMPLE
        BREG_LATCHED_SAMPLE_CNT */

	/* First read of counters */
	x = gwloc_read_counter(brd, &ref1);
	SX1301AR_ERR_PROPAGATE(x);

	/* Second read of counters */
	x = gwloc_read_counter(brd, &ref2);
	SX1301AR_ERR_PROPAGATE(x);

	/* Compare the 2 sets of counters and determine if read is valid */
	/*  This is to ensure that we don't have a partial update, as read is not atomic */
	/* All FPGA counters are ensured to change at the same time, so we just
        need to compare one FPGA counter with the SX1301 one */
	/* Possible cases:
        (1) FPGA(x) SX1301(x) | FPGA(x) SX1301(x) => OK
        (2) FPGA(x) SX1301(x) | FPGA(x) SX1301(y) => Read again
        (3) FPGA(x) SX1301(x) | FPGA(y) SX1301(y) => OK
        (4) FPGA(x) SX1301(y) | FPGA(y) SX1301(y) => OK
        So, need to check for case (2) to read again, otherwise ref2 should always be valid */
	if ((ref1.sapps == ref2.sapps) && (ref1.sx1301pps != ref2.sx1301pps)) {
		/* Second read of counters */
		x = gwloc_read_counter(brd, &ref2);
		SX1301AR_ERR_PROPAGATE(x);
	}

	/* Update global ref */
	gwloc_hstime_ref = ref2;

	/* validate current reference */
	gwloc_hstime_ref.isvalid = true;

	/* Compute time spent in this function */
	_meas_time_stop(2, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_cnt2ls(uint32_t cnt, uint32_t *ls)
{
	uint32_t offset;

	if (ls == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if (gwloc_hstime_ref.isvalid == false) {
		error("%s: invalid high speed time reference", __FUNCTION__);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	offset = (gwloc_hstime_ref.sx1301pps - gwloc_hstime_ref.sapps * 8) % (1 << 27);

	*ls = ((cnt - offset) / 8) % (1 << 24);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int dsp_boot(uint8_t brd, uint8_t dsp)
{
	int x;
	int32_t val;
	uint8_t dsp_mic_status;
	sx1301ar_breg_t dsp_gpio_out, dsp_gpio_in, dsp_spi_reg_en, dsp_rstn, dsp_boot_en;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (dsp >= GWLOC_BOARD_DSP_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_DSP_NB);
	}

	debug("[BRD%u] Booting DSP%d ", brd, dsp);

	switch (dsp) {
	case 0:
		dsp_gpio_out = BREG_DSP1_GPIO_OUT;
		dsp_gpio_in = BREG_DSP1_GPIO_IN;
		dsp_spi_reg_en = BREG_DSP1_SPI_REG_EN;
		dsp_rstn = BREG_DSP1_RSTN;
		dsp_boot_en = BREG_DSP1_BOOT_EN;
		break;
	case 1:
		dsp_gpio_out = BREG_DSP2_GPIO_OUT;
		dsp_gpio_in = BREG_DSP2_GPIO_IN;
		dsp_spi_reg_en = BREG_DSP2_SPI_REG_EN;
		dsp_rstn = BREG_DSP2_RSTN;
		dsp_boot_en = BREG_DSP2_BOOT_EN;
		break;
	case 2:
		dsp_gpio_out = BREG_DSP3_GPIO_OUT;
		dsp_gpio_in = BREG_DSP3_GPIO_IN;
		dsp_spi_reg_en = BREG_DSP3_SPI_REG_EN;
		dsp_rstn = BREG_DSP3_RSTN;
		dsp_boot_en = BREG_DSP3_BOOT_EN;
		break;
	}

	x = sx1301ar_reg_brd_w(brd, dsp_gpio_out, 1);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_spi_reg_en, 0);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_rstn, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Start boot sequence */
	x = sx1301ar_reg_brd_w(brd, dsp_boot_en, 1);
	SX1301AR_ERR_PROPAGATE(x);

	/* Start FPGA DSP PLL */
	x = sx1301ar_reg_brd_w(brd, BREG_DSP_PLL_RESET, 0);
	SX1301AR_ERR_PROPAGATE(x);
	sx1301ar_wait_us(1000);
	x = sx1301ar_reg_brd_r(brd, BREG_DSP_PLL_LOCKED, &val);
	SX1301AR_ERR_PROPAGATE(x);
	if (val != 1) {
		SX1301AR_ERR_RETURN(ERR_FPGA_DSP_PLL);
	}

	/* Put DSP out of reset */
	x = sx1301ar_reg_brd_w(brd, dsp_rstn, 1);
	SX1301AR_ERR_PROPAGATE(x);

	/* Wait until DSP drive his GPIO to 0 */
	x = dsp_poll_reg_timeout(brd, dsp, dsp_gpio_in, 0, 10000);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_gpio_out, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Wait until DSP drive his GPIO to 1 */
	x = dsp_poll_reg_timeout(brd, dsp, dsp_gpio_in, 1, 10000);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_gpio_out, 1);
	SX1301AR_ERR_PROPAGATE(x);

	/* Wait until DSP drive his GPIO to 0 */
	x = dsp_poll_reg_timeout(brd, dsp, dsp_gpio_in, 0, 10000);
	SX1301AR_ERR_PROPAGATE(x);

	x = sx1301ar_reg_brd_w(brd, dsp_spi_reg_en, 1);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_brd_w(brd, dsp_gpio_out, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* End of boot sequence */
	x = sx1301ar_reg_brd_w(brd, dsp_boot_en, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Check if DSP signature is correct */
	if (dsp < 2) {
		sx1301ar_reg_brd_r(brd, BREG_DSP_MIC_STATUS, &val);
		dsp_mic_status = (uint8_t)TAKE_N_BITS_FROM((uint8_t)val, dsp * 4, 4);
	} else {
		sx1301ar_reg_brd_r(brd, BREG_DSP3_MIC_STATUS, &val);
		dsp_mic_status = (uint8_t)TAKE_N_BITS_FROM((uint8_t)val, 0, 4);
	}
	switch (dsp_mic_status) {
	case 0x5:
		debug("[BRD%u] DSP%d signature check succeeded", brd, dsp);
		break;
	case 0x6:
		debug("[BRD%u] DSP%d signature check succeeded, for production test usage (Check your AES key)", brd, dsp);
		break;
	/* Error cases */
	case 0x2:
		error("[BRD%u] DSP%d time elapsed for production test usage", brd, dsp);
	/* Falls through. */
	default:
		error("[BRD%u] DSP%d signature check failed (MIC_STATUS=0x%x)", brd, dsp, dsp_mic_status);
		SX1301AR_ERR_RETURN(ERR_DSP_SIG);
	}

	debug("[BRD%u] DSP%d boot done", brd, dsp);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int dsp_poll_reg_timeout(uint8_t brd, uint8_t dsp, sx1301ar_breg_t idx, int32_t expected_val, int32_t timeout_ms)
{
	int x;
	int32_t val;

	do {
		sx1301ar_wait_ms(100);
		timeout_ms -= 100;

		x = sx1301ar_reg_brd_r(brd, idx, &val);
		SX1301AR_ERR_PROPAGATE(x);
	} while ((val != expected_val) && (timeout_ms > 0));

	if ((timeout_ms <= 0) && (val != expected_val)) {
		error("Failed to boot DSP%d - TIMEOUT (reg = 0x%X)", dsp, val);
		SX1301AR_ERR_RETURN(ERR_DSP_BOOT);
	}

	return 0;
}

/* --- EOF ------------------------------------------------------------------ */
