/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    DSP functions for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

#ifndef _SX1301_ARRAY_DSP_H
#define _SX1301_ARRAY_DSP_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */

#include "hal.h"

/* -------------------------------------------------------------------------- */
/* --- PUBLIC CONSTANTS ----------------------------------------------------- */

#define GWLOC_BOARD_DSP_NB 3	   /* number of DSPs per board */
#define GWLOC_BOARD_DSP_FIFO_SIZE 4096 /* Size of read/write DSP FIFOs in bytes */

#define GWLOC_TMST_QUEUE_MAX 255

#define GWLOC_MSG_SYNCWORD_1 0xBF
#define GWLOC_MSG_SYNCWORD_2 0xA1

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

/**
@brief DSP message type
*/
typedef enum {
	GWLOC_DSP_MSG_TYPE_NONE = 0x00,
	GWLOC_DSP_MSG_TYPE_LOG_DEBUG = 0x01,
	GWLOC_DSP_MSG_TYPE_LOG_LOC = 0x02,
	GWLOC_DSP_MSG_TYPE_TIMESTAMP = 0x03,
	GWLOC_DSP_MSG_TYPE_CONFIG = 0x04,
	GWLOC_DSP_MSG_TYPE_STATUS = 0x05
} gwloc_dsp_msg_type_t;
#define GWLOC_DSP_MSG_TYPE_NB 6 /* size of the gwloc_dsp_msg_id_t type */

/**
@brief DSP message structure
        | ID (1byte) | length (2bytes) | data (nbytes) | CRC_A (1byte) | CRC_B (1byte) |
*/
typedef struct
{
	gwloc_dsp_msg_type_t id; /*!> message type */
	uint16_t payload_length; /*!> length of message payload */
	uint8_t *payload;	/*!> pointer to payload raw data, in static buffer array */
	uint8_t crc_a;		 /*!> first part of message CRC */
	uint8_t crc_b;		 /*!> second part of message CRC */
} gwloc_dsp_msg_t;

#define GWLOC_DSP_MSG_HEADER_SIZE 5 /* 5 bytes: SYNCWORD1+2 (2 bytes) - ID (1 byte) - Payload Length (2bytes) */
#define GWLOC_DSP_MSG_CRC_SIZE 2    /* 2 bytes: CRC_A (1byte) - CRC_B (1 byte) */

/**
@brief Structure containing the fine timestamp and related information
*/
typedef struct
{
	uint8_t brd;					       /*!> Board from which the fine timestamp has been received */
	uint32_t ts_lsp_ls_int;				       /*!> Integer part of the timestamp */
	uint8_t ts_lsp_frac_enc[SX1301AR_BOARD_AES_DATA_SIZE]; /*!> Encrypted fractionnal part of the timestamp */
	uint32_t ts_lsp_frac_dec;			       /*!> Decrypted fractionnal part of the timestamp, number of nanoseconds since last PPS */
	int8_t rssi_chan;				       /*!> RSSI in dB of the channel on which packet has been received */
	int8_t rssi_sig;				       /*!> RSSI in dB of the signal received */
	uint8_t rssi_sig_std;				       /*!> Preambule amplitude standard deviation */
	int8_t snr;					       /*!> average packet SNR, in dB (LoRa only) */
	float freq_offset;				       /*!> frequency offset compared to channel frequency */
	sx1301ar_modr_t modrate;			       /*!> TX modulation rate of the packet */
	uint8_t chan;					       /*!> by which chip & channel was packet received */
	uint8_t status;					       /*!> status of the fine timestamp */
	int16_t ts2_lsp_delta;				       /*!> offset between main and alternative fine timestamps, in nanoseconds */
	uint16_t ts_debug1;				       /*!> fine timestamp debug information 1 */
	uint16_t ts_debug2;				       /*!> fine timestamp debug information 2 */
} gwloc_dsp_tmst_t;

/**
@brief Structure containing the DSP configuration information
*/
typedef struct {
	uint8_t dsp_version;
	bool lorawan_public;
	bool diversity;
	uint8_t debug_tmst;
	uint8_t main_tmst_version;
	uint32_t pll0;
	uint32_t pll1;
	uint8_t chan_nb;
	float chan_freq[SX1301AR_BOARD_NB_CHIP_PER_DSP * SX1301AR_CHIP_MULTI_NB];
	uint32_t report_period;
} gwloc_dsp_cfg_t;

/**
@brief Structure containing the DSP status
*/
typedef struct
{
	uint32_t log_counter;
	uint32_t nb_jobs_added;
	uint32_t nb_jobs_executed;
	uint32_t nb_jobs_in_queue;
	uint32_t nb_jobs_max_in_queue; /*!> since last report */
	uint32_t nb_jobs_dropped;      /*!> since last report */
	uint32_t nb_jobs_dropped_acc;  /*!> accumulated */
	uint32_t nb_valid_timestamps;
	uint32_t nb_lost_sync;      /*!> since last report */
	uint32_t nb_lost_sync_acc;  /*!> accumulated */
	uint32_t sx1301_0_max_fifo; /*!> since last report */
	uint32_t sx1301_1_max_fifo; /*!> since last report */
	uint32_t dsp_max_fifo;      /*!> since last report */
	uint32_t dsp_max_enc_fifo;  /*!> since last report */
} gwloc_dsp_status_t;

/**
@brief Structure containing a timestamp node of timestamp queue
*/
typedef struct
{
	unsigned long long systime; /*!> system time when the node was enqueued */
	unsigned long long age_us;  /*!> age of the node at last purge check (microseconds) */
	gwloc_dsp_tmst_t tmst;      /*!> fine timestamp */
	bool is_sent;		    /*!> has the fine timestamp been sent already or not */
} gwloc_tmst_node_t;

/**
@brief Structure containing a timestamp queue
*/
typedef struct
{
	uint8_t num_tmst;
	gwloc_tmst_node_t nodes[GWLOC_TMST_QUEUE_MAX];
} gwloc_tmst_queue_t;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */

/**
@brief Boot the requested DSP
@param brd [in]             Target board number
@param dsp [in]             Target dsp number to boot
@param retry [in]           Number of retries in case of boot failure
@return 0 if DSP booted successfully, -1 otherwise
*/
int gwloc_dsp_boot(uint8_t brd, uint8_t dsp, uint8_t retry);

/**
@brief Stop the requested DSP, and stop the DSPs PLL if all DSPs are stopped.
@param brd [in]             Target board number
@param dsp [in]             Target dsp number to stop
@param nb_dsp [in]          Target nb dsp of the board
@return 0 if DSP booted successfully, -1 otherwise
*/
int gwloc_dsp_stop(uint8_t brd, uint8_t dsp, uint8_t nb_dsp);

/**
@brief Fetch all messages waiting in the given DSP's "read" FIFO (raw data to be parsed)
@param brd [in]             Target board number
@param dsp [in]             Target dsp number to get messages from
@param messages [in/out]    Pointer to array of messages
@param max_nb [in]          Size of array of messages
@param nb_msg [out]         Actual number of messages found
@return 0 if DSP Read FIFO was accessed, whether messages were waiting or not, -1 else
*/
int gwloc_dsp_fetch(uint8_t brd, uint8_t dsp, gwloc_dsp_msg_t *messages, uint8_t max_nb, uint8_t *nb_msg);

/**
@brief Fetch all messages waiting in the FPGA "read" FIFO (raw data to be parsed)
@param brd [in]             Target board number
@param dsp [in]             Target dsp number to get messages from
@param messages [in/out]    Pointer to array of messages
@param max_nb [in]          Size of array of messages
@param nb_msg [out]         Actual number of messages found
@return 0 if FPGA Read FIFO was accessed, whether messages were waiting or not, -1 else
*/
int gwloc_fpga_fetch(uint8_t brd, uint8_t dsp, gwloc_dsp_msg_t *messages, uint8_t max_nb, uint8_t *nb_msg);

/**
@brief build fine timestamp structure from raw payload
@param brd [in]             Target board number
@param brd_type [in]        Target board type
@param dsp [in]             DSP number from which fts has been received
@param tmst_msg [in]        Raw message to be parsed to fill fine timestamp structure
@param tmst [out]           Resulting fine timestamp structure
@return 0 if parsing was successfull, -1 else
*/
int gwloc_dsp_parse_timestamp(uint8_t brd, sx1301ar_btype_t brd_type, uint8_t dsp, gwloc_dsp_msg_t *tmst_msg, gwloc_dsp_tmst_t *tmst);

/**
@brief Get fine timestamp validity from its status
@param status [in]      Full status bitfield of a timestamp
@param is_valid [out]   Validity of the timestamp parsed from the status
@return 0 if parsing was successfull, -1 else
*/
int gwloc_dsp_parse_timestamp_status(uint8_t status, bool *isvalid);

/**
@brief build DSP configuration structure from raw payload
@param msg [in]    Raw message to be parsed to fill configuration structure
@param cfg [out]   Resulting configuration structure
@return 0 if parsing was successfull, -1 else
*/
int gwloc_dsp_parse_config(gwloc_dsp_msg_t *msg, gwloc_dsp_cfg_t *cfg);

/**
@brief build DSP status structure from raw payload
@param msg [in]     Raw message to be parsed to fill status structure
@param stat [out]   Resulting configuration structure
@return 0 if parsing was successfull, -1 else
*/
int gwloc_dsp_parse_status(gwloc_dsp_msg_t *msg, gwloc_dsp_status_t *stat);

/**
@brief Update all reference counters used to timing conversions
@param brd [in]     Target board number
@return 0 if update was successfull, -1 else
*/
int gwloc_update_ref_counter(uint8_t brd);

/*
    Fine timestamps queue functions
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

void gwloc_tmst_queue_init(gwloc_tmst_queue_t *queue);

int gwloc_tmst_enqueue(gwloc_tmst_queue_t *queue, const gwloc_dsp_tmst_t *fine_tmst);

int gwloc_tmst_dequeue(gwloc_tmst_queue_t *queue, int index, gwloc_dsp_tmst_t *fine_tmst);

int gwloc_tmst_peek(gwloc_tmst_queue_t *queue, const sx1301ar_rx_pkt_t *pkt, bool diversity, int *index, uint8_t *nb_match_tmst);

int gwloc_tmst_purge(gwloc_tmst_queue_t *queue);

bool gwloc_tmst_queue_is_full(const gwloc_tmst_queue_t *queue);

bool gwloc_tmst_queue_is_empty(const gwloc_tmst_queue_t *queue);

#endif

/* --- EOF ------------------------------------------------------------------ */
