/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Error management for 'LoRa Gateway v2' library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include "err.h"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES DEFINITION ------------------------------------------ */

sx1301ar_rcode_t sx1301ar_errno = ERR_NO_ERROR;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

const char *sx1301ar_err_message(sx1301ar_rcode_t code)
{
	switch (code) {
	case ERR_NO_ERROR:
		return "No error";
	case ERR_UNSPECIFIED:
		return "Unspecified error code";

	/* SPI callbacks issues */
	case ERR_READ_SPI:
		return "SPI function failed to read data";
	case ERR_WRITE_SPI:
		return "SPI function failed to write data";

	/* Errors in functions parameters */
	case ERR_NULL_POINTER:
		return "Invalid null pointer in function parameter";
	case ERR_BAD_BOARD_NB:
		return "Invalid board number in function parameter";
	case ERR_BAD_BOARD_TYPE:
		return "Invalid board type in function parameter";
	case ERR_BAD_BOARD_REG:
		return "Invalid board register in function parameter";
	case ERR_BAD_CHIP_NB:
		return "Invalid chip number in function parameter";
	case ERR_BAD_DSP_NB:
		return "Invalid DSP number in function parameter";
	case ERR_BAD_CHAN_NB:
		return "Invalid channel number in function parameter";
	case ERR_BAD_RFCHAIN_NB:
		return "Invalid RF chain number in function parameter";

	/* Board initialization code */
	case ERR_FPGA_VZERO:
		return "Version reg for FPGA returned zero, check SPI communication";
	case ERR_FPGA_VERSION:
		return "Invalid version number for the top FPGA";
	case ERR_CHIP_VZERO:
		return "Version reg for a SX1301 returned zero, check SPI communication";
	case ERR_CHIP_VERSION:
		return "Invalid version number for one of the SX1301 chip";
	case ERR_CHIP_ID:
		return "Invalid chip ID for one of the SX1301 chip";

	/* Codes returned by the sx1301ar_conf_* functions */
	case ERR_BAD_CFG:
		return "Invalid configuration structure";
	case ERR_CANT_CFG:
		return "Board is running, cannot be configured";
	case ERR_OUT_OF_BAND:
		return "Frequency is out of configured band";
	case ERR_NO_RX_ENABLE:
		return "At least 1 RX RF chain must be enabled";
	case ERR_NO_TX_ENABLE:
		return "No TX RF chain enabled enabled";
	case ERR_IF_LIMIT:
		return "Maximum supported IF exceeded";
	case ERR_CHIP_DISABLE:
		return "Can't enable a channel if the corresponding chip is disabled";
	case ERR_INVALID_SF:
		return "Datarate setting not supported";
	case ERR_INVALID_BR:
		return "Bitrate setting not supportedl";
	case ERR_INVALID_BW:
		return "Bandwidth setting not supported";
	case ERR_INVALID_CH:
		return "Channel is invalid";
	case ERR_INVALID_LUT:
		return "Tx gain LUT is invalid (size or gain value)";

	/* Codes returned by HAL */
	case ERR_FPGA_PLL:
		return "FPGA PLL failed to lock";
	case ERR_AGC_VERSION:
		return "Invalid version of AGC firmware";
	case ERR_ARB_VERSION:
		return "Invalid version of arbiter firmware";
	case ERR_BOARD_STATE:
		return "Board is in wrong state";
	case ERR_FPGA_DSP_PLL:
		return "FPGA DSP PLL failed to lock";
	case ERR_DSP_BOOT:
		return "Failed to boot DSP";
	case ERR_DSP_SIG:
		return "DSP signature check failed";
	case ERR_LOC_FORBIDDEN:
		return "Fine timestamp not allowed";
	case ERR_AGC_INIT:
		return "SX1301 AGC failed to initialise";
	case ERR_UBX_CMD_SEND:
		return "Failed to send UBX command to GPS";
	case ERR_UBX_CMD_NACK:
		return "Failed to get UBX command acknowledge";
	case ERR_UBX_CMD_TIMEOUT:
		return "Failed to get UBX command response (timeout)";
	case ERR_UBX_INCOMPLETE:
		return "Failed to get a complete UBX frame";
	case ERR_UBX_INVALID:
		return "Failed to get a valid UBX frame (CRC error)";
	case ERR_LBT_TMST_SYNC:
		return "FPGA and SX1301 timestamp counter are not in sync";
	case ERR_LBT_SCAN_TIME:
		return "Listen-Before-Talk scan time not supported";
	case ERR_LBT_INIT:
		return "Failed to initialize Listen-Before-Talk";
	case ERR_LBT_START:
		return "Failed to start/stop Listen-Before-Talk";
	case ERR_LBT_FORBIDDEN:
		return "TX is not allowed (Listen-Before-Talk)";

	/* Codes retuned by radio driver */
	case ERR_AD9361_VERSION:
		return "Invalid version of AD9361";
	case ERR_AD9361_ALERT:
		return "AD9361 failed to enter ALERT state";
	case ERR_AD9361_PLL_CAL:
		return "Calibration of AD9361 PLL charge pump failed";
	case ERR_AD9361_BBPLL_LOCK:
		return "AD9361 BB PLL failed to lock";
	case ERR_AD9361_RXPLL_LOCK:
		return "AD9361 Rx PLL failed to lock";
	case ERR_AD9361_TXPLL_LOCK:
		return "AD9361 Tx PLL failed to lock";
	case ERR_AD9361_RXFILT_CAL:
		return "Calibration of AD9361 Rx BB filter failed";
	case ERR_AD9361_TXFILT_CAL:
		return "Calibration of AD9361 Tx BB filter failed";
	case ERR_AD9361_GAIN_TABLE:
		return "Failed to write AD9361 AGC gain table";
	case ERR_AD9361_RX_FIR:
		return "Failed to write AD9361 Rx FIR filter coefficients";
	case ERR_AD9361_TX_FIR:
		return "Failed to write AD9361 Tx FIR filter coefficients";
	case ERR_AD9361_BBDC_CAL:
		return "Calibration of AD9361 Rx BB DC offsets failed";
	case ERR_AD9361_RFDC_CAL:
		return "Calibration of AD9361 Rx RF DC offsets failed";
	case ERR_AD9361_RXQUAD_CAL:
		return "Calibration of AD9361 Rx quadrature correction failed";
	case ERR_AD9361_TXQUAD_CAL:
		return "Calibration of AD9361 Tx quadrature correction failed";
	case ERR_AD9361_TDD:
		return "AD9361 TDD test failed";
	case ERR_AD9361_FREQBAND:
		return "Invalid frequency band";

	/* Default code */
	default:
		return "** not error message, or invalid error code **";
	}
}

/* --- EOF ------------------------------------------------------------------ */
