/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Error management for 'LoRa Gateway v2' library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

#ifndef _SX1301_ARRAY_ERR_H
#define _SX1301_ARRAY_ERR_H

#include <stdint.h> /* C99 types */
#include <stdlib.h> /* size_t */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

typedef enum {
	ERR_NO_ERROR = 0, /*!> Operation successful */

	/* SPI callbacks issues */
	ERR_READ_SPI,  /*!> SPI function failed to read data */
	ERR_WRITE_SPI, /*!> SPI function failed to write data */

	/* Errors in functions parameters */
	ERR_NULL_POINTER,   /*!> Invalid null pointer in function parameter */
	ERR_BAD_BOARD_NB,   /*!> Invalid board number in function parameter */
	ERR_BAD_BOARD_TYPE, /*!> Invalid board type in function parameter */
	ERR_BAD_BOARD_REG,  /*!> Invalid board register in function parameter */
	ERR_BAD_CHIP_NB,    /*!> Invalid chip number in function parameter */
	ERR_BAD_DSP_NB,     /*!> Invalid dsp number in function parameter */
	ERR_BAD_CHAN_NB,    /*!> Invalid channel number in function parameter */
	ERR_BAD_RFCHAIN_NB, /*!> Invalid RF chain number in function parameter */

	/* Board initialization code */
	ERR_FPGA_VZERO,   /*!> Version reg for FPGA returned zero, check SPI communication */
	ERR_FPGA_VERSION, /*!> Invalid version number for the top FPGA */
	ERR_CHIP_VZERO,   /*!> Version reg for a SX1301 returned zero, check SPI communication */
	ERR_CHIP_VERSION, /*!> Invalid version number for one of the SX1301 chip */
	ERR_CHIP_ID,      /*!> Invalid chip ID for one of the SX1301 chip */

	/* Codes returned by the sx1301ar_conf_* functions */
	ERR_BAD_CFG,      /*!> Invalid configuration structure */
	ERR_CANT_CFG,     /*!> Board is running, cannot be configured */
	ERR_OUT_OF_BAND,  /*!> Frequency is out of configured band */
	ERR_NO_RX_ENABLE, /*!> At least 1 RX RF chain must be enabled */
	ERR_NO_TX_ENABLE, /*!> No TX RF chain enabled */
	ERR_IF_LIMIT,     /*!> Maximum supported IF exceeded */
	ERR_CHIP_DISABLE, /*!> Can't enable a channel if the corresponding chip is disabled */
	ERR_INVALID_SF,   /*!> Spreading factor setting not supported by the channel */
	ERR_INVALID_BR,   /*!> Bitrate setting not supported by the channel */
	ERR_INVALID_BW,   /*!> Bandwidth setting not supported by the channel */
	ERR_INVALID_CH,   /*!> Channel is invalid */
	ERR_INVALID_LUT,  /*!> Tx gain LUT is invalid (size or gain value) */

	/* Codes retuned by HAL */
	ERR_FPGA_PLL,	/*!> FPGA PLL failed to lock */
	ERR_AGC_VERSION,     /*!> Invalid version of AGC firmware */
	ERR_ARB_VERSION,     /*!> Invalid version of arbiter firmware */
	ERR_BOARD_STATE,     /*!> Board is not in proper state */
	ERR_FPGA_DSP_PLL,    /*!> FPGA DSP PLL failed to lock */
	ERR_DSP_BOOT,	/*!> Failed to boot DSP */
	ERR_DSP_SIG,	 /*!> DSP signature check failed */
	ERR_LOC_FORBIDDEN,   /*!> Fine timestamp not allowed */
	ERR_AGC_INIT,	/*!> SX1301 AGC failed to initialise */
	ERR_UBX_CMD_SEND,    /*!> Failed to send UBX command to GPS */
	ERR_UBX_CMD_NACK,    /*!> Failed to get UBX command acknowledge */
	ERR_UBX_CMD_TIMEOUT, /*!> Failed to get UBX command response (timeout) */
	ERR_UBX_INCOMPLETE,  /*!> Failed to get a complete UBX frame */
	ERR_UBX_INVALID,     /*!> Failed to get a valid UBX frame (CRC error) */
	ERR_LBT_TMST_SYNC,   /*!> FPGA and SX1301 timestamp counter are not in sync */
	ERR_LBT_SCAN_TIME,   /*!> LBT scan time not supported */
	ERR_LBT_INIT,	/*!> Failed to initialize Listen-Before-Talk */
	ERR_LBT_START,       /*!> Failed to start/stop Listen-Before-Talk */
	ERR_LBT_FORBIDDEN,   /*!> TX is not allowed (Listen-Before-Talk) */

	/* Codes retuned by radio driver */
	ERR_AD9361_VERSION,    /*!> Invalid version of AD9361 */
	ERR_AD9361_ALERT,      /*!> AD9361 failed to enter ALERT state */
	ERR_AD9361_PLL_CAL,    /*!> Calibration of AD9361 PLL charge pump failed */
	ERR_AD9361_BBPLL_LOCK, /*!> AD9361 BB PLL failed to lock */
	ERR_AD9361_RXPLL_LOCK, /*!> AD9361 Rx PLL failed to lock */
	ERR_AD9361_TXPLL_LOCK, /*!> AD9361 Tx PLL failed to lock */
	ERR_AD9361_RXFILT_CAL, /*!> Calibration of AD9361 Rx BB filter failed */
	ERR_AD9361_TXFILT_CAL, /*!> Calibration of AD9361 Tx BB filter failed */
	ERR_AD9361_GAIN_TABLE, /*!> Failed to write AD9361 AGC gain table */
	ERR_AD9361_RX_FIR,     /*!> Failed to write AD9361 Rx FIR filter coefficients */
	ERR_AD9361_TX_FIR,     /*!> Failed to write AD9361 Tx FIR filter coefficients */
	ERR_AD9361_BBDC_CAL,   /*!> Calibration of AD9361 Rx BB DC offsets failed */
	ERR_AD9361_RFDC_CAL,   /*!> Calibration of AD9361 Rx RF DC offsets failed */
	ERR_AD9361_RXQUAD_CAL, /*!> Calibration of AD9361 Rx quadrature correction failed */
	ERR_AD9361_TXQUAD_CAL, /*!> Calibration of AD9361 Tx quadrature correction failed */
	ERR_AD9361_TDD,	/*!> AD9361 TDD test failed */
	ERR_AD9361_FREQBAND,   /*!> Invalid frequency band */

	/* Default code */
	ERR_UNSPECIFIED /*!> Unspecified error */
} sx1301ar_rcode_t;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC MACROS -------------------------------------------------------- */

#define SX1301AR_ERR_RETURN(a)        \
	do {                          \
		sx1301ar_errno = (a); \
		return -1;            \
	} while (0)

#define SX1301AR_ERR_PROPAGATE(a) \
	if ((a) < 0)              \
	return -1

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES PROTOTYPES ------------------------------------------ */

extern sx1301ar_rcode_t sx1301ar_errno;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */
/** @warning This module has no internal state. */

/**
@brief Transform error code into user readable error information
@param code  error code returned by a function of that library
@return pointer to a string containing an explicit error message for display
*/
const char *sx1301ar_err_message(sx1301ar_rcode_t code);

#endif

/* --- EOF ------------------------------------------------------------------ */
