/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    GPS timestamping functions for LoRa Gateway v2

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <float.h>
#include <limits.h>  /* INT_MAX */
#include <math.h>    /* modf */
#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <stdio.h>   /* NULL */
#include <string.h>  /* memset */
#include <time.h>    /* struct timespec, tzset, mktime */
#include <unistd.h>  /* read */

#include "aux.h"
#include "err.h"
#include "gps.h"
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#define TS_CPS 1E6      /* nominal count-per-second of timestamp counter */
#define HS_CPS 256E6    /* nominal count-per-second of high speed counter */
#define PPM_LIMIT 20e-6 /* +/- X ppm of clock error tolerated */
#define UP_PPM_BOUND (1.0 + PPM_LIMIT)
#define DOWN_PPM_BOUND (1.0 - PPM_LIMIT)

#define ROUND_2_INT(f) ((int)(f >= 0.0 ? (f + 0.5) : (f - 0.5)))

#define UBX_SYNC_BYTE_1 0xB5
#define UBX_SYNC_BYTE_2 0x62
#define UBX_PKT_HDR_LEN 6 /* Sync chars + class + ID + length */
#define UBX_PKT_CRC_LEN 2

/* UBX message class */
#define UBX_MSG_CLASS_NAV 0x01
#define UBX_MSG_CLASS_ACK 0x05
#define UBX_MSG_CLASS_CFG 0x06

/* UBX message ID */
/* NAV */
#define UBX_MSG_ID_NAV_POSLLH 0x02
#define UBX_MSG_ID_NAV_TIMEGPS 0x20
/* ACK */
#define UBX_MSG_ID_ACK 0x01
#define UBX_MSG_ID_NACK 0x00
/* CFG */
#define UBX_MSG_ID_CFG_PRT 0x00
#define UBX_MSG_ID_CFG_SBAS 0x16
#define UBX_MSG_ID_CFG_NAV5 0x24
#define UBX_MSG_ID_CFG_TP5 0x31
#define UBX_MSG_ID_CFG_INF 0x02
#define UBX_MSG_ID_CFG_MSG 0x01
#define UBX_MSG_ID_CFG_CFG 0x09

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

typedef enum {
	UBX_MSG_FRAME_ACKNOWLEDGED,
	UBX_MSG_FRAME_ONLY,
	UBX_MSG_ACK_ONLY
} ubx_msgType_t;

typedef enum {
	UBX_DYNMODEL_PORTABLE,
	UBX_DYNMODEL_UNUSED,
	UBX_DYNMODEL_STATIONARY,
	UBX_DYNMODEL_PEDESTRIAN,
	UBX_DYNMODEL_AUTOMOTIVE,
	UBX_DYNMODEL_SEA,
	UBX_DYNMODEL_AIRBORNE_1G,
	UBX_DYNMODEL_AIRBORNE_2G,
	UBX_DYNMODEL_AIRBORNE_4G
} ubx_dynModel_t;

typedef enum {
	UBX_FIXMODE_UNUSED,
	UBX_FIXMODE_2D_ONLY,
	UBX_FIXMODE_3D_ONLY,
	UBX_FIXMODE_AUTO
} ubx_fixMode_t;

typedef enum {
	UBX_PORT_I2C,
	UBX_PORT_UART1,
	UBX_PORT_UART2,
	UBX_PORT_USB,
	UBX_PORT_SPI
} ubx_portId_t;

typedef enum {
	UBX_PROTOCOLID_UBX,
	UBX_PROTOCOLID_NMEA
} ubx_protocolId_t;

/**
@brief UBlox GPS UBX packet structure
*/
typedef struct
{
	uint8_t class;       /*!> Message class */
	uint8_t id;	  /*!> Message ID */
	size_t length;       /*!> Message length in byte */
	uint8_t payload[50]; /*!> Message payload */
	uint16_t crc;	/*!> Message CRC */
} ubx_pkt_t;

/**
@brief UBlox GPS UBX GPS Time Solution
*/
typedef struct
{
	uint32_t iTOW; /*!> GPS time of week in milliseconds */
	int32_t fTOW;  /*!> Fractional part of iTOW (+/-500000) in nanosec */
	int16_t week;  /*!> GPS week number of the navigation epoch */
	int8_t leap;   /*!> GPS leap seconds (GPS-UTC) */
	bool valid;    /*!> Validity of the time solution */
} ubx_timegps_t;

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

static struct
{
	bool bfix_altitude;
	uint16_t fix_altitude;
} gps_altitude_priv;

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

/**
@brief Parse the given buffer and return the UBS packet which has been found
@param buff         Pointer to buffer containing UBX frame
@param buff_size    Size of buffer pointed by 'buff' (in bytes)
@param p_ubx_msg    Pointer to the UBX packet which has been found
@return value >=0 to indicate the position of the UBX packet in the buffer, -1 if an error occured
*/
static int get_ubx_pkt(const uint8_t *buff, size_t buff_size, ubx_pkt_t *p_ubx_msg);

/**
@brief Transform a nibble (half byte) into an hexadecimal ASCII character
@param a  Nibble (4 bit) to convert
@return ASCII character [0-9,A-F] if ok, '?' if the nibble was invalid (>=16)
*/
static char nibble_to_hexchar(uint8_t nibble);

/**
@brief Calculate the checksum for a UBX message
@param buff         Pointer to buffer containing UBX frame
@param buff_size    Size of buffer pointed by 'buff' (in bytes)
@param nmea_length  Optional pointer to return index of '*' NMEA terminator
@return Checksum value (0-255), -1 if error
*/
static void calc_ubx_checksum(const uint8_t *buff, size_t buff_size, uint16_t *checksum);

/**
@brief Calculate the checksum for a NMEA string
@param buff         Pointer to character buffer containing NMEA frame
@param buff_size    Size of buffer pointed by 'buff' (in bytes)
@param nmea_length  Optional pointer to return index of '*' NMEA terminator
@return Checksum value (0-255), -1 if error
*/
static int calc_nmea_checksum(const char *buff, size_t buff_size, int *nmea_term);

/**
@brief Check if the checksum of an NMEA frame is valid or not
@param buff       Pointer to character buffer containing NMEA frame
@param buff_size  Size of buffer pointed by 'buff' (in bytes)
@return true if NMEA checksum is valid, false else
*/
static bool validate_nmea_checksum(const char *buff, size_t buff_size);

/**
@brief Chop a string in small null-terminated tokens
@param buff       Pointer to a string to chop in tokens
@param buff_size  Size of buffer pointer by 'buff' (in bytes)
@param separator  Separator character
@param idx        Pointer to array that will receive the token offsets
@param idx_size   Size of the array pointed by 'idx' (in bytes)
@return number of tokens found, -1 if error (N separator -> N+1 tokens)
*/
static int strnchopchop(char *buff, size_t buff_size, char separator, int *idx, size_t idx_size);

/**
@brief Set Ublox GPS I/O port configuration
@param fd           File descriptor for GPS serial access
@param port_id      Port identifier number
@param ubx_input    Enable/Disable UBX protocol input on selected I/O port
@param ubx_output   Enable/Disable UBX protocol output on selected I/O port
@param nmea_input   Enable/Disable NMEA protocol input on selected I/O port
@param nmea_output  Enable/Disable NMEA protocol output on selected I/O port
@return 0 if success, -1 if error
*/
static int set_port_configuration(int fd, ubx_portId_t port_id, bool ubx_input, bool ubx_output, bool nmea_input, bool nmea_output);

/**
@brief Set Ublox GPS SBAS configuration
@param fd           File descriptor for GPS serial access
@param enabled      Enable/Disable SBAS mode
@return 0 if success, -1 if error
*/
static int set_sbas_mode(int fd, bool enabled);

/**
@brief Set Ublox GPS Navigation settings
@param fd           File descriptor for GPS serial access
@param dyn_model    Dynamic platform model
@param fix_mode     Position fixing mode
@param fixed_alt    Fixed altitude
@return 0 if success, -1 if error
*/
static int set_nav5_param(int fd, ubx_dynModel_t dyn_model, ubx_fixMode_t fix_mode, uint16_t fixed_alt);

/**
@brief Set Ublox GPS information message configuration
@param fd           File descriptor for GPS serial access
@param port_id      I/O port for which information message are being configured
@param protocol_id  Protocol identifier for which the configuration is set
@return 0 if success, -1 if error
*/
static int set_info_msg_configuration(int fd, ubx_portId_t port_id, ubx_protocolId_t protocol_id);

/**
@brief Set Ublox GPS message rates
@param fd           File descriptor for GPS serial access
@param port_id      I/O port for which information message are being configured
@param msg_class    Message class for which rate is set
@param msg_id       Message ID for which rate is set
@param rate         Rate to be applied for message, in seconds [0..255]
@return 0 if success, -1 if error
*/
static int set_message_rates(int fd, ubx_portId_t port_id, uint8_t msg_class, uint8_t msg_id, uint8_t rate);

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

sx1301ar_gps_cfg_t sx1301ar_init_gps_cfg(void)
{
	sx1301ar_gps_cfg_t a;

	a.bfix_altitude = false;
	a.fix_altitude = 0;

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_gps(const sx1301ar_gps_cfg_t *cfg)
{
	/* Check input parameters */
	if (cfg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	gps_altitude_priv.bfix_altitude = cfg->bfix_altitude;
	gps_altitude_priv.fix_altitude = cfg->fix_altitude;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_init_gps(int fd)
{
	int ret;

	/* Disable GPS Time Solution message on UART during initialization */
	ret = set_message_rates(fd, UBX_PORT_UART1, UBX_MSG_CLASS_NAV, UBX_MSG_ID_NAV_TIMEGPS, 0);
	SX1301AR_ERR_PROPAGATE(ret);

	/* Configuration of the UART port to output both UBX and NMEA, and receive UBX */
	ret = set_port_configuration(fd, UBX_PORT_UART1, true, true, false, true);
	SX1301AR_ERR_PROPAGATE(ret);

	/* Disable SBAS */
	ret = set_sbas_mode(fd, true);
	SX1301AR_ERR_PROPAGATE(ret);

	if (gps_altitude_priv.bfix_altitude == true) {
		/* Define the navigation engine settings with 2D mode to fix the altitude */
		ret = set_nav5_param(fd, UBX_DYNMODEL_STATIONARY, UBX_FIXMODE_2D_ONLY, gps_altitude_priv.fix_altitude);
		SX1301AR_ERR_PROPAGATE(ret);
	} else {
		/* Define the navigation engine settings with 3D mode in stationary mode */
		ret = set_nav5_param(fd, UBX_DYNMODEL_STATIONARY, UBX_FIXMODE_3D_ONLY, gps_altitude_priv.fix_altitude);
		SX1301AR_ERR_PROPAGATE(ret);
	}

	/* Disable Warning message. Enable only Error message */
	ret = set_info_msg_configuration(fd, UBX_PORT_UART1, 1);
	SX1301AR_ERR_PROPAGATE(ret);

	/* Enable GPS Time Solution message on UART, with rate of 1 second */
	ret = set_message_rates(fd, UBX_PORT_UART1, UBX_MSG_CLASS_NAV, UBX_MSG_ID_NAV_TIMEGPS, 1);
	SX1301AR_ERR_PROPAGATE(ret);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_term_gps(int fd)
{
	int ret;

	/* Disable GPS Time Solution message on UART */
	ret = set_message_rates(fd, UBX_PORT_UART1, UBX_MSG_CLASS_NAV, UBX_MSG_ID_NAV_TIMEGPS, 0);
	SX1301AR_ERR_PROPAGATE(ret);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_nmea_type(const char *buff, size_t buff_size, sx1301ar_nmea_msg_t *type)
{
	/* Check input parameters */
	if (buff == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (type == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((buff_size / sizeof(char)) < 10) {
		*type = NMEA_INVALID;
		return 0;
	}

	/* Set default value */
	*type = NMEA_UNDEFINED;

	/* Identification tree */
	if ((buff[0] == '$') && (buff[1] == 'G')) {
		if ((buff[2] == 'P') || (buff[2] == 'L') || (buff[2] == 'A') || (buff[2] == 'B') || (buff[2] == 'N')) {
			/*
            $GPyyy -> GPS (US)
            $GLyyy -> GLONASS (RU)
            $GAyyy -> Galileo (EU)
            $GByyy -> BeiDou/COMPASS (CN)
            $GNyyy -> Mixed between several systems
            */
			switch (buff[3]) {
			case 'G':
				switch (buff[4]) {
				case 'B':
					if (buff[5] == 'S')
						*type = NMEA_GBS;
					break;
				case 'G':
					if (buff[5] == 'A')
						*type = NMEA_GGA;
					break;
				case 'L':
					if (buff[5] == 'L')
						*type = NMEA_GLL;
					break;
				case 'N':
					if (buff[5] == 'S')
						*type = NMEA_GNS;
					break;
				case 'S':
					if (buff[5] == 'A')
						*type = NMEA_GSA;
					else if (buff[5] == 'T')
						*type = NMEA_GST;
					else if (buff[5] == 'V')
						*type = NMEA_GSV;
					break;
				}
				break;
			case 'R':
				if ((buff[4] == 'M') && (buff[5] == 'C'))
					*type = NMEA_RMC;
				break;
			case 'T':
				if ((buff[4] == 'X') && (buff[5] == 'T'))
					*type = NMEA_TXT;
				break;
			case 'V':
				if ((buff[4] == 'T') && (buff[5] == 'G'))
					*type = NMEA_VTG;
				break;
			case 'Z':
				if ((buff[4] == 'D') && (buff[5] == 'A'))
					*type = NMEA_ZDA;
				break;
			}
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_parse_rmc(char *buff, size_t buff_size, struct timespec *utc, sx1301ar_coord_t *pos)
{
	int i, j;
	bool checksum_ok;
	int tok_idx[13];
	char gps_mode;
	char parser_buff[256]; /* parsing modifies buffer so need a local copy */
	size_t parser_buff_size;

	/* Variables for time parsing */
	short yea;   /* year (2 or 4 digits) */
	short mon;   /* month (1-12) */
	short day;   /* day of the month (1-31) */
	short hou;   /* hours (0-23) */
	short min;   /* minutes (0-59) */
	double sec;  /* seconds (0-59, 60? 61?) */
	double fra;  /* fractions of seconds (<1) */
	struct tm x; /* broken-up time */

	/* Variables for coordinate parsing (2D) */
	short dla, dlo;  /* degrees of latitude/longitude */
	double mla, mlo; /* minutes of latitude/longitude */
	char ola;	/* orientation (N-S) of latitude */
	char olo;	/* orientation (E-W) of longitude */

	/*
    RMC FORMAT:
           1         2 3       4 5        6 7   8   9    10 11 12 13
           |         | |       | |        | |   |   |    |   | |  |
    $--RMC,hhmmss.ss,A,llll.ll,a,yyyyy.yy,a,x.x,x.x,xxxx,x.x,a,m,*hh<CR><LF>

    1.  UTC Time
    2.  Status, V=Navigation receiver warning A=Valid
    3.  Latitude
    4.  N or S
    5.  Longitude
    6.  E or W
    7.  Speed over ground (knots ?)
    8.  Track made good, degrees true
    9.  Date, ddmmyy
    10. Magnetic Variation, degrees
    11. E or W
    12. FAA mode indicator (NMEA 2.3 and later)
    13. Checksum

    Min: $GPRMC,,V,,,,,,,,,,*XX
    */

	/* Check input parameters */
	if (buff == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (buff_size < 23) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Validate checksum */
	checksum_ok = validate_nmea_checksum(buff, buff_size);
	if (checksum_ok == false) {
		return WARN_NMEA_BAD_CHECKSUM;
	}

	/* Add string termination character to buffer */
	memcpy(parser_buff, buff, buff_size);
	parser_buff[buff_size] = '\0';
	parser_buff_size = buff_size + 1;

	/* Chop the message in tokens */
	i = strnchopchop(parser_buff, parser_buff_size, ',', tok_idx, sizeof tok_idx);
	if (i < 12) {
		return WARN_NMEA_BAD_FORMAT;
	}

	/* Check NMEA frame validity indicator */
	gps_mode = parser_buff[tok_idx[2]];
	if (gps_mode == 'V') {
		return WARN_NMEA_INVALID;
	} else if (gps_mode != 'A') {
		return WARN_NMEA_BAD_FORMAT;
	}

	/* Get and return the time information */
	if (utc != NULL) {
		i = sscanf(parser_buff + tok_idx[9], "%2hd%2hd%hd", &day, &mon, &yea);
		j = sscanf(parser_buff + tok_idx[1], "%2hd%2hd%lf", &hou, &min, &fra); /* fra catch seconds + optional fractional seconds */
		if ((i != 3) || (j != 3)) {
			return WARN_NMEA_INCOMPLETE;
		}
		fra = modf(fra, &sec); /* split integer seconds and optional fractional part */

		/* Initialize broken-up time structure */
		memset((void *)&x, 0, sizeof x);

		/* Format date (won't work for years before 2000) */
		if (yea < 100) {	       /* 2-digits year, 20xx */
			x.tm_year = yea + 100; /* 100 years offset to 1900 */
		} else {		       /* 4-digits year, Gregorian calendar */
			x.tm_year = yea - 1900;
		}
		x.tm_mon = mon - 1; /* tm_mon is [0,11], gps_mon is [1,12] */
		x.tm_mday = day;

		/* Format time */
		x.tm_hour = hou;
		x.tm_min = min;
		x.tm_sec = (int)sec;

		/* Return time */
		utc->tv_sec = mktime(&x);
		tzset();
		utc->tv_sec -= timezone; /* correct for timezone bc mktime assumes local time */
		utc->tv_nsec = (long)(fra * 1e9);
	}

	/* Get and return the 2D position information */
	if (pos != NULL) {
		i = sscanf(parser_buff + tok_idx[3], "%2hd%lf", &dla, &mla);
		j = sscanf(parser_buff + tok_idx[5], "%3hd%lf", &dlo, &mlo);
		if ((i != 2) || (j != 2)) {
			return WARN_NMEA_INCOMPLETE;
		}
		ola = parser_buff[tok_idx[4]];
		olo = parser_buff[tok_idx[6]];
		if (((ola != 'N') && (ola != 'S')) || ((olo != 'E') && (olo != 'W'))) {
			return WARN_NMEA_INCOMPLETE;
		}

		/* Return geodesic coordiantes */
		pos->lat = ((double)dla + (mla / 60.0)) * ((ola == 'N') ? 1.0 : -1.0);
		pos->lon = ((double)dlo + (mlo / 60.0)) * ((olo == 'E') ? 1.0 : -1.0);
		pos->alt = 0;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_parse_gga(char *buff, size_t buff_size, sx1301ar_coord_t *pos, int *n_sat, float *hdop)
{
	int i, j;
	bool checksum_ok;
	int tok_idx[15];
	char parser_buff[256]; /* parsing modifies buffer so need a local copy */
	size_t parser_buff_size;

	/* Variables for coordinate parsing (3D) */
	short dla, dlo;  /* degrees of latitude/longitude */
	double mla, mlo; /* minutes of latitude/longitude */
	char ola;	/* orientation (N-S) of latitude */
	char olo;	/* orientation (E-W) of longitude */
	int alt;	 /* altitude in meters */

	/* Misc variables */
	short sat_nb;
	float prec_dil;

	/*
    GGA FORMAT:
           1         2       3 4        5 6 7  8   9  10 11 12 13  14   15
           |         |       | |        | | |  |   |   | |   | |   |    |
    $--GGA,hhmmss.ss,llll.ll,a,yyyyy.yy,a,x,xx,x.x,x.x,M,x.x,M,x.x,xxxx*hh<CR><LF>

    1.  Universal Time Coordinated (UTC)
    2.  Latitude
    3.  N or S (North or South)
    4.  Longitude
    5.  E or W (East or West)
    6.  GPS Quality Indicator
    7.  Number of satellites in view
    8.  Horizontal Dilution of precision
    9.  Antenna Altitude above/below mean-sea-level (geoid)
    10. Units of antenna altitude
    11. Geoidal separation
    12. Units of geoidal separation,
    13. Age of differential GPS data
    14. Differential reference station ID
    15. Checksum

    Min: $GPGGA,,,,,,0,00,,,,,,,*XX
    */

	/* Check input parameters */
	if (buff == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (buff_size < 27) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Validate checksum */
	checksum_ok = validate_nmea_checksum(buff, buff_size);
	if (checksum_ok == false) {
		return WARN_NMEA_BAD_CHECKSUM;
	}

	/* Add string termination character to buffer */
	memcpy(parser_buff, buff, buff_size);
	parser_buff[buff_size] = '\0';
	parser_buff_size = buff_size + 1;

	/* Chop the message in tokens */
	i = strnchopchop(parser_buff, parser_buff_size, ',', tok_idx, sizeof tok_idx);
	if (i != 15) {
		return WARN_NMEA_BAD_FORMAT;
	}

	/* Get and return the 3D position information */
	if (pos != NULL) {
		i = sscanf(parser_buff + tok_idx[2], "%2hd%lf", &dla, &mla);
		j = sscanf(parser_buff + tok_idx[4], "%3hd%lf", &dlo, &mlo);
		if ((i != 2) || (j != 2)) {
			return WARN_NMEA_INCOMPLETE;
		}
		ola = parser_buff[tok_idx[3]];
		olo = parser_buff[tok_idx[5]];
		if (((ola != 'N') && (ola != 'S')) || ((olo != 'E') && (olo != 'W'))) {
			return WARN_NMEA_INCOMPLETE;
		}
		i = sscanf(parser_buff + tok_idx[9], "%d", &alt);
		if (i != 1) {
			return WARN_NMEA_INCOMPLETE;
		}

		/* Return geodesic coordinates */
		pos->lat = ((double)dla + (mla / 60.0)) * ((ola == 'N') ? 1.0 : -1.0);
		pos->lon = ((double)dlo + (mlo / 60.0)) * ((olo == 'E') ? 1.0 : -1.0);
		pos->alt = alt;
	}

	/* Return number of satellites */
	if (n_sat != NULL) {
		i = sscanf(parser_buff + tok_idx[7], "%hd", &sat_nb);
		if (i != 1) {
			return WARN_NMEA_INCOMPLETE;
		}
		*n_sat = sat_nb;
	}

	/* Return horizontal dilution of precision */
	if (hdop != NULL) {
		i = sscanf(parser_buff + tok_idx[8], "%f", &prec_dil);
		if (i != 1) {
			return WARN_NMEA_INCOMPLETE;
		}
		*hdop = prec_dil;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_ubx_type(const char *buff, size_t buff_size, sx1301ar_ubx_msg_t *type, size_t *length)
{
	int i;
	ubx_pkt_t ubx_msg;

	/* Check input parameters */
	if (buff == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (type == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (buff_size < LGW_GPS_MIN_MSG_SIZE) {
		*type = UBX_INVALID;
		return 0;
	}

	/* Set default value */
	*type = UBX_UNDEFINED;

	/* Parse buffer to get UBX packet */
	if ((i = get_ubx_pkt((const uint8_t *)buff, buff_size, &ubx_msg)) >= 0) {
		if ((ubx_msg.class == UBX_MSG_CLASS_NAV) && (ubx_msg.id == UBX_MSG_ID_NAV_TIMEGPS)) {
			debug("GPS: found UBX NAV-TIMEGPS message");
			*type = UBX_NAV_TIMEGPS;
			*length = UBX_PKT_HDR_LEN + ubx_msg.length + UBX_PKT_CRC_LEN;
		} else {
			debug("GPS: unsupported UBX message %02X %02X", ubx_msg.class, ubx_msg.id);
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_parse_ubx_timegps(const char *buff, size_t buff_size, struct timespec *timegps)
{
	ubx_timegps_t tgps;
	double intpart, fractpart;

	/* Check input parameters */
	if (buff == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (timegps == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (buff_size < LGW_GPS_MIN_MSG_SIZE) {
		SX1301AR_ERR_RETURN(ERR_UBX_INVALID);
	}

	/* Check validity of information */
	tgps.valid = buff[UBX_PKT_HDR_LEN + 11] & 0x3; /* towValid, weekValid */
	if (tgps.valid == true) {
		/* Parse buffer to extract GPS time */
		tgps.iTOW = (uint32_t)buff[UBX_PKT_HDR_LEN + 0];
		tgps.iTOW |= (uint32_t)buff[UBX_PKT_HDR_LEN + 1] << 8;
		tgps.iTOW |= (uint32_t)buff[UBX_PKT_HDR_LEN + 2] << 16;
		tgps.iTOW |= (uint32_t)buff[UBX_PKT_HDR_LEN + 3] << 24;

		tgps.fTOW = (int32_t)buff[UBX_PKT_HDR_LEN + 4];
		tgps.fTOW |= (int32_t)buff[UBX_PKT_HDR_LEN + 5] << 8;
		tgps.fTOW |= (int32_t)buff[UBX_PKT_HDR_LEN + 6] << 16;
		tgps.fTOW |= (int32_t)buff[UBX_PKT_HDR_LEN + 7] << 24;

		tgps.week = (int16_t)buff[UBX_PKT_HDR_LEN + 8];
		tgps.week |= (int16_t)buff[UBX_PKT_HDR_LEN + 9] << 8;

		tgps.leap = (int8_t)buff[UBX_PKT_HDR_LEN + 10];

		debug("GPS: GPS time solution: iTOW=%u,fTOW=%d,week=%d,leap=%d", tgps.iTOW, tgps.fTOW, tgps.week, tgps.leap);

		/* Convert to timespec format */
		fractpart = modf(((double)tgps.iTOW / 1E3) + ((double)tgps.fTOW / 1E9), &intpart);
		/* Number of seconds since beginning on current GPS week */
		(*timegps).tv_sec = (time_t)intpart;
		/* Number of seconds since GPS epoch 06.Jan.1980 */
		(*timegps).tv_sec += (time_t)tgps.week * 604800; /* day*hours*minutes*secondes: 7*24*60*60; */
		/* Fractional part in nanoseconds */
		(*timegps).tv_nsec = (long)(fractpart * 1E9);

#if 0
        /* For debug */
        {
            uint8_t ubx_gps_hou = 0; /* hours (0-23) */
            uint8_t ubx_gps_min = 0; /* minutes (0-59) */
            uint8_t ubx_gps_sec = 0; /* seconds (0-59) */

            /* Format GPS time in hh:mm:ss based on iTOW */
            ubx_gps_sec = (tgps.iTOW / 1000) % 60;
            ubx_gps_min = (tgps.iTOW / 1000 / 60) % 60;
            ubx_gps_hou = (tgps.iTOW / 1000 / 60 / 60) % 24;
            debug("NAV-TIMEGPS:%02u:%02u:%02u (leap:%ds)", ubx_gps_hou, ubx_gps_min, ubx_gps_sec, tgps.leap);
        }
#endif
	} else {
		debug("GPS: GPS time solution is not valid");
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_synchronize(sx1301ar_tref_t *ref, struct timespec utc,
#if 0
			 struct timespec gps,
#endif
			 uint32_t count_us, uint32_t hs_pps, int *pps_delay)
{
	double cnt_diff;    /* internal concentrator time difference (in seconds) */
	double hs_diff;     /* high speed counter time difference (in seconds) */
	double utc_diff;    /* UTC time difference (in seconds) */
	double slope = 1.0; /* time slope between new reference and old reference */
	double slope_hs = 1.0;
	bool slope_ok = false;    /* is the update value for synchronization aberrant or not ? */
	bool slope_hs_ok = false; /* is the update value for synchronization aberrant or not ? */

	/* Check input parameters */
	if (ref == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if (pps_delay != NULL) {
		*pps_delay = 0;
	}

	/* Calculate slope if there's a previous reference point (unless forced) */
	if (ref->sync_cnt > 0) {
		/* Calculate time difference between new points and existing ref */
		cnt_diff = (double)(count_us - ref->count_us) / (double)(TS_CPS); /* uncorrected by xtal_err */
		hs_diff = (double)(hs_pps - ref->hs_pps) / (double)(HS_CPS);      /* uncorrected by xtal_err */
		utc_diff = (double)(utc.tv_sec - (ref->utc).tv_sec);
		utc_diff += 1E-9 * (double)(utc.tv_nsec - (ref->utc).tv_nsec);

		if (pps_delay != NULL) {
			*pps_delay = ROUND_2_INT(cnt_diff);
			debug("UTC diff %.*g - cnt diff %.*g - hs_diff %.*g - PPS delay %d", DBL_DIG, utc_diff, DBL_DIG, cnt_diff, DBL_DIG, hs_diff, *pps_delay);
		}

		if (utc_diff != 0.0) /* prevent divide by 0 */
		{
			/* Calculate the slope */
			slope = cnt_diff / utc_diff;
			slope_hs = hs_diff / utc_diff;

			/* Validate that slope is in the acceptable tolerance range */
			if ((slope < UP_PPM_BOUND) && (slope > DOWN_PPM_BOUND)) {
				slope_ok = true;
			}
			if ((slope_hs < UP_PPM_BOUND) && (slope_hs > DOWN_PPM_BOUND)) {
				slope_hs_ok = true;
			}
		}

		/* Cannot sync if slope is out of range */
		if ((slope_ok == false) || (slope_hs_ok == false)) {
			/* Waiting for two PPS before resynchronization */
			if (cnt_diff > 1.5) {
				/* PPS was not generated by the GPS module => desynchronization between UTC and PPS counter */
				/* Now update all counters to synchronize UTC and PPS */
				/* Record system time of updated reference for ageing limit */
				ref->systime = time(NULL);

				/* Use new UTC/counter points for reference */
				ref->count_us = count_us;
				ref->hs_pps = hs_pps;
				ref->utc = utc;
#if 0
				ref->gps = gps;
				debug("Synchronize: count us=%u, hs_pps=%u, utc sec=%ld, utc nsec=%ld, gps sec=%ld, gps nsec=%ld", ref->count_us, ref->hs_pps, utc.tv_sec, utc.tv_nsec, gps.tv_sec, gps.tv_nsec);
#endif

				/* Reinitialization of  Xtal errors */
				ref->xtal_err = 1.0;
				ref->xtal_hs_err = 1.0;
			}
			return WARN_SYNC_OUT_OF_RANGE;
		} else {
			/* Record slope for XTAL error compensation (after averaging) */
			ref->xtal_err = ((15.0 * ref->xtal_err) + slope) / 16.0;
			ref->xtal_hs_err = ((15.0 * ref->xtal_hs_err) + slope_hs) / 16.0;

			//debug("XTAL error %.*g - HS XTAL error %.*g", DBL_DIG, ref->xtal_err, DBL_DIG, ref->xtal_hs_err);

			/* Record system time of updated reference for ageing limit */
			ref->systime = time(NULL);

			/* Use new UTC/counter points for reference */
			ref->count_us = count_us;
			ref->hs_pps = hs_pps;
			ref->utc = utc;
#if 0
			ref->gps = gps;
			debug("Synchronize: count us=%u, hs_pps=%u, utc sec=%ld, utc nsec=%ld, gps sec=%ld, gps nsec=%ld", ref->count_us, ref->hs_pps, utc.tv_sec, utc.tv_nsec, gps.tv_sec, gps.tv_nsec);
#endif
		}
	} else {
		/* Record system time of updated reference for ageing limit */
		ref->systime = time(NULL);

		/* Use new UTC/counter points for reference */
		ref->count_us = count_us;
		ref->hs_pps = hs_pps;
		ref->utc = utc;
#if 0
		ref->gps = gps;
		debug("Synchronize: count us=%u, hs_pps=%u, utc sec=%ld, utc nsec=%ld, gps sec=%ld, gps nsec=%ld", ref->count_us, ref->hs_pps, utc.tv_sec, utc.tv_nsec, gps.tv_sec, gps.tv_nsec);
#endif
	}

	/* Increment sync_cnt with saturation */
	if (ref->sync_cnt < UINT32_MAX) {
		ref->sync_cnt++;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_set_xtal_err(uint8_t brd, sx1301ar_tref_t ref)
{
	int ret;
	uint32_t alpha_coeff;
	double alpha_coeff_f;
	uint8_t alpha_coeff_u8[4];

	/* Send Xtal error to FPGA for more precise fine timestamp calculation */
	alpha_coeff_f = (double)1E9 / (ref.xtal_hs_err * (double)256E6) * (double)pow(2, 26);
	alpha_coeff = (uint32_t)round(alpha_coeff_f);
	alpha_coeff_u8[0] = (uint8_t)(alpha_coeff & 0xFF);
	alpha_coeff_u8[1] = (uint8_t)((alpha_coeff >> 8) & 0xFF);
	alpha_coeff_u8[2] = (uint8_t)((alpha_coeff >> 16) & 0xFF);
	alpha_coeff_u8[3] = (uint8_t)((alpha_coeff >> 24) & 0xFF);
	ret = sx1301ar_reg_brd_w(brd, BREG_ALPHA_BETA_COEFF_VALID, 0);
	if (ret != 0) {
		error("Failed to set BREG_ALPHA_BETA_COEFF_VALID to 0");
		return -1;
	}
	ret = sx1301ar_reg_brd_wb(brd, BREG_ALPHA_COEFF, alpha_coeff_u8, 4);
	if (ret != 0) {
		error("Failed to set ALPHA COEFF");
		return -1;
	}
	ret = sx1301ar_reg_brd_w(brd, BREG_ALPHA_BETA_COEFF_VALID, 1);
	if (ret != 0) {
		error("Failed to set BREG_ALPHA_BETA_COEFF_VALID to 1");
		return -1;
	}

	return 0;
}

#if 0
int sx1301ar_cnt2utc(sx1301ar_tref_t ref, uint32_t count_us, struct timespec *utc)
{
	double delta_sec;
	double intpart, fractpart;
	long x;

	/* Check input parameters */
	if (utc == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (ref.sync_cnt == 0) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Calculate delta in seconds between reference timestamp and target timestamp */
	delta_sec = ((double)count_us - (double)ref.count_us) / (TS_CPS * ref.xtal_err);
	debug("count us = %u, ref count us = %u, XTAL ERR = %.*g => delta_sec = %.*g", count_us, ref.count_us, DBL_DIG, ref.xtal_hs_err, DBL_DIG, delta_sec);
	fractpart = modf(delta_sec, &intpart);

	/* Now add that delta to reference UTC time */
	x = ref.utc.tv_nsec + (long)(fractpart * 1E9);
	if (x >= 0) {
		if (x < (long)1E9) { /* the nanosecond part doesn't overflow */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart;
			utc->tv_nsec = x;
		} else { /* must carry one second */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart + 1;
			utc->tv_nsec = x - (long)1E9;
		}
	} else {
		/* x is negative, we need to go back in time of 1 or 2 seconds */
		if (abs(x) < (long)1E9) { /* must carry one second */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart - 1;
			utc->tv_nsec = (long)1E9 + x;
		} else { /* must carry two seconds */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart - 2;
			utc->tv_nsec = (long)2E9 + x;
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_utc2cnt(sx1301ar_tref_t ref, struct timespec utc, uint32_t *count_us)
{
	double delta_sec;

	/* Check input parameters */
	if (count_us == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (ref.sync_cnt == 0) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Calculate delta in seconds between reference utc and target utc */
	delta_sec = (double)(utc.tv_sec - ref.utc.tv_sec);
	delta_sec += 1E-9 * (double)(utc.tv_nsec - ref.utc.tv_nsec);

	/* Now convert that to internal counter tics and add that to reference counter value */
	*count_us = ref.count_us + (uint32_t)((delta_sec * TS_CPS * ref.xtal_err) + 0.5); /* add 0.5 because integer cast does a floor, not a round */

	return 0;
}
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int gwloc_hs2utc(sx1301ar_tref_t ref, uint32_t ts_hs, struct timespec *utc)
{
	double delta_sec;
	double intpart, fractpart;
	long x;
	uint32_t delta_hs;

	/* Check input parameters */
	if (utc == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (ref.sync_cnt == 0) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Calculate delta in seconds between reference timestamp and target timestamp. */

	/* First, we want to detect if ref.hs_pps refer to a PPS which occured after
        the fine timestamp. */
	if (ts_hs > ref.hs_pps) {
		delta_hs = ts_hs - ref.hs_pps;
	} else {
		delta_hs = ref.hs_pps - ts_hs;
	}

	/* We cannot just compare the values because the counters may have rolled over. */
	/* If (ts_hs - hs_pps) > (UINT_MAX - 8seconds), it means that the PPS taken
        a reference happened after the fine timestamp, so in this case we need
        to handle a negative delta. */
	if (delta_hs > (UINT_MAX - 8 * HS_CPS)) {
		/* Roll-over case: soustract then cast */
		delta_sec = (double)(ts_hs - ref.hs_pps) / (HS_CPS * ref.xtal_hs_err);
	} else {
		/* General case: cast then soustract to get the positive or negative sign. */
		delta_sec = ((double)ts_hs - (double)ref.hs_pps) / (HS_CPS * ref.xtal_hs_err);
	}

	debug("Ts hs = %u, HSPPS = %u, XTAL ERR = %.9f => delta_sec = %.9f", ts_hs, ref.hs_pps, ref.xtal_hs_err, delta_sec);

	fractpart = modf(delta_sec, &intpart);

	/* Now add that delta to reference UTC time */
	x = ref.utc.tv_nsec + (long)(fractpart * 1E9);
	if (x >= 0) {
		if (x < (long)1E9) { /* the nanosecond part doesn't overflow */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart;
			utc->tv_nsec = x;
		} else { /* must carry one second */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart + 1;
			utc->tv_nsec = x - (long)1E9;
		}
	} else {
		/* x is negative, we need to go back in time of 1 or 2 seconds */
		if (abs(x) < (long)1E9) { /* must carry one second */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart - 1;
			utc->tv_nsec = (long)1E9 + x;
		} else { /* must carry two seconds */
			utc->tv_sec = ref.utc.tv_sec + (time_t)intpart - 2;
			utc->tv_nsec = (long)2E9 + x;
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_cnt2gps(sx1301ar_tref_t ref, uint32_t count_us, struct timespec *gps_time)
{
	double delta_sec;
	double intpart, fractpart;
	long tmp;

	/* Check input parameters */
	if (gps_time == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (ref.sync_cnt == 0) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* calculate delta in milliseconds between reference count_us and target count_us */
	delta_sec = (double)(count_us - ref.count_us) / (TS_CPS * ref.xtal_err);

	/* now add that delta to reference GPS time */
	fractpart = modf(delta_sec, &intpart);
	tmp = ref.gps.tv_nsec + (long)(fractpart * 1E9);
	if (tmp < (long)1E9) { /* the nanosecond part doesn't overflow */
		gps_time->tv_sec = ref.gps.tv_sec + (time_t)intpart;
		gps_time->tv_nsec = tmp;
	} else { /* must carry one second */
		gps_time->tv_sec = ref.gps.tv_sec + (time_t)intpart + 1;
		gps_time->tv_nsec = tmp - (long)1E9;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_gps2cnt(sx1301ar_tref_t ref, struct timespec gps_time, uint32_t *count_us)
{
	double delta_sec;

	/* Check input parameters */
	if (count_us == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (ref.sync_cnt == 0) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* calculate delta in seconds between reference gps time and target gps time */
	delta_sec = (double)(gps_time.tv_sec - ref.gps.tv_sec);
	delta_sec += (double)(gps_time.tv_nsec - ref.gps.tv_nsec) * 1E-9;

	/* now convert that to internal counter tics and add that to reference counter value */
	*count_us = ref.count_us + (uint32_t)((delta_sec * TS_CPS * ref.xtal_err) + 0.5); /* add 0.5 because integer cast does a floor, not a round */

	return 0;
}
#endif

sx1301ar_tref_t sx1301ar_init_tref(void)
{
	sx1301ar_tref_t a;

	a.systime = (time_t)(-1);
	a.hs_pps = 0;
	a.count_us = 0;
	a.utc.tv_sec = (time_t)(-1);
	a.utc.tv_nsec = 0;
	a.xtal_err = 1.0;
	a.xtal_hs_err = 1.0;
	a.sync_cnt = 0;

	return a;
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

bool check_ubx_pkt_class(uint8_t class)
{
	switch (class) {
	case 0x01:
	case 0x02:
	case 0x04:
	case 0x05:
	case 0x06:
	case 0x0A:
	case 0x0B:
	case 0x0D:
	case 0x21:
		/* supported */
		return true;
	default:
		/* not supported */
		return false;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int get_ubx_pkt(const uint8_t *buff, size_t buff_size, ubx_pkt_t *p_ubx_msg)
{
	int i;
	size_t msg_size;
	uint16_t msg_crc, rcv_crc;
	bool msg_found = false;

	/* Check input parameters */
	if ((buff == NULL) || (p_ubx_msg == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (buff_size < LGW_GPS_MIN_MSG_SIZE) {
		SX1301AR_ERR_RETURN(ERR_UBX_INCOMPLETE);
	}

#if 0
    debug( "Parsing buffer for UBX (%d):", buff_size );
    for( i = 0; i < (int)buff_size; i++ )
    {
        debug( "%02X ", buff[i] );
    }
    debug( "" );
#endif

	/* Parsing */
	for (i = 0; i < (int)buff_size; i++) {
		if (buff[i] == UBX_SYNC_BYTE_1) {
			/* Found first UBX sync char */
			if ((buff_size - i) >= UBX_PKT_HDR_LEN) {
				/* We have a full UBX header to parse */
				if (buff[i + 1] == UBX_SYNC_BYTE_2) {
					/* Found a UBX frame syncword */
					/* Now check if it is a valid UBX class */
					if (check_ubx_pkt_class(buff[i + 2]) == false) {
						continue;
					}
					/* Continue parsing */
					msg_size = buff[i + 4] | (buff[i + 5] << 8);
					if ((i + UBX_PKT_HDR_LEN + msg_size + UBX_PKT_CRC_LEN) <= buff_size) {
						/* We got the complete frame, check its integrity */
						msg_crc = (buff[i + UBX_PKT_HDR_LEN + msg_size] << 8) | buff[i + UBX_PKT_HDR_LEN + msg_size + 1]; /* what we should have */
						calc_ubx_checksum((const uint8_t *)(buff + i + 2), msg_size + 4, &rcv_crc);			  /* what we have */
						if (msg_crc == rcv_crc) {
							msg_found = true;
							p_ubx_msg->class = buff[i + 2];
							p_ubx_msg->id = buff[i + 3];
							p_ubx_msg->length = msg_size;
							p_ubx_msg->crc = rcv_crc;
							memcpy(p_ubx_msg->payload, &buff[i + UBX_PKT_HDR_LEN], p_ubx_msg->length * sizeof(uint8_t));
							/* Exit */
							break;
						} else {
							error("Received UBX message with CRC ERROR (msg:%04X - rcv:%04X)", msg_crc, rcv_crc);
							SX1301AR_ERR_RETURN(ERR_UBX_INVALID);
						}
					} else {
						SX1301AR_ERR_RETURN(ERR_UBX_INCOMPLETE);
					}
				}
			} else {
				SX1301AR_ERR_RETURN(ERR_UBX_INCOMPLETE);
			}
		}
	}

	/* Returns the UBX message start position in buffer if found */
	return ((msg_found == true) ? i : -1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int get_ubx_response(int fd, uint8_t msg_class, uint8_t msg_id, ubx_msgType_t ack_mode, ubx_pkt_t *p_ubx_msg)
{
	int i;
	char buff[32 * LGW_GPS_MIN_MSG_SIZE]; /* buffer to receive GPS data */
	ssize_t nb_char;
	size_t rd_idx = 0;
	size_t wr_idx = 0; /* pointer to end of chars in buffer */
	bool msg_found = false;
	bool ack_found = false;
	bool nack_found = false;
	ubx_pkt_t ubx_msg;
	struct timeval tm_start, tm_now;
	time_t timeout_sec = 5; /* 5 seconds timeout */
	bool timeout_exit = false;

	/* Check input parameters */
	if (p_ubx_msg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	debug("GPS: Waiting for UBX response for %02X %02X...", msg_class, msg_id);

	switch (ack_mode) {
	case UBX_MSG_FRAME_ONLY:
		ack_found = true;
		break;
	case UBX_MSG_ACK_ONLY:
		msg_found = true;
		break;
	case UBX_MSG_FRAME_ACKNOWLEDGED:
	default:
		break;
	}

	gettimeofday(&tm_start, NULL);
	while (((msg_found == false) || ((ack_found == false) && (nack_found == false))) && (timeout_exit == false)) {
		gettimeofday(&tm_now, NULL);
		timeout_exit = ((tm_now.tv_sec - tm_start.tv_sec) >= timeout_sec) ? true : false;

		/* Wait for NMEA/UBX frames (blocking non-canonical read on serial port) */
		nb_char = read(fd, buff + wr_idx, LGW_GPS_MIN_MSG_SIZE);
		if (nb_char <= 0) {
			warning("serial port read() returned value %zd", nb_char);
			continue;
		}
		wr_idx += (size_t)nb_char;

		/* Parse buffer to get UBX packet */
		if ((i = get_ubx_pkt((const uint8_t *)buff, wr_idx, &ubx_msg)) >= 0) {
			if ((msg_found == false) && ((ack_mode == UBX_MSG_FRAME_ACKNOWLEDGED) || (ack_mode == UBX_MSG_FRAME_ONLY))) {
				/* Check if it is the message we are looking for */
				if ((msg_class == ubx_msg.class) && (msg_id == ubx_msg.id)) {
					dump(buff + i, UBX_PKT_HDR_LEN + ubx_msg.length + UBX_PKT_CRC_LEN, "UBX rsp");

					msg_found = true;
					memcpy(p_ubx_msg, &ubx_msg, sizeof(ubx_pkt_t));
				} else {
					debug("GPS: Received UBX message (IGNORED): %02X %02X", ubx_msg.class, ubx_msg.id);
				}
			}
			if ((ack_mode == UBX_MSG_FRAME_ACKNOWLEDGED) || (ack_mode == UBX_MSG_ACK_ONLY)) {
				/* Check if we got an ACK for our message */
				if ((ubx_msg.class == UBX_MSG_CLASS_ACK) && (ubx_msg.id == UBX_MSG_ID_ACK)) {
					if ((msg_class == ubx_msg.payload[0]) && (msg_id == ubx_msg.payload[1])) {
						debug("GPS: Received UBX ACK for %02X %02X", ubx_msg.payload[0], ubx_msg.payload[1]);
						ack_found = true;
					} else {
						debug("GPS: Received UBX ACK for %02X %02X (IGNORED)", ubx_msg.payload[0], ubx_msg.payload[1]);
					}
				}
				/* Check if we got an NACK for our message */
				if ((ubx_msg.class == UBX_MSG_CLASS_ACK) && (ubx_msg.id == UBX_MSG_ID_NACK)) {
					if ((msg_class == ubx_msg.payload[0]) && (msg_id == ubx_msg.payload[1])) {
						debug("GPS: Received UBX NACK for %02X %02X", ubx_msg.payload[0], ubx_msg.payload[1]);
						nack_found = true;
					} else {
						debug("GPS: Received UBX NACK for 0x%02X 0x%02X (IGNORED)", ubx_msg.payload[0], ubx_msg.payload[1]);
					}
				}
			}

			/* Removed parsed UBX packet from buffer */
			rd_idx = i + UBX_PKT_HDR_LEN + ubx_msg.length + UBX_PKT_CRC_LEN;
			wr_idx -= rd_idx;
			memcpy(buff, &buff[rd_idx], wr_idx);
		} else {
			//debug("GPS: No valid UBX packet found..." );
			/* continue reading... */
		}

		/* Prevent buffer overflow */
		if ((sizeof(buff) - wr_idx) < LGW_GPS_MIN_MSG_SIZE) {
			wr_idx -= LGW_GPS_MIN_MSG_SIZE;
			memcpy(buff, &buff[LGW_GPS_MIN_MSG_SIZE], wr_idx);
		}
	}

	if (nack_found == true) {
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_NACK);
	}

	if (timeout_exit == true) {
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_TIMEOUT);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int send_ubx_command(int fd, ubx_pkt_t *p_cmd_msg)
{
	int ret;
	uint16_t msg_crc;

	if (p_cmd_msg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	uint8_t cmd_buff[p_cmd_msg->length + UBX_PKT_HDR_LEN + UBX_PKT_CRC_LEN]; /* declared after null pointer check */

	cmd_buff[0] = UBX_SYNC_BYTE_1;
	cmd_buff[1] = UBX_SYNC_BYTE_2;
	cmd_buff[2] = p_cmd_msg->class;
	cmd_buff[3] = p_cmd_msg->id;
	cmd_buff[4] = (uint8_t)(p_cmd_msg->length & 0xFF);
	cmd_buff[5] = (uint8_t)((p_cmd_msg->length >> 8) & 0xFF);
	if (p_cmd_msg->length) {
		memcpy(&cmd_buff[6], p_cmd_msg->payload, p_cmd_msg->length * sizeof(uint8_t));
	}

	/* CRC computation of the received UBX message */
	calc_ubx_checksum(&cmd_buff[2], p_cmd_msg->length + 4, &msg_crc);

	cmd_buff[6 + p_cmd_msg->length] = (uint8_t)((msg_crc >> 8) & 0xFF);
	cmd_buff[7 + p_cmd_msg->length] = (uint8_t)(msg_crc & 0xFF);

	dump(cmd_buff, p_cmd_msg->length + UBX_PKT_HDR_LEN + UBX_PKT_CRC_LEN, "UBX cmd");

	ret = write(fd, &cmd_buff[0], p_cmd_msg->length + UBX_PKT_HDR_LEN + UBX_PKT_CRC_LEN);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int set_port_configuration(int fd, ubx_portId_t port_id, bool ubx_input, bool ubx_output, bool nmea_input, bool nmea_output)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	uint16_t input_mask = 0, output_mask = 0;
	int ret;

	debug("GPS: I/O port configuration");

	/* Get the current configuration */
	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_PRT;
	ubx_cmd_msg.length = 1;
	ubx_cmd_msg.payload[0] = port_id;
	/* Poll the CFG-PRT message */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_PRT, UBX_MSG_FRAME_ACKNOWLEDGED, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Define mask describing which input protocols are active */
	input_mask = ubx_resp_msg.payload[12] | (ubx_resp_msg.payload[13] << 8);
	output_mask = ubx_resp_msg.payload[14] | (ubx_resp_msg.payload[15] << 8);
	debug("GPS: input mask %x - output mask %x", input_mask, output_mask);
	if (ubx_input == true) {
		input_mask = input_mask | 0x0001; /* set first bit */
	} else {
		input_mask = input_mask & 0xFFFE; /* clear first bit */
	}

	if (nmea_input == true) {
		input_mask = input_mask | 0x0002; /* set second bit */
	} else {
		input_mask = input_mask & 0xFFFD; /* clear second bit */
	}

	if (ubx_output == true) {
		output_mask = output_mask | 0x0001; /* set first bit */
	} else {
		output_mask = output_mask & 0xFFFE; /* clear first bit */
	}

	if (nmea_output == true) {
		output_mask = output_mask | 0x0002; /* set second bit */
	} else {
		output_mask = output_mask & 0xFFFD; /* clear second bit */
	}

	/* Now configure the port with new parameters */
	memcpy(ubx_cmd_msg.payload, ubx_resp_msg.payload, ubx_resp_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.length = ubx_resp_msg.length;

	debug("GPS: input mask %x - output mask %x", input_mask, output_mask);
	ubx_cmd_msg.payload[12] = input_mask & 0xFF;
	ubx_cmd_msg.payload[13] = (input_mask >> 8) & 0xFF;
	ubx_cmd_msg.payload[14] = output_mask & 0xFF;
	ubx_cmd_msg.payload[15] = (output_mask >> 8) & 0xFF;
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_PRT, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int set_sbas_mode(int fd, bool enabled)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_SBAS;
	ubx_cmd_msg.length = 8;

	if (enabled == true) {
		ubx_cmd_msg.payload[0] = 0x0;  /* mode */
		ubx_cmd_msg.payload[1] = 0x07; /* usage */
		ubx_cmd_msg.payload[2] = 0x03; /* maxSBAS */
	} else {
		ubx_cmd_msg.payload[0] = 0x1; /* mode */
		ubx_cmd_msg.payload[1] = 0x0; /* usage */
		ubx_cmd_msg.payload[2] = 0x0; /* maxSBAS */
	}
	ubx_cmd_msg.payload[3] = 0x0; /* scanmode2 */
	ubx_cmd_msg.payload[4] = 0x0; /* scanmode1 */

	debug("GPS: Disable SBAS");

	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_SBAS, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int set_nav5_param(int fd, ubx_dynModel_t dyn_model, ubx_fixMode_t fix_mode, uint16_t fixed_alt)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	debug("GPS: Set NAV5 parameters: Dynamic platform model: %d, Pos fixing mode: %d, Fixed altitude: %d", dyn_model, fix_mode, fixed_alt);

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_NAV5;
	ubx_cmd_msg.length = 0;
	/* Poll the CFG-NAV5 message */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_NAV5, UBX_MSG_FRAME_ACKNOWLEDGED, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Now configure the device with new parameters */
	memcpy(ubx_cmd_msg.payload, ubx_resp_msg.payload, ubx_resp_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.length = ubx_resp_msg.length;

	ubx_cmd_msg.payload[2] = dyn_model;
	ubx_cmd_msg.payload[3] = fix_mode;
	if (fix_mode == UBX_FIXMODE_2D_ONLY) {
		ubx_cmd_msg.payload[4] = fixed_alt & 0xFF;
		ubx_cmd_msg.payload[5] = (fixed_alt >> 8) & 0xFF;
		ubx_cmd_msg.payload[6] = 0;
		ubx_cmd_msg.payload[7] = 0;
	}
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_NAV5, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int set_time_pulse(int fd, int32_t rf_group_delay)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	debug("GPS: Set Time Pulse parameters");

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_TP5;
	ubx_cmd_msg.length = 0;
	/* Poll the CFG-TP5 message */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_TP5, UBX_MSG_FRAME_ACKNOWLEDGED, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Now configure the device with new parameters */
	memcpy(ubx_cmd_msg.payload, ubx_resp_msg.payload, ubx_resp_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.length = ubx_resp_msg.length;

	ubx_cmd_msg.payload[6] = rf_group_delay & 0xFF;
	ubx_cmd_msg.payload[7] = (rf_group_delay >> 8) & 0xFF;
	ubx_cmd_msg.payload[8] = 0x40;
	ubx_cmd_msg.payload[9] = 0x42;
	ubx_cmd_msg.payload[10] = 0x0F;

	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_TP5, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int set_info_msg_configuration(int fd, ubx_portId_t port_id, ubx_protocolId_t protocol_id)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	debug("GPS: Information message configuration");

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_INF;
	ubx_cmd_msg.length = 1;
	ubx_cmd_msg.payload[0] = protocol_id;
	/* Poll the CFG-INF message */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_INF, UBX_MSG_FRAME_ACKNOWLEDGED, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Now configure the device with new parameters */
	memcpy(ubx_cmd_msg.payload, ubx_resp_msg.payload, ubx_resp_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.length = ubx_resp_msg.length;

	ubx_cmd_msg.payload[4 + port_id] = ubx_cmd_msg.payload[4 + port_id] & 0xFFFD; /* clear second bit: Warning */

	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_INF, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int set_message_rates(int fd, ubx_portId_t port_id, uint8_t msg_class, uint8_t msg_id, uint8_t rate)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_MSG;
	ubx_cmd_msg.length = 2;
	ubx_cmd_msg.payload[0] = msg_class;
	ubx_cmd_msg.payload[1] = msg_id;
	/* Poll the current message rates configuration */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_MSG, UBX_MSG_FRAME_ACKNOWLEDGED, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Now configure the device with new parameters */
	memcpy(ubx_cmd_msg.payload, ubx_resp_msg.payload, ubx_resp_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.length = ubx_resp_msg.length;
	ubx_cmd_msg.payload[2 + port_id] = rate; /* New message rate for the current port */
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_MSG, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int save_config(int fd)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_CFG;
	ubx_cmd_msg.length = 12;
	memset(ubx_cmd_msg.payload, 0, ubx_cmd_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.payload[4] = 0x1F;
	ubx_cmd_msg.payload[5] = 0x06;
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_CFG, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int factory_reset(int fd)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	ubx_cmd_msg.class = UBX_MSG_CLASS_CFG;
	ubx_cmd_msg.id = UBX_MSG_ID_CFG_CFG;
	ubx_cmd_msg.length = 12;
	memset(ubx_cmd_msg.payload, 0, ubx_cmd_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.payload[0] = 0x1F;
	ubx_cmd_msg.payload[1] = 0x06;
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_CFG, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	memset(ubx_cmd_msg.payload, 0, ubx_cmd_msg.length * sizeof(uint8_t));
	ubx_cmd_msg.payload[8] = 0x1F;
	ubx_cmd_msg.payload[9] = 0x06;
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the acknowledge */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_CFG, UBX_MSG_ID_CFG_CFG, UBX_MSG_ACK_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int get_geodetic_pos(int fd, sx1301ar_coord_t *coord)
{
	ubx_pkt_t ubx_cmd_msg, ubx_resp_msg;
	int ret;

	ubx_cmd_msg.class = UBX_MSG_CLASS_NAV;
	ubx_cmd_msg.id = UBX_MSG_ID_NAV_POSLLH;
	ubx_cmd_msg.length = 0;
	ret = send_ubx_command(fd, &ubx_cmd_msg);
	if (ret == -1) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	/* Get the result */
	ret = get_ubx_response(fd, UBX_MSG_CLASS_NAV, UBX_MSG_ID_NAV_POSLLH, UBX_MSG_FRAME_ONLY, &ubx_resp_msg);
	if (ret != 0) {
		error("%s (line %d) - %s", __FUNCTION__, __LINE__, sx1301ar_err_message(sx1301ar_errno));
		SX1301AR_ERR_RETURN(ERR_UBX_CMD_SEND);
	}

	coord->lon = (double)(ubx_resp_msg.payload[4] | (ubx_resp_msg.payload[5] << 8) | (ubx_resp_msg.payload[6] << 16) | (ubx_resp_msg.payload[7] << 24));
	coord->lat = (double)(ubx_resp_msg.payload[8] | (ubx_resp_msg.payload[9] << 8) | (ubx_resp_msg.payload[10] << 16) | (ubx_resp_msg.payload[11] << 24));
	coord->alt = (short)(ubx_resp_msg.payload[12] | (ubx_resp_msg.payload[13] << 8));

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static char nibble_to_hexchar(uint8_t nibble)
{
	if (nibble < 10) {
		return '0' + nibble;
	} else if (nibble < 16) {
		return 'A' + (nibble - 10);
	} else {
		return '?';
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void calc_ubx_checksum(const uint8_t *buff, size_t buff_size, uint16_t *checksum)
{
	uint8_t i;
	uint8_t chka = 0;
	uint8_t chkb = 0;

	for (i = 0; i < buff_size; i++) {
		chka = chka + buff[i];
		chkb = chkb + chka;
	}

	//debug("CK_A = %x CK_B = %x", chka, chkb );
	*checksum = (chka << 8) | chkb;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int calc_nmea_checksum(const char *buff, size_t buff_size, int *nmea_term)
{
	const int buff_size_nb = buff_size / sizeof(char);
	int i = 0;
	uint8_t check_num = 0;

	/* Check input parameters */
	if ((buff == NULL) || (buff_size_nb <= 2)) {
		return -1;
	}

	/* Skip the optional '$' preamble character */
	if (buff[i] == '$') {
		i += 1;
	}

	/* XOR until '*' terminator, null char or max length is reached */
	while ((i < buff_size_nb) && (buff[i] != '*') && (buff[i] != '\0')) {
		check_num ^= (uint8_t)buff[i];
		++i;
	}

	/* Return error if '*' terminator was not reached */
	if ((i >= buff_size_nb) || (buff[i] == '\0')) {
		return -1;
	}

	/* Send back optional '*' nmea string terminator index */
	if (nmea_term != NULL) {
		*nmea_term = i;
	}

	return (int)check_num;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static bool validate_nmea_checksum(const char *buff, size_t buff_size)
{
	const int buff_size_nb = buff_size / sizeof(char);
	int cs_calc;		 /* calculated checksum */
	int nmea_term = INT_MAX; /* index of NMEA '*' character */
	char cs1, cs2;		 /* checksum characters (in the buffer) */
	char cs1_calc, cs2_calc; /* checksum characters (calculated) */

	/* Check input parameters */
	if ((buff == NULL) || (buff_size_nb <= 2)) {
		return false;
	}

	/* Calculate checksum */
	cs_calc = calc_nmea_checksum(buff, buff_size, &nmea_term);
	if (cs_calc < 0) {
		return false;
	}
	if (buff_size_nb <= nmea_term + 2) {
		return false; /* not enough room for checksum characters */
	}

	/* Compare with characters at the end of the string */
	cs1_calc = nibble_to_hexchar((0xF0 & cs_calc) >> 4);
	cs2_calc = nibble_to_hexchar(0x0F & cs_calc);
	cs1 = buff[nmea_term + 1];
	cs2 = buff[nmea_term + 2];
	if ((cs1 == cs1_calc) && (cs2 == cs2_calc)) {
		return true;
	} else {
		return false;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int strnchopchop(char *buff, size_t buff_size, char separator, int *idx, size_t idx_size)
{
	const int buff_size_nb = buff_size / sizeof(char);
	const int idx_size_nb = idx_size / sizeof(int);
	int i = 0; /* index in the string */
	int j = 0; /* index in the token offset array */

	/* Check input parameters */
	if ((buff == NULL) || (buff_size_nb < 1) || (idx == NULL) || (idx_size_nb < 1)) {
		return -1;
	}

	/* First token start at the beginning */
	idx[j] = 0;

	/* Search & replace loop */
	while ((i < buff_size_nb) && (j < idx_size_nb) && (buff[i] != '\0')) {
		if (buff[i] == separator) {
			buff[i] = '\0'; /* replace separator by string terminator */
			++j;
			if (j < idx_size_nb) {
				idx[j] = i + 1; /* next token begin after current terminator */
			}
		}
		++i;
	}

	/* Return error if end of string was not reached */
	if ((i >= buff_size_nb) || (j >= idx_size_nb)) {
		return -1;
	}

	return j + 1;
}

/* --- EOF ------------------------------------------------------------------ */
