/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    GPS timestamping functions for 'LoRa Gateway v2' library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Sylvain Miermont
*/

#ifndef _SX1301_ARRAY_GPS_H
#define _SX1301_ARRAY_GPS_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <time.h>    /* struct timespec */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC CONSTANTS ----------------------------------------------------- */

#define WARN_NMEA_BAD_CHECKSUM 1 /* checksum is invalid */
#define WARN_NMEA_BAD_FORMAT 2   /* frame doesn't follow NMEA specification */
#define WARN_NMEA_INVALID 4      /* the NMEA frame is marked invalid */
#define WARN_NMEA_INCOMPLETE 8   /* some values could not be returned */

#define WARN_SYNC_OUT_OF_RANGE 1

#define LGW_GPS_MIN_MSG_SIZE (8)
#define LGW_GPS_UBX_SYNC_CHAR (0xB5)
#define LGW_GPS_NMEA_SYNC_CHAR (0x24)

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

/**
@brief Type of NMEA sentences from GPS or other GNSS
*/
typedef enum {
	NMEA_UNDEFINED,
	NMEA_INVALID,
	/* NMEA messages of interest */
	NMEA_RMC, /*!> Recommended Minimum data (time, date, 2D pos) */
	NMEA_GGA, /*!> Global positioning system fix data (pos + alt) */
	/* NMEA message useful for time reference quality assessment */
	NMEA_GNS, /*!> GNSS fix data (pos + alt, sat number) */
	NMEA_ZDA, /*!> Time and Date */
	NMEA_GBS, /*!> GNSS Satellite Fault Detection */
	NMEA_GST, /*!> GNSS Pseudo Range Error Statistics */
	NMEA_GSA, /*!> GNSS DOP and Active Satellites (sat number) */
	NMEA_GSV, /*!> GNSS Satellites in View (sat SNR) */
	/* Misc. NMEA messages */
	NMEA_GLL, /*!> Latitude and longitude, with time fix and status */
	NMEA_TXT, /*!> Text Transmission */
	NMEA_VTG  /*!> Course over ground and Ground speed */
} sx1301ar_nmea_msg_t;

typedef enum {
	UBX_UNDEFINED,
	UBX_INVALID,
	UBX_INCOMPLETE,
	UBX_NAV_TIMEGPS
} sx1301ar_ubx_msg_t;

/**
@brief Structure containing the user configuration for the GPS
*/
typedef struct
{
	bool bfix_altitude;
	uint16_t fix_altitude;
} sx1301ar_gps_cfg_t;

/**
@brief Time reference for internal timestamp from/to UTC time conversion
*/
typedef struct
{
	time_t systime;      /*!> system time when solution was calculated */
	uint32_t hs_pps;     /*!> reference high speed counter */
	uint32_t count_us;   /*!> reference concentrator internal timestamp */
	struct timespec utc; /*!> reference UTC time (from GPS) */
#if 0
	struct timespec gps; /*!> reference GPS time (from GPS), no leap second */
#endif
	double xtal_err;     /*!> raw clock error (eg. <1 'slow' XTAL) */
	double xtal_hs_err;  /*!> raw high speed clock error (eg. <1 'slow' XTAL) */
	uint32_t sync_cnt;   /*!> count nb of updates (saturation (2^32)-1) */
} sx1301ar_tref_t;

/**
@brief Geodesic 3D coordinates
*/
typedef struct
{
	double lat; /*!> latitude [-90,90] (North +, South -) */
	double lon; /*!> longitude [-180,180] (East +, West -)*/
	short alt;  /*!> altitude in meters (WGS 84 geoid ref.) */
} sx1301ar_coord_t;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */
/** @warning This module has no internal state. */

/**
@brief Return a sx1301ar_gps_cfg structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_gps_cfg_t sx1301ar_init_gps_cfg(void);

/**
@brief Set  Configuration of the GPS
@param cfg  Pointer to configuration structure for the GPS
@return 0 if the GPS configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_gps(const sx1301ar_gps_cfg_t *cfg);

/**
@brief Configuration of the GPS
@param fd   File descriptor of GPS serial device
@return 0 if the GPS configuration was done, -1 else
*/
int sx1301ar_init_gps(int fd);

/**
@brief Revert configuration of the GPS
@param fd   File descriptor of GPS serial device
@return 0 if the GPS configuration was done, -1 else
*/
int sx1301ar_term_gps(int fd);

/**
@brief Identify the type of NMEA frame, for parsing or to trigger actions
@param buff       Pointer to a character buffer containing the NMEA string
@param buff_size  Size of the character buffer (= max size of the string)
@param type       Pointer to return the identified type of the NMEA frame
@return 0 if the frame could be parsed (NMEA or not), -1 else

This function only checks the first few characters of the string to identify
a NMEA frame.
Accepts
*/
int sx1301ar_get_nmea_type(const char *buff, size_t buff_size, sx1301ar_nmea_msg_t *type);

/**
@brief Parse a NMEA frame of type RMC
@param buff       Pointer to a character buffer containing the NMEA string
@param buff_size  Size of the character buffer (= max size of the string)
@param utc        Pointer to return UTC in Unix time + subsecond in ns format
@param pos        Pointer to return geodesic coordinates (2D only in RMC)
@return 0 if RMC frame is valid, >0 if 'invalid', -1 for critical errors

If NULL pointers are passed as 'utc' or 'pos' parameters, those values are not
parsed from the NMEA frame.
Validates the NMEA cheksum and returns WARN_NMEA_BAD_CHECKSUM if bad.
If the NMEA frame validity indicator is V (invalid), returns WARN_NMEA_INVALID.
*/
int sx1301ar_parse_rmc(char *buff, size_t buff_size, struct timespec *utc, sx1301ar_coord_t *pos);

/**
@brief Parse a NMEA frame of type GGA
@param buff       Pointer to a character buffer containing the NMEA string
@param buff_size  Size of the character buffer (= max size of the string)
@param pos        Pointer to return geodesic coordinates (3D if enough sats)
@param n_sat      Pointer to return the number of satellites used for the fix
@param hdop       Pointer to return horizontal dilution of precision
@return 0 if GGA frame is valid, >0 if 'invalid', -1 for critical errors

If NULL pointers are passed as 'pos', 'n_sat' or 'hdop' parameters, those
values are not parsed from the NMEA frame.
Validates the NMEA cheksum and returns WARN_NMEA_BAD_CHECKSUM if bad.
If the NMEA quality indicator is not good, returns WARN_NMEA_BAD_QUALITY.
*/
int sx1301ar_parse_gga(char *buff, size_t buff_size, sx1301ar_coord_t *pos, int *n_sat, float *hdop);

/**
@brief Identify the type of UBX frame, for parsing or to trigger actions
@param buff       Pointer to a buffer containing the UBX packet
@param buff_size  Size of the buffer (= max size of the string)
@param type       Pointer to return the identified type of the UBX frame
@param length     Length of the identified UBX frame
@return 0 if the frame could be parsed (UBX or not), -1 else
*/
int sx1301ar_get_ubx_type(const char *buff, size_t buff_size, sx1301ar_ubx_msg_t *type, size_t *length);

/**
@brief Parse a UBX frame of type NAV-TIMEGPS
@param buff       Pointer to a buffer containing the UBX packet
@param buff_size  Size of the buffer (= max size of the string)
@param timegps    Pointer to return GPS time with ns precision
@return 0 if GPS time is valid, -1 for critical errors or invalid GPS time
*/
int sx1301ar_parse_ubx_timegps(const char *buff, size_t buff_size, struct timespec *timegps);

/**
@brief Use a UTC time and matched internal timestamp to update a time reference
@param ref        Pointer to time reference to update
@param utc        UTC time, with ns precision
@param gps        GPS time, with ns precision (no leap second)
@param count_us   Internal concetrator timestamp latched on the PPS pulse
@param hs_pps     Internal high speed counter latched on the PPS pulse
@param pps_delay  Return the number of PPS pulses (Difference between the last reference)
@return 0 if reference was updated, >0 if problems, -1 for critical errors

Check the new synchronization point compared to the old reference. If the slope
is out of expected frequency error (+/- X ppm) the function return the positive
value WARN_SYNC_OUT_OF_RANGE and DO NOT UPDATE the time reference.
*/
int sx1301ar_synchronize(sx1301ar_tref_t *ref, struct timespec utc,
#if 0
			 struct timespec gps,
#endif
			 uint32_t count_us, uint32_t hs_pps, int *pps_delay);

/**
@brief Set the board with XTAL error value, and validate value
@param brd        Target board number
@param ref        Time reference used for conversion
@return 0 if conversion was successful, -1 else
*/
int sx1301ar_set_xtal_err(uint8_t brd, sx1301ar_tref_t ref);

#if 0
/**
@brief Convert an internal concentrator timestamp into an absolute UTC time
@param ref        Time reference used for conversion
@param count_us   Internal timestamp to convert
@param utc        Pointer to return corresponding absolute UTC time
@return 0 if conversion was successful, -1 else
*/
int sx1301ar_cnt2utc(sx1301ar_tref_t ref, uint32_t count_us, struct timespec *utc);

/**
@brief Convert an absolute UTC time into an internal concentrator timestamp
@param ref        Time reference used for conversion
@param utc        Absolute UTC time to convert
@param count_us   Pointer to return corresponding internal timestamp
@return 0 if conversion was successful, -1 else
*/
int sx1301ar_utc2cnt(sx1301ar_tref_t ref, struct timespec utc, uint32_t *count_us);

/**
@brief Convert concentrator timestamp counter value to GPS time

@param ref      Time reference used for conversion
@param count_us Internal timestamp counter to convert
@param gps_time Pointer to store GPS time, with ns precision (leap seconds ignored)
@return 0 if the function was able to convert timestamp to GPS time, -1 else

This function is typically used when a packet is received to transform the
internal counter-based timestamp in an absolute timestamp with an accuracy in
the order of a millisecond.
*/
int sx1301ar_cnt2gps(sx1301ar_tref_t ref, uint32_t count_us, struct timespec *gps_time);

/**
@brief Convert GPS time to concentrator timestamp counter value

@param ref      Time reference used for conversion
@param gps_time Absolute GPS time to convert
@param count_us Pointer to store internal timestamp counter
@return 0 if the function was able to convert GPS time to timestamp, -1 else

This function is typically used when a packet must be sent at an accurate time
(eg. to send a Class-B downlink) to transform an absolute GPS time into a
matching internal concentrator timestamp.
*/
int sx1301ar_gps2cnt(sx1301ar_tref_t ref, struct timespec gps_time, uint32_t *count_us);

/**
@brief Convert an internal high speed counter into an absolute UTC time
@param ref        Time reference used for conversion
@param ts_hs      High speed counter value to convert
@param utc        Pointer to return corresponding absolute UTC time
@return 0 if conversion was successful, -1 else
*/
int gwloc_hs2utc(sx1301ar_tref_t ref, uint32_t ts_hs, struct timespec *utc);
#endif

/**
@brief Return a sx1301ar_tref_t structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_tref_t sx1301ar_init_tref(void);

#endif

/* --- EOF ------------------------------------------------------------------ */
