/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Hardware Abstraction Layer for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <math.h>    /* sin, cos, sqrt */
#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <stdio.h>   /* NULL */
#include <string.h>  /* memcpy */

#include "aux.h"
#include "dsp.h"
#include "err.h"
#include "hal.h"
#include "lbt.h"
#include "radio.h"
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#ifndef VERSION_STRING
#define VERSION_STRING "v0.0.?"
#endif

#define MCU_ARB_FW_BYTE 8192 /* size of fw IN BYTES (2x nb of 14b words) */
#define MCU_AGC_FW_BYTE 8192 /* size of fw IN BYTES (2x nb of 14b words) */
#define FW_VERSION_ADDR 0x20 /* Address of firmware version in data memory */
#define FW_VERSION_AGC 13    /* Expected version of AGC firmware */
#define FW_VERSION_ARB 3     /* Expected version of arbiter firmware */

#include "firmware/sx1301/agc.h" /* external definition of the variable */
#include "firmware/sx1301/arb.h" /* external definition of the variable */

#define MULTI_DEFAULT_SF MR_SF7_10  /* LoRa multi-SF default SFs */
#define LSA_DEFAULT_BW BW_125K      /* LoRa stand-alone default bandwidth */
#define LSA_DEFAULT_SF 10	   /* LoRa stand-alone default SF */
#define FSK_DEFAULT_BW BW_125K      /* FSK default bandwidth */
#define FSK_DEFAULT_MR MR_64K       /* FSK default modulation rate */
#define FSK_DEF_SYNC_SIZE 4	 /* FSK default sync word length (bytes) */
#define FSK_DEF_SYNC_MSB 0x01010101 /* FSK default sync word (MSB first) */
#define FSK_DEF_SYNC_LSB 0x01010101 /* FSK default sync word (MSB first) */

typedef enum fw_proc {
	MCU_ARB,
	MCU_AGC
} fw_proc_t;

#define RSSI_MULTI_BIAS 19.5    /* difference between "multi" modem RSSI offset and "stand-alone" modem RSSI offset */
#define RSSI_FSK_POLY_0 54.5494 /* polynomiam coefficients to linearize FSK RSSI */
#define RSSI_FSK_POLY_1 1.5351
#define RSSI_FSK_POLY_2 0.003

#define TX_METADATA_NB 16
#define RX_METADATA_NB 16

#define CHIP_XTAL_FREQU 32000000 /* frequency of the RF reference oscillator */

#define MIN_LORA_PREAMBLE 6
#define STD_LORA_PREAMBLE 8
#define MIN_FSK_PREAMBLE 3
#define STD_FSK_PREAMBLE 5

#define TX_NOTCH_TUNE 50
#define TX_START_DELAY_DEFAULT 1494U /* Calibrated value for 500KHz BW and notch filter disabled */

/* IF hardware limits, practical max IF (taking filtering into account) is lower */
#define CHAN_IF_MAX 1500000 /* filter has ~3 MHz bandwidth */

#define CAPTURE_RAM_SIZE 4096

#define SNR_THRESHOLD_MAX 15.0 /* Clipping value of SNR */

#define SX1301_DSP_RSSI_OFFSET 61.5

/* DSP boot options */
#define GWLOC_BOOT_RETRY 5

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

/* -------------------------------------------------------------------------- */
/* --- PRIVATE TYPES -------------------------------------------------------- */

typedef struct
{
	uint8_t chan;
	float rssi;
	float snr;
} pkt_rsig_t;

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

/* Array of structures to keep track of internal configuration parameters */
/* We rely on static variables being null-initialized at launch */
static struct
{
	sx1301ar_btype_t btype; /* Board type: Master or slave */
	int16_t fpga_version;   /* FPGA version */
	bool is_started;	/* Is the board started or not */
	/* Radio configuration */
	uint32_t rx_freq; /* Center RX frequency of the radio, in Hz */
	uint32_t rx_bw;   /* RX bandwidth of the radio, in Hz */
	bool full_duplex; /* Full-duplex / Half-duplex mode */
	/* Radio chains configuration */
	sx1301ar_rfchain_t rf_chain[SX1301AR_BOARD_RFCHAIN_NB];
	bool rf_diversity; /* indicates if antenna diversity is used on this board */
	/* Phy parameters that can be protocol-specific */
	uint32_t fsk_sync_msb; /* FSK sync word register value (aligned left) */
	uint32_t fsk_sync_lsb; /* FSK sync word register value (aligned left) */
	uint8_t fsk_sync_size; /* FSK sync word size */
	bool loramac_public;   /* Enable ONLY for *public* networks using the LoRa MAC protocol */
	/* SX1301 settings */
	uint8_t chip_nb;				/* Number of chip per board */
	bool chip_en[SX1301AR_BOARD_CHIPS_NB];		/* Is SX1301 enabled or not */
	uint8_t chip_rf_chain[SX1301AR_BOARD_CHIPS_NB]; /* Select Rx RF path */
	uint32_t chip_freq[SX1301AR_BOARD_CHIPS_NB];    /* Center RX frequency of the SX1301 chips, in Hz */
	/* Channels settings */
	uint32_t chan_en[SX1301AR_BOARD_CHIPS_NB][SX1301AR_CHIP_CHAN_NB];   /* Is RX channel enabled or not */
	uint32_t chan_freq[SX1301AR_BOARD_CHIPS_NB][SX1301AR_CHIP_CHAN_NB]; /* Center RX frequency of the channels, in Hz */
	uint8_t multi_sf[SX1301AR_BOARD_CHIPS_NB][SX1301AR_CHIP_MULTI_NB];  /* SF mask for LoRa multi-SF channels */
	sx1301ar_bandw_t lsa_bw[SX1301AR_BOARD_CHIPS_NB];		    /* RX bandwidth for the LoRa stand-alone channel */
	uint8_t lsa_sf[SX1301AR_BOARD_CHIPS_NB];			    /* Spreading factor for the LoRa stand-alone channel */
	sx1301ar_bandw_t fsk_bw[SX1301AR_BOARD_CHIPS_NB];		    /* RX bandwidth for the FSK channel */
	uint32_t fsk_br[SX1301AR_BOARD_CHIPS_NB];			    /* Baud rate for the FSK channel */
	/* DSP settings */
	uint8_t dsp_nb;      /* Number of DSP */
	int16_t dsp_version; /* DSP version, -1 if unknown */
	uint8_t dsp_stat_interval;
	int8_t room_temp_ref;    /* reference room temperature (Tref) */
	uint8_t ad9361_temp_ref; /* temperature code returned by radio sensor when room is at Tref */
	/* Options */
	bool match_tmst_crc_err;   /* match fine timestamps for packets with CRC error */
	uint8_t main_tmst_version; /* version of the encrypted/main fine timestamp to be sent by the DSP */
	bool debug_tmst;	   /* enable/disable fine timestamp debug mode of DSP */
	bool fine_timestamp_decrypt_enable;
} brd_cfg_priv[SX1301AR_MAX_BOARD_NB];

/* Internal: Board AES key for fine timestamp decryption */
uint8_t brd_aes_key[SX1301AR_MAX_BOARD_NB][SX1301AR_BOARD_AES_KEY_SIZE / 8];

/* Internal channel diversity table:
    Single Antenna:
            - No Diversity: no channel in common between SX1301_0 and SX1301_1
    Dual Antenna:
            - Full Diversity: same channels defined on SX1301_0 and SX1301_1
            - Partial Diversity: only some channels are identical
            - No Diversity: as for single antenna (useless)
    Ex: table: { 0, 0, 0, 1, 1, 0, 1, 0, 0 } means:
            channel 0|16: no diversity
            channel 1|17: no diversity
            channel 2|18: diversity
            channel 3|19: diversity
            channel 4|20: no diversity
            channel 5|21: diversity
            channel 6|22: no diversity
            channel 7|23: no diversity
*/
static bool brd_channel_diversity_table[SX1301AR_MAX_BOARD_NB][SX1301AR_CHIP_CHAN_NB];

/* Internal: Frequency error compensation due to round of IF freq */
static int16_t brd_chip_if_bias[SX1301AR_MAX_BOARD_NB][SX1301AR_BOARD_CHIPS_NB];
static int16_t brd_chan_if_bias[SX1301AR_MAX_BOARD_NB][SX1301AR_BOARD_CHIPS_NB * SX1301AR_CHIP_CHAN_NB];

/* Internal received packet history */
static struct
{
	uint32_t count_us;
	uint8_t modrate;
	uint8_t chan;
} rxpk_history[SX1301AR_MAX_BOARD_NB][SX1301AR_MAX_PKT_NB];
static uint8_t rxpk_history_nb = 0;

/* Fine timestamps Queue */
static gwloc_tmst_queue_t fine_timestamp_queue;

/* Fine timestamps statistics variables */
static uint32_t nb_pkt_sx1301_received_crc_ok = 0;
static uint32_t nb_pkt_sx1301_received_crc_err = 0;
static uint32_t nb_fine_tmst_received = 0;
static uint32_t nb_match = 0;
static uint32_t nb_no_fine_timestamp = 0;
static uint32_t nb_invalid_fine_timestamp = 0;

#if 0
static uint32_t debug_history = 0;
#endif

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

/**
@brief Load a firmware inside a controller program memory
@param brd     Target board number
@param chip    Target chip number in the array
@param target  Target controller inside the chip
@param fw      Pointer to byte array containing the firmware
@param size    Size of the firmware (in bytes))
@return 0 if firmware was loaded without error, -1 else
*/
static int load_firmware(uint8_t brd, uint8_t chip, fw_proc_t target, const uint8_t *fw, uint32_t size);

/**
@brief Calculate RSSI by correcting raw measurement
@param brd       Target board number
@param cx        Local address of the channel (in the chip)
@param raw_rssi  RAW unsigned RSSI (0 ~ noise floor)
@param temperature_code Current temperature code, for RSSI correction
@return calibrated RSSI in absolute dBm
*/
static float calc_rssi(uint8_t brd, uint8_t cx, uint8_t raw_rssi, uint8_t rx_chain, uint8_t temperature_code);

/**
@brief Determine if 'reduced PPM mode' must be activated in LoRa
@param bw  LoRa modulation bandwidth
@param sf  LoRa spreading factor
@return true if 'reduce PPM mode' must be activated
*/
static bool ppm_mode_en(sx1301ar_bandw_t bw, uint8_t sf);

/**
@brief Get the index of the TX gain LUT corresponding to the requested power
@param brd       Target board number
@param rf_chain  Target RF chain number
@param rf_power  TX power for which we want the index in the LUT
@param lut_index Pointer to the LUT index
@return 0 if success
*/
static int get_tx_gain_lut_index(uint8_t brd, uint8_t rf_chain, int8_t rf_power, uint8_t *lut_index);

/**
@brief Get the temperature compensation to be applied for given TX gain LUT index
@param brd       Target board number
@param rf_chain  Target RF chain number
@param lut_index LUT index for which to apply compensation
@param atten     Pointer to atten correction to be applied
@param auxdac    Pointer to auxdac correction to be applied
@return 0 if success
*/
static int get_tx_gain_temperature_compensation(uint8_t brd, uint8_t rf_chain, uint8_t lut_index, int8_t *atten, int8_t *auxdac);

/**
@brief Get the temperature compensation to be applied to RX RSSI for given temperature code
@param brd              Target board number
@param rf_chain         Target RF chain number
@param temperature_code Temperature code measured by AD9361
@return the RSSI offset to be applied to compensate given temperature
*/
static float get_rx_rssi_temperature_compensation(uint8_t brd, uint8_t rf_chain, uint8_t temperature_code);

/**
@brief Get the TX start delay to be applied based on TX notch filter activation
@param tx_notch_enable  Is TX notch filter enabled for this transmission ?
@param bw               Bandwidth of the signal to be transmitted
@return TX start delay to be applied.
*/
static uint16_t get_tx_start_delay(bool tx_notch_enable, sx1301ar_bandw_t bw);

#if 0 /* Not used for now */
/**
@brief Capture I/Q samples from AD9361
@param brd          Target board number
@param sig_i        Pointer to I vector
@param sig_q        Pointer to Q vector
@param capture_len  Number of samples to capture
@return 0
*/
static int ad9361_capture( uint8_t brd, int16_t *sig_i, int16_t *sig_q, int capture_len );

/**
@brief Perform a correlation on AD9361 capture at a given frequency
@param brd        Target board number
@param f          Frequency to correlate in Hz
@param avg_len    Number of samples to correlate
@return absolute value of correlation
*/
static uint16_t ad9361_sig_analyser( uint8_t brd, float f, int avg_len );
#endif

/**
@brief Configure the board before starting it
@param brd              Target board number to be configured
@return 0 if success.
*/
static int configure_board(uint8_t brd);

/**
@brief Fetch packets from SX1301s
@param brd              Target board number
@param p                Pointer to array for receiving packet data and metadata
@param p_rsig           Pointer to array of received signal information corresponding to
                            SX1301 packets (same index)
@param max_nb           Maximum number of packets to fetch (0 is accepted)
@param nb_pkt           Pointer to a uint8_t to receive number of returned packets
@param temperature_code Current temperature code, for RSSI correction
@return 0 if success.
*/
static int fetch_packets(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t max_nb, uint8_t *nb_pkt, uint8_t temperature_code);

/**
@brief Merge duplicated SX1301 packets received on different antennas
@param p        Pointer to array for receiving packet data and metadata
@param p_rsig   Pointer to the array of received signal information corresponding to
                    SX1301 packets (same index)
@param nb_pkt   Pointer to the updated number of returned packets after merge
@return 0 if success.
*/
static int merge_packets(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt);

/**
@brief Fetch FPGA FIFO for fine timestamps
@param brd      Target board number
@param queue    Queue to store fetched fine timestamps
@return 0 if success.
*/
static int fetch_fpga(uint8_t brd, gwloc_tmst_queue_t *queue);

/**
@brief Fetch DSP for messages (logs, status, config)
@param brd      Target board number
@return 0 if success.
*/
static int fetch_dsp(uint8_t brd);

/**
@brief Merge packets with matching fine timestamps
@param brd              Target board number
@param p                Pointer to array for receiving packet data and metadata
@param p_rsig           Pointer to array of received signal information corresponding to
                            SX1301 packets (same index)
@param nb_pkt           Pointer to the number of returned packets in array
@param queue            Queue to get fine timestamps from, for matching
@param temperature_code Current temperature code, for RSSI correction
@return 0 if success.
*/
static int merge_timestamps(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt, gwloc_tmst_queue_t *queue, uint8_t temperature_code);

/**
@brief Remove a packet from received packet array
@param p            Pointer to array for receiving packet data and metadata
@param p_rsig       Pointer to array of received signal information corresponding to
                        SX1301 packets (same index)
@param nb_pkt       Pointer to the number of packets in array, updated after removal
@param pkt_index    Index of the packet to be removed from the array
@return 0 if success.
*/
static int remove_pkt(sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt, uint8_t pkt_index);

/**
@brief Check if 2 packets are identical
@param p1            Pointer to first packet
@param p1_rsig       Pointer to first packet's received signal information
@param p2            Pointer to second packet
@param p2_rsig       Pointer to second packet's received signal information
@return true if packets are identical, false otherwise.
*/
static bool is_same_pkt(sx1301ar_rx_pkt_t *p1, pkt_rsig_t *p1_rsig, sx1301ar_rx_pkt_t *p2, pkt_rsig_t *p2_rsig);

/**
@brief Save received packets history
@param brd      Target board number
@param p        Pointer to array of received packet data and metadata
@param nb_pkt   number of packets in the array
@return 0 if success.
*/
static int history_save(uint8_t brd, sx1301ar_rx_pkt_t *p, uint8_t nb_pkt);

/**
@brief Check received packets history to find ghost packets (packets which have already been sent)
@param brd      Target board number
@param p        Pointer to array of received packet data and metadata
@param p_rsig   Pointer to array of received signal information corresponding to
                        SX1301 packets (same index)
@param nb_pkt   Pointer to the number of packets in array, updated after removal of ghost packets
@return 0 if success.
*/
static int history_check(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt);

/**
@brief Set the configuration of TX gain table
@param brd      Target board number
@param rf_chain Target RF chain number
@param cfg      Pointer to configuration structure for TX gain table
@return 0 if the configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_tx_gain(uint8_t brd, uint8_t rf_chain, const sx1301ar_tx_gain_lut_t *cfg);

/**
@brief Internal function to compute RX & TX packet time on air of LoRa packets, in milliseconds
@param
@param
@return the packet time on air in milliseconds
*/
static double sx1301ar_i_lora_time_on_air(sx1301ar_bandw_t pkt_bw, sx1301ar_modr_t pkt_modr, uint8_t pkt_size, sx1301ar_coder_t pkt_cr, bool no_header, uint16_t preamb_size, double *t_preamble, double *t_syncword, double *t_payload);

/**
@brief Internal function to compute RX & TX packet time on air of FSK packets, in milliseconds
@param
@param
@return the packet time on air in milliseconds
*/
static double sx1301ar_i_fsk_time_on_air(uint8_t brd, sx1301ar_modr_t pkt_modr, uint16_t preamb_size, uint8_t pkt_size, bool no_crc);

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

const char *sx1301ar_version_info(uint8_t brd, int16_t *fpga_version, int16_t *dsp_version)
{
	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		error("Invalid board number (%d)", brd);
	} else {
		/* FPGA */
		if (fpga_version != NULL) {
			*fpga_version = brd_cfg_priv[brd].fpga_version;
		}

		/* DSP */
		if (dsp_version != NULL) {
			*dsp_version = brd_cfg_priv[brd].dsp_version;
		}
	}

	/* HAL */
	return VERSION_STRING;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/* ~ Configuration functions ~~~~~~ */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_board(uint8_t brd, const sx1301ar_board_cfg_t *cfg)
{
	int i, x;
	uint64_t sync_word_reg;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		error("Invalid board number");
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cfg->rx_freq_hz < SX1301AR_MIN_FREQ) {
		error("Invalid board frequency");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}
	if ((cfg->spi_read == NULL) || (cfg->spi_write == NULL)) {
		error("Invalid board spi function pointer");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}
	if (cfg->fsk_sync_size > 8) {
		error("Invalid board fsk sync size");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}
	if ((cfg->rx_bw_hz != 4e6) && (cfg->rx_bw_hz != 7e6) && (cfg->rx_bw_hz != 8e6) && (cfg->rx_bw_hz != 13e6)) {
		error("Invalid board bandwidth");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}
	if ((cfg->board_type != BRD_MASTER) && (cfg->board_type != BRD_SLAVE)) {
		error("Invalid board type");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}
	if ((cfg->nb_chip == 0) || (cfg->nb_chip > SX1301AR_BOARD_CHIPS_NB)) {
		error("Invalid board chip number");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}

	/* initialize memory for global variables */
	memset(&brd_channel_diversity_table[brd], 0, sizeof brd_channel_diversity_table[brd]);

	/* No reconfiguration if board is running */
	if (brd_cfg_priv[brd].is_started == true) {
		SX1301AR_ERR_RETURN(ERR_CANT_CFG);
	}

	/* Default values */
	brd_cfg_priv[brd].fpga_version = -1;    /* unknown yet */
	brd_cfg_priv[brd].dsp_version = -1;     /* unknown yet */
	brd_cfg_priv[brd].rf_diversity = false; /* unknown yet */

	/* Record configuration values */
	brd_cfg_priv[brd].btype = cfg->board_type;
	brd_cfg_priv[brd].loramac_public = cfg->loramac_public;
	brd_cfg_priv[brd].full_duplex = cfg->full_duplex;

	info("board #%zu: Using board: type=%s loramac_public=%s full_duplex=%s",
	     brd,
	     (brd_cfg_priv[brd].btype == BRD_MASTER) ? "MASTER" : (brd_cfg_priv[brd].btype == BRD_SLAVE) ? "SLAVE" : "UNKNOWN",
	     (brd_cfg_priv[brd].loramac_public) ? "true" : "false",
	     (brd_cfg_priv[brd].full_duplex) ? "true" : "false");

	brd_cfg_priv[brd].rx_freq = cfg->rx_freq_hz;
	brd_cfg_priv[brd].rx_bw = cfg->rx_bw_hz;

	info("board #%zu: Using board: rx_freq_hz=%u rx_bw_hz=%u",
	     brd,
	     brd_cfg_priv[brd].rx_freq,
	     brd_cfg_priv[brd].rx_bw);

	brd_cfg_priv[brd].chip_nb = cfg->nb_chip;
	brd_cfg_priv[brd].dsp_nb = cfg->nb_dsp;
	brd_cfg_priv[brd].dsp_stat_interval = cfg->dsp_stat_interval;

	info("board #%zu: Using board: nb_chip=%u nb_dsp=%u dsp_stat_interval=%u",
	     brd,
	     brd_cfg_priv[brd].chip_nb,
	     brd_cfg_priv[brd].dsp_nb,
	     brd_cfg_priv[brd].dsp_stat_interval);

	for (i = 0; i < SX1301AR_BOARD_RFCHAIN_NB; i++) {
		brd_cfg_priv[brd].rf_chain[i].rx_enable = cfg->rf_chain[i].rx_enable;
		brd_cfg_priv[brd].rf_chain[i].tx_enable = cfg->rf_chain[i].tx_enable;
		brd_cfg_priv[brd].rf_chain[i].rssi_offset = cfg->rf_chain[i].rssi_offset;
		brd_cfg_priv[brd].rf_chain[i].rssi_offset_coeff_a = cfg->rf_chain[i].rssi_offset_coeff_a;
		brd_cfg_priv[brd].rf_chain[i].rssi_offset_coeff_b = cfg->rf_chain[i].rssi_offset_coeff_b;

		if (brd_cfg_priv[brd].rf_chain[i].rx_enable) {
			info("board #%zu: Using chain #%zu: enable=true rssi_offset=%g rssi_offset_coeff_a=%i rssi_offset_coeff_b=%i tx_enable=%s",
			     brd, i,
			     brd_cfg_priv[brd].rf_chain[i].rssi_offset,
			     brd_cfg_priv[brd].rf_chain[i].rssi_offset_coeff_a,
			     brd_cfg_priv[brd].rf_chain[i].rssi_offset_coeff_b,
			     (brd_cfg_priv[brd].rf_chain[i].tx_enable) ? "true" : "false");
		} else {
			info("board #%zu: Using chain #%zu: enable=false", brd, i);
		}

		if (brd_cfg_priv[brd].rf_chain[i].rx_enable && brd_cfg_priv[brd].rf_chain[i].tx_enable) {
			if (cfg->rf_chain[i].tx_lut.size > 0) {
				x = sx1301ar_conf_tx_gain(brd, i, &(cfg->rf_chain[i].tx_lut));
				if (x != 0) {
					error("failed to configure TX Gain LUT for RF chain %d (%s)", i, sx1301ar_err_message(sx1301ar_errno));
				}
			}
		}
	}

	brd_cfg_priv[brd].room_temp_ref = cfg->room_temp_ref;
	brd_cfg_priv[brd].ad9361_temp_ref = cfg->ad9361_temp_ref;

	info("board #%zu: Using board: room_temp_ref=%i ad9361_temp_ref=%u",
	     brd,
	     brd_cfg_priv[brd].room_temp_ref,
	     brd_cfg_priv[brd].ad9361_temp_ref);

	brd_cfg_priv[brd].fine_timestamp_decrypt_enable = cfg->fine_timestamp_decrypt_enable;
	if (brd_cfg_priv[brd].fine_timestamp_decrypt_enable) {
		memcpy(brd_aes_key[brd], cfg->aes_key, sizeof brd_aes_key[brd]);
		info("board #%zu: Using board: fts_decrypt_enable=true "
		     "fts_aes_key=%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X",
		     brd,
		     brd_aes_key[brd][0],
		     brd_aes_key[brd][1],
		     brd_aes_key[brd][2],
		     brd_aes_key[brd][3],
		     brd_aes_key[brd][4],
		     brd_aes_key[brd][5],
		     brd_aes_key[brd][6],
		     brd_aes_key[brd][7],
		     brd_aes_key[brd][8],
		     brd_aes_key[brd][9],
		     brd_aes_key[brd][10],
		     brd_aes_key[brd][11],
		     brd_aes_key[brd][12],
		     brd_aes_key[brd][13],
		     brd_aes_key[brd][14],
		     brd_aes_key[brd][15]);
	} else {
		info("board #%zu: Using board: fts_decrypt_enable=false", brd);
	}

	brd_cfg_priv[brd].match_tmst_crc_err = cfg->match_tmst_crc_err;
	brd_cfg_priv[brd].main_tmst_version = cfg->main_tmst_version;
	brd_cfg_priv[brd].debug_tmst = cfg->debug_tmst;

	info("board #%zu: Using board: fts_version=%u fts_match_crc_err=%s fts_debug=%s",
	     brd,
	     brd_cfg_priv[brd].main_tmst_version,
	     (brd_cfg_priv[brd].match_tmst_crc_err) ? "true" : "false",
	     (brd_cfg_priv[brd].debug_tmst) ? "true" : "false");

	brd_cfg_priv[brd].fsk_sync_size = cfg->fsk_sync_size;
	if (cfg->fsk_sync_size > 0) {
		sync_word_reg = cfg->fsk_sync_word << (8 - cfg->fsk_sync_size) * 8; /* left-align the sync word */
		brd_cfg_priv[brd].fsk_sync_msb = 0xFFFFFFFF & (sync_word_reg >> 32);
		brd_cfg_priv[brd].fsk_sync_lsb = 0xFFFFFFFF & sync_word_reg;

		info("board #%zu: Using board: fsk_sync=%llX",
		     brd,
		     sync_word_reg >> (8 * (8 - brd_cfg_priv[brd].fsk_sync_size)));
	} else {
		info("board #%zu: Using board: fsk_sync=none", brd);
	}

	/* Record SPI callbacks */
	return sx1301ar_reg_setup(brd, cfg->board_type, cfg->spi_read, cfg->spi_write);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_conf_board(uint8_t brd, sx1301ar_board_cfg_t *cfg)
{
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cfg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Get recorded configuration values */
	cfg->board_type = brd_cfg_priv[brd].btype;
	cfg->nb_chip = brd_cfg_priv[brd].chip_nb;
	cfg->nb_dsp = brd_cfg_priv[brd].dsp_nb;
	cfg->rx_freq_hz = brd_cfg_priv[brd].rx_freq;
	cfg->rx_bw_hz = brd_cfg_priv[brd].rx_bw;
	cfg->full_duplex = brd_cfg_priv[brd].full_duplex;
	for (i = 0; i < SX1301AR_BOARD_RFCHAIN_NB; i++) {
		cfg->rf_chain[i].rx_enable = brd_cfg_priv[brd].rf_chain[i].rx_enable;
		cfg->rf_chain[i].tx_enable = brd_cfg_priv[brd].rf_chain[i].tx_enable;
	}

	cfg->fine_timestamp_decrypt_enable = brd_cfg_priv[brd].fine_timestamp_decrypt_enable;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_chip(uint8_t brd, uint8_t chip, const sx1301ar_chip_cfg_t *cfg)
{
	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		error("Invalid board number");
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		error("Invalid chip number");
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if ((cfg->enable == true) && (cfg->freq_hz < SX1301AR_MIN_FREQ)) {
		error("Invalid chip frequency");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}

	/* No reconfiguration if board is running */
	if (brd_cfg_priv[brd].is_started == true) {
		SX1301AR_ERR_RETURN(ERR_CANT_CFG);
	}

	/* Record enable setting */
	brd_cfg_priv[brd].chip_en[chip] = cfg->enable;
	brd_cfg_priv[brd].chip_rf_chain[chip] = cfg->rf_chain;

	if (chip < brd_cfg_priv[brd].chip_nb) {
		if (cfg->enable == true) {
			/* Record frequency setting */
			brd_cfg_priv[brd].chip_freq[chip] = cfg->freq_hz;

			info("board #%zu: Using chip #%zu: enable=true rf_chain=%u freq_hz=%u",
			     brd, chip,
			     brd_cfg_priv[brd].chip_rf_chain[chip],
			     brd_cfg_priv[brd].chip_freq[chip]);
		} else {
			info("board #%zu: Using chip #%zu: enable=false", brd, chip);
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_chan(uint8_t brd, uint8_t chan, const sx1301ar_chan_cfg_t *cfg)
{
	uint8_t chip = chan >> 4;
	uint8_t cx = chan & 0x0F; /* local channel address in the chip */
	int sf;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		error("Invalid board number");
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		error("Invalid chip number");
		SX1301AR_ERR_RETURN(ERR_BAD_CHAN_NB);
	}
	if (cx >= SX1301AR_CHIP_CHAN_NB) {
		error("Invalid chan number");
		SX1301AR_ERR_RETURN(ERR_BAD_CHAN_NB);
	}
	if ((cfg->enable == true) && (cfg->freq_hz < SX1301AR_MIN_FREQ)) {
		error("Invalid chan frequency");
		SX1301AR_ERR_RETURN(ERR_BAD_CFG);
	}

	/* No reconfiguration if board is running */
	if (brd_cfg_priv[brd].is_started == true) {
		SX1301AR_ERR_RETURN(ERR_CANT_CFG);
	}

	/* Record enable setting */
	brd_cfg_priv[brd].chan_en[chip][cx] = cfg->enable;

	if (chip < brd_cfg_priv[brd].chip_nb) {
		if (cfg->enable == true) {
			/* Check that the corresponding chip is enabled */
			if (brd_cfg_priv[brd].chip_en[chip] == false) {
				SX1301AR_ERR_RETURN(ERR_CHIP_DISABLE);
			}

			/* Check channel frequency against radio & chip configuration:
         *  => we want to ensure that all channels fit in the the sx1301 band
         *     AND in the radio band.
         *  Note: the sx1301 bandwidth could be partly outside from the radio
         *        band as long as there is no channel in this "outer" band.
         *
         *                             rx_freq (ad9361)
         *                                ^              chip_freq (sx1301)
         *                                |                    ^ chan_freq
         *                                |                    |    ^
         *                                |                    |    |
         *                                |<----------rx_bw/2-------->
         *   -----------------------------|--------------------|----|--
         *  /                             |                    |    |  \
         * /                              |         -----------|----|---\--
         *                                |        /<--------->|    |    \ \
         *                                |       / CHAN_IF_MAX|    |     \ \
         *                                |      /             |    |      \ \
         *--------------------------------|-----/--------------|-|--|--|----\-\----->f
         *                                |    /                 |<--->|     \ \
         *                                |                      chan_bw
         */
			/* Check that the channel is in the SX1301 band */
			if (cfg->freq_hz > (brd_cfg_priv[brd].chip_freq[chip] + CHAN_IF_MAX - sx1301ar_bw_enum2nb(cfg->bandwidth) / 2)) {
				SX1301AR_ERR_RETURN(ERR_IF_LIMIT);
			}
			if (cfg->freq_hz < (brd_cfg_priv[brd].chip_freq[chip] - CHAN_IF_MAX + sx1301ar_bw_enum2nb(cfg->bandwidth) / 2)) {
				SX1301AR_ERR_RETURN(ERR_IF_LIMIT);
			}
			/* Check that the channel is in the radio band */
			if (cfg->freq_hz > (brd_cfg_priv[brd].rx_freq + brd_cfg_priv[brd].rx_bw / 2 - sx1301ar_bw_enum2nb(cfg->bandwidth) / 2)) {
				SX1301AR_ERR_RETURN(ERR_IF_LIMIT);
			}
			if (cfg->freq_hz < (brd_cfg_priv[brd].rx_freq - brd_cfg_priv[brd].rx_bw / 2 + sx1301ar_bw_enum2nb(cfg->bandwidth) / 2)) {
				SX1301AR_ERR_RETURN(ERR_IF_LIMIT);
			}

			/* Record frequency setting */
			brd_cfg_priv[brd].chan_freq[chip][cx] = cfg->freq_hz;

			/* Check and record channel-specific settings */
			if (cx < SX1301AR_CHIP_MULTI_NB) {
				/* LoRa multi-SF channel */
				if ((cfg->bandwidth != BW_UNDEFINED) && (cfg->bandwidth != BW_125K)) {
					SX1301AR_ERR_RETURN(ERR_INVALID_BW); /* only 125 kHz bandwidth is supported */
				}

				if (cfg->modrate == MR_UNDEFINED) {
					brd_cfg_priv[brd].multi_sf[chip][cx] = MULTI_DEFAULT_SF; /* force default setting */
				} else if ((cfg->modrate & ~MR_SF7_12) == 0) {
					brd_cfg_priv[brd].multi_sf[chip][cx] = (uint8_t)(cfg->modrate & MR_SF7_12);
				} else {
					SX1301AR_ERR_RETURN(ERR_INVALID_SF);
				}

				info("board #%zu: chip #%zu: Using channel#%zu: enable=true type=MULTI freq_hz=%u bw=125000 sf=%u-%u",
				     brd, chip, cx,
				     brd_cfg_priv[brd].chan_freq[chip][cx],
				     sx1301ar_sf_min_enum2nb(brd_cfg_priv[brd].multi_sf[chip][cx]),
				     sx1301ar_sf_max_enum2nb(brd_cfg_priv[brd].multi_sf[chip][cx]));

			} else if (cx == SX1301AR_CHIP_LSA_IDX) {
				/* LoRa stand-alone channel */
				if (cfg->bandwidth == BW_UNDEFINED) {
					brd_cfg_priv[brd].lsa_bw[chip] = LSA_DEFAULT_BW; /* force default setting */
				} else if ((cfg->bandwidth == BW_125K) || (cfg->bandwidth == BW_250K) || (cfg->bandwidth == BW_500K)) {
					brd_cfg_priv[brd].lsa_bw[chip] = cfg->bandwidth;
				} else {
					SX1301AR_ERR_RETURN(ERR_INVALID_BW);
				}

				if (cfg->modrate == MR_UNDEFINED) {
					brd_cfg_priv[brd].lsa_sf[chip] = LSA_DEFAULT_SF; /* force default setting */
				} else {
					sf = sx1301ar_sf_enum2nb(cfg->modrate);
					if (sf != -1) {
						brd_cfg_priv[brd].lsa_sf[chip] = sf;
					} else {
						SX1301AR_ERR_RETURN(ERR_INVALID_SF);
					}
				}

				info("board #%zu: chip #%zu: Using channel#%zu: enable=true type=STD   freq_hz=%u bw=%lu sf=%u",
				     brd, chip, cx,
				     brd_cfg_priv[brd].chan_freq[chip][cx],
				     sx1301ar_bw_enum2nb(brd_cfg_priv[brd].lsa_bw[chip]),
				     brd_cfg_priv[brd].lsa_sf[chip]);

			} else if (cx == SX1301AR_CHIP_FSK_IDX) {
				/* FSK channel */
				if (cfg->bandwidth == BW_UNDEFINED) {
					brd_cfg_priv[brd].fsk_bw[chip] = FSK_DEFAULT_BW; /* force default setting */
				} else if (cfg->bandwidth <= BW_7K8)			 /* bandwidth index in the enum is inversely proportional to bandwidth */
				{
					brd_cfg_priv[brd].fsk_bw[chip] = cfg->bandwidth;
				} else {
					SX1301AR_ERR_RETURN(ERR_INVALID_BW);
				}

				if (cfg->modrate == MR_UNDEFINED) {
					brd_cfg_priv[brd].fsk_br[chip] = FSK_DEFAULT_MR;
				} else if ((cfg->modrate >= MR_300) && (cfg->modrate <= MR_250K)) {
					brd_cfg_priv[brd].fsk_br[chip] = cfg->modrate;
				} else {
					SX1301AR_ERR_RETURN(ERR_INVALID_BR);
				}

				info("board #%zu: chip #%zu: Using channel#%zu: enable=true type=FSK   freq_hz=%u bw=%lu br=%u",
				     brd, chip, cx,
				     brd_cfg_priv[brd].chan_freq[chip][cx],
				     sx1301ar_bw_enum2nb(brd_cfg_priv[brd].fsk_bw[chip]),
				     brd_cfg_priv[brd].fsk_br[chip]);
			}
		} else {
			info("board #%zu: chip #%zu: Using channel#%zu: enable=false", brd, chip, cx);
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_lbt(uint8_t brd, const sx1301ar_lbt_cfg_t *cfg)
{
	int x;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		error("Invalid board number");
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cfg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* No reconfiguration if board is running */
	if (brd_cfg_priv[brd].is_started == true) {
		SX1301AR_ERR_RETURN(ERR_CANT_CFG);
	}

	x = sx1301ar_i_conf_lbt(brd, cfg);
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_conf_tx_gain(uint8_t brd, uint8_t rf_chain, const sx1301ar_tx_gain_lut_t *cfg)
{
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cfg == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((cfg->size < 1) || (cfg->size > SX1301AR_BOARD_MAX_LUT_NB)) {
		SX1301AR_ERR_RETURN(ERR_INVALID_LUT);
	}
	if (rf_chain >= SX1301AR_BOARD_RFCHAIN_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_RFCHAIN_NB);
	}

	/* Update internal TX gain LUT with given values */
	brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.size = cfg->size;

	for (i = 0; i < cfg->size; i++) {
		/* Check parameter range */
		if (cfg->lut[i].fpga_dig_gain > 13) {
			SX1301AR_ERR_RETURN(ERR_INVALID_LUT);
		}
		if (cfg->lut[i].ad9361_gain.auxdac_word > 1023) {
			SX1301AR_ERR_RETURN(ERR_INVALID_LUT);
		}

		/* Set internal LUT */
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].rf_power = cfg->lut[i].rf_power;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].fpga_dig_gain = cfg->lut[i].fpga_dig_gain;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.atten = cfg->lut[i].ad9361_gain.atten;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.auxdac_vref = cfg->lut[i].ad9361_gain.auxdac_vref;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.auxdac_word = cfg->lut[i].ad9361_gain.auxdac_word;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_tcomp.coeff_a = cfg->lut[i].ad9361_tcomp.coeff_a;
		brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_tcomp.coeff_b = cfg->lut[i].ad9361_tcomp.coeff_b;

		info("board #%zu: Using chain #%zu: lut=%-2zu atten=%-3u vref=%u word=%u a=%-3i b=%-5i dig=%u rf_power=%i",
		     brd,
		     rf_chain,
		     i,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.atten,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.auxdac_vref,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_gain.auxdac_word,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_tcomp.coeff_a,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].ad9361_tcomp.coeff_b,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].fpga_dig_gain,
		     brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[i].rf_power);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/* ~ Core functions ~~~~~~~~~~~~~~~ */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_start(uint8_t nb_brd)
{
	int brd_idx;
	uint8_t brd = 0;
	int x;
	int32_t val;
	uint8_t chip;
	uint8_t path_enable[SX1301AR_MAX_BOARD_NB];

	/* Check input parameters */
	if ((nb_brd > SX1301AR_MAX_BOARD_NB) || (nb_brd == 0)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	memset(path_enable, 0, sizeof path_enable);

	/* Configure all boards */
	for (brd = 0; brd < nb_brd; brd++) {
		x = configure_board(brd);
		SX1301AR_ERR_PROPAGATE(x);

		/* Record path to be enabled later */
		for (chip = 0; chip < brd_cfg_priv[brd].chip_nb; ++chip) {
			if (brd_cfg_priv[brd].chip_en[chip] == true) {
				path_enable[brd] |= (1 << chip);
			}
		}

		if (brd_cfg_priv[brd].btype == BRD_MASTER) {
			/* Enable clocks - KEEP THIS ORDER FOR SYNCHRONIZATION */
			sx1301ar_reg_brd_w(brd, BREG_SX1301_TIMESTAMP_CPY_EN, 1); /* Start FPGA 1 MHz clock along with SX1301 clocks - used for LBT */
			sx1301ar_reg_brd_w(brd, BREG_SX1301_CLK_32M_EN, 1);       /* Start all SX1301 clocks at the same time */
		}
	}

	/* Enable paths on all boards, MASTER last */
	for (brd_idx = (nb_brd - 1); brd_idx >= 0; brd_idx--) {
		brd = (uint8_t)brd_idx;

		sx1301ar_reg_brd_w(brd, BREG_ENABLE_SX1301_PATH, path_enable[brd]);

		/* Check that chip firmwares have been loaded properly */
		for (chip = 0; chip < brd_cfg_priv[brd].chip_nb; ++chip) {
			if (brd_cfg_priv[brd].chip_en[chip] == true) {
				/* Check AGC firmware version */
				sx1301ar_reg_chip_w(brd, chip, CREG_DBG_AGC_MCU_RAM_ADDR, FW_VERSION_ADDR);
				sx1301ar_reg_chip_r(brd, chip, CREG_DBG_AGC_MCU_RAM_DATA, &val);
				if (val != FW_VERSION_AGC) {
					error("[BRD%u][CHIP%u] AGC firmware version: %i (expecting %d)", brd, chip, val, FW_VERSION_AGC);
					SX1301AR_ERR_RETURN(ERR_AGC_VERSION);
				}

				/* Check ARB firmware version */
				sx1301ar_reg_chip_w(brd, chip, CREG_DBG_ARB_MCU_RAM_ADDR, FW_VERSION_ADDR);
				sx1301ar_reg_chip_r(brd, chip, CREG_DBG_ARB_MCU_RAM_DATA, &val);
				if (val != FW_VERSION_ARB) {
					SX1301AR_ERR_RETURN(ERR_ARB_VERSION);
				}
			}
		}

		/* The board has started now */
		brd_cfg_priv[brd].is_started = true;
		debug("--> board %u started", brd);
	}

	/* Initialize fine timestamps queue */
	gwloc_tmst_queue_init(&fine_timestamp_queue);

#if 0
    sx1301ar_reg_brd_r( 0, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_0] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_reg_brd_r( 1, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_1] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_wait_ms(1000);
    sx1301ar_reg_brd_r( 0, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_0] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_reg_brd_r( 1, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_1] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_wait_ms(1000);
    sx1301ar_reg_brd_r( 0, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_0] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_reg_brd_r( 1, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_1] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_wait_ms(1000);
    sx1301ar_reg_brd_r( 0, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_0] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_reg_brd_r( 1, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_1] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_wait_ms(1000);
    sx1301ar_reg_brd_r( 0, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_0] BREG_LATCHED_SAMPLE_CNT : %d", val );
    sx1301ar_reg_brd_r( 1, BREG_LATCHED_SAMPLE_CNT, &val);
    debug( "[BRD_1] BREG_LATCHED_SAMPLE_CNT : %d", val );
#endif

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_stop(uint8_t nb_brd)
{
	uint8_t brd = 0;
	int x;
	sx1301ar_tstat_t tx_status;
	int retry = 0;
	uint8_t dsp;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	for (brd = 0; brd < nb_brd; brd++) {
		if (brd_cfg_priv[brd].is_started == false) {
			/* Return immediately if board is already stopped */
			SX1301AR_ERR_RETURN(ERR_BOARD_STATE);
		}

		if ((brd_cfg_priv[brd].rf_chain[0].tx_enable != false) || (brd_cfg_priv[brd].rf_chain[1].tx_enable != false)) {
			/* Abort TX */
			x = sx1301ar_abort_tx(brd);
			SX1301AR_ERR_PROPAGATE(x);

			/* Check that TX stopped */
			do {
				sx1301ar_wait_us(100);
				x = sx1301ar_tx_status(brd, &tx_status);
				SX1301AR_ERR_PROPAGATE(x);
				retry++;
			} while ((tx_status != TX_FREE) && (retry < 10));

			if (tx_status != TX_FREE) {		      /* Failed to free TX */
				SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
			}
		}

		/* Stop Listen-Before-Talk */
		if (sx1301ar_i_lbt_is_enabled(brd) == true) {
			x = sx1301ar_i_stop_lbt(brd);
			if (x != 0) {
				SX1301AR_ERR_RETURN(ERR_LBT_START);
			}
		}

		/* Stop the DSPs */
		for (dsp = 0; dsp < brd_cfg_priv[brd].dsp_nb; dsp++) {
			x = gwloc_dsp_stop(brd, dsp, brd_cfg_priv[brd].dsp_nb);
			SX1301AR_ERR_PROPAGATE(x);
		}

		/* Reset registers, should reduce power consumption a lot */
		x = sx1301ar_reg_reset(brd);
		SX1301AR_ERR_PROPAGATE(x);

		brd_cfg_priv[brd].is_started = false;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_fetch(uint8_t brd, sx1301ar_rx_pkt_t *p, uint8_t max_nb, uint8_t *nb_pkt)
{
	int x;
	uint8_t temperature_code = SX1301AR_DEFAULT_AD9361_TEMP_REF;
	uint8_t brd_master;
	/* performances variables */
	struct timeval tm;

	/* Allocate memory for SX1301 packet fetching */
	pkt_rsig_t p_rsig[SX1301AR_MAX_PKT_NB];

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (p == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (nb_pkt == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (max_nb > SX1301AR_MAX_PKT_NB) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Return immediately if board is not running */
	if (brd_cfg_priv[brd].is_started == false) {
		SX1301AR_ERR_RETURN(ERR_BOARD_STATE);
	}

	/* Record fetch start time */
	_meas_time_start(&tm);

	/* Get current temperature from master board for RSSI correction */
	brd_master = (brd_cfg_priv[brd].btype == BRD_MASTER) ? brd : SX1301AR_BOARD_MASTER;
	x = sx1301ar_get_temperature(brd_master, &temperature_code, NULL);
	if (x != 0) {
		error("[BRD%u] Failed to read current temperature", brd);
	}

	/* Fetch packets from SX1301s */
	memset(p_rsig, 0, sizeof p_rsig);
	x = fetch_packets(brd, p, p_rsig, max_nb, nb_pkt, temperature_code);
	if (x != 0) {
		error("[BRD%u] Failed to fetch packets from SX1301s (error=%d)", brd, x);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Check history to remove ghost packet (packet received on SX1301 which has already been sent on previous fetch round) */
	x = history_check(brd, p, p_rsig, nb_pkt);
	if (x != 0) {
		error("Failed to check history to find ghost packet(s)");
	}

	/* Merge duplicated packets received on different antennas (diversity mode) */
	if (brd_cfg_priv[brd].chip_rf_chain[0] != brd_cfg_priv[brd].chip_rf_chain[1]) {
		x = merge_packets(brd, p, p_rsig, nb_pkt);
		if (x != 0) {
			error("Failed to merge duplicated packets (error=%d)", x);
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
		}
	}

	/* We wait a moment before fetching DSP, due to DSP latency for timestamp generation */
	/* This is needed for low SFs 7->9 */
	sx1301ar_wait_us(65000);

	/* Fetch FPGA to get encrypted fine timestamps */
	x = fetch_fpga(brd, &fine_timestamp_queue);
	if (x != 0) {
		error("Failed to fetch FPGA for timestamps (error=%d)", x);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Fetch DSPs to get clear fine timestamps, logs etc... */
	x = fetch_dsp(brd);
	if (x != 0) {
		error("Failed to fetch DSP for timestamps, log messages etc... (error=%d)", x);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* FPGA/DSP messages fetched, we can now try to match with received packet */
	x = merge_timestamps(brd, p, p_rsig, nb_pkt, &fine_timestamp_queue, temperature_code);
	if (x != 0) {
		error("Failed to fine timestamps (error=%d)", x);
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Save packet history */
	x = history_save(brd, p, *nb_pkt);
	if (x != 0) {
		error("failed to save history");
	}

	/* Compute fetch time */
	if (*nb_pkt > 0) {
		/* Record fetch end time */
		_meas_time_stop(1, tm, __FUNCTION__);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_send(uint8_t brd, const sx1301ar_tx_pkt_t *p)
{
	int x;
	uint8_t buff[256 + TX_METADATA_NB]; /* buffer to prepare the packet to send + metadata before SPI write burst */
	uint16_t fsk_dr_div;		    /* divider to configure for target datarate */
	int transfer_size = 0;		    /* data to transfer from host to TX databuffer */
	int payload_offset = 0;		    /* start of the payload content in the databuffer */
	uint8_t pow_index;		    /* 4-bit value to set the firmware TX power */
	uint32_t count_trig;		    /* timestamp value in trigger mode corrected for TX start delay */
	uint16_t preamble_size;
	int sf, cr;
	uint64_t n_int, n_frac;
	int16_t atten, auxdac;
	uint8_t lsb, msb;
	int8_t tcomp_atten, tcomp_auxdac;
	sx1301ar_tx_gain_t lut;
	bool tx_allowed;
	uint16_t tx_start_delay;
	bool tx_notch_enable = false;
	static uint8_t agc_tread_disable = 0;
	/* performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (p == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((brd_cfg_priv[brd].rf_chain[0].tx_enable == false) && (brd_cfg_priv[brd].rf_chain[1].tx_enable == false)) {
		SX1301AR_ERR_RETURN(ERR_NO_TX_ENABLE);
	}
	if (brd_cfg_priv[brd].rf_chain[p->rf_chain].tx_enable == false) {
		SX1301AR_ERR_RETURN(ERR_NO_TX_ENABLE);
	}

	/* Return immediately if board is not running */
	if (brd_cfg_priv[brd].is_started == false) {
		SX1301AR_ERR_RETURN(ERR_BOARD_STATE);
	}

	/* Record send start time */
	_meas_time_start(&tm);

	/* Enable notch filter for LoRa 125kHz */
	if ((p->modulation == MOD_LORA) && (p->bandwidth == BW_125K)) {
		tx_notch_enable = true;
	}

	/* Get the TX start delay to be applied for this TX */
	tx_start_delay = get_tx_start_delay(tx_notch_enable, p->bandwidth);

	/* Check if this is not too late for this packet to be programmed for TX */
	if (p->tx_mode == TX_TIMESTAMPED) {
		sx1301ar_get_instcnt(brd, &count_trig);
		if (((p->count_us - count_trig) < (uint32_t)tx_start_delay) || ((p->count_us - count_trig) > 1E6)) /* unsigned arithmetic */
		{
			warning("[%llu] it is already too late to send this packet (%dµs)", sx1301ar_timestamp(), (int32_t)(p->count_us - tx_start_delay) - (int32_t)count_trig);
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
		}
	}

	/* Fixed metadata, useful payload and misc metadata compositing */
	transfer_size = TX_METADATA_NB + p->size;
	payload_offset = TX_METADATA_NB; /* start the payload just after the metadata */

	/* Get TX gain LUT index corresponding to requested TX power */
	x = get_tx_gain_lut_index(brd, p->rf_chain, p->rf_power, &pow_index);
	SX1301AR_ERR_PROPAGATE(x);

	/* Get temperature compensation gains */
	x = get_tx_gain_temperature_compensation(brd, p->rf_chain, pow_index, &tcomp_atten, &tcomp_auxdac);
	SX1301AR_ERR_PROPAGATE(x);

	/* Set Tx gain registers accessed from SX1301 AGC */
	lut = brd_cfg_priv[brd].rf_chain[p->rf_chain].tx_lut.lut[pow_index];
	atten = lut.ad9361_gain.atten + tcomp_atten;
	if (atten < 0) {
		warning("TX gain \'atten\' is out of range (%d), clipping to %d", atten, 0);
		atten = 0;
	}
	if (atten > AD9361_TXGAIN_ATTEN_MAX) {
		warning("TX gain \'atten\' is out of range (%d), clipping to %d", atten, AD9361_TXGAIN_ATTEN_MAX);
		atten = AD9361_TXGAIN_ATTEN_MAX;
	}

	auxdac = lut.ad9361_gain.auxdac_word + tcomp_auxdac;
	if (auxdac < 0) {
		warning("TX gain \'auxdac\' is out of range (%d), clipping to %d", auxdac, 0);
		auxdac = 0;
	}
	if (auxdac > AD9361_TXGAIN_AUXDAC_MAX) {
		warning("TX gain \'auxdac\' is out of range (%d), clipping to %d", auxdac, AD9361_TXGAIN_AUXDAC_MAX);
		auxdac = AD9361_TXGAIN_AUXDAC_MAX;
	}

	/* Set AUXDAC (10-bits) */
	msb = (uint8_t)(auxdac >> 2);
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_AUXDAC_8MSB, (int32_t)msb);
	lsb = (uint8_t)(auxdac & 3);
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_AUXDAC_2LSB, (int32_t)lsb);

	/* Set AUXDAC VREF */
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_AUXDAC_VREF, (int32_t)lut.ad9361_gain.auxdac_vref);

	/* Set ATTEN (9-bits) */
	msb = (uint8_t)((atten >> 8) & 1);
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_AD_ATTEN_MSB, (int32_t)msb);
	lsb = (uint8_t)(atten & 0xFF);
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_AD_ATTEN_8LSB, (int32_t)lsb);

	/* Set FPGA gain (not controlled by AGC anymore) */
	sx1301ar_reg_brd_w(brd, BREG_TX_GAIN, (int32_t)lut.fpga_dig_gain);

	/* Enable notch filter if required */
	sx1301ar_reg_brd_w(brd, BREG_TX_NOTCH_EN, (tx_notch_enable == true) ? 1 : 0);

	/* Select RF path (not controlled by AGC anymore) */
	sx1301ar_reg_brd_w(brd, BREG_TX_RF_PATH_SELECT, (int32_t)p->rf_chain);

	/* Metadata 0 to 2, TX PLL fractional frequency */
	n_int = ((uint64_t)(p->freq_hz) * AD9361_PLL_DIV) / (AD9361_REF_FREQ_MHz * 1e6);
	n_frac = (AD9361_PLL_MODULUS * (uint64_t)p->freq_hz * AD9361_PLL_DIV) / (AD9361_REF_FREQ_MHz * 1e6) - AD9361_PLL_MODULUS * n_int;

	buff[0] = (uint8_t)((n_frac & 0xFF0000) >> 16);
	buff[1] = (uint8_t)((n_frac & 0x00FF00) >> 8);
	buff[2] = (uint8_t)(n_frac & 0x0000FF);

	/* Interger word 0 */
	sx1301ar_reg_brd_w(brd, BREG_TX_LUT_FREQ_INTEGER_0, (int32_t)(n_int & 0xFF));

	/* Metadata 3 to 6, timestamp trigger value */
	if (p->tx_mode == TX_TIMESTAMPED) {
		count_trig = p->count_us - tx_start_delay;
		buff[3] = 0xFF & (count_trig >> 24);
		buff[4] = 0xFF & (count_trig >> 16);
		buff[5] = 0xFF & (count_trig >> 8);
		buff[6] = 0xFF & count_trig;
	}

	/* Indicate the the AGC firmware to disable temperature reading until this TX is finished */
	agc_tread_disable += 1;

	/* Parameters depending on modulation  */
	if (p->modulation == MOD_LORA) {
		/* Metadata 7, modulation type, radio chain selection and TX power */
		buff[7] = (p->rf_chain << 5) | (0x0F & agc_tread_disable); /* b4=0 -> LoRa modulation, b5=0 -> radio 0 */
		buff[8] = 0;						   /* metadata 8, not used */

		/* Metadata 9, CRC, LoRa CR & SF */
		cr = sx1301ar_cr_enum2nb(p->coderate);
		if (cr == -1) {
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
		sf = sx1301ar_sf_enum2nb(p->modrate);
		if (sf == -1) {
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
		buff[9] = (p->no_crc) ? 0 : 0x80; /* set 'CRC enable' bit */
		buff[9] |= 0x70 & (cr << 4);
		buff[9] |= 0x0F & sf;

		/* Metadata 10, payload size */
		buff[10] = p->size;

		/* Metadata 11, implicit header, modulation bandwidth, PPM offset & polarity */
		switch (p->bandwidth) {
		case BW_125K:
			buff[11] = 0x03 & 0;
			break;
		case BW_250K:
			buff[11] = 0x03 & 1;
			break;
		case BW_500K:
			buff[11] = 0x03 & 2;
			break;
		default:
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
		buff[11] |= (p->no_header) ? 0x04 : 0;		      /* set 'implicit header' bit */
		buff[11] |= ppm_mode_en(p->bandwidth, sf) ? 0x08 : 0; /* set 'PPM offset' bit at 1 */
		buff[11] |= (p->invert_pol) ? 0x10 : 0;		      /* set 'TX polarity' bit at 1 */

		/* Metadata 12 & 13, LoRa preamble size */
		if (p->preamble == 0) {
			/* If not explicit, use recommended LoRa preamble size */
			preamble_size = STD_LORA_PREAMBLE;
		} else if (p->preamble < MIN_LORA_PREAMBLE) {
			/* Enforce minimum preamble size */
			preamble_size = MIN_LORA_PREAMBLE;
		} else {
			preamble_size = p->preamble;
		}
		buff[12] = 0xFF & (preamble_size >> 8);
		buff[13] = 0xFF & preamble_size;

		/* Metadata 14 & 15, not used */
		buff[14] = 0;
		buff[15] = 0;
	} else if (p->modulation == MOD_FSK) {
		/* Metadata 7, modulation type, radio chain selection and TX power */
		buff[7] = (p->rf_chain << 5) | 0x10 | (0x0F & agc_tread_disable); /* b4=1 -> FSK modulation, b5=0 -> radio 0 */

		buff[8] = 0; /* metadata 8, not used */

		/* Metadata 9, frequency deviation */
		buff[9] = p->f_dev;

		/* Metadata 10, payload size */
		buff[10] = p->size + 1; /* add a byte to encode payload length in the packet */
		// TODO: handle fixed packet length ?
		// TODO: handle 255 bytes packets ?

		/* Metadata 11, packet mode, CRC, encoding */
		buff[11] = 0x01 | ((p->no_crc) ? 0 : 0x02) | (0x02 << 2); /* always in variable length packet mode, no DC-free encoding, CCITT CRC if CRC is not disabled  */

		/* Metadata 12 & 13, FSK preamble size */
		if (p->preamble == 0) {
			/* If not explicit, use LoRa MAC FSK preamble size */
			preamble_size = STD_FSK_PREAMBLE;
		} else if (p->preamble < MIN_FSK_PREAMBLE) {
			/* Enforce minimum preamble size */
			preamble_size = MIN_FSK_PREAMBLE;
		} else {
			preamble_size = p->preamble;
		}
		buff[12] = 0xFF & (preamble_size >> 8);
		buff[13] = 0xFF & preamble_size;

		/* Metadata 14 & 15, FSK baudrate */
		fsk_dr_div = (uint16_t)((uint32_t)CHIP_XTAL_FREQU / p->modrate); /* Ok for datarate between 500bps and 250kbps */
		buff[14] = 0xFF & (fsk_dr_div >> 8);
		buff[15] = 0xFF & fsk_dr_div;

		/* Insert payload size in the packet for variable mode */
		buff[16] = p->size;
		++transfer_size;  /* one more byte to transfer to the TX modem */
		++payload_offset; /* start the payload with one more byte of offset */
	} else {
		/* Unknown modulation */
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Copy payload from user struct to buffer containing metadata */
	memcpy((void *)(buff + payload_offset), (void *)(p->payload), p->size);

	/* Configure TX start delay based on TX bandwidth */
	x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_START_DELAY, tx_start_delay);
	SX1301AR_ERR_PROPAGATE(x);

	/* Reset TX command flags */
	x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_TRIG_ALL, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Put metadata + payload in the TX data buffer */
	x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_DATA_BUF_ADDR, 0);
	SX1301AR_ERR_PROPAGATE(x);
	x = sx1301ar_reg_chip_wb(brd, 0, CREG_TX_DATA_BUF_DATA, buff, transfer_size);
	SX1301AR_ERR_PROPAGATE(x);

	/* Listen-Before-Talk: check if TX is allowed */
	x = sx1301ar_i_lbt_is_tx_allowed(brd, p, tx_start_delay, &tx_allowed);
	SX1301AR_ERR_PROPAGATE(x);

	/* Tx trigger */
	if (tx_allowed == true) {
		switch (p->tx_mode) {
		case TX_IMMEDIATE:
			x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_TRIG_IMMEDIATE, 1);
			SX1301AR_ERR_PROPAGATE(x);
			break;
		case TX_TIMESTAMPED:
			x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_TRIG_DELAYED, 1);
			SX1301AR_ERR_PROPAGATE(x);
			break;
		case TX_ON_GPS:
			x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_TRIG_GPS, 1);
			SX1301AR_ERR_PROPAGATE(x);
			break;
		default:
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
	} else {
		SX1301AR_ERR_RETURN(ERR_LBT_FORBIDDEN);
	}

	/* Record send end time */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_tx_status(uint8_t brd, sx1301ar_tstat_t *s)
{
	int x;
	int32_t read_value;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (s == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((brd_cfg_priv[brd].rf_chain[0].tx_enable == false) && (brd_cfg_priv[brd].rf_chain[1].tx_enable == false)) {
		SX1301AR_ERR_RETURN(ERR_NO_TX_ENABLE);
	}

	/* Return immediately if board is not running */
	if (brd_cfg_priv[brd].is_started == false) {
		*s = TX_OFF;
		return 0;
	}

	x = sx1301ar_reg_chip_r(brd, 0, CREG_TX_STATUS, &read_value);
	SX1301AR_ERR_PROPAGATE(x);

	if ((read_value & 0x10) == 0) /* bit 4 @1: TX programmed */
	{
		*s = TX_FREE;
	} else if ((read_value & 0x60) != 0) /* bit 5 or 6 @1: TX sequence */
	{
		*s = TX_EMITTING;
	} else {
		*s = TX_SCHEDULED;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_abort_tx(uint8_t brd)
{
	int x;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Reset TX_TRIG_IMMEDIATE, TX_TRIG_DELAYED & TX_TRIG_GPS */
	x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_TRIG_ALL, 0);
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_brd_status(uint8_t brd, sx1301ar_bstat_t *s)
{
	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (s == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	if (brd_cfg_priv[brd].is_started == true) {
		*s = BRD_STARTED;
	} else {
		*s = BRD_STOPPED;
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_instcnt(uint8_t brd, uint32_t *cnt_us)
{
	int x;
	int retry = 0;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cnt_us == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	/* sx1301ar_reg_chip_r will catch an unconfigured board */

	/* Counter of chip SX1301AR_FREE_RUN_TIMER_CHIP is free-running */
	/* A retry loop is needed because SPI read sometimes fails on that specific operation */
	do {
		x = sx1301ar_reg_chip_r(brd, SX1301AR_FREE_RUN_TIMER_CHIP, CREG_TIMESTAMP, (int32_t *)cnt_us);
		retry++;
	} while ((x != 0) && (retry < 3));
	// if( retry > 1 ) // DEBUG
	// {
	//     warning("sx1301ar_get_instcnt took %i retries", retry );
	// }
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_trigcnt(uint8_t brd, uint32_t *cnt_us)
{
	int x;
	int retry = 0;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (cnt_us == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	/* sx1301ar_reg_chip_r will catch an unconfigured board */

	/* Counter of chip SX1301AR_GPS_TIMER_CHIP is latched on external trigger (eg. PPS signal from GPS) */
	/* A retry loop is needed because SPI read sometimes fails on that specific operation */
	do {
		x = sx1301ar_reg_chip_r(brd, SX1301AR_GPS_TIMER_CHIP, CREG_TIMESTAMP, (int32_t *)cnt_us);
		retry++;
	} while ((x != 0) && (retry < 3));
	// if( retry > 1 ) // DEBUG
	// {
	//     warning("sx1301ar_get_trigcnt took %i retries", retry );
	// }
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_get_temperature(uint8_t brd, uint8_t *temp_code, int16_t *temp_celsius)
{
	int x;
	int32_t val;
	float temp_f;
	int16_t tcode;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Request SX1301 AGC firmware to read AD9361 temperature sensor */
	x = sx1301ar_reg_chip_w(brd, 0, CREG_DBG_AGC_MCU_RAM_ADDR, 0x21);
	SX1301AR_ERR_PROPAGATE(x);

	/* Get temperature value (unsigned 8bits value) [-40..127°C] */
	x = sx1301ar_reg_chip_r(brd, 0, CREG_DBG_AGC_MCU_RAM_DATA, &val);
	SX1301AR_ERR_PROPAGATE(x);

	tcode = (int16_t)val;

	/* Return raw temperature code read from AD9361 radio */
	if (temp_code != NULL) {
		*temp_code = (uint8_t)tcode;
	}

	/* Return temperature in degree celsius, converted from temperature code */
	/* Temp(°C) = pTemp * (T - Tref_ad9361) + Tref_room */
	if (temp_celsius != NULL) {
		temp_f = (SX1301AR_DEFAULT_AD9361_TEMP_SLOPE * (tcode - brd_cfg_priv[brd].ad9361_temp_ref)) / 1024 + brd_cfg_priv[brd].room_temp_ref; /* Slope is given multiplied by 1024, so we rescale */
		*temp_celsius = (int16_t)round(temp_f);

		/*
         * FPGA chip junction temperature is 105°C:
         * Assume mesured temperature will not be higher than 110°C ... but can be lower than -40°C
         * And so re-scale temp range (-41°C code == 127°C code)
         */
		if (*temp_celsius > 110) {
			/* Invert tcode and re-compute */
			tcode = tcode - 256;
			temp_f = (SX1301AR_DEFAULT_AD9361_TEMP_SLOPE * (tcode - brd_cfg_priv[brd].ad9361_temp_ref)) / 1024 + brd_cfg_priv[brd].room_temp_ref;
			*temp_celsius = (int16_t)round(temp_f);
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

double sx1301ar_rx_pkt_time_on_air(const sx1301ar_rx_pkt_t *pkt, double *t_preamble, double *t_syncword, double *t_payload)
{
	/* Check input parameters */
	if (pkt == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	switch (pkt->modulation) {
	case MOD_LORA:
		return sx1301ar_i_lora_time_on_air(pkt->bandwidth, pkt->modrate, pkt->size, pkt->coderate, false, STD_LORA_PREAMBLE, t_preamble, t_syncword, t_payload);
	default:
		error("modulation not supported for RX time on air calculation (0x%02X)", pkt->modulation);
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

double sx1301ar_tx_pkt_time_on_air(uint8_t brd, const sx1301ar_tx_pkt_t *pkt, double *t_preamble, double *t_syncword, double *t_payload)
{
	/* Check input parameters */
	if (pkt == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	switch (pkt->modulation) {
	case MOD_LORA:
		return sx1301ar_i_lora_time_on_air(pkt->bandwidth, pkt->modrate, pkt->size, pkt->coderate, pkt->no_header, pkt->preamble, t_preamble, t_syncword, t_payload);
	case MOD_FSK:
		return sx1301ar_i_fsk_time_on_air(brd, pkt->modrate, pkt->preamble, pkt->size, pkt->no_crc);
	default:
		error("modulation not supported for TX time on air calculation (0x%02X)", pkt->modulation);
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
/* ~ Helper functions ~~~~~~~~~~~~~ */
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_board_cfg_t sx1301ar_init_board_cfg(void)
{
	sx1301ar_board_cfg_t a;
	int i;

	a.board_type = BRD_TYPE_UNKNOWN;
	a.spi_read = NULL;
	a.spi_write = NULL;
	a.rx_freq_hz = 0;
	a.rx_bw_hz = 0;
	a.full_duplex = false;

	/* RF chain initialization */
	for (i = 0; i < SX1301AR_BOARD_RFCHAIN_NB; i++) {
		a.rf_chain[i].rx_enable = false;
		a.rf_chain[i].tx_enable = false;
		a.rf_chain[i].antenna_gain = 0;
		a.rf_chain[i].insertion_loss = 0;
		a.rf_chain[i].rssi_offset = SX1301AR_DEFAULT_RSSI_OFFSET;
		a.rf_chain[i].rssi_offset_coeff_a = 0;
		a.rf_chain[i].rssi_offset_coeff_b = 0;
		a.rf_chain[i].tx_lut = sx1301ar_init_tx_gain_lut();
	}

	/* No board-specific corrections */
	a.room_temp_ref = SX1301AR_DEFAULT_ROOM_TEMP_REF;
	a.ad9361_temp_ref = SX1301AR_DEFAULT_AD9361_TEMP_REF;

	/* Phy parameters that are protocol dependant */
	a.fsk_sync_word = 0;
	a.fsk_sync_size = 0;
	a.loramac_public = false;

	/* Localization parameters */
	a.dsp_stat_interval = 0; /* Turn off dsp report by default */
	a.fine_timestamp_decrypt_enable = false;
	memset(a.aes_key, 0, sizeof a.aes_key);
	a.match_tmst_crc_err = false;
	a.main_tmst_version = SX1301AR_DEFAULT_FTS_VERSION;
	a.debug_tmst = false;

	a.nb_chip = 0;
	a.nb_dsp = 0;

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_chip_cfg_t sx1301ar_init_chip_cfg(void)
{
	sx1301ar_chip_cfg_t a;

	a.enable = false;
	a.freq_hz = 0;
	a.rf_chain = 0;

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_chan_cfg_t sx1301ar_init_chan_cfg(void)
{
	sx1301ar_chan_cfg_t a;

	a.enable = false;
	a.freq_hz = 0;
	a.modrate = MR_UNDEFINED;
	a.bandwidth = BW_UNDEFINED;

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_lbt_cfg_t sx1301ar_init_lbt_cfg(void)
{
	sx1301ar_lbt_cfg_t a;

	a.enable = false;
	a.rssi_target = 0;
	a.rssi_shift = 0;
	a.nb_channel = 0;
	memset(a.channels, 0, sizeof a.channels);

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_tx_pkt_t sx1301ar_init_tx_pkt(void)
{
	sx1301ar_tx_pkt_t a;

	a.tx_mode = TX_IMMEDIATE;
	a.count_us = 0;
	a.freq_hz = 0;
	a.rf_power = 0;
	a.modulation = MOD_UNDEFINED;
	a.bandwidth = BW_UNDEFINED;
	a.modrate = MR_UNDEFINED;
	a.coderate = CR_UNDEFINED;
	a.f_dev = 0;
	a.preamble = 0;
	a.invert_pol = false;
	a.no_crc = false;
	a.no_header = false;
	a.size = 0;
	memset((void *)(a.payload), 0, ARRAY_SIZE(a.payload));

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_tx_gain_t sx1301ar_init_tx_gain(void)
{
	sx1301ar_tx_gain_t a;

	a.rf_power = 0;
	a.fpga_dig_gain = 0;
	a.ad9361_gain.atten = 0;
	a.ad9361_gain.auxdac_vref = 0;
	a.ad9361_gain.auxdac_word = 0;
	a.ad9361_tcomp.coeff_a = 0;
	a.ad9361_tcomp.coeff_b = 0;

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_tx_gain_lut_t sx1301ar_init_tx_gain_lut(void)
{
	sx1301ar_tx_gain_lut_t a;

	a.size = SX1301AR_BOARD_MAX_LUT_NB;
	memset(a.lut, 0, sizeof(sx1301ar_tx_gain_lut_t));

	return a;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_modr_t sx1301ar_sf_nb2enum(int x)
{
	switch (x) {
	case 7:
		return MR_SF7;
	case 8:
		return MR_SF8;
	case 9:
		return MR_SF9;
	case 10:
		return MR_SF10;
	case 11:
		return MR_SF11;
	case 12:
		return MR_SF12;
	default:
		return MR_UNDEFINED;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_sf_enum2nb(sx1301ar_modr_t x)
{
	switch (x) {
	case MR_SF7:
		return 7;
	case MR_SF8:
		return 8;
	case MR_SF9:
		return 9;
	case MR_SF10:
		return 10;
	case MR_SF11:
		return 11;
	case MR_SF12:
		return 12;
	default:
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_modr_t sx1301ar_sf_range_nb2enum(int a, int b)
{
	switch (a) {
	case 7:
		switch (b) {
		case 7:
			return MR_SF7;
		case 8:
			return MR_SF7_8;
		case 9:
			return MR_SF7_9;
		case 10:
			return MR_SF7_10;
		case 11:
			return MR_SF7_11;
		case 12:
			return MR_SF7_12;
		default:
			return MR_UNDEFINED;
		}
	case 8:
		switch (b) {
		case 7:
			return MR_SF7_8;
		case 8:
			return MR_SF8;
		case 9:
			return MR_SF8_9;
		case 10:
			return MR_SF8_10;
		case 11:
			return MR_SF8_11;
		case 12:
			return MR_SF8_12;
		default:
			return MR_UNDEFINED;
		}
	case 9:
		switch (b) {
		case 7:
			return MR_SF7_9;
		case 8:
			return MR_SF8_9;
		case 9:
			return MR_SF9;
		case 10:
			return MR_SF9_10;
		case 11:
			return MR_SF9_11;
		case 12:
			return MR_SF9_12;
		default:
			return MR_UNDEFINED;
		}
	case 10:
		switch (b) {
		case 7:
			return MR_SF7_10;
		case 8:
			return MR_SF8_10;
		case 9:
			return MR_SF9_10;
		case 10:
			return MR_SF10;
		case 11:
			return MR_SF10_11;
		case 12:
			return MR_SF10_12;
		default:
			return MR_UNDEFINED;
		}
	case 11:
		switch (b) {
		case 7:
			return MR_SF7_11;
		case 8:
			return MR_SF8_11;
		case 9:
			return MR_SF9_11;
		case 10:
			return MR_SF10_11;
		case 11:
			return MR_SF11;
		case 12:
			return MR_SF11_12;
		default:
			return MR_UNDEFINED;
		}
	case 12:
		switch (b) {
		case 7:
			return MR_SF7_12;
		case 8:
			return MR_SF8_12;
		case 9:
			return MR_SF9_12;
		case 10:
			return MR_SF10_12;
		case 11:
			return MR_SF11_12;
		case 12:
			return MR_SF12;
		default:
			return MR_UNDEFINED;
		}
	default:
		return MR_UNDEFINED;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_sf_min_enum2nb(sx1301ar_modr_t x)
{
	switch (x) {
	case MR_SF7:
		return 7;
	case MR_SF7_8:
		return 7;
	case MR_SF7_9:
		return 7;
	case MR_SF7_10:
		return 7;
	case MR_SF7_11:
		return 7;
	case MR_SF7_12:
		return 7;
	case MR_SF8:
		return 8;
	case MR_SF8_9:
		return 8;
	case MR_SF8_10:
		return 8;
	case MR_SF8_11:
		return 8;
	case MR_SF8_12:
		return 8;
	case MR_SF9:
		return 9;
	case MR_SF9_10:
		return 9;
	case MR_SF9_11:
		return 9;
	case MR_SF9_12:
		return 9;
	case MR_SF10:
		return 10;
	case MR_SF10_11:
		return 10;
	case MR_SF10_12:
		return 10;
	case MR_SF11:
		return 11;
	case MR_SF11_12:
		return 11;
	case MR_SF12:
		return 12;
	default:
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_sf_max_enum2nb(sx1301ar_modr_t x)
{
	switch (x) {
	case MR_SF7:
		return 7;
	case MR_SF7_8:
		return 8;
	case MR_SF7_9:
		return 9;
	case MR_SF7_10:
		return 10;
	case MR_SF7_11:
		return 11;
	case MR_SF7_12:
		return 12;
	case MR_SF8:
		return 8;
	case MR_SF8_9:
		return 9;
	case MR_SF8_10:
		return 10;
	case MR_SF8_11:
		return 11;
	case MR_SF8_12:
		return 12;
	case MR_SF9:
		return 9;
	case MR_SF9_10:
		return 10;
	case MR_SF9_11:
		return 11;
	case MR_SF9_12:
		return 12;
	case MR_SF10:
		return 10;
	case MR_SF10_11:
		return 11;
	case MR_SF10_12:
		return 12;
	case MR_SF11:
		return 11;
	case MR_SF11_12:
		return 12;
	case MR_SF12:
		return 12;
	default:
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_coder_t sx1301ar_cr_nb2enum(int x)
{
	switch (x) {
	case 1:
		return CR_4_5;
	case 2:
		return CR_4_6;
	case 3:
		return CR_4_7;
	case 4:
		return CR_4_8;
	default:
		return CR_UNDEFINED;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_cr_enum2nb(sx1301ar_coder_t x)
{
	switch (x) {
	case CR_4_5:
		return 1;
	case CR_4_6:
		return 2;
	case CR_4_7:
		return 3;
	case CR_4_8:
		return 4;
	default:
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

const char *sx1301ar_cr_enum2str(sx1301ar_coder_t x)
{
	switch (x) {
	case CR_4_5:
		return "4/5";
	case CR_4_6:
		return "4/6";
	case CR_4_7:
		return "4/7";
	case CR_4_8:
		return "4/8";
	default:
		return "???";
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

sx1301ar_bandw_t sx1301ar_bw_nb2enum(long x)
{
	if (x <= 0)
		return BW_UNDEFINED;
	else if (x <= 7800)
		return BW_7K8;
	else if (x <= 15600)
		return BW_15K6;
	else if (x <= 31200)
		return BW_31K2;
	else if (x <= 62500)
		return BW_62K5;
	else if (x <= 125000)
		return BW_125K;
	else if (x <= 250000)
		return BW_250K;
	else if (x <= 500000)
		return BW_500K;
	else
		return BW_UNDEFINED;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

long sx1301ar_bw_enum2nb(sx1301ar_bandw_t x)
{
	switch (x) {
	case BW_500K:
		return 500000;
	case BW_250K:
		return 250000;
	case BW_125K:
		return 125000;
	case BW_62K5:
		return 62500;
	case BW_31K2:
		return 31200;
	case BW_15K6:
		return 15600;
	case BW_7K8:
		return 7800;
	default:
		return -1;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int8_t sx1301ar_ad9361_temp2atten(uint8_t T, uint8_t Tref, int16_t a, int16_t b)
{
	int32_t atten_i32;
	float atten_f;

	/* Calculate attenuation correction to be applied for given temperature */
	/* atten = a.(T²-Tref²) + b.(T-Tref) */
	atten_i32 = a * (pow(T, 2) - pow(Tref, 2)) + b * (T - Tref); /* dB step */

	/* Remove the factor 2^16 added by coeff_a and coeff_b */
	atten_f = (float)atten_i32 / pow(2, 16);

	/* Scale to 0.25dB step, round to 8bits integer */
	return ((int8_t)roundf(atten_f * 4));
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

float sx1301ar_ad9361_temp2rssi(uint8_t T, uint8_t Tref, int16_t a, int16_t b)
{
	int32_t rssi_i32;

	/* Calculate rssi correction to be applied for given temperature */
	/* rssi = a.(T-Tref)² + b.(T-Tref) */
	rssi_i32 = a * (pow(T - Tref, 2)) + b * (T - Tref);

	/* Remove the factor 2^16 added by coeff_a and coeff_b */
	return ((float)rssi_i32 / pow(2, 16));
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

static int load_firmware(uint8_t brd, uint8_t chip, fw_proc_t target, const uint8_t *fw, uint32_t size)
{
	sx1301ar_creg_t index_creg_reset;
	sx1301ar_creg_t index_creg_select;
	uint8_t fw_check[8192];
	int32_t dummy;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if (fw == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Check target controller */
	if (target == MCU_ARB) {
		if (size > MCU_ARB_FW_BYTE) {
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
		index_creg_reset = CREG_MCU_RST_0;
		index_creg_select = CREG_MCU_SELECT_MUX_0;
	} else if (target == MCU_AGC) {
		if (size > MCU_AGC_FW_BYTE) {
			SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
		}
		index_creg_reset = CREG_MCU_RST_1;
		index_creg_select = CREG_MCU_SELECT_MUX_1;
	} else {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Reset the target controller */
	sx1301ar_reg_chip_w(brd, chip, index_creg_reset, 1);

	/* Set mux to access controller program RAM and set address to 0 */
	sx1301ar_reg_chip_w(brd, chip, index_creg_select, 0);
	sx1301ar_reg_chip_w(brd, chip, CREG_MCU_PROM_ADDR, 0);

	/* Write the controller firmware in one burst */
	sx1301ar_reg_chip_wb(brd, chip, CREG_MCU_PROM_DATA, fw, size);

	/* Read back firmware code for check */
	sx1301ar_reg_chip_r(brd, chip, CREG_MCU_PROM_DATA, &dummy); /* bug workaround */
	sx1301ar_reg_chip_rb(brd, chip, CREG_MCU_PROM_DATA, fw_check, size);
	if (memcmp(fw, fw_check, size) != 0) {
		error("Failed to load fw %d on chip %d", (int)target, chip);
	}

	/* Give back control of the program ram to the controller */
	/* /!\ controller still in reset, reset must be de-asserted separately */
	sx1301ar_reg_chip_w(brd, chip, index_creg_select, 1);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static float calc_rssi(uint8_t brd, uint8_t cx, uint8_t raw_rssi, uint8_t rx_chain, uint8_t temperature_code)
{
	float a = (float)raw_rssi;
	uint8_t brd_master;

	/* Get RSSI offset and temperature compensation from master board */
	brd_master = (brd_cfg_priv[brd].btype == BRD_MASTER) ? brd : SX1301AR_BOARD_MASTER;

	/* RSSI offset for single SF LoRa */
	a += brd_cfg_priv[brd_master].rf_chain[rx_chain].rssi_offset;

	/* Channel specific correction (LoRa stand-alone is reference) */
	if (cx < SX1301AR_CHIP_MULTI_NB) { /* LoRa multi-sf channels */
		a += RSSI_MULTI_BIAS;
	} else if (cx == SX1301AR_CHIP_FSK_IDX) { /* FSK channel */
		a = RSSI_FSK_POLY_0 + RSSI_FSK_POLY_1 * a + RSSI_FSK_POLY_2 * pow(a, 2);
	}

	/* Correction of FPGA bandpass filters */
	// TODO: correction of chip RX bandwidth profile f(IF_channel)

	/* Correction of radio IF filter */
	// TODO: correction of radio RX bandwidth profile f(IF_chip)

	/* Correction for radio temperature */
	a += get_rx_rssi_temperature_compensation(brd_master, rx_chain, temperature_code);

	return roundf(a); /* 1dBm precision */
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static bool ppm_mode_en(sx1301ar_bandw_t bw, uint8_t sf)
{
	if (bw == BW_125K) {
		if ((sf == 12) || (sf == 11)) {
			return true;
		}
	} else if (bw == BW_250K) {
		if (sf == 12) {
			return true;
		}
	}

	return false;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int get_tx_gain_lut_index(uint8_t brd, uint8_t rf_chain, int8_t rf_power, uint8_t *lut_index)
{
	uint8_t pow_index;
	int current_best_index = -1;
	uint8_t current_best_match = 0xFF;
	int diff;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (rf_chain >= SX1301AR_BOARD_RFCHAIN_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_RFCHAIN_NB);
	}
	if (lut_index == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Search requested power in TX gain LUT */
	for (pow_index = 0; pow_index < brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.size; pow_index++) {
		diff = rf_power - brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[pow_index].rf_power;
		if (diff < 0) {
			/* The selected power must be lower or equal to requested one */
			continue;
		} else {
			/* Record the index corresponding to the closest rf_power available in LUT */
			if ((current_best_index == -1) || (diff < current_best_match)) {
				current_best_match = diff;
				current_best_index = pow_index;
			}
		}
	}

	/* Return corresponding index */
	if (current_best_index > -1) {
		*lut_index = (uint8_t)current_best_index;
		debug("Board #%u: Using power: %i dB", brd, brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[current_best_index].rf_power);

	} else {
		*lut_index = 0;
		SX1301AR_ERR_RETURN(ERR_INVALID_LUT);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int get_tx_gain_temperature_compensation(uint8_t brd, uint8_t rf_chain, uint8_t lut_index, int8_t *atten, int8_t *auxdac)
{
	int x;
	int16_t coeff_a, coeff_b;
	uint8_t temperature_code;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (rf_chain >= SX1301AR_BOARD_RFCHAIN_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_RFCHAIN_NB);
	}
	if (lut_index > (brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.size - 1)) {
		SX1301AR_ERR_RETURN(ERR_INVALID_LUT);
	}

	/* Get coeff, already multiplied by 2^16 */
	coeff_a = brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[lut_index].ad9361_tcomp.coeff_a;
	coeff_b = brd_cfg_priv[brd].rf_chain[rf_chain].tx_lut.lut[lut_index].ad9361_tcomp.coeff_b;

	/* Read current raw temperature */
	x = sx1301ar_get_temperature(brd, &temperature_code, NULL);
	SX1301AR_ERR_PROPAGATE(x);

	/* calculate attenuation to be applied to compensate temperature_code */
	*atten = sx1301ar_ad9361_temp2atten(temperature_code, brd_cfg_priv[brd].ad9361_temp_ref, coeff_a, coeff_b);

	/* Not used for now */
	*auxdac = 0;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static float get_rx_rssi_temperature_compensation(uint8_t brd, uint8_t rf_chain, uint8_t temperature_code)
{
	float temp_correction;
	int16_t coeff_a, coeff_b;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		return 0.0;
	}
	if (rf_chain >= SX1301AR_BOARD_RFCHAIN_NB) {
		return 0.0;
	}

	/* Get coeff, already multiplied by 2^16 */
	coeff_a = brd_cfg_priv[brd].rf_chain[rf_chain].rssi_offset_coeff_a;
	coeff_b = brd_cfg_priv[brd].rf_chain[rf_chain].rssi_offset_coeff_b;

	temp_correction = sx1301ar_ad9361_temp2rssi(temperature_code, brd_cfg_priv[brd].ad9361_temp_ref, coeff_a, coeff_b);

	return temp_correction;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static uint16_t get_tx_start_delay(bool tx_notch_enable, sx1301ar_bandw_t bw)
{
	float notch_delay_us = 0.0;
	float bw_delay_us = 0.0;
	float tx_start_delay;

	/* Notch filtering performed by FPGA adds a constant delay (group delay) that we need to compensate */
	if (tx_notch_enable) {
		notch_delay_us = (31.25 * ((64 + TX_NOTCH_TUNE) / 2)) / 1E3; /* 32MHz => 31.25ns */
	}

	/* Calibrated delay brought by SX1301 depending on signal bandwidth */
	switch (bw) {
	case BW_125K:
		bw_delay_us = 1.3;
		break;
	case BW_500K:
	/* Intended fall-through: it is the calibrated reference */
	default:
		break;
	}

	tx_start_delay = (float)TX_START_DELAY_DEFAULT - bw_delay_us - notch_delay_us;

	//debug("tx_start_delay=%u (%f) - (%u, bw_delay=%f, notch_delay=%f)", (uint16_t)(tx_start_delay + 0.5), tx_start_delay, TX_START_DELAY_DEFAULT, bw_delay_us, notch_delay_us);

	return (uint16_t)(tx_start_delay + 0.5); /* floor rounding */
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#if 0 /* Not used for now */
static int ad9361_capture( uint8_t brd, int16_t *sig_i, int16_t *sig_q, int capture_len )
{
    int i;
    uint8_t read_burst[CAPTURE_RAM_SIZE*4];
    uint16_t data_i_c2;
    uint16_t data_q_c2;

    sx1301ar_reg_brd_w( brd, BREG_CAPTURE_SOURCE , 0 );
    sx1301ar_reg_brd_w( brd, BREG_CAPTURE_PERIOD , 0 );

    sx1301ar_reg_brd_w( brd, BREG_CAPTURE_START, 1 );
    sx1301ar_wait_us( 100 );
    sx1301ar_reg_brd_w( brd, BREG_CAPTURE_START, 0 );

    sx1301ar_reg_brd_w( brd, BREG_CAPTURE_RAM_ADDR, 0 );
    sx1301ar_reg_brd_rb( brd, BREG_CAPTURE_RAM_DATA, read_burst, capture_len*4 );

    for( i=0 ; i<capture_len ; i++ )
    {
        data_i_c2 = ((uint16_t)read_burst[i*4+3] << 4) + ((uint16_t)read_burst[i*4+2] >> 4);
        data_q_c2 = ((uint16_t)read_burst[i*4+1] << 4) + ((uint16_t)read_burst[i*4  ] >> 4);
        sig_i[i] = (int16_t)((data_i_c2 > 2047) ? data_i_c2 - 4096 : data_i_c2);
        sig_q[i] = (int16_t)((data_q_c2 > 2047) ? data_q_c2 - 4096 : data_q_c2);
    }

    return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
static uint16_t ad9361_sig_analyser( uint8_t brd, float f, int avg_len  )
{
    int i;
    int16_t sig_i [CAPTURE_RAM_SIZE];
    int16_t sig_q [CAPTURE_RAM_SIZE];
    double phase;
    double corr_sig_i, corr_sig_q;
    double corr_sig_abs;

    ad9361_capture( brd, sig_i, sig_q, avg_len );

    corr_sig_i = 0;
    corr_sig_q = 0;

    for( i=0 ; i<avg_len ; i++ )
    {
        phase = -6.28318530717959*i*f/32e6;
        corr_sig_i += (double)sig_i[i] * cos( phase ) - (double)sig_q[i] * sin( phase );
        corr_sig_q += (double)sig_q[i] * cos( phase ) + (double)sig_i[i] * sin( phase );
    }

    corr_sig_abs = sqrt( corr_sig_i*corr_sig_i + corr_sig_q*corr_sig_q );

    return (uint16_t)(corr_sig_abs / avg_len);
}
#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int configure_board(uint8_t brd)
{
	int x;
	int32_t val;
	uint8_t chip, cx, dsp;
	int32_t radio_freq;
	int32_t chip_freq;
	int16_t chip_if_reg, chan_if_reg;
	uint32_t fttd_mult, fttd_lsa; /* used for frequency-to-time-drift calculation */
	unsigned bw, sf, ppm_en, br_div;
	uint8_t rx_en, tx_en;
	uint8_t rx_rf_chain;
	bool dsp_diversity = false;
	uint32_t chip_freq_base_rounded_mhz;
	int32_t chan_freq_base_offset_hz;
	uint8_t buff[TX_METADATA_NB];
	uint8_t dsp_cfg;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Return immediately if board is already running */
	if (brd_cfg_priv[brd].is_started == true) {
		SX1301AR_ERR_RETURN(ERR_BOARD_STATE);
	}

	/*------ Configure concentrator hardware ------*/

	/* Initialize board (check VERSION registers) */
	x = sx1301ar_reg_init(brd, brd_cfg_priv[brd].chip_nb, &brd_cfg_priv[brd].fpga_version);
	SX1301AR_ERR_PROPAGATE(x);

	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		/* Write 0 to the register that triggers a AD936SPI transaction. Otherwise the first transaction will fail. */
		x = sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_SEL, 0);
		SX1301AR_ERR_PROPAGATE(x);

		/* Configure SPI bridge */
		sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_CPOL, 0);
		sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_CPHA, 1);
		sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_CHIP_SELECT_POLARITY, 0);
		sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_SPEED, 5);

		/* Put the FPGA PLL in reset before AD9361 conf */
		sx1301ar_reg_brd_w(brd, BREG_PLL_RESET, 1);

		/* Get control of AD9361 Rx and Tx enables */
		sx1301ar_reg_brd_w(brd, BREG_AD9361_TXRX_HOST_CTRL, 1);
		sx1301ar_reg_brd_w(brd, BREG_AD9361_ENABLE, 0);
		sx1301ar_reg_brd_w(brd, BREG_AD9361_TXNRX, 0);

		/* RF Rx enable gating (if one Rx RF channel is not used) */
		sx1301ar_reg_brd_w(brd, BREG_FE_HOST_CTRL, 1);
		sx1301ar_reg_brd_w(brd, BREG_FE_RX_A_EN, (int32_t)brd_cfg_priv[brd].rf_chain[0].rx_enable);
		sx1301ar_reg_brd_w(brd, BREG_FE_RX_B_EN, (int32_t)brd_cfg_priv[brd].rf_chain[1].rx_enable);

		/* Hard reset AD9361 and SX1301 */
		sx1301ar_reg_brd_w(brd, BREG_AD9361_RESETB, 0);
		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 1);
		sx1301ar_wait_us(1000); //wait for reset settling time
		sx1301ar_reg_brd_w(brd, BREG_AD9361_RESETB, 1);
		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 0);
		sx1301ar_wait_us(1000); //wait for reset settling time
	} else {
		/* Hard reset SX1301 */
		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 1);
		sx1301ar_wait_us(1000); //wait for reset settling time
		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 0);
		sx1301ar_wait_us(1000); //wait for reset settling time

		/* Unreset FPGA PLL, needed for feedback clock */
		sx1301ar_reg_brd_w(brd, BREG_PLL_RESET, 0);
		sx1301ar_wait_us(1000);
		sx1301ar_reg_brd_r(brd, BREG_PLL_LOCKED, &val);
		if (val != 1) {
			SX1301AR_ERR_RETURN(ERR_FPGA_PLL);
		}

#if 1
		/*************************************************************
WORKAROUND TO HAVE LS COUNTER SYNCHRONIZED BETWEEN BOARDS
**************************************************************/
		sx1301ar_reg_brd_w(brd, BREG_SOFT_RESET, 1);
		sx1301ar_reg_brd_w(brd, BREG_SOFT_RESET, 0);

		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 1);
		sx1301ar_wait_us(1000); //wait for reset settling time
		sx1301ar_reg_brd_w(brd, BREG_SX1301_RESET, 0);
		sx1301ar_wait_us(1000); //wait for reset settling time

		/* Unreset FPGA PLL, needed for feedback clock */
		sx1301ar_reg_brd_w(brd, BREG_PLL_RESET, 0);
		sx1301ar_wait_us(1000);
		sx1301ar_reg_brd_r(brd, BREG_PLL_LOCKED, &val);
		if (val != 1) {
			SX1301AR_ERR_RETURN(ERR_FPGA_PLL);
		}
#endif /* END OF WORKAROUND */
	}

	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		/* Start and configure the radio */
		info("Configuring radio...");
		rx_en = (uint8_t)brd_cfg_priv[brd].rf_chain[0].rx_enable | ((uint8_t)brd_cfg_priv[brd].rf_chain[1].rx_enable << 1);
		tx_en = (uint8_t)brd_cfg_priv[brd].rf_chain[0].tx_enable | ((uint8_t)brd_cfg_priv[brd].rf_chain[1].tx_enable << 1);
		x = ad9361_start(brd, (uint64_t)brd_cfg_priv[brd].rx_freq, brd_cfg_priv[brd].rx_bw, rx_en, tx_en, brd_cfg_priv[brd].full_duplex);
		SX1301AR_ERR_PROPAGATE(x);

		/* Enable extension interface */
		sx1301ar_reg_brd_w(brd, BREG_EXT_IF_EN, 1);
	}

	/* DSP configuration */
	info("Configuring DSP...");
	sx1301ar_reg_brd_w(brd, BREG_UPP_CLK_DELAY, 128);
	sx1301ar_reg_brd_w(brd, BREG_UPP_LINE_SIZE, 255);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_0_IQ_INVERT, 0);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_1_IQ_INVERT, 0);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_CLIP_12B, 0);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_CHANN_GAIN, 11);
	sx1301ar_reg_brd_w(brd, BREG_DSP_ATTEN, 15);
	sx1301ar_reg_brd_w(brd, BREG_DBG_DSP_CHAN_SELECT, 128);

	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		/* Enable FPGA Tx I/Q path */
		sx1301ar_reg_brd_w(brd, BREG_TX_EN, (int32_t)(brd_cfg_priv[brd].rf_chain[0].tx_enable || brd_cfg_priv[brd].rf_chain[1].tx_enable));

		/* Tx notch filter */
		sx1301ar_reg_brd_w(brd, BREG_TX_NOTCH_TUNE, TX_NOTCH_TUNE);
	}

	/* Configure FPGA AGC */
	sx1301ar_reg_brd_w(brd, BREG_RSSI_THRESHOLD_HIGH, 17000);
	sx1301ar_reg_brd_w(brd, BREG_RSSI_THRESHOLD_LOW, 6000);
	sx1301ar_reg_brd_w(brd, BREG_RSSI_MAX_ATTEN, 12);
	sx1301ar_reg_brd_w(brd, BREG_RSSI_MIN_ATTEN, 6);
	sx1301ar_reg_brd_w(brd, BREG_RSSI_DEC_FILTER_ALPHA, 8);
	sx1301ar_reg_brd_w(brd, BREG_FIR_GAIN_STEP, 1);

	/* Enable DC notch */
	sx1301ar_reg_brd_w(brd, BREG_DC_NOTCH_EN, 255);

	/* Invert clocks to SX1301 to reduce consumption peaks */
	sx1301ar_reg_brd_w(brd, BREG_SX1301_CLK_32M_INVERT, 0);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_CLK_HS_INVERT, 0);

	/* SX1301 Rx interface */
	sx1301ar_reg_brd_w(brd, BREG_SX1301_IQ_INTERFACE_DELAY, 15);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_A_I_RX_SEL, 2);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_A_Q_RX_SEL, 3);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_B_I_RX_SEL, 0);
	sx1301ar_reg_brd_w(brd, BREG_SX1301_B_Q_RX_SEL, 1);

	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		/* RSSI histogram */
		sx1301ar_reg_brd_w(brd, BREG_RSSI_FILTER_1_GAIN, 15);
		sx1301ar_reg_brd_w(brd, BREG_RSSI_FILTER_1_DC_NOTCH_EN, 0);
		sx1301ar_reg_brd_w(brd, BREG_RSSI_FILTER_2_GAIN, 13);
		sx1301ar_reg_brd_w(brd, BREG_RSSI_HISTO_SELECT_RF_PATH, 0);
	}

	/*----- Deal with registers that impact timing: clock and filter enables, timer sync between SX1301s -----*/

	rx_rf_chain = 0;

	sx1301ar_reg_brd_w(brd, BREG_SX1301_CLK_32M_EN, 0); /* Ensure SX1301 clocks are off for now */

	for (chip = 0; chip < brd_cfg_priv[brd].chip_nb; ++chip) {
		if (((brd_cfg_priv[brd].btype == BRD_MASTER) && ((brd_cfg_priv[brd].chip_en[chip] == true) || (chip == SX1301AR_FREE_RUN_TIMER_CHIP) || (chip == SX1301AR_GPS_TIMER_CHIP))) ||
		    ((brd_cfg_priv[brd].btype == BRD_SLAVE) && (brd_cfg_priv[brd].chip_en[chip] == true))) {
			debug("[BRD%u] enabling chip %u", brd, chip);
			sx1301ar_reg_chip_w(brd, chip, CREG_GLOBAL_EN, 1);

			/* Force GPIO OUT to 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_GPIO_PIN_REG_OUT, 0);
			/* Force GPIO to OUTPUT mode */
			sx1301ar_reg_chip_w(brd, chip, CREG_GPIO_MODE, 31);
			/* Force GPIO value to CREG_GPIO_PIN_REG_OUT value */
			sx1301ar_reg_chip_w(brd, chip, CREG_GPIO_SELECT_OUTPUT, 8);
		}
		if (brd_cfg_priv[brd].chip_rf_chain[chip] > 0) {
			rx_rf_chain |= (1 << chip);
		}
	}
	sx1301ar_reg_brd_w(brd, BREG_RF_PATH_SELECT, rx_rf_chain);

	/* Enable latch on GPS PPS on 1 SX1301 */
	sx1301ar_reg_chip_w(brd, SX1301AR_GPS_TIMER_CHIP, CREG_GPS_EN, 1);

	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		/*------ First SX1301 (chip 0) now handles Rx/Tx controls on both RF chains so it needs a necessary minimum setup ------*/

		/* Set metadata part of TX data buffer to 0, to initialize temperature reading context */
		memset(buff, 0, sizeof buff);
		x = sx1301ar_reg_chip_w(brd, 0, CREG_TX_DATA_BUF_ADDR, 0);
		SX1301AR_ERR_PROPAGATE(x);
		x = sx1301ar_reg_chip_wb(brd, 0, CREG_TX_DATA_BUF_DATA, buff, sizeof buff);
		SX1301AR_ERR_PROPAGATE(x);

		/* AGC */
		/* Load firmware */
		x = load_firmware(brd, 0, MCU_AGC, agc_firmware, MCU_AGC_FW_BYTE);
		SX1301AR_ERR_PROPAGATE(x);

		/* Give control to MCU */
		sx1301ar_reg_brd_w(brd, BREG_FE_HOST_CTRL, 0);		    // Control of RF enables
		sx1301ar_reg_brd_w(brd, BREG_AD9361_TXRX_HOST_CTRL, 0);     // Control of AD9361 ENABLE and TXNRX pins
		sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_HOST_CTRL, 0);      // Control of SPI bridge to AD9361
		sx1301ar_reg_chip_w(brd, 0, CREG_FORCE_HOST_FE_CTRL, 0);    // Control of SX1301 RF enable pins
		sx1301ar_reg_chip_w(brd, 0, CREG_FORCE_HOST_RADIO_CTRL, 0); // Control of SX1301 SPI master A (to dedicated FPGA registers)

		/* Release AGC MCU reset*/
		sx1301ar_reg_chip_w(brd, 0, CREG_MCU_RST_1, 0);

		/* Tx settings */
		/* I/Q path setup */
		sx1301ar_reg_chip_w(brd, 0, CREG_TX_SWAP_IQ, 1); /* default 0 */
		//sx1301ar_reg_chip_w( brd, 0, CREG_TX_EDGE_SELECT_TOP, (chip+1) % 2 ); // To fit the inverted clocks
		//NB: Not applied for EU gateway V2

		/* TX general parameters */
		sx1301ar_reg_chip_w(brd, 0, CREG_TX_START_DELAY, TX_START_DELAY_DEFAULT); /* default 0 */

		/* TX FSK */
		sx1301ar_reg_chip_w(brd, 0, CREG_FSK_TX_GAUSSIAN_SELECT_BT, 2); /* Gaussian filter always on TX, default 0 */

		/* LoRa preamble for network cohabitation */
		if (brd_cfg_priv[brd].loramac_public == true) {
			sx1301ar_reg_chip_w(brd, 0, CREG_TX_FRAME_SYNCH_PEAK1_POS, 3); /* else == 1 */
			sx1301ar_reg_chip_w(brd, 0, CREG_TX_FRAME_SYNCH_PEAK2_POS, 4); /* else == 2 */
		}

		/* FSK user-settable sync word configuration */
		if (brd_cfg_priv[brd].fsk_sync_size > 0) {
			sx1301ar_reg_chip_w(brd, 0, CREG_FSK_TX_PSIZE, brd_cfg_priv[brd].fsk_sync_size - 1);
			sx1301ar_reg_chip_w(brd, 0, CREG_FSK_REF_PATTERN_LSB, brd_cfg_priv[brd].fsk_sync_lsb);
			sx1301ar_reg_chip_w(brd, 0, CREG_FSK_REF_PATTERN_MSB, brd_cfg_priv[brd].fsk_sync_msb);
		}
	}

	/*------ Configure the SX1301s ------*/

	/* Compute optimum freq-to-time-drift parameters */
	fttd_mult = 4096000000U / (brd_cfg_priv[brd].rx_freq >> 1); /* dividend: (4*2048*1000000) >> 1, rescaled to avoid 32b overflow */
	fttd_mult = (fttd_mult > 63) ? 63 : fttd_mult;		    /* saturation */
	fttd_lsa = 4096000000U / (brd_cfg_priv[brd].rx_freq >> 3);  /* dividend: (16*2048*1000000) >> 3, rescaled to avoid 32b overflow */
	fttd_lsa = (fttd_lsa > 63) ? 63 : fttd_lsa;		    /* saturation */

	/* Intermediate variables */
	radio_freq = (int32_t)(brd_cfg_priv[brd].rx_freq);
	chip_freq = 0;

	for (chip = 0; chip < brd_cfg_priv[brd].chip_nb; ++chip) {
		/* Shared settings between Rx and Tx */
		if (((brd_cfg_priv[brd].chip_en[chip] == true) && (chip > 0) && (brd_cfg_priv[brd].btype == BRD_MASTER)) ||
		    ((brd_cfg_priv[brd].chip_en[chip] == true) && (brd_cfg_priv[brd].btype == BRD_SLAVE))) {
			/* FSK user-settable sync word configuration */
			if (brd_cfg_priv[brd].fsk_sync_size > 0) {
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_REF_PATTERN_LSB, brd_cfg_priv[brd].fsk_sync_lsb);
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_REF_PATTERN_MSB, brd_cfg_priv[brd].fsk_sync_msb);
			}

			/* AGC */
			/* Control of SX1301 SPI master A (to dedicated FPGA registers) */
			sx1301ar_reg_chip_w(brd, chip, CREG_FORCE_HOST_RADIO_CTRL, 0);

			/* Load firmware */
			x = load_firmware(brd, chip, MCU_AGC, agc_firmware, MCU_AGC_FW_BYTE);
			SX1301AR_ERR_PROPAGATE(x);

			/* Release AGC MCU reset */
			sx1301ar_reg_chip_w(brd, chip, CREG_MCU_RST_1, 0);
		}

		if (brd_cfg_priv[brd].chip_en[chip] == true) {
			/* Set chip RX IF in FPGA */
			chip_if_reg = chip_if_freq2reg(radio_freq - (int32_t)(brd_cfg_priv[brd].chip_freq[chip]), &brd_chip_if_bias[brd][chip]);
			sx1301ar_reg_brd_w(brd, BREG_IF_FREQ_0 + chip, chip_if_reg);
			chip_freq = chip_if_reg2freq(chip_if_reg) + radio_freq;

			/* Enable High-Speed clock */
			sx1301ar_reg_brd_r(brd, BREG_SX1301_CLK_HS_EN, &val);
			val = val | (1 << chip);
			sx1301ar_reg_brd_w(brd, BREG_SX1301_CLK_HS_EN, val);

			/* Adjust RX clock polarity */
			sx1301ar_reg_chip_w(brd, chip, CREG_RX_EDGE_SELECT, 0); // to fit the inverted clocks

			/* Radio selection (only one radio) */
			sx1301ar_reg_chip_w(brd, chip, CREG_RADIO_SELECT, 0);  /* default 240 */
			sx1301ar_reg_chip_w(brd, chip, CREG_MISC_RADIO_EN, 1); /* default 0 */

			/* AGC */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_BB_FILTER_ALPHA, 6);      /* default 7 */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_DEC_FILTER_ALPHA, 7);     /* default 5 */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_CHANN_FILTER_ALPHA, 7);   /* default 8 */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_BB_DEFAULT_VALUE, 23);    /* default 32 */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_CHANN_DEFAULT_VALUE, 85); /* default 100 */
			sx1301ar_reg_chip_w(brd, chip, CREG_RSSI_DEC_DEFAULT_VALUE, 66);   /* default 100 */
			sx1301ar_reg_chip_w(brd, chip, CREG_DEC_GAIN_OFFSET, 7);	   /* default 8 */
			sx1301ar_reg_chip_w(brd, chip, CREG_CHAN_GAIN_OFFSET, 4);	  /* default 7 */

			/* LoRa 'multi' demodulators setup */
			sx1301ar_reg_chip_w(brd, chip, CREG_SNR_AVG_CST, 3); /* default 2 */

			/* Improvement of reference clock frequency error tolerance */
			sx1301ar_reg_chip_w(brd, chip, CREG_ADJUST_MODEM_START_OFFSET_RDX4, 1);
			sx1301ar_reg_chip_w(brd, chip, CREG_ADJUST_MODEM_START_OFFSET_SF12_RDX4, 4094);
			sx1301ar_reg_chip_w(brd, chip, CREG_CORR_MAC_GAIN, 7);

			/* FSK demodulator setup */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_RSSI_LENGTH, 4);	   /* default 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_PKT_MODE, 1);		   /* variable length, default 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_CRC_EN, 1);		   /* default 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_DCFREE_ENC, 2);		   /* default 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_ERROR_OSR_TOL, 10);	/* default 0 */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_PKT_LENGTH, 255);	  /* max packet length in variable length mode */
			sx1301ar_reg_chip_w(brd, chip, CREG_FSK_PATTERN_TIMEOUT_CFG, 128); /* sync timeout (allow 8 bytes preamble + 8 bytes sync word, default 0 */

			/* LoRa preamble for network cohabitation */
			if (brd_cfg_priv[brd].loramac_public == true) {
				/* Configure SX1301 */
				sx1301ar_reg_chip_w(brd, chip, CREG_FRAME_SYNCH_PEAK1_POS, 3);	/* else == 1 */
				sx1301ar_reg_chip_w(brd, chip, CREG_FRAME_SYNCH_PEAK2_POS, 4);	/* else == 2 */
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_FRAME_SYNCH_PEAK1_POS, 3); /* else == 1 */
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_FRAME_SYNCH_PEAK2_POS, 4); /* else == 2 */

				/* Configure DSP syncword */
				sx1301ar_reg_brd_w(brd, BREG_DSP_CFG_SYNC_WORD, 1);
			} else {
				/* Configure DSP syncword */
				sx1301ar_reg_brd_w(brd, BREG_DSP_CFG_SYNC_WORD, 0);
			}

			/* Configure DSP status report interval */
			sx1301ar_reg_brd_w(brd, BREG_DSP_STATUS_PERIOD, (int32_t)brd_cfg_priv[brd].dsp_stat_interval);

			/* FSK user-settable sync word configuration */
			if (brd_cfg_priv[brd].fsk_sync_size > 0) {
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_PSIZE, brd_cfg_priv[brd].fsk_sync_size - 1);
			}

			/* Freq-to-time-drift adjustment */
			sx1301ar_reg_chip_w(brd, chip, CREG_FREQ_TO_TIME_DRIFT, fttd_mult);
			sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_FREQ_TO_TIME_DRIFT, fttd_lsa);

			/* Set digital IF frequency for all enabled channels */
			for (cx = 0; cx < SX1301AR_CHIP_CHAN_NB; ++cx) {
				if (brd_cfg_priv[brd].chan_en[chip][cx] == true) {
					chan_if_reg = chan_if_freq2reg(chip_freq - (int32_t)(brd_cfg_priv[brd].chan_freq[chip][cx]), &brd_chan_if_bias[brd][cx + chip * SX1301AR_CHIP_CHAN_NB]);
					/* Set digital IF freq in SX1301 */
					sx1301ar_reg_chip_w(brd, chip, CREG_IF_FREQ_0 + cx, chan_if_reg);
				}
			}

			/* Set DSP base frequency, in MHz (integer) */
			chip_freq_base_rounded_mhz = brd_cfg_priv[brd].chip_freq[chip] / 1E6; /* precision below 1MHz is lost */
			sx1301ar_reg_brd_w(brd, BREG_SX1301_0_FREQ_BASE + ((8 + 1) * chip), (int32_t)chip_freq_base_rounded_mhz);

			/* Set FPGA digital IF frequency and DSP channel frequencies offsets for all LoRa channels */
			for (cx = 0; cx < SX1301AR_CHIP_MULTI_NB; ++cx) {
				if (brd_cfg_priv[brd].chan_en[chip][cx] == true) {
					/* Set FPGA digital IF frequency */
					chan_if_reg = chan_if_freq2reg(chip_freq - (int32_t)(brd_cfg_priv[brd].chan_freq[chip][cx]), &brd_chan_if_bias[brd][cx + chip * SX1301AR_CHIP_CHAN_NB]);
					sx1301ar_reg_brd_w(brd, BREG_SX1301_0_IF_FREQ_0 + cx + 8 * chip, chan_if_reg);

					/* Set DSP channel frequency offset (+/- 10KHz unit), compared to rounded base frequency */
					chan_freq_base_offset_hz = (int32_t)brd_cfg_priv[brd].chan_freq[chip][cx] - (int32_t)(chip_freq_base_rounded_mhz * 1E6);
					sx1301ar_reg_brd_w(brd, BREG_SX1301_0_CH0_FREQ_OFFSET + cx + ((8 + 1) * chip), chan_freq_base_offset_hz / 10E3);
				}
			}

			/* Setup multi-SF channels */
			for (cx = 0; cx < SX1301AR_CHIP_MULTI_NB; ++cx) {
				if (brd_cfg_priv[brd].chan_en[chip][cx] == true) {
					/* Switch channel on */
					sx1301ar_reg_chip_w(brd, chip, CREG_CORR0_DETECT_EN + cx, brd_cfg_priv[brd].multi_sf[chip][cx]); /* SF 7 to 12 enabled */
				} else {
					sx1301ar_reg_chip_w(brd, chip, CREG_CORR0_DETECT_EN + cx, 0); /* all SF disabled */
				}
			}
			sx1301ar_reg_chip_w(brd, chip, CREG_PPM_OFFSET, 0x60); /* 0x60 enable ppm_offset for SF12 and SF11 @125kHz */
			sx1301ar_reg_chip_w(brd, chip, CREG_CONCENTRATOR_MODEM_ENABLE, 1);

			/* Setup LoRa stand-alone channel */
			cx = SX1301AR_CHIP_LSA_IDX;
			if (brd_cfg_priv[brd].chan_en[chip][cx] == true) {
				/* Configure BW, SF and 'ppm offset' mode */
				bw = (unsigned)brd_cfg_priv[brd].lsa_bw[chip];
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_MODEM_BW, 3 - bw);

				sf = brd_cfg_priv[brd].lsa_sf[chip];
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_RATE_SF, sf);

				ppm_en = ppm_mode_en(bw, sf) ? 1 : 0;
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_PPM_OFFSET, ppm_en);

				/* Switch channel on */
				sx1301ar_reg_chip_w(brd, chip, CREG_MBWSSF_MODEM_ENABLE, 1);
			}

			/* Setup FSK channel */
			cx = SX1301AR_CHIP_FSK_IDX;
			if (brd_cfg_priv[brd].chan_en[chip][cx] == true) {
				/* Configure BW and data-rate */
				bw = (unsigned)brd_cfg_priv[brd].fsk_bw[chip];
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_CH_BW_EXPO, bw);

				br_div = CHIP_XTAL_FREQU / brd_cfg_priv[brd].fsk_br[chip];
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_BR_RATIO, br_div);

				/* Switch channel on */
				sx1301ar_reg_chip_w(brd, chip, CREG_FSK_MODEM_ENABLE, 1);
			}

			/* Load arbiter firmware */
			x = load_firmware(brd, chip, MCU_ARB, arb_firmware, MCU_ARB_FW_BYTE);
			SX1301AR_ERR_PROPAGATE(x);

			/* Get controller out of reset */
			sx1301ar_reg_chip_w(brd, chip, CREG_MCU_RST_0, 0);

			/* Set GPIO to output arbiter GPO */
			sx1301ar_reg_chip_w(brd, chip, CREG_GPIO_MODE, 31);
			sx1301ar_reg_chip_w(brd, chip, CREG_GPIO_SELECT_OUTPUT, 3);
		}
	}

	/* Configure DSP diversity (global for all DSP on the board) */
	if (brd_cfg_priv[brd].btype == BRD_MASTER) {
		if (brd_cfg_priv[brd].chip_rf_chain[0] == brd_cfg_priv[brd].chip_rf_chain[1]) /* No more than 2 SX1301s per DSP */
		{
			/* Single antenna mode - No diversity */
			dsp_diversity = false;
		} else {
			/* Dual antenna mode - Full diversity or not ? */
			dsp_diversity = true;
			brd_cfg_priv[brd].rf_diversity = false;
			for (cx = 0; cx < SX1301AR_CHIP_CHAN_NB; ++cx) {
				if (brd_cfg_priv[brd].chan_freq[0][cx] != brd_cfg_priv[brd].chan_freq[1][cx]) {
					/* Fill channel diversity table */
					brd_channel_diversity_table[brd][cx] = false;
					/* If one channel frequency differs from one antenna to the other, disable diversity */
					dsp_diversity = false;
				} else {
					/* Fill channel diversity table */
					brd_channel_diversity_table[brd][cx] = true;
					/* If one channel has diversity, set the whole board in diversity mode for fine timestamp matching */
					brd_cfg_priv[brd].rf_diversity = true;
				}
			}
		}
	} else {
		/* Slave board - No diversity (only one RF chain sent to SLAVE board) */
		dsp_diversity = false;
	}
	sx1301ar_reg_brd_w(brd, BREG_DSP_CFG_DIVERSITY_PARAM, (dsp_diversity == true) ? 1 : 0);

	/* Configure DSP fine timestamp debug and version */
	dsp_cfg = (brd_cfg_priv[brd].debug_tmst == true) ? 1 : 0;       /* enable/disable fine timestamp debug mode - 1 bit */
	dsp_cfg |= ((brd_cfg_priv[brd].main_tmst_version << 1) & 0x06); /* main fine timestamp version - 2 bits */
	sx1301ar_reg_brd_w(brd, BREG_DSP_CFG_SPARE, (int32_t)dsp_cfg);

	/* Boot the DSPs */
	if (brd_cfg_priv[brd].dsp_nb) {
		/* Boot the DSPs */
		for (dsp = 0; dsp < brd_cfg_priv[brd].dsp_nb; dsp++) {
			x = gwloc_dsp_boot(brd, dsp, GWLOC_BOOT_RETRY);
			SX1301AR_ERR_PROPAGATE(x);
		}
	} else {
		/* Start FPGA DSP PLL */
		x = sx1301ar_reg_brd_w(brd, BREG_DSP_PLL_RESET, 0);
		SX1301AR_ERR_PROPAGATE(x);
		sx1301ar_wait_us(1000);
		x = sx1301ar_reg_brd_r(brd, BREG_DSP_PLL_LOCKED, &val);
		SX1301AR_ERR_PROPAGATE(x);
		if (val != 1) {
			SX1301AR_ERR_RETURN(ERR_FPGA_DSP_PLL);
		}
	}

	/* Listen-Before-Talk */
	if (sx1301ar_i_lbt_is_enabled(brd) == true) {
		x = sx1301ar_i_init_lbt(brd, brd_cfg_priv[brd].rx_freq);
		SX1301AR_ERR_PROPAGATE(x);

		x = sx1301ar_i_start_lbt(brd);
		SX1301AR_ERR_PROPAGATE(x);
	}

	/* Print channel diversity table based on current configuration */
	debug("[BRD%d] channel diversity table:", brd);
	dump(brd_channel_diversity_table[brd], SX1301AR_CHIP_CHAN_NB, "diversity");

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int fetch_packets(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t max_nb, uint8_t *nb_pkt, uint8_t temperature_code)
{
	uint8_t cpt = 0; /* number of retrieved packets */
	uint8_t chip;
	uint8_t buff[255 + RX_METADATA_NB];      /* buffer to store the result of SPI read bursts */
	unsigned sz;				 /* size of the payload, uses to address metadata */
	uint8_t cx;				 /* local (chip level) channel ID */
	unsigned sf, cr, bw_pow, crc_en, ppm_en; /* used to calculate timestamp correction */
	uint32_t raw_timestamp;			 /* timestamp when internal 'RX finished' was triggered */
	uint32_t delay_x, delay_y, delay_z;      /* temporary variable for timestamp offset calculation */
	uint32_t timestamp_correction;		 /* correction to account for processing delay */
	int pkt_idx;
	uint8_t rsig_antenna = 0;
	/* performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((p == NULL) || (p_rsig == NULL) || (nb_pkt == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (max_nb > SX1301AR_MAX_PKT_NB) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Fetch SX1301 chips to get packets */
	chip = 0;
	pkt_idx = 0;
	while ((cpt < max_nb) && (chip < brd_cfg_priv[brd].chip_nb)) {
		/* If chip is disabled, skip and go to next chip */
		if (brd_cfg_priv[brd].chip_en[chip] != true) {
			++chip;
			continue;
		}

		/* Fetch all the RX FIFO data, including FIFO fill counter */
		/* Get FIFO_NUM_STORED, FIFO_ADDR_POINTER, FIFO_STATUS and FIFO_PAYLOAD_SIZE in one go */
		sx1301ar_reg_chip_rb(brd, chip, CREG_RX_PACKET_DATA_FIFO_NUM_STORED, buff, 5);

		/* Go to next chip if no packets */
		if (buff[0] == 0) {
			++chip;
			continue;
		} else if (buff[0] > 15) {
			warning("sx1301_%u packet FIFO is FULL, packets may have been lost", chip);
		}

		/* Process FIFO informations */
		p[pkt_idx].size = buff[4];
		sz = buff[4];

		switch (buff[3] & 0x07) {
		/* The packet status as indicated in the FIFO */
		case 5:
			p[pkt_idx].status = STAT_CRC_OK;
			crc_en = 1;
			nb_pkt_sx1301_received_crc_ok++; /* ftime stats */
			break;
		case 7:
			p[pkt_idx].status = STAT_CRC_ERR;
			crc_en = 1;
			nb_pkt_sx1301_received_crc_err++; /* ftime stats */
			break;
		case 1:
			p[pkt_idx].status = STAT_NO_CRC;
			crc_en = 0;
			break;
		default:
			p[pkt_idx].status = STAT_UNDEFINED;
			crc_en = 0;
		}

		/* /!\ RX_PACKET_DATA_FIFO_ADDR_POINTER is automatically copied to RX_DATA_BUF_ADDR */

		/* Get payload + metadata, copy payload to RX packet structure */
		sx1301ar_reg_chip_rb(brd, chip, CREG_RX_DATA_BUF_DATA, buff, sz + RX_METADATA_NB);
		memcpy((void *)p[pkt_idx].payload, (void *)buff, sz);

		/* Processing of generic meta-data */
		cx = 0x0F & buff[sz + 0];												 /* local channel ID */
		p_rsig[pkt_idx].chan = cx + (chip << 4);										 /* board-level channel ID */
		p_rsig[pkt_idx].chan += ((brd_cfg_priv[brd].btype == BRD_SLAVE) ? 16 * brd_cfg_priv[SX1301AR_BOARD_MASTER].chip_nb : 0); /* multiboard-level channel ID */
		p[pkt_idx].freq_hz = brd_cfg_priv[brd].chan_freq[chip][cx];

		/* Process RAW RSSI information */
		p_rsig[pkt_idx].rssi = calc_rssi(brd, cx, buff[sz + 5], brd_cfg_priv[brd].chip_rf_chain[chip], temperature_code); /* use global config to compensate for various offsets */

		/* Processing of modulation-dependant meta-data */
		if (cx <= SX1301AR_CHIP_LSA_IDX) /* LoRa modulation */
		{
			p[pkt_idx].modulation = MOD_LORA;
			p[pkt_idx].bandwidth = (cx == SX1301AR_CHIP_LSA_IDX) ? brd_cfg_priv[brd].lsa_bw[chip] : BW_125K; /* variable BW for LoRa stand-alone channel */
			p_rsig[pkt_idx].snr = ((float)((int8_t)buff[sz + 2])) / 4;

			sf = (buff[sz + 1] >> 4) & 0x0F; /* LoRa SF encoded on bits 4 to 7 */
			p[pkt_idx].modrate = sx1301ar_sf_nb2enum(sf);

			cr = (buff[sz + 1] >> 1) & 0x07; /* LoRa coding rate encoded on bits 1 to 3 */
			p[pkt_idx].coderate = sx1301ar_cr_nb2enum(cr);

			/* Timestamp correction (base delay) */
			if (cx == SX1301AR_CHIP_LSA_IDX) { /* LoRa stand-alone channel */
				switch (p[pkt_idx].bandwidth) {
				case BW_125K:
					delay_x = 64;
					bw_pow = 1;
					break;
				case BW_250K:
					delay_x = 32;
					bw_pow = 2;
					break;
				case BW_500K:
					delay_x = 16;
					bw_pow = 4;
					break;
				default:
					delay_x = 0;
					bw_pow = 0;
				}
			} else { /* multi-SF channel */
				delay_x = 114;
				bw_pow = 1;
			}

			/* Timestamp correction (variable delay) */
			ppm_en = ppm_mode_en(p[pkt_idx].bandwidth, sf) ? 1 : 0;
			if ((sf >= 6) && (sf <= 12) && (bw_pow > 0)) {
				if ((2 * (sz + 2 * crc_en) - (sf - 7)) <= 0) { /* payload fits entirely in first 8 symbols */
					delay_y = (((1 << (sf - 1)) * (sf + 1)) + (3 * (1 << (sf - 4)))) / bw_pow;
					delay_z = 32 * (2 * (sz + 2 * crc_en) + 5) / bw_pow;
				} else {
					delay_y = (((1 << (sf - 1)) * (sf + 1)) + ((4 - ppm_en) * (1 << (sf - 4)))) / bw_pow;
					delay_z = (16 + 4 * cr) * (((2 * (sz + 2 * crc_en) - sf + 6) % (sf - 2 * ppm_en)) + 1) / bw_pow;
				}
			} else {
				delay_y = 0;
				delay_z = 0;
			}

			timestamp_correction = delay_x + delay_y + delay_z;
		} else if (cx == SX1301AR_CHIP_FSK_IDX) /* FSK modulation */
		{
			p[pkt_idx].modulation = MOD_FSK;
			p[pkt_idx].bandwidth = brd_cfg_priv[brd].fsk_bw[chip];
			p_rsig[pkt_idx].snr = -128.0;
			p[pkt_idx].modrate = (sx1301ar_modr_t)(brd_cfg_priv[brd].fsk_br[chip]);
			p[pkt_idx].coderate = CR_UNDEFINED;

			timestamp_correction = 0; // TODO: implement FSK timestamp correction
		} else				  /* channel does not exist */
		{
			SX1301AR_ERR_RETURN(ERR_INVALID_CH);
		}

		/* Calculating packet timestamp */
		raw_timestamp = (uint32_t)buff[sz + 6] + ((uint32_t)buff[sz + 7] << 8) + ((uint32_t)buff[sz + 8] << 16) + ((uint32_t)buff[sz + 9] << 24);
		p[pkt_idx].count_us = raw_timestamp - timestamp_correction;

		/* Initialize rsig structure */
		memset(&(p[pkt_idx].rsig), 0, sizeof p[pkt_idx].rsig);

		/* Fill rsig for this channel */
		rsig_antenna = brd_cfg_priv[brd].chip_rf_chain[chip];
		p[pkt_idx].rsig[rsig_antenna].is_valid = true;
		p[pkt_idx].rsig[rsig_antenna].chan = p_rsig[pkt_idx].chan;
		p[pkt_idx].rsig[rsig_antenna].rssi_chan = p_rsig[pkt_idx].rssi;
		p[pkt_idx].rsig[rsig_antenna].snr = p_rsig[pkt_idx].snr;

		/* Count one packet and advance packet FIFO */
		cpt++;
		pkt_idx++;

		sx1301ar_reg_chip_w(brd, chip, CREG_RX_PACKET_DATA_FIFO_NUM_STORED, 0);
	}

	*nb_pkt = cpt;

	/* Compute time spent in this function */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int merge_packets(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt)
{
	uint8_t cpt;
	int j, k, pkt_idx, pkt_dup_idx, x;
	uint8_t rsig_antenna = 0;
	bool dup_restart = false;
	/* Performances variables */
	struct timeval tm;

	/* Check input parameters */
	if ((p == NULL) || (p_rsig == NULL) || (nb_pkt == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Init number of packets in array before merge */
	cpt = *nb_pkt;

	/* Remove duplicates */
	j = 0;
	while (j < cpt) {
		for (k = (j + 1); k < cpt; k++) {
			/* Check if current packet's channel exists on both RF chains and so can generate duplicates */
			if (brd_channel_diversity_table[brd][p_rsig[j].chan % 16] == false) {
				continue;
			}

			/* Searching for duplicated packets:
                -- count_us should be equal or can have up to 24µs of difference (3 samples)
                -- abs(chan_ant0 - chan_ant1) == 16 : this is how channels are mapped on SX1301s
                -- modrate should be same
            */
			if (is_same_pkt(&p[j], &p_rsig[j], &p[k], &p_rsig[k])) {
				/* We keep the packet which has CRC checked */
				if ((p[j].status == STAT_CRC_OK) && (p[k].status == STAT_CRC_ERR)) {
					pkt_dup_idx = k;
					pkt_idx = j;
				} else if ((p[j].status == STAT_CRC_ERR) && (p[k].status == STAT_CRC_OK)) {
					pkt_dup_idx = j;
					pkt_idx = k;
				} else {
					/* we keep the packet which has best SNR */
					if (p_rsig[j].snr >= p_rsig[k].snr) {
						pkt_dup_idx = k;
						pkt_idx = j;
					} else {
						pkt_dup_idx = j;
						pkt_idx = k;
					}
				}
				/* pkt_dup_idx contains the index to be deleted */
				/* Now, merge the rsig paths info */
				rsig_antenna = p_rsig[pkt_dup_idx].chan / 16; /* second rsig path */
				p[pkt_idx].rsig[rsig_antenna].is_valid = true;
				p[pkt_idx].rsig[rsig_antenna].chan = p_rsig[pkt_dup_idx].chan;
				p[pkt_idx].rsig[rsig_antenna].rssi_chan = p_rsig[pkt_dup_idx].rssi;
				p[pkt_idx].rsig[rsig_antenna].snr = p_rsig[pkt_dup_idx].snr;
				/* Remove duplicated packet from packet array */
				x = remove_pkt(p, p_rsig, &cpt, pkt_dup_idx);
				if (x != 0) {
					error("failed to remove packet from array (%d)", x);
				}
				dup_restart = true;
				break;
			}
		}
		if (dup_restart == true) {
			/* Duplicate found, restart searching for duplicate from first element */
			j = 0;
			dup_restart = false;
		} else {
			/* No duplicate found, continue... */
			j += 1;
		}
	}

	/* Update number of packets contained in packet array */
	*nb_pkt = cpt;

	/* Compute time spent in this function */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int fetch_fpga(uint8_t brd, gwloc_tmst_queue_t *queue)
{
	int i, j;
	int x;
	gwloc_dsp_msg_t fpga_messages[255];
	gwloc_dsp_tmst_t fine_timestamp;
	uint8_t nb_msg;
	/* Performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (queue == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (brd_cfg_priv[brd].dsp_nb == 0) {
		/* No need to fetch FPGA for fine timestamp if no DSP on the board */
		return 0;
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Fetch all DSPs available on the board */
	for (i = 0; i < brd_cfg_priv[brd].dsp_nb; i++) {
		/* Fetch messages from FPGA FIFO */
		memset(&fpga_messages, 0, sizeof fpga_messages);
		nb_msg = 0;
		x = gwloc_fpga_fetch(brd, i, fpga_messages, 255, &nb_msg);
		if (x != 0) {
			error("failed to fetch FPGA");
		} else {
			if (nb_msg > 0) {
				/* Update reference for time conversions */
				x = gwloc_update_ref_counter(brd);
				if (x != 0) {
					error("%s: Failed to update reference for time conversion", __FUNCTION__);
					SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
				}

				for (j = 0; j < nb_msg; j++) {
					switch (fpga_messages[j].id) {
					case GWLOC_DSP_MSG_TYPE_TIMESTAMP:
						/* Convert raw data to fine timestamp structure */
						x = gwloc_dsp_parse_timestamp(brd, brd_cfg_priv[brd].btype, i, &fpga_messages[j], &fine_timestamp);
						if (x != 0) {
							error("failed to parse fine timestamp");
						} else {
							nb_fine_tmst_received++; /* ftime stats */

							/* Enqueue fine timestamp in queue */
							x = gwloc_tmst_enqueue(queue, &fine_timestamp);
							if (x != 0) {
								error("failed to enqueue fine timestamp");
							}
						}
						break;
					default:
						error("Unknown DSP message");
						break;
					}
				}
			}
		}
	}

	/* Compute time spent in this function */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int fetch_dsp(uint8_t brd)
{
	int i, j, k;
	int x;
	uint8_t nb_msg;
	gwloc_dsp_msg_t dsp_messages[255];
	gwloc_dsp_cfg_t dsp_cfg;
	gwloc_dsp_status_t dsp_status;
	char log[GWLOC_BOARD_DSP_FIFO_SIZE]; /* there could be only one log in the whole FIFO */
	/* Performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Fetch all DSPs available on the board */
	for (i = 0; i < brd_cfg_priv[brd].dsp_nb; i++) {
		/* Fetch DSP messages */
		memset(&dsp_messages, 0, sizeof dsp_messages);
		nb_msg = 0;
		x = gwloc_dsp_fetch(brd, i, dsp_messages, 255, &nb_msg);
		if (x != 0) {
			error("failed to fetch dsp %d", i);
			continue;
		}

		/* Parse messages */
		if (nb_msg > 0) {
			for (j = 0; j < nb_msg; j++) {
				switch (dsp_messages[j].id) {
				case GWLOC_DSP_MSG_TYPE_LOG_DEBUG:
					strncpy(log, (char *)(dsp_messages[j].payload), dsp_messages[j].payload_length);
					log[dsp_messages[j].payload_length] = '\0';
					debug("[BRD%d][*DSP%d*] LOG_DEBUG (length=%d):\t%s", brd, i, dsp_messages[j].payload_length, log);
					break;
				case GWLOC_DSP_MSG_TYPE_LOG_LOC:
					strncpy(log, (char *)(dsp_messages[j].payload), dsp_messages[j].payload_length);
					log[dsp_messages[j].payload_length] = '\0';
					debug("[BRD%d][*DSP%d*] LOG_LOC   (length=%d):\t%s", brd, i, dsp_messages[j].payload_length, log);
					break;
				case GWLOC_DSP_MSG_TYPE_TIMESTAMP:
					/* DSP should not send clear fine timestamp */
					SX1301AR_ERR_RETURN(ERR_LOC_FORBIDDEN);
					break;
				case GWLOC_DSP_MSG_TYPE_CONFIG:
					/* Convert raw data to dsp config structure */
					x = gwloc_dsp_parse_config(&dsp_messages[j], &dsp_cfg);
					if (x != 0) {
						error("failed to parse DSP configuration message");
					} else {
						/* Record DSP version when we get this message */
						brd_cfg_priv[brd].dsp_version = (int16_t)dsp_cfg.dsp_version;
						debug("[BRD%d][*DSP%d*] CONFIG (length=%d) received:", brd, i, dsp_messages[j].payload_length);
						debug("[BRD%d][*DSP%d*] -- DSP version=%u", brd, i, dsp_cfg.dsp_version);
						debug("[BRD%d][*DSP%d*] -- LoRaWAN Public=%d", brd, i, dsp_cfg.lorawan_public);
						debug("[BRD%d][*DSP%d*] -- Diversity=%d", brd, i, dsp_cfg.diversity);
						debug("[BRD%d][*DSP%d*] -- PLL0=%uMHz", brd, i, dsp_cfg.pll0);
						debug("[BRD%d][*DSP%d*] -- PLL1=%uMHz", brd, i, dsp_cfg.pll1);
						debug("[BRD%d][*DSP%d*] -- number of channels=%u", brd, i, dsp_cfg.chan_nb);
						for (k = 0; k < (SX1301AR_BOARD_NB_CHIP_PER_DSP * SX1301AR_CHIP_MULTI_NB); k++) {
							debug("[BRD%d][*DSP%d*] -- CH%d freq=%.3fMHz", brd, i, k, dsp_cfg.chan_freq[k]);
						}
						debug("[BRD%d][*DSP%d*] -- Report period=%us", brd, i, dsp_cfg.report_period * 10);
					}
					break;
				case GWLOC_DSP_MSG_TYPE_STATUS:
					/* Convert raw data to dsp status structure */
					x = gwloc_dsp_parse_status(&dsp_messages[j], &dsp_status);
					if (x != 0) {
						error("failed to parse DSP status message");
					} else {
						debug("[BRD%d][*DSP%d*] STATUS (length=%d) received:", brd, i, dsp_messages[j].payload_length);
						debug("[BRD%d][*DSP%d*] -- log counter=%u", brd, i, dsp_status.log_counter);
						debug("[BRD%d][*DSP%d*] -- nb jobs added=%u", brd, i, dsp_status.nb_jobs_added);
						debug("[BRD%d][*DSP%d*] -- nb jobs executed=%u", brd, i, dsp_status.nb_jobs_executed);
						debug("[BRD%d][*DSP%d*] -- nb jobs in queue=%u", brd, i, dsp_status.nb_jobs_in_queue);
						debug("[BRD%d][*DSP%d*] -- nb jobs in queue max=%u", brd, i, dsp_status.nb_jobs_max_in_queue);
						debug("[BRD%d][*DSP%d*] -- nb jobs dropped=%u", brd, i, dsp_status.nb_jobs_dropped);
						debug("[BRD%d][*DSP%d*] -- nb jobs dropped acc=%u", brd, i, dsp_status.nb_jobs_dropped_acc);
						debug("[BRD%d][*DSP%d*] -- nb valid ts=%u", brd, i, dsp_status.nb_valid_timestamps);
						debug("[BRD%d][*DSP%d*] -- nb sync lost=%u", brd, i, dsp_status.nb_lost_sync);
						debug("[BRD%d][*DSP%d*] -- nb sync lost acc=%u", brd, i, dsp_status.nb_lost_sync_acc);
						debug("[BRD%d][*DSP%d*] -- sx1301 0 max fifo=%u", brd, i, dsp_status.sx1301_0_max_fifo);
						debug("[BRD%d][*DSP%d*] -- sx1301 1 max fifo=%u", brd, i, dsp_status.sx1301_1_max_fifo);
						debug("[BRD%d][*DSP%d*] -- dsp max fifo=%u", brd, i, dsp_status.dsp_max_fifo);
						debug("[BRD%d][*DSP%d*] -- dsp max enc fifo=%u", brd, i, dsp_status.dsp_max_enc_fifo);

						/* UPP sync loss is quite critical, we should never get any */
						if (dsp_status.nb_lost_sync != 0) {
							warning("[%llu] [BRD%d][*DSP%d*] UPP sync lost=%u (accumulated:%u)", sx1301ar_timestamp(), brd, i, dsp_status.nb_lost_sync, dsp_status.nb_lost_sync_acc);
						}
					}
					break;
				default:
					error("Unknown DSP message");
					break;
				}
			}
		}
	}

	/* Compute time spent in this function */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int merge_timestamps(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt, gwloc_tmst_queue_t *queue, uint8_t temperature_code)
{
	int j, k, x;
	int tmst_idx[SX1301AR_BOARD_RFCHAIN_NB], ret_peek;
	uint8_t rsig_antenna = 0;
	float rssi_dsp_offset;
	uint8_t nb_matched_tmst;
	gwloc_dsp_tmst_t fine_timestamp = {0};
	bool isvalid_fts;
	uint8_t brd_master;
	/* Performances variables */
	struct timeval tm;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((p == NULL) || (queue == NULL) || (p_rsig == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (brd_cfg_priv[brd].dsp_nb == 0) {
		/* No fine timestamp to merge if no DSP on the board */
		return 0;
	}

	/* Record function start time */
	_meas_time_start(&tm);

	/* Look for matching fine timestamps */
	for (j = 0; j < (*nb_pkt); j++) {
		if (gwloc_tmst_queue_is_empty(queue)) {
			nb_no_fine_timestamp++; /* ftime stats */
			debug("No fine timestamp available for matching");
		} else {
			/* Do not try to merge a timestamp if packet CRC is not OK */
			if ((p[j].status != STAT_CRC_OK) && (brd_cfg_priv[brd].match_tmst_crc_err == false)) {
				debug("Skip fine timestamp matching for packet with CRC error");
				continue;
			}
			/* Search for fine timestamp correlation in queue (we can find up to 2 correlated
                fine timestamp, 0 or 1 per antenna) */
			ret_peek = gwloc_tmst_peek(queue, &p[j], brd_cfg_priv[brd].rf_diversity, &tmst_idx[0], &nb_matched_tmst);
			if (ret_peek < 0) {
				error("failed to peek fine timestamp");
				continue;
			}

			nb_match += nb_matched_tmst; /* ftime stats */

			/* Dequeue fine timestamp and update packet structure */
			if (nb_matched_tmst > 0) {
				for (k = 0; k < nb_matched_tmst; k++) {
					/* Found a matching fine timestamp, dequeue it */
					if (gwloc_tmst_dequeue(queue, tmst_idx[k], &fine_timestamp) == 0) {
						/* Get antenna on which the corresponding packet has been received */
						if (brd_cfg_priv[brd].chip_rf_chain[0] != brd_cfg_priv[brd].chip_rf_chain[1]) {
							/* Dual antenna */
							rsig_antenna = fine_timestamp.chan / 16; /* 16 channels per SX1301 */
						} else {
							/* Single antenna */
							rsig_antenna = brd_cfg_priv[brd].chip_rf_chain[0];
						}

						/* It could happen that we received a fine timestamp from the DSP, without receiving
                            a packet from SX1301, in this case we enable the antenna path and add
                            fine timestamp and signal information */
						if (p[j].rsig[rsig_antenna].is_valid != true) {
							debug("No packet received on antenna %d, adding fine timestamp for channel %u", rsig_antenna, fine_timestamp.chan);
							p[j].rsig[rsig_antenna].is_valid = true;
							/* Add missing channel information */
							p[j].rsig[rsig_antenna].chan = fine_timestamp.chan;
						}

						/* Drop the fine timestamp if invalid, but keep RSSI/SNR coming from DSP */
						gwloc_dsp_parse_timestamp_status(fine_timestamp.status, &isvalid_fts);
						if (isvalid_fts == true) {
							/* Indicate that a fine timestamp is available for this radio path */
							p[j].rsig[rsig_antenna].fine_received = true;

							/* Encrypted TS correponds to last preambule symbol, no need to adjust */
							p[j].rsig[rsig_antenna].fine_tmst = fine_timestamp.ts_lsp_frac_dec;
							memcpy(p[j].rsig[rsig_antenna].fine_tmst_enc, fine_timestamp.ts_lsp_frac_enc, 16);

							/* Set sub band fine timestamp info */
							p[j].rsig[rsig_antenna].fine_tmst_alt = fine_timestamp.ts2_lsp_delta;
							p[j].rsig[rsig_antenna].fine_tmst_debug1 = fine_timestamp.ts_debug1;
							p[j].rsig[rsig_antenna].fine_tmst_debug2 = fine_timestamp.ts_debug2;

							/* Set encrypted/main fine timestamp version */
							p[j].rsig[rsig_antenna].fine_tmst_version = brd_cfg_priv[brd].main_tmst_version;
						} else {
							/* Indicate that no fine timestamp is available for this radio path */
							p[j].rsig[rsig_antenna].fine_received = false;
							error("fine timestamp (SF=%d,CH=%u,ts_lsp_ls_int=%u) has been dropped, status=0x%02X", sx1301ar_sf_enum2nb(fine_timestamp.modrate), fine_timestamp.chan, fine_timestamp.ts_lsp_ls_int, fine_timestamp.status);
							nb_invalid_fine_timestamp += 1;
						}

						/* Indicate that a signal info is available for this radio path */
						p[j].rsig[rsig_antenna].sig_info_received = true;

						/* Always indicate the status of the fine timestamp, even if invalid */
						p[j].rsig[rsig_antenna].fine_tmst_status = fine_timestamp.status;

						/* Set Frequency Offset of this radio path */
						p[j].rsig[rsig_antenna].freq_offset = (int16_t)(fine_timestamp.freq_offset * 125.0E3) + brd_chan_if_bias[brd][(fine_timestamp.chan >> 4) * SX1301AR_CHIP_CHAN_NB + (fine_timestamp.chan & 0xF)] + brd_chip_if_bias[brd][rsig_antenna];

						/* overload RSSI/SNR info with more precise info */
						brd_master = (brd_cfg_priv[brd].btype == BRD_MASTER) ? brd : SX1301AR_BOARD_MASTER;														   /* get offset from master board */
						rssi_dsp_offset = roundf(brd_cfg_priv[brd_master].rf_chain[rsig_antenna].rssi_offset + SX1301_DSP_RSSI_OFFSET + get_rx_rssi_temperature_compensation(brd_master, rsig_antenna, temperature_code)); /* 1dBm precision */
						p[j].rsig[rsig_antenna].rssi_chan = fine_timestamp.rssi_chan + rssi_dsp_offset;
						p[j].rsig[rsig_antenna].rssi_sig = fine_timestamp.rssi_sig + rssi_dsp_offset;
						p[j].rsig[rsig_antenna].rssi_sig_std = fine_timestamp.rssi_sig_std;
						p[j].rsig[rsig_antenna].snr = ((fine_timestamp.snr > SNR_THRESHOLD_MAX) ? SNR_THRESHOLD_MAX : fine_timestamp.snr);
					}
				}

				/* Sanity checks on returned packet */
				/* 1: If 2 antennas received, channel numbers must be (x, x+16): */
				if ((p[j].rsig[0].is_valid == true) && (p[j].rsig[1].is_valid == true)) {
					if (abs(p[j].rsig[1].chan - p[j].rsig[0].chan) != 16) {
						error("wrong channel information (ant0:%u, ant1:%u), dropping packet %d", p[j].rsig[0].chan, p[j].rsig[1].chan, j);
						x = remove_pkt(p, p_rsig, nb_pkt, j);
						if (x != 0) {
							error("failed to remove packet from array (%d)", x);
						}
					}
				}
			} else {
				if (p[j].status == STAT_CRC_OK) {
					debug("no fine timestamp found for this packet count_us=%u (crc:%d)", p[j].count_us, p[j].status);
				}
			}

			/* Purge obsolete fine timestamps */
			gwloc_tmst_purge(queue);
		}
	}

	/* Compute time spent in this function */
	_meas_time_stop(1, tm, __FUNCTION__);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static bool is_same_pkt(sx1301ar_rx_pkt_t *p1, pkt_rsig_t *p1_rsig, sx1301ar_rx_pkt_t *p2, pkt_rsig_t *p2_rsig)
{
	if ((p1 != NULL) && (p2 != NULL) && (p1_rsig != NULL) && (p2_rsig != NULL)) {
		/* Criterias to determine if packets are identical:
            -- count_us should be equal or can have up to 24µs of difference (3 samples)
            -- abs(chan_ant0 - chan_ant1) == 16 : this is how channels are mapped on SX1301s
            -- modrate should be same
        */
		if (((abs(p1->count_us - p2->count_us) <= 24)) &&
		    (abs(p1_rsig->chan - p2_rsig->chan) == 16) &&
		    (p1->modrate == p2->modrate)) {
			return true;
		}
	}

	return false;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int remove_pkt(sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt, uint8_t pkt_index)
{
	/* Check input parameters */
	if ((p == NULL) || (p_rsig == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (pkt_index > ((*nb_pkt) - 1)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED);
	}

	/* Remove pkt from array, by replacing it with last packet of array */
	if (pkt_index == ((*nb_pkt) - 1)) {
		/* If we remove last element, just decrement nb packet counter */
		/* Do nothing */
	} else {
		/* Copy last packet onto the packet to be removed */
		memcpy(p + pkt_index, p + (*nb_pkt) - 1, sizeof(sx1301ar_rx_pkt_t));
		/* Update rsig arrays accordingly */
		p_rsig[pkt_index].chan = p_rsig[(*nb_pkt) - 1].chan;
		p_rsig[pkt_index].rssi = p_rsig[(*nb_pkt) - 1].rssi;
		p_rsig[pkt_index].snr = p_rsig[(*nb_pkt) - 1].snr;
	}

	*nb_pkt -= 1;

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int history_save(uint8_t brd, sx1301ar_rx_pkt_t *p, uint8_t nb_pkt)
{
	int i, j;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (p == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (nb_pkt == 0) {
		return 0;
	}

	/* reset */
	rxpk_history_nb = 0;
	memset(&rxpk_history[brd], 0, sizeof rxpk_history[brd]);

	/* save */
	for (i = 0; i < nb_pkt; i++) {
		for (j = 0; j < SX1301AR_BOARD_RFCHAIN_NB; j++) {
			if (p[i].rsig[j].is_valid == true) {
				rxpk_history[brd][rxpk_history_nb].count_us = p[i].count_us;
				rxpk_history[brd][rxpk_history_nb].modrate = p[i].modrate;
				rxpk_history[brd][rxpk_history_nb].chan = p[i].rsig[j].chan;
				rxpk_history_nb += 1;
			}
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int history_check(uint8_t brd, sx1301ar_rx_pkt_t *p, pkt_rsig_t *p_rsig, uint8_t *nb_pkt)
{
	int i, j;
	sx1301ar_rx_pkt_t pkt_dummy;
	pkt_rsig_t p_rsig_dummy;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}

	/* Run through the given packet array */
	for (i = 0; i < *nb_pkt; i++) {
		/* Compare with each record of the history to find match */
		for (j = 0; j < rxpk_history_nb; j++) {
			pkt_dummy.count_us = rxpk_history[brd][j].count_us;
			pkt_dummy.modrate = rxpk_history[brd][j].modrate;
			p_rsig_dummy.chan = rxpk_history[brd][j].chan;

			/* If the same packet has already been sent, discard it */
			if (is_same_pkt(&p[i], &p_rsig[i], &pkt_dummy, &p_rsig_dummy)) {
				debug("HISTORY: found ghost packet at index %d : tmst=%u, chan=%u, SF=%d", i, p[i].count_us, p_rsig[i].chan, sx1301ar_sf_enum2nb(p[i].modrate));
				remove_pkt(p, p_rsig, nb_pkt, i);
			}
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static double sx1301ar_i_lora_time_on_air(sx1301ar_bandw_t pkt_bw, sx1301ar_modr_t pkt_modr, uint8_t pkt_size, sx1301ar_coder_t pkt_cr, bool no_header, uint16_t preamb_size, double *t_preamble, double *t_syncword, double *t_payload)
{
	int val_i;
	long val_l;
	uint8_t SF, H, DE;
	uint16_t BW;
	uint32_t payloadSymbNb;
	double Tsym, Tpreamble, Tsyncword, Tpayload, Tpacket;

	/* Get bandwidth value */
	val_l = sx1301ar_bw_enum2nb(pkt_bw);
	if (val_l != -1) {
		BW = (uint16_t)(val_l / 1E3);
	} else {
		error("Cannot compute time on air for this packet, unsupported bandwidth (0x%02X)", pkt_bw);
		return 0;
	}

	/* Get modulation rate value */
	val_i = sx1301ar_sf_enum2nb(pkt_modr);
	if (val_i != -1) {
		SF = (uint8_t)val_i;
	} else {
		error("Cannot compute time on air for this packet, unsupported datarate (0x%02X)", pkt_modr);
		return 0;
	}

	/* Duration of 1 symbol */
	Tsym = pow(2, SF) / BW;

	/* Duration of preamble */
	Tpreamble = (double)preamb_size * Tsym;

	/* Duration of syncword */
	Tsyncword = 4.25 * Tsym;

	/* Duration of payload */
	H = (no_header == false) ? 0 : 1; /* header is always enabled, except for beacons */
	DE = (SF >= 11) ? 1 : 0;	  /* Low datarate optimization enabled for SF11 and SF12 */

	payloadSymbNb = 8 + ceil((double)(8 * pkt_size - 4 * SF + 28 + 16 - 20 * H) / (double)(4 * (SF - 2 * DE))) * (pkt_cr + 4); /* Explicitely cast to double to keep precision of the division */

	Tpayload = payloadSymbNb * Tsym;

	Tpacket = Tpreamble + Tsyncword + Tpayload;

	/* Return values */
	if (t_preamble != NULL) {
		*t_preamble = Tpreamble;
	}
	if (t_syncword != NULL) {
		*t_syncword = Tsyncword;
	}
	if (t_payload != NULL) {
		*t_payload = Tpayload;
	}

	return Tpacket;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static double sx1301ar_i_fsk_time_on_air(uint8_t brd, sx1301ar_modr_t pkt_modr, uint16_t preamb_size, uint8_t pkt_size, bool no_crc)
{
	double Tpacket;

	Tpacket = (8 * (double)(preamb_size + brd_cfg_priv[brd].fsk_sync_size + 1 + pkt_size + ((no_crc == true) ? 0 : 2)) / (double)pkt_modr) * 1E3;
	Tpacket = Tpacket + 1; /* rounding with margin */

	return Tpacket;
}

/* --- EOF ------------------------------------------------------------------ */
