/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Hardware Abstraction Layer for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

#ifndef _SX1301_ARRAY_HAL_H
#define _SX1301_ARRAY_HAL_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC CONSTANTS ----------------------------------------------------- */

#define SX1301AR_MAX_BOARD_NB 4		 /* library can support board index from 0 to 1 */
#define SX1301AR_BOARD_RFCHAIN_NB 2      /* number of RF chains per board */
#define SX1301AR_BOARD_CHIPS_NB 6	/* number of SX1301 chips per board */
#define SX1301AR_BOARD_NB_CHIP_PER_DSP 2 /* number of SX1301 chips per DSP on the board */
#define SX1301AR_BOARD_VERSION_MASTER 61 /* Master board FPGA version */
#define SX1301AR_BOARD_VERSION_SLAVE 8   /* Slave board FPGA version */
#define SX1301AR_BOARD_AES_KEY_SIZE 128  /* AES key size in bits */
#define SX1301AR_BOARD_AES_DATA_SIZE 16  /* AES data size in bytes */
#define SX1301AR_BOARD_MAX_LUT_NB 32     /* maximum size of the TX gain LUT */
#define SX1301AR_BOARD_MASTER 0		 /* index of the master board (with radio) */

#define SX1301AR_CHIP_MULTI_NB 8      /* number of LoRa multi-SF RX channel(s) per chip */
#define SX1301AR_CHIP_LSA_IDX 8       /* index of first LoRa stand-alone RX channel */
#define SX1301AR_CHIP_FSK_IDX 9       /* index of first FSK RX channel */
#define SX1301AR_CHIP_CHAN_NB 10      /* number of RX channels per chip */
#define SX1301AR_CHIP_FIFO_SIZE 8     /* RX FIFO depth of the SX1301 */
#define SX1301AR_CHIP_RXBUF_SIZE 1024 /* SX1301 RX data buffer size */

#define SX1301AR_MAX_PKT_NB (16 * SX1301AR_BOARD_CHIPS_NB) /* Maximum number of packet to be fetched */
#define SX1301AR_XTAL_FREQU 32000000			   /* frequency in Hz of the RF reference oscillator */

#define SX1301AR_MIN_FREQ (uint32_t)(10E6) /* if something is set bellow 10 MHz, it's probably an error (eg. value in kHz or MHz) */

#define SX1301AR_GPS_TIMER_CHIP 0      /* chip that has a GPS-latched timestamp counter */
#define SX1301AR_FREE_RUN_TIMER_CHIP 1 /* chip that has a free-running timestamp counter */

#define SX1301AR_DEFAULT_RSSI_OFFSET -202.5

#define SX1301AR_DEFAULT_ROOM_TEMP_REF 22	/* Reference room temperature Tref */
#define SX1301AR_DEFAULT_AD9361_TEMP_REF 77      /* AD9361 temperature returned for room at Tref */
#define SX1301AR_DEFAULT_AD9361_TEMP_SLOPE 909.0 /* Slope to convert radio temperature to room temp in °C (x1024) */

#define SX1301AR_LBT_CHANNEL_NB_MAX 16 /* Maximum number of LBT channels */

#define SX1301AR_DEFAULT_FTS_VERSION 1 /* The default version of the encrypted/main fine timestamp (for FPGA >= v59) */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

/* SPI callbacks function pointer definition
    address  SPI register address (between 0 and 127)
    status   Pointer to return 8-bit 'status' value (first byte read); NULL = don't care about status
    size     Size of the data to transfer, use 0 to only read status
    data     Pointer to a uint8_t buffer to read or write data
The functions must return 0 if operation was a success, -1 otherwise
/!\ WARNING: types also defined in sx1301ar_reg.h /!\
*/
typedef int (*spi_r_t)(uint8_t header, uint16_t address, uint8_t *data, uint32_t size, uint8_t *status);
typedef int (*spi_w_t)(uint8_t header, uint16_t address, const uint8_t *data, uint32_t size, uint8_t *status);

/**
@brief Status of the received packet regarding error correction
*/
typedef enum {
	STAT_UNDEFINED = 0,
	STAT_NO_CRC,
	STAT_CRC_ERR,
	STAT_CRC_OK
} sx1301ar_pstat_t;

/**
@brief Types of modulation supported by the modems in RX an TX
*/
typedef enum {
	MOD_UNDEFINED = 0,
	MOD_LORA,
	MOD_FSK
} sx1301ar_modul_t;

/**
@brief Modulation rate (SF for LoRa, bitrate for FSK)
*/
typedef enum {
	MR_UNDEFINED = 0,
	MR_SF7 = 0x02,
	MR_SF8 = 0x04,
	MR_SF9 = 0x08,
	MR_SF10 = 0x10,
	MR_SF11 = 0x20,
	MR_SF12 = 0x40,
	MR_SF7_8 = MR_SF7 | MR_SF8,
	MR_SF7_9 = MR_SF7 | MR_SF8 | MR_SF9,
	MR_SF7_10 = MR_SF7 | MR_SF8 | MR_SF9 | MR_SF10,
	MR_SF7_11 = MR_SF7 | MR_SF8 | MR_SF9 | MR_SF10 | MR_SF11,
	MR_SF7_12 = MR_SF7 | MR_SF8 | MR_SF9 | MR_SF10 | MR_SF11 | MR_SF12,
	MR_SF8_9 = MR_SF8 | MR_SF9,
	MR_SF8_10 = MR_SF8 | MR_SF9 | MR_SF10,
	MR_SF8_11 = MR_SF8 | MR_SF9 | MR_SF10 | MR_SF11,
	MR_SF8_12 = MR_SF8 | MR_SF9 | MR_SF10 | MR_SF11 | MR_SF12,
	MR_SF9_10 = MR_SF9 | MR_SF10,
	MR_SF9_11 = MR_SF9 | MR_SF10 | MR_SF11,
	MR_SF9_12 = MR_SF9 | MR_SF10 | MR_SF11 | MR_SF12,
	MR_SF10_11 = MR_SF10 | MR_SF11,
	MR_SF10_12 = MR_SF10 | MR_SF11 | MR_SF12,
	MR_SF11_12 = MR_SF11 | MR_SF12,
	MR_300 = 300,
	MR_600 = 600,
	MR_1000 = 1000,
	MR_1K = 1000,
	MR_1200 = 1200,
	MR_1K2 = 1200,
	MR_2400 = 2400,
	MR_2K4 = 2400,
	MR_4800 = 4800,
	MR_4K8 = 4800,
	MR_9600 = 9600,
	MR_9K6 = 9600,
	MR_10000 = 10000,
	MR_10K = 10000,
	MR_14400 = 14400,
	MR_14K4 = 14400,
	MR_19200 = 19200,
	MR_19K2 = 19200,
	MR_28800 = 28800,
	MR_28K8 = 28800,
	MR_38400 = 38400,
	MR_38K4 = 38400,
	MR_56000 = 56000,
	MR_56K = 56000,
	MR_57600 = 57600,
	MR_57K6 = 57600,
	MR_64000 = 64000,
	MR_64K = 64000,
	MR_100000 = 100000,
	MR_100K = 100000,
	MR_115200 = 115200,
	MR_115K2 = 115200,
	MR_128000 = 128000,
	MR_128K = 128000,
	MR_230400 = 230400,
	MR_230K4 = 230400,
	MR_250000 = 250000,
	MR_250K = 250000
} sx1301ar_modr_t;

/**
@brief RX bandwidth for FSK, RX/TX modulation bandwidth for LoRa
/!\ directly encode FSK RX bandwidth, do not change indexes
*/
typedef enum {
	BW_UNDEFINED = 0,
	BW_500K = 1,
	BW_250K = 2,
	BW_125K = 3,
	BW_62K5 = 4,
	BW_31K2 = 5,
	BW_15K6 = 6,
	BW_7K8 = 7
} sx1301ar_bandw_t;

/**
@brief Error correction level for LoRa modulation
*/
typedef enum {
	CR_UNDEFINED = 0,
	CR_4_5 = 1,
	CR_4_6 = 2,
	CR_2_3 = CR_4_6,
	CR_4_7 = 3,
	CR_4_8 = 4,
	CR_1_2 = CR_4_8
} sx1301ar_coder_t;

/**
@brief TX mode, select event that triggers packet emission
*/
typedef enum {
	TX_IMMEDIATE,
	TX_TIMESTAMPED,
	TX_ON_GPS
} sx1301ar_tmode_t;

/**
@brief Status of the TX modem(s)
*/
typedef enum {
	TX_STATUS_UNKNOWN = 0,
	TX_OFF,
	TX_FREE,
	TX_SCHEDULED,
	TX_EMITTING
} sx1301ar_tstat_t;

/**
@brief Status of the board (ie. the whole concentrator)
*/
typedef enum {
	BRD_STATUS_UNKNOWN = 0,
	BRD_STOPPED,
	BRD_STARTED
} sx1301ar_bstat_t;

/**
@brief Type of the board
*/
typedef enum {
	BRD_TYPE_UNKNOWN = 0,
	BRD_MASTER,
	BRD_SLAVE
} sx1301ar_btype_t;

/**
@brief Structure containing the different TX available gains for AD9361 radio
*/
typedef struct
{
	uint16_t atten;
	uint8_t auxdac_vref;
	uint16_t auxdac_word;
} sx1301ar_ad9361_tx_gain_t;

/**
@brief Structure containing the coefficients to be used to calculate the temperature compensation to be applied
*/
typedef struct
{
	int16_t coeff_a;
	int16_t coeff_b;
} sx1301ar_ad9361_temperature_comp_t;

/**
@brief Structure containing the different TX gains to be applied to get the wanted RF power
*/
typedef struct
{
	int8_t rf_power;
	uint8_t fpga_dig_gain;
	sx1301ar_ad9361_tx_gain_t ad9361_gain;
	sx1301ar_ad9361_temperature_comp_t ad9361_tcomp;
} sx1301ar_tx_gain_t;

/**
@brief Structure containing the table of TX gains
*/
typedef struct
{
	sx1301ar_tx_gain_t lut[SX1301AR_BOARD_MAX_LUT_NB]; /*!> TX gain table */
	uint8_t size;					   /*!> Size of the TX gain table */
} sx1301ar_tx_gain_lut_t;

/**
@brief Structure containing a radio chain configuration parameters
*/
typedef struct
{
	bool rx_enable;
	bool tx_enable;
	double antenna_gain;
	double insertion_loss;
	float rssi_offset;
	int16_t rssi_offset_coeff_a;
	int16_t rssi_offset_coeff_b;
	sx1301ar_tx_gain_lut_t tx_lut;
} sx1301ar_rfchain_t;

/**
@brief Structure containing the user configuration for a SX1301 array board
*/
typedef struct
{
	spi_r_t spi_read;					/*!> SPI read callback */
	spi_w_t spi_write;					/*!> SPI write callback */
	sx1301ar_btype_t board_type;				/*!> Type of the board: Master or slave */
	uint32_t rx_freq_hz;					/*!> Center RX frequency of the radio, in Hz */
	uint32_t rx_bw_hz;					/*!> RX bandwidth of the radio, in Hz */
	bool full_duplex;					/*!> full-duplex / half-duplex radio mode */
	sx1301ar_rfchain_t rf_chain[SX1301AR_BOARD_RFCHAIN_NB]; /*!> radio chains configuration */
	uint64_t fsk_sync_word;					/*!> FSK sync word (ALIGN RIGHT, eg. 0xC194C1) */
	uint8_t fsk_sync_size;					/*!> size of FSK sync word (in bytes, 0 = use default sync word) */
	bool loramac_public;					/*!> Enable ONLY for *public* networks using the LoRa MAC protocol */
	uint8_t dsp_stat_interval;
	bool fine_timestamp_decrypt_enable;		  /*!> Fine timestamp decryption enabled or not */
	uint8_t aes_key[SX1301AR_BOARD_AES_KEY_SIZE / 8]; /*!> AES-128 decryption key associated with the board */
	int8_t room_temp_ref;				  /*!> Reference room temperature Tref used for calibration (°C) */
	uint8_t ad9361_temp_ref;			  /*!> Temperature code returned by AD9361 radio when room temperature is Tref [0..255]*/
	uint8_t nb_chip;				  /*!> Number of SX1301 chips per board */
	uint8_t nb_dsp;					  /*!> Number of DSP per board */
	bool match_tmst_crc_err;			  /*!> Enable fine timestamp matching with packets with CRC error */
	uint8_t main_tmst_version;			  /*!> Version of the encrypted/main fine timestamp to be sent by the DSP */
	bool debug_tmst;				  /*!> Enable fine timestamp debug mode in DSP */
} sx1301ar_board_cfg_t;

/**
@brief Structure containing the user configuration for a SX1301
*/
typedef struct
{
	bool enable;      /*!> Enable or disable that chip */
	uint32_t freq_hz; /*!> Center RX frequency of the SX1301 chip, in Hz */
	uint8_t rf_chain; /*!> Select Rx RF chain */
} sx1301ar_chip_cfg_t;

/**
@brief Structure containing the user configuration for a SX1301 receive channel
*/
typedef struct
{
	bool enable;		    /*!> Enable or disable that RX channel and IF chain */
	uint32_t freq_hz;	   /*!> Center RX frequency of the channel, in Hz */
	sx1301ar_modr_t modrate;    /*!> RX modulation rate, 0 for default */
	sx1301ar_bandw_t bandwidth; /*!> Channel bandwidth */
} sx1301ar_chan_cfg_t;

/**
@brief Structure containing a Listen-Before-Talk channel configuration
*/
typedef struct
{
	uint32_t freq_hz;	  /*!> LBT channel frequency */
	uint16_t scan_time_us;     /*!> LBT channel carrier sense time */
	uint16_t transmit_time_ms; /*!> LBT channel transmission duration when allowed */
} sx1301ar_lbt_chan_cfg_t;

/**
@brief Structure containing the Listen-Before-Talk configuration for a board
*/
typedef struct
{
	bool enable;						       /*!> enable or disable LBT */
	int8_t rssi_target;					       /*!> RSSI threshold to detect if channel is busy or not (dBm) */
	uint8_t rssi_shift;					       /*!> value to be applied to RSSI to convert RSSI in dBm to register value */
	uint8_t nb_channel;					       /*!> number of LBT channels */
	sx1301ar_lbt_chan_cfg_t channels[SX1301AR_LBT_CHANNEL_NB_MAX]; /*!> LBT channels configuration */
} sx1301ar_lbt_cfg_t;

/**
@brief Structure containing the signal information of the received packet
*/
typedef struct
{
	bool is_valid;					     /*!> Is there a signal on this antenna? */
	bool fine_received;				     /*!> Have we received a valid fine timestamp? */
	bool sig_info_received;				     /*!> Have we received signal information from DSP (RSSI, SNR...)? */
	uint8_t chan;					     /*!> Channel on which packet was received */
	int16_t freq_offset;				     /*!> Frequency offset, in Hz */
	uint8_t fine_tmst_enc[SX1301AR_BOARD_AES_DATA_SIZE]; /*!> Main fine timestamp of packet arrival (encrypted) */
	uint32_t fine_tmst;				     /*!> Main fine timestamp of packet arrival (clear) */
	uint8_t fine_tmst_status;			     /*!> Main fine timestamp status */
	uint8_t fine_tmst_version;			     /*!> Version of the main fine timestamp */
	float rssi_chan;				     /*!> Channel RSSI in dB */
	float rssi_sig;					     /*!> Signal RSSI in dB */
	uint32_t rssi_sig_std;				     /*!> Standard deviation of RSSI during preamble */
	float snr;					     /*!> Average packet SNR, in dB (LoRa only) */
	int16_t fine_tmst_alt;				     /*!> Alternative fine timestamp: delta in nanoseconds compared to main fine timestamp */
	uint16_t fine_tmst_debug1;			     /*!> Fine timestamp debug info 1 */
	uint16_t fine_tmst_debug2;			     /*!> Fine timestamp debug info 2 */
} gwloc_rx_pkt_rsig_t;

/**
@brief Structure containing the metadata and the payload of a received packet
*/
typedef struct
{
	sx1301ar_pstat_t status;			     /*!> status of the received packet */
	uint32_t freq_hz;				     /*!> Center RX frequency of the channel, in Hz */
	uint32_t count_us;				     /*!> internal concentrator timestamping, in us */
	sx1301ar_modul_t modulation;			     /*!> modulation used by the packet */
	sx1301ar_bandw_t bandwidth;			     /*!> modulation bandwidth (LoRa only) */
	sx1301ar_modr_t modrate;			     /*!> RX modulation rate of the packet */
	sx1301ar_coder_t coderate;			     /*!> error-correcting code of the packet (LoRa only) */
	uint8_t size;					     /*!> received payload size, in bytes */
	uint8_t payload[255];				     /*!> buffer containing the payload */
	gwloc_rx_pkt_rsig_t rsig[SX1301AR_BOARD_RFCHAIN_NB]; /*!> received signal information */
} sx1301ar_rx_pkt_t;

/**
@brief Structure containing the configuration and the payload of a packet to send
*/
typedef struct
{
	sx1301ar_tmode_t tx_mode;    /*!> select on what event/time the TX is triggered */
	uint32_t count_us;	   /*!> target timestamp for TIMESTAMPED TX mode */
	uint32_t freq_hz;	    /*!> Center TX frequency, in Hz */
	int8_t rf_power;	     /*!> TX power, in dBm */
	uint8_t rf_chain;	    /*!> RF chain */
	sx1301ar_modul_t modulation; /*!> modulation used by the packet */
	sx1301ar_bandw_t bandwidth;  /*!> modulation bandwidth (LoRa only) */
	sx1301ar_modr_t modrate;     /*!> TX modulation rate of the packet */
	sx1301ar_coder_t coderate;   /*!> error-correcting code of the packet (LoRa only) */
	uint8_t f_dev;		     /*!> frequency deviation, in kHz (FSK only) */
	uint16_t preamble;	   /*!> set the preamble length, 0 for default */
	bool invert_pol;	     /*!> invert signal polarity (LoRa only) */
	bool no_crc;		     /*!> if true, do not send a CRC in the packet */
	bool no_header;		     /*!> if true, enable implicit header mode (LoRa), fixed length (FSK) */
	uint8_t size;		     /*!> payload size to send, in bytes */
	uint8_t payload[255];	/*!> buffer containing the payload */
	uint32_t message_id;
	uint32_t transmission_id;
} sx1301ar_tx_pkt_t;

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */

/**
@warning This module has internal states: for each board are stored all the
variables settings that are applied to the hardware when the _start() function
is called.
*/

/**
@brief Get library version information (including config options)
@param brd Target board number
@param fpga_version pointer to an integer containing FPGA version
@param dsp_version pointer to an integer containing DSP version
@return pointer to a string describing version informations of the library
*/
const char *sx1301ar_version_info(uint8_t brd, int16_t *fpga_version, int16_t *dsp_version);

/*
    Configuration functions
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Affect the internal configuration state of the library.
    Check to consistency of user-set parameters.
*/

/**
@brief Set the configuration of the top level of a SX1301 array board
@param brd  Target board number
@param cfg  Pointer to configuration structure for the top level of a SX1301 array board
@return 0 if the board configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_board(uint8_t brd, const sx1301ar_board_cfg_t *cfg);

/**
@brief Get the configuration of the top level of a SX1301 array board
@param brd  Target board number
@param cfg  Pointer to the current configuration structure for the top level of a SX1301 array board
@return 0 if the board configuration was valid and coherent, -1 else
*/
int sx1301ar_get_conf_board(uint8_t brd, sx1301ar_board_cfg_t *cfg);

/**
@brief Set the configuration of a SX1301 chip in an array
@param brd   Target board number
@param chip  Target chip number in the array
@param cfg   Pointer to configuration structure for a SX1301 chip
@return 0 if the chip configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_chip(uint8_t brd, uint8_t chip, const sx1301ar_chip_cfg_t *cfg);

/**
@brief Set the configuration of a channel of a SX1301 in an array
@param brd   Target board number
@param chan  Target channel number (the 4 MSb are the chip number)
@param cfg   Pointer to configuration structure for a RX channel
@return 0 if the channel configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_chan(uint8_t brd, uint8_t chan, const sx1301ar_chan_cfg_t *cfg);

/**
@brief Set the configuration of Listen-Before-Talk
@param conf structure containing the configuration parameters
@return 0 if the LBT configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_lbt(uint8_t brd, const sx1301ar_lbt_cfg_t *cfg);

/*
    Core functions
    ~~~~~~~~~~~~~~~

    Interact with the hardware concentrator board.
*/

/**
@brief Connect, configure and start all boards
@param nb_brd  Number of target boards to be initialized
@return 0 if all SX1301 array boards could be started (and was not already started), -1 else
*/
int sx1301ar_start(uint8_t nb_brd);

/**
@brief Stop and disconnect all boards
@param nb_brd  Number of target boards to be stopped
@return 0 if all SX1301 array boards were stopped (and was not already stopped), -1 else
*/
int sx1301ar_stop(uint8_t nb_brd);

/**
@brief Fetch all packets waiting in the SX1301 FIFOs, up to a specified number
@param brd     Target board number
@param p       Pointer to array for receiving packet data and metadata
@param max_nb  Maximum number of packets to fetch (0 is accepted)
@param nb_pkt  Pointer to a uint8_t to receive number of returned packets
@return 0 if RX FIFO was accessed, whether packets were waiting or not, -1 else
*/
int sx1301ar_fetch(uint8_t brd, sx1301ar_rx_pkt_t *p, uint8_t max_nb, uint8_t *nb_pkt);

/**
@brief Send a packet to a TX modem for immediate or delayed sending
@param brd   Target board number
@param p     Pointer to TX parameters and payload of the packet to be sent
@return 0 if packet was scheduled for TX, -1 else

/!\ When sending a packet, there is a 1.5 ms delay for the analog circuitry to start and be stable (TX_START_DELAY).
In 'timestamp' mode, this is transparent: the modem is started 1.5ms before the user-set timestamp value is reached, the preamble of the packet start right when the internal timestamp counter reach target value.
In 'immediate' mode, the packet is emitted as soon as possible: transferring the packet (and its parameters) from the host to the concentrator takes some time, then there is the TX_START_DELAY, then the packet is emitted.
In 'triggered' mode (aka PPS/GPS mode), the packet, typically a beacon, is emitted 1.5ms after a rising edge of the trigger signal. Because there is no way to anticipate the triggering event and start the analog circuitry beforehand, that delay must be taken into account in the protocol.
*/
int sx1301ar_send(uint8_t brd, const sx1301ar_tx_pkt_t *p);

/**
@brief Check if TX modem is free, if a packet is scheduled or being emitted
@param brd   Target board number
@param s     Pointer to return the value of the TX hardware state-machine
@return 0 if the TX modem status could be determined, -1 else

This is a non-blocking function, it returns as soon as the packet is scheduled
in the TX modem, not when the packet has effectively been emitted.
*/
int sx1301ar_tx_status(uint8_t brd, sx1301ar_tstat_t *s);

/**
@brief Abort a currently scheduled or ongoing TX
@param brd   Target board number
@return 0 if the TX modem could be stopped, -1 else
*/
int sx1301ar_abort_tx(uint8_t brd);

/**
@brief Check if the target board is started or not
@param brd  Target board number
@param s    Pointer to return the value of the board status
@return 0 if the board status could be determined, -1 else
*/
int sx1301ar_brd_status(uint8_t brd, sx1301ar_bstat_t *s);

/**
@brief Return instantaneous value of internal timestamp counter
@param brd     Target board number
@param cnt_us  Pointer to receive timestamp value
@return 0 if the instantaneous timestamp counter was returned correctly, -1 else
*/
int sx1301ar_get_instcnt(uint8_t brd, uint32_t *cnt_us);

/**
@brief Return value of internal timestamp counter when latest event (eg GPS pulse) was captured
@param brd     Target board number
@param cnt_us  Pointer to receive timestamp value
@return 0 if the latched timestamp counter was returned correctly, -1 else
*/
int sx1301ar_get_trigcnt(uint8_t brd, uint32_t *cnt_us);

/**
@brief Return value of internal high speed timestamp counter when latest event (eg GPS pulse) was captured
@param brd     Target board number
@param cnt     Pointer to receive timestamp value
@return 0 if the latched timestamp counter was returned correctly, -1 else
*/
int sx1301ar_get_trighs(uint8_t brd, uint32_t *cnt_hs);

/**
@brief Return current temperature measured by the radio temperature sensor
@param brd          Target board number
@param temp_code    Pointer to received temperature code value (from AD9361 radio)
@param temp_celsius Pointer to received temperature value in °C (converted from the temperature code, using Tref values)
@return 0 if successfully read, -1 otherwise
*/
int sx1301ar_get_temperature(uint8_t brd, uint8_t *temp_code, int16_t *temp_celsius);

/**
@brief Return time on air of given RX packet, in milliseconds
@param
@param
@return the packet time on air in milliseconds
*/
double sx1301ar_rx_pkt_time_on_air(const sx1301ar_rx_pkt_t *pkt, double *t_preamble, double *t_syncword, double *t_payload);

/**
@brief Return time on air of given TX packet, in milliseconds
@param brd          Target board number
@param
@param
@return the packet time on air in milliseconds
*/
double sx1301ar_tx_pkt_time_on_air(uint8_t brd, const sx1301ar_tx_pkt_t *pkt, double *t_preamble, double *t_syncword, double *t_payload);

/*
    Helper functions
    ~~~~~~~~~~~~~~~~~

    Assist for internal types manipulation.
    Initialize structures defined by the library.
    Covert to/from abstract (enum) types.
*/

/**
@brief Return a sx1301ar_board_cfg structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_board_cfg_t sx1301ar_init_board_cfg(void);

/**
@brief Return a sx1301ar_chip_cfg structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_chip_cfg_t sx1301ar_init_chip_cfg(void);

/**
@brief Return a sx1301ar_chan_cfg structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_chan_cfg_t sx1301ar_init_chan_cfg(void);

/**
@brief Return a sx1301ar_lbt_cfg structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_lbt_cfg_t sx1301ar_init_lbt_cfg(void);

/**
@brief Return a sx1301ar_tx_pkt structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_tx_pkt_t sx1301ar_init_tx_pkt(void);

/**
@brief Return a sx1301ar_tx_gain structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_tx_gain_t sx1301ar_init_tx_gain(void);

/**
@brief Return a sx1301ar_tx_gain_lut structure initialized to safe default parameters
@return Structure with safe default parameters
*/
sx1301ar_tx_gain_lut_t sx1301ar_init_tx_gain_lut(void);

/**
@brief Transform SF value from number to enum modrate format
@param x  SF as an unsigned number between 7 and 12
@return SF in the format define by the 'modrate' enumeration (MR_UNDEFINED for error)
*/
sx1301ar_modr_t sx1301ar_sf_nb2enum(int x);

/**
@brief Transform SF value from enum modrate format to number
@param x  SF in the format define by the 'modrate' enumeration
@return SF as an unsigned number between 7 and 12 (-1 for error)
*/
int sx1301ar_sf_enum2nb(sx1301ar_modr_t x);

/**
@brief Transform two SF values from number to enum modrate format
@param a  SF as an unsigned number between 7 and 12, defining an SF range
@param b  SF as an unsigned number between 7 and 12, defining an SF range
@return SF in the format define by the 'modrate' enumeration (MR_UNDEFINED for error)
*/
sx1301ar_modr_t sx1301ar_sf_range_nb2enum(int a, int b);

/**
@brief Transform a SF range enum (modrate format) to min SF number
@param x  SF range in the format define by the 'modrate' enumeration
@return minimum SF as an unsigned number between 7 and 12 (-1 for error)
*/
int sx1301ar_sf_min_enum2nb(sx1301ar_modr_t x);

/**
@brief Transform a SF range enum (modrate format) to max SF number
@param x  SF range in the format define by the 'modrate' enumeration
@return maximum SF as an unsigned number between 7 and 12 (-1 for error)
*/
int sx1301ar_sf_max_enum2nb(sx1301ar_modr_t x);

/**
@brief Transform CR value from number to enum coderate format
@param x  CR as a number between 1 and 4, following Semtech LoRa convetion
@return CR in the format define by the 'coderate' enumeration (CR_UNDEFINED for error)
*/
sx1301ar_coder_t sx1301ar_cr_nb2enum(int x);

/**
@brief Transform CR value from enum coderate format to number
@param x  CR in the format define by the 'coderate' enumeration
@return CR as a number between 1 and 4, following Semtech LoRa convention (-1 for error)
*/
int sx1301ar_cr_enum2nb(sx1301ar_coder_t x);

/**
@brief Transform CR value from enum coderate format to ASCII string
@param x  CR in the format define by the 'coderate' enumeration
@return CR as a 3-characters string representing coding ratio ("???" for error)

The string value represent the radio of 'useful' data over the size of the
payload after encoding. Eg. 4/8 mean that every useful bit is represented as
two bits once encoded.
*/
const char *sx1301ar_cr_enum2str(sx1301ar_coder_t x);

/**
@brief Transform BW value from number to enum coderate format
@param x  bandwidth in hertz (will be rounded to closest available BW above)
@return BW in the format define by the 'bandwidth' enumeration (BW_UNDEFINED for error)
*/
sx1301ar_bandw_t sx1301ar_bw_nb2enum(long x);

/**
@brief Transform CR value from enum coderate format to number
@param x  BW in the format define by the 'bandwidth' enumeration
@return bandwidth in hertz (-1 for error)
*/
long sx1301ar_bw_enum2nb(sx1301ar_bandw_t x);

/**
@brief Transfor a temperature code given by AD9361 radio to corresponding attenuation correction
@brief      atten_correction = a * (T² - Tref²) + b * (T - Tref)
@param T    AD9361 temperature code to be converted
@param Tref AD9361 reference temperature code used for calibration
@param a    First coefficient to be used in conversion polynome
@param b    Second coefficient to be used in conversion polynome
@return attenuation correction to be applied to AD9361 atten for given temperature.
*/
int8_t sx1301ar_ad9361_temp2atten(uint8_t T, uint8_t Tref, int16_t a, int16_t b);

/**
@brief Transfor a temperature code given by AD9361 radio to corresponding RSSI correction
@brief      rssi_correction = a * (T - Tref)² + b * (T - Tref)
@param T    AD9361 temperature code to be converted
@param Tref AD9361 reference temperature code used for calibration
@param a    First coefficient to be used in conversion polynome
@param b    Second coefficient to be used in conversion polynome
@return RSSI correction to be applied.
*/
float sx1301ar_ad9361_temp2rssi(uint8_t T, uint8_t Tref, int16_t a, int16_t b);

/**
@brief Set the configuration of TX gain table
@param brd      Target board number
@param rf_chain Target RF chain number
@param cfg      Pointer to configuration structure for TX gain table
@return 0 if the configuration was valid and coherent, -1 else
*/
int sx1301ar_conf_tx_gain(uint8_t brd, uint8_t rf_chain, const sx1301ar_tx_gain_lut_t *cfg);

#endif

/* --- EOF ------------------------------------------------------------------ */
