/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    I2C functions using Linux device driver

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Semtech
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <errno.h>
#include <fcntl.h>  /* open */
#include <stdint.h> /* C99 types */
#include <string.h> /* memset */
#include <unistd.h> /* lseek, close */

#include <linux/i2c-dev.h>
#include <linux/i2c.h>
#include <sys/ioctl.h>

#include "i2c.h"

#include "utils/logger.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#define I2C_SUCCESS 0
#define I2C_ERROR -1

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int i2c_linuxdev_open(const char *path, uint8_t device_addr, int *i2c_ref)
{
	int dev;

	/* Check input variables */
	if (path == NULL) {
		error("null pointer path");
		return I2C_ERROR;
	}
	if (i2c_ref == NULL) {
		error("null pointer i2c_ref");
		return I2C_ERROR;
	}

	/* Open I2C device */
	dev = open(path, O_RDWR);
	if (dev < 0) {
		error("Failed to open I2C %s - %s", path, strerror(errno));
		return I2C_ERROR;
	}

	/* Setting I2C device mode to slave */
	if (ioctl(dev, I2C_SLAVE, device_addr) < 0) {
		error("Failed to acquire bus access and/or talk to slave - %s", strerror(errno));
		return I2C_ERROR;
	}

	debug("I2C port opened successfully");
	*i2c_ref = dev; /* return file descriptor index */

	return I2C_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int i2c_linuxdev_close(int i2c_ref)
{
	int i;

	i = close(i2c_ref);

	if (i == 0) {
		debug("I2C port closed successfully");
		return I2C_SUCCESS;
	} else {
		error("Failed to close I2C - %s", strerror(errno));
		return I2C_ERROR;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int i2c_linuxdev_read_reg(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t *data, uint16_t nb_bytes)
{
	uint8_t *inbuff, outbuff;
	struct i2c_rdwr_ioctl_data packets;
	struct i2c_msg messages[2];

	outbuff = reg_addr;
	messages[0].addr = device_addr;
	messages[0].flags = 0;
	messages[0].len = sizeof(outbuff);
	messages[0].buf = &outbuff;

	inbuff = data;
	messages[1].addr = device_addr;
	messages[1].flags = I2C_M_RD;
	messages[1].len = nb_bytes;
	messages[1].buf = inbuff;

	packets.msgs = messages;
	packets.nmsgs = 2;

	if (ioctl(i2c_ref, I2C_RDWR, &packets) < 0) {
		error("Read from I2C Device failed (%d, 0x%02x, 0x%02x) - %s", i2c_ref, device_addr, reg_addr, strerror(errno));
		return I2C_ERROR;
	}

	return I2C_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int i2c_linuxdev_read(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t *data)
{
	uint8_t *inbuff, outbuff;
	struct i2c_rdwr_ioctl_data packets;
	struct i2c_msg messages[2];

	outbuff = reg_addr;
	messages[0].addr = device_addr;
	messages[0].flags = 0;
	messages[0].len = sizeof(outbuff);
	messages[0].buf = &outbuff;

	inbuff = data;
	messages[1].addr = device_addr;
	messages[1].flags = I2C_M_RD;
	messages[1].len = sizeof(*inbuff);
	messages[1].buf = inbuff;

	packets.msgs = messages;
	packets.nmsgs = 2;

	if (ioctl(i2c_ref, I2C_RDWR, &packets) < 0) {
		error("Read from I2C Device failed (%d, 0x%02x, 0x%02x) - %s", i2c_ref, device_addr, reg_addr, strerror(errno));
		return I2C_ERROR;
	}

	return I2C_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int i2c_linuxdev_write_burst(int i2c_ref, uint8_t device_addr, uint8_t *data, uint16_t nb_bytes)
{
	struct i2c_rdwr_ioctl_data packets;
	struct i2c_msg messages[1];

	messages[0].addr = device_addr;
	messages[0].flags = 0;
	messages[0].len = nb_bytes;
	messages[0].buf = data;

	packets.msgs = messages;
	packets.nmsgs = 1;

	if (ioctl(i2c_ref, I2C_RDWR, &packets) < 0) {
		error("Write to I2C Device failed (%d, 0x%02x) - %s", i2c_ref, device_addr, strerror(errno));
		return I2C_ERROR;
	}

	return I2C_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int i2c_linuxdev_write(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t data)
{
	unsigned char buff[2];
	struct i2c_rdwr_ioctl_data packets;
	struct i2c_msg messages[1];

	buff[0] = reg_addr;
	buff[1] = data;

	messages[0].addr = device_addr;
	messages[0].flags = 0;
	messages[0].len = sizeof(buff);
	messages[0].buf = buff;

	packets.msgs = messages;
	packets.nmsgs = 1;

	if (ioctl(i2c_ref, I2C_RDWR, &packets) < 0) {
		error("Write to I2C Device failed (%d, 0x%02x, 0x%02x) - %s", i2c_ref, device_addr, reg_addr, strerror(errno));
		return I2C_ERROR;
	}

	return I2C_SUCCESS;
}

/* --- EOF ------------------------------------------------------------------ */
