/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    I2C functions using Linux device driver

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Semtech
*/

#ifndef _I2C_LINUXDEV_H
#define _I2C_LINUXDEV_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdint.h> /* C99 types */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */

/**
@brief Open I2C port
@param path         Path to the I2C device driver (absolute or relative)
@param device_addr  Address of the device
@param i2c_ref      Pointer to receive I2C port file descriptor index
@return 0 if I2C port was open successfully, -1 else
*/
int i2c_linuxdev_open(const char *path, uint8_t device_addr, int *i2c_ref);

/**
@brief Close I2C port
@param spi_ref      I2C port file descriptor index
@return 0 if I2C port was closed successfully, -1 else
*/
int i2c_linuxdev_close(int i2c_ref);

/**
@brief Read data in a register from an I2C port
@param i2c_ref      I2C port file descriptor index
@param device_addr  I2C device address
@param reg_addr        Register to read
@param data         Pointer to a buffer to store read data
@param nb_bytes     Number of bytes to read
@return 0 if I2C data read is successful, -1 else
*/
int i2c_linuxdev_read_reg(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t *data, uint16_t nb_bytes);

/**
@brief Read data from an I2C port
@param i2c_ref      I2C port file descriptor index
@param device_addr  I2C device address
@param reg_addr     Address of the register to be read
@param data         Pointer to a buffer to store read data
@return 0 if I2C data read is successful, -1 else
*/
int i2c_linuxdev_read(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t *data);

/**
@brief Write burst of data to an I2C port
@param i2c_ref      I2C port file descriptor index
@param device_addr  I2C device address
@param data         Data to be written
@param nb_bytes     Number of bytes to be written
@return 0 if I2C data write is successful, -1 else
*/
int i2c_linuxdev_write_burst(int i2c_ref, uint8_t device_addr, uint8_t *data, uint16_t nb_bytes);

/**
@brief Write data to an I2C port
@param i2c_ref      I2C port file descriptor index
@param device_addr  I2C device address
@param reg_addr     Address of the register to write to
@param data         byte to write in the register
@return 0 if I2C data write is successful, -1 else
*/
int i2c_linuxdev_write(int i2c_ref, uint8_t device_addr, uint8_t reg_addr, uint8_t data);

#endif

/* --- EOF ------------------------------------------------------------------ */
