/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Functions used to handle the Listen Before Talk feature

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDANCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <stdio.h>   /* printf fprintf */
#include <stdlib.h>  /* abs, labs, llabs */

#include "aux.h"
#include "err.h"
#include "hal.h"
#include "lbt.h"
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE TYPES -------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS ---------------------------------------------------- */

#define LBT_DELAY 10000 /* in microseconds */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

extern void *lgw_spi_target;     /*! generic pointer to the SPI device */
extern uint8_t lgw_spi_mux_mode; /*! current SPI mux mode used */

static sx1301ar_lbt_cfg_t lbt_cfg[SX1301AR_MAX_BOARD_NB];

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

static bool is_equal_freq(uint32_t a, uint32_t b);

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int sx1301ar_i_conf_lbt(uint8_t brd, const sx1301ar_lbt_cfg_t *cfg)
{
	int i;

	/* set internal config according to parameters */
	lbt_cfg[brd].enable = cfg->enable;
	lbt_cfg[brd].rssi_target = cfg->rssi_target;
	lbt_cfg[brd].rssi_shift = cfg->rssi_shift;
	lbt_cfg[brd].nb_channel = cfg->nb_channel;

	if (lbt_cfg[brd].enable) {
		info("board #%zu: Using lbt: enable=true rssi_target=%i rssi_shift=%u",
		     brd,
		     lbt_cfg[brd].rssi_target,
		     lbt_cfg[brd].rssi_shift);
	} else {
		info("board #%zu: Using lbt: enable=false", brd);
		return 0;
	}

	for (i = 0; i < lbt_cfg[brd].nb_channel; i++) {
		lbt_cfg[brd].channels[i].freq_hz = cfg->channels[i].freq_hz;
		lbt_cfg[brd].channels[i].scan_time_us = cfg->channels[i].scan_time_us;
		lbt_cfg[brd].channels[i].transmit_time_ms = cfg->channels[i].transmit_time_ms;

		info("board #%zu: Using lbt: channel=%-2u freq_hz=%u scan_time_us=%u transmit_time_ms=%u",
		     brd,
		     i,
		     lbt_cfg[brd].channels[i].freq_hz,
		     lbt_cfg[brd].channels[i].scan_time_us,
		     lbt_cfg[brd].channels[i].transmit_time_ms);
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_i_init_lbt(uint8_t brd, uint32_t rx_freq)
{
	int x, i;
	int16_t if_reg;
	uint16_t scan_time_reg = 0x0;

	/* Configure FPGA for Listen-Before-Talk */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_FILTER_1_GAIN, 15);
	x |= sx1301ar_reg_brd_w(brd, BREG_LBT_FILTER_1_DC_NOTCH_EN, 0);
	x |= sx1301ar_reg_brd_w(brd, BREG_LBT_FILTER_2_GAIN, 13);
	SX1301AR_ERR_PROPAGATE(x);

	/* Select radio path to be used for LBT scanning */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_SELECT_RF_PATH, 0);
	SX1301AR_ERR_PROPAGATE(x);

	/* Set the time we want to wait for a stable RSSI measure before starting LBT scan */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_RSSI_START_DELAY, 500); /* µs */
	SX1301AR_ERR_PROPAGATE(x);

	/* Configure LBT scan times */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_RSSI_READ_DELAY, 10);      /* scan time = RSSI_READ_DELAY * READ_RSSI_TIMEx */
	x |= sx1301ar_reg_brd_w(brd, BREG_LBT_NB_READ_RSSI_TIME1, 13);  /* 128µs */
	x |= sx1301ar_reg_brd_w(brd, BREG_LBT_NB_READ_RSSI_TIME2, 500); /* 5000µs */
	SX1301AR_ERR_PROPAGATE(x);

	/* Configure LBT channels */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_NB_CHANNEL, lbt_cfg[brd].nb_channel - 1);
	SX1301AR_ERR_PROPAGATE(x);
	for (i = 0; i < lbt_cfg[brd].nb_channel; i++) {
		/* Set channel frequency */
		if_reg = chip_if_freq2reg((int32_t)rx_freq - (int32_t)(lbt_cfg[brd].channels[i].freq_hz), NULL);
		x = sx1301ar_reg_brd_w(brd, BREG_LBT_CHANNEL_0_IF_FREQ + i, if_reg);
		SX1301AR_ERR_PROPAGATE(x);

		/* Set channel scan time */
		switch (lbt_cfg[brd].channels[i].scan_time_us) {
		case 128:
			/* clean corresponding channel bit */
			scan_time_reg &= ~(1 << i);
			break;
		case 5000:
			/* set corresponding channel bit */
			scan_time_reg |= (1 << i);
			break;
		default:
			error("LBT scan time %u is not supported (channel %d)", lbt_cfg[brd].channels[i].scan_time_us, i);
			SX1301AR_ERR_RETURN(ERR_LBT_SCAN_TIME);
		}
	}
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_CHANNEL_TIME_MODE, (int32_t)scan_time_reg);
	SX1301AR_ERR_PROPAGATE(x);

	/* Enable all channels, and rely on NB_CHANNEL */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_CHANNEL_ENABLE, 0xFFFF);
	SX1301AR_ERR_PROPAGATE(x);

	/* Set LBT RSSI target threshold */
	x = sx1301ar_reg_brd_w(brd, BREG_LBT_RSSI_TARGET, (int32_t)(lbt_cfg[brd].rssi_target + lbt_cfg[brd].rssi_shift));
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_i_start_lbt(uint8_t brd)
{
	int x;
#if 0 /* FOR DEBUG */
    int32_t val, val2;

    /* Check that FPGA counter is in sync with SX1301 counter */
    x  = sx1301ar_reg_chip_r( brd, SX1301AR_GPS_TIMER_CHIP, CREG_TIMESTAMP, &val );
    x |= sx1301ar_reg_brd_r( brd, BREG_SX1301_TIMESTAMP_CPY_GPS, &val2 );
    SX1301AR_ERR_PROPAGATE( x );
    if( (val == 0) || (abs(val - val2) > 1) ) /* tolerate 1µs */
    {
        error( "timestamp counter out of sync (%i %i - %dµs)", val, val2, abs(val - val2) );
        SX1301AR_ERR_RETURN( ERR_LBT_TMST_SYNC );
    }
#endif

	/* Start Listen-Before-Talk */
	x = sx1301ar_reg_brd_w(brd, BREG_START_LBT, 1);
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_i_stop_lbt(uint8_t brd)
{
	int x;

	/* Stop Listen-Before-Talk */
	x = sx1301ar_reg_brd_w(brd, BREG_START_LBT, 0);
	SX1301AR_ERR_PROPAGATE(x);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

bool sx1301ar_i_lbt_is_enabled(uint8_t brd)
{
	return lbt_cfg[brd].enable;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_i_lbt_is_tx_allowed(uint8_t brd, const sx1301ar_tx_pkt_t *pkt, uint16_t tx_start_delay, bool *tx_allowed)
{
	int i;
	int32_t val;
	uint32_t tx_start_time = 0;
	uint32_t delta_time = 0;
	uint32_t sx1301_time = 0;
	uint32_t sx1301_now = 0;
	uint32_t lbt_time = 0;
	uint32_t lbt_time1 = 0;
	uint32_t lbt_time2 = 0;
	uint32_t tx_max_time = 0;
	int lbt_channel_decod_1 = -1;
	int lbt_channel_decod_2 = -1;
	uint32_t packet_duration = 0;

	/* Check input parameters */
	if ((pkt == NULL) || (tx_allowed == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Check if TX is allowed */
	if (lbt_cfg[brd].enable == true) {
		/* TX allowed for LoRa only */
		if (pkt->modulation != MOD_LORA) {
			*tx_allowed = false;
			error("TX is not allowed for this modulation (%x)", pkt->modulation);
			return 0;
		}

		/* Get SX1301 time at last PPS */
		sx1301ar_get_trigcnt(brd, &sx1301_time);

		switch (pkt->tx_mode) {
		case TX_TIMESTAMPED:
			tx_start_time = pkt->count_us;
			break;
		case TX_ON_GPS:
			tx_start_time = sx1301_time + tx_start_delay + 1000000;
			break;
		case TX_IMMEDIATE:
			error("tx_mode IMMEDIATE is not supported when LBT is enabled");
		/* FALLTHROUGH  */
		default:
			*tx_allowed = false;
			return 0;
		}

		/* Select LBT Channel corresponding to required TX frequency */
		lbt_channel_decod_1 = -1;
		lbt_channel_decod_2 = -1;
		if (pkt->bandwidth == BW_125K) {
			for (i = 0; i < lbt_cfg[brd].nb_channel; i++) {
				if (is_equal_freq(pkt->freq_hz, lbt_cfg[brd].channels[i].freq_hz) == true) {
					lbt_channel_decod_1 = i;
					lbt_channel_decod_2 = i;
					tx_max_time = lbt_cfg[brd].channels[i].transmit_time_ms * 1E3;
					break;
				}
			}
		} else if (pkt->bandwidth == BW_250K) {
			/* In case of 250KHz, the TX freq has to be in between 2 consecutive channels of 200KHz BW. The TX can only be over 2 channels, not more */
			for (i = 0; i < (lbt_cfg[brd].nb_channel - 1); i++) {
				/* TODO: to be checked */
				if ((is_equal_freq(pkt->freq_hz, (lbt_cfg[brd].channels[i].freq_hz + lbt_cfg[brd].channels[i + 1].freq_hz) / 2) == true) && ((lbt_cfg[brd].channels[i + 1].freq_hz - lbt_cfg[brd].channels[i].freq_hz) == 200E3)) {
					lbt_channel_decod_1 = i;
					lbt_channel_decod_2 = i + 1;
					tx_max_time = lbt_cfg[brd].channels[i].transmit_time_ms * 1E3;
					break;
				}
			}
		} else {
			/* Nothing to do for now */
		}

		/* We want to know the status of the channel on which we are going to transmit just before the
        actual time of departure of the packet. So we wait until that time (with margin LBT_DELAY) before
        reading the LBT timestamp. */
		sx1301ar_get_instcnt(brd, &sx1301_now);
		if (((tx_start_time - sx1301_now) < LBT_DELAY) || ((tx_start_time - sx1301_now) > 1E6)) /* unsigned arithmetic */
		{
			error("too late for LBT check, rejecting TX request (sx1301_now=%u, tx_start_time=%u, diff=%d)", sx1301_now, tx_start_time, (int32_t)tx_start_time - (int32_t)sx1301_now);
			*tx_allowed = false;
			return 0;
		}
		sx1301ar_wait_us(tx_start_time - sx1301_now - LBT_DELAY); /* remaining time before departure, with LBT_DELAY margin */

		/* Get last time when selected channel was free */
		if ((lbt_channel_decod_1 >= 0) && (lbt_channel_decod_2 >= 0)) {
			sx1301ar_reg_brd_w(brd, BREG_LBT_SELECT_CH, (int32_t)lbt_channel_decod_1);
			sx1301ar_reg_brd_r(brd, BREG_LBT_TIMESTAMP, &val);
			lbt_time = lbt_time1 = (uint32_t)val;

			if (lbt_channel_decod_1 != lbt_channel_decod_2) {
				sx1301ar_reg_brd_w(brd, BREG_LBT_SELECT_CH, (int32_t)lbt_channel_decod_2);
				sx1301ar_reg_brd_r(brd, BREG_LBT_TIMESTAMP, &val);
				lbt_time2 = (uint32_t)val;

				if (lbt_time2 < lbt_time1) {
					lbt_time = lbt_time2;
				}
			}
		} else {
			lbt_time = 0;
		}

		packet_duration = (uint32_t)(sx1301ar_tx_pkt_time_on_air(brd, pkt, NULL, NULL, NULL) * 1000UL);
		delta_time = (tx_start_time + packet_duration) - lbt_time;

		debug("lbt: T0=%u Tsensing=%u Dwindow=%u Tstart=%u Dpacket=%u Delta=%i",
		      sx1301_time,
		      lbt_time,
		      tx_max_time,
		      tx_start_time,
		      packet_duration,
		      delta_time);

		/* send data if allowed */
		/* lbt_time: last time when channel was free */
		/* tx_max_time: maximum time allowed to send packet since last free time */
		if ((delta_time < tx_max_time) && (lbt_time != 0)) {
			*tx_allowed = true;
		} else {
			error("TX request rejected (LBT) - %u", pkt->freq_hz);
			*tx_allowed = false;
		}
	} else {
		/* Always allow if LBT is disabled */
		*tx_allowed = true;
	}

	return 0;
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

/* As given frequencies have been converted from float to integer, some aliasing
issues can appear, so we can't simply check for equality, but have to take some
margin */
static bool is_equal_freq(uint32_t a, uint32_t b)
{
	int64_t diff;
	int64_t a64 = (int64_t)a;
	int64_t b64 = (int64_t)b;

	/* Calculate the difference */
	diff = llabs(a64 - b64);

	/* Check for acceptable diff range */
	if (diff <= 10000) {
		return true;
	}

	return false;
}

/* --- EOF ------------------------------------------------------------------ */
