/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Functions used to handle the Listen Before Talk feature

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

#ifndef _LORAGW_LBT_H
#define _LORAGW_LBT_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDANCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */

#include "hal.h"

/* -------------------------------------------------------------------------- */
/* --- PUBLIC CONSTANTS ----------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */

/**
@brief Set the configuration parameters for Listen-Before-Talk feature
@param brd  Target board number
@param cfg  Structure containing the configuration parameters
@return 0 if the board has been successfully configured, -1 else
*/
int sx1301ar_i_conf_lbt(uint8_t brd, const sx1301ar_lbt_cfg_t *cfg);

/**
@brief Initialize the board for Listen-Before-Talk feature based on current configuration
@param brd      Target board number
@param rx_freq  Center RX frequency of the radio, in Hz
@return 0 if the board has been successfully initialized for Listen-Before-Talk, -1 else
*/
int sx1301ar_i_init_lbt(uint8_t brd, uint32_t rx_freq);

/**
@brief Start the Listen-Before-Talk state machine
@param brd  Target board number
@return 0 if the Listen-Before-Talk has successfully started, -1 else
*/
int sx1301ar_i_start_lbt(uint8_t brd);

/**
@brief Stop the Listen-Before-Talk state machine
@param brd  Target board number
@return 0 if the Listen-Before-Talk has successfully stopped, -1 else
*/
int sx1301ar_i_stop_lbt(uint8_t brd);

/**
@brief Check if Listen-Before-Talk is enabled for this board
@param brd  Target board number
@return true if Listen-Before-Talk is enabled, false else
*/
bool sx1301ar_i_lbt_is_enabled(uint8_t brd);

/**
@brief Check if TX is allowed when Listen-Before-Talk is enabled
@param brd              Target board number
@param pkt              pointer to downlink packet to be transmitted
@param tx_start_delay   TX start delay applied for this transmit
@param tx_allowed       pointer to receive permission for transmission
@return 0 if the check could successfully be performed, -1 else
*/
int sx1301ar_i_lbt_is_tx_allowed(uint8_t brd, const sx1301ar_tx_pkt_t *pkt, uint16_t tx_start_delay, bool *tx_allowed);

#endif

/* --- EOF ------------------------------------------------------------------ */
