/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Radio driver for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Matthieu Leurent
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <math.h>
#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <string.h>

#include "aux.h"
#include "err.h"
#include "hal.h"
#include "radio.h"
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

/* AD9361 */
#define XTAL_FREQ_MHz 32
#define BB_FREQ_MHz 1024
#define ADC_FREQ_MHz 256
#define CLK_OUT_FREQ_MHz 128
#define TX_BW_MHz 0.5 // SSB, min:0.2 max:28
#define RX_FIR_SIZE 128
#define TX_FIR_SIZE 128
#define GAIN_TABLE_SIZE 2

#define PI 3.14159265358979

#define TX_FREQ_DEFAULT 900000000ULL /* dummy TX frequency for init */

struct ad9361_gain_s {
	uint8_t index;
	uint8_t lna;
	uint8_t mixer;
	uint8_t tia;
	uint8_t lpf;
	uint8_t digital;
	bool calibrate;
};

struct ad9361_offset_s {
	int16_t cgin_1_i;
	int16_t cgin_1_q;
	int16_t cgin_2_i;
	int16_t cgin_2_q;
	int16_t cgout_1_i;
	int16_t cgout_1_q;
	int16_t cgout_2_i;
	int16_t cgout_2_q;
	int16_t dig_1_i;
	int16_t dig_1_q;
	int16_t dig_2_i;
	int16_t dig_2_q;
};

struct tx_pwr_s {
	int8_t pwr_dbm;
	uint8_t fpga_gain;
	uint8_t ad_atten;
	double v_pa;
};

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

static const struct ad9361_gain_s gain_table[GAIN_TABLE_SIZE] =
    {
	{0, 2, 4, 1, 1, 0, true},
	{1, 3, 7, 1, 16, 0, true}};

/* 4 MHz band */
static const int16_t rx_fir_coef_4mhz[RX_FIR_SIZE] =
    {
	0, 0, 0, 0, 0, -2, -4, -7,
	-10, -10, -8, -2, 7, 17, 28, 36,
	37, 32, 17, -4, -30, -56, -77, -85,
	-79, -56, -17, 33, 85, 129, 156, 157,
	128, 70, -10, -101, -186, -249, -274, -251,
	-177, -58, 89, 242, 373, 453, 461, 382,
	219, -15, -288, -557, -773, -885, -850, -640,
	-246, 318, 1017, 1792, 2574, 3286, 3856, 4225,
	4352, 4225, 3856, 3286, 2574, 1792, 1017, 318,
	-246, -640, -850, -885, -773, -557, -288, -15,
	219, 382, 461, 453, 373, 242, 89, -58,
	-177, -251, -274, -249, -186, -101, -10, 70,
	128, 157, 156, 129, 85, 33, -17, -56,
	-79, -85, -77, -56, -30, -4, 17, 32,
	37, 36, 28, 17, 7, -2, -8, -10,
	-10, -7, -4, -2, 0, 0, 0, 0,
};

/* 7 MHz band */
static const int16_t rx_fir_coef_7mhz[RX_FIR_SIZE] =
    {
	0, 0, 0, -1, -2, -4, -2, 2,
	8, 13, 10, 0, -14, -26, -26, -11,
	16, 41, 50, 33, -6, -52, -79, -68,
	-19, 51, 106, 115, 63, -32, -125, -168,
	-128, -14, 125, 219, 213, 93, -93, -259,
	-312, -210, 18, 271, 420, 371, 119, -237,
	-528, -588, -346, 126, 624, 891, 735, 136,
	-701, -1401, -1543, -835, 751, 2924, 5148, 6809,
	7424, 6809, 5148, 2924, 751, -835, -1543, -1401,
	-701, 136, 735, 891, 624, 126, -346, -588,
	-528, -237, 119, 371, 420, 271, 18, -210,
	-312, -259, -93, 93, 213, 219, 125, -14,
	-128, -168, -125, -32, 63, 115, 106, 51,
	-19, -68, -79, -52, -6, 33, 50, 41,
	16, -11, -26, -26, -14, 0, 10, 13,
	8, 2, -2, -4, -2, -1, 0, 0};

/* 8 MHz band */
/* TODO: TO BE UPDATED */
static const int16_t rx_fir_coef_8mhz[RX_FIR_SIZE] =
    {
	0, 0, 0, -1, -2, -4, -2, 2,
	8, 13, 10, 0, -14, -26, -26, -11,
	16, 41, 50, 33, -6, -52, -79, -68,
	-19, 51, 106, 115, 63, -32, -125, -168,
	-128, -14, 125, 219, 213, 93, -93, -259,
	-312, -210, 18, 271, 420, 371, 119, -237,
	-528, -588, -346, 126, 624, 891, 735, 136,
	-701, -1401, -1543, -835, 751, 2924, 5148, 6809,
	7424, 6809, 5148, 2924, 751, -835, -1543, -1401,
	-701, 136, 735, 891, 624, 126, -346, -588,
	-528, -237, 119, 371, 420, 271, 18, -210,
	-312, -259, -93, 93, 213, 219, 125, -14,
	-128, -168, -125, -32, 63, 115, 106, 51,
	-19, -68, -79, -52, -6, 33, 50, 41,
	16, -11, -26, -26, -14, 0, 10, 13,
	8, 2, -2, -4, -2, -1, 0, 0};

/* 13 MHz band */
static const int16_t rx_fir_coef_13mhz[RX_FIR_SIZE] =
    {
	0, 0, -1, -1, 1, 4, 0, -7,
	-6, 8, 14, -3, -22, -10, 25, 28,
	-16, -46, -8, 53, 41, -41, -74, 6,
	93, 47, -84, -104, 39, 144, 38, -146,
	-128, 96, 203, 6, -231, -138, 186, 265,
	-62, -341, -123, 321, 327, -184, -484, -65,
	528, 381, -401, -686, 79, 878, 424, -842,
	-1050, 456, 1705, 451, -2281, -2386, 2676, 10047,
	13568, 10047, 2676, -2386, -2281, 451, 1705, 456,
	-1050, -842, 424, 878, 79, -686, -401, 381,
	528, -65, -484, -184, 327, 321, -123, -341,
	-62, 265, 186, -138, -231, 6, 203, 96,
	-128, -146, 38, 144, 39, -104, -84, 47,
	93, 6, -74, -41, 41, 53, -8, -46,
	-16, 28, 25, -10, -22, -3, 14, 8,
	-6, -7, 0, 4, 1, -1, -1, 0};

/* 0.5 MHz band (previous Rx filter) */
static const int16_t tx_fir_coef[TX_FIR_SIZE] =
    {
	0, 0, -1, -2, -3, -5, -7, -9,
	-11, -13, -15, -17, -19, -20, -21, -21,
	-21, -19, -17, -13, -8, -2, 5, 14,
	25, 37, 51, 66, 84, 103, 124, 146,
	171, 197, 224, 253, 284, 315, 348, 382,
	417, 452, 488, 524, 560, 597, 633, 668,
	703, 737, 770, 802, 832, 861, 888, 912,
	935, 955, 973, 988, 1001, 1011, 1018, 1023,
	1024, 1023, 1018, 1011, 1001, 988, 973, 955,
	935, 912, 888, 861, 832, 802, 770, 737,
	703, 668, 633, 597, 560, 524, 488, 452,
	417, 382, 348, 315, 284, 253, 224, 197,
	171, 146, 124, 103, 84, 66, 51, 37,
	25, 14, 5, -2, -8, -13, -17, -19,
	-21, -21, -21, -20, -19, -17, -15, -13,
	-11, -9, -7, -5, -3, -2, -1, 0};

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

static int ad9361_bb_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz);

static void ad9361_rx_pll_setup(uint8_t brd);

static int ad9361_rx_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz);

static int ad9361_tx_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz);

static int ad9361_rx_bb_filter_cal(uint8_t brd, uint64_t bw_khz, uint64_t f_hz, uint32_t *C3_fF, uint32_t *R2346);

static int ad9361_tx_bb_filter_cal(uint8_t brd, uint64_t bw_khz, uint64_t f_hz);

static void ad9361_rx_tia_setup(uint8_t brd, uint32_t C3_fF, uint32_t R2346, uint64_t bw_khz);

static void ad9361_adc_setup(uint8_t brd, uint32_t C3_fF, uint32_t R2346, uint64_t bw_khz, uint64_t f_hz);

static int ad9361_gain_table_write(uint8_t brd, const struct ad9361_gain_s *gain_table, uint8_t size, uint8_t rf_chain, bool read_back);

static int ad9361_rx_fir_set_coef(uint8_t brd, const int16_t *coef, uint8_t size, uint16_t rf_chain, bool read_back);

static int ad9361_tx_fir_set_coef(uint8_t brd, const int16_t *coef, uint8_t size, uint16_t rf_chain, bool read_back);

static int ad9361_rx_bb_dc_offset_cal(uint8_t brd);

static int ad9361_rx_rf_dc_offset_cal(uint8_t brd);

static int ad9361_rx_quadrature_cal(uint8_t brd);

static int ad9361_tx_quadrature_cal(uint8_t brd);

#if 0 /* Not used for now */
static void ad9361_dc_offset_read( uint8_t brd, struct ad9361_offset_s * offsets );

/**
@brief Set AD9361 auxiliary DAC output voltage
@param brd     Target board number
@param dac_nb  Target 1 -> AUX_DAC1, 2 -> AUX_DAC2
@param v       Target output voltage of the DAC
@return 0 if the DAC was set successfully, -1 else
*/
void ad9361_auxdac_set_val( uint8_t brd, uint8_t rf_chain, double voltage );
#endif

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int ad9361_start(uint8_t brd, uint64_t rx_freq_hz, uint32_t rx_bw_hz, uint8_t rx_en, uint8_t tx_en, bool full_duplex)
{
	int i, x;
	uint8_t r, r2;
	int32_t r3;
	uint8_t buf[3] = {0, 0, 0};
	uint32_t C3_fF, R2346;
	uint64_t rx_bw_khz;
	int16_t rx_fir_coef[RX_FIR_SIZE];
	//struct ad9361_offset_s o;

	/* Soft reset */
	sx1301ar_ad9361_w(brd, RREG_SPI_CONF__SOFT_RESET, 1);
	sx1301ar_ad9361_w(brd, RREG_SPI_CONF___SOFT_RESET, 1);
	sx1301ar_ad9361_w(brd, RREG_SPI_CONF__SOFT_RESET, 0);
	sx1301ar_ad9361_w(brd, RREG_SPI_CONF___SOFT_RESET, 0);

	/* Check version */
	sx1301ar_ad9361_r(brd, RREG_PRODUCT_ID, &r);
	if (r != 10) {
		SX1301AR_ERR_RETURN(ERR_AD9361_VERSION);
	}

	/* Select RX FIR based on give RX bandwidth */
	switch (rx_bw_hz) {
	case 4000000:
		memcpy(rx_fir_coef, rx_fir_coef_4mhz, sizeof rx_fir_coef);
		break;
	case 7000000:
		memcpy(rx_fir_coef, rx_fir_coef_7mhz, sizeof rx_fir_coef);
		break;
	case 8000000:
		memcpy(rx_fir_coef, rx_fir_coef_8mhz, sizeof rx_fir_coef);
		break;
	case 13000000:
		memcpy(rx_fir_coef, rx_fir_coef_13mhz, sizeof rx_fir_coef);
		break;
	}
	rx_bw_khz = rx_bw_hz / 1e3 / 2; /* format needed for subsequent calls */

	/* --- Misc ---------- */

	sx1301ar_ad9361_w(brd, RREG_BANDGAP_CONFIG0__MASTER_BIAS_TRIM, 13);
	sx1301ar_ad9361_w(brd, RREG_BANDGAP_CONFIG1__BANDGAP_TEMP_TRIM, 13);
	sx1301ar_ad9361_w(brd, RREG_CLOCK_ENABLE__BBPLL_ENABLE, 1);
	sx1301ar_ad9361_w(brd, RREG_CLOCK_ENABLE__CLOCK_ENABLE_DFLT, 1);
	sx1301ar_ad9361_w(brd, RREG_CLOCK_ENABLE__DIGITAL_POWER_UP, 1);
	sx1301ar_ad9361_w(brd, RREG_CLOCK_ENABLE__XO_BYPASS, 1); // Use external TCXO
	sx1301ar_ad9361_w(brd, RREG_INPUT_SELECT__RX_INPUT, 3);  // Balanced input A on both chains
	sx1301ar_ad9361_w(brd, RREG_EXTERNAL_LNA_CTRL__AUXDAC_MANUAL_SELECT, 1);
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_ENABLE_CTRL__AUXDAC_MANUAL_BAR, 3); // disable AuxDAC outputs
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_1_CONFIG__AUXDAC1_STP_FACTOR, 1);   // lowest step
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_2_CONFIG__AUXDAC2_STP_FACTOR, 1);   // lowest step
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_1_CONFIG__AUXDAC_1_VREF, 2);	// Always use Vref MSB=1
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_2_CONFIG__AUXDAC_2_VREF, 2);	// Always use Vref MSB=1
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_1_CONFIG__COMP_CTRL_1, 1);	  // Adds 400fF of capacitance to the AuxDAC to increase stability
	sx1301ar_ad9361_w(brd, RREG_AUXDAC_2_CONFIG__COMP_CTRL_2, 1);	  // Adds 400fF of capacitance to the AuxDAC to increase stability
	sx1301ar_ad9361_w(brd, RREG_EXTERNAL_LNA_CTRL__GPO_MANUAL_SELECT, 1);
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CONFIG_2, 96); // Bits must be 1 but 0 by default !?
	sx1301ar_ad9361_w(brd, RREG_RX_ENABLE_FILTER_CTRL__RX_CHANNEL_ENABLE, rx_en);
	sx1301ar_ad9361_w(brd, RREG_TX_ENABLE_FILTER_CTRL__TX_CHANNEL_ENABLE, tx_en);
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL, 0);
	sx1301ar_ad9361_w(brd, RREG_TEMP_SENSE2__TEMP_SENSE_PERIODIC_ENABLE, 0);

	/* --- Interface --------- */

	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_1, 0);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_1__PP_RX_SWAP_IQ, 1);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_1__PP_TX_SWAP_IQ, 1);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_1__RX_FRAME_PULSE_MODE, 1);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_1__R2T2_TIMING, 1);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_2, 0);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_3, 0);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_3__FULL_PORT, 1);
	sx1301ar_ad9361_w(brd, RREG_PARALLEL_PORT_CONF_3__SWAP_PORTS, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_CLOCK_DATA_DELAY__RX_DATA_DELAY, 15);
	sx1301ar_ad9361_w(brd, RREG_RX_CLOCK_DATA_DELAY__DATA_CLK_DELAY, 0);
	sx1301ar_ad9361_w(brd, RREG_TX_CLOCK_DATA_DELAY__TX_DATA_DELAY, 15);
	sx1301ar_ad9361_w(brd, RREG_TX_CLOCK_DATA_DELAY__FB_CLK_DELAY, 0);

	/* --- Enable state machine --------- */

	sx1301ar_ad9361_w(brd, RREG_ENSM_MODE__FDD_MODE, 1);
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_1, 0);
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_1__ENABLE_ENSM_PIN_CTRL, 1); // Control state machine with ENABLE/TXNRX pins
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_1__LEVEL_MODE, 1);	   // Control of ENABLE/TXNRX is level-based, not pulse-based
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_1__TO_ALERT, 1);	     // Go back to alert when leaving Rx/Tx
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_2, 0);
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_2__SYNTH_ENABLE_PIN_CTRL_MODE, 1);
	//sx1301ar_ad9361_w( brd, RREG_ENSM_CONFIG_2__TX_SYNTH_READY_MASK, 1 );
	//sx1301ar_ad9361_w( brd, RREG_ENSM_CONFIG_2__RX_SYNTH_READY_MASK, 1 );

	/* --- Reference Freq, BB PLL, ADC clock, and clock output -------- */

	sx1301ar_ad9361_w(brd, RREG_REFERENCE_CLOCK_CYCLES__REFERENCE_CLOCK_CYCLES_PER_US, XTAL_FREQ_MHz - 1);

	/* Ref */
	sx1301ar_ad9361_w(brd, RREG_REF_DIVIDE_CONFIG_1__RX_REF_DIVIDER_MSB, 1); // 3: 2*XTAL
	sx1301ar_ad9361_w(brd, RREG_REF_DIVIDE_CONFIG_2__RX_REF_DIVIDER_LSB, 1);
	sx1301ar_ad9361_w(brd, RREG_REF_DIVIDE_CONFIG_2__TX_REF_DIVIDER, 3); // 3: 2*XTAL
	sx1301ar_ad9361_w(brd, RREG_REF_DIVIDE_CONFIG_1__RX_REF_RESET_BAR, 1);
	sx1301ar_ad9361_w(brd, RREG_REF_DIVIDE_CONFIG_2__TX_REF_RESET_BAR, 1);
	sx1301ar_ad9361_w(brd, RREG_CLOCK_CTRL__REF_FREQ_SCALER, 3); // 3: 2*XTAL

	/* BB PLL */
	x = ad9361_bb_pll_set_freq(brd, BB_FREQ_MHz * 1e6, AD9361_REF_FREQ_MHz * 1e6);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_BBPLL_LOCK);
	}

	/* ADC clock */
	sx1301ar_ad9361_w(brd, RREG_BBPLL__BBPLL_DIVIDER, 2); //log2(BB_FREQ_MHz/ADC_FREQ_MHz)

	/* CLK OUT */
	sx1301ar_ad9361_w(brd, RREG_BBPLL__CLKOUT_SELECT, 3);
	sx1301ar_ad9361_w(brd, RREG_BBPLL__CLKOUT_ENABLE, 0);

	/* --- Enter ALERT state -------- */

	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_1__FORCE_ALERT_STATE, 1);
	sx1301ar_wait_us(1000);
	sx1301ar_ad9361_r(brd, RREG_STATE__ENSM_STATE, &r);
	if (r != 5) {
		SX1301AR_ERR_RETURN(ERR_AD9361_ALERT);
	}

	/* --- RF PLLs --------- */

	sx1301ar_ad9361_w(brd, RREG_RFPLL_DIVIDERS__RX_VCO_DIVIDER, 2); //log2(AD9361_PLL_DIV)-1
	sx1301ar_ad9361_w(brd, RREG_RFPLL_DIVIDERS__TX_VCO_DIVIDER, 2); //log2(AD9361_PLL_DIV)-1

	/* Rx PLL setup */
	ad9361_rx_pll_setup(brd);

	/* Tx PLL setup: copy from Rx */
	for (i = 0x230; i <= 0x251; i++) {
		// Read
		buf[0] = (uint8_t)(i >> 8);
		buf[1] = (uint8_t)(i & 0xFF);
		sx1301ar_spibdg_w_r(brd, buf, 2, &r, 1);
		// Write
		buf[0] = (1 << 7) | (uint8_t)((i + 64) >> 8);
		buf[1] = (uint8_t)((i + 64) & 0xFF);
		buf[2] = r;
		sx1301ar_spibdg_w_r(brd, buf, 3, NULL, 0);
	}

	/* Set this register to enable both Rx and Tx synthesisers */
	/* It will be only used here for calibration */
	sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_2__DUAL_SYNTH_MODE, 1);

	/* Charge pump calibration */
	sx1301ar_ad9361_w(brd, RREG_RX_CP_CONFIG__CP_CAL_ENABLE, 1);
	sx1301ar_ad9361_w(brd, RREG_TX_CP_CONFIG__CP_CAL_ENABLE, 1);
	sx1301ar_wait_us(1000);

	/* Check calibration status */
	sx1301ar_ad9361_r(brd, RREG_RX_CAL_STATUS__CP_CAL_VALID, &r);
	sx1301ar_ad9361_r(brd, RREG_TX_CAL_STATUS__CP_CAL_VALID, &r2);
	if ((r == 0) || (r2 == 0)) {
		SX1301AR_ERR_RETURN(ERR_AD9361_PLL_CAL);
	}

	/* Do calibration only once */
	sx1301ar_ad9361_w(brd, RREG_RX_CP_CONFIG__CP_CAL_ENABLE, 0);
	sx1301ar_ad9361_w(brd, RREG_TX_CP_CONFIG__CP_CAL_ENABLE, 0);

	/* Lock Rx PLL */
	x = ad9361_rx_pll_set_freq(brd, rx_freq_hz * AD9361_PLL_DIV, AD9361_REF_FREQ_MHz * 1e6);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_RXPLL_LOCK);
	}

	/* Lock Tx PLL */
	x = ad9361_tx_pll_set_freq(brd, TX_FREQ_DEFAULT * AD9361_PLL_DIV, AD9361_REF_FREQ_MHz * 1e6);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_TXPLL_LOCK);
	}

	if (full_duplex == false) {
		/* Unset this register to have only one synthesiser on */
		sx1301ar_ad9361_w(brd, RREG_ENSM_CONFIG_2__DUAL_SYNTH_MODE, 0);
	}

	/* --- Rx BB filter ---------- */
	x = ad9361_rx_bb_filter_cal(brd, rx_bw_khz, BB_FREQ_MHz * 1e6, &C3_fF, &R2346);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_RXFILT_CAL);
	}
	//debug( "C3 = %d fF", C3_fF );
	//debug( "R2346 = %d Ohm", R2346 );

	/* --- Rx TIA setup ---------- */

	ad9361_rx_tia_setup(brd, C3_fF, R2346, rx_bw_khz);

	/* --- Rx ADC setup ----------- */

	ad9361_adc_setup(brd, C3_fF, R2346, rx_bw_khz, ADC_FREQ_MHz * 1e6);

	/* --- Tx BB filter ---------- */

	x = ad9361_tx_bb_filter_cal(brd, TX_BW_MHz * 1e3, BB_FREQ_MHz * 1e6);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_TXFILT_CAL);
	}

	/* --- Gain control ----------- */

	sx1301ar_ad9361_w(brd, RREG_AGC_CONFIG_1__RX1_GAIN_CTRL_SETUP, 0); // Manual gain control
	sx1301ar_ad9361_w(brd, RREG_AGC_CONFIG_1__RX2_GAIN_CTRL_SETUP, 0);
	sx1301ar_ad9361_w(brd, RREG_AGC_CONFIG_2__AGC_USE_FULL_GAIN_TABLE, 1);
	sx1301ar_ad9361_w(brd, RREG_AGC_CONFIG_2__MAN_GAIN_CTRL_RX1, 0); // SPI control of gain index
	sx1301ar_ad9361_w(brd, RREG_AGC_CONFIG_2__MAN_GAIN_CTRL_RX2, 0);
	sx1301ar_ad9361_w(brd, RREG_MAX_LMT_FULL_GAIN__MAXIMUM_FULL_TABLELMT_TABLE_INDEX, GAIN_TABLE_SIZE - 1); // Set max index to stop RF DC calibration

	/* Write back table on both chains */
	x = ad9361_gain_table_write(brd, &gain_table[0], GAIN_TABLE_SIZE, 3, true);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_GAIN_TABLE);
	}

	/* Set gain */
	sx1301ar_ad9361_w(brd, RREG_RX1_MANUAL_LMT_FULL_GAIN__RX1_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX, 0);
	sx1301ar_ad9361_w(brd, RREG_RX2_MANUAL_LMT_FULL_GAIN__RX2_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX, 0);

	/* --- Tx power control ----------- */

	// Apply instantaneously attenuation, attenuation is 0 by default
	sx1301ar_ad9361_w(brd, RREG_TX_ATTEN_OFFSET__MASK_CLR_ATTEN_UPDATE, 1);
	sx1301ar_ad9361_w(brd, RREG_TX2_DIG_ATTEN__IMMEDIATELY_UPDATE_TPC_ATTEN, 1);

	/* --- Rx Digital filter ------------- */

	sx1301ar_ad9361_w(brd, RREG_RX_ENABLE_FILTER_CTRL__DEC3_ENABLE_DECIMATION, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_ENABLE_FILTER_CTRL__RHB2_EN, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_ENABLE_FILTER_CTRL__RHB1_EN, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_ENABLE_FILTER_CTRL__RX_FIR_ENABLE_DECIMATION, 2);
	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_GAIN__FILTER_GAIN, 2); // 0:+6, 1:0, 2:-6, 3:-12

	x = ad9361_rx_fir_set_coef(brd, rx_fir_coef, RX_FIR_SIZE, 3, true);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_RX_FIR);
	}

	/* --- Tx Digital filter ---------- */

	sx1301ar_ad9361_w(brd, RREG_TX_ENABLE_FILTER_CTRL__TX_FIR_ENABLE_INTERPOLATION, 2);
	sx1301ar_ad9361_w(brd, RREG_TX_ENABLE_FILTER_CTRL__THB1_EN, 1);
	sx1301ar_ad9361_w(brd, RREG_TX_ENABLE_FILTER_CTRL__THB2_EN, 1);
	sx1301ar_ad9361_w(brd, RREG_TX_ENABLE_FILTER_CTRL__THB3_ENABLE_INTERP, 1);

	x = ad9361_tx_fir_set_coef(brd, tx_fir_coef, TX_FIR_SIZE, 3, true);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_TX_FIR);
	}

	/* --- Rx BB DC offset calibration ---------- */

	x = ad9361_rx_bb_dc_offset_cal(brd);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_BBDC_CAL);
	}

	/* --- Rx RF DC offset calibration ---------- */

	x = ad9361_rx_rf_dc_offset_cal(brd);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_RFDC_CAL);
	}

	/* --- Rx quadrature calibration ------------ */

	x = ad9361_rx_quadrature_cal(brd);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_RXQUAD_CAL);
	}

	/* --- Tx quadrature calibration ------------ */

	x = ad9361_tx_quadrature_cal(brd);
	if (x != 0) {
		SX1301AR_ERR_RETURN(ERR_AD9361_TXQUAD_CAL);
	}

	/* --- Unreset FPGA PLL, needed for feedback clock -------- */

	sx1301ar_reg_brd_w(brd, BREG_PLL_RESET, 0);
	sx1301ar_wait_us(1000);
	sx1301ar_reg_brd_r(brd, BREG_PLL_LOCKED, &r3);
	if (r3 != 1) {
		SX1301AR_ERR_RETURN(ERR_FPGA_PLL);
	}

	/* --- Test FDD ----------------*/

	sx1301ar_reg_brd_w(brd, BREG_AD9361_ENABLE, 1);
	sx1301ar_wait_us(1000);
	sx1301ar_ad9361_r(brd, RREG_STATE__ENSM_STATE, &r);
	if (r != 10) {
		SX1301AR_ERR_RETURN(ERR_AD9361_TDD);
	}

	/* --- Read back gains and offsets ------------- */

	/*
    sx1301ar_ad9361_r( brd, RREG_GAIN_RX1__FULL_TABLE_GAIN_INDEX, &r );
    debug( "LMT gain index: %d", r );
    sx1301ar_ad9361_r( brd, RREG_LPF_GAIN_RX1__LPF_GAIN_RX, &r );
    debug( "LPF gain (dB) : %d", r );
    sx1301ar_ad9361_r( brd, RREG_DIG_GAIN_RX1__DIGITAL_GAIN_RX, &r );
    debug( "Dig gain index: %d", r );

    ad9361_dc_offset_read( brd, &o );
    debug( "CGIN  DAC: I:%5d Q:%5d", o.cgin_1_i, o.cgin_1_q );
    debug( "CGOUT DAC: I:%5d Q:%5d", o.cgout_1_i, o.cgout_1_q );
    debug( "DC WORDS : I:%5d Q:%5d", o.dig_1_i, o.dig_1_q );
    */
	return 0;
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

static int ad9361_bb_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz)
{
	uint8_t r;
	uint64_t n_int, n_frac;

	/* Integer and fractional word */
	n_int = f_hz / fref_hz;
	n_frac = (2088960 * f_hz) / fref_hz - 2088960 * n_int;
	sx1301ar_ad9361_w(brd, RREG_FRACT_BB_FREQ_WORD_1, (uint8_t)(n_frac & 0x0000FF));
	sx1301ar_ad9361_w(brd, RREG_FRACT_BB_FREQ_WORD_2, (uint8_t)((n_frac & 0x00FF00) >> 8));
	sx1301ar_ad9361_w(brd, RREG_FRACT_BB_FREQ_WORD_3, (uint8_t)((n_frac & 0xFF0000) >> 16));
	sx1301ar_ad9361_w(brd, RREG_INTEGER_BB_FREQ_WORD, (uint8_t)n_int);

	/* from ad9361_bbpll_set_rate */
	sx1301ar_ad9361_w(brd, RREG_VCO_CTRL__FREQ_CAL_COUNT_LENGTH, 3); // Longest cal time
	sx1301ar_ad9361_w(brd, RREG_VCO_CTRL__FREQ_CAL_ENABLE, 1);
	sx1301ar_ad9361_w(brd, RREG_SDM_CTRL__CAL_CLOCK_DIV_4, 1);
	sx1301ar_ad9361_w(brd, RREG_SDM_CTRL_1__BBPLL_RESET_BAR, 1);

	/* from ad9361_bbpll_set_rate: Increase BBPLL KV and phase margin */
	sx1301ar_ad9361_w(brd, RREG_VCO_PROGRAM_1, 0x86);
	sx1301ar_ad9361_w(brd, RREG_VCO_PROGRAM_2, 0x01);
	sx1301ar_ad9361_w(brd, RREG_VCO_PROGRAM_2, 0x05);

	/* BB VCO calibration */
	sx1301ar_ad9361_w(brd, RREG_SDM_CTRL_1__INIT_BB_FO_CAL, 1);
	sx1301ar_ad9361_w(brd, RREG_SDM_CTRL_1__INIT_BB_FO_CAL, 0);
	sx1301ar_wait_us(1000);

	sx1301ar_ad9361_r(brd, RREG_CH_1_OVERFLOW__BBPLL_LOCK, &r);

	return ((int)r - 1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void ad9361_rx_pll_setup(uint8_t brd)
{
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_OUTPUT__PORB_VCO_LOGIC, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_LOCK_DETECT_CONFIG__LOCK_DETECT_COUNT, 3); // max lock time

	/* from ad9361_txrx_synth_cp_calib */
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_PD_OVERRIDES__POWER_DOWN_CAL_TCF, 1); // ?
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_LDO__VCO_LDO_SEL, 2);		 // Low noise, high current
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_CAL__VCO_CAL_EN, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_CAL__VCO_CAL_COUNT, 3); // max calibration time

	/* from ad9361_rfpll_vco_init */
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_CAL_REF__VCO_CAL_REF_TCF, 0);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_OFFSET, 0);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_REFERENCE_TCF, 7);

	/* from index 32 of LUT FDD 60MHz */
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_OUTPUT__VCO_OUTPUT_LEVEL, 10);
	sx1301ar_ad9361_w(brd, RREG_RX_ALC_VARACTOR__VCO_VARACTOR, 1);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_BIAS_1__VCO_BIAS_REF, 7);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_BIAS_1__VCO_BIAS_TCF, 2);
	sx1301ar_ad9361_w(brd, RREG_RX_FORCE_VCO_TUNE_1__VCO_CAL_OFFSET, 15);
	sx1301ar_ad9361_w(brd, RREG_RX_VCO_VARACTOR_CTRL_1__VCO_VARACTOR_REFERENCE, 12);

	/* Optimised PLL BW */
	sx1301ar_ad9361_w(brd, RREG_RX_CP_CURRENT, 128 + 2);
	sx1301ar_ad9361_w(brd, RREG_RX_LOOP_FILTER_1__LOOP_FILTER_C2, 0);
	sx1301ar_ad9361_w(brd, RREG_RX_LOOP_FILTER_1__LOOP_FILTER_C1, 15);
	sx1301ar_ad9361_w(brd, RREG_RX_LOOP_FILTER_2__LOOP_FILTER_R1, 13);
	sx1301ar_ad9361_w(brd, RREG_RX_LOOP_FILTER_2__LOOP_FILTER_C3, 15);
	sx1301ar_ad9361_w(brd, RREG_RX_LOOP_FILTER_3__LOOP_FILTER_R3, 15);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz)
{
	uint8_t r;
	uint64_t n_int, n_frac;

	/* Integer and fractional word */
	n_int = f_hz / fref_hz;
	n_frac = (AD9361_PLL_MODULUS * f_hz) / fref_hz - AD9361_PLL_MODULUS * n_int;
	sx1301ar_ad9361_w(brd, RREG_RX_FRACT_BYTE_0, (uint8_t)(n_frac & 0x0000FF));
	sx1301ar_ad9361_w(brd, RREG_RX_FRACT_BYTE_1, (uint8_t)((n_frac & 0x00FF00) >> 8));
	sx1301ar_ad9361_w(brd, RREG_RX_FRACT_BYTE_2__SYNTH_FRACT_WORD, (uint8_t)((n_frac & 0xFF0000) >> 16));
	sx1301ar_ad9361_w(brd, RREG_RX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD, (uint8_t)((n_int & 0xFF00) >> 8));
	sx1301ar_ad9361_w(brd, RREG_RX_INTEGER_BYTE_0, (uint8_t)(n_int & 0x00FF));

	/* INFO: VCO cal starts automatically after writting RX_INTEGER_BYTE_0 */
	sx1301ar_wait_us(1000);

	sx1301ar_ad9361_r(brd, RREG_RX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK, &r);

	return ((int)r - 1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_tx_pll_set_freq(uint8_t brd, uint64_t f_hz, uint64_t fref_hz)
{
	uint8_t r;
	uint64_t n_int, n_frac;

	/* Integer and fractional word */
	n_int = f_hz / fref_hz;
	n_frac = (AD9361_PLL_MODULUS * f_hz) / fref_hz - AD9361_PLL_MODULUS * n_int;
	sx1301ar_ad9361_w(brd, RREG_TX_FRACT_BYTE_0, (uint8_t)(n_frac & 0x0000FF));
	sx1301ar_ad9361_w(brd, RREG_TX_FRACT_BYTE_1, (uint8_t)((n_frac & 0x00FF00) >> 8));
	sx1301ar_ad9361_w(brd, RREG_TX_FRACT_BYTE_2__SYNTH_FRACT_WORD, (uint8_t)((n_frac & 0xFF0000) >> 16));
	sx1301ar_ad9361_w(brd, RREG_TX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD, (uint8_t)((n_int & 0xFF00) >> 8));
	sx1301ar_ad9361_w(brd, RREG_TX_INTEGER_BYTE_0, (uint8_t)(n_int & 0x00FF));

	/* INFO: VCO cal starts automatically after writting TX_INTEGER_BYTE_0 */
	sx1301ar_wait_us(1000);

	sx1301ar_ad9361_r(brd, RREG_TX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK, &r);

	return ((int)r - 1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_bb_filter_cal(uint8_t brd, uint64_t bw_khz, uint64_t f_hz, uint32_t *C3_fF, uint32_t *R2346)
{
	uint8_t r;
	uint64_t TuneDiv;
	uint8_t c3_msb, c3_lsb, r2346;
	uint64_t bw_mhz_round;

	/* Configure clock tuner */
	//TuneDiv = (uint16_t) ceil( (f_MHz * log(2)) / (bw_MHz * 1.4 * 2 * PI) );
	TuneDiv = f_hz / (bw_khz * 12691) + 1;
	bw_mhz_round = (uint64_t)(bw_khz / 1e3);
	sx1301ar_ad9361_w(brd, RREG_RX_BBF_TUNE_DIVIDE, (uint8_t)(TuneDiv & 0x00FF));
	sx1301ar_ad9361_w(brd, RREG_RX_BBF_TUNE_CONFIG__RX_BBF_TUNE_DIVIDE, (uint8_t)((TuneDiv & 0x0100) >> 8));
	sx1301ar_ad9361_w(brd, RREG_RX_BBBW_MHZ__RX_TUNE_BBBW_MHZ, (uint8_t)bw_mhz_round);
	sx1301ar_ad9361_w(brd, RREG_RX_BBBW_KHZ__RX_TUNE_BBBW_KHZ, (uint8_t)((double)(bw_khz - bw_mhz_round * 1e3) / 7.8125 + 0.5));
	sx1301ar_ad9361_w(brd, RREG_RX_BBF_TUNE_CONFIG__RX_TUNE_EVALTIME, 1); // Maximum delay before tune comparator is sampled
	sx1301ar_ad9361_w(brd, RREG_RX_BBF_TUNE_CONFIG__RX_TUNE_MODE, 7);     // Algorithm stops when all comparator outputs are low

	/* Enable tune circuit of both chains */
	sx1301ar_ad9361_w(brd, RREG_RX1_TUNE_CTRL__RX1_PD_TUNE, 0);
	sx1301ar_ad9361_w(brd, RREG_RX2_TUNE_CTRL__RX2_PD_TUNE, 0);

	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__RX_BB_TUNE_CAL, 1); // Start calibration
	sx1301ar_wait_us(1000);
	sx1301ar_ad9361_r(brd, RREG_CALIBRATION_CTRL__RX_BB_TUNE_CAL, &r); // Check if calibration is finished
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__RX_BB_TUNE_CAL, 0);  // Disable calibration

	/* Disable tune circuit of both chains */
	sx1301ar_ad9361_w(brd, RREG_RX1_TUNE_CTRL__RX1_PD_TUNE, 1);
	sx1301ar_ad9361_w(brd, RREG_RX2_TUNE_CTRL__RX2_PD_TUNE, 1);

	/* Read filter calibration results for TIA and ADC setup */
	sx1301ar_ad9361_r(brd, RREG_RX_BBF_C3_MSB, &c3_msb);
	sx1301ar_ad9361_r(brd, RREG_RX_BBF_C3_LSB, &c3_lsb);
	sx1301ar_ad9361_r(brd, RREG_RX_BBF_R2346__RX_BBF_R2346, &r2346);
	*C3_fF = (uint32_t)c3_msb * 160 + (uint32_t)c3_lsb * 10 + 140;
	*R2346 = (uint32_t)r2346 * 18300;

	return ((int)r * -1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_tx_bb_filter_cal(uint8_t brd, uint64_t bw_khz, uint64_t f_hz)
{
	uint8_t r;
	uint64_t TuneDiv;

	/* Configure clock tuner */
	//TuneDiv = (uint16_t) ceil( (f_MHz * log(2)) / (bw_MHz * 1.6 * 2 * PI) );
	TuneDiv = f_hz / (bw_khz * 14504) + 1;
	sx1301ar_ad9361_w(brd, RREG_TX_BBF_TUNE_DIVIDER, (uint8_t)(TuneDiv & 0x00FF));
	sx1301ar_ad9361_w(brd, RREG_TX_BBF_TUNE_MODE__TX_BBF_TUNE_DIVIDER, (uint8_t)((TuneDiv & 0x0100) >> 8));
	sx1301ar_ad9361_w(brd, RREG_TX_BBF_TUNE_MODE__EVALTIME, 1);   // Maximum delay before tune comparator is sampled
	sx1301ar_ad9361_w(brd, RREG_TX_BBF_TUNE_MODE__TUNER_MODE, 7); // Algorithm stops when all comparator outputs are low
	sx1301ar_ad9361_w(brd, RREG_TX_TUNE_CTRL__TUNER_RESAMPLE, 1);

	sx1301ar_ad9361_w(brd, RREG_TX_TUNE_CTRL__PD_TUNE, 0);		  // Enable tune circuits
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__TX_BB_TUNE_CAL, 1); // Start calibration
	sx1301ar_wait_us(1000);
	sx1301ar_ad9361_r(brd, RREG_CALIBRATION_CTRL__TX_BB_TUNE_CAL, &r); // Check if calibration is finished
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__TX_BB_TUNE_CAL, 0);  // Disable calibration
	sx1301ar_ad9361_w(brd, RREG_TX_TUNE_CTRL__PD_TUNE, 1);		   // Disable tune circuit

	return ((int)r * -1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void ad9361_rx_tia_setup(uint8_t brd, uint32_t C3_fF, uint32_t R2346, uint64_t bw_khz)
{
	double CTIA_fF;
	uint8_t c_tia;

	if (bw_khz <= 3e3) {
		sx1301ar_ad9361_w(brd, RREG_RX_TIA_CONFIG__TIA_SEL_CC, 7);
	} else if (bw_khz <= 10e3) {
		sx1301ar_ad9361_w(brd, RREG_RX_TIA_CONFIG__TIA_SEL_CC, 3);
	} else {
		sx1301ar_ad9361_w(brd, RREG_RX_TIA_CONFIG__TIA_SEL_CC, 1);
	}

	CTIA_fF = (double)C3_fF * (double)R2346 * 0.56 / 3500;
	if (CTIA_fF > 2920) {
		c_tia = (uint8_t)(((CTIA_fF - 400) / 320) + 0.5);
		sx1301ar_ad9361_w(brd, RREG_TIA1_C_MSB, c_tia);
		sx1301ar_ad9361_w(brd, RREG_TIA2_C_MSB, c_tia);
		sx1301ar_ad9361_w(brd, RREG_TIA1_C_LSB, 0);
		sx1301ar_ad9361_w(brd, RREG_TIA2_C_LSB, 0);
	} else {
		c_tia = (uint8_t)(((CTIA_fF - 400) / 40) + 0.5);
		sx1301ar_ad9361_w(brd, RREG_TIA1_C_MSB, 0);
		sx1301ar_ad9361_w(brd, RREG_TIA2_C_MSB, 0);
		sx1301ar_ad9361_w(brd, RREG_TIA1_C_LSB, c_tia);
		sx1301ar_ad9361_w(brd, RREG_TIA2_C_LSB, c_tia);
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void ad9361_adc_setup(uint8_t brd, uint32_t C3_fF, uint32_t R2346, uint64_t bw_khz, uint64_t f_hz)
{
	int i;
	uint8_t buf[3] = {0, 0, 0};
	double data[40];
	double rc_timeConst;
	double scale_res, scale_cap, scale_snr;
	double max_snr;
	double bw_mhz;
	double fs_mhz;

	bw_mhz = (double)bw_khz / 1e3;
	fs_mhz = (double)f_hz / 1e6;
	if (bw_mhz < 18) {
		rc_timeConst = 1 / (1.4 * 2 * PI * (double)R2346 * (double)C3_fF * 1e-15 * bw_mhz * 1e6);
	} else {
		rc_timeConst = 1 / (1.4 * 2 * PI * (double)R2346 * (double)C3_fF * 1e-15 * bw_mhz * 1e6 * (1 + 0.01 * (bw_mhz - 18)));
	}
	scale_res = sqrt(1 / rc_timeConst);
	scale_cap = sqrt(1 / rc_timeConst);
	scale_snr = (fs_mhz < 80) ? 1 : 1.58489319246111; //10^(2/10)
	max_snr = 640 / 160;

	/* Copied from calibration guide */
	data[0] = 0;											       //address x200
	data[1] = 0;											       //address x201
	data[2] = 0;											       //address x202
	data[3] = 0x24;											       //address x203
	data[4] = 0x24;											       //address x204
	data[5] = 0;											       //addr x205
	data[6] = 0;											       //address x206
	data[7] = fmin(124, floor(-0.5 + 80 * scale_snr * scale_res * fmin(1, sqrt(max_snr * fs_mhz / 640)))); //addr x207
	data[8] = fmin(255, floor(0.5 + 20 * (640 / fs_mhz) * (data[7] / 80) / (scale_res * scale_cap)));      //addr x208
	data[10] = fmin(127, floor(-0.5 + 77 * scale_res * fmin(1, sqrt(max_snr * fs_mhz / 640))));	    //addr x20A
	data[9] = fmin(127, floor(0.8 * data[10]));							       //addr x209
	data[11] = fmin(255, floor(0.5 + 20 * (640 / fs_mhz) * (data[10] / 77) / (scale_res * scale_cap)));    //addr x20B
	data[12] = fmin(127, floor(-0.5 + 80 * scale_res * fmin(1, sqrt(max_snr * fs_mhz / 640))));	    //addr x20C
	data[13] = fmin(255, floor(-1.5 + 20 * (640 / fs_mhz) * (data[12] / 80) / (scale_res * scale_cap)));   //addr x20D
	data[14] = 21 * floor(0.1 * 640 / fs_mhz);							       //addrx20E
	data[15] = fmin(127, 1.025 * data[7]);								       //addr x20F
	data[16] = fmin(127, floor(data[15] * (0.98 + 0.02 * fmax(1, (640 / fs_mhz) / max_snr))));	     //addr x210
	data[17] = data[15];										       //addr x211
	data[18] = fmin(127, 0.975 * data[10]);								       //addr x212
	data[19] = fmin(127, floor(data[18] * (0.98 + 0.02 * fmax(1, (640 / fs_mhz) / max_snr))));	     //addr x213
	data[20] = data[18];										       //addr x214
	data[21] = fmin(127, 0.975 * data[12]);								       //addr x215
	data[22] = fmin(127, floor(data[21] * (0.98 + 0.02 * fmax(1, (640 / fs_mhz) / max_snr))));	     //addr x216
	data[23] = data[21];										       //addr x217
	data[24] = 0x2E;										       //addr x218
	data[25] = floor(128 + fmin(63, 63 * (fs_mhz / 640)));						       //addr x219
	data[26] = floor(0 + fmin(63, 63 * (fs_mhz / 640) * (0.92 + 0.08 * (640 / fs_mhz))));		       //addr x21A
	data[27] = floor(0 + fmin(63, 32 * sqrt(fs_mhz / 640)));					       //addr x21B
	data[28] = floor(128 + fmin(63, 63 * (fs_mhz / 640)));						       //addr x21C
	data[29] = floor(0 + fmin(63, 63 * (fs_mhz / 640) * (0.92 + 0.08 * (640 / fs_mhz))));		       //addr x21D
	data[30] = floor(0 + fmin(63, 32 * sqrt(fs_mhz / 640)));					       //addr x21E
	data[31] = floor(128 + fmin(63, 63 * (fs_mhz / 640)));						       //addr x21F
	data[32] = floor(0 + fmin(63, 63 * (fs_mhz / 640) * (0.92 + 0.08 * (640 / fs_mhz))));		       //addr x220
	data[33] = floor(0 + fmin(63, 63 * sqrt(fs_mhz / 640)));					       //addr x221
	data[34] = fmin(127, floor(64 * sqrt(fs_mhz / 640)));						       //addr x222
	data[35] = 0x40;										       //addr x223
	data[36] = 0x40;										       //addr x224
	data[37] = 0x2C;										       //addr x225
	data[38] = 0x00;										       //addr x226
	data[39] = 0x00;										       //addr x227

	/* Write ADC config */
	for (i = 0; i < 40; i++) {
		buf[0] = (1 << 7) | (uint8_t)((0x200 + i) >> 8);
		buf[1] = (uint8_t)((0x200 + i) & 0xFF);
		buf[2] = (uint8_t)data[i];
		sx1301ar_spibdg_w_r(brd, buf, 3, NULL, 0);
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_gain_table_write(uint8_t brd, const struct ad9361_gain_s *gain_table, uint8_t size, uint8_t rf_chain, bool read_back)
{
	int i, j;
	uint8_t r;
	struct ad9361_gain_s g;

	sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__RECEIVER_SELECT, rf_chain);
	sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__WRITE_GAIN_TABLE, 0);
	sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__START_GAIN_TABLE_CLOCK, 1);
	for (i = 0; i < size; i++) {
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_ADDRESS__GAIN_TABLE_ADDRESS, (gain_table + i)->index);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA1__LNA_GAIN, (gain_table + i)->lna);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA1__MIXER_GM_GAIN, (gain_table + i)->mixer);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA2__TIA_GAIN, (gain_table + i)->tia);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA2__LPF_GAIN, (gain_table + i)->lpf);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA3__DIGITAL_GAIN, (gain_table + i)->digital);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_WRITE_DATA3__RF_DC_CAL, (uint8_t)(gain_table + i)->calibrate);
		sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__WRITE_GAIN_TABLE, 1);
		sx1301ar_wait_us(100);
	}
	sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__WRITE_GAIN_TABLE, 0);
	sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__START_GAIN_TABLE_CLOCK, 0);

	/* Read back table */
	if (read_back) {
		for (j = 1; j <= 2; j++) {
			if (rf_chain & j) {
				sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_CONFIG__RECEIVER_SELECT, j);
				for (i = 0; i < size; i++) {
					sx1301ar_ad9361_w(brd, RREG_GAIN_TABLE_ADDRESS__GAIN_TABLE_ADDRESS, (gain_table + i)->index);
					sx1301ar_ad9361_r(brd, RREG_GAIN_TABLE_READ_DATA1, &r);
					g.lna = (r & 0x60) >> 5;
					g.mixer = r & 0x1F;
					sx1301ar_ad9361_r(brd, RREG_GAIN_TABLE_READ_DATA2, &r);
					g.tia = (r & 0x20) >> 5;
					g.lpf = r & 0x1F;
					sx1301ar_ad9361_r(brd, RREG_GAIN_TABLE_READ_DATA3, &r);
					g.calibrate = (bool)((r & 0x20) >> 5);
					g.digital = r & 0x1F;
					if ((g.lna != (gain_table + i)->lna) || (g.mixer != (gain_table + i)->mixer) || (g.tia != (gain_table + i)->tia) || (g.lpf != (gain_table + i)->lpf) || (g.digital != (gain_table + i)->digital) || (g.calibrate != (gain_table + i)->calibrate)) {
						return -1;
					}
				}
			}
		}
	}
	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_fir_set_coef(uint8_t brd, const int16_t *coef, uint8_t size, uint16_t rf_chain, bool read_back)
{
	int i, j;
	int32_t c;
	uint8_t r, r2;

	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__NUMBER_OF_TAPS, (size >> 4) - 1);
	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__SELECT_RX_CH, rf_chain);
	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__START_RX_CLOCK, 1);
	for (i = 0; i < size; i++) {
		c = (int32_t)coef[i];
		if (c < 0) {
			c = c + 65536;
		}
		sx1301ar_ad9361_w(brd, RREG_RX_FILTER_COEF_ADDR, i);
		sx1301ar_ad9361_w(brd, RREG_RX_FILTER_COEF_DATA_1, (uint8_t)(c & 0x00FF));
		sx1301ar_ad9361_w(brd, RREG_RX_FILTER_COEF_DATA_2, (uint8_t)((c & 0xFF00) >> 8));
		sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__WRITE_RX, 1);
		sx1301ar_wait_us(100);
	}
	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__WRITE_RX, 0);
	sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__START_RX_CLOCK, 0);

	/* Read back coefficients */
	if (read_back) {
		for (j = 1; j <= 2; j++) {
			if (rf_chain & j) {
				sx1301ar_ad9361_w(brd, RREG_RX_FILTER_CONFIG__SELECT_RX_CH, j);
				for (i = 0; i < size; i++) {
					sx1301ar_ad9361_w(brd, RREG_RX_FILTER_COEF_ADDR, i);
					sx1301ar_ad9361_r(brd, RREG_RX_FILTER_COEF_READ_DATA_1, &r);
					sx1301ar_ad9361_r(brd, RREG_RX_FILTER_COEF_READ_DATA_2, &r2);
					c = (int32_t)r + ((int32_t)r2 << 8);
					if (c > 32767) {
						c = c - 65536;
					}
					if ((int16_t)c != coef[i]) {
						return -1;
					}
				}
			}
		}
	}
	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_tx_fir_set_coef(uint8_t brd, const int16_t *coef, uint8_t size, uint16_t rf_chain, bool read_back)
{
	int i, j;
	int32_t c;
	uint8_t r, r2;

	sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_NUM_TAPS, (size >> 4) - 1);
	sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_SELECT, rf_chain);
	sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_START_CLK, 1);
	for (i = 0; i < size; i++) {
		c = (int32_t)coef[i];
		if (c < 0) {
			c = c + 65536;
		}
		sx1301ar_ad9361_w(brd, RREG_TX_FILTER_COEF_ADDR, i);
		sx1301ar_ad9361_w(brd, RREG_TX_FILTER_COEF_WRITE_DATA_1, (uint8_t)(c & 0x00FF));
		sx1301ar_ad9361_w(brd, RREG_TX_FILTER_COEF_WRITE_DATA_2, (uint8_t)((c & 0xFF00) >> 8));
		sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_WRITE, 1);
		sx1301ar_wait_us(100);
	}
	sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_WRITE, 0);
	sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_START_CLK, 0);

	/* Read back coefficients */
	if (read_back) {
		for (j = 1; j <= 2; j++) {
			if (rf_chain & j) {
				sx1301ar_ad9361_w(brd, RREG_TX_FILTER_CONF__FIR_SELECT, j);
				for (i = 0; i < size; i++) {
					sx1301ar_ad9361_w(brd, RREG_TX_FILTER_COEF_ADDR, i);
					sx1301ar_ad9361_r(brd, RREG_TX_FILTER_COEF_READ_DATA_1, &r);
					sx1301ar_ad9361_r(brd, RREG_TX_FILTER_COEF_READ_DATA_2, &r2);
					c = (int32_t)r + ((int32_t)r2 << 8);
					if (c > 32767) {
						c = c - 65536;
					}
					if ((int16_t)c != coef[i]) {
						return -1;
					}
				}
			}
		}
	}
	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_bb_dc_offset_cal(uint8_t brd)
{
	uint8_t r;

	sx1301ar_ad9361_w(brd, RREG_WAIT_COUNT, 255);					 // Wait time for gain to settle: set to max for better estimation
	sx1301ar_ad9361_w(brd, RREG_BB_DC_OFFSET_COUNT, 255);				 // default:63
	sx1301ar_ad9361_w(brd, RREG_BB_DC_OFFSET_SHIFT__INCREASE_COUNT_DURATION, 1);     // default:0
	sx1301ar_ad9361_w(brd, RREG_DC_OFFSET_CONFIG2__ENABLE_BB_DC_OFFSET_TRACKING, 1); // Enable tracking

	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__BBDC_CAL, 1); // Start calibration
	sx1301ar_wait_us(500000);
	sx1301ar_ad9361_r(brd, RREG_CALIBRATION_CTRL__BBDC_CAL, &r); // Check if calibration is finished
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__BBDC_CAL, 0);  // Disable calibration

	return ((int)r * -1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_rf_dc_offset_cal(uint8_t brd)
{
	uint8_t r;

	sx1301ar_ad9361_w(brd, RREG_WAIT_COUNT, 255);					 // Wait time for gain to settle: set to max for better estimation
	sx1301ar_ad9361_w(brd, RREG_RF_DC_OFFSET_COUNT, 255);				 // default:180, ad9361_rf_dc_offset_calib:50
	sx1301ar_ad9361_w(brd, RREG_RF_DC_OFFSET_CONFIG_1__RF_DC_CALIBRATION_COUNT, 15); // Number of measurement averages, default:12, ad9361_rf_dc_offset_calib:4
	sx1301ar_ad9361_w(brd, RREG_INVERT_BITS__INVERT_RX1_RF_DC_CGOUT_WORD, 1);	// default:0, ad9361_rf_dc_offset_calib:1
	sx1301ar_ad9361_w(brd, RREG_INVERT_BITS__INVERT_RX2_RF_DC_CGOUT_WORD, 1);
	sx1301ar_ad9361_w(brd, RREG_DC_OFFSET_CONFIG2__ENABLE_RF_OFFSET_TRACKING, 0);

	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__RFDC_CAL, 1); // Start calibration
	sx1301ar_wait_us(500000);
	sx1301ar_ad9361_r(brd, RREG_CALIBRATION_CTRL__RFDC_CAL, &r); // Check if calibration is finished
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CTRL__RFDC_CAL, 0);  // Disable calibration

	return ((int)r * -1);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_rx_quadrature_cal(uint8_t brd)
{
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CONFIG_1, 0); // Disable all
	sx1301ar_ad9361_w(brd, RREG_CALIBRATION_CONFIG_2__SOFT_RESET, 1);
	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_tx_quadrature_cal(uint8_t brd)
{
	sx1301ar_ad9361_w(brd, RREG_QUAD_CAL_CTRL, 0); // Disable all
	sx1301ar_ad9361_w(brd, RREG_QUAD_CAL_CTRL__QUAD_CAL_SOFT_RESET, 1);
	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#if 0 /* Not used for now */
static void ad9361_dc_offset_read( uint8_t brd, struct ad9361_offset_s * offsets )
{
    uint8_t r, r2;
    int32_t i, q;

    sx1301ar_ad9361_r( brd, RREG_INPUT_A_MSBS__INPUT_A_RX1_I, &r );
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_RX1_I, &r2 );
    i = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_MSBS__INPUT_A_RX1_Q, &r );
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_RX1_Q, &r2 );
    q = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INVERT_BITS__INVERT_RX1_RF_DC_CGIN_WORD, &r );
    if( r )
    {
        i = (~i) & 0x7FFF;
        q = (~q) & 0x7FFF;
    }
    if( i > 16383 ) {i = i - 32768;}
    if( q > 16383 ) {q = q - 32768;}
    offsets->cgin_1_i = (int16_t) i;
    offsets->cgin_1_q = (int16_t) q;

    sx1301ar_ad9361_r( brd, RREG_INPUT_A_MSBS__INPUT_A_RX2_I, &r );
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_RX2_I, &r2 );
    i = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_MSBS__INPUT_A_RX2_Q, &r );
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_RX2_Q, &r2 );
    q = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INVERT_BITS__INVERT_RX2_RF_DC_CGIN_WORD, &r );
    if( r )
    {
        i = (~i) & 0x7FFF;
        q = (~q) & 0x7FFF;
    }
    if( i > 16383 ) {i = i - 32768;}
    if( q > 16383 ) {q = q - 32768;}
    offsets->cgin_2_i = (int16_t) i;
    offsets->cgin_2_q = (int16_t) q;

    sx1301ar_ad9361_r( brd, RREG_INPUT_A_OFFSETS_1__RX1_INPUT_A_I_DC_OFFSET_MSB, &r );
    sx1301ar_ad9361_r( brd, RREG_RX1_INPUT_A_OFFSETS__RX1_INPUT_A_I_DC_OFFSET_LSB, &r2 );
    i = ((int32_t)r << 6) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_RX1_INPUT_A_OFFSETS__RX1_INPUT_A_Q_DC_OFFSET, &r );
    sx1301ar_ad9361_r( brd, RREG_RX1_INPUT_A_Q_OFFSET, &r2 );
    q = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INVERT_BITS__INVERT_RX1_RF_DC_CGOUT_WORD, &r );
    if( r )
    {
        i = (~i) & 0x3FF;
        q = (~q) & 0x3FF;
    }
    if( i > 511 ) {i = i - 1024;}
    if( q > 511 ) {q = q - 1024;}
    offsets->cgout_1_i = (int16_t) i;
    offsets->cgout_1_q = (int16_t) q;

    sx1301ar_ad9361_r( brd, RREG_RX2_INPUT_A_I_OFFSET, &r );
    sx1301ar_ad9361_r( brd, RREG_RX2_INPUT_A_OFFSETS__RX2_INPUT_A_I_DC_OFFSET, &r2 );
    i = ((int32_t)r << 2) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_RX2_INPUT_A_OFFSETS__RX2_INPUT_A_Q_DC_OFFSET_MSB, &r );
    sx1301ar_ad9361_r( brd, RREG_INPUT_A_OFFSETS_1__RX2_INPUT_A_Q_DC_OFFSET_LSB, &r2 );
    q = ((int32_t)r << 4) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_INVERT_BITS__INVERT_RX2_RF_DC_CGOUT_WORD, &r );
    if( r )
    {
        i = (~i) & 0x3FF;
        q = (~q) & 0x3FF;
    }
    if( i > 511 ) {i = i - 1024;}
    if( q > 511 ) {q = q - 1024;}
    offsets->cgout_2_i = (int16_t) i;
    offsets->cgout_2_q = (int16_t) q;

    sx1301ar_ad9361_r( brd, RREG_RX1_BB_DC_WORD_I_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_I, &r );
    sx1301ar_ad9361_r( brd, RREG_RX1_BB_DC_WORD_I_LSB, &r2 );
    i = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_RX1_BB_DC_WORD_Q_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_Q, &r );
    sx1301ar_ad9361_r( brd, RREG_RX1_BB_DC_WORD_Q_LSB, &r2 );
    q = ((int32_t)r << 8) + (int32_t)r2;
    if( i > 16383 ) {i = i - 32768;}
    if( q > 16383 ) {q = q - 32768;}
    offsets->dig_1_i = (int16_t) i;
    offsets->dig_1_q = (int16_t) q;

    sx1301ar_ad9361_r( brd, RREG_RX2_BB_DC_WORD_I_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_I, &r );
    sx1301ar_ad9361_r( brd, RREG_RX2_BB_DC_WORD_I_LSB, &r2 );
    i = ((int32_t)r << 8) + (int32_t)r2;
    sx1301ar_ad9361_r( brd, RREG_RX2_BB_DC_WORD_Q_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_Q, &r );
    sx1301ar_ad9361_r( brd, RREG_RX2_BB_DC_WORD_Q_LSB, &r2 );
    q = ((int32_t)r << 8) + (int32_t)r2;
    if( i > 16383 ) {i = i - 32768;}
    if( q > 16383 ) {q = q - 32768;}
    offsets->dig_2_i = (int16_t) i;
    offsets->dig_2_q = (int16_t) q;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

void ad9361_auxdac_set_val( uint8_t brd, uint8_t rf_chain, double voltage )
{
    uint8_t r;
    double step;
    double v_ref, v_max;
    double a, b;
    int16_t word;

    if( rf_chain == 1 )
    {
        sx1301ar_ad9361_r( brd, RREG_AUXDAC_1_CONFIG__AUXDAC1_STP_FACTOR, &r );
        step = 2 - (double)r;
    }
    else if( rf_chain == 2 )
    {
        sx1301ar_ad9361_r( brd, RREG_AUXDAC_2_CONFIG__AUXDAC2_STP_FACTOR, &r );
        step = 2 - (double)r;
    }
    else
    {
        step = 2;
    }

    // voltage = a * word + b
    a = 0;
    b = 0;
    v_ref = 0.5;
    v_max = 0;
    while( (voltage > v_max) && (v_ref < 2.5) )
    {
        v_ref = v_ref + 0.5;
        a = (0.000738 + (9e-6 * ((v_ref * 1.6) - 2))) * step;
        b = (0.97 * v_ref) - (0.3572 * step) + 0.05;
        v_max = (a * 1023) + b;
    }

    word = (int16_t) ((voltage - b) / a);
    if( word < 0 ) {word = 0;}
    if( word > 1023 ) {word = 1023;}

    if( rf_chain == 1 )
    {
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_1_CONFIG__AUXDAC_1_VREF, (uint8_t) (2*(v_ref - 1)) );
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_1_WORD, (uint8_t) (word >> 2) );
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_1_CONFIG__AUXDAC_1_WORD_LSB, (uint8_t) (word & 0x03) );
    }
    else if( rf_chain == 2 )
    {
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_2_CONFIG__AUXDAC_2_VREF, (uint8_t) (2*(v_ref - 1)) );
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_2_WORD, (uint8_t) (word >> 2) );
        sx1301ar_ad9361_w( brd, RREG_AUXDAC_2_CONFIG__AUXDAC_2_WORD_LSB, (uint8_t) (word & 0x03) );
    }
}
#endif

/* --- EOF ------------------------------------------------------------------ */
