/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Register access functions for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

#ifndef _SX1301_ARRAY_REG_H
#define _SX1301_ARRAY_REG_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdint.h> /* C99 types */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC TYPES, ENUMERATIONS AND STRUCTURES ---------------------------- */

/* SPI callbacks function pointer definition
    address  SPI register address (between 0 and 127)
    status   Pointer to return 8-bit 'status' value (first byte read); NULL = don't care about status
    size     Size of the data to transfer, use 0 to only read status
    data     Pointer to a uint8_t buffer to read or write data
The functions must return 0 if operation was a success, -1 otherwise
/!\ WARNING: types also defined in sx1301ar_hal.h /!\
*/
typedef int (*spi_r_t)(uint8_t header, uint16_t address, uint8_t *data, uint32_t size, uint8_t *status);
typedef int (*spi_w_t)(uint8_t header, uint16_t address, const uint8_t *data, uint32_t size, uint8_t *status);

/**
@brief SPI Mux headers for slave selection
*/
typedef enum {
	SPI_HDR_SX1301_0 = 0x00,
	SPI_HDR_SX1301_1 = 0x01,
	SPI_HDR_REG_FPGA = 0x02,
	SPI_HDR_SX1301_2 = 0x08,
	SPI_HDR_SX1301_3 = 0x09,
	SPI_HDR_SX1301_4 = 0x0A,
	SPI_HDR_SX1301_5 = 0x0B,
	SPI_HDR_SX1301_6 = 0x0C,
	SPI_HDR_SX1301_7 = 0x0D,
	SPI_HDR_REG_FPGA_SLAVE = 0x0E,
} sx1301ar_spi_header_t;

/**
@brief Index of board (top FPGA interface) registers
*/
typedef enum {
	BREG_SOFT_RESET = 0,
	BREG_VERSION = 1,
	BREG_ALT_CHIP_ID_LSB = 2,
	BREG_ALT_CHIP_ID_MSB = 3,
	BREG_AD9361_SPI_DATA_OUT = 4,
	BREG_AD9361_SPI_DATA_IN = 5,
	BREG_AD9361_SPI_CHIP_SELECT_POLARITY = 6,
	BREG_AD9361_SPI_CPOL = 7,
	BREG_AD9361_SPI_CPHA = 8,
	BREG_AD9361_SPI_LOOPBACK_EN = 9,
	BREG_AD9361_SPI_HOST_CTRL = 10,
	BREG_AD9361_SPI_SPEED = 11,
	BREG_AD9361_SPI_DBG_WR_CNT = 12,
	BREG_AD9361_SPI_SEL = 13,
	BREG_AD9361_SPI_DBG_STATE = 14,
	BREG_AD9361_RESETB = 15,
	BREG_AD9361_ENABLE = 16,
	BREG_AD9361_TXNRX = 17,
	BREG_AD9361_EN_AGC = 18,
	BREG_AD9361_TXRX_HOST_CTRL = 19,
	BREG_AD9361_CTRL_IN = 20,
	BREG_AD9361_CTRL_OUT = 21,
	BREG_IF_FREQ_0 = 22,
	BREG_IF_FREQ_1 = 23,
	BREG_IF_FREQ_2 = 24,
	BREG_IF_FREQ_3 = 25,
	BREG_IF_FREQ_4 = 26,
	BREG_IF_FREQ_5 = 27,
	BREG_RSSI_THRESHOLD_HIGH = 28,
	BREG_RSSI_THRESHOLD_LOW = 29,
	BREG_RSSI_DEC_FILTER_ALPHA = 30,
	BREG_RSSI_MIN_ATTEN = 31,
	BREG_RSSI_MAX_ATTEN = 32,
	BREG_RSSI_ATTEN_0 = 33,
	BREG_RSSI_ATTEN_1 = 34,
	BREG_RSSI_ATTEN_2 = 35,
	BREG_RSSI_ATTEN_3 = 36,
	BREG_RSSI_ATTEN_4 = 37,
	BREG_RSSI_ATTEN_5 = 38,
	BREG_FIR_GAIN = 39,
	BREG_OVERRIDE_GAIN = 40,
	BREG_FIR_GAIN_STEP = 41,
	BREG_DC_NOTCH_EN = 42,
	BREG_DSP_ATTEN = 43,
	BREG_SX1301_RESET = 44,
	BREG_SX1301_IQ_INTERFACE_DELAY = 45,
	BREG_SX1301_CLK_32M_EN = 46,
	BREG_SX1301_CLK_32M_INVERT = 47,
	BREG_SX1301_CLK_HS_EN = 48,
	BREG_SX1301_CLK_HS_INVERT = 49,
	BREG_SX1301_STATUS = 50,
	BREG_ENABLE_SX1301_PATH = 51,
	BREG_RF_PATH_SELECT = 52,
	BREG_SX1301_A_I_RX_SEL = 53,
	BREG_SX1301_A_Q_RX_SEL = 54,
	BREG_SX1301_B_I_RX_SEL = 55,
	BREG_SX1301_B_Q_RX_SEL = 56,
	BREG_FE_RX_A_EN = 57,
	BREG_FE_RX_B_EN = 58,
	BREG_FE_TX_A_EN = 59,
	BREG_FE_TX_B_EN = 60,
	BREG_FE_HOST_CTRL = 61,
	BREG_LNA_A_CTRL = 62,
	BREG_PA_A_CTRL = 63,
	BREG_LNA_B_CTRL = 64,
	BREG_PA_B_CTRL = 65,
	BREG_PA_GAIN = 66,
	BREG_CAPTURE_RAM_ADDR = 67,
	BREG_CAPTURE_RAM_DATA = 68,
	BREG_CAPTURE_SOURCE = 69,
	BREG_CAPTURE_START = 70,
	BREG_CAPTURE_PERIOD = 71,
	BREG_PLL_RESET = 72,
	BREG_PLL_LOCKED = 73,
	BREG_DSP_PLL_RESET = 74,
	BREG_DSP_PLL_LOCKED = 75,
	BREG_SX1301_TX_EDGE_SELECT = 76,
	BREG_SX1301_TX_SELECT = 77,
	BREG_SX1301_TX_INVERT_IQ = 78,
	BREG_SX1301_TX_CTRL = 79,
	BREG_SIGGEN_FREQ = 80,
	BREG_TX_NOTCH_EN = 81,
	BREG_TX_NOTCH_TUNE = 82,
	BREG_TX_EN = 83,
	BREG_TX_SIGGEN_EN = 84,
	BREG_TX_RF_PATH_SELECT = 85,
	BREG_TX_INVERT_CLK = 86,
	BREG_TX_GAIN = 87,
	BREG_TX_LUT_AUXDAC_8MSB = 88,
	BREG_TX_LUT_AUXDAC_2LSB = 89,
	BREG_TX_LUT_AUXDAC_VREF = 90,
	BREG_TX_LUT_AD_ATTEN_MSB = 91,
	BREG_TX_LUT_SPARE = 92,
	BREG_TX_LUT_AD_ATTEN_8LSB = 93,
	BREG_TX_LUT_FREQ_INTEGER_0 = 94,
	BREG_IF_FREQ_RSSI = 95,
	BREG_RSSI_FILTER_1_GAIN = 96,
	BREG_RSSI_FILTER_1_DC_NOTCH_EN = 97,
	BREG_RSSI = 98,
	BREG_RSSI_HISTO_RATE_DIV = 99,
	BREG_RSSI_HISTO_LEN = 100,
	BREG_RSSI_HISTO_START = 101,
	BREG_RSSI_HISTO_VALID = 102,
	BREG_RSSI_HISTO_SELECT_RF_PATH = 103,
	BREG_RSSI_STATE = 104,
	BREG_RSSI_HISTO_RAM_ADDR = 105,
	BREG_RSSI_HISTO_RAM_DATA = 106,
	BREG_RSSI_FILTER_2_GAIN = 107,
	BREG_DBG_CHAN_SELECT = 108,
	BREG_DBG_TEST_UPP = 109,
	BREG_DBG_TEST_DSP_SPI = 110,
	BREG_FPGA_GPO_SELECT = 111,
	BREG_FPGA_GPO_VAL = 112,
	BREG_DBG_DSP_CHAN_SELECT = 113,
	BREG_DSP_STATUS_PERIOD = 114,
	BREG_DSP_CFG_DIVERSITY_PARAM = 115,
	BREG_DSP_CFG_SYNC_WORD = 116,
	BREG_DSP_CFG_SPARE = 117,
	BREG_DSP_1_FIFO_WR_DATA = 118,
	BREG_DSP_1_FIFO_RD_DATA = 119,
	BREG_DSP_2_FIFO_WR_DATA = 120,
	BREG_DSP_2_FIFO_RD_DATA = 121,
	BREG_DSP_3_FIFO_WR_DATA = 122,
	BREG_DSP_3_FIFO_RD_DATA = 123,
	BREG_DSP_1_FIFO_WR_NB_BYTE = 124,
	BREG_DSP_1_FIFO_RD_NB_BYTE = 125,
	BREG_DSP_2_FIFO_WR_NB_BYTE = 126,
	BREG_DSP_2_FIFO_RD_NB_BYTE = 127,
	BREG_DSP_3_FIFO_WR_NB_BYTE = 128,
	BREG_DSP_3_FIFO_RD_NB_BYTE = 129,
	BREG_UPP_LINE_SIZE = 130,
	BREG_UPP_CLK_DELAY = 131,
	BREG_DSP1_RSTN = 132,
	BREG_DSP1_NMI = 133,
	BREG_DSP1_SPI_REG_EN = 134,
	BREG_DSP2_RSTN = 135,
	BREG_DSP2_NMI = 136,
	BREG_DSP2_SPI_REG_EN = 137,
	BREG_DSP3_RSTN = 138,
	BREG_DSP3_NMI = 139,
	BREG_DSP3_SPI_REG_EN = 140,
	BREG_DSP1_BOOT_EN = 141,
	BREG_DSP2_BOOT_EN = 142,
	BREG_DSP3_BOOT_EN = 143,
	BREG_DSP1_GPIO_IN = 144,
	BREG_DSP1_GPIO_OUT = 145,
	BREG_DSP2_GPIO_IN = 146,
	BREG_DSP2_GPIO_OUT = 147,
	BREG_DSP3_GPIO_IN = 148,
	BREG_DSP3_GPIO_OUT = 149,
	BREG_SX1301_0_IF_FREQ_0 = 150,
	BREG_SX1301_0_IF_FREQ_1 = 151,
	BREG_SX1301_0_IF_FREQ_2 = 152,
	BREG_SX1301_0_IF_FREQ_3 = 153,
	BREG_SX1301_0_IF_FREQ_4 = 154,
	BREG_SX1301_0_IF_FREQ_5 = 155,
	BREG_SX1301_0_IF_FREQ_6 = 156,
	BREG_SX1301_0_IF_FREQ_7 = 157,
	BREG_SX1301_1_IF_FREQ_0 = 158,
	BREG_SX1301_1_IF_FREQ_1 = 159,
	BREG_SX1301_1_IF_FREQ_2 = 160,
	BREG_SX1301_1_IF_FREQ_3 = 161,
	BREG_SX1301_1_IF_FREQ_4 = 162,
	BREG_SX1301_1_IF_FREQ_5 = 163,
	BREG_SX1301_1_IF_FREQ_6 = 164,
	BREG_SX1301_1_IF_FREQ_7 = 165,
	BREG_SX1301_2_IF_FREQ_0 = 166,
	BREG_SX1301_2_IF_FREQ_1 = 167,
	BREG_SX1301_2_IF_FREQ_2 = 168,
	BREG_SX1301_2_IF_FREQ_3 = 169,
	BREG_SX1301_2_IF_FREQ_4 = 170,
	BREG_SX1301_2_IF_FREQ_5 = 171,
	BREG_SX1301_2_IF_FREQ_6 = 172,
	BREG_SX1301_2_IF_FREQ_7 = 173,
	BREG_SX1301_3_IF_FREQ_0 = 174,
	BREG_SX1301_3_IF_FREQ_1 = 175,
	BREG_SX1301_3_IF_FREQ_2 = 176,
	BREG_SX1301_3_IF_FREQ_3 = 177,
	BREG_SX1301_3_IF_FREQ_4 = 178,
	BREG_SX1301_3_IF_FREQ_5 = 179,
	BREG_SX1301_3_IF_FREQ_6 = 180,
	BREG_SX1301_3_IF_FREQ_7 = 181,
	BREG_SX1301_4_IF_FREQ_0 = 182,
	BREG_SX1301_4_IF_FREQ_1 = 183,
	BREG_SX1301_4_IF_FREQ_2 = 184,
	BREG_SX1301_4_IF_FREQ_3 = 185,
	BREG_SX1301_4_IF_FREQ_4 = 186,
	BREG_SX1301_4_IF_FREQ_5 = 187,
	BREG_SX1301_4_IF_FREQ_6 = 188,
	BREG_SX1301_4_IF_FREQ_7 = 189,
	BREG_SX1301_5_IF_FREQ_0 = 190,
	BREG_SX1301_5_IF_FREQ_1 = 191,
	BREG_SX1301_5_IF_FREQ_2 = 192,
	BREG_SX1301_5_IF_FREQ_3 = 193,
	BREG_SX1301_5_IF_FREQ_4 = 194,
	BREG_SX1301_5_IF_FREQ_5 = 195,
	BREG_SX1301_5_IF_FREQ_6 = 196,
	BREG_SX1301_5_IF_FREQ_7 = 197,
	BREG_SX1301_CHANN_GAIN = 198,
	BREG_SX1301_CLIP_12B = 199,
	BREG_SX1301_0_IQ_INVERT = 200,
	BREG_SX1301_1_IQ_INVERT = 201,
	BREG_SX1301_0_FREQ_BASE = 202,
	BREG_SX1301_0_CH0_FREQ_OFFSET = 203,
	BREG_SX1301_0_CH1_FREQ_OFFSET = 204,
	BREG_SX1301_0_CH2_FREQ_OFFSET = 205,
	BREG_SX1301_0_CH3_FREQ_OFFSET = 206,
	BREG_SX1301_0_CH4_FREQ_OFFSET = 207,
	BREG_SX1301_0_CH5_FREQ_OFFSET = 208,
	BREG_SX1301_0_CH6_FREQ_OFFSET = 209,
	BREG_SX1301_0_CH7_FREQ_OFFSET = 210,
	BREG_SX1301_1_FREQ_BASE = 211,
	BREG_SX1301_1_CH0_FREQ_OFFSET = 212,
	BREG_SX1301_1_CH1_FREQ_OFFSET = 213,
	BREG_SX1301_1_CH2_FREQ_OFFSET = 214,
	BREG_SX1301_1_CH3_FREQ_OFFSET = 215,
	BREG_SX1301_1_CH4_FREQ_OFFSET = 216,
	BREG_SX1301_1_CH5_FREQ_OFFSET = 217,
	BREG_SX1301_1_CH6_FREQ_OFFSET = 218,
	BREG_SX1301_1_CH7_FREQ_OFFSET = 219,
	BREG_SX1301_2_FREQ_BASE = 220,
	BREG_SX1301_2_CH0_FREQ_OFFSET = 221,
	BREG_SX1301_2_CH1_FREQ_OFFSET = 222,
	BREG_SX1301_2_CH2_FREQ_OFFSET = 223,
	BREG_SX1301_2_CH3_FREQ_OFFSET = 224,
	BREG_SX1301_2_CH4_FREQ_OFFSET = 225,
	BREG_SX1301_2_CH5_FREQ_OFFSET = 226,
	BREG_SX1301_2_CH6_FREQ_OFFSET = 227,
	BREG_SX1301_2_CH7_FREQ_OFFSET = 228,
	BREG_SX1301_3_FREQ_BASE = 229,
	BREG_SX1301_3_CH0_FREQ_OFFSET = 230,
	BREG_SX1301_3_CH1_FREQ_OFFSET = 231,
	BREG_SX1301_3_CH2_FREQ_OFFSET = 232,
	BREG_SX1301_3_CH3_FREQ_OFFSET = 233,
	BREG_SX1301_3_CH4_FREQ_OFFSET = 234,
	BREG_SX1301_3_CH5_FREQ_OFFSET = 235,
	BREG_SX1301_3_CH6_FREQ_OFFSET = 236,
	BREG_SX1301_3_CH7_FREQ_OFFSET = 237,
	BREG_SX1301_4_FREQ_BASE = 238,
	BREG_SX1301_4_CH0_FREQ_OFFSET = 239,
	BREG_SX1301_4_CH1_FREQ_OFFSET = 240,
	BREG_SX1301_4_CH2_FREQ_OFFSET = 241,
	BREG_SX1301_4_CH3_FREQ_OFFSET = 242,
	BREG_SX1301_4_CH4_FREQ_OFFSET = 243,
	BREG_SX1301_4_CH5_FREQ_OFFSET = 244,
	BREG_SX1301_4_CH6_FREQ_OFFSET = 245,
	BREG_SX1301_4_CH7_FREQ_OFFSET = 246,
	BREG_SX1301_5_FREQ_BASE = 247,
	BREG_SX1301_5_CH0_FREQ_OFFSET = 248,
	BREG_SX1301_5_CH1_FREQ_OFFSET = 249,
	BREG_SX1301_5_CH2_FREQ_OFFSET = 250,
	BREG_SX1301_5_CH3_FREQ_OFFSET = 251,
	BREG_SX1301_5_CH4_FREQ_OFFSET = 252,
	BREG_SX1301_5_CH5_FREQ_OFFSET = 253,
	BREG_SX1301_5_CH6_FREQ_OFFSET = 254,
	BREG_SX1301_5_CH7_FREQ_OFFSET = 255,
	BREG_HIGH_SPEED_TIMESTAMP = 256,
	BREG_HIGH_SPEED_TIMESTAMP_GPS = 257,
	BREG_HIGH_SPEED_TIMESTAMP_SAMPLE = 258,
	BREG_LATCHED_SAMPLE_CNT = 259,
	BREG_ALPHA_COEFF = 260,
	BREG_BETA_COEFF = 261,
	BREG_ALPHA_BETA_COEFF_VALID = 262,
	BREG_DSP_1_FIFO_RD_DATA_CRYPTO = 263,
	BREG_DSP_1_FIFO_RD_NB_BYTE_CRYPTO = 264,
	BREG_DSP_2_FIFO_RD_DATA_CRYPTO = 265,
	BREG_DSP_2_FIFO_RD_NB_BYTE_CRYPTO = 266,
	BREG_DSP_3_FIFO_RD_DATA_CRYPTO = 267,
	BREG_DSP_3_FIFO_RD_NB_BYTE_CRYPTO = 268,
	BREG_DBG_CRYPTO_STATE = 269,
	BREG_DBG_CRYPTO_BYTE_CPT = 270,
	BREG_DBG_CRYPTO_BUFFER_0 = 271,
	BREG_HASH_DATA0 = 272,
	BREG_HASH_DATA1 = 273,
	BREG_HASH_DATA2 = 274,
	BREG_HASH_DATA3 = 275,
	BREG_HASH_NB_BYTE = 276,
	BREG_HASH_NB_WORD = 277,
	BREG_DSP_MIC_STATUS = 278,
	BREG_DSP3_MIC_STATUS = 279,
	BREG_DSP_FORCE_MIC_ENABLE = 280,
	BREG_DSP_FORCE_MIC_MODE = 281,
	BREG_DSP2_IF_FREQ_HIGH = 282,
	BREG_DSP2_IF_FREQ_LOW = 283,
	BREG_DSP2_UPP_LINE_SIZE = 284,
	BREG_DSP2_UPP_CLK_DELAY = 285,
	BREG_DSP2_RF_SELECT = 286,
	BREG_START_LBT = 287,
	BREG_LBT_FILTER_1_GAIN = 288,
	BREG_LBT_FILTER_1_DC_NOTCH_EN = 289,
	BREG_LBT_FILTER_2_GAIN = 290,
	BREG_LBT_SELECT_RF_PATH = 291,
	BREG_LBT_CHANNEL_0_IF_FREQ = 292,
	BREG_LBT_CHANNEL_1_IF_FREQ = 293,
	BREG_LBT_CHANNEL_2_IF_FREQ = 294,
	BREG_LBT_CHANNEL_3_IF_FREQ = 295,
	BREG_LBT_CHANNEL_4_IF_FREQ = 296,
	BREG_LBT_CHANNEL_5_IF_FREQ = 297,
	BREG_LBT_CHANNEL_6_IF_FREQ = 298,
	BREG_LBT_CHANNEL_7_IF_FREQ = 299,
	BREG_LBT_CHANNEL_8_IF_FREQ = 300,
	BREG_LBT_CHANNEL_9_IF_FREQ = 301,
	BREG_LBT_CHANNEL_10_IF_FREQ = 302,
	BREG_LBT_CHANNEL_11_IF_FREQ = 303,
	BREG_LBT_CHANNEL_12_IF_FREQ = 304,
	BREG_LBT_CHANNEL_13_IF_FREQ = 305,
	BREG_LBT_CHANNEL_14_IF_FREQ = 306,
	BREG_LBT_CHANNEL_15_IF_FREQ = 307,
	BREG_LBT_NB_CHANNEL = 308,
	BREG_LBT_CHANNEL_ENABLE = 309,
	BREG_LBT_NB_READ_RSSI_TIME1 = 310,
	BREG_LBT_NB_READ_RSSI_TIME2 = 311,
	BREG_LBT_CHANNEL_TIME_MODE = 312,
	BREG_LBT_RSSI_TARGET = 313,
	BREG_LBT_RSSI_START_DELAY = 314,
	BREG_LBT_RSSI_READ_DELAY = 315,
	BREG_LBT_SELECT_CH = 316,
	BREG_LBT_TIMESTAMP = 317,
	BREG_SX1301_TIMESTAMP_CPY = 318,
	BREG_SX1301_TIMESTAMP_CPY_EN = 319,
	BREG_SX1301_TIMESTAMP_CPY_GPS = 320,
	BREG_EXT_IF_EN = 321,
	BREG_EXT_IF_DELAY = 322,
	BREG_ = -1 /* force the use of a signed integer */
} sx1301ar_breg_t;

#define SX1301AR_BREG_NB 324

/**
@brief Index of chip (SX1301) registers
*/
typedef enum {
	CREG_PAGE_REG = 0,
	CREG_SOFT_RESET = 1,
	CREG_VERSION = 2,
	CREG_RX_DATA_BUF_ADDR = 3,
	CREG_RX_DATA_BUF_DATA = 4,
	CREG_TX_DATA_BUF_ADDR = 5,
	CREG_TX_DATA_BUF_DATA = 6,
	CREG_CAPTURE_RAM_ADDR = 7,
	CREG_CAPTURE_RAM_DATA = 8,
	CREG_MCU_PROM_ADDR = 9,
	CREG_MCU_PROM_DATA = 10,
	CREG_RX_PACKET_DATA_FIFO_NUM_STORED = 11,
	CREG_RX_PACKET_DATA_FIFO_ADDR_POINTER = 12,
	CREG_RX_PACKET_DATA_FIFO_STATUS = 13,
	CREG_RX_PACKET_DATA_FIFO_PAYLOAD_SIZE = 14,
	CREG_MBWSSF_MODEM_ENABLE = 15,
	CREG_CONCENTRATOR_MODEM_ENABLE = 16,
	CREG_FSK_MODEM_ENABLE = 17,
	CREG_GLOBAL_EN = 18,
	CREG_CLK32M_EN = 19,
	CREG_CLKHS_EN = 20,
	CREG_START_BIST0 = 21,
	CREG_START_BIST1 = 22,
	CREG_CLEAR_BIST0 = 23,
	CREG_CLEAR_BIST1 = 24,
	CREG_BIST0_FINISHED = 25,
	CREG_BIST1_FINISHED = 26,
	CREG_MCU_AGC_PROG_RAM_BIST_STATUS = 27,
	CREG_MCU_ARB_PROG_RAM_BIST_STATUS = 28,
	CREG_CAPTURE_RAM_BIST_STATUS = 29,
	CREG_CHAN_FIR_RAM0_BIST_STATUS = 30,
	CREG_CHAN_FIR_RAM1_BIST_STATUS = 31,
	CREG_CORR0_RAM_BIST_STATUS = 32,
	CREG_CORR1_RAM_BIST_STATUS = 33,
	CREG_CORR2_RAM_BIST_STATUS = 34,
	CREG_CORR3_RAM_BIST_STATUS = 35,
	CREG_CORR4_RAM_BIST_STATUS = 36,
	CREG_CORR5_RAM_BIST_STATUS = 37,
	CREG_CORR6_RAM_BIST_STATUS = 38,
	CREG_CORR7_RAM_BIST_STATUS = 39,
	CREG_MODEM0_RAM0_BIST_STATUS = 40,
	CREG_MODEM1_RAM0_BIST_STATUS = 41,
	CREG_MODEM2_RAM0_BIST_STATUS = 42,
	CREG_MODEM3_RAM0_BIST_STATUS = 43,
	CREG_MODEM4_RAM0_BIST_STATUS = 44,
	CREG_MODEM5_RAM0_BIST_STATUS = 45,
	CREG_MODEM6_RAM0_BIST_STATUS = 46,
	CREG_MODEM7_RAM0_BIST_STATUS = 47,
	CREG_MODEM0_RAM1_BIST_STATUS = 48,
	CREG_MODEM1_RAM1_BIST_STATUS = 49,
	CREG_MODEM2_RAM1_BIST_STATUS = 50,
	CREG_MODEM3_RAM1_BIST_STATUS = 51,
	CREG_MODEM4_RAM1_BIST_STATUS = 52,
	CREG_MODEM5_RAM1_BIST_STATUS = 53,
	CREG_MODEM6_RAM1_BIST_STATUS = 54,
	CREG_MODEM7_RAM1_BIST_STATUS = 55,
	CREG_MODEM0_RAM2_BIST_STATUS = 56,
	CREG_MODEM1_RAM2_BIST_STATUS = 57,
	CREG_MODEM2_RAM2_BIST_STATUS = 58,
	CREG_MODEM3_RAM2_BIST_STATUS = 59,
	CREG_MODEM4_RAM2_BIST_STATUS = 60,
	CREG_MODEM5_RAM2_BIST_STATUS = 61,
	CREG_MODEM6_RAM2_BIST_STATUS = 62,
	CREG_MODEM7_RAM2_BIST_STATUS = 63,
	CREG_MODEM_MBWSSF_RAM0_BIST_STATUS = 64,
	CREG_MODEM_MBWSSF_RAM1_BIST_STATUS = 65,
	CREG_MODEM_MBWSSF_RAM2_BIST_STATUS = 66,
	CREG_MCU_AGC_DATA_RAM_BIST0_STATUS = 67,
	CREG_MCU_AGC_DATA_RAM_BIST1_STATUS = 68,
	CREG_MCU_ARB_DATA_RAM_BIST0_STATUS = 69,
	CREG_MCU_ARB_DATA_RAM_BIST1_STATUS = 70,
	CREG_TX_TOP_RAM_BIST0_STATUS = 71,
	CREG_TX_TOP_RAM_BIST1_STATUS = 72,
	CREG_DATA_MNGT_RAM_BIST0_STATUS = 73,
	CREG_DATA_MNGT_RAM_BIST1_STATUS = 74,
	CREG_GPIO_SELECT_INPUT = 75,
	CREG_GPIO_SELECT_OUTPUT = 76,
	CREG_GPIO_MODE = 77,
	CREG_GPIO_PIN_REG_IN = 78,
	CREG_GPIO_PIN_REG_OUT = 79,
	CREG_MCU_AGC_STATUS = 80,
	CREG_MCU_ARB_STATUS = 81,
	CREG_CHIP_ID = 82,
	CREG_EMERGENCY_FORCE_HOST_CTRL = 83,
	CREG_RX_INVERT_IQ = 84,
	CREG_MODEM_INVERT_IQ = 85,
	CREG_MBWSSF_MODEM_INVERT_IQ = 86,
	CREG_RX_EDGE_SELECT = 87,
	CREG_MISC_RADIO_EN = 88,
	CREG_FSK_MODEM_INVERT_IQ = 89,
	CREG_FILTER_GAIN = 90,
	CREG_RADIO_SELECT = 91,
	CREG_IF_FREQ_0 = 92,
	CREG_IF_FREQ_1 = 93,
	CREG_IF_FREQ_2 = 94,
	CREG_IF_FREQ_3 = 95,
	CREG_IF_FREQ_4 = 96,
	CREG_IF_FREQ_5 = 97,
	CREG_IF_FREQ_6 = 98,
	CREG_IF_FREQ_7 = 99,
	CREG_IF_FREQ_8 = 100,
	CREG_IF_FREQ_9 = 101,
	CREG_CHANN_OVERRIDE_AGC_GAIN = 102,
	CREG_CHANN_AGC_GAIN = 103,
	CREG_CORR0_DETECT_EN = 104,
	CREG_CORR1_DETECT_EN = 105,
	CREG_CORR2_DETECT_EN = 106,
	CREG_CORR3_DETECT_EN = 107,
	CREG_CORR4_DETECT_EN = 108,
	CREG_CORR5_DETECT_EN = 109,
	CREG_CORR6_DETECT_EN = 110,
	CREG_CORR7_DETECT_EN = 111,
	CREG_CORR_SAME_PEAKS_OPTION_SF6 = 112,
	CREG_CORR_SAME_PEAKS_OPTION_SF7 = 113,
	CREG_CORR_SAME_PEAKS_OPTION_SF8 = 114,
	CREG_CORR_SAME_PEAKS_OPTION_SF9 = 115,
	CREG_CORR_SAME_PEAKS_OPTION_SF10 = 116,
	CREG_CORR_SAME_PEAKS_OPTION_SF11 = 117,
	CREG_CORR_SAME_PEAKS_OPTION_SF12 = 118,
	CREG_CORR_SIG_NOISE_RATIO_SF6 = 119,
	CREG_CORR_SIG_NOISE_RATIO_SF7 = 120,
	CREG_CORR_SIG_NOISE_RATIO_SF8 = 121,
	CREG_CORR_SIG_NOISE_RATIO_SF9 = 122,
	CREG_CORR_SIG_NOISE_RATIO_SF10 = 123,
	CREG_CORR_SIG_NOISE_RATIO_SF11 = 124,
	CREG_CORR_SIG_NOISE_RATIO_SF12 = 125,
	CREG_CORR_NUM_SAME_PEAK = 126,
	CREG_CORR_MAC_GAIN = 127,
	CREG_ADJUST_MODEM_START_OFFSET_RDX4 = 128,
	CREG_ADJUST_MODEM_START_OFFSET_SF12_RDX4 = 129,
	CREG_DBG_CORR_SELECT_SF = 130,
	CREG_DBG_CORR_SELECT_CHANNEL = 131,
	CREG_DBG_DETECT_CPT = 132,
	CREG_DBG_SYMB_CPT = 133,
	CREG_CHIRP_INVERT_RX = 134,
	CREG_DC_NOTCH_EN = 135,
	CREG_IMPLICIT_CRC_EN = 136,
	CREG_IMPLICIT_CODING_RATE = 137,
	CREG_IMPLICIT_PAYLOAD_LENGHT = 138,
	CREG_FREQ_TO_TIME_INVERT = 139,
	CREG_FREQ_TO_TIME_DRIFT = 140,
	CREG_PAYLOAD_FINE_TIMING_GAIN = 141,
	CREG_PREAMBLE_FINE_TIMING_GAIN = 142,
	CREG_TRACKING_INTEGRAL = 143,
	CREG_FRAME_SYNCH_PEAK1_POS = 144,
	CREG_FRAME_SYNCH_PEAK2_POS = 145,
	CREG_PREAMBLE_SYMB1_NB = 146,
	CREG_FRAME_SYNCH_GAIN = 147,
	CREG_SYNCH_DETECT_TH = 148,
	CREG_LLR_SCALE = 149,
	CREG_SNR_AVG_CST = 150,
	CREG_PPM_OFFSET = 151,
	CREG_MAX_PAYLOAD_LEN = 152,
	CREG_ONLY_CRC_EN = 153,
	CREG_ZERO_PAD = 154,
	CREG_DEC_GAIN_OFFSET = 155,
	CREG_CHAN_GAIN_OFFSET = 156,
	CREG_FORCE_HOST_RADIO_CTRL = 157,
	CREG_FORCE_HOST_FE_CTRL = 158,
	CREG_FORCE_DEC_FILTER_GAIN = 159,
	CREG_MCU_RST_0 = 160,
	CREG_MCU_RST_1 = 161,
	CREG_MCU_SELECT_MUX_0 = 162,
	CREG_MCU_SELECT_MUX_1 = 163,
	CREG_MCU_CORRUPTION_DETECTED_0 = 164,
	CREG_MCU_CORRUPTION_DETECTED_1 = 165,
	CREG_MCU_SELECT_EDGE_0 = 166,
	CREG_MCU_SELECT_EDGE_1 = 167,
	CREG_CHANN_SELECT_RSSI = 168,
	CREG_RSSI_BB_DEFAULT_VALUE = 169,
	CREG_RSSI_DEC_DEFAULT_VALUE = 170,
	CREG_RSSI_CHANN_DEFAULT_VALUE = 171,
	CREG_RSSI_BB_FILTER_ALPHA = 172,
	CREG_RSSI_DEC_FILTER_ALPHA = 173,
	CREG_RSSI_CHANN_FILTER_ALPHA = 174,
	CREG_IQ_MISMATCH_A_AMP_COEFF = 175,
	CREG_IQ_MISMATCH_A_PHI_COEFF = 176,
	CREG_IQ_MISMATCH_B_AMP_COEFF = 177,
	CREG_IQ_MISMATCH_B_SEL_I = 178,
	CREG_IQ_MISMATCH_B_PHI_COEFF = 179,
	CREG_TX_TRIG_IMMEDIATE = 180,
	CREG_TX_TRIG_DELAYED = 181,
	CREG_TX_TRIG_GPS = 182,
	CREG_TX_START_DELAY = 183,
	CREG_TX_FRAME_SYNCH_PEAK1_POS = 184,
	CREG_TX_FRAME_SYNCH_PEAK2_POS = 185,
	CREG_TX_RAMP_DURATION = 186,
	CREG_TX_OFFSET_I = 187,
	CREG_TX_OFFSET_Q = 188,
	CREG_TX_MODE = 189,
	CREG_TX_ZERO_PAD = 190,
	CREG_TX_EDGE_SELECT = 191,
	CREG_TX_EDGE_SELECT_TOP = 192,
	CREG_TX_GAIN = 193,
	CREG_TX_CHIRP_LOW_PASS = 194,
	CREG_TX_FCC_WIDEBAND = 195,
	CREG_TX_SWAP_IQ = 196,
	CREG_MBWSSF_IMPLICIT_HEADER = 197,
	CREG_MBWSSF_IMPLICIT_CRC_EN = 198,
	CREG_MBWSSF_IMPLICIT_CODING_RATE = 199,
	CREG_MBWSSF_IMPLICIT_PAYLOAD_LENGHT = 200,
	CREG_MBWSSF_AGC_FREEZE_ON_DETECT = 201,
	CREG_MBWSSF_FRAME_SYNCH_PEAK1_POS = 202,
	CREG_MBWSSF_FRAME_SYNCH_PEAK2_POS = 203,
	CREG_MBWSSF_PREAMBLE_SYMB1_NB = 204,
	CREG_MBWSSF_FRAME_SYNCH_GAIN = 205,
	CREG_MBWSSF_SYNCH_DETECT_TH = 206,
	CREG_MBWSSF_DETECT_MIN_SINGLE_PEAK = 207,
	CREG_MBWSSF_DETECT_TRIG_SAME_PEAK_NB = 208,
	CREG_MBWSSF_FREQ_TO_TIME_INVERT = 209,
	CREG_MBWSSF_FREQ_TO_TIME_DRIFT = 210,
	CREG_MBWSSF_PPM_CORRECTION = 211,
	CREG_MBWSSF_PAYLOAD_FINE_TIMING_GAIN = 212,
	CREG_MBWSSF_PREAMBLE_FINE_TIMING_GAIN = 213,
	CREG_MBWSSF_TRACKING_INTEGRAL = 214,
	CREG_MBWSSF_ZERO_PAD = 215,
	CREG_MBWSSF_MODEM_BW = 216,
	CREG_MBWSSF_RADIO_SELECT = 217,
	CREG_MBWSSF_RX_CHIRP_INVERT = 218,
	CREG_MBWSSF_LLR_SCALE = 219,
	CREG_MBWSSF_SNR_AVG_CST = 220,
	CREG_MBWSSF_PPM_OFFSET = 221,
	CREG_MBWSSF_RATE_SF = 222,
	CREG_MBWSSF_ONLY_CRC_EN = 223,
	CREG_MBWSSF_MAX_PAYLOAD_LEN = 224,
	CREG_TX_STATUS = 225,
	CREG_FSK_CH_BW_EXPO = 226,
	CREG_FSK_RSSI_LENGTH = 227,
	CREG_FSK_RX_INVERT = 228,
	CREG_FSK_PKT_MODE = 229,
	CREG_FSK_PSIZE = 230,
	CREG_FSK_CRC_EN = 231,
	CREG_FSK_DCFREE_ENC = 232,
	CREG_FSK_CRC_IBM = 233,
	CREG_FSK_ERROR_OSR_TOL = 234,
	CREG_FSK_RADIO_SELECT = 235,
	CREG_FSK_BR_RATIO = 236,
	CREG_FSK_REF_PATTERN_LSB = 237,
	CREG_FSK_REF_PATTERN_MSB = 238,
	CREG_FSK_PKT_LENGTH = 239,
	CREG_FSK_TX_GAUSSIAN_EN = 240,
	CREG_FSK_TX_GAUSSIAN_SELECT_BT = 241,
	CREG_FSK_TX_PATTERN_EN = 242,
	CREG_FSK_TX_PREAMBLE_SEQ = 243,
	CREG_FSK_TX_PSIZE = 244,
	CREG_FSK_NODE_ADRS = 245,
	CREG_FSK_BROADCAST = 246,
	CREG_FSK_AUTO_AFC_ON = 247,
	CREG_FSK_PATTERN_TIMEOUT_CFG = 248,
	CREG_SPI_RADIO_A__DATA = 249,
	CREG_SPI_RADIO_A__DATA_READBACK = 250,
	CREG_SPI_RADIO_A__ADDR = 251,
	CREG_SPI_RADIO_A__CS = 252,
	CREG_SPI_RADIO_B__DATA = 253,
	CREG_SPI_RADIO_B__DATA_READBACK = 254,
	CREG_SPI_RADIO_B__ADDR = 255,
	CREG_SPI_RADIO_B__CS = 256,
	CREG_RADIO_A_EN = 257,
	CREG_RADIO_B_EN = 258,
	CREG_RADIO_RST = 259,
	CREG_LNA_A_EN = 260,
	CREG_PA_A_EN = 261,
	CREG_LNA_B_EN = 262,
	CREG_PA_B_EN = 263,
	CREG_PA_GAIN = 264,
	CREG_LNA_A_CTRL_LUT = 265,
	CREG_PA_A_CTRL_LUT = 266,
	CREG_LNA_B_CTRL_LUT = 267,
	CREG_PA_B_CTRL_LUT = 268,
	CREG_CAPTURE_SOURCE = 269,
	CREG_CAPTURE_START = 270,
	CREG_CAPTURE_FORCE_TRIGGER = 271,
	CREG_CAPTURE_WRAP = 272,
	CREG_CAPTURE_PERIOD = 273,
	CREG_MODEM_STATUS = 274,
	CREG_VALID_HEADER_COUNTER_0 = 275,
	CREG_VALID_PACKET_COUNTER_0 = 276,
	CREG_VALID_HEADER_COUNTER_MBWSSF = 277,
	CREG_VALID_HEADER_COUNTER_FSK = 278,
	CREG_VALID_PACKET_COUNTER_MBWSSF = 279,
	CREG_VALID_PACKET_COUNTER_FSK = 280,
	CREG_CHANN_RSSI = 281,
	CREG_BB_RSSI = 282,
	CREG_DEC_RSSI = 283,
	CREG_DBG_MCU_DATA = 284,
	CREG_DBG_ARB_MCU_RAM_DATA = 285,
	CREG_DBG_AGC_MCU_RAM_DATA = 286,
	CREG_NEXT_PACKET_CNT = 287,
	CREG_ADDR_CAPTURE_COUNT = 288,
	CREG_TIMESTAMP = 289,
	CREG_DBG_CHANN0_GAIN = 290,
	CREG_DBG_CHANN1_GAIN = 291,
	CREG_DBG_CHANN2_GAIN = 292,
	CREG_DBG_CHANN3_GAIN = 293,
	CREG_DBG_CHANN4_GAIN = 294,
	CREG_DBG_CHANN5_GAIN = 295,
	CREG_DBG_CHANN6_GAIN = 296,
	CREG_DBG_CHANN7_GAIN = 297,
	CREG_DBG_DEC_FILT_GAIN = 298,
	CREG_SPI_DATA_FIFO_PTR = 299,
	CREG_PACKET_DATA_FIFO_PTR = 300,
	CREG_DBG_ARB_MCU_RAM_ADDR = 301,
	CREG_DBG_AGC_MCU_RAM_ADDR = 302,
	CREG_SPI_MASTER_CHIP_SELECT_POLARITY = 303,
	CREG_SPI_MASTER_CPOL = 304,
	CREG_SPI_MASTER_CPHA = 305,
	CREG_SIG_GEN_ANALYSER_MUX_SEL = 306,
	CREG_SIG_GEN_EN = 307,
	CREG_SIG_ANALYSER_EN = 308,
	CREG_SIG_ANALYSER_AVG_LEN = 309,
	CREG_SIG_ANALYSER_PRECISION = 310,
	CREG_SIG_ANALYSER_VALID_OUT = 311,
	CREG_SIG_GEN_FREQ = 312,
	CREG_SIG_ANALYSER_FREQ = 313,
	CREG_SIG_ANALYSER_I_OUT = 314,
	CREG_SIG_ANALYSER_Q_OUT = 315,
	CREG_GPS_EN = 316,
	CREG_GPS_POL = 317,
	CREG_SW_TEST_REG1 = 318,
	CREG_SW_TEST_REG2 = 319,
	CREG_SW_TEST_REG3 = 320,
	CREG_DATA_MNGT_STATUS = 321,
	CREG_DATA_MNGT_CPT_FRAME_ALLOCATED = 322,
	CREG_DATA_MNGT_CPT_FRAME_FINISHED = 323,
	CREG_DATA_MNGT_CPT_FRAME_READEN = 324,
	CREG_TX_TRIG_ALL = 325,
	CREG_ = -1 /* force the use of a signed integer */
} sx1301ar_creg_t;

#define SX1301AR_CREG_NB 326

/**
@brief Address of radio (AD9361) registers
*/
typedef enum {
	RREG_SPI_CONF = 0,
	RREG_SPI_CONF__SOFT_RESET = 1,
	RREG_SPI_CONF__WIRE3_SPI = 2,
	RREG_SPI_CONF__LSB_FIRST = 3,
	RREG_SPI_CONF___LSB_FIRST = 4,
	RREG_SPI_CONF___WIRE3_SPI = 5,
	RREG_SPI_CONF___SOFT_RESET = 6,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL = 7,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__TX2_MONITOR_ENABLE = 8,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__TX1_MONITOR_ENABLE = 9,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_RF_ENABLE = 10,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_BBPLL_ENABLE = 11,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_DIGITAL_CLK_ENABLE = 12,
	RREG_MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_BB_ENABLE = 13,
	RREG_TX_ENABLE_FILTER_CTRL = 14,
	RREG_TX_ENABLE_FILTER_CTRL__THB2_EN = 15,
	RREG_TX_ENABLE_FILTER_CTRL__THB1_EN = 16,
	RREG_TX_ENABLE_FILTER_CTRL__TX_CHANNEL_ENABLE = 17,
	RREG_TX_ENABLE_FILTER_CTRL__THB3_ENABLE_INTERP = 18,
	RREG_TX_ENABLE_FILTER_CTRL__TX_FIR_ENABLE_INTERPOLATION = 19,
	RREG_RX_ENABLE_FILTER_CTRL = 20,
	RREG_RX_ENABLE_FILTER_CTRL__RHB2_EN = 21,
	RREG_RX_ENABLE_FILTER_CTRL__RHB1_EN = 22,
	RREG_RX_ENABLE_FILTER_CTRL__RX_CHANNEL_ENABLE = 23,
	RREG_RX_ENABLE_FILTER_CTRL__DEC3_ENABLE_DECIMATION = 24,
	RREG_RX_ENABLE_FILTER_CTRL__RX_FIR_ENABLE_DECIMATION = 25,
	RREG_INPUT_SELECT = 26,
	RREG_INPUT_SELECT__TX_OUTPUT = 27,
	RREG_INPUT_SELECT__RX_INPUT = 28,
	RREG_RFPLL_DIVIDERS = 29,
	RREG_RFPLL_DIVIDERS__TX_VCO_DIVIDER = 30,
	RREG_RFPLL_DIVIDERS__RX_VCO_DIVIDER = 31,
	RREG_RX_CLOCK_DATA_DELAY = 32,
	RREG_RX_CLOCK_DATA_DELAY__DATA_CLK_DELAY = 33,
	RREG_RX_CLOCK_DATA_DELAY__RX_DATA_DELAY = 34,
	RREG_TX_CLOCK_DATA_DELAY = 35,
	RREG_TX_CLOCK_DATA_DELAY__FB_CLK_DELAY = 36,
	RREG_TX_CLOCK_DATA_DELAY__TX_DATA_DELAY = 37,
	RREG_CLOCK_ENABLE = 38,
	RREG_CLOCK_ENABLE__XO_BYPASS = 39,
	RREG_CLOCK_ENABLE__DIGITAL_POWER_UP = 40,
	RREG_CLOCK_ENABLE__CLOCK_ENABLE_DFLT = 41,
	RREG_CLOCK_ENABLE__BBPLL_ENABLE = 42,
	RREG_BBPLL = 43,
	RREG_BBPLL__CLKOUT_ENABLE = 44,
	RREG_BBPLL__DAC_CLK_DIV2 = 45,
	RREG_BBPLL__CLKOUT_SELECT = 46,
	RREG_BBPLL__BBPLL_DIVIDER = 47,
	RREG_TEMP_OFFSET = 48,
	RREG_START_TEMP_READING = 49,
	RREG_START_TEMP_READING__START_TEMP_READING = 50,
	RREG_TEMP_SENSE2 = 51,
	RREG_TEMP_SENSE2__TEMP_SENSE_PERIODIC_ENABLE = 52,
	RREG_TEMP_SENSE2__MEASUREMENT_TIME_INTERVAL = 53,
	RREG_TEMPERATURE = 54,
	RREG_TEMP_SENSOR_CONFIG = 55,
	RREG_TEMP_SENSOR_CONFIG__TEMP_SENSOR_DECIMATION = 56,
	RREG_PARALLEL_PORT_CONF_1 = 57,
	RREG_PARALLEL_PORT_CONF_1__PP_TX_SWAP_IQ = 58,
	RREG_PARALLEL_PORT_CONF_1__PP_RX_SWAP_IQ = 59,
	RREG_PARALLEL_PORT_CONF_1__TX_CHANNEL_SWAP = 60,
	RREG_PARALLEL_PORT_CONF_1__RX_CHANNEL_SWAP = 61,
	RREG_PARALLEL_PORT_CONF_1__RX_FRAME_PULSE_MODE = 62,
	RREG_PARALLEL_PORT_CONF_1__R2T2_TIMING = 63,
	RREG_PARALLEL_PORT_CONF_1__INVERT_DATA_BUS = 64,
	RREG_PARALLEL_PORT_CONF_1__INVERT_DATA_CLK = 65,
	RREG_PARALLEL_PORT_CONF_2 = 66,
	RREG_PARALLEL_PORT_CONF_2__FDD_ALT_WORD_ORDER = 67,
	RREG_PARALLEL_PORT_CONF_2__INVERT_RX1 = 68,
	RREG_PARALLEL_PORT_CONF_2__INVERT_RX2 = 69,
	RREG_PARALLEL_PORT_CONF_2__INVERT_TX1 = 70,
	RREG_PARALLEL_PORT_CONF_2__INVERT_TX2 = 71,
	RREG_PARALLEL_PORT_CONF_2__INVERT_RX_FRAME = 72,
	RREG_PARALLEL_PORT_CONF_2__DELAY_RX_DATA = 73,
	RREG_PARALLEL_PORT_CONF_3 = 74,
	RREG_PARALLEL_PORT_CONF_3__FDD_RX_RATE_2TX_RATE = 75,
	RREG_PARALLEL_PORT_CONF_3__SWAP_PORTS = 76,
	RREG_PARALLEL_PORT_CONF_3__SINGLE_DATA_RATE = 77,
	RREG_PARALLEL_PORT_CONF_3__LVDS_MODE = 78,
	RREG_PARALLEL_PORT_CONF_3__HALF_DUPLEX_MODE = 79,
	RREG_PARALLEL_PORT_CONF_3__SINGLE_PORT_MODE = 80,
	RREG_PARALLEL_PORT_CONF_3__FULL_PORT = 81,
	RREG_PARALLEL_PORT_CONF_3__FULL_DUPLEX_SWAP_BITS = 82,
	RREG_ENSM_MODE = 83,
	RREG_ENSM_MODE__FDD_MODE = 84,
	RREG_ENSM_CONFIG_1 = 85,
	RREG_ENSM_CONFIG_1__ENABLE_RX_DATA_PORT_FOR_CAL = 86,
	RREG_ENSM_CONFIG_1__FORCE_RX_ON = 87,
	RREG_ENSM_CONFIG_1__FORCE_TX_ON = 88,
	RREG_ENSM_CONFIG_1__ENABLE_ENSM_PIN_CTRL = 89,
	RREG_ENSM_CONFIG_1__LEVEL_MODE = 90,
	RREG_ENSM_CONFIG_1__FORCE_ALERT_STATE = 91,
	RREG_ENSM_CONFIG_1__AUTO_GAIN_LOCK = 92,
	RREG_ENSM_CONFIG_1__TO_ALERT = 93,
	RREG_ENSM_CONFIG_2 = 94,
	RREG_ENSM_CONFIG_2__FDD_EXTERNAL_CTRL_ENABLE = 95,
	RREG_ENSM_CONFIG_2__POWER_DOWN_RX_SYNTH = 96,
	RREG_ENSM_CONFIG_2__POWER_DOWN_TX_SYNTH = 97,
	RREG_ENSM_CONFIG_2__TXNRX_SPI_CTRL = 98,
	RREG_ENSM_CONFIG_2__SYNTH_ENABLE_PIN_CTRL_MODE = 99,
	RREG_ENSM_CONFIG_2__DUAL_SYNTH_MODE = 100,
	RREG_ENSM_CONFIG_2__RX_SYNTH_READY_MASK = 101,
	RREG_ENSM_CONFIG_2__TX_SYNTH_READY_MASK = 102,
	RREG_CALIBRATION_CTRL = 103,
	RREG_CALIBRATION_CTRL__RX_BB_TUNE_CAL = 104,
	RREG_CALIBRATION_CTRL__TX_BB_TUNE_CAL = 105,
	RREG_CALIBRATION_CTRL__RX_QUAD_CAL = 106,
	RREG_CALIBRATION_CTRL__TX_QUAD_CAL = 107,
	RREG_CALIBRATION_CTRL__RX_GAIN_STEP_CAL = 108,
	RREG_CALIBRATION_CTRL__TXMON_CAL = 109,
	RREG_CALIBRATION_CTRL__RFDC_CAL = 110,
	RREG_CALIBRATION_CTRL__BBDC_CAL = 111,
	RREG_STATE = 112,
	RREG_STATE__CALIBRATION_SEQUENCE_STATE = 113,
	RREG_STATE__ENSM_STATE = 114,
	RREG_AUXDAC_1_WORD = 115,
	RREG_AUXDAC_2_WORD = 116,
	RREG_AUXDAC_1_CONFIG = 117,
	RREG_AUXDAC_1_CONFIG__COMP_CTRL_1 = 118,
	RREG_AUXDAC_1_CONFIG__AUXDAC1_STP_FACTOR = 119,
	RREG_AUXDAC_1_CONFIG__AUXDAC_1_VREF = 120,
	RREG_AUXDAC_1_CONFIG__AUXDAC_1_WORD_LSB = 121,
	RREG_AUXDAC_2_CONFIG = 122,
	RREG_AUXDAC_2_CONFIG__COMP_CTRL_2 = 123,
	RREG_AUXDAC_2_CONFIG__AUXDAC2_STP_FACTOR = 124,
	RREG_AUXDAC_2_CONFIG__AUXDAC_2_VREF = 125,
	RREG_AUXDAC_2_CONFIG__AUXDAC_2_WORD_LSB = 126,
	RREG_AUXADC_CLOCK_DIVIDER = 127,
	RREG_AUXADC_CLOCK_DIVIDER__AUXADC_CLOCK_DIVIDER = 128,
	RREG_AUXADC_CONFIG = 129,
	RREG_AUXADC_CONFIG__AUXADC_POWER_DOWN = 130,
	RREG_AUXADC_CONFIG__AUX_ADC_DECIMATION = 131,
	RREG_AUXADC_WORD_MSB = 132,
	RREG_AUXADC_LSB = 133,
	RREG_AUXADC_LSB__AUXADC_WORD_LSB = 134,
	RREG_AUTO_GPO = 135,
	RREG_AUTO_GPO__GPO_ENABLE_AUTO_RX = 136,
	RREG_AUTO_GPO__GPO_ENABLE_AUTO_TX = 137,
	RREG_AGC_GAIN_LOCK_DELAY = 138,
	RREG_AGC_ATTACK_DELAY = 139,
	RREG_AGC_ATTACK_DELAY__INVERT_BYPASSED_LNA_POLARITY = 140,
	RREG_AGC_ATTACK_DELAY__AGC_ATTACK_DELAY = 141,
	RREG_AUXDAC_ENABLE_CTRL = 142,
	RREG_AUXDAC_ENABLE_CTRL__AUXDAC_MANUAL_BAR = 143,
	RREG_AUXDAC_ENABLE_CTRL__AUXDAC_AUTO_TX_BAR = 144,
	RREG_AUXDAC_ENABLE_CTRL__AUXDAC_AUTO_RX_BAR = 145,
	RREG_AUXDAC_ENABLE_CTRL__AUXDAC_INIT_BAR = 146,
	RREG_RX_LOAD_SYNTH_DELAY = 147,
	RREG_TX_LOAD_SYNTH_DELAY = 148,
	RREG_EXTERNAL_LNA_CTRL = 149,
	RREG_EXTERNAL_LNA_CTRL__AUXDAC_MANUAL_SELECT = 150,
	RREG_EXTERNAL_LNA_CTRL__EXTERNAL_LNA2_CTRL = 151,
	RREG_EXTERNAL_LNA_CTRL__EXTERNAL_LNA1_CTRL = 152,
	RREG_EXTERNAL_LNA_CTRL__GPO_MANUAL_SELECT = 153,
	RREG_EXTERNAL_LNA_CTRL__OPEN = 154,
	RREG_GPO_FORCE_AND_INIT = 155,
	RREG_GPO_FORCE_AND_INIT__GPO_MANUAL_CTRL = 156,
	RREG_GPO_FORCE_AND_INIT__GPO_INIT_STATE = 157,
	RREG_GPO0_RX_DELAY = 158,
	RREG_GPO1_RX_DELAY = 159,
	RREG_GPO2_RX_DELAY = 160,
	RREG_GPO3_RX_DELAY = 161,
	RREG_GPO0_TX_DELAY = 162,
	RREG_GPO1_TX_DELAY = 163,
	RREG_GPO2_TX_DELAY = 164,
	RREG_GPO3_TX_DELAY = 165,
	RREG_AUXDAC1_RX_DELAY = 166,
	RREG_AUXDAC1_TX_DELAY = 167,
	RREG_AUXDAC2_RX_DELAY = 168,
	RREG_AUXDAC2_TX_DELAY = 169,
	RREG_CTRL_OUTPUT_POINTER = 170,
	RREG_CTRL_OUTPUT_ENABLE = 171,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL7 = 172,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL6 = 173,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL5 = 174,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL4 = 175,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL3 = 176,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL2 = 177,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL1 = 178,
	RREG_CTRL_OUTPUT_ENABLE__EN_CTRL0 = 179,
	RREG_PRODUCT_ID = 180,
	RREG_REFERENCE_CLOCK_CYCLES = 181,
	RREG_REFERENCE_CLOCK_CYCLES__REFERENCE_CLOCK_CYCLES_PER_US = 182,
	RREG_DIGITAL_IO_CTRL = 183,
	RREG_DIGITAL_IO_CTRL__CLK_OUT_DRIVE = 184,
	RREG_DIGITAL_IO_CTRL__DATACLK_DRIVE = 185,
	RREG_DIGITAL_IO_CTRL__DATA_PORT_DRIVE = 186,
	RREG_DIGITAL_IO_CTRL__DATACLK_SLEW = 187,
	RREG_DIGITAL_IO_CTRL__DATA_PORT_SLEW = 188,
	RREG_LVDS_BIAS_CTRL = 189,
	RREG_LVDS_BIAS_CTRL__RX_ON_CHIP_TERM = 190,
	RREG_LVDS_BIAS_CTRL__LVDS_BYPASS_BIAS_R = 191,
	RREG_LVDS_BIAS_CTRL__LVDS_TX_LO_VCM = 192,
	RREG_LVDS_BIAS_CTRL__CLK_OUT_SLEW = 193,
	RREG_LVDS_BIAS_CTRL__LVDS_BIAS = 194,
	RREG_LVDS_INVERT_CTRL1 = 195,
	RREG_LVDS_INVERT_CTRL2 = 196,
	RREG_SDM_CTRL_1 = 197,
	RREG_SDM_CTRL_1__INIT_BB_FO_CAL = 198,
	RREG_SDM_CTRL_1__BBPLL_RESET_BAR = 199,
	RREG_FRACT_BB_FREQ_WORD_1 = 200,
	RREG_FRACT_BB_FREQ_WORD_2 = 201,
	RREG_FRACT_BB_FREQ_WORD_3 = 202,
	RREG_INTEGER_BB_FREQ_WORD = 203,
	RREG_CLOCK_CTRL = 204,
	RREG_CLOCK_CTRL__REF_FREQ_SCALER = 205,
	RREG_CP_CURRENT = 206,
	RREG_CP_CURRENT__CHARGE_PUMP_CURRENT = 207,
	RREG_CP_BLEED_CURRENT = 208,
	RREG_CP_BLEED_CURRENT__MCS_REFCLK_SCALE_EN = 209,
	RREG_LOOP_FILTER_1 = 210,
	RREG_LOOP_FILTER_1__C1_WORD = 211,
	RREG_LOOP_FILTER_1__R1_WORD = 212,
	RREG_LOOP_FILTER_2 = 213,
	RREG_LOOP_FILTER_2__R2_WORD = 214,
	RREG_LOOP_FILTER_2__C2_WORD = 215,
	RREG_LOOP_FILTER_2__C1_WORD_LSB = 216,
	RREG_LOOP_FILTER_3 = 217,
	RREG_LOOP_FILTER_3__BYPASS_C3 = 218,
	RREG_LOOP_FILTER_3__BYPASS_R2 = 219,
	RREG_LOOP_FILTER_3__C3_WORD = 220,
	RREG_LOOP_FILTER_3__R2_WORD_LSB = 221,
	RREG_VCO_CTRL = 222,
	RREG_VCO_CTRL__FREQ_CAL_ENABLE = 223,
	RREG_VCO_CTRL__FREQ_CAL_RESET = 224,
	RREG_VCO_CTRL__FREQ_CAL_COUNT_LENGTH = 225,
	RREG_VCO_PROGRAM_1 = 226,
	RREG_VCO_PROGRAM_2 = 227,
	RREG_SDM_CTRL = 228,
	RREG_SDM_CTRL__CAL_CLOCK_DIV_4 = 229,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE = 230,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE__RX_LO_POWER_DOWN = 231,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_ALC_POWER_DOWN = 232,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_PTAT_POWER_DOWN = 233,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_POWER_DOWN = 234,
	RREG_RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_LDO_POWER_DOWN = 235,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE = 236,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE__TX_LO_POWER_DOWN = 237,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_ALC_POWER_DOWN = 238,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_PTAT_POWER_DOWN = 239,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_POWER_DOWN = 240,
	RREG_TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_LDO_POWER_DOWN = 241,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_1 = 242,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_OFFSET_DAC_CGIN_POWER_DOWN = 243,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_LMT_OVERLOAD_POWER_DOWN = 244,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_MIXER_GM_POWER_DOWN = 245,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_CGB_POWER_DOWN = 246,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_2 = 247,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_BBF_POWER_DOWN = 248,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_TIA_POWER_DOWN = 249,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_MIXER_POWER_DOWN = 250,
	RREG_RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_OFFSET_DAC_CGOUT_POWER_DOWN = 251,
	RREG_RX1_ADC_POWER_DOWN_OVERRIDE = 252,
	RREG_RX2_ADC_POWER_DOWN_OVERRIDE = 253,
	RREG_TX_ANALOG_POWER_DOWN_OVERRIDE_1 = 254,
	RREG_TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_SECONDARY_FILTER_POWER_DOWN = 255,
	RREG_TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_BBF_POWER_DOWN = 256,
	RREG_TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_DAC_POWER_DOWN = 257,
	RREG_TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_DAC_BIAS_POWER_DOWN = 258,
	RREG_ANALOG_POWER_DOWN_OVERRIDE = 259,
	RREG_ANALOG_POWER_DOWN_OVERRIDE__RX_EXT_VCO_BUFFER_POWER_DOWN = 260,
	RREG_ANALOG_POWER_DOWN_OVERRIDE__TX_EXT_VCO_BUFFER_POWER_DOWN = 261,
	RREG_ANALOG_POWER_DOWN_OVERRIDE__TX_MONITOR_POWER_DOWN = 262,
	RREG_ANALOG_POWER_DOWN_OVERRIDE__TX_UPCONVERTER_POWER_DOWN = 263,
	RREG_MISC_POWER_DOWN_OVERRIDE = 264,
	RREG_MISC_POWER_DOWN_OVERRIDE__RX_LNA_POWER_DOWN = 265,
	RREG_MISC_POWER_DOWN_OVERRIDE__DCXO_POWER_DOWN = 266,
	RREG_MISC_POWER_DOWN_OVERRIDE__MASTER_BIAS_POWER_DOWN = 267,
	RREG_MISC_POWER_DOWN_OVERRIDE__RX_CALIBRATION_POWER_DOWN = 268,
	RREG_CH_1_OVERFLOW = 269,
	RREG_CH_1_OVERFLOW__BBPLL_LOCK = 270,
	RREG_CH_1_OVERFLOW__CH_1_INT3 = 271,
	RREG_CH_1_OVERFLOW__CH1_HB3 = 272,
	RREG_CH_1_OVERFLOW__CH1_HB2 = 273,
	RREG_CH_1_OVERFLOW__CH1_QEC = 274,
	RREG_CH_1_OVERFLOW__CH1_HB1 = 275,
	RREG_CH_1_OVERFLOW__CH1_TFIR = 276,
	RREG_CH_1_OVERFLOW__CH1_RFIR = 277,
	RREG_CH_2_OVERFLOW = 278,
	RREG_CH_2_OVERFLOW__CH2_INT3 = 279,
	RREG_CH_2_OVERFLOW__CH2_HB3 = 280,
	RREG_CH_2_OVERFLOW__CH2_HB2 = 281,
	RREG_CH_2_OVERFLOW__CH2_QEC = 282,
	RREG_CH_2_OVERFLOW__CH2_HB1 = 283,
	RREG_CH_2_OVERFLOW__CH2_TFIR = 284,
	RREG_CH_2_OVERFLOW__CH2_RFIR = 285,
	RREG_TX_FILTER_COEF_ADDR = 286,
	RREG_TX_FILTER_COEF_WRITE_DATA_1 = 287,
	RREG_TX_FILTER_COEF_WRITE_DATA_2 = 288,
	RREG_TX_FILTER_COEF_READ_DATA_1 = 289,
	RREG_TX_FILTER_COEF_READ_DATA_2 = 290,
	RREG_TX_FILTER_CONF = 291,
	RREG_TX_FILTER_CONF__TX_FIR_GAIN_6DB = 292,
	RREG_TX_FILTER_CONF__FIR_START_CLK = 293,
	RREG_TX_FILTER_CONF__FIR_WRITE = 294,
	RREG_TX_FILTER_CONF__FIR_SELECT = 295,
	RREG_TX_FILTER_CONF__FIR_NUM_TAPS = 296,
	RREG_TX_MON_LOW_GAIN = 297,
	RREG_TX_MON_LOW_GAIN__TX_MON_TRACK = 298,
	RREG_TX_MON_LOW_GAIN__TX_MON_LOW_GAIN = 299,
	RREG_TX_MON_HIGH_GAIN = 300,
	RREG_TX_MON_HIGH_GAIN__TX_MON_HIGH_GAIN = 301,
	RREG_TX_MON_DELAY = 302,
	RREG_TX_LEVEL_THRESH = 303,
	RREG_TX_LEVEL_THRESH__TX_LEVEL_THRESH = 304,
	RREG_TX_LEVEL_THRESH__TX_MON_DELAY_COUNTER = 305,
	RREG_TX_RSSI1 = 306,
	RREG_TX_RSSI2 = 307,
	RREG_TX_RSSI_LSB = 308,
	RREG_TX_RSSI_LSB__TX_RSSI_2 = 309,
	RREG_TX_RSSI_LSB__TX_RSSI_1 = 310,
	RREG_TPM_MODE_ENABLE = 311,
	RREG_TPM_MODE_ENABLE__ONE_SHOT_MODE = 312,
	RREG_TPM_MODE_ENABLE__TX_MON_DURATION = 313,
	RREG_TX_MON_TEMP_GAIN_COEF = 314,
	RREG_TX_MON_1_CONFIG = 315,
	RREG_TX_MON_1_CONFIG__TX_MON_1_LO_CM = 316,
	RREG_TX_MON_1_CONFIG__TX_MON_1_GAIN = 317,
	RREG_TX_MON_2_CONFIG = 318,
	RREG_TX_MON_2_CONFIG__TX_MON_2_LO_CM = 319,
	RREG_TX_MON_2_CONFIG__TX_MON_2_GAIN = 320,
	RREG_TX1_ATTEN_0 = 321,
	RREG_TX1_ATTEN_1 = 322,
	RREG_TX1_ATTEN_1__TX_1_ATTEN = 323,
	RREG_TX2_ATTEN_0 = 324,
	RREG_TX2_ATTEN_1 = 325,
	RREG_TX2_ATTEN_1__TX_2_ATTEN = 326,
	RREG_TX_ATTEN_OFFSET = 327,
	RREG_TX_ATTEN_OFFSET__MASK_CLR_ATTEN_UPDATE = 328,
	RREG_TX_ATTEN_OFFSET__TX_ATTEN_OFFSET = 329,
	RREG_TX_ATTEN_THRESH = 330,
	RREG_TX1_DIG_ATTEN = 331,
	RREG_TX1_DIG_ATTEN__SEL_TX1_TX2 = 332,
	RREG_TX2_DIG_ATTEN = 333,
	RREG_TX2_DIG_ATTEN__IMMEDIATELY_UPDATE_TPC_ATTEN = 334,
	RREG_TX1_SYMBOL_ATTEN = 335,
	RREG_TX1_SYMBOL_ATTEN__TX_1_SYMBOL_ATTEN = 336,
	RREG_TX2_SYMBOL_ATTEN = 337,
	RREG_TX2_SYMBOL_ATTEN__TX_2_SYMBOL_ATTEN = 338,
	RREG_TX_SYMBOL_ATTEN_CONFIG = 339,
	RREG_TX_SYMBOL_ATTEN_CONFIG__USE_TX1_PIN_SYMBOL_ATTEN = 340,
	RREG_TX_SYMBOL_ATTEN_CONFIG__USE_CTRL_IN_FOR_SYMBOL_ATTEN = 341,
	RREG_TX_SYMBOL_ATTEN_CONFIG__ENABLE_SYMBOL_ATTEN = 342,
	RREG_TX1_OUT_1_PHASE_CORR = 343,
	RREG_TX1_OUT_1_GAIN_CORR = 344,
	RREG_TX2_OUT_1_PHASE_CORR = 345,
	RREG_TX2_OUT_1_GAIN_CORR = 346,
	RREG_TX1_OUT_1_OFFSET_I = 347,
	RREG_TX1_OUT_1_OFFSET_Q = 348,
	RREG_TX2_OUT_1_OFFSET_I = 349,
	RREG_TX2_OUT_1_OFFSET_Q = 350,
	RREG_TX1_OUT_2_PHASE_CORR = 351,
	RREG_TX1_OUT_2_GAIN_CORR = 352,
	RREG_TX2_OUT_2_PHASE_CORR = 353,
	RREG_TX2_OUT_2_GAIN_CORR = 354,
	RREG_TX1_OUT_2_OFFSET_I = 355,
	RREG_TX1_OUT_2_OFFSET_Q = 356,
	RREG_TX2_OUT_2_OFFSET_I = 357,
	RREG_TX2_OUT_2_OFFSET_Q = 358,
	RREG_TX_FORCE_BITS = 359,
	RREG_TX_FORCE_BITS__FORCE_OUT_2_TX2_OFFSET = 360,
	RREG_TX_FORCE_BITS__FORCE_OUT_2_TX1_OFFSET = 361,
	RREG_TX_FORCE_BITS__FORCE_OUT_2_TX2_PHASE_GAIN = 362,
	RREG_TX_FORCE_BITS__FORCE_OUT_2_TX1_PHASE_GAIN = 363,
	RREG_TX_FORCE_BITS__FORCE_OUT_1_TX2_OFFSET = 364,
	RREG_TX_FORCE_BITS__FORCE_OUT_1_TX1_OFFSET = 365,
	RREG_TX_FORCE_BITS__FORCE_OUT_1_TX2_PHASE_GAIN = 366,
	RREG_TX_FORCE_BITS__FORCE_OUT_1_TX1_PHASE_GAIN = 367,
	RREG_QUAD_CAL_NCO_FREQ_PHASE_OFFSET = 368,
	RREG_QUAD_CAL_NCO_FREQ_PHASE_OFFSET__RX_NCO_FREQ = 369,
	RREG_QUAD_CAL_NCO_FREQ_PHASE_OFFSET__RX_NCO_PHASE_OFFSET = 370,
	RREG_QUAD_CAL_CTRL = 371,
	RREG_QUAD_CAL_CTRL__FREE_RUN_ENABLE = 372,
	RREG_QUAD_CAL_CTRL__SETTLE_MAIN_ENABLE = 373,
	RREG_QUAD_CAL_CTRL__DC_OFFSET_ENABLE = 374,
	RREG_QUAD_CAL_CTRL__GAIN_ENABLE = 375,
	RREG_QUAD_CAL_CTRL__PHASE_ENABLE = 376,
	RREG_QUAD_CAL_CTRL__QUAD_CAL_SOFT_RESET = 377,
	RREG_QUAD_CAL_CTRL__M_DECIM = 378,
	RREG_KEXP_1 = 379,
	RREG_KEXP_1__KEXP_TX = 380,
	RREG_KEXP_1__KEXP_TX_COMP = 381,
	RREG_KEXP_1__KEXP_DC_I = 382,
	RREG_KEXP_1__KEXP_DC_Q = 383,
	RREG_KEXP_2 = 384,
	RREG_KEXP_2__INVERT_I_DATA = 385,
	RREG_KEXP_2__INVERT_Q_DATA = 386,
	RREG_KEXP_2__TX_NCO_FREQ = 387,
	RREG_KEXP_2__KEXP_PHASE = 388,
	RREG_KEXP_2__KEXP_AMP = 389,
	RREG_QUAD_SETTLE_COUNT = 390,
	RREG_MAG_FTEST_THRESH = 391,
	RREG_MAG_FTEST_THRESH_2 = 392,
	RREG_QUAD_CAL_STATUS_TX1 = 393,
	RREG_QUAD_CAL_STATUS_TX1__TX1_LO_CONV = 394,
	RREG_QUAD_CAL_STATUS_TX1__TX1_SSB_CONV = 395,
	RREG_QUAD_CAL_STATUS_TX1__TX1_CONVERGENCE_COUNT = 396,
	RREG_QUAD_CAL_STATUS_TX2 = 397,
	RREG_QUAD_CAL_STATUS_TX2__TX2_LO_CONV = 398,
	RREG_QUAD_CAL_STATUS_TX2__TX2_SSB_CONV = 399,
	RREG_QUAD_CAL_STATUS_TX2__TX2_CONVERGENCE_COUNT = 400,
	RREG_QUAD_CAL_COUNT = 401,
	RREG_TX_QUAD_FULL_LMT_GAIN = 402,
	RREG_TX_QUAD_FULL_LMT_GAIN__RX_FULL_TABLELMT_TABLE_GAIN = 403,
	RREG_SQUARER_CONFIG = 404,
	RREG_SQUARER_CONFIG__GM_STAGE_TIME_CON_OVERRIDE = 405,
	RREG_SQUARER_CONFIG__GM_STAGE_MV_HP_POLE = 406,
	RREG_SQUARER_CONFIG__GM_STAGE_LOWER_CM = 407,
	RREG_SQUARER_CONFIG__BYPASS_BIAS_R = 408,
	RREG_SQUARER_CONFIG__VBIAS_CTRL = 409,
	RREG_TX_QUAD_CAL_ATTEN = 410,
	RREG_THRESH_ACCUM = 411,
	RREG_THRESH_ACCUM__THRESH_ACCUMULATOR = 412,
	RREG_TX_QUAD_LPF_GAIN = 413,
	RREG_TX_QUAD_LPF_GAIN__RX_LPF_GAIN = 414,
	RREG_TXDAC_VDS_I = 415,
	RREG_TXDAC_VDS_I__TXDAC_VDS_I = 416,
	RREG_TXDAC_VDS_Q = 417,
	RREG_TXDAC_VDS_Q__TXDAC_VDS_Q = 418,
	RREG_TXDAC_GN_I = 419,
	RREG_TXDAC_GN_I__TXDAC_GN_I = 420,
	RREG_TXDAC_GN_Q = 421,
	RREG_TXDAC_GN_Q__TXDAC_GN_Q = 422,
	RREG_TXBBF_OPAMP_A = 423,
	RREG_TXBBF_OPAMP_A__OPAMPA_OUTPUT_BIAS = 424,
	RREG_TXBBF_OPAMP_A__OPAMPA_RZ = 425,
	RREG_TXBBF_OPAMP_A__OPAMP_A_CC = 426,
	RREG_TXBBF_OPAMP_B = 427,
	RREG_TXBBF_OPAMP_B__OPAMPB_OUTPUT_BIAS = 428,
	RREG_TXBBF_OPAMP_B__OPAMPB_RZ = 429,
	RREG_TXBBF_OPAMP_B__OPAMP_B_CC = 430,
	RREG_TX_BBF_R1 = 431,
	RREG_TX_BBF_R1__OVERRIDE_ENABLE = 432,
	RREG_TX_BBF_R1__R1 = 433,
	RREG_TX_BBF_R2 = 434,
	RREG_TX_BBF_R2__R2 = 435,
	RREG_TX_BBF_R3 = 436,
	RREG_TX_BBF_R3__R3 = 437,
	RREG_TX_BBF_R4 = 438,
	RREG_TX_BBF_R4__R4 = 439,
	RREG_TX_BBF_RP = 440,
	RREG_TX_BBF_RP__RP = 441,
	RREG_TX_BBF_C1 = 442,
	RREG_TX_BBF_C1__C1 = 443,
	RREG_TX_BBF_C2 = 444,
	RREG_TX_BBF_C2__C2 = 445,
	RREG_TX_BBF_CP = 446,
	RREG_TX_BBF_CP__CP = 447,
	RREG_TX_TUNE_CTRL = 448,
	RREG_TX_TUNE_CTRL__PD_TUNE = 449,
	RREG_TX_TUNE_CTRL__TUNER_RESAMPLE = 450,
	RREG_TX_TUNE_CTRL__TUNER_RESAMPLE_PHASE = 451,
	RREG_TX_TUNE_CTRL__TUNE_CTRL = 452,
	RREG_TX_BBF_R2B = 453,
	RREG_TX_BBF_R2B__TX_BBF_BYPASS_BIAS_R = 454,
	RREG_TX_BBF_R2B__R2B_OVR = 455,
	RREG_TX_BBF_R2B__R2B = 456,
	RREG_TX_BBF_TUNE = 457,
	RREG_TX_BBF_TUNE__BBF1_COMP_I = 458,
	RREG_TX_BBF_TUNE__BBF1_COMP_Q = 459,
	RREG_TX_BBF_TUNE__BBF2_COMP_I = 460,
	RREG_TX_BBF_TUNE__BBF2_COMP_Q = 461,
	RREG_CONFIG0 = 462,
	RREG_CONFIG0__BIAS = 463,
	RREG_CONFIG0__RGM = 464,
	RREG_CONFIG0__CC = 465,
	RREG_CONFIG0__AMPBIAS = 466,
	RREG_RESISTOR = 467,
	RREG_RESISTOR__RESISTOR = 468,
	RREG_CAPACITOR = 469,
	RREG_CAPACITOR__CAPACITOR = 470,
	RREG_LO_CM = 471,
	RREG_LO_CM__LO_COMMON_MODE = 472,
	RREG_TX_BBF_TUNE_DIVIDER = 473,
	RREG_TX_BBF_TUNE_MODE = 474,
	RREG_TX_BBF_TUNE_MODE__EVALTIME = 475,
	RREG_TX_BBF_TUNE_MODE__TX_BBF_TUNE_DIVIDER = 476,
	RREG_TX_BBF_TUNE_MODE__TUNE_COMP_MASK = 477,
	RREG_TX_BBF_TUNE_MODE__TUNER_MODE = 478,
	RREG_RX_FILTER_COEF_ADDR = 479,
	RREG_RX_FILTER_COEF_DATA_1 = 480,
	RREG_RX_FILTER_COEF_DATA_2 = 481,
	RREG_RX_FILTER_COEF_READ_DATA_1 = 482,
	RREG_RX_FILTER_COEF_READ_DATA_2 = 483,
	RREG_RX_FILTER_CONFIG = 484,
	RREG_RX_FILTER_CONFIG__WRITE_RX = 485,
	RREG_RX_FILTER_CONFIG__START_RX_CLOCK = 486,
	RREG_RX_FILTER_CONFIG__NUMBER_OF_TAPS = 487,
	RREG_RX_FILTER_CONFIG__SELECT_RX_CH = 488,
	RREG_RX_FILTER_GAIN = 489,
	RREG_RX_FILTER_GAIN__FILTER_GAIN = 490,
	RREG_AGC_CONFIG_1 = 491,
	RREG_AGC_CONFIG_1__DEC_PWR_FOR_LOW_PWR = 492,
	RREG_AGC_CONFIG_1__DEC_PWR_FOR_LOCK_LEVEL = 493,
	RREG_AGC_CONFIG_1__DEC_PWR_FOR_GAIN_LOCK_EXIT = 494,
	RREG_AGC_CONFIG_1__SLOW_ATTACK_HYBRID_MODE = 495,
	RREG_AGC_CONFIG_1__RX2_GAIN_CTRL_SETUP = 496,
	RREG_AGC_CONFIG_1__RX1_GAIN_CTRL_SETUP = 497,
	RREG_AGC_CONFIG_2 = 498,
	RREG_AGC_CONFIG_2__AGC_SOFT_RESET = 499,
	RREG_AGC_CONFIG_2__AGC_GAIN_UNLOCK_CTRL = 500,
	RREG_AGC_CONFIG_2__AGC_USE_FULL_GAIN_TABLE = 501,
	RREG_AGC_CONFIG_2__DIG_GAIN_EN = 502,
	RREG_AGC_CONFIG_2__MAN_GAIN_CTRL_RX2 = 503,
	RREG_AGC_CONFIG_2__MAN_GAIN_CTRL_RX1 = 504,
	RREG_AGC_CONFIG_3 = 505,
	RREG_AGC_CONFIG_3__INCDEC_LMT_GAIN = 506,
	RREG_AGC_CONFIG_3__USE_AGC_FOR_LMTLPF_GAIN = 507,
	RREG_AGC_CONFIG_3__MANUAL_INCR_STEP_SIZE = 508,
	RREG_AGC_CONFIG_3__ADC_OVERRANGE_SAMPLE_SIZE = 509,
	RREG_MAX_LMT_FULL_GAIN = 510,
	RREG_MAX_LMT_FULL_GAIN__MAXIMUM_FULL_TABLELMT_TABLE_INDEX = 511,
	RREG_PEAK_WAIT_TIME = 512,
	RREG_PEAK_WAIT_TIME__MANUAL_CTRL_IN_DECR_GAIN_STP_SIZE = 513,
	RREG_PEAK_WAIT_TIME__PEAK_OVERLOAD_WAIT_TIME = 514,
	RREG_DIGITAL_GAIN = 515,
	RREG_DIGITAL_GAIN__DIG_GAIN_STP_SIZE = 516,
	RREG_DIGITAL_GAIN__MAXIMUM_DIGITAL_GAIN = 517,
	RREG_AGC_LOCK_LEVEL = 518,
	RREG_AGC_LOCK_LEVEL__ENABLE_DIG_SAT_OVRG = 519,
	RREG_AGC_LOCK_LEVEL__AGC_LOCK_LEVEL_FAST_AGC_INNER_HIGH_THRESH_SLOW = 520,
	RREG_ADC_NOISE_CORRECTION_FACTOR = 521,
	RREG_GAIN_STP_CONFIG1 = 522,
	RREG_GAIN_STP_CONFIG1__LMT_DETECTOR_SETTLING_TIME = 523,
	RREG_GAIN_STP_CONFIG1__DEC_STP_SIZE_FOR_LARGE_LMT_OVERLOAD = 524,
	RREG_GAIN_STP_CONFIG1__ADC_NOISE_CORRECTION_FACTOR = 525,
	RREG_ADC_SMALL_OVERLOAD_THRESH = 526,
	RREG_ADC_LARGE_OVERLOAD_THRESH = 527,
	RREG_GAIN_STP_CONFIG_2 = 528,
	RREG_GAIN_STP_CONFIG_2__DECREMENT_STP_SIZE_FOR_SMALL_LPF_GAIN_CHANGE = 529,
	RREG_GAIN_STP_CONFIG_2__LARGE_LPF_GAIN_STEP = 530,
	RREG_SMALL_LMT_OVERLOAD_THRESH = 531,
	RREG_SMALL_LMT_OVERLOAD_THRESH__FORCE_PD_RESET_RX2 = 532,
	RREG_SMALL_LMT_OVERLOAD_THRESH__FOR_PD_RESET_RX1 = 533,
	RREG_SMALL_LMT_OVERLOAD_THRESH__SMALL_LMT_OVERLOAD_THRESH = 534,
	RREG_LARGE_LMT_OVERLOAD_THRESH = 535,
	RREG_LARGE_LMT_OVERLOAD_THRESH__LARGE_LMT_OVERLOAD_THRESH = 536,
	RREG_RX1_MANUAL_LMT_FULL_GAIN = 537,
	RREG_RX1_MANUAL_LMT_FULL_GAIN__POWER_MEAS_IN_STATE_5_MSB = 538,
	RREG_RX1_MANUAL_LMT_FULL_GAIN__RX1_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX = 539,
	RREG_RX1_MANUAL_LPF_GAIN = 540,
	RREG_RX1_MANUAL_LPF_GAIN__POWER_MEAS_IN_STATE_5 = 541,
	RREG_RX1_MANUAL_LPF_GAIN__RX1_MANUAL_LPF_GAIN = 542,
	RREG_RX1_MANUAL_DIGITALFORCED_GAIN = 543,
	RREG_RX1_MANUAL_DIGITALFORCED_GAIN__FORCE_RX1_DIGITAL_GAIN = 544,
	RREG_RX1_MANUAL_DIGITALFORCED_GAIN__RX1_MANUALFORCED_DIGITAL_GAIN = 545,
	RREG_RX2_MANUAL_LMT_FULL_GAIN = 546,
	RREG_RX2_MANUAL_LMT_FULL_GAIN__RX2_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX = 547,
	RREG_RX2_MANUAL_LPF_GAIN = 548,
	RREG_RX2_MANUAL_LPF_GAIN__RX2_MANUAL_LPF_GAIN = 549,
	RREG_RX2_MANUAL_DIGITALFORCED_GAIN = 550,
	RREG_RX2_MANUAL_DIGITALFORCED_GAIN__FORCE_RX2_DIGITAL_GAIN = 551,
	RREG_RX2_MANUAL_DIGITALFORCED_GAIN__RX2_MANUALFORCED_DIGITAL_GAIN = 552,
	RREG_FAST_CONFIG_1 = 553,
	RREG_FAST_CONFIG_1__ENABLE_GAIN_INC_AFTER_GAIN_LOCK = 554,
	RREG_FAST_CONFIG_1__GOTO_OPT_GAIN_IF_ENERGY_LOST_OR_EN_AGC_HIGH = 555,
	RREG_FAST_CONFIG_1__GOTO_SET_GAIN_IF_EN_AGC_HIGH = 556,
	RREG_FAST_CONFIG_1__GOTO_SET_GAIN_IF_EXIT_RX_STATE = 557,
	RREG_FAST_CONFIG_1__DONT_UNLOCK_GAIN_IF_ENERGY_LOST = 558,
	RREG_FAST_CONFIG_1__GOTO_OPTIMIZED_GAIN_IF_EXIT_RX_STATE = 559,
	RREG_FAST_CONFIG_1__DONT_UNLOCK_GAIN_IF_LG_ADC_OR_LMT_OVRG = 560,
	RREG_FAST_CONFIG_1__ENABLE_INCR_GAIN = 561,
	RREG_FAST_CONFIG_2_SETTLING_DELAY = 562,
	RREG_FAST_CONFIG_2_SETTLING_DELAY__USE_LAST_LOCK_LEVEL_FOR_SET_GAIN = 563,
	RREG_FAST_CONFIG_2_SETTLING_DELAY__ENABLE_LMT_GAIN_INC_FOR_LOCK_LEVEL = 564,
	RREG_FAST_CONFIG_2_SETTLING_DELAY__GOTO_MAX_GAIN_OR_OPT_GAIN_IF_EN_AGC_HIGH = 565,
	RREG_FAST_CONFIG_2_SETTLING_DELAY__SETTLING_DELAY = 566,
	RREG_FAST_ENERGY_LOST_THRESH = 567,
	RREG_FAST_ENERGY_LOST_THRESH__POST_LOCK_LEVEL_STP_SIZE_FOR_LPF_TABLE_FULL_TABLE = 568,
	RREG_FAST_ENERGY_LOST_THRESH__ENERGY_LOST_THRESH = 569,
	RREG_FAST_STRONGER_SIGNAL_THRESH = 570,
	RREG_FAST_STRONGER_SIGNAL_THRESH__POST_LOCK_LEVEL_STP_FOR_LMT_TABLE = 571,
	RREG_FAST_STRONGER_SIGNAL_THRESH__STRONGER_SIGNAL_THRESH = 572,
	RREG_FAST_LOW_POWER_THRESH = 573,
	RREG_FAST_LOW_POWER_THRESH__DONT_UNLOCK_GAIN_IF_ADC_OVRG = 574,
	RREG_FAST_LOW_POWER_THRESH__LOW_POWER_THRESH = 575,
	RREG_FAST_STRONG_SIGNAL_FREEZE = 576,
	RREG_FAST_STRONG_SIGNAL_FREEZE__DONT_UNLOCK_GAIN_IF_STRONGER_SIGNAL = 577,
	RREG_FAST_FINAL_OVER_RANGE_AND_OPT_GAIN = 578,
	RREG_FAST_FINAL_OVER_RANGE_AND_OPT_GAIN__FINAL_OVER_RANGE_COUNT = 579,
	RREG_FAST_FINAL_OVER_RANGE_AND_OPT_GAIN__OPTIMIZE_GAIN_OFFSET = 580,
	RREG_FAST_ENERGY_DETECT_COUNT = 581,
	RREG_FAST_ENERGY_DETECT_COUNT__INCREMENT_GAIN_STP_LPFLMT = 582,
	RREG_FAST_ENERGY_DETECT_COUNT__ENERGY_DETECT_COUNT = 583,
	RREG_FAST_AGCLL_UPPER_LIMIT = 584,
	RREG_FAST_AGCLL_UPPER_LIMIT__AGCLL_MAX_INCREASE = 585,
	RREG_FAST_GAIN_LOCK_EXIT_COUNT = 586,
	RREG_FAST_GAIN_LOCK_EXIT_COUNT__GAIN_LOCK_EXIT_COUNT = 587,
	RREG_FAST_INITIAL_LMT_GAIN_LIMIT = 588,
	RREG_FAST_INITIAL_LMT_GAIN_LIMIT__INITIAL_LMT_GAIN_LIMIT = 589,
	RREG_FAST_INCREMENT_TIME = 590,
	RREG_AGC_INNER_LOW_THRESH = 591,
	RREG_AGC_INNER_LOW_THRESH__PREVENT_GAIN_INC = 592,
	RREG_AGC_INNER_LOW_THRESH__AGC_INNER_LOW_THRESH = 593,
	RREG_LMT_OVERLOAD_COUNTERS = 594,
	RREG_LMT_OVERLOAD_COUNTERS__LARGE_LMT_OVERLOAD_EXED_COUNTER = 595,
	RREG_LMT_OVERLOAD_COUNTERS__SMALL_LMT_OVERLOAD_EXED_COUNTER = 596,
	RREG_ADC_OVERLOAD_COUNTERS = 597,
	RREG_ADC_OVERLOAD_COUNTERS__LARGE_ADC_OVERLOAD_EXED_COUNTER = 598,
	RREG_ADC_OVERLOAD_COUNTERS__SMALL_ADC_OVERLOAD_EXED_COUNTER = 599,
	RREG_GAIN_STP1 = 600,
	RREG_GAIN_STP1__IMMED_GAIN_CHANGE_IF_LG_LMT_OVERLOAD = 601,
	RREG_GAIN_STP1__IMMED_GAIN_CHANGE_IF_LG_ADC_OVERLOAD = 602,
	RREG_GAIN_STP1__AGC_INNER_HIGH_THRESH_EXED_STP_SIZE = 603,
	RREG_GAIN_STP1__AGC_INNER_LOW_THRESH_EXED_STP_SIZE = 604,
	RREG_GAIN_UPDATE_COUNTER1 = 605,
	RREG_GAIN_UPDATE_COUNTER2 = 606,
	RREG_DIGITAL_SAT_COUNTER = 607,
	RREG_DIGITAL_SAT_COUNTER__DOUBLE_GAIN_COUNTER = 608,
	RREG_DIGITAL_SAT_COUNTER__ENABLE_SYNC_FOR_GAIN_COUNTER = 609,
	RREG_DIGITAL_SAT_COUNTER__DIG_SATURATION_EXED_COUNTER = 610,
	RREG_OUTER_POWER_THRESHS = 611,
	RREG_OUTER_POWER_THRESHS__AGC_OUTER_HIGH_THRESH = 612,
	RREG_OUTER_POWER_THRESHS__AGC_OUTER_LOW_THRESH = 613,
	RREG_GAIN_STP_2 = 614,
	RREG_GAIN_STP_2__AGC_OUTER_HIGH_THRESH_EXED_STP_SIZE = 615,
	RREG_GAIN_STP_2__AGC_OUTER_LOW_THRESH_EXED_STP_SIZE = 616,
	RREG_EXT_LNA_HIGH_GAIN = 617,
	RREG_EXT_LNA_HIGH_GAIN__EXT_LNA_HIGH_GAIN = 618,
	RREG_EXT_LNA_LOW_GAIN = 619,
	RREG_EXT_LNA_LOW_GAIN__EXT_LNA_LOW_GAIN = 620,
	RREG_GAIN_TABLE_ADDRESS = 621,
	RREG_GAIN_TABLE_ADDRESS__GAIN_TABLE_ADDRESS = 622,
	RREG_GAIN_TABLE_WRITE_DATA1 = 623,
	RREG_GAIN_TABLE_WRITE_DATA1__EXT_LNA_CTRL = 624,
	RREG_GAIN_TABLE_WRITE_DATA1__LNA_GAIN = 625,
	RREG_GAIN_TABLE_WRITE_DATA1__MIXER_GM_GAIN = 626,
	RREG_GAIN_TABLE_WRITE_DATA2 = 627,
	RREG_GAIN_TABLE_WRITE_DATA2__TIA_GAIN = 628,
	RREG_GAIN_TABLE_WRITE_DATA2__LPF_GAIN = 629,
	RREG_GAIN_TABLE_WRITE_DATA3 = 630,
	RREG_GAIN_TABLE_WRITE_DATA3__RF_DC_CAL = 631,
	RREG_GAIN_TABLE_WRITE_DATA3__DIGITAL_GAIN = 632,
	RREG_GAIN_TABLE_READ_DATA1 = 633,
	RREG_GAIN_TABLE_READ_DATA2 = 634,
	RREG_GAIN_TABLE_READ_DATA3 = 635,
	RREG_GAIN_TABLE_CONFIG = 636,
	RREG_GAIN_TABLE_CONFIG__WRITE_GAIN_TABLE = 637,
	RREG_GAIN_TABLE_CONFIG__START_GAIN_TABLE_CLOCK = 638,
	RREG_GAIN_TABLE_CONFIG__RECEIVER_SELECT = 639,
	RREG_GM_SUB_TABLE_ADDRESS = 640,
	RREG_GM_SUB_TABLE_GAIN_WRITE = 641,
	RREG_GM_SUB_TABLE_GAIN_WRITE__GM_SUB_TABLE_GAIN_WRITE = 642,
	RREG_GM_SUB_TABLE_BIAS_WRITE = 643,
	RREG_GM_SUB_TABLE_BIAS_WRITE__GM_SUB_TABLE_BIAS_WRITE = 644,
	RREG_GM_SUB_TABLE_CTRL_WRITE = 645,
	RREG_GM_SUB_TABLE_CTRL_WRITE__GM_SUB_TABLE_CTRL_WRITE = 646,
	RREG_GM_SUB_TABLE_GAIN_READ = 647,
	RREG_GM_SUB_TABLE_GAIN_READ__GM_SUB_TABLE_GAIN_READ = 648,
	RREG_GM_SUB_TABLE_BIAS_READ = 649,
	RREG_GM_SUB_TABLE_BIAS_READ__GM_SUB_TABLE_BIAS_READ = 650,
	RREG_GM_SUB_TABLE_CTRL_READ = 651,
	RREG_GM_SUB_TABLE_CTRL_READ__GM_SUB_TABLE_CTRL_READ = 652,
	RREG_GM_SUB_TABLE_CONFIG = 653,
	RREG_GM_SUB_TABLE_CONFIG__WRITE_GM_SUB_TABLE = 654,
	RREG_GM_SUB_TABLE_CONFIG__START_GM_SUB_TABLE_CLOCK = 655,
	RREG_WORD_ADDRESS = 656,
	RREG_GAIN_DIFF_WORDERROR_WRITE = 657,
	RREG_GAIN_DIFF_WORDERROR_WRITE__CALIB_TABLE_GAIN_DIFFERROR_WORD = 658,
	RREG_GAIN_ERROR_READ = 659,
	RREG_GAIN_ERROR_READ__CALIB_TABLE_GAIN_ERROR = 660,
	RREG_CONFIG = 661,
	RREG_CONFIG__READ_SELECT = 662,
	RREG_CONFIG__WRITE_MIXER_ERROR_TABLE = 663,
	RREG_CONFIG__WRITE_LNA_ERROR_TABLE = 664,
	RREG_CONFIG__WRITE_LNA_GAIN_DIFF = 665,
	RREG_CONFIG__START_CALIB_TABLE_CLOCK = 666,
	RREG_CONFIG__CALIB_TABLE_SELECT = 667,
	RREG_LNA_GAIN_DIFF_READ_BACK = 668,
	RREG_LNA_GAIN_DIFF_READ_BACK__LNA_CALIB_TABLE_GAIN_DIFFERENCE_WORD = 669,
	RREG_MAX_MIXER_CALIBRATION_GAIN_INDEX = 670,
	RREG_MAX_MIXER_CALIBRATION_GAIN_INDEX__MAX_MIXER_CALIBRATION_GAIN_INDEX = 671,
	RREG_TEMP_GAIN_COEF = 672,
	RREG_SETTLE_TIME = 673,
	RREG_SETTLE_TIME__ENABLE_DIG_GAIN_CORR = 674,
	RREG_SETTLE_TIME__FORCE_TEMP_SENSOR_FOR_CAL = 675,
	RREG_SETTLE_TIME__SETTLE_TIME = 676,
	RREG_MEASURE_DURATION = 677,
	RREG_MEASURE_DURATION__GAIN_CAL_MEAS_DURATION = 678,
	RREG_CAL_TEMP_SENSOR_WORD = 679,
	RREG_MEASURE_DURATION_01 = 680,
	RREG_MEASURE_DURATION_01__MEASUREMENT_DURATION_1 = 681,
	RREG_MEASURE_DURATION_01__MEASUREMENT_DURATION_0 = 682,
	RREG_MEASURE_DURATION_23 = 683,
	RREG_MEASURE_DURATION_23__MEASUREMENT_DURATION_3 = 684,
	RREG_MEASURE_DURATION_23__MEASUREMENT_DURATION_2 = 685,
	RREG_RSSI_WEIGHT_0 = 686,
	RREG_RSSI_WEIGHT_1 = 687,
	RREG_RSSI_WEIGHT_2 = 688,
	RREG_RSSI_WEIGHT_3 = 689,
	RREG_RSSI_DELAY = 690,
	RREG_RSSI_WAIT_TIME = 691,
	RREG_RSSI_CONFIG = 692,
	RREG_RSSI_CONFIG__START_RSSI_MEAS = 693,
	RREG_RSSI_CONFIG__ENABLE_ADC_POWER_MEAS = 694,
	RREG_RSSI_CONFIG__DEFAULT_RSSI_MEAS_MODE = 695,
	RREG_RSSI_CONFIG__RFIR_FOR_RSSI_MEASUREMENT = 696,
	RREG_RSSI_CONFIG__RSSI_MODE_SELECT = 697,
	RREG_ADC_MEASURE_DURATION_01 = 698,
	RREG_ADC_MEASURE_DURATION_01__ADC_POWER_MEASUREMENT_DURATION_1 = 699,
	RREG_ADC_MEASURE_DURATION_01__ADC_POWER_MEASUREMENT_DURATION_0 = 700,
	RREG_ADC_WEIGHT_0 = 701,
	RREG_ADC_WEIGHT_1 = 702,
	RREG_DEC_POWER_MEASURE_DURATION_0 = 703,
	RREG_DEC_POWER_MEASURE_DURATION_0__USE_HB3_OUT_FOR_ADC_PWR_MEAS = 704,
	RREG_DEC_POWER_MEASURE_DURATION_0__USE_HB1_OUT_FOR_DEC_PWR_MEAS = 705,
	RREG_DEC_POWER_MEASURE_DURATION_0__ENABLE_DEC_PWR_MEAS = 706,
	RREG_DEC_POWER_MEASURE_DURATION_0__DEFAULT_MODE_ADC_POWER = 707,
	RREG_DEC_POWER_MEASURE_DURATION_0__DEC_POWER_MEASUREMENT_DURATION = 708,
	RREG_LNA_GAIN = 709,
	RREG_LNA_GAIN__DB_GAIN_READBACK_CHANNEL = 710,
	RREG_LNA_GAIN__MAX_LNA_GAIN = 711,
	RREG_CH1_ADC_POWER = 712,
	RREG_CH1_RX_FILTER_POWER = 713,
	RREG_CH2_ADC_POWER = 714,
	RREG_CH2_RX_FILTER_POWER = 715,
	RREG_RX_QUAD_CAL_LEVEL = 716,
	RREG_RX_QUAD_CAL_LEVEL__RX_QUAD_CAL_LEVEL = 717,
	RREG_CALIBRATION_CONFIG_1 = 718,
	RREG_CALIBRATION_CONFIG_1__ENABLE_PHASE_CORR = 719,
	RREG_CALIBRATION_CONFIG_1__ENABLE_GAIN_CORR = 720,
	RREG_CALIBRATION_CONFIG_1__USE_SETTLE_COUNT_FOR_DC_CAL_WAIT = 721,
	RREG_CALIBRATION_CONFIG_1__FIXED_DC_CAL_WAIT_TIME = 722,
	RREG_CALIBRATION_CONFIG_1__FREE_RUN_MODE = 723,
	RREG_CALIBRATION_CONFIG_1__ENABLE_CORR_WORD_DECIMATION = 724,
	RREG_CALIBRATION_CONFIG_1__ENABLE_TRACKING_MODE_CH2 = 725,
	RREG_CALIBRATION_CONFIG_1__ENABLE_TRACKING_MODE_CH1 = 726,
	RREG_CALIBRATION_CONFIG_2 = 727,
	RREG_CALIBRATION_CONFIG_2__SOFT_RESET = 728,
	RREG_CALIBRATION_CONFIG_2__K_EXP_PHASE = 729,
	RREG_CALIBRATION_CONFIG_3 = 730,
	RREG_CALIBRATION_CONFIG_3__PREVENT_POS_LOOP_GAIN = 731,
	RREG_CALIBRATION_CONFIG_3__K_EXP_AMPLITUDE = 732,
	RREG_CALIB_COUNT = 733,
	RREG_SETTLE_COUNT = 734,
	RREG_RX_QUAD_GAIN1 = 735,
	RREG_RX_QUAD_GAIN1__RX_FULL_TABLELMT_TABLE_GAIN = 736,
	RREG_RX_QUAD_GAIN2 = 737,
	RREG_RX_QUAD_GAIN2__CORRECTION_WORD_DECIMATION_M = 738,
	RREG_RX_QUAD_GAIN2__RX_LPF_GAIN = 739,
	RREG_RX1_INPUT_A_PHASE_CORR = 740,
	RREG_RX1_INPUT_A_GAIN_CORR = 741,
	RREG_RX2_INPUT_A_PHASE_CORR = 742,
	RREG_RX2_INPUT_A_GAIN_CORR = 743,
	RREG_RX1_INPUT_A_Q_OFFSET = 744,
	RREG_RX1_INPUT_A_OFFSETS = 745,
	RREG_RX1_INPUT_A_OFFSETS__RX1_INPUT_A_I_DC_OFFSET_LSB = 746,
	RREG_RX1_INPUT_A_OFFSETS__RX1_INPUT_A_Q_DC_OFFSET = 747,
	RREG_INPUT_A_OFFSETS_1 = 748,
	RREG_INPUT_A_OFFSETS_1__RX2_INPUT_A_Q_DC_OFFSET_LSB = 749,
	RREG_INPUT_A_OFFSETS_1__RX1_INPUT_A_I_DC_OFFSET_MSB = 750,
	RREG_RX2_INPUT_A_OFFSETS = 751,
	RREG_RX2_INPUT_A_OFFSETS__RX2_INPUT_A_I_DC_OFFSET = 752,
	RREG_RX2_INPUT_A_OFFSETS__RX2_INPUT_A_Q_DC_OFFSET_MSB = 753,
	RREG_RX2_INPUT_A_I_OFFSET = 754,
	RREG_RX1_INPUT_BC_PHASE_CORR = 755,
	RREG_RX1_INPUT_BC_GAIN_CORR = 756,
	RREG_RX2_INPUT_BC_PHASE_CORR = 757,
	RREG_RX2_INPUT_BC_GAIN_CORR = 758,
	RREG_RX1_INPUT_BC_Q_OFFSET = 759,
	RREG_RX1_INPUT_BC_OFFSETS = 760,
	RREG_RX1_INPUT_BC_OFFSETS__RX1_INPUT_BC_I_DC_OFFSET_LSB = 761,
	RREG_RX1_INPUT_BC_OFFSETS__RX1_INPUT_BC_Q_DC_OFFSET = 762,
	RREG_INPUT_BC_OFFSETS_1 = 763,
	RREG_INPUT_BC_OFFSETS_1__RX2_INPUT_BC_Q_DC_OFFSET_LSB = 764,
	RREG_INPUT_BC_OFFSETS_1__RX1_INPUT_BC_I_DC_OFFSET_MSB = 765,
	RREG_RX2_INPUT_BC_OFFSETS = 766,
	RREG_RX2_INPUT_BC_OFFSETS__RX2_INPUT_BC_I_DC_OFFSET = 767,
	RREG_RX2_INPUT_BC_OFFSETS__RX2_INPUT_BC_Q_DC_OFFSET_MSB = 768,
	RREG_RX2_INPUT_BC_I_OFFSET = 769,
	RREG_FORCE_BITS = 770,
	RREG_FORCE_BITS__RX2_INPUT_BC_FORCE_OFFSET = 771,
	RREG_FORCE_BITS__RX1_INPUT_BC_FORCE_OFFSET = 772,
	RREG_FORCE_BITS__RX2_INPUT_BC_FORCE_PHGAIN = 773,
	RREG_FORCE_BITS__RX1_INPUT_BC_FORCE_PHGAIN = 774,
	RREG_FORCE_BITS__RX2_INPUT_A_FORCE_OFFSET = 775,
	RREG_FORCE_BITS__RX1_INPUT_A_FORCE_OFFSET = 776,
	RREG_FORCE_BITS__RX2_INPUT_A_FORCE_PHGAIN = 777,
	RREG_FORCE_BITS__RX1_INPUT_A_FORCE_PHGAIN = 778,
	RREG_WAIT_COUNT = 779,
	RREG_RF_DC_OFFSET_COUNT = 780,
	RREG_RF_DC_OFFSET_CONFIG_1 = 781,
	RREG_RF_DC_OFFSET_CONFIG_1__DAC_FS = 782,
	RREG_RF_DC_OFFSET_CONFIG_1__RF_DC_CALIBRATION_COUNT = 783,
	RREG_RF_DC_OFFSET_ATTEN = 784,
	RREG_RF_DC_OFFSET_ATTEN__RF_DC_OFFSET_TABLE_UPDATE_COUNT = 785,
	RREG_RF_DC_OFFSET_ATTEN__RF_DC_OFFSET_ATTEN = 786,
	RREG_INVERT_BITS = 787,
	RREG_INVERT_BITS__INVERT_RX2_RF_DC_CGIN_WORD = 788,
	RREG_INVERT_BITS__INVERT_RX1_RF_DC_CGIN_WORD = 789,
	RREG_INVERT_BITS__INVERT_RX2_RF_DC_CGOUT_WORD = 790,
	RREG_INVERT_BITS__INVERT_RX1_RF_DC_CGOUT_WORD = 791,
	RREG_DC_OFFSET_CONFIG2 = 792,
	RREG_DC_OFFSET_CONFIG2__USE_WAIT_COUNTER_FOR_RF_DC_INIT_CAL = 793,
	RREG_DC_OFFSET_CONFIG2__ENABLE_FAST_SETTLE_MODE = 794,
	RREG_DC_OFFSET_CONFIG2__ENABLE_BB_DC_OFFSET_TRACKING = 795,
	RREG_DC_OFFSET_CONFIG2__RESET_ACC_ON_GAIN_CHANGE = 796,
	RREG_DC_OFFSET_CONFIG2__ENABLE_RF_OFFSET_TRACKING = 797,
	RREG_DC_OFFSET_CONFIG2__DC_OFFSET_UPDATE = 798,
	RREG_RF_CAL_GAIN_INDEX = 799,
	RREG_RF_CAL_GAIN_INDEX__RF_MINIMUM_CALIBRATION_GAIN_INDEX = 800,
	RREG_SOI_THRESH = 801,
	RREG_SOI_THRESH__RF_SOI_THRESH = 802,
	RREG_BB_DC_OFFSET_SHIFT = 803,
	RREG_BB_DC_OFFSET_SHIFT__INCREASE_COUNT_DURATION = 804,
	RREG_BB_DC_OFFSET_SHIFT__BB_TRACKING_DECIMATE = 805,
	RREG_BB_DC_OFFSET_SHIFT__BB_DC_M_SHIFT = 806,
	RREG_BB_DC_OFFSET_FAST_SETTLE_SHIFT = 807,
	RREG_BB_DC_OFFSET_FAST_SETTLE_SHIFT__READ_BACK_CH_SEL = 808,
	RREG_BB_DC_OFFSET_FAST_SETTLE_SHIFT__UPDATE_TRACKING_WORD = 809,
	RREG_BB_DC_OFFSET_FAST_SETTLE_SHIFT__FORCE_RX_NULL = 810,
	RREG_BB_DC_OFFSET_FAST_SETTLE_SHIFT__BB_DC_TRACKING_FAST_SETTLE_M_SHIFT = 811,
	RREG_BB_FAST_SETTLE_DUR = 812,
	RREG_BB_DC_OFFSET_COUNT = 813,
	RREG_BB_DC_OFFSET_ATTEN = 814,
	RREG_BB_DC_OFFSET_ATTEN__BB_DC_OFFSET_ATTEN = 815,
	RREG_RX1_BB_DC_WORD_I_MSB = 816,
	RREG_RX1_BB_DC_WORD_I_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_I = 817,
	RREG_RX1_BB_DC_WORD_I_LSB = 818,
	RREG_RX1_BB_DC_WORD_Q_MSB = 819,
	RREG_RX1_BB_DC_WORD_Q_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_Q = 820,
	RREG_RX1_BB_DC_WORD_Q_LSB = 821,
	RREG_RX2_BB_DC_WORD_I_MSB = 822,
	RREG_RX2_BB_DC_WORD_I_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_I = 823,
	RREG_RX2_BB_DC_WORD_I_LSB = 824,
	RREG_RX2_BB_DC_WORD_Q_MSB = 825,
	RREG_RX2_BB_DC_WORD_Q_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_Q = 826,
	RREG_RX2_BB_DC_WORD_Q_LSB = 827,
	RREG_BB_TRACK_CORR_WORD_I_MSB = 828,
	RREG_BB_TRACK_CORR_WORD_I_MSB__RX1RX2_BB_DC_OFFSET_TRACKING_CORRECTION_WORD_I = 829,
	RREG_BB_TRACK_CORR_WORD_I_LSB = 830,
	RREG_BB_TRACK_CORR_WORD_Q_MSB = 831,
	RREG_BB_TRACK_CORR_WORD_Q_MSB__RX1RX2_BB_DC_OFFSET_TRACKING_CORRECTION_WORD_Q = 832,
	RREG_BB_TRACK_CORR_WORD_Q_LSB = 833,
	RREG_RX1_RSSI_SYMBOL = 834,
	RREG_RX1_RSSI_PREAMBLE = 835,
	RREG_RX2_RSSI_SYMBOL = 836,
	RREG_RX2_RSSI_PREAMBLE = 837,
	RREG_SYMBOL_LSB = 838,
	RREG_SYMBOL_LSB__RX2_RSSI_SYMBOL = 839,
	RREG_SYMBOL_LSB__RX1_RSSI_SYMBOL = 840,
	RREG_PREAMBLE_LSB = 841,
	RREG_PREAMBLE_LSB__RX2_RSSI_PREAMBLE = 842,
	RREG_PREAMBLE_LSB__RX1_RSSI_PREAMBLE = 843,
	RREG_RX_PATH_GAIN_MSB = 844,
	RREG_RX_PATH_GAIN_LSB = 845,
	RREG_RX_PATH_GAIN_LSB__RX_PATH_GAIN = 846,
	RREG_RX_DIFF_LNA_FORCE = 847,
	RREG_RX_DIFF_LNA_FORCE__FORCE_RX2_LNA_GAIN = 848,
	RREG_RX_DIFF_LNA_FORCE__RX2_LNA_BYPASS = 849,
	RREG_RX_DIFF_LNA_FORCE__FORCE_RX1_LNA_GAIN = 850,
	RREG_RX_DIFF_LNA_FORCE__RX1_LNA_BYPASS = 851,
	RREG_RX_DIFF_LNA_FORCE__RX2_LNA_GAIN = 852,
	RREG_RX_DIFF_LNA_FORCE__RX1_LNA_GAIN = 853,
	RREG_RX_LNA_BIAS_COARSE = 854,
	RREG_RX_LNA_BIAS_COARSE__RX_LNA_BIAS_COARSE = 855,
	RREG_RX_LNA_BIAS_FINE_0 = 856,
	RREG_RX_LNA_BIAS_FINE_0__RX_LNA_PCASCODE_BIAS = 857,
	RREG_RX_LNA_BIAS_FINE_0__RX_LNA_BIAS = 858,
	RREG_RX_LNA_BIAS_FINE_1 = 859,
	RREG_RX_LNA_BIAS_FINE_1__RX_LNA_P_CASCODE_BIAS_FINE = 860,
	RREG_RX_MIX_GM_CONFIG = 861,
	RREG_RX_MIX_GM_CONFIG__RX_MIX_GM_CM_OUT = 862,
	RREG_RX_MIX_GM_CONFIG__RX_MIX_GM_PLOAD = 863,
	RREG_RX1_MIX_GM_FORCE = 864,
	RREG_RX1_MIX_GM_FORCE__FORCE_RX1_MIX_GM = 865,
	RREG_RX1_MIX_GM_FORCE__RX1_MIX_GM_GAIN = 866,
	RREG_RX1_MIX_GM_BIAS_FORCE = 867,
	RREG_RX1_MIX_GM_BIAS_FORCE__RX1_MIX_GM_BIAS = 868,
	RREG_RX2_MIX_GM_FORCE = 869,
	RREG_RX2_MIX_GM_FORCE__FORCE_RX2_MIX_GM = 870,
	RREG_RX2_MIX_GM_FORCE__RX2_MIX_GM_GAIN = 871,
	RREG_RX2_MIX_GM_BIAS_FORCE = 872,
	RREG_RX2_MIX_GM_BIAS_FORCE__RX2_MIX_GM_BIAS = 873,
	RREG_INPUT_A_MSBS = 874,
	RREG_INPUT_A_MSBS__INPUT_A_RX1_Q = 875,
	RREG_INPUT_A_MSBS__INPUT_A_RX1_I = 876,
	RREG_INPUT_A_MSBS__INPUT_A_RX2_I = 877,
	RREG_INPUT_A_MSBS__INPUT_A_RX2_Q = 878,
	RREG_INPUT_A_RX1_I = 879,
	RREG_INPUT_A_RX1_Q = 880,
	RREG_INPUT_A_RX2_I = 881,
	RREG_INPUT_A_RX2_Q = 882,
	RREG_INPUTS_BC_RX1_I = 883,
	RREG_BAND1_RX1_Q = 884,
	RREG_INPUTS_BC_RX2_I = 885,
	RREG_INPUTS_BC_RX2_Q = 886,
	RREG_INPUTS_BC_MSBS = 887,
	RREG_INPUTS_BC_MSBS__INPUTS_BC_RX1_Q = 888,
	RREG_INPUTS_BC_MSBS__INPUTS_BC_RX1_I = 889,
	RREG_INPUTS_BC_MSBS__INPUTS_BC_RX2_I = 890,
	RREG_INPUTS_BC_MSBS__INPUTS_BC_RX2_Q = 891,
	RREG_FORCE_OS_DAC = 892,
	RREG_FORCE_OS_DAC__FORCE_CGIN_DAC = 893,
	RREG_RX_MIX_LO_CM = 894,
	RREG_RX_MIX_LO_CM__RX_MIX_LO_CM = 895,
	RREG_RX_CGB_SEG_ENABLE = 896,
	RREG_RX_CGB_SEG_ENABLE__RX_CGB_SEG_ENABLE = 897,
	RREG_RX_MIX_INPUTBIAS = 898,
	RREG_RX_MIX_INPUTBIAS__RX_CGB_INPUT_CM_SEL = 899,
	RREG_RX_MIX_INPUTBIAS__RX_CGB_BIAS = 900,
	RREG_RX_TIA_CONFIG = 901,
	RREG_RX_TIA_CONFIG__TIA2_OVERRIDE_C = 902,
	RREG_RX_TIA_CONFIG__TIA2_OVERRIDE_R = 903,
	RREG_RX_TIA_CONFIG__TIA1_OVERRIDE_C = 904,
	RREG_RX_TIA_CONFIG__TIA1_OVERRIDE_R = 905,
	RREG_RX_TIA_CONFIG__TIA_SEL_CC = 906,
	RREG_TIA1_C_LSB = 907,
	RREG_TIA1_C_LSB__TIA1_RF = 908,
	RREG_TIA1_C_LSB__TIA1_C_LSB = 909,
	RREG_TIA1_C_MSB = 910,
	RREG_TIA1_C_MSB__TIA1_C_MSB = 911,
	RREG_TIA2_C_LSB = 912,
	RREG_TIA2_C_LSB__TIA2_RF = 913,
	RREG_TIA2_C_LSB__TIA2_C_LSB = 914,
	RREG_TIA2_C_MSB = 915,
	RREG_TIA2_C_MSB__TIA2_C_MSB = 916,
	RREG_RX1_BBF_R1A = 917,
	RREG_RX1_BBF_R1A__FORCE_RX1_RESISTORS = 918,
	RREG_RX1_BBF_R1A__RX1_BBF_R1A = 919,
	RREG_RX2_BBF_R1A = 920,
	RREG_RX2_BBF_R1A__FORCE_RX2_RESISTORS = 921,
	RREG_RX2_BBF_R1A__RX2_BBF_R1A = 922,
	RREG_RX1_TUNE_CTRL = 923,
	RREG_RX1_TUNE_CTRL__RX1_TUNE_RESAMPLE_PHASE = 924,
	RREG_RX1_TUNE_CTRL__RX1_TUNE_RESAMPLE = 925,
	RREG_RX1_TUNE_CTRL__RX1_PD_TUNE = 926,
	RREG_RX2_TUNE_CTRL = 927,
	RREG_RX2_TUNE_CTRL__RX2_TUNE_RESAMPLE_PHASE = 928,
	RREG_RX2_TUNE_CTRL__RX2_TUNE_RESAMPLE = 929,
	RREG_RX2_TUNE_CTRL__RX2_PD_TUNE = 930,
	RREG_RX1_BBF_R5 = 931,
	RREG_RX2_BBF_R5 = 932,
	RREG_RX_BBF_R2346 = 933,
	RREG_RX_BBF_R2346__TUNE_OVERRIDE = 934,
	RREG_RX_BBF_R2346__RX_BBF_R2346 = 935,
	RREG_RX_BBF_C1_MSB = 936,
	RREG_RX_BBF_C1_MSB__RX_BBF_C1_MSB = 937,
	RREG_RX_BBF_C1_LSB = 938,
	RREG_RX_BBF_C1_LSB__RX_BBF_C1_LSB = 939,
	RREG_RX_BBF_C2_MSB = 940,
	RREG_RX_BBF_C2_MSB__RX_BBF_C2_MSB = 941,
	RREG_RX_BBF_C2_LSB = 942,
	RREG_RX_BBF_C2_LSB__RX_BBF_C2_LSB = 943,
	RREG_RX_BBF_C3_MSB = 944,
	RREG_RX_BBF_C3_MSB__RX_BBF_C3_MSB = 945,
	RREG_RX_BBF_C3_LSB = 946,
	RREG_RX_BBF_C3_LSB__RX_BBF_C3_LSB = 947,
	RREG_RX_BBF_CC1_CTR = 948,
	RREG_RX_BBF_CC1_CTR__RX_BBF_CC1_CTR = 949,
	RREG_RX_BBF_POW_RZ_BYTE0 = 950,
	RREG_RX_BBF_POW_RZ_BYTE0__MUST_BE_ZERO = 951,
	RREG_RX_BBF_POW_RZ_BYTE0__RX1_BBF_POW_CTR = 952,
	RREG_RX_BBF_POW_RZ_BYTE0__RX_BBF_RZ1_CTR = 953,
	RREG_RX_BBF_CC2_CTR = 954,
	RREG_RX_BBF_CC2_CTR__RX_BBF_CC2_CTR = 955,
	RREG_RX_BBF_POW_RZ_BYTE1 = 956,
	RREG_RX_BBF_POW_RZ_BYTE1__RX_BBF_POW3_CTR = 957,
	RREG_RX_BBF_POW_RZ_BYTE1__RX_BBF_RZ3_CTR = 958,
	RREG_RX_BBF_POW_RZ_BYTE1__RX_BBF_POW2_CTR = 959,
	RREG_RX_BBF_POW_RZ_BYTE1__RX_BBF_RZ2_CTR = 960,
	RREG_RX_BBF_CC3_CTR = 961,
	RREG_RX_BBF_CC3_CTR__RX_BBF_CC3_CTR = 962,
	RREG_RX_BBF_R5_TUNE = 963,
	RREG_RX_BBF_TUNE = 964,
	RREG_RX_BBF_TUNE__RXBBF_BYPASS_BIAS_R = 965,
	RREG_RX_BBF_TUNE__RX_BBF_R5_TUNE = 966,
	RREG_RX_BBF_TUNE__RX1_BBF_TUNE_COMP_I = 967,
	RREG_RX_BBF_TUNE__RX1_BBF_TUNE_COMP_Q = 968,
	RREG_RX_BBF_TUNE__RX2_BBF_TUNE_COMP_I = 969,
	RREG_RX_BBF_TUNE__RX2_BBF_TUNE_COMP_Q = 970,
	RREG_RX_BBF_TUNE__RX_BBF_TUNE_CTR = 971,
	RREG_RX1_BBF_MAN_GAIN = 972,
	RREG_RX1_BBF_MAN_GAIN__RX1_BBF_FORCE_GAIN = 973,
	RREG_RX1_BBF_MAN_GAIN__RX1_BBF_BQ_GAIN = 974,
	RREG_RX1_BBF_MAN_GAIN__RX1_BBF_POLE_GAIN = 975,
	RREG_RX2_BBF_MAN_GAIN = 976,
	RREG_RX2_BBF_MAN_GAIN__RX2_BBF_FORCE_GAIN = 977,
	RREG_RX2_BBF_MAN_GAIN__RX2_BBF_BQ_GAIN = 978,
	RREG_RX2_BBF_MAN_GAIN__RX2_BBF_POLE_GAIN = 979,
	RREG_RX_BBF_TUNE_DIVIDE = 980,
	RREG_RX_BBF_TUNE_CONFIG = 981,
	RREG_RX_BBF_TUNE_CONFIG__RX_TUNE_EVALTIME = 982,
	RREG_RX_BBF_TUNE_CONFIG__RX_BBF_TUNE_DIVIDE = 983,
	RREG_RX_BBF_TUNE_CONFIG__TUNE_COMP_MASK = 984,
	RREG_RX_BBF_TUNE_CONFIG__RX_TUNE_MODE = 985,
	RREG_POLE_GAIN = 986,
	RREG_POLE_GAIN__POLE_GAIN_TUNE = 987,
	RREG_RX_BBBW_MHZ = 988,
	RREG_RX_BBBW_MHZ__RX_TUNE_BBBW_MHZ = 989,
	RREG_RX_BBBW_KHZ = 990,
	RREG_RX_BBBW_KHZ__RX_TUNE_BBBW_KHZ = 991,
	RREG_FB_DAC_CLK_DELAY1 = 992,
	RREG_FB_DAC_CLK_DELAY2 = 993,
	RREG_FLASH_SAMPLE_CLK_DELAY_3P = 994,
	RREG_FLASH_SAMPLE_CLK_DELAY_3N = 995,
	RREG_TEST_MUX_2I = 996,
	RREG_TEST_MUX_2Q = 997,
	RREG_INTEGRATOR_1_RESISTANCE = 998,
	RREG_INTEGRATOR_1_CAPACITANCE = 999,
	RREG_INTEGRATOR_23_RESISTANCE = 1000,
	RREG_INTEGRATOR_2_RESISTANCE = 1001,
	RREG_INTEGRATOR_2_CAPACITANCE = 1002,
	RREG_INTEGRATOR_3_RESISTANCE = 1003,
	RREG_INTEGRATOR_3_CAPACITANCE = 1004,
	RREG_INTEGRATOR_AMP_CC = 1005,
	RREG_INT_1_FB_DAC_NMOS_CURRENT_SOURCE = 1006,
	RREG_INT_1_FB_DAC_NMOS_CASOADE_BIAS_CURRENT = 1007,
	RREG_INT_1_FB_DAC_PMOS_CURRENT_SOURCE = 1008,
	RREG_INT_2_FB_DAC_NMOS_CURRENT_SOURCE = 1009,
	RREG_INT_2_FB_DAC_NMOS_CASCODE_BIAS_CURRENT = 1010,
	RREG_INT_2_FB_DAC_PMOS_CURRENT_SOURCE = 1011,
	RREG_INT_3_FB_DAC_NMOS_CURRENT_SOURCE = 1012,
	RREG_INT_3_FB_DAC_NMOS_CASCODE_BIAS_CURRENT = 1013,
	RREG_INT_3_FB_DAC_PMOS_CURRENT_SOURCE = 1014,
	RREG_FB_DAC_BIAS_CURRENT = 1015,
	RREG_INT_1_1ST_STAGE_CURRENT = 1016,
	RREG_INT_1_1ST_STAGE_CASCODE_CURRENT = 1017,
	RREG_INT_1_2ND_STAGE_CURRENT = 1018,
	RREG_INTEGRATOR_2_1ST_STAGE_CURRENT = 1019,
	RREG_INT_2_1ST_STAGE_CASCODE_CURRENT = 1020,
	RREG_INT_2_2ND_STAGE_CURRENT = 1021,
	RREG_INT_3_1ST_STAGE_CURRENT = 1022,
	RREG_INT_3_1ST_STAGE_CASCODE_CURRENT = 1023,
	RREG_INT_3_2ND_STAGE_CURRENT = 1024,
	RREG_FLASH_BIAS_CURRENT = 1025,
	RREG_FLASH_LADDER_BIAS = 1026,
	RREG_FLASH_LADDER_CASCODE_CURRENT = 1027,
	RREG_FLASH_LADDER_BIAS2 = 1028,
	RREG_RESET = 1029,
	RREG_RX_PFD_CONFIG = 1030,
	RREG_RX_PFD_CONFIG__BYPASS_LD_SYNTH = 1031,
	RREG_RX_INTEGER_BYTE_0 = 1032,
	RREG_RX_INTEGER_BYTE_1 = 1033,
	RREG_RX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD = 1034,
	RREG_RX_FRACT_BYTE_0 = 1035,
	RREG_RX_FRACT_BYTE_1 = 1036,
	RREG_RX_FRACT_BYTE_2 = 1037,
	RREG_RX_FRACT_BYTE_2__SYNTH_FRACT_WORD = 1038,
	RREG_RX_FORCE_ALC = 1039,
	RREG_RX_FORCE_VCO_TUNE_0 = 1040,
	RREG_RX_FORCE_VCO_TUNE_1 = 1041,
	RREG_RX_FORCE_VCO_TUNE_1__VCO_CAL_OFFSET = 1042,
	RREG_RX_ALC_VARACTOR = 1043,
	RREG_RX_ALC_VARACTOR__INIT_ALC_VALUE = 1044,
	RREG_RX_ALC_VARACTOR__VCO_VARACTOR = 1045,
	RREG_RX_VCO_OUTPUT = 1046,
	RREG_RX_VCO_OUTPUT__PORB_VCO_LOGIC = 1047,
	RREG_RX_VCO_OUTPUT__VCO_OUTPUT_LEVEL = 1048,
	RREG_RX_CP_CURRENT = 1049,
	RREG_RX_CP_CURRENT__CHARGE_PUMP_CURRENT = 1050,
	RREG_RX_CP_OFFSET = 1051,
	RREG_RX_CP_OFFSET__SYNTH_RECAL = 1052,
	RREG_RX_CP_CONFIG = 1053,
	RREG_RX_CP_CONFIG__HALF_VCO_CAL_CLK = 1054,
	RREG_RX_CP_CONFIG__F_CPCAL = 1055,
	RREG_RX_CP_CONFIG__CP_CAL_ENABLE = 1056,
	RREG_RX_LOOP_FILTER_1 = 1057,
	RREG_RX_LOOP_FILTER_1__LOOP_FILTER_C2 = 1058,
	RREG_RX_LOOP_FILTER_1__LOOP_FILTER_C1 = 1059,
	RREG_RX_LOOP_FILTER_2 = 1060,
	RREG_RX_LOOP_FILTER_2__LOOP_FILTER_R1 = 1061,
	RREG_RX_LOOP_FILTER_2__LOOP_FILTER_C3 = 1062,
	RREG_RX_LOOP_FILTER_3 = 1063,
	RREG_RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R3 = 1064,
	RREG_RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R1 = 1065,
	RREG_RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C2 = 1066,
	RREG_RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C1 = 1067,
	RREG_RX_LOOP_FILTER_3__LOOP_FILTER_R3 = 1068,
	RREG_RX_DITHERCP_CAL = 1069,
	RREG_RX_DITHERCP_CAL__FORCED_CP_CAL_WORD = 1070,
	RREG_RX_VCO_BIAS_1 = 1071,
	RREG_RX_VCO_BIAS_1__VCO_BIAS_TCF = 1072,
	RREG_RX_VCO_BIAS_1__VCO_BIAS_REF = 1073,
	RREG_RX_CAL_STATUS = 1074,
	RREG_RX_CAL_STATUS__CP_CAL_VALID = 1075,
	RREG_RX_CAL_STATUS__CP_CAL_DONE = 1076,
	RREG_RX_CAL_STATUS__VCO_CAL_BUSY = 1077,
	RREG_RX_CAL_STATUS__CP_CAL_WORD = 1078,
	RREG_RX_VCO_CAL_REF = 1079,
	RREG_RX_VCO_CAL_REF__VCO_CAL_REF_TCF = 1080,
	RREG_RX_VCO_PD_OVERRIDES = 1081,
	RREG_RX_VCO_PD_OVERRIDES__POWER_DOWN_VARACTOR_REF = 1082,
	RREG_RX_VCO_PD_OVERRIDES__PWR_DOWN_VARACT_REF_TCF = 1083,
	RREG_RX_VCO_PD_OVERRIDES__POWER_DOWN_CAL_TCF = 1084,
	RREG_RX_VCO_PD_OVERRIDES__POWER_DOWN_VCO_BUFFFER = 1085,
	RREG_RX_CP_OVERRANGE_VCO_LOCK = 1086,
	RREG_RX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_HIGH = 1087,
	RREG_RX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_LOW = 1088,
	RREG_RX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK = 1089,
	RREG_RX_VCO_LDO = 1090,
	RREG_RX_VCO_LDO__VCO_LDO_BYPASS = 1091,
	RREG_RX_VCO_LDO__VCO_LDO_INRUSH = 1092,
	RREG_RX_VCO_LDO__VCO_LDO_SEL = 1093,
	RREG_RX_VCO_LDO__VCO_LDO_VDROP_SEL = 1094,
	RREG_RX_VCO_CAL = 1095,
	RREG_RX_VCO_CAL__VCO_CAL_EN = 1096,
	RREG_RX_VCO_CAL__VCO_CAL_ALC_WAIT = 1097,
	RREG_RX_VCO_CAL__VCO_CAL_COUNT = 1098,
	RREG_RX_LOCK_DETECT_CONFIG = 1099,
	RREG_RX_LOCK_DETECT_CONFIG__LOCK_DETECT_COUNT = 1100,
	RREG_RX_LOCK_DETECT_CONFIG__LOCK_DETECT_MODE = 1101,
	RREG_RX_CP_LEVEL_DETECT = 1102,
	RREG_RX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_POWER_DOWN = 1103,
	RREG_RX_CP_LEVEL_DETECT__CP_LEVEL_THRESH_LOW = 1104,
	RREG_RX_CP_LEVEL_DETECT__CP_LEVEL_THRESH_HIGH = 1105,
	RREG_RX_DSM_SETUP_0 = 1106,
	RREG_RX_DSM_SETUP_0__DSM_PROG = 1107,
	RREG_RX_DSM_SETUP_1 = 1108,
	RREG_RX_DSM_SETUP_1__SIF_CLOCK = 1109,
	RREG_RX_DSM_SETUP_1__SIF_RESET_BAR = 1110,
	RREG_RX_DSM_SETUP_1__SIF_ADDR = 1111,
	RREG_RX_CORRECTION_WORD0 = 1112,
	RREG_RX_CORRECTION_WORD0__UPDATE_FREQ_WORD = 1113,
	RREG_RX_CORRECTION_WORD0__READ_EFFECTIVE_TUNING_WORD = 1114,
	RREG_RX_CORRECTION_WORD0__FREQ_CORRECTION_WORD_MSB = 1115,
	RREG_RX_CORRECTION_WORD1 = 1116,
	RREG_RX_CORRECTION_WORD1__UPDATE_FREQ_WORD = 1117,
	RREG_RX_CORRECTION_WORD1__FREQ_CORRECTION_WORD_LSB = 1118,
	RREG_RX_VCO_VARACTOR_CTRL_0 = 1119,
	RREG_RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_REFERENCE_TCF = 1120,
	RREG_RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_OFFSET = 1121,
	RREG_RX_VCO_VARACTOR_CTRL_1 = 1122,
	RREG_RX_VCO_VARACTOR_CTRL_1__VCO_VARACTOR_REFERENCE = 1123,
	RREG_RX_FAST_LOCK_SETUP = 1124,
	RREG_RX_FAST_LOCK_SETUP__RX_FAST_LOCK_LOAD_SYNTH = 1125,
	RREG_RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE_INIT = 1126,
	RREG_RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE_PIN_SELECT = 1127,
	RREG_RX_FAST_LOCK_SETUP__RX_FAST_LOCK_MODE_ENABLE = 1128,
	RREG_RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE = 1129,
	RREG_RX_FAST_LOCK_SETUP_INIT_DELAY = 1130,
	RREG_RX_FAST_LOCK_PROGRAM_ADDR = 1131,
	RREG_RX_FAST_LOCK_PROGRAM_ADDR__RX_FAST_LOCK_PROFILE_ADDR = 1132,
	RREG_RX_FAST_LOCK_PROGRAM_ADDR__RX_FAST_LOCK_PROFILE_WORD = 1133,
	RREG_RX_FAST_LOCK_PROGRAM_DATA = 1134,
	RREG_RX_FAST_LOCK_PROGRAM_READ = 1135,
	RREG_RX_FAST_LOCK_PROGRAM_CTRL = 1136,
	RREG_RX_FAST_LOCK_PROGRAM_CTRL__RX_FAST_LOCK_PROGRAM_WRITE = 1137,
	RREG_RX_FAST_LOCK_PROGRAM_CTRL__RX_FAST_LOCK_PROGRAM_CLOCK_ENABLE = 1138,
	RREG_RX_LO_GEN_POWER_MODE = 1139,
	RREG_RX_LO_GEN_POWER_MODE__RX_LO_GEN_POWER_MODE = 1140,
	RREG_TX_PFD_CONFIG = 1141,
	RREG_TX_PFD_CONFIG__DIV_TEST_EN = 1142,
	RREG_TX_PFD_CONFIG__PFD_CLK_EDGE = 1143,
	RREG_TX_PFD_CONFIG__BYPASS_LD_SYNTH = 1144,
	RREG_TX_PFD_CONFIG__PFD_WIDTH = 1145,
	RREG_TX_INTEGER_BYTE_0 = 1146,
	RREG_TX_INTEGER_BYTE_1 = 1147,
	RREG_TX_INTEGER_BYTE_1__SDM_BYPASS = 1148,
	RREG_TX_INTEGER_BYTE_1__SDM_POWER_DOWN = 1149,
	RREG_TX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD = 1150,
	RREG_TX_FRACT_BYTE_0 = 1151,
	RREG_TX_FRACT_BYTE_1 = 1152,
	RREG_TX_FRACT_BYTE_2 = 1153,
	RREG_TX_FRACT_BYTE_2__SYNTH_FRACT_WORD = 1154,
	RREG_TX_FORCE_ALC = 1155,
	RREG_TX_FORCE_ALC__FORCE_ALC_ENABLE = 1156,
	RREG_TX_FORCE_ALC__FORCE_ALC_WORD = 1157,
	RREG_TX_FORCE_VCO_TUNE_0 = 1158,
	RREG_TX_FORCE_VCO_TUNE_1 = 1159,
	RREG_TX_FORCE_VCO_TUNE_1__BYPASS_LOAD_DELAY = 1160,
	RREG_TX_FORCE_VCO_TUNE_1__FORCE_VCO_TUNE_ENABLE = 1161,
	RREG_TX_FORCE_VCO_TUNE_1__FORCE_VCO_TUNE = 1162,
	RREG_TX_FORCE_VCO_TUNE_1__VCO_CAL_OFFSET = 1163,
	RREG_TX_ALCVARACT_OR = 1164,
	RREG_TX_ALCVARACT_OR__INIT_ALC_VALUE = 1165,
	RREG_TX_ALCVARACT_OR__VCO_VARACTOR = 1166,
	RREG_TX_VCO_OUTPUT = 1167,
	RREG_TX_VCO_OUTPUT__PORB_VCO_LOGIC = 1168,
	RREG_TX_VCO_OUTPUT__VCO_OUTPUT_LEVEL = 1169,
	RREG_TX_CP_CURRENT = 1170,
	RREG_TX_CP_CURRENT__TX_CP_CURRENT_DFLT = 1171,
	RREG_TX_CP_CURRENT__VTUNE_FORCE = 1172,
	RREG_TX_CP_CURRENT__CHARGE_PUMP_CURRENT = 1173,
	RREG_TX_CP_OFFSET = 1174,
	RREG_TX_CP_OFFSET__SYNTH_RECAL = 1175,
	RREG_TX_CP_OFFSET__CHARGE_PUMP_OFFSET = 1176,
	RREG_TX_CP_CONFIG = 1177,
	RREG_TX_CP_CONFIG__HALF_VCO_CAL_CLK = 1178,
	RREG_TX_CP_CONFIG__DITHER_MODE = 1179,
	RREG_TX_CP_CONFIG__CP_OFFSET_OFF = 1180,
	RREG_TX_CP_CONFIG__F_CPCAL = 1181,
	RREG_TX_CP_CONFIG__CP_CAL_ENABLE = 1182,
	RREG_TX_CP_CONFIG__CP_TEST = 1183,
	RREG_TX_LOOP_FILTER_1 = 1184,
	RREG_TX_LOOP_FILTER_1__LOOP_FILTER_C2 = 1185,
	RREG_TX_LOOP_FILTER_1__LOOP_FILTER_C1 = 1186,
	RREG_TX_LOOP_FILTER_2 = 1187,
	RREG_TX_LOOP_FILTER_2__LOOP_FILTER_R1 = 1188,
	RREG_TX_LOOP_FILTER_2__LOOP_FILTER_C3 = 1189,
	RREG_TX_LOOP_FILTER_3 = 1190,
	RREG_TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R3 = 1191,
	RREG_TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R1 = 1192,
	RREG_TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C2 = 1193,
	RREG_TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C1 = 1194,
	RREG_TX_LOOP_FILTER_3__LOOP_FILTER_R3 = 1195,
	RREG_TX_DITHERCP_CAL = 1196,
	RREG_TX_DITHERCP_CAL__NUMBER_SDM_DITHER_BITS = 1197,
	RREG_TX_DITHERCP_CAL__FORCED_CP_CAL_WORD = 1198,
	RREG_TX_VCO_BIAS_1 = 1199,
	RREG_TX_VCO_BIAS_1__MUST_BE_ZEROS = 1200,
	RREG_TX_VCO_BIAS_1__VCO_BIAS_TCF = 1201,
	RREG_TX_VCO_BIAS_1__VCO_BIAS_REF = 1202,
	RREG_TX_VCO_BIAS_2 = 1203,
	RREG_TX_VCO_BIAS_2__VCO_BYPASS_BIAS_DAC_R = 1204,
	RREG_TX_VCO_BIAS_2__VCO_COMP_BYPASS_BIAS_R = 1205,
	RREG_TX_VCO_BIAS_2__BYPASS_PRESCALE_R = 1206,
	RREG_TX_VCO_BIAS_2__LAST_ALC_ENABLE = 1207,
	RREG_TX_VCO_BIAS_2__PRESCALE_BIAS = 1208,
	RREG_TX_CAL_STATUS = 1209,
	RREG_TX_CAL_STATUS__CP_CAL_VALID = 1210,
	RREG_TX_CAL_STATUS__COMP_OUT = 1211,
	RREG_TX_CAL_STATUS__CP_CAL_DONE = 1212,
	RREG_TX_CAL_STATUS__VCO_CAL_BUSY = 1213,
	RREG_TX_CAL_STATUS__CP_CAL_WORD = 1214,
	RREG_TX_VCO_CAL_REF = 1215,
	RREG_TX_VCO_CAL_REF__VCO_CAL_REF_MONITOR = 1216,
	RREG_TX_VCO_CAL_REF__VCO_CAL_REF_TCF = 1217,
	RREG_TX_VCO_PD_OVERRIDES = 1218,
	RREG_TX_VCO_PD_OVERRIDES__POWER_DOWN_VARACTOR_REF = 1219,
	RREG_TX_VCO_PD_OVERRIDES__POWER_DOWN_VARACT_REF_TCF = 1220,
	RREG_TX_VCO_PD_OVERRIDES__POWER_DOWN_CAL_TCF = 1221,
	RREG_TX_VCO_PD_OVERRIDES__POWER_DOWN_VCO_BUFFFER = 1222,
	RREG_TX_CP_OVERRANGE_VCO_LOCK = 1223,
	RREG_TX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_HIGH = 1224,
	RREG_TX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_LOW = 1225,
	RREG_TX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK = 1226,
	RREG_TX_VCO_LDO = 1227,
	RREG_TX_VCO_LDO__VCO_LDO_BYPASS = 1228,
	RREG_TX_VCO_LDO__VCO_LDO_INRUSH = 1229,
	RREG_TX_VCO_LDO__VCO_LDO_VOUT_SEL = 1230,
	RREG_TX_VCO_LDO__VCO_LDO_VDROP_SEL = 1231,
	RREG_TX_VCO_CAL = 1232,
	RREG_TX_VCO_CAL__VCO_CAL_EN = 1233,
	RREG_TX_VCO_CAL__VCO_CAL_ALC_WAIT = 1234,
	RREG_TX_VCO_CAL__VCO_CAL_COUNT = 1235,
	RREG_TX_VCO_CAL__FB_CLOCK_ADV = 1236,
	RREG_TX_LOCK_DETECT_CONFIG = 1237,
	RREG_TX_LOCK_DETECT_CONFIG__LOCK_DETECT_COUNT = 1238,
	RREG_TX_LOCK_DETECT_CONFIG__LOCK_DETECT_MODE = 1239,
	RREG_TX_CP_LEVEL_DETECT = 1240,
	RREG_TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_POWER_DOWN = 1241,
	RREG_TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_THRESH_LOW = 1242,
	RREG_TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_THRESH_HIGH = 1243,
	RREG_TX_DSM_SETUP_0 = 1244,
	RREG_TX_DSM_SETUP_0__DSM_PROG = 1245,
	RREG_TX_DSM_SETUP_1 = 1246,
	RREG_TX_DSM_SETUP_1__SIF_CLOCK = 1247,
	RREG_TX_DSM_SETUP_1__SIF_RESET_BAR = 1248,
	RREG_TX_DSM_SETUP_1__SIF_ADDR = 1249,
	RREG_TX_CORRECTION_WORD0 = 1250,
	RREG_TX_CORRECTION_WORD0__UPDATE_FREQ_WORD = 1251,
	RREG_TX_CORRECTION_WORD0__READ_EFFECTIVE_TUNING_WORD = 1252,
	RREG_TX_CORRECTION_WORD0__FREQ_CORRECTION_WORD_MSB = 1253,
	RREG_TX_CORRECTION_WORD1 = 1254,
	RREG_TX_CORRECTION_WORD1__UPDATE_FREQ_WORD = 1255,
	RREG_TX_CORRECTION_WORD1__FREQ_CORRECTION_WORD_LSB = 1256,
	RREG_TX_VCO_VARACTOR_CTRL_0 = 1257,
	RREG_TX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_REFERENCE_TCF = 1258,
	RREG_TX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_OFFSET = 1259,
	RREG_TX_VCO_VARACTOR_CTRL_1 = 1260,
	RREG_TX_VCO_VARACTOR_CTRL_1__VCO_VARACTOR_REFERENCE = 1261,
	RREG_DCXO_COARSE_TUNE = 1262,
	RREG_DCXO_COARSE_TUNE__DCXO_TUNE_COARSE = 1263,
	RREG_DCXO_FINE_TUNE_HIGH = 1264,
	RREG_DCXO_FINE_TUNE_LOW = 1265,
	RREG_DCXO_FINE_TUNE_LOW__DCXO_TUNE_FINE_LOW = 1266,
	RREG_DCXO_CONFIG = 1267,
	RREG_DCXO_CONFIG__MUST_BE_ZERO = 1268,
	RREG_DCXO_CONFIG__DCXO_RTAIL = 1269,
	RREG_DCXO_CONFIG__DCXO_RD = 1270,
	RREG_DCXO_TEMPCO_WRITE = 1271,
	RREG_DCXO_TEMPCO_READ = 1272,
	RREG_DCXO_TEMPCO_ADDR = 1273,
	RREG_DCXO_TEMPCO_ADDR__DCXO_TEMPCO_EN = 1274,
	RREG_DCXO_TEMPCO_ADDR__DCXO_TEMPCO_CLK = 1275,
	RREG_DCXO_TEMPCO_ADDR__DCXO_TEMPERATURE_COEF_ADDRESS = 1276,
	RREG_DELTA_T_READ = 1277,
	RREG_TX_FAST_LOCK_SETUP = 1278,
	RREG_TX_FAST_LOCK_SETUP__TX_FAST_LOCK_LOAD_SYNTH = 1279,
	RREG_TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE_INIT = 1280,
	RREG_TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE_PIN_SELECT = 1281,
	RREG_TX_FAST_LOCK_SETUP__TX_FAST_LOCK_MODE_ENABLE = 1282,
	RREG_TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE = 1283,
	RREG_TX_FAST_LOCK_SETUP_INIT_DELAY = 1284,
	RREG_TX_FAST_LOCK_PROGRAM_ADDR = 1285,
	RREG_TX_FAST_LOCK_PROGRAM_DATA = 1286,
	RREG_TX_FAST_LOCK_PROGRAM_READ = 1287,
	RREG_TX_FAST_LOCK_PROGRAM_CTRL = 1288,
	RREG_TX_FAST_LOCK_PROGRAM_CTRL__TX_FAST_LOCK_PROGRAM_WRITE = 1289,
	RREG_TX_FAST_LOCK_PROGRAM_CTRL__TX_FAST_LOCK_PROGRAM_CLOCK_ENABLE = 1290,
	RREG_TX_LO_GEN_POWER_MODE = 1291,
	RREG_TX_LO_GEN_POWER_MODE__TX_LO_GEN_POWER_MODE = 1292,
	RREG_BANDGAP_CONFIG0 = 1293,
	RREG_BANDGAP_CONFIG0__POWER_DOWN_BANDGAP_REF = 1294,
	RREG_BANDGAP_CONFIG0__MASTER_BIAS_FILTER_BYPASS = 1295,
	RREG_BANDGAP_CONFIG0__MASTER_BIAS_REF_SEL = 1296,
	RREG_BANDGAP_CONFIG0__MASTER_BIAS_TRIM = 1297,
	RREG_BANDGAP_CONFIG1 = 1298,
	RREG_BANDGAP_CONFIG1__VCO_LDO_FILTER_BYPASS = 1299,
	RREG_BANDGAP_CONFIG1__VCO_LDO_REF_SEL = 1300,
	RREG_BANDGAP_CONFIG1__BANDGAP_REF_RESET = 1301,
	RREG_BANDGAP_CONFIG1__BANDGAP_TEMP_TRIM = 1302,
	RREG_REF_DIVIDE_CONFIG_1 = 1303,
	RREG_REF_DIVIDE_CONFIG_1__REF_DIVIDE_CONFIG_1_DFLT = 1304,
	RREG_REF_DIVIDE_CONFIG_1__RX_REF_RESET_BAR = 1305,
	RREG_REF_DIVIDE_CONFIG_1__RX_REF_DIVIDER_MSB = 1306,
	RREG_REF_DIVIDE_CONFIG_2 = 1307,
	RREG_REF_DIVIDE_CONFIG_2__RX_REF_DIVIDER_LSB = 1308,
	RREG_REF_DIVIDE_CONFIG_2__TX_REF_RESET_BAR = 1309,
	RREG_REF_DIVIDE_CONFIG_2__RX_REF_DOUBLER_FB_DELAY = 1310,
	RREG_REF_DIVIDE_CONFIG_2__TX_REF_DIVIDER = 1311,
	RREG_REF_DIVIDE_CONFIG_2__TX_REF_DOUBLER_FB_DELAY = 1312,
	RREG_GAIN_RX1 = 1313,
	RREG_GAIN_RX1__FULL_TABLE_GAIN_INDEX = 1314,
	RREG_LPF_GAIN_RX1 = 1315,
	RREG_LPF_GAIN_RX1__LPF_GAIN_RX = 1316,
	RREG_DIG_GAIN_RX1 = 1317,
	RREG_DIG_GAIN_RX1__DIGITAL_GAIN_RX = 1318,
	RREG_FAST_ATTACK_STATE = 1319,
	RREG_FAST_ATTACK_STATE__FAST_ATTACK_STATE_RX2 = 1320,
	RREG_FAST_ATTACK_STATE__FAST_ATTACK_STATE_RX1 = 1321,
	RREG_SLOW_LOOP_STATE = 1322,
	RREG_SLOW_LOOP_STATE__SLOW_LOOP_STATE_RX2 = 1323,
	RREG_SLOW_LOOP_STATE__SLOW_LOOP_STATE_RX1 = 1324,
	RREG_GAIN_RX2 = 1325,
	RREG_GAIN_RX2__FULL_TABLE_GAIN_INDEX = 1326,
	RREG_LPF_GAIN_RX2 = 1327,
	RREG_LPF_GAIN_RX2__LPF_GAIN_RX = 1328,
	RREG_DIG_GAIN_RX2 = 1329,
	RREG_DIG_GAIN_RX2__DIGITAL_GAIN_RX = 1330,
	RREG_OVRG_SIGS_RX1 = 1331,
	RREG_OVRG_SIGS_RX1__GAIN_LOCK_1 = 1332,
	RREG_OVRG_SIGS_RX1__LOW_POWER_1 = 1333,
	RREG_OVRG_SIGS_RX1__LARGE_LMT_OL = 1334,
	RREG_OVRG_SIGS_RX1__SMALL_LMT_OL = 1335,
	RREG_OVRG_SIGS_RX1__LARGE_ADC_OL = 1336,
	RREG_OVRG_SIGS_RX1__SMALL_ADC_OL = 1337,
	RREG_OVRG_SIGS_RX1__DIG_SAT = 1338,
	RREG_OVRG_SIGS_RX2 = 1339,
	RREG_OVRG_SIGS_RX2__GAIN_LOCK_1 = 1340,
	RREG_OVRG_SIGS_RX2__LOW_POWER_1 = 1341,
	RREG_OVRG_SIGS_RX2__LARGE_LMT_OL = 1342,
	RREG_OVRG_SIGS_RX2__SMALL_LMT_OL = 1343,
	RREG_OVRG_SIGS_RX2__LARGE_ADC_OL = 1344,
	RREG_OVRG_SIGS_RX2__SMALL_ADC_OL = 1345,
	RREG_OVRG_SIGS_RX2__DIG_SAT = 1346,
	RREG_CTRL = 1347,
	RREG_CTRL__CTRL_ENABLE = 1348,
	RREG_BIST_CONFIG = 1349,
	RREG_BIST_CONFIG__TONE_PRBS = 1350,
	RREG_BIST_CONFIG__BIST_ENABLE = 1351,
	RREG_BIST_CONFIG__TONE_FREQ = 1352,
	RREG_BIST_CONFIG__TONE_LEVEL = 1353,
	RREG_BIST_CONFIG__BIST_CTRL_POINT = 1354,
	RREG_OBSERVE_CONFIG = 1355,
	RREG_OBSERVE_CONFIG__DATA_PORT_SP_HD_LOOP_TEST_OE = 1356,
	RREG_OBSERVE_CONFIG__RX_MASK = 1357,
	RREG_OBSERVE_CONFIG__CHANNEL = 1358,
	RREG_OBSERVE_CONFIG__DATA_PORT_LOOP_TEST_ENABLE = 1359,
	RREG_OBSERVE_CONFIG__OBSERVATION_POINT = 1360,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG = 1361,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_2_Q_DATA = 1362,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_2_I_DATA = 1363,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_1_Q_DATA = 1364,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_1_I_DATA = 1365,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__DATA_PORT_HILOW = 1366,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__USE_DATA_PORT = 1367,
	RREG_BIST_AND_DATA_PORT_TEST_CONFIG__TEMP_SENSE_VBE_TEST = 1368,
	RREG_DAC_TEST_0 = 1369,
	RREG_DAC_TEST_1 = 1370,
	RREG_DAC_TEST_2 = 1371,
	RREG_DAC_TEST_2__DAC_TEST_ENABLE = 1372,
	RREG_DAC_TEST_2__DAC_TEST_WORD = 1373,
	RREG_ = -1 /* force the use of a signed integer */
} sx1301ar_rreg_t;

#define SX1301AR_RREG_NB 1375

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */

/**
@warning This module has internal states: for each board are stored functions
pointer for SPI access, and internal state of the SPi interface (top level SPI
multiplexer state, FPGA page selection sate, and page state of each SX1301).
Because of that, it's not possible to combine direct SPI access to the
concentrator board and access through that module of the library.
*/

/**
@brief Setup callback to communicate with the board
@param brd        Target board number
@param brd_type   Type of board (Master or Slave)
@param spi_read   Pointer to a SPI read function
@param spi_write  Pointer to a SPI write function
@return 0 if board setup parameters are ok, -1 else
*/
int sx1301ar_reg_setup(uint8_t brd, uint8_t brd_type, spi_r_t spi_read, spi_w_t spi_write);

/**
@brief Initiate communication with the board and check hardware version
@param brd          Target board number
@param nb_chip      Number of SX1301 chips
@param fpga_version FPGA version
@return 0 if the communication with the board is working, -1 else
*/
int sx1301ar_reg_init(uint8_t brd, uint8_t nb_chip, int16_t *fpga_version);

/**
@brief Do a soft-reset of the board registers
@param brd  Target board number
@return 0 if the board was reset successfully, -1 else
*/
int sx1301ar_reg_reset(uint8_t brd);

/**
@brief Read a register of the top FPGA of the board
@param brd  Target board number
@param idx  Target *fpga* register index
@param val  Pointer to a variable that will contain the read value
@return 0 if the register was read successfully, -1 else
*/
int sx1301ar_reg_brd_r(uint8_t brd, sx1301ar_breg_t idx, int32_t *val);

/**
@brief Write a register of the top FPGA of the board
@param brd  Target board number
@param idx  Target *fpga* register index
@param val  Value to write to the register
@return 0 if the register was written successfully, -1 else
*/
int sx1301ar_reg_brd_w(uint8_t brd, sx1301ar_breg_t idx, int32_t val);

/**
@brief Do a 'burst read' on a register of the top FPGA of the board
@param brd   Target board number
@param idx   Target *fpga* register index
@param data  Pointer to a buffer that will receive data read from the register
@param size  Number of bytes to read
@return 0 if the 'burst read' was successful, -1 else
*/
int sx1301ar_reg_brd_rb(uint8_t brd, sx1301ar_breg_t idx, uint8_t *data, uint32_t size);

/**
@brief Do a 'burst write' on a register of the top FPGA of the board
@param brd   Target board number
@param idx   Target *fpga* register index
@param data  Pointer to a buffer containing bytes to write to the register
@param size  Number of bytes to write
@return 0 if the 'burst write' was successful, -1 else
*/
int sx1301ar_reg_brd_wb(uint8_t brd, sx1301ar_breg_t idx, const uint8_t *data, uint32_t size);

/**
@brief Read a register of one of the SX1301 of the board
@param brd   Target board number
@param chip  Target chip number in the array
@param idx   Target *sx1301* register index
@param val   Pointer to a variable that will contain the read value
@return 0 if the register was read successfully, -1 else
*/
int sx1301ar_reg_chip_r(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, int32_t *val);

/**
@brief Write a register of one of the SX1301 of the board
@param brd   Target board number
@param chip  Target chip number in the array
@param idx   Target *sx1301* register index
@param val   Value to write to the register
@return 0 if the register was written successfully, -1 else
*/
int sx1301ar_reg_chip_w(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, int32_t val);

/**
@brief Do a 'burst read' on a register of one of the SX1301 of the board
@param brd   Target board number
@param chip  Target chip number in the array
@param idx   Target *sx1301* register index
@param data  Pointer to a buffer that will receive data read from the register
@param size  Number of bytes to read
@return 0 if the 'burst read' was successful, , -1 else
*/
int sx1301ar_reg_chip_rb(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, uint8_t *data, uint32_t size);

/**
@brief Do a 'burst write' on a register of one of the SX1301 of the board
@param brd   Target board number
@param chip  Target chip number in the array
@param idx   Target *sx1301* register index
@param data  Pointer to a buffer containing bytes to write to the register
@param size  Number of bytes to write
@return 0 if the 'burst write' was successful, -1 else
*/
int sx1301ar_reg_chip_wb(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, const uint8_t *data, uint32_t size);

/**
@brief Read a register of the AD9361 radio chip via the FPGA SPI bridge
@param brd   Target board number
@param addr  Radio register address (12 bits)
@param data  Pointer to a buffer that will receive data read from the radio
@param size  Data transfer size (1 to 8 bytes)
@return 0 if the radio register was read successfully, -1 else
*/
int sx1301ar_ad9361_r(uint8_t brd, sx1301ar_rreg_t idx, uint8_t *val);

/**
@brief Write a register of the AD9361 radio chip
@param brd   Target board number
@param addr  Radio register address (12 bits)
@param data  Pointer to a buffer containing bytes to write to the radio
@param size  Data transfer size (1 to 8 bytes)
@return 0 if the radio register was written successfully, -1 else
*/
int sx1301ar_ad9361_w(uint8_t brd, sx1301ar_rreg_t idx, uint8_t val);

/**
@brief Low-level SPI bridge function needed by the AD9361 driver
@param brd      Target board number
@param tx_data  Pointer to a buffer containing bytes to write to SPI bus
@param tx_size  Number of bytes to write
@param rx_data  Pointer to a buffer that will receive the received data
@param rx_size  Number of bytes to read
@return 0 if the transfer was successful, -1 else

In a single SPI transaction, write X bytes (while ignoring received data), then
read Y bytes (while sending null data).
*/
int sx1301ar_spibdg_w_r(uint8_t brd, const uint8_t *tx_data, uint8_t tx_size, uint8_t *rx_data, uint8_t rx_size);

#endif

/* --- EOF ------------------------------------------------------------------ */
