/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    SPI functions using Linux device driver (typically /dev/spidev{number})

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin

*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <fcntl.h>  /* open */
#include <stdint.h> /* C99 types */
#include <string.h> /* memset */
#include <sys/mman.h>
#include <unistd.h> /* lseek, close */

#include <linux/spi/spidev.h>
#include <sys/ioctl.h>

#include "spi.h"

#include "utils/logger.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#define SPI_SUCCESS 0
#define SPI_ERROR -1

#define READ_ACCESS 0x00
#define WRITE_ACCESS 0x80
#define MAX_CHUNK 1024
#define MAX_SPI_BLK_SIZE 4095
#define FAST_SIZE 10 /* 4 for SMTC regs, 10 to also accelerate AD reg access */

#define SPI_DEFAULT_SPEED 1000000

/* SPI mux header constants */
#define SPI_HDR_REG_FPGA 0x02 /* SPI header for FPGA registers */
#define SPI_HDR_REG_FPGA2 0x0E

/* Mem Mapping constants */
#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int spi_linuxdev_open(const char *path, int speed_mhz, int *spi_ref)
{
	int i, j; /* return values */
	int dev;
	int val; /* setting value */

	/* Check input variables */
	if (path == NULL) {
		error("null pointer path");
		return SPI_ERROR;
	}
	if (spi_ref == NULL) {
		error("null pointer spi_ref");
		return SPI_ERROR;
	}

	/* Open SPI device */
	dev = open(path, O_RDWR);
	if (dev < 0) {
		error("open() function failed - %s (%s)", strerror(errno), path);
		return SPI_ERROR;
	}

	/* Setting SPI mode to 'mode 0' */
	val = SPI_MODE_0;
	i = ioctl(dev, SPI_IOC_WR_MODE, &val);
	j = ioctl(dev, SPI_IOC_RD_MODE, &val);
	if ((i < 0) || (j < 0)) {
		error("ioctl() failed to set SPI mode - %s", strerror(errno));
		close(dev);
		return SPI_ERROR;
	}

	/* Setting SPI maximum clock speed (in Hz) */
	val = (speed_mhz <= 0) ? SPI_DEFAULT_SPEED : speed_mhz * 1000000;
	i = ioctl(dev, SPI_IOC_WR_MAX_SPEED_HZ, &val);
	j = ioctl(dev, SPI_IOC_RD_MAX_SPEED_HZ, &val);
	if ((i < 0) || (j < 0)) {
		error("ioctl() failed to set SPI maximum clock speed - %s", strerror(errno));
		close(dev);
		return SPI_ERROR;
	} else {
		debug("SPI speed set to %d Hz", val);
	}

	/* Setting SPI bit endianness (most significant bit first) */
	val = 0;
	i = ioctl(dev, SPI_IOC_WR_LSB_FIRST, &val);
	j = ioctl(dev, SPI_IOC_RD_LSB_FIRST, &val);
	if ((i < 0) || (j < 0)) {
		error("ioctl() failed to set SPI bit endianness - %s", strerror(errno));
		close(dev);
		return SPI_ERROR;
	}

	/* Setting SPI word size (8-bit words) */
	val = 0;
	i = ioctl(dev, SPI_IOC_WR_BITS_PER_WORD, &val);
	j = ioctl(dev, SPI_IOC_RD_BITS_PER_WORD, &val);
	if ((i < 0) || (j < 0)) {
		error("ioctl() failed to set SPI word size - %s", strerror(errno));
		close(dev);
		return SPI_ERROR;
	}

	debug("SPI port opened successfully");
	*spi_ref = dev; /* return file descriptor index */
	return SPI_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int spi_linuxdev_close(int spi_ref)
{
	int i;

	i = close(spi_ref);

	if (i == 0) {
		debug("SPI port closed successfully");
		return SPI_SUCCESS;
	} else {
		error("close( ) function failed");
		return SPI_ERROR;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int spi_linuxdev_read(uint8_t header, int spi_ref, uint16_t address, uint8_t *data, uint32_t size, uint8_t *status)
{
	int i;
	uint8_t bufo[FAST_SIZE + 3]; /* add header and address bytes */
	uint8_t bufi[FAST_SIZE + 3];
	struct spi_ioc_transfer k[2];
	uint8_t addr_size; /* number of bytes to code address (1 or 2) */
	uint32_t size_to_do, chunk_size, offset;
	uint32_t byte_transfered = 0;

	/* Check input variables */
	if ((size > 0) && (data == NULL)) {
		error("null pointer data");
		return SPI_ERROR;
	}
	/* 16bits address for FPGA, 8bits address for everyhting else */
	if ((((header == SPI_HDR_REG_FPGA) || (header == SPI_HDR_REG_FPGA2)) && (address > 32768)) || ((header != SPI_HDR_REG_FPGA) && (header != SPI_HDR_REG_FPGA2) && (address > 128))) {
		error("invalid SPI address %i", address);
		return SPI_ERROR;
	}

	/* Clear spi_ioc_transfer structure (ensure compatibility) */
	memset(&k, 0, sizeof k);

	/* Setup command/status transfer */
	bufo[0] = header;
	if ((header == SPI_HDR_REG_FPGA) || (header == SPI_HDR_REG_FPGA2)) {
		/* 16 bits address */
		addr_size = 2;
		bufo[1] = READ_ACCESS | ((address >> 8) & 0x7F);
		bufo[2] = address & 0xFF;
	} else {
		/* 8 bits address */
		addr_size = 1;
		bufo[1] = READ_ACCESS | (address & 0x7F);
	}
	memset(bufo + addr_size + 1, 0, FAST_SIZE);

	if (size <= FAST_SIZE) {
		/* Accelerate registers access (transfers of 0 to FAST_SIZE bytes ) */
		k[0].tx_buf = (uintptr_t)bufo;
		k[0].rx_buf = (uintptr_t)bufi;
		k[0].len = 1 + addr_size + size; /* add header + address bytes */
		k[0].cs_change = 0;
		i = ioctl(spi_ref, SPI_IOC_MESSAGE(1), &k);
		if (i != (int)k[0].len) {
			error("spi_linuxdev_read failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
		if (size > 0) {
			memcpy((void *)data, (void *)(bufi + 1 + addr_size), size); /* skip first 2 or 3 bytes */
		}
		if (status != NULL) {
			*status = bufi[1];
		}
	} else {
		/* I/O transaction */
		k[0].tx_buf = (uintptr_t)&bufo[0];
		k[0].len = 1 + addr_size;
		k[0].cs_change = 0;
		k[1].cs_change = 0;
		size_to_do = size;
		for (i = 0; size_to_do > 0; ++i) {
			chunk_size = (size_to_do < MAX_CHUNK) ? size_to_do : MAX_CHUNK;
			offset = i * MAX_CHUNK;
			k[1].rx_buf = (uintptr_t)(data + offset);
			k[1].len = chunk_size;
			byte_transfered += (ioctl(spi_ref, SPI_IOC_MESSAGE(2), &k) - k[0].len);
			size_to_do -= chunk_size; /* subtract the quantity of data already transferred */
		}

		/* determine return code */
		if (byte_transfered != size) {
			error("SPI BURST READ FAILURE");
			return SPI_ERROR;
		}
	}

	return SPI_SUCCESS;
}

int spi_linuxdev_read_flash_mem(int spi_ref, uint8_t cmd, uint8_t *address, uint8_t address_size, uint8_t *data, uint32_t size)
{
	int i;
	uint8_t bufo[FAST_SIZE + 4]; /* add header and address bytes */
	uint8_t bufi[FAST_SIZE + 4];
	struct spi_ioc_transfer k[2];

	/* Check input variables */
	if ((size > 0) && (data == NULL)) {
		error("null pointer data");
		return SPI_ERROR;
	}

	if (size >= MAX_SPI_BLK_SIZE) {
		error("Maximum SPI block size");
		return SPI_ERROR;
	}

	/* Clear spi_ioc_transfer structure (ensure compatibility) */
	memset(&k, 0, sizeof k);

	/* Setup command transfer */
	bufo[0] = cmd;
	if (address_size)
		memcpy(bufo + 1, address, address_size);

	memset(bufo + 1 + address_size, 0, FAST_SIZE);

	k[0].delay_usecs = 0;
	if (size <= FAST_SIZE) {
		/* Accelerate registers access (transfers of 0 to FAST_SIZE bytes ) */
		k[0].tx_buf = (uintptr_t)bufo;
		k[0].rx_buf = (uintptr_t)bufi;
		k[0].len = 1 + address_size + size; /* address bytes */
		k[0].cs_change = 0;

		i = ioctl(spi_ref, SPI_IOC_MESSAGE(1), k);
		if (i != (int)k[0].len) {
			error("spi_linuxdev_read failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
		if (size > 0) {
			memcpy((void *)data, (void *)(bufi + 1 + address_size), size); /* skip first 2 or 3 bytes */
		}
	} else {
		/* I/O transaction */
		k[0].tx_buf = (uintptr_t)&bufo[0];
		k[1].rx_buf = (uintptr_t)data;
		k[0].len = 1 + address_size;
		k[0].cs_change = 0;
		k[1].cs_change = 0;
		k[1].len = size;

		i = ioctl(spi_ref, SPI_IOC_MESSAGE(2), &k);
		if (i != (int)(k[0].len + k[1].len)) {
			error("spi_linuxdev_read failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
	}

	return SPI_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int spi_linuxdev_write(uint8_t header, int spi_ref, uint16_t address, const uint8_t *data, uint32_t size, uint8_t *status)
{
	int i;
	uint8_t bufo[FAST_SIZE + 3]; /* add header and address bytes */
	uint8_t bufi[FAST_SIZE + 3];
	struct spi_ioc_transfer k[2];
	uint8_t addr_size;
	uint32_t size_to_do, chunk_size, offset;
	uint32_t byte_transfered = 0;

	/* Check input variables */
	if ((size > 0) && (data == NULL)) {
		error("null pointer data");
		return SPI_ERROR;
	}
	/* 16bits address for FPGA, 8bits address for everyhting else */
	if ((((header == SPI_HDR_REG_FPGA) || (header == SPI_HDR_REG_FPGA2)) && (address > 32768)) || ((header != SPI_HDR_REG_FPGA) && (header != SPI_HDR_REG_FPGA2) && (address > 128))) {
		error("invalid SPI address %i", address);
		return SPI_ERROR;
	}

	/* Clear spi_ioc_transfer structure (ensure compatibility) */
	memset(&k, 0, sizeof k);

	/* Setup command/status transfer */
	bufo[0] = header;
	if ((header == SPI_HDR_REG_FPGA) || (header == SPI_HDR_REG_FPGA2)) {
		/* 16 bits address */
		addr_size = 2;
		bufo[1] = WRITE_ACCESS | ((address >> 8) & 0x7F);
		bufo[2] = address & 0xFF;
	} else {
		/* 8 bits address */
		addr_size = 1;
		bufo[1] = WRITE_ACCESS | (address & 0x7F);
	}

	if (size <= FAST_SIZE) {
		/* Accelerate registers access (transfer of 0 to FAST_SIZE bytes ) */
		if (size > 0) {
			memcpy((void *)(bufo + 1 + addr_size), (void *)data, size);
		}
		k[0].tx_buf = (uintptr_t)bufo;
		k[0].rx_buf = (uintptr_t)bufi;
		k[0].len = 1 + addr_size + size; /* add header + address bytes */
		k[0].cs_change = 0;
		i = ioctl(spi_ref, SPI_IOC_MESSAGE(1), &k);
		if (i != (int)k[0].len) {
			error("spi_linuxdev_write failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
		if (status != NULL) {
			*status = bufi[1];
		}
	} else {
		/* I/O transaction */
		k[0].tx_buf = (uintptr_t)&bufo[0];
		k[0].len = 1 + addr_size;
		k[0].cs_change = 0;
		k[1].cs_change = 0;
		size_to_do = size;
		for (i = 0; size_to_do > 0; ++i) {
			chunk_size = (size_to_do < MAX_CHUNK) ? size_to_do : MAX_CHUNK;
			offset = i * MAX_CHUNK;
			k[1].tx_buf = (uintptr_t)(data + offset);
			k[1].len = chunk_size;
			byte_transfered += (ioctl(spi_ref, SPI_IOC_MESSAGE(2), &k) - k[0].len);
			size_to_do -= chunk_size; /* subtract the quantity of data already transferred */
		}

		/* determine return code */
		if (byte_transfered != size) {
			error("SPI BURST WRITE FAILURE");
			return SPI_ERROR;
		}
	}

	return SPI_SUCCESS;
}

int spi_linuxdev_write_flash_mem(int spi_ref, uint8_t cmd, uint8_t *address, uint8_t address_size, const uint8_t *data, uint32_t size)
{
	int i;
	uint8_t bufo[FAST_SIZE + 3]; /* add header and address bytes */
	uint8_t bufi[FAST_SIZE + 3];
	struct spi_ioc_transfer k[2];

	/* Check input variables */
	if ((size > 0) && (data == NULL)) {
		error("null pointer data");
		return SPI_ERROR;
	}

	if (size >= MAX_SPI_BLK_SIZE) {
		error("Maximum SPI block size");
		return SPI_ERROR;
	}

	/* Clear spi_ioc_transfer structure (ensure compatibility) */
	memset(&k, 0, sizeof k);

	/* Setup command transfer */
	bufo[0] = cmd;
	if (address_size)
		memcpy(bufo + 1, address, address_size);

	memset(bufo + 1 + address_size, 0, FAST_SIZE);

	k[0].delay_usecs = 0;

	if (size <= FAST_SIZE) {
		/* Accelerate registers access (transfer of 0 to FAST_SIZE bytes ) */
		if (size > 0) {
			memcpy((void *)(bufo + 1 + address_size), (void *)data, size);
		}
		k[0].tx_buf = (uintptr_t)bufo;
		k[0].rx_buf = (uintptr_t)bufi;
		k[0].len = 1 + size + address_size; /* cmd + address bytes */
		k[0].cs_change = 0;

		i = ioctl(spi_ref, SPI_IOC_MESSAGE(1), &k);
		if (i != (int)k[0].len) {
			error("spi_linuxdev_write failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
	} else {
		/* I/O transaction */
		k[0].tx_buf = (unsigned long)&bufo[0];
		k[0].len = 1 + address_size; /* cmd + address bytes */
		k[0].cs_change = 0;
		k[1].tx_buf = (unsigned long)data;
		k[1].len = size;

		i = ioctl(spi_ref, SPI_IOC_MESSAGE(2), &k);
		if (i != (int)(k[0].len + k[1].len)) {
			error("spi_linuxdev_write failed, ioctl returned %i (errno %i)", i, errno);
			return SPI_ERROR;
		}
	}

	return SPI_SUCCESS;
}

/* --- EOF ------------------------------------------------------------------ */
