// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "daemon.h"

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/logger.h"

// The daemonize() function will follow the steps define here:
// https://www.freedesktop.org/software/systemd/man/daemon.html
//
// This function should be called at the very beginning of the process, this way
// the steps 1 to 4 can be skipped.
//
// This function handle the steps 5 to 11. The step 12 is handled by the
// create_pidfile function.
//
// The step 13 is not applicable. And to make it easier the steps 14 and 15 are
// not implemented.
int daemonize(void)
{
	pid_t pid;
	int fd;

	if ((pid = fork()) < 0) {
		eerror("Failed to create a child process");
		return -1;

	} else if (pid > 0) {
		return pid;

	} else if (setsid() < 0) {
		eerror("Failed to create a new session");
		return -1;

	} else if ((pid = fork()) < 0) {
		eerror("Failed to create a second child process");
		return -1;

	} else if (pid > 0) {
		return pid;

	} else if (close(STDIN_FILENO) < 0) {
		eerror("Failed to close stdin");
		return -1;

	} else if ((fd = open("/dev/null", O_RDONLY)) < 0) {
		eerror("Failed to open stdin");
		return -1;

	} else if (fd != STDIN_FILENO) {
		eerror("Mismatch stdin file descriptor");
		return -1;

	} else if (close(STDOUT_FILENO) < 0) {
		eerror("Failed to close stdout");
		return -1;

	} else if ((fd = open("/dev/null", O_WRONLY)) < 0) {
		eerror("Failed to open stdout");
		return -1;

	} else if (fd != STDOUT_FILENO) {
		eerror("Mismatch stdout file descriptor");
		return -1;

	} else if (close(STDERR_FILENO) < 0) {
		eerror("Failed to close stderr");
		return -1;

	} else if ((fd = dup(STDOUT_FILENO)) < 0) {
		eerror("Failed to duplicate stderr");
		return -1;

	} else if (fd != STDERR_FILENO) {
		eerror("Mismatch stderr file descriptor");
		return -1;

	} else if (chdir("/") < 0) {
		eerror("Failed to change working directory");
		return -1;

	} else {
		umask(0);
	}

	return 0;
}

int handle_signals(const struct signal *signals)
{
	for (; signals->num != 0 && signals->handler != NULL; signals++) {
		if (signal(signals->num, signals->handler) == SIG_ERR) {
			eerror("Failed to handle signal: %s",
			       strsignal(signals->num));
			return -1;
		}
	}

	return 0;
}

// The lock on the pid file ensure that the process cannot be started more than
// once.
int create_pidfile(const char *file)
{
	char pidstr[32];
	int fd;

	if ((fd = open(file, O_CREAT | O_WRONLY, 0644)) < 0) {
		eerror("Failed to open pid file: %s", file);
		return -1;

	} else if (lockf(fd, F_TLOCK, 0) < 0) {
		eerror("Failed to lock pid file: %s", file);
		close(fd);
		return -1;

	} else if (ftruncate(fd, 0) < 0) {
		eerror("Failed to clear pid file: %s", file);
		close(fd);
		return -1;
	}

	snprintf(pidstr, sizeof(pidstr), "%u\n", getpid());

	if (write(fd, pidstr, strlen(pidstr)) < 0) {
		eerror("Failed to write pid file: %s", file);
		close(fd);
		return -1;
	}

	return fd;
}
