/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Register access functions for 'LoRa Gateway v2' concentrator driver library

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdbool.h> /* bool type */
#include <stdint.h>  /* C99 types */
#include <string.h>

#include "aux.h"
#include "err.h"
#include "hal.h" /* for constants */
#include "reg.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS & TYPES -------------------------------------------- */

#define SPI_BG_FIFO_SIZE 16
#define SPI_BG_TIMEOUT 50

/**
@brief Register definition structure
*/
struct reg {
	int8_t page;   /*!< page containing the register (-1 for all pages) */
	uint16_t addr; /*!< base address of the register (7 or 15 bits) */
	uint8_t offs;  /*!< position of the register LSB (between 0 to 7) */
	bool sign;     /*!< 1 indicates the register is signed (2 complem.) */
	uint8_t leng;  /*!< number of bits in the register */
	bool rdon;     /*!< 1 indicates a read-only register */
	int32_t dflt;  /*!< register default value */
};

/**
@brief Another register definition structure: no page, no sign and no default
*/
struct reg2 {
	uint16_t addr; /*!< base address of the register */
	uint8_t offs;  /*!< position of the register LSB (between 0 to 7) */
	uint8_t leng;  /*!< number of bits in the register */
};

struct brd_reg {
	sx1301ar_breg_t breg_idx;
	struct reg breg;
};

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#define RETURN_SPI_W_ERR(a) \
	if ((a) != 0)       \
	SX1301AR_ERR_RETURN(ERR_WRITE_SPI)
#define RETURN_SPI_R_ERR(a) \
	if ((a) != 0)       \
	SX1301AR_ERR_RETURN(ERR_READ_SPI)

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

/**
@brief Array of struct, contains the complete configuration for all boards
*/
static struct
{
	spi_r_t spi_r;					 /* SPI read callback for a specific board */
	spi_w_t spi_w;					 /* SPI write callback for a specific board */
	int8_t fpga_page_state;				 /* state of the FPGA page register */
	int8_t chip_page_state[SX1301AR_BOARD_CHIPS_NB]; /* state of the page register for each SX1301 */
	uint8_t header_reg_fpga;			 /* header value to access FPGA registers */
	uint8_t header_reg_chip;			 /* header value to access chip registers */
} brd_stat_priv[SX1301AR_MAX_BOARD_NB];

/**
@brief Registers of the FPGA (top level of the board hierarchy)
*/
static struct reg breg_table[SX1301AR_MAX_BOARD_NB][SX1301AR_BREG_NB];

static const struct brd_reg breg_master[] =
    {
	{BREG_SOFT_RESET, {-1, 0, 7, 0, 1, 0, 0}},			    /*!< SOFT_RESET */
	{BREG_VERSION, {-1, 1, 0, 0, 8, 1, SX1301AR_BOARD_VERSION_MASTER}}, /*!< VERSION */
	{BREG_ALT_CHIP_ID_LSB, {-1, 2, 0, 0, 32, 1, 0}},		    /*!< ALT_CHIP_ID_LSB */
	{BREG_ALT_CHIP_ID_MSB, {-1, 6, 0, 0, 32, 1, 0}},		    /*!< ALT_CHIP_ID_MSB */
	{BREG_AD9361_SPI_DATA_OUT, {-1, 64, 0, 0, 8, 0, 0}},		    /*!< AD9361_SPI_DATA_OUT */
	{BREG_AD9361_SPI_DATA_IN, {-1, 65, 0, 0, 8, 1, 0}},		    /*!< AD9361_SPI_DATA_IN */
	{BREG_AD9361_SPI_CHIP_SELECT_POLARITY, {-1, 66, 1, 0, 1, 0, 0}},    /*!< AD9361_SPI_CHIP_SELECT_POLARITY */
	{BREG_AD9361_SPI_CPOL, {-1, 66, 2, 0, 1, 0, 0}},		    /*!< AD9361_SPI_CPOL */
	{BREG_AD9361_SPI_CPHA, {-1, 66, 3, 0, 1, 0, 0}},		    /*!< AD9361_SPI_CPHA */
	{BREG_AD9361_SPI_LOOPBACK_EN, {-1, 66, 4, 0, 1, 0, 0}},		    /*!< AD9361_SPI_LOOPBACK_EN */
	{BREG_AD9361_SPI_HOST_CTRL, {-1, 66, 5, 0, 1, 0, 1}},		    /*!< AD9361_SPI_HOST_CTRL */
	{BREG_AD9361_SPI_SPEED, {-1, 67, 0, 0, 4, 0, 0}},		    /*!< AD9361_SPI_SPEED */
	{BREG_AD9361_SPI_DBG_WR_CNT, {-1, 67, 4, 0, 4, 1, 0}},		    /*!< AD9361_SPI_DBG_WR_CNT */
	{BREG_AD9361_SPI_SEL, {-1, 68, 0, 0, 1, 0, 0}},			    /*!< AD9361_SPI_SEL */
	{BREG_AD9361_SPI_DBG_STATE, {-1, 69, 0, 0, 3, 1, 0}},		    /*!< AD9361_SPI_DBG_STATE */
	{BREG_AD9361_RESETB, {-1, 70, 0, 0, 1, 0, 0}},			    /*!< AD9361_RESETB */
	{BREG_AD9361_ENABLE, {-1, 70, 1, 0, 1, 0, 0}},			    /*!< AD9361_ENABLE */
	{BREG_AD9361_TXNRX, {-1, 70, 2, 0, 1, 0, 0}},			    /*!< AD9361_TXNRX */
	{BREG_AD9361_EN_AGC, {-1, 70, 3, 0, 1, 0, 0}},			    /*!< AD9361_EN_AGC */
	{BREG_AD9361_TXRX_HOST_CTRL, {-1, 70, 4, 0, 1, 0, 1}},		    /*!< AD9361_TXRX_HOST_CTRL */
	{BREG_AD9361_CTRL_IN, {-1, 71, 0, 0, 4, 0, 1}},			    /*!< AD9361_CTRL_IN */
	{BREG_AD9361_CTRL_OUT, {-1, 72, 0, 0, 8, 1, 1}},		    /*!< AD9361_CTRL_OUT */
	{BREG_IF_FREQ_0, {-1, 128, 0, 1, 13, 0, -384}},			    /*!< IF_FREQ_0 */
	{BREG_IF_FREQ_1, {-1, 130, 0, 1, 13, 0, -128}},			    /*!< IF_FREQ_1 */
	{BREG_RSSI_THRESHOLD_HIGH, {-1, 192, 0, 0, 16, 0, 0}},		    /*!< RSSI_THRESHOLD_HIGH */
	{BREG_RSSI_THRESHOLD_LOW, {-1, 194, 0, 0, 16, 0, 0}},		    /*!< RSSI_THRESHOLD_LOW */
	{BREG_RSSI_DEC_FILTER_ALPHA, {-1, 196, 0, 0, 4, 0, 7}},		    /*!< RSSI_DEC_FILTER_ALPHA */
	{BREG_RSSI_MIN_ATTEN, {-1, 197, 0, 0, 5, 0, 7}},		    /*!< RSSI_MIN_ATTEN */
	{BREG_RSSI_MAX_ATTEN, {-1, 198, 0, 0, 5, 0, 11}},		    /*!< RSSI_MAX_ATTEN */
	{BREG_RSSI_ATTEN_0, {-1, 199, 0, 0, 5, 1, 0}},			    /*!< RSSI_ATTEN_0 */
	{BREG_RSSI_ATTEN_1, {-1, 200, 0, 0, 5, 1, 0}},			    /*!< RSSI_ATTEN_1 */
	{BREG_FIR_GAIN, {-1, 207, 0, 0, 5, 0, 0}},			    /*!< FIR_GAIN */
	{BREG_OVERRIDE_GAIN, {-1, 207, 5, 0, 1, 0, 0}},			    /*!< OVERRIDE_GAIN */
	{BREG_FIR_GAIN_STEP, {-1, 208, 0, 0, 5, 0, 0}},			    /*!< FIR_GAIN_STEP */
	{BREG_DC_NOTCH_EN, {-1, 209, 0, 0, 8, 0, 0}},			    /*!< DC_NOTCH_EN */
	{BREG_DSP_ATTEN, {-1, 210, 0, 0, 5, 0, 10}},			    /*!< DSP_ATTEN */
	{BREG_SX1301_RESET, {-1, 256, 0, 0, 1, 0, 1}},			    /*!< SX1301_RESET */
	{BREG_SX1301_IQ_INTERFACE_DELAY, {-1, 256, 1, 0, 4, 0, 0}},	 /*!< SX1301_IQ_INTERFACE_DELAY */
	{BREG_SX1301_CLK_32M_EN, {-1, 257, 0, 0, 8, 0, 0}},		    /*!< SX1301_CLK_32M_EN */
	{BREG_SX1301_CLK_32M_INVERT, {-1, 258, 0, 0, 8, 0, 0}},		    /*!< SX1301_CLK_32M_INVERT */
	{BREG_SX1301_CLK_HS_EN, {-1, 259, 0, 0, 8, 0, 0}},		    /*!< SX1301_CLK_HS_EN */
	{BREG_SX1301_CLK_HS_INVERT, {-1, 260, 0, 0, 8, 0, 0}},		    /*!< SX1301_CLK_HS_INVERT */
	{BREG_SX1301_STATUS, {-1, 261, 0, 0, 8, 1, 0}},			    /*!< SX1301_STATUS */
	{BREG_ENABLE_SX1301_PATH, {-1, 262, 0, 0, 8, 0, 0}},		    /*!< ENABLE_SX1301_PATH */
	{BREG_RF_PATH_SELECT, {-1, 263, 0, 0, 8, 0, 0}},		    /*!< RF_PATH_SELECT */
	{BREG_SX1301_A_I_RX_SEL, {-1, 264, 0, 0, 2, 0, 0}},		    /*!< SX1301_A_I_RX_SEL */
	{BREG_SX1301_A_Q_RX_SEL, {-1, 264, 2, 0, 2, 0, 1}},		    /*!< SX1301_A_Q_RX_SEL */
	{BREG_SX1301_B_I_RX_SEL, {-1, 264, 4, 0, 2, 0, 2}},		    /*!< SX1301_B_I_RX_SEL */
	{BREG_SX1301_B_Q_RX_SEL, {-1, 264, 6, 0, 2, 0, 3}},		    /*!< SX1301_B_Q_RX_SEL */
	{BREG_FE_RX_A_EN, {-1, 320, 0, 0, 1, 0, 0}},			    /*!< FE_RX_A_EN */
	{BREG_FE_RX_B_EN, {-1, 320, 1, 0, 1, 0, 0}},			    /*!< FE_RX_B_EN */
	{BREG_FE_TX_A_EN, {-1, 320, 2, 0, 1, 0, 0}},			    /*!< FE_TX_A_EN */
	{BREG_FE_TX_B_EN, {-1, 320, 3, 0, 1, 0, 0}},			    /*!< FE_TX_B_EN */
	{BREG_FE_HOST_CTRL, {-1, 320, 4, 0, 1, 0, 1}},			    /*!< FE_HOST_CTRL */
	{BREG_LNA_A_CTRL, {-1, 321, 0, 0, 1, 1, 0}},			    /*!< LNA_A_CTRL */
	{BREG_PA_A_CTRL, {-1, 321, 1, 0, 1, 1, 0}},			    /*!< PA_A_CTRL */
	{BREG_LNA_B_CTRL, {-1, 321, 2, 0, 1, 1, 0}},			    /*!< LNA_B_CTRL */
	{BREG_PA_B_CTRL, {-1, 321, 3, 0, 1, 1, 0}},			    /*!< PA_B_CTRL */
	{BREG_PA_GAIN, {-1, 321, 4, 0, 2, 1, 0}},			    /*!< PA_GAIN */
	{BREG_CAPTURE_RAM_ADDR, {-1, 384, 0, 0, 8, 0, 0}},		    /*!< CAPTURE_RAM_ADDR */
	{BREG_CAPTURE_RAM_DATA, {-1, 385, 0, 0, 8, 1, 0}},		    /*!< CAPTURE_RAM_DATA */
	{BREG_CAPTURE_SOURCE, {-1, 386, 0, 0, 5, 0, 0}},		    /*!< CAPTURE_SOURCE */
	{BREG_CAPTURE_START, {-1, 386, 5, 0, 1, 0, 0}},			    /*!< CAPTURE_START */
	{BREG_CAPTURE_PERIOD, {-1, 387, 0, 0, 16, 0, 0}},		    /*!< CAPTURE_PERIOD */
	{BREG_PLL_RESET, {-1, 389, 0, 0, 1, 0, 1}},			    /*!< PLL_RESET */
	{BREG_PLL_LOCKED, {-1, 389, 1, 0, 1, 1, 0}},			    /*!< PLL_LOCKED */
	{BREG_DSP_PLL_RESET, {-1, 389, 2, 0, 1, 0, 1}},			    /*!< DSP_PLL_RESET */
	{BREG_DSP_PLL_LOCKED, {-1, 389, 3, 0, 1, 1, 0}},		    /*!< DSP_PLL_LOCKED */
	{BREG_SX1301_TX_EDGE_SELECT, {-1, 448, 0, 0, 2, 0, 0}},		    /*!< SX1301_TX_EDGE_SELECT */
	{BREG_SX1301_TX_SELECT, {-1, 448, 2, 0, 1, 0, 0}},		    /*!< SX1301_TX_SELECT */
	{BREG_SX1301_TX_INVERT_IQ, {-1, 448, 3, 0, 1, 0, 0}},		    /*!< SX1301_TX_INVERT_IQ */
	{BREG_SX1301_TX_CTRL, {-1, 448, 4, 0, 1, 0, 0}},		    /*!< SX1301_TX_CTRL */
	{BREG_SIGGEN_FREQ, {-1, 449, 0, 1, 13, 0, 0}},			    /*!< SIGGEN_FREQ */
	{BREG_TX_NOTCH_EN, {-1, 451, 0, 0, 1, 0, 0}},			    /*!< TX_NOTCH_EN */
	{BREG_TX_NOTCH_TUNE, {-1, 451, 1, 0, 6, 0, 0}},			    /*!< TX_NOTCH_TUNE */
	{BREG_TX_EN, {-1, 452, 0, 0, 1, 0, 0}},				    /*!< TX_EN */
	{BREG_TX_SIGGEN_EN, {-1, 452, 1, 0, 1, 0, 0}},			    /*!< TX_SIGGEN_EN */
	{BREG_TX_RF_PATH_SELECT, {-1, 452, 2, 0, 1, 0, 0}},		    /*!< TX_RF_PATH_SELECT */
	{BREG_TX_INVERT_CLK, {-1, 452, 3, 0, 1, 0, 0}},			    /*!< TX_INVERT_CLK */
	{BREG_TX_GAIN, {-1, 452, 4, 0, 4, 0, 3}},			    /*!< TX_GAIN */
	{BREG_TX_LUT_AUXDAC_8MSB, {-1, 453, 0, 0, 8, 0, 0}},		    /*!< TX_LUT_AUXDAC_8MSB */
	{BREG_TX_LUT_AUXDAC_2LSB, {-1, 454, 0, 0, 2, 0, 0}},		    /*!< TX_LUT_AUXDAC_2LSB */
	{BREG_TX_LUT_AUXDAC_VREF, {-1, 454, 2, 0, 2, 0, 0}},		    /*!< TX_LUT_AUXDAC_VREF */
	{BREG_TX_LUT_AD_ATTEN_MSB, {-1, 454, 4, 0, 1, 0, 0}},		    /*!< TX_LUT_AD_ATTEN_MSB */
	{BREG_TX_LUT_SPARE, {-1, 454, 5, 0, 3, 0, 0}},			    /*!< TX_LUT_SPARE */
	{BREG_TX_LUT_AD_ATTEN_8LSB, {-1, 455, 0, 0, 8, 0, 0}},		    /*!< TX_LUT_AD_ATTEN_8LSB */
	{BREG_TX_LUT_FREQ_INTEGER_0, {-1, 456, 0, 0, 8, 0, 0}},		    /*!< TX_LUT_FREQ_INTEGER_0 */
	{BREG_IF_FREQ_RSSI, {-1, 512, 0, 1, 13, 0, 384}},		    /*!< IF_FREQ_RSSI */
	{BREG_RSSI_FILTER_1_GAIN, {-1, 514, 0, 0, 5, 0, 0}},		    /*!< RSSI_FILTER_1_GAIN */
	{BREG_RSSI_FILTER_1_DC_NOTCH_EN, {-1, 514, 5, 0, 1, 0, 0}},	 /*!< RSSI_FILTER_1_DC_NOTCH_EN */
	{BREG_RSSI, {-1, 515, 0, 0, 8, 1, 0}},				    /*!< RSSI */
	{BREG_RSSI_HISTO_RATE_DIV, {-1, 516, 0, 0, 16, 0, 1000}},	   /*!< RSSI_HISTO_RATE_DIV */
	{BREG_RSSI_HISTO_LEN, {-1, 518, 0, 0, 16, 0, 65535}},		    /*!< RSSI_HISTO_LEN */
	{BREG_RSSI_HISTO_START, {-1, 520, 0, 0, 1, 0, 0}},		    /*!< RSSI_HISTO_START */
	{BREG_RSSI_HISTO_VALID, {-1, 520, 1, 0, 1, 1, 0}},		    /*!< RSSI_HISTO_VALID */
	{BREG_RSSI_HISTO_SELECT_RF_PATH, {-1, 520, 2, 0, 1, 0, 0}},	 /*!< RSSI_HISTO_SELECT_RF_PATH */
	{BREG_RSSI_STATE, {-1, 520, 3, 0, 3, 1, 0}},			    /*!< RSSI_STATE */
	{BREG_RSSI_HISTO_RAM_ADDR, {-1, 521, 0, 0, 8, 0, 0}},		    /*!< RSSI_HISTO_RAM_ADDR */
	{BREG_RSSI_HISTO_RAM_DATA, {-1, 522, 0, 0, 8, 1, 0}},		    /*!< RSSI_HISTO_RAM_DATA */
	{BREG_RSSI_FILTER_2_GAIN, {-1, 523, 0, 0, 5, 0, 0}},		    /*!< RSSI_FILTER_2_GAIN */
	{BREG_DBG_CHAN_SELECT, {-1, 576, 0, 0, 4, 0, 0}},		    /*!< DBG_CHAN_SELECT */
	{BREG_DBG_TEST_UPP, {-1, 576, 4, 0, 1, 0, 0}},			    /*!< DBG_TEST_UPP */
	{BREG_DBG_TEST_DSP_SPI, {-1, 576, 5, 0, 1, 0, 0}},		    /*!< DBG_TEST_DSP_SPI */
	{BREG_FPGA_GPO_SELECT, {-1, 577, 0, 0, 8, 0, 0}},		    /*!< FPGA_GPO_SELECT */
	{BREG_FPGA_GPO_VAL, {-1, 578, 0, 0, 8, 0, 0}},			    /*!< FPGA_GPO_VAL */
	{BREG_DBG_DSP_CHAN_SELECT, {-1, 579, 0, 0, 8, 0, 0}},		    /*!< DBG_DSP_CHAN_SELECT */
	{BREG_DSP_STATUS_PERIOD, {-1, 580, 0, 0, 8, 0, 1}},		    /*!< DSP_STATUS_PERIOD */
	{BREG_DSP_CFG_DIVERSITY_PARAM, {-1, 581, 0, 0, 1, 0, 1}},	   /*!< DSP_CFG_DIVERSITY_PARAM */
	{BREG_DSP_CFG_SYNC_WORD, {-1, 581, 1, 0, 1, 0, 1}},		    /*!< DSP_CFG_SYNC_WORD */
	{BREG_DSP_CFG_SPARE, {-1, 581, 2, 0, 6, 0, 0}},			    /*!< DSP_CFG_SPARE */
	{BREG_DSP_1_FIFO_WR_DATA, {-1, 640, 0, 0, 8, 0, 0}},		    /*!< DSP_1_FIFO_WR_DATA */
	{BREG_DSP_1_FIFO_RD_DATA, {-1, 641, 0, 0, 8, 1, 0}},		    /*!< DSP_1_FIFO_RD_DATA */
	{BREG_DSP_2_FIFO_WR_DATA, {-1, 642, 0, 0, 8, 0, 0}},		    /*!< DSP_2_FIFO_WR_DATA */
	{BREG_DSP_2_FIFO_RD_DATA, {-1, 643, 0, 0, 8, 1, 0}},		    /*!< DSP_2_FIFO_RD_DATA */
	{BREG_DSP_1_FIFO_WR_NB_BYTE, {-1, 644, 0, 0, 16, 1, 0}},	    /*!< DSP_1_FIFO_WR_NB_BYTE */
	{BREG_DSP_1_FIFO_RD_NB_BYTE, {-1, 646, 0, 0, 16, 1, 0}},	    /*!< DSP_1_FIFO_RD_NB_BYTE */
	{BREG_DSP_2_FIFO_WR_NB_BYTE, {-1, 648, 0, 0, 16, 1, 0}},	    /*!< DSP_2_FIFO_WR_NB_BYTE */
	{BREG_DSP_2_FIFO_RD_NB_BYTE, {-1, 650, 0, 0, 16, 1, 0}},	    /*!< DSP_2_FIFO_RD_NB_BYTE */
	{BREG_UPP_LINE_SIZE, {-1, 652, 0, 0, 16, 0, 512}},		    /*!< UPP_LINE_SIZE */
	{BREG_UPP_CLK_DELAY, {-1, 654, 0, 0, 8, 0, 0}},			    /*!< UPP_CLK_DELAY */
	{BREG_DSP1_RSTN, {-1, 655, 0, 0, 1, 0, 0}},			    /*!< DSP1_RSTN */
	{BREG_DSP1_NMI, {-1, 655, 1, 0, 1, 0, 0}},			    /*!< DSP1_NMI */
	{BREG_DSP1_SPI_REG_EN, {-1, 655, 2, 0, 1, 0, 0}},		    /*!< DSP1_SPI_REG_EN */
	{BREG_DSP2_RSTN, {-1, 655, 3, 0, 1, 0, 0}},			    /*!< DSP2_RSTN */
	{BREG_DSP2_NMI, {-1, 655, 4, 0, 1, 0, 0}},			    /*!< DSP2_NMI */
	{BREG_DSP2_SPI_REG_EN, {-1, 655, 5, 0, 1, 0, 0}},		    /*!< DSP2_SPI_REG_EN */
	{BREG_DSP1_BOOT_EN, {-1, 655, 6, 0, 1, 0, 0}},			    /*!< DSP1_BOOT_EN */
	{BREG_DSP2_BOOT_EN, {-1, 655, 7, 0, 1, 0, 0}},			    /*!< DSP2_BOOT_EN */
	{BREG_DSP1_GPIO_IN, {-1, 656, 0, 0, 2, 1, 0}},			    /*!< DSP1_GPIO_IN */
	{BREG_DSP1_GPIO_OUT, {-1, 656, 2, 0, 2, 0, 3}},			    /*!< DSP1_GPIO_OUT */
	{BREG_DSP2_GPIO_IN, {-1, 656, 4, 0, 2, 1, 0}},			    /*!< DSP2_GPIO_IN */
	{BREG_DSP2_GPIO_OUT, {-1, 656, 6, 0, 2, 0, 3}},			    /*!< DSP2_GPIO_OUT */
	{BREG_SX1301_0_IF_FREQ_0, {-1, 704, 0, 1, 13, 0, -384}},	    /*!< SX1301_0_IF_FREQ_0 */
	{BREG_SX1301_0_IF_FREQ_1, {-1, 706, 0, 1, 13, 0, -128}},	    /*!< SX1301_0_IF_FREQ_1 */
	{BREG_SX1301_0_IF_FREQ_2, {-1, 708, 0, 1, 13, 0, 128}},		    /*!< SX1301_0_IF_FREQ_2 */
	{BREG_SX1301_0_IF_FREQ_3, {-1, 710, 0, 1, 13, 0, 384}},		    /*!< SX1301_0_IF_FREQ_3 */
	{BREG_SX1301_0_IF_FREQ_4, {-1, 712, 0, 1, 13, 0, -384}},	    /*!< SX1301_0_IF_FREQ_4 */
	{BREG_SX1301_0_IF_FREQ_5, {-1, 714, 0, 1, 13, 0, -128}},	    /*!< SX1301_0_IF_FREQ_5 */
	{BREG_SX1301_0_IF_FREQ_6, {-1, 716, 0, 1, 13, 0, 128}},		    /*!< SX1301_0_IF_FREQ_6 */
	{BREG_SX1301_0_IF_FREQ_7, {-1, 718, 0, 1, 13, 0, 384}},		    /*!< SX1301_0_IF_FREQ_7 */
	{BREG_SX1301_1_IF_FREQ_0, {-1, 720, 0, 1, 13, 0, -384}},	    /*!< SX1301_1_IF_FREQ_0 */
	{BREG_SX1301_1_IF_FREQ_1, {-1, 722, 0, 1, 13, 0, -128}},	    /*!< SX1301_1_IF_FREQ_1 */
	{BREG_SX1301_1_IF_FREQ_2, {-1, 724, 0, 1, 13, 0, 128}},		    /*!< SX1301_1_IF_FREQ_2 */
	{BREG_SX1301_1_IF_FREQ_3, {-1, 726, 0, 1, 13, 0, 384}},		    /*!< SX1301_1_IF_FREQ_3 */
	{BREG_SX1301_1_IF_FREQ_4, {-1, 728, 0, 1, 13, 0, -384}},	    /*!< SX1301_1_IF_FREQ_4 */
	{BREG_SX1301_1_IF_FREQ_5, {-1, 730, 0, 1, 13, 0, -128}},	    /*!< SX1301_1_IF_FREQ_5 */
	{BREG_SX1301_1_IF_FREQ_6, {-1, 732, 0, 1, 13, 0, 128}},		    /*!< SX1301_1_IF_FREQ_6 */
	{BREG_SX1301_1_IF_FREQ_7, {-1, 734, 0, 1, 13, 0, 384}},		    /*!< SX1301_1_IF_FREQ_7 */
	{BREG_SX1301_CHANN_GAIN, {-1, 736, 0, 0, 5, 0, 29}},		    /*!< SX1301_CHANN_GAIN */
	{BREG_SX1301_CLIP_12B, {-1, 736, 5, 0, 1, 0, 0}},		    /*!< SX1301_CLIP_12B */
	{BREG_SX1301_0_IQ_INVERT, {-1, 737, 0, 0, 1, 0, 0}},		    /*!< SX1301_0_IQ_INVERT */
	{BREG_SX1301_1_IQ_INVERT, {-1, 737, 1, 0, 1, 0, 0}},		    /*!< SX1301_1_IQ_INVERT */
	{BREG_SX1301_0_FREQ_BASE, {-1, 738, 0, 0, 12, 0, 832}},		    /*!< SX1301_0_FREQ_BASE */
	{BREG_SX1301_0_CH0_FREQ_OFFSET, {-1, 740, 0, 1, 8, 0, -15}},	/*!< SX1301_0_CH0_FREQ_OFFSET */
	{BREG_SX1301_0_CH1_FREQ_OFFSET, {-1, 741, 0, 1, 8, 0, -13}},	/*!< SX1301_0_CH1_FREQ_OFFSET */
	{BREG_SX1301_0_CH2_FREQ_OFFSET, {-1, 742, 0, 1, 8, 0, -11}},	/*!< SX1301_0_CH2_FREQ_OFFSET */
	{BREG_SX1301_0_CH3_FREQ_OFFSET, {-1, 743, 0, 1, 8, 0, -9}},	 /*!< SX1301_0_CH3_FREQ_OFFSET */
	{BREG_SX1301_0_CH4_FREQ_OFFSET, {-1, 744, 0, 1, 8, 0, -7}},	 /*!< SX1301_0_CH4_FREQ_OFFSET */
	{BREG_SX1301_0_CH5_FREQ_OFFSET, {-1, 745, 0, 1, 8, 0, -5}},	 /*!< SX1301_0_CH5_FREQ_OFFSET */
	{BREG_SX1301_0_CH6_FREQ_OFFSET, {-1, 746, 0, 1, 8, 0, -3}},	 /*!< SX1301_0_CH6_FREQ_OFFSET */
	{BREG_SX1301_0_CH7_FREQ_OFFSET, {-1, 747, 0, 1, 8, 0, -1}},	 /*!< SX1301_0_CH7_FREQ_OFFSET */
	{BREG_SX1301_1_FREQ_BASE, {-1, 748, 0, 0, 12, 0, 832}},		    /*!< SX1301_1_FREQ_BASE */
	{BREG_SX1301_1_CH0_FREQ_OFFSET, {-1, 750, 0, 1, 8, 0, 1}},	  /*!< SX1301_1_CH0_FREQ_OFFSET */
	{BREG_SX1301_1_CH1_FREQ_OFFSET, {-1, 751, 0, 1, 8, 0, 3}},	  /*!< SX1301_1_CH1_FREQ_OFFSET */
	{BREG_SX1301_1_CH2_FREQ_OFFSET, {-1, 752, 0, 1, 8, 0, 5}},	  /*!< SX1301_1_CH2_FREQ_OFFSET */
	{BREG_SX1301_1_CH3_FREQ_OFFSET, {-1, 753, 0, 1, 8, 0, 7}},	  /*!< SX1301_1_CH3_FREQ_OFFSET */
	{BREG_SX1301_1_CH4_FREQ_OFFSET, {-1, 754, 0, 1, 8, 0, 9}},	  /*!< SX1301_1_CH4_FREQ_OFFSET */
	{BREG_SX1301_1_CH5_FREQ_OFFSET, {-1, 755, 0, 1, 8, 0, 11}},	 /*!< SX1301_1_CH5_FREQ_OFFSET */
	{BREG_SX1301_1_CH6_FREQ_OFFSET, {-1, 756, 0, 1, 8, 0, 13}},	 /*!< SX1301_1_CH6_FREQ_OFFSET */
	{BREG_SX1301_1_CH7_FREQ_OFFSET, {-1, 757, 0, 1, 8, 0, 15}},	 /*!< SX1301_1_CH7_FREQ_OFFSET */
	{BREG_HIGH_SPEED_TIMESTAMP, {-1, 768, 0, 0, 32, 1, 0}},		    /*!< HIGH_SPEED_TIMESTAMP */
	{BREG_HIGH_SPEED_TIMESTAMP_GPS, {-1, 772, 0, 0, 32, 1, 0}},	 /*!< HIGH_SPEED_TIMESTAMP_GPS */
	{BREG_HIGH_SPEED_TIMESTAMP_SAMPLE, {-1, 776, 0, 0, 32, 1, 0}},      /*!< HIGH_SPEED_TIMESTAMP_SAMPLE */
	{BREG_LATCHED_SAMPLE_CNT, {-1, 780, 0, 0, 32, 1, 0}},		    /*!< LATCHED_SAMPLE_CNT */
	{BREG_ALPHA_COEFF, {-1, 784, 0, 0, 32, 0, 262144000}},		    /*!< ALPHA_COEFF */
	{BREG_BETA_COEFF, {-1, 788, 0, 0, 32, 0, 0}},			    /*!< BETA_COEFF */
	{BREG_ALPHA_BETA_COEFF_VALID, {-1, 792, 0, 0, 1, 0, 0}},	    /*!< ALPHA_BETA_COEFF_VALID */
	{BREG_DSP_1_FIFO_RD_DATA_CRYPTO, {-1, 832, 0, 0, 8, 1, 0}},	 /*!< DSP_1_FIFO_RD_DATA_CRYPTO */
	{BREG_DSP_1_FIFO_RD_NB_BYTE_CRYPTO, {-1, 833, 0, 0, 16, 1, 0}},     /*!< DSP_1_FIFO_RD_NB_BYTE_CRYPTO */
	{BREG_DBG_CRYPTO_STATE, {-1, 835, 0, 0, 8, 1, 0}},		    /*!< DBG_CRYPTO_STATE */
	{BREG_DBG_CRYPTO_BYTE_CPT, {-1, 836, 0, 0, 8, 1, 0}},		    /*!< DBG_CRYPTO_BYTE_CPT */
	{BREG_DBG_CRYPTO_BUFFER_0, {-1, 837, 0, 0, 8, 1, 0}},		    /*!< DBG_CRYPTO_BUFFER_0 */
	{BREG_HASH_DATA0, {-1, 896, 0, 0, 32, 1, 0}},			    /*!< HASH_DATA0 */
	{BREG_HASH_DATA1, {-1, 900, 0, 0, 32, 1, 0}},			    /*!< HASH_DATA1 */
	{BREG_HASH_DATA2, {-1, 904, 0, 0, 32, 1, 0}},			    /*!< HASH_DATA2 */
	{BREG_HASH_DATA3, {-1, 908, 0, 0, 32, 1, 0}},			    /*!< HASH_DATA3 */
	{BREG_HASH_NB_BYTE, {-1, 912, 0, 0, 16, 1, 0}},			    /*!< HASH_NB_BYTE */
	{BREG_HASH_NB_WORD, {-1, 914, 0, 0, 16, 1, 0}},			    /*!< HASH_NB_WORD */
	{BREG_DSP_MIC_STATUS, {-1, 916, 0, 0, 8, 1, 0}},		    /*!< DSP_MIC_STATUS */
	{BREG_DSP_FORCE_MIC_ENABLE, {-1, 917, 0, 0, 1, 0, 0}},		    /*!< DSP_FORCE_MIC_ENABLE */
	{BREG_DSP_FORCE_MIC_MODE, {-1, 917, 1, 0, 2, 0, 0}},		    /*!< DSP_FORCE_MIC_MODE */
	{BREG_DSP2_IF_FREQ_HIGH, {-1, 960, 0, 1, 13, 0, 384}},		    /*!< DSP2_IF_FREQ_HIGH */
	{BREG_DSP2_IF_FREQ_LOW, {-1, 962, 0, 1, 13, 0, 384}},		    /*!< DSP2_IF_FREQ_LOW */
	{BREG_DSP2_UPP_LINE_SIZE, {-1, 964, 0, 0, 16, 0, 255}},		    /*!< DSP2_UPP_LINE_SIZE */
	{BREG_DSP2_UPP_CLK_DELAY, {-1, 966, 0, 0, 8, 0, 1}},		    /*!< DSP2_UPP_CLK_DELAY */
	{BREG_DSP2_RF_SELECT, {-1, 967, 0, 0, 1, 0, 0}},		    /*!< DSP2_RF_SELECT */
	{BREG_START_LBT, {-1, 1024, 0, 0, 1, 0, 0}},			    /*!< START_LBT */
	{BREG_LBT_FILTER_1_GAIN, {-1, 1025, 0, 0, 5, 0, 0}},		    /*!< LBT_FILTER_1_GAIN */
	{BREG_LBT_FILTER_1_DC_NOTCH_EN, {-1, 1025, 5, 0, 1, 0, 0}},	 /*!< LBT_FILTER_1_DC_NOTCH_EN */
	{BREG_LBT_FILTER_2_GAIN, {-1, 1026, 0, 0, 5, 0, 0}},		    /*!< LBT_FILTER_2_GAIN */
	{BREG_LBT_SELECT_RF_PATH, {-1, 1026, 5, 0, 1, 0, 0}},		    /*!< LBT_SELECT_RF_PATH */
	{BREG_LBT_CHANNEL_0_IF_FREQ, {-1, 1027, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_0_IF_FREQ */
	{BREG_LBT_CHANNEL_1_IF_FREQ, {-1, 1029, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_1_IF_FREQ */
	{BREG_LBT_CHANNEL_2_IF_FREQ, {-1, 1031, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_2_IF_FREQ */
	{BREG_LBT_CHANNEL_3_IF_FREQ, {-1, 1033, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_3_IF_FREQ */
	{BREG_LBT_CHANNEL_4_IF_FREQ, {-1, 1035, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_4_IF_FREQ */
	{BREG_LBT_CHANNEL_5_IF_FREQ, {-1, 1037, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_5_IF_FREQ */
	{BREG_LBT_CHANNEL_6_IF_FREQ, {-1, 1039, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_6_IF_FREQ */
	{BREG_LBT_CHANNEL_7_IF_FREQ, {-1, 1041, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_7_IF_FREQ */
	{BREG_LBT_CHANNEL_8_IF_FREQ, {-1, 1043, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_8_IF_FREQ */
	{BREG_LBT_CHANNEL_9_IF_FREQ, {-1, 1045, 0, 1, 13, 0, 0}},	   /*!< LBT_CHANNEL_9_IF_FREQ */
	{BREG_LBT_CHANNEL_10_IF_FREQ, {-1, 1047, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_10_IF_FREQ */
	{BREG_LBT_CHANNEL_11_IF_FREQ, {-1, 1049, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_11_IF_FREQ */
	{BREG_LBT_CHANNEL_12_IF_FREQ, {-1, 1051, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_12_IF_FREQ */
	{BREG_LBT_CHANNEL_13_IF_FREQ, {-1, 1053, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_13_IF_FREQ */
	{BREG_LBT_CHANNEL_14_IF_FREQ, {-1, 1055, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_14_IF_FREQ */
	{BREG_LBT_CHANNEL_15_IF_FREQ, {-1, 1057, 0, 1, 13, 0, 0}},	  /*!< LBT_CHANNEL_15_IF_FREQ */
	{BREG_LBT_NB_CHANNEL, {-1, 1059, 0, 0, 8, 0, 7}},		    /*!< LBT_NB_CHANNEL */
	{BREG_LBT_CHANNEL_ENABLE, {-1, 1060, 0, 0, 16, 0, 0}},		    /*!< LBT_CHANNEL_ENABLE */
	{BREG_LBT_NB_READ_RSSI_TIME1, {-1, 1062, 0, 0, 16, 0, 7}},	  /*!< LBT_NB_READ_RSSI_TIME1 */
	{BREG_LBT_NB_READ_RSSI_TIME2, {-1, 1064, 0, 0, 16, 0, 7}},	  /*!< LBT_NB_READ_RSSI_TIME2 */
	{BREG_LBT_CHANNEL_TIME_MODE, {-1, 1066, 0, 0, 16, 0, 0}},	   /*!< LBT_CHANNEL_TIME_MODE */
	{BREG_LBT_RSSI_TARGET, {-1, 1068, 0, 0, 8, 0, 7}},		    /*!< LBT_RSSI_TARGET */
	{BREG_LBT_RSSI_START_DELAY, {-1, 1069, 0, 0, 16, 0, 0}},	    /*!< LBT_RSSI_START_DELAY */
	{BREG_LBT_RSSI_READ_DELAY, {-1, 1071, 0, 0, 16, 0, 0}},		    /*!< LBT_RSSI_READ_DELAY */
	{BREG_LBT_SELECT_CH, {-1, 1073, 0, 0, 8, 0, 0}},		    /*!< LBT_SELECT_CH */
	{BREG_LBT_TIMESTAMP, {-1, 1074, 0, 0, 32, 1, 0}},		    /*!< LBT_TIMESTAMP */
	{BREG_SX1301_TIMESTAMP_CPY, {-1, 1078, 0, 0, 32, 1, 0}},	    /*!< SX1301_TIMESTAMP_CPY */
	{BREG_SX1301_TIMESTAMP_CPY_EN, {-1, 1082, 0, 0, 1, 0, 0}},	  /*!< SX1301_TIMESTAMP_CPY_EN */
	{BREG_SX1301_TIMESTAMP_CPY_GPS, {-1, 1083, 0, 0, 32, 1, 0}},	/*!< SX1301_TIMESTAMP_CPY_GPS */
	{BREG_EXT_IF_EN, {-1, 1152, 0, 0, 1, 0, 0}},			    /*!< EXT_IF_EN */
	{BREG_EXT_IF_DELAY, {-1, 1152, 1, 0, 3, 0, 0}},			    /*!< EXT_IF_DELAY */
	{BREG_, {-1, 0, 0, 0, 0, 0, 0}}};

static const struct brd_reg breg_slave[] =
    {
	{BREG_SOFT_RESET, {-1, 0, 7, 0, 1, 0, 0}},			   /*!< SOFT_RESET */
	{BREG_VERSION, {-1, 1, 0, 0, 8, 1, SX1301AR_BOARD_VERSION_SLAVE}}, /*!< VERSION */
	{BREG_ALT_CHIP_ID_LSB, {-1, 2, 0, 0, 32, 1, 0}},		   /*!< ALT_CHIP_ID_LSB */
	{BREG_ALT_CHIP_ID_MSB, {-1, 6, 0, 0, 32, 1, 0}},		   /*!< ALT_CHIP_ID_MSB */
	{BREG_IF_FREQ_0, {-1, 128, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_0 */
	{BREG_IF_FREQ_1, {-1, 130, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_1 */
	{BREG_IF_FREQ_2, {-1, 132, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_2 */
	{BREG_IF_FREQ_3, {-1, 134, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_3 */
	{BREG_IF_FREQ_4, {-1, 136, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_4 */
	{BREG_IF_FREQ_5, {-1, 138, 0, 1, 13, 0, 0}},			   /*!< IF_FREQ_5 */
	{BREG_RSSI_THRESHOLD_HIGH, {-1, 192, 0, 0, 16, 0, 0}},		   /*!< IF_FREQ_6 */
	{BREG_RSSI_THRESHOLD_LOW, {-1, 194, 0, 0, 16, 0, 0}},		   /*!< IF_FREQ_7 */
	{BREG_RSSI_DEC_FILTER_ALPHA, {-1, 196, 0, 0, 4, 0, 7}},		   /*!< RSSI_DEC_FILTER_ALPHA */
	{BREG_RSSI_MIN_ATTEN, {-1, 197, 0, 0, 5, 0, 7}},		   /*!< RSSI_MIN_ATTEN */
	{BREG_RSSI_MAX_ATTEN, {-1, 198, 0, 0, 5, 0, 11}},		   /*!< RSSI_MAX_ATTEN */
	{BREG_RSSI_ATTEN_0, {-1, 199, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_0 */
	{BREG_RSSI_ATTEN_1, {-1, 200, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_1 */
	{BREG_RSSI_ATTEN_2, {-1, 201, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_2 */
	{BREG_RSSI_ATTEN_3, {-1, 202, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_3 */
	{BREG_RSSI_ATTEN_4, {-1, 203, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_4 */
	{BREG_RSSI_ATTEN_5, {-1, 204, 0, 0, 5, 1, 0}},			   /*!< RSSI_ATTEN_5 */
	{BREG_FIR_GAIN, {-1, 207, 0, 0, 5, 0, 0}},			   /*!< FIR_GAIN */
	{BREG_OVERRIDE_GAIN, {-1, 207, 5, 0, 1, 0, 0}},			   /*!< OVERRIDE_GAIN */
	{BREG_FIR_GAIN_STEP, {-1, 208, 0, 0, 5, 0, 0}},			   /*!< FIR_GAIN_STEP */
	{BREG_DC_NOTCH_EN, {-1, 209, 0, 0, 8, 0, 0}},			   /*!< DC_NOTCH_EN */
	{BREG_DSP_ATTEN, {-1, 210, 0, 0, 5, 0, 10}},			   /*!< DSP_ATTEN */
	{BREG_SX1301_RESET, {-1, 256, 0, 0, 1, 0, 1}},			   /*!< SX1301_RESET */
	{BREG_SX1301_IQ_INTERFACE_DELAY, {-1, 256, 1, 0, 4, 0, 0}},	/*!< SX1301_IQ_INTERFACE_DELAY */
	{BREG_SX1301_CLK_32M_EN, {-1, 257, 0, 0, 8, 0, 0}},		   /*!< SX1301_CLK_32M_EN */
	{BREG_SX1301_CLK_32M_INVERT, {-1, 258, 0, 0, 8, 0, 0}},		   /*!< SX1301_CLK_32M_INVERT */
	{BREG_SX1301_CLK_HS_EN, {-1, 259, 0, 0, 8, 0, 0}},		   /*!< SX1301_CLK_HS_EN */
	{BREG_SX1301_CLK_HS_INVERT, {-1, 260, 0, 0, 8, 0, 0}},		   /*!< SX1301_CLK_HS_INVERT */
	{BREG_SX1301_STATUS, {-1, 261, 0, 0, 8, 1, 0}},			   /*!< SX1301_STATUS */
	{BREG_ENABLE_SX1301_PATH, {-1, 262, 0, 0, 8, 0, 0}},		   /*!< ENABLE_SX1301_PATH */
	{BREG_RF_PATH_SELECT, {-1, 263, 0, 0, 8, 0, 0}},		   /*!< RF_PATH_SELECT */
	{BREG_SX1301_A_I_RX_SEL, {-1, 264, 0, 0, 2, 0, 0}},		   /*!< SX1301_A_I_RX_SEL */
	{BREG_SX1301_A_Q_RX_SEL, {-1, 264, 2, 0, 2, 0, 1}},		   /*!< SX1301_A_Q_RX_SEL */
	{BREG_SX1301_B_I_RX_SEL, {-1, 264, 4, 0, 2, 0, 2}},		   /*!< SX1301_B_I_RX_SEL */
	{BREG_SX1301_B_Q_RX_SEL, {-1, 264, 6, 0, 2, 0, 3}},		   /*!< SX1301_B_Q_RX_SEL */
	{BREG_CAPTURE_RAM_ADDR, {-1, 384, 0, 0, 8, 0, 0}},		   /*!< CAPTURE_RAM_ADDR */
	{BREG_CAPTURE_RAM_DATA, {-1, 385, 0, 0, 8, 1, 0}},		   /*!< CAPTURE_RAM_DATA */
	{BREG_CAPTURE_SOURCE, {-1, 386, 0, 0, 5, 0, 0}},		   /*!< CAPTURE_SOURCE */
	{BREG_CAPTURE_START, {-1, 386, 5, 0, 1, 0, 0}},			   /*!< CAPTURE_START */
	{BREG_CAPTURE_PERIOD, {-1, 387, 0, 0, 16, 0, 0}},		   /*!< CAPTURE_PERIOD */
	{BREG_PLL_RESET, {-1, 389, 0, 0, 1, 0, 1}},			   /*!< PLL_RESET */
	{BREG_PLL_LOCKED, {-1, 389, 1, 0, 1, 1, 0}},			   /*!< PLL_LOCKED */
	{BREG_DSP_PLL_RESET, {-1, 389, 2, 0, 1, 0, 1}},			   /*!< DSP_PLL_RESET */
	{BREG_DSP_PLL_LOCKED, {-1, 389, 3, 0, 1, 1, 0}},		   /*!< DSP_PLL_LOCKED */
	{BREG_DBG_CHAN_SELECT, {-1, 576, 0, 0, 4, 0, 0}},		   /*!< DBG_CHAN_SELECT */
	{BREG_DBG_TEST_UPP, {-1, 576, 4, 0, 1, 0, 0}},			   /*!< DBG_TEST_UPP */
	{BREG_DBG_TEST_DSP_SPI, {-1, 576, 5, 0, 1, 0, 0}},		   /*!< DBG_TEST_DSP_SPI */
	{BREG_FPGA_GPO_SELECT, {-1, 577, 0, 0, 8, 0, 0}},		   /*!< FPGA_GPO_SELECT */
	{BREG_FPGA_GPO_VAL, {-1, 578, 0, 0, 8, 0, 0}},			   /*!< FPGA_GPO_VAL */
	{BREG_DBG_DSP_CHAN_SELECT, {-1, 579, 0, 0, 8, 0, 0}},		   /*!< DBG_DSP_CHAN_SELECT */
	{BREG_DSP_STATUS_PERIOD, {-1, 580, 0, 0, 8, 0, 1}},		   /*!< DSP_STATUS_PERIOD */
	{BREG_DSP_CFG_DIVERSITY_PARAM, {-1, 581, 0, 0, 1, 0, 1}},	  /*!< DSP_CFG_DIVERSITY_PARAM */
	{BREG_DSP_CFG_SYNC_WORD, {-1, 581, 1, 0, 1, 0, 1}},		   /*!< DSP_CFG_SYNC_WORD */
	{BREG_DSP_CFG_SPARE, {-1, 581, 2, 0, 6, 0, 0}},			   /*!< DSP_CFG_SPARE */
	{BREG_DSP_1_FIFO_WR_DATA, {-1, 640, 0, 0, 8, 0, 0}},		   /*!< DSP_1_FIFO_WR_DATA */
	{BREG_DSP_1_FIFO_RD_DATA, {-1, 641, 0, 0, 8, 1, 0}},		   /*!< DSP_1_FIFO_RD_DATA */
	{BREG_DSP_2_FIFO_WR_DATA, {-1, 642, 0, 0, 8, 0, 0}},		   /*!< DSP_2_FIFO_WR_DATA */
	{BREG_DSP_2_FIFO_RD_DATA, {-1, 643, 0, 0, 8, 1, 0}},		   /*!< DSP_2_FIFO_RD_DATA */
	{BREG_DSP_1_FIFO_WR_NB_BYTE, {-1, 644, 0, 0, 16, 1, 0}},	   /*!< DSP_1_FIFO_WR_NB_BYTE */
	{BREG_DSP_1_FIFO_RD_NB_BYTE, {-1, 646, 0, 0, 16, 1, 0}},	   /*!< DSP_1_FIFO_RD_NB_BYTE */
	{BREG_DSP_2_FIFO_WR_NB_BYTE, {-1, 648, 0, 0, 16, 1, 0}},	   /*!< DSP_2_FIFO_WR_NB_BYTE */
	{BREG_DSP_2_FIFO_RD_NB_BYTE, {-1, 650, 0, 0, 16, 1, 0}},	   /*!< DSP_2_FIFO_RD_NB_BYTE */
	{BREG_DSP_3_FIFO_WR_NB_BYTE, {-1, 659, 0, 0, 16, 1, 0}},	   /*!< DSP_3_FIFO_WR_NB_BYTE */
	{BREG_DSP_3_FIFO_RD_NB_BYTE, {-1, 661, 0, 0, 16, 1, 0}},	   /*!< DSP_3_FIFO_RD_NB_BYTE */
	{BREG_UPP_LINE_SIZE, {-1, 652, 0, 0, 16, 0, 255}},		   /*!< UPP_LINE_SIZE */
	{BREG_UPP_CLK_DELAY, {-1, 654, 0, 0, 8, 0, 128}},		   /*!< UPP_CLK_DELAY */
	{BREG_DSP1_RSTN, {-1, 655, 0, 0, 1, 0, 0}},			   /*!< DSP1_RSTN */
	{BREG_DSP1_NMI, {-1, 655, 1, 0, 1, 0, 0}},			   /*!< DSP1_NMI */
	{BREG_DSP1_SPI_REG_EN, {-1, 655, 2, 0, 1, 0, 0}},		   /*!< DSP1_SPI_REG_EN */
	{BREG_DSP2_RSTN, {-1, 655, 3, 0, 1, 0, 0}},			   /*!< DSP2_RSTN */
	{BREG_DSP2_NMI, {-1, 655, 4, 0, 1, 0, 0}},			   /*!< DSP2_NMI */
	{BREG_DSP2_SPI_REG_EN, {-1, 655, 5, 0, 1, 0, 0}},		   /*!< DSP2_SPI_REG_EN */
	{BREG_DSP1_BOOT_EN, {-1, 655, 6, 0, 1, 0, 0}},			   /*!< DSP1_BOOT_EN */
	{BREG_DSP2_BOOT_EN, {-1, 655, 7, 0, 1, 0, 0}},			   /*!< DSP2_BOOT_EN */
	{BREG_DSP1_GPIO_IN, {-1, 656, 0, 0, 2, 1, 0}},			   /*!< DSP1_GPIO_IN */
	{BREG_DSP1_GPIO_OUT, {-1, 656, 2, 0, 2, 0, 3}},			   /*!< DSP1_GPIO_OUT */
	{BREG_DSP2_GPIO_IN, {-1, 656, 4, 0, 2, 1, 0}},			   /*!< DSP2_GPIO_IN */
	{BREG_DSP2_GPIO_OUT, {-1, 656, 6, 0, 2, 0, 3}},			   /*!< DSP2_GPIO_OUT */
	{BREG_DSP_3_FIFO_WR_DATA, {-1, 657, 0, 0, 8, 0, 0}},		   /*!< DSP_3_FIFO_WR_DATA */
	{BREG_DSP_3_FIFO_RD_DATA, {-1, 658, 0, 0, 8, 1, 0}},		   /*!< DSP_3_FIFO_RD_DATA */
	{BREG_DSP3_RSTN, {-1, 663, 0, 0, 1, 0, 0}},			   /*!< DSP3_RSTN */
	{BREG_DSP3_NMI, {-1, 663, 1, 0, 1, 0, 0}},			   /*!< DSP3_NMI */
	{BREG_DSP3_SPI_REG_EN, {-1, 663, 2, 0, 1, 0, 0}},		   /*!< DSP3_SPI_REG_EN */
	{BREG_DSP3_BOOT_EN, {-1, 663, 3, 0, 1, 0, 0}},			   /*!< DSP3_BOOT_EN */
	{BREG_DSP3_GPIO_IN, {-1, 664, 0, 0, 2, 1, 0}},			   /*!< DSP2_GPIO_IN */
	{BREG_DSP3_GPIO_OUT, {-1, 664, 2, 0, 2, 0, 3}},			   /*!< DSP2_GPIO_OUT */
	{BREG_SX1301_0_IF_FREQ_0, {-1, 704, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_0 */
	{BREG_SX1301_0_IF_FREQ_1, {-1, 706, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_1 */
	{BREG_SX1301_0_IF_FREQ_2, {-1, 708, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_2 */
	{BREG_SX1301_0_IF_FREQ_3, {-1, 710, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_3 */
	{BREG_SX1301_0_IF_FREQ_4, {-1, 712, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_4 */
	{BREG_SX1301_0_IF_FREQ_5, {-1, 714, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_5 */
	{BREG_SX1301_0_IF_FREQ_6, {-1, 716, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_6 */
	{BREG_SX1301_0_IF_FREQ_7, {-1, 718, 0, 1, 13, 0, 0}},		   /*!< SX1301_0_IF_FREQ_7 */
	{BREG_SX1301_1_IF_FREQ_0, {-1, 720, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_0 */
	{BREG_SX1301_1_IF_FREQ_1, {-1, 722, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_1 */
	{BREG_SX1301_1_IF_FREQ_2, {-1, 724, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_2 */
	{BREG_SX1301_1_IF_FREQ_3, {-1, 726, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_3 */
	{BREG_SX1301_1_IF_FREQ_4, {-1, 728, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_4 */
	{BREG_SX1301_1_IF_FREQ_5, {-1, 730, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_5 */
	{BREG_SX1301_1_IF_FREQ_6, {-1, 732, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_6 */
	{BREG_SX1301_1_IF_FREQ_7, {-1, 734, 0, 1, 13, 0, 0}},		   /*!< SX1301_1_IF_FREQ_7 */
	{BREG_SX1301_CHANN_GAIN, {-1, 736, 0, 0, 5, 0, 29}},		   /*!< SX1301_CHANN_GAIN */
	{BREG_SX1301_CLIP_12B, {-1, 736, 5, 0, 1, 0, 0}},		   /*!< SX1301_CLIP_12B */
	{BREG_SX1301_0_IQ_INVERT, {-1, 737, 0, 0, 1, 0, 0}},		   /*!< SX1301_0_IQ_INVERT */
	{BREG_SX1301_1_IQ_INVERT, {-1, 737, 1, 0, 1, 0, 0}},		   /*!< SX1301_1_IQ_INVERT */
	{BREG_SX1301_0_FREQ_BASE, {-1, 738, 0, 0, 12, 0, 832}},		   /*!< SX1301_0_FREQ_BASE */
	{BREG_SX1301_0_CH0_FREQ_OFFSET, {-1, 740, 0, 1, 8, 0, -15}},       /*!< SX1301_0_CH0_FREQ_OFFSET */
	{BREG_SX1301_0_CH1_FREQ_OFFSET, {-1, 741, 0, 1, 8, 0, -13}},       /*!< SX1301_0_CH1_FREQ_OFFSET */
	{BREG_SX1301_0_CH2_FREQ_OFFSET, {-1, 742, 0, 1, 8, 0, -11}},       /*!< SX1301_0_CH2_FREQ_OFFSET */
	{BREG_SX1301_0_CH3_FREQ_OFFSET, {-1, 743, 0, 1, 8, 0, -9}},	/*!< SX1301_0_CH3_FREQ_OFFSET */
	{BREG_SX1301_0_CH4_FREQ_OFFSET, {-1, 744, 0, 1, 8, 0, -7}},	/*!< SX1301_0_CH4_FREQ_OFFSET */
	{BREG_SX1301_0_CH5_FREQ_OFFSET, {-1, 745, 0, 1, 8, 0, -5}},	/*!< SX1301_0_CH5_FREQ_OFFSET */
	{BREG_SX1301_0_CH6_FREQ_OFFSET, {-1, 746, 0, 1, 8, 0, -3}},	/*!< SX1301_0_CH6_FREQ_OFFSET */
	{BREG_SX1301_0_CH7_FREQ_OFFSET, {-1, 747, 0, 1, 8, 0, -1}},	/*!< SX1301_0_CH7_FREQ_OFFSET */
	{BREG_SX1301_1_FREQ_BASE, {-1, 748, 0, 0, 12, 0, 832}},		   /*!< SX1301_1_FREQ_BASE */
	{BREG_SX1301_1_CH0_FREQ_OFFSET, {-1, 750, 0, 1, 8, 0, -15}},       /*!< SX1301_1_CH0_FREQ_OFFSET */
	{BREG_SX1301_1_CH1_FREQ_OFFSET, {-1, 751, 0, 1, 8, 0, -13}},       /*!< SX1301_1_CH1_FREQ_OFFSET */
	{BREG_SX1301_1_CH2_FREQ_OFFSET, {-1, 752, 0, 1, 8, 0, -11}},       /*!< SX1301_1_CH2_FREQ_OFFSET */
	{BREG_SX1301_1_CH3_FREQ_OFFSET, {-1, 753, 0, 1, 8, 0, -9}},	/*!< SX1301_1_CH3_FREQ_OFFSET */
	{BREG_SX1301_1_CH4_FREQ_OFFSET, {-1, 754, 0, 1, 8, 0, -7}},	/*!< SX1301_1_CH4_FREQ_OFFSET */
	{BREG_SX1301_1_CH5_FREQ_OFFSET, {-1, 755, 0, 1, 8, 0, -5}},	/*!< SX1301_1_CH5_FREQ_OFFSET */
	{BREG_SX1301_1_CH6_FREQ_OFFSET, {-1, 756, 0, 1, 8, 0, -3}},	/*!< SX1301_1_CH6_FREQ_OFFSET */
	{BREG_SX1301_1_CH7_FREQ_OFFSET, {-1, 757, 0, 1, 8, 0, -1}},	/*!< SX1301_1_CH7_FREQ_OFFSET */
	{BREG_HIGH_SPEED_TIMESTAMP, {-1, 768, 0, 0, 32, 1, 0}},		   /*!< HIGH_SPEED_TIMESTAMP */
	{BREG_HIGH_SPEED_TIMESTAMP_GPS, {-1, 772, 0, 0, 32, 1, 0}},	/*!< HIGH_SPEED_TIMESTAMP_GPS */
	{BREG_HIGH_SPEED_TIMESTAMP_SAMPLE, {-1, 776, 0, 0, 32, 1, 0}},     /*!< HIGH_SPEED_TIMESTAMP_SAMPLE */
	{BREG_LATCHED_SAMPLE_CNT, {-1, 780, 0, 0, 32, 1, 0}},		   /*!< LATCHED_SAMPLE_CNT */
	{BREG_ALPHA_COEFF, {-1, 784, 0, 0, 32, 0, 262144000}},		   /*!< ALPHA_COEFF */
	{BREG_BETA_COEFF, {-1, 788, 0, 0, 32, 0, 0}},			   /*!< BETA_COEFF */
	{BREG_ALPHA_BETA_COEFF_VALID, {-1, 792, 0, 0, 1, 0, 0}},	   /*!< ALPHA_BETA_COEFF_VALID */
	{BREG_DSP_1_FIFO_RD_DATA_CRYPTO, {-1, 832, 0, 0, 8, 1, 0}},	/*!< DSP_1_FIFO_RD_DATA_CRYPTO */
	{BREG_DSP_1_FIFO_RD_NB_BYTE_CRYPTO, {-1, 833, 0, 0, 16, 1, 0}},    /*!< DSP_1_FIFO_RD_NB_BYTE_CRYPTO */
	{BREG_DBG_CRYPTO_STATE, {-1, 835, 0, 0, 8, 1, 0}},		   /*!< DBG_CRYPTO_STATE */
	{BREG_DBG_CRYPTO_BYTE_CPT, {-1, 836, 0, 0, 8, 1, 0}},		   /*!< DBG_CRYPTO_BYTE_CPT */
	{BREG_DBG_CRYPTO_BUFFER_0, {-1, 837, 0, 0, 8, 1, 0}},		   /*!< DBG_CRYPTO_BUFFER_0 */
	{BREG_DSP_2_FIFO_RD_DATA_CRYPTO, {-1, 838, 0, 0, 8, 1, 0}},	/*!< DSP_2_FIFO_RD_DATA_CRYPTO */
	{BREG_DSP_2_FIFO_RD_NB_BYTE_CRYPTO, {-1, 839, 0, 0, 16, 1, 0}},    /*!< DSP_2_FIFO_RD_NB_BYTE_CRYPTO */
	{BREG_DSP_3_FIFO_RD_DATA_CRYPTO, {-1, 841, 0, 0, 8, 1, 0}},	/*!< DSP_3_FIFO_RD_DATA_CRYPTO */
	{BREG_DSP_3_FIFO_RD_NB_BYTE_CRYPTO, {-1, 842, 0, 0, 16, 1, 0}},    /*!< DSP_3_FIFO_RD_NB_BYTE_CRYPTO */
	{BREG_HASH_DATA0, {-1, 896, 0, 0, 32, 1, 0}},			   /*!< HASH_DATA0 */
	{BREG_HASH_DATA1, {-1, 900, 0, 0, 32, 1, 0}},			   /*!< HASH_DATA1 */
	{BREG_HASH_DATA2, {-1, 904, 0, 0, 32, 1, 0}},			   /*!< HASH_DATA2 */
	{BREG_HASH_DATA3, {-1, 908, 0, 0, 32, 1, 0}},			   /*!< HASH_DATA3 */
	{BREG_HASH_NB_BYTE, {-1, 912, 0, 0, 16, 1, 0}},			   /*!< HASH_NB_BYTE */
	{BREG_HASH_NB_WORD, {-1, 914, 0, 0, 16, 1, 0}},			   /*!< HASH_NB_WORD */
	{BREG_DSP_MIC_STATUS, {-1, 916, 0, 0, 8, 1, 0}},		   /*!< DSP_MIC_STATUS */
	{BREG_DSP_FORCE_MIC_ENABLE, {-1, 917, 0, 0, 1, 0, 0}},		   /*!< DSP_FORCE_MIC_ENABLE */
	{BREG_DSP_FORCE_MIC_MODE, {-1, 917, 1, 0, 2, 0, 0}},		   /*!< DSP_FORCE_MIC_MODE */
	{BREG_DSP3_MIC_STATUS, {-1, 918, 0, 0, 8, 1, 0}},		   /*!< DSP3_MIC_STATUS */
	{BREG_EXT_IF_EN, {-1, 1152, 0, 0, 1, 0, 0}},			   /*!< EXT_IF_EN */
	{BREG_EXT_IF_DELAY, {-1, 1152, 1, 0, 3, 0, 0}},			   /*!< EXT_IF_DELAY */
	{BREG_SX1301_2_IF_FREQ_0, {-1, 1216, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_0 */
	{BREG_SX1301_2_IF_FREQ_1, {-1, 1218, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_1 */
	{BREG_SX1301_2_IF_FREQ_2, {-1, 1220, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_2 */
	{BREG_SX1301_2_IF_FREQ_3, {-1, 1222, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_3 */
	{BREG_SX1301_2_IF_FREQ_4, {-1, 1224, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_4 */
	{BREG_SX1301_2_IF_FREQ_5, {-1, 1226, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_5 */
	{BREG_SX1301_2_IF_FREQ_6, {-1, 1228, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_6 */
	{BREG_SX1301_2_IF_FREQ_7, {-1, 1230, 0, 1, 13, 0, 0}},		   /*!< SX1301_2_IF_FREQ_7 */
	{BREG_SX1301_3_IF_FREQ_0, {-1, 1232, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_0 */
	{BREG_SX1301_3_IF_FREQ_1, {-1, 1234, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_1 */
	{BREG_SX1301_3_IF_FREQ_2, {-1, 1236, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_2 */
	{BREG_SX1301_3_IF_FREQ_3, {-1, 1238, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_3 */
	{BREG_SX1301_3_IF_FREQ_4, {-1, 1240, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_4 */
	{BREG_SX1301_3_IF_FREQ_5, {-1, 1242, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_5 */
	{BREG_SX1301_3_IF_FREQ_6, {-1, 1244, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_6 */
	{BREG_SX1301_3_IF_FREQ_7, {-1, 1246, 0, 1, 13, 0, 0}},		   /*!< SX1301_3_IF_FREQ_7 */
	{BREG_SX1301_4_IF_FREQ_0, {-1, 1248, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_0 */
	{BREG_SX1301_4_IF_FREQ_1, {-1, 1250, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_1 */
	{BREG_SX1301_4_IF_FREQ_2, {-1, 1252, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_2 */
	{BREG_SX1301_4_IF_FREQ_3, {-1, 1254, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_3 */
	{BREG_SX1301_4_IF_FREQ_4, {-1, 1256, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_4 */
	{BREG_SX1301_4_IF_FREQ_5, {-1, 1258, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_5 */
	{BREG_SX1301_4_IF_FREQ_6, {-1, 1260, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_6 */
	{BREG_SX1301_4_IF_FREQ_7, {-1, 1262, 0, 1, 13, 0, 0}},		   /*!< SX1301_4_IF_FREQ_7 */
	{BREG_SX1301_5_IF_FREQ_0, {-1, 1264, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_0 */
	{BREG_SX1301_5_IF_FREQ_1, {-1, 1266, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_1 */
	{BREG_SX1301_5_IF_FREQ_2, {-1, 1268, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_2 */
	{BREG_SX1301_5_IF_FREQ_3, {-1, 1270, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_3 */
	{BREG_SX1301_5_IF_FREQ_4, {-1, 1272, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_4 */
	{BREG_SX1301_5_IF_FREQ_5, {-1, 1274, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_5 */
	{BREG_SX1301_5_IF_FREQ_6, {-1, 1276, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_6 */
	{BREG_SX1301_5_IF_FREQ_7, {-1, 1278, 0, 1, 13, 0, 0}},		   /*!< SX1301_5_IF_FREQ_7 */
	{BREG_SX1301_2_FREQ_BASE, {-1, 1280, 0, 0, 12, 0, 832}},	   /*!< SX1301_2_FREQ_BASE */
	{BREG_SX1301_2_CH0_FREQ_OFFSET, {-1, 1282, 0, 1, 8, 0, -15}},      /*!< SX1301_2_CH0_FREQ_OFFSET */
	{BREG_SX1301_2_CH1_FREQ_OFFSET, {-1, 1283, 0, 1, 8, 0, -13}},      /*!< SX1301_2_CH1_FREQ_OFFSET */
	{BREG_SX1301_2_CH2_FREQ_OFFSET, {-1, 1284, 0, 1, 8, 0, -11}},      /*!< SX1301_2_CH2_FREQ_OFFSET */
	{BREG_SX1301_2_CH3_FREQ_OFFSET, {-1, 1285, 0, 1, 8, 0, -9}},       /*!< SX1301_2_CH3_FREQ_OFFSET */
	{BREG_SX1301_2_CH4_FREQ_OFFSET, {-1, 1286, 0, 1, 8, 0, -7}},       /*!< SX1301_2_CH4_FREQ_OFFSET */
	{BREG_SX1301_2_CH5_FREQ_OFFSET, {-1, 1287, 0, 1, 8, 0, -5}},       /*!< SX1301_2_CH5_FREQ_OFFSET */
	{BREG_SX1301_2_CH6_FREQ_OFFSET, {-1, 1288, 0, 1, 8, 0, -3}},       /*!< SX1301_2_CH6_FREQ_OFFSET */
	{BREG_SX1301_2_CH7_FREQ_OFFSET, {-1, 1289, 0, 1, 8, 0, -1}},       /*!< SX1301_2_CH7_FREQ_OFFSET */
	{BREG_SX1301_3_FREQ_BASE, {-1, 1290, 0, 0, 12, 0, 832}},	   /*!< SX1301_3_FREQ_BASE */
	{BREG_SX1301_3_CH0_FREQ_OFFSET, {-1, 1292, 0, 1, 8, 0, -15}},      /*!< SX1301_3_CH0_FREQ_OFFSET */
	{BREG_SX1301_3_CH1_FREQ_OFFSET, {-1, 1293, 0, 1, 8, 0, -13}},      /*!< SX1301_3_CH1_FREQ_OFFSET */
	{BREG_SX1301_3_CH2_FREQ_OFFSET, {-1, 1294, 0, 1, 8, 0, -11}},      /*!< SX1301_3_CH2_FREQ_OFFSET */
	{BREG_SX1301_3_CH3_FREQ_OFFSET, {-1, 1295, 0, 1, 8, 0, -9}},       /*!< SX1301_3_CH3_FREQ_OFFSET */
	{BREG_SX1301_3_CH4_FREQ_OFFSET, {-1, 1296, 0, 1, 8, 0, -7}},       /*!< SX1301_3_CH4_FREQ_OFFSET */
	{BREG_SX1301_3_CH5_FREQ_OFFSET, {-1, 1297, 0, 1, 8, 0, -5}},       /*!< SX1301_3_CH5_FREQ_OFFSET */
	{BREG_SX1301_3_CH6_FREQ_OFFSET, {-1, 1298, 0, 1, 8, 0, -3}},       /*!< SX1301_3_CH6_FREQ_OFFSET */
	{BREG_SX1301_3_CH7_FREQ_OFFSET, {-1, 1299, 0, 1, 8, 0, -1}},       /*!< SX1301_3_CH7_FREQ_OFFSET */
	{BREG_SX1301_4_FREQ_BASE, {-1, 1300, 0, 0, 12, 0, 832}},	   /*!< SX1301_4_FREQ_BASE */
	{BREG_SX1301_4_CH0_FREQ_OFFSET, {-1, 1302, 0, 1, 8, 0, -15}},      /*!< SX1301_4_CH0_FREQ_OFFSET */
	{BREG_SX1301_4_CH1_FREQ_OFFSET, {-1, 1303, 0, 1, 8, 0, -13}},      /*!< SX1301_4_CH1_FREQ_OFFSET */
	{BREG_SX1301_4_CH2_FREQ_OFFSET, {-1, 1304, 0, 1, 8, 0, -11}},      /*!< SX1301_4_CH2_FREQ_OFFSET */
	{BREG_SX1301_4_CH3_FREQ_OFFSET, {-1, 1305, 0, 1, 8, 0, -9}},       /*!< SX1301_4_CH3_FREQ_OFFSET */
	{BREG_SX1301_4_CH4_FREQ_OFFSET, {-1, 1306, 0, 1, 8, 0, -7}},       /*!< SX1301_4_CH4_FREQ_OFFSET */
	{BREG_SX1301_4_CH5_FREQ_OFFSET, {-1, 1307, 0, 1, 8, 0, -5}},       /*!< SX1301_4_CH5_FREQ_OFFSET */
	{BREG_SX1301_4_CH6_FREQ_OFFSET, {-1, 1308, 0, 1, 8, 0, -3}},       /*!< SX1301_4_CH6_FREQ_OFFSET */
	{BREG_SX1301_4_CH7_FREQ_OFFSET, {-1, 1309, 0, 1, 8, 0, -1}},       /*!< SX1301_4_CH7_FREQ_OFFSET */
	{BREG_SX1301_5_FREQ_BASE, {-1, 1310, 0, 0, 12, 0, 832}},	   /*!< SX1301_5_FREQ_BASE */
	{BREG_SX1301_5_CH0_FREQ_OFFSET, {-1, 1312, 0, 1, 8, 0, -15}},      /*!< SX1301_5_CH0_FREQ_OFFSET */
	{BREG_SX1301_5_CH1_FREQ_OFFSET, {-1, 1313, 0, 1, 8, 0, -13}},      /*!< SX1301_5_CH1_FREQ_OFFSET */
	{BREG_SX1301_5_CH2_FREQ_OFFSET, {-1, 1314, 0, 1, 8, 0, -11}},      /*!< SX1301_5_CH2_FREQ_OFFSET */
	{BREG_SX1301_5_CH3_FREQ_OFFSET, {-1, 1315, 0, 1, 8, 0, -9}},       /*!< SX1301_5_CH3_FREQ_OFFSET */
	{BREG_SX1301_5_CH4_FREQ_OFFSET, {-1, 1316, 0, 1, 8, 0, -7}},       /*!< SX1301_5_CH4_FREQ_OFFSET */
	{BREG_SX1301_5_CH5_FREQ_OFFSET, {-1, 1317, 0, 1, 8, 0, -5}},       /*!< SX1301_5_CH5_FREQ_OFFSET */
	{BREG_SX1301_5_CH6_FREQ_OFFSET, {-1, 1318, 0, 1, 8, 0, -3}},       /*!< SX1301_5_CH6_FREQ_OFFSET */
	{BREG_SX1301_5_CH7_FREQ_OFFSET, {-1, 1319, 0, 1, 8, 0, -1}},       /*!< SX1301_5_CH7_FREQ_OFFSET */
	{BREG_, {-1, 0, 0, 0, 0, 0, 0}}};

/**
@brief Registers of the SX1301
*/
static const struct reg creg[SX1301AR_CREG_NB] =
    {
	{-1, 0, 0, 0, 2, 0, 0},     /*!< PAGE_REG */
	{-1, 0, 7, 0, 1, 0, 0},     /*!< SOFT_RESET */
	{-1, 1, 0, 0, 8, 1, 103},   /*!< VERSION */
	{-1, 2, 0, 0, 16, 0, 0},    /*!< RX_DATA_BUF_ADDR */
	{-1, 4, 0, 0, 8, 0, 0},     /*!< RX_DATA_BUF_DATA */
	{-1, 5, 0, 0, 8, 0, 0},     /*!< TX_DATA_BUF_ADDR */
	{-1, 6, 0, 0, 8, 0, 0},     /*!< TX_DATA_BUF_DATA */
	{-1, 7, 0, 0, 8, 0, 0},     /*!< CAPTURE_RAM_ADDR */
	{-1, 8, 0, 0, 8, 1, 0},     /*!< CAPTURE_RAM_DATA */
	{-1, 9, 0, 0, 8, 0, 0},     /*!< MCU_PROM_ADDR */
	{-1, 10, 0, 0, 8, 0, 0},    /*!< MCU_PROM_DATA */
	{-1, 11, 0, 0, 8, 0, 0},    /*!< RX_PACKET_DATA_FIFO_NUM_STORED */
	{-1, 12, 0, 0, 16, 1, 0},   /*!< RX_PACKET_DATA_FIFO_ADDR_POINTER */
	{-1, 14, 0, 0, 8, 1, 0},    /*!< RX_PACKET_DATA_FIFO_STATUS */
	{-1, 15, 0, 0, 8, 1, 0},    /*!< RX_PACKET_DATA_FIFO_PAYLOAD_SIZE */
	{-1, 16, 0, 0, 1, 0, 0},    /*!< MBWSSF_MODEM_ENABLE */
	{-1, 16, 1, 0, 1, 0, 0},    /*!< CONCENTRATOR_MODEM_ENABLE */
	{-1, 16, 2, 0, 1, 0, 0},    /*!< FSK_MODEM_ENABLE */
	{-1, 16, 3, 0, 1, 0, 0},    /*!< GLOBAL_EN */
	{-1, 17, 0, 0, 1, 0, 1},    /*!< CLK32M_EN */
	{-1, 17, 1, 0, 1, 0, 1},    /*!< CLKHS_EN */
	{-1, 18, 0, 0, 1, 0, 0},    /*!< START_BIST0 */
	{-1, 18, 1, 0, 1, 0, 0},    /*!< START_BIST1 */
	{-1, 18, 2, 0, 1, 0, 0},    /*!< CLEAR_BIST0 */
	{-1, 18, 3, 0, 1, 0, 0},    /*!< CLEAR_BIST1 */
	{-1, 19, 0, 0, 1, 1, 0},    /*!< BIST0_FINISHED */
	{-1, 19, 1, 0, 1, 1, 0},    /*!< BIST1_FINISHED */
	{-1, 20, 0, 0, 1, 1, 0},    /*!< MCU_AGC_PROG_RAM_BIST_STATUS */
	{-1, 20, 1, 0, 1, 1, 0},    /*!< MCU_ARB_PROG_RAM_BIST_STATUS */
	{-1, 20, 2, 0, 1, 1, 0},    /*!< CAPTURE_RAM_BIST_STATUS */
	{-1, 20, 3, 0, 1, 1, 0},    /*!< CHAN_FIR_RAM0_BIST_STATUS */
	{-1, 20, 4, 0, 1, 1, 0},    /*!< CHAN_FIR_RAM1_BIST_STATUS */
	{-1, 21, 0, 0, 1, 1, 0},    /*!< CORR0_RAM_BIST_STATUS */
	{-1, 21, 1, 0, 1, 1, 0},    /*!< CORR1_RAM_BIST_STATUS */
	{-1, 21, 2, 0, 1, 1, 0},    /*!< CORR2_RAM_BIST_STATUS */
	{-1, 21, 3, 0, 1, 1, 0},    /*!< CORR3_RAM_BIST_STATUS */
	{-1, 21, 4, 0, 1, 1, 0},    /*!< CORR4_RAM_BIST_STATUS */
	{-1, 21, 5, 0, 1, 1, 0},    /*!< CORR5_RAM_BIST_STATUS */
	{-1, 21, 6, 0, 1, 1, 0},    /*!< CORR6_RAM_BIST_STATUS */
	{-1, 21, 7, 0, 1, 1, 0},    /*!< CORR7_RAM_BIST_STATUS */
	{-1, 22, 0, 0, 1, 1, 0},    /*!< MODEM0_RAM0_BIST_STATUS */
	{-1, 22, 1, 0, 1, 1, 0},    /*!< MODEM1_RAM0_BIST_STATUS */
	{-1, 22, 2, 0, 1, 1, 0},    /*!< MODEM2_RAM0_BIST_STATUS */
	{-1, 22, 3, 0, 1, 1, 0},    /*!< MODEM3_RAM0_BIST_STATUS */
	{-1, 22, 4, 0, 1, 1, 0},    /*!< MODEM4_RAM0_BIST_STATUS */
	{-1, 22, 5, 0, 1, 1, 0},    /*!< MODEM5_RAM0_BIST_STATUS */
	{-1, 22, 6, 0, 1, 1, 0},    /*!< MODEM6_RAM0_BIST_STATUS */
	{-1, 22, 7, 0, 1, 1, 0},    /*!< MODEM7_RAM0_BIST_STATUS */
	{-1, 23, 0, 0, 1, 1, 0},    /*!< MODEM0_RAM1_BIST_STATUS */
	{-1, 23, 1, 0, 1, 1, 0},    /*!< MODEM1_RAM1_BIST_STATUS */
	{-1, 23, 2, 0, 1, 1, 0},    /*!< MODEM2_RAM1_BIST_STATUS */
	{-1, 23, 3, 0, 1, 1, 0},    /*!< MODEM3_RAM1_BIST_STATUS */
	{-1, 23, 4, 0, 1, 1, 0},    /*!< MODEM4_RAM1_BIST_STATUS */
	{-1, 23, 5, 0, 1, 1, 0},    /*!< MODEM5_RAM1_BIST_STATUS */
	{-1, 23, 6, 0, 1, 1, 0},    /*!< MODEM6_RAM1_BIST_STATUS */
	{-1, 23, 7, 0, 1, 1, 0},    /*!< MODEM7_RAM1_BIST_STATUS */
	{-1, 24, 0, 0, 1, 1, 0},    /*!< MODEM0_RAM2_BIST_STATUS */
	{-1, 24, 1, 0, 1, 1, 0},    /*!< MODEM1_RAM2_BIST_STATUS */
	{-1, 24, 2, 0, 1, 1, 0},    /*!< MODEM2_RAM2_BIST_STATUS */
	{-1, 24, 3, 0, 1, 1, 0},    /*!< MODEM3_RAM2_BIST_STATUS */
	{-1, 24, 4, 0, 1, 1, 0},    /*!< MODEM4_RAM2_BIST_STATUS */
	{-1, 24, 5, 0, 1, 1, 0},    /*!< MODEM5_RAM2_BIST_STATUS */
	{-1, 24, 6, 0, 1, 1, 0},    /*!< MODEM6_RAM2_BIST_STATUS */
	{-1, 24, 7, 0, 1, 1, 0},    /*!< MODEM7_RAM2_BIST_STATUS */
	{-1, 25, 0, 0, 1, 1, 0},    /*!< MODEM_MBWSSF_RAM0_BIST_STATUS */
	{-1, 25, 1, 0, 1, 1, 0},    /*!< MODEM_MBWSSF_RAM1_BIST_STATUS */
	{-1, 25, 2, 0, 1, 1, 0},    /*!< MODEM_MBWSSF_RAM2_BIST_STATUS */
	{-1, 26, 0, 0, 1, 1, 0},    /*!< MCU_AGC_DATA_RAM_BIST0_STATUS */
	{-1, 26, 1, 0, 1, 1, 0},    /*!< MCU_AGC_DATA_RAM_BIST1_STATUS */
	{-1, 26, 2, 0, 1, 1, 0},    /*!< MCU_ARB_DATA_RAM_BIST0_STATUS */
	{-1, 26, 3, 0, 1, 1, 0},    /*!< MCU_ARB_DATA_RAM_BIST1_STATUS */
	{-1, 26, 4, 0, 1, 1, 0},    /*!< TX_TOP_RAM_BIST0_STATUS */
	{-1, 26, 5, 0, 1, 1, 0},    /*!< TX_TOP_RAM_BIST1_STATUS */
	{-1, 26, 6, 0, 1, 1, 0},    /*!< DATA_MNGT_RAM_BIST0_STATUS */
	{-1, 26, 7, 0, 1, 1, 0},    /*!< DATA_MNGT_RAM_BIST1_STATUS */
	{-1, 27, 0, 0, 4, 0, 0},    /*!< GPIO_SELECT_INPUT */
	{-1, 28, 0, 0, 4, 0, 0},    /*!< GPIO_SELECT_OUTPUT */
	{-1, 29, 0, 0, 5, 0, 0},    /*!< GPIO_MODE */
	{-1, 30, 0, 0, 5, 1, 0},    /*!< GPIO_PIN_REG_IN */
	{-1, 31, 0, 0, 5, 0, 0},    /*!< GPIO_PIN_REG_OUT */
	{-1, 32, 0, 0, 8, 1, 0},    /*!< MCU_AGC_STATUS */
	{-1, 125, 0, 0, 8, 1, 0},   /*!< MCU_ARB_STATUS */
	{-1, 126, 0, 0, 8, 1, 1},   /*!< CHIP_ID */
	{-1, 127, 0, 0, 1, 0, 1},   /*!< EMERGENCY_FORCE_HOST_CTRL */
	{0, 33, 0, 0, 1, 0, 0},     /*!< RX_INVERT_IQ */
	{0, 33, 1, 0, 1, 0, 1},     /*!< MODEM_INVERT_IQ */
	{0, 33, 2, 0, 1, 0, 0},     /*!< MBWSSF_MODEM_INVERT_IQ */
	{0, 33, 3, 0, 1, 0, 0},     /*!< RX_EDGE_SELECT */
	{0, 33, 4, 0, 1, 0, 0},     /*!< MISC_RADIO_EN */
	{0, 33, 5, 0, 1, 0, 0},     /*!< FSK_MODEM_INVERT_IQ */
	{0, 34, 0, 0, 4, 0, 7},     /*!< FILTER_GAIN */
	{0, 35, 0, 0, 8, 0, 240},   /*!< RADIO_SELECT */
	{0, 36, 0, 1, 13, 0, -384}, /*!< IF_FREQ_0 */
	{0, 38, 0, 1, 13, 0, -128}, /*!< IF_FREQ_1 */
	{0, 40, 0, 1, 13, 0, 128},  /*!< IF_FREQ_2 */
	{0, 42, 0, 1, 13, 0, 384},  /*!< IF_FREQ_3 */
	{0, 44, 0, 1, 13, 0, -384}, /*!< IF_FREQ_4 */
	{0, 46, 0, 1, 13, 0, -128}, /*!< IF_FREQ_5 */
	{0, 48, 0, 1, 13, 0, 128},  /*!< IF_FREQ_6 */
	{0, 50, 0, 1, 13, 0, 384},  /*!< IF_FREQ_7 */
	{0, 52, 0, 1, 13, 0, 0},    /*!< IF_FREQ_8 */
	{0, 54, 0, 1, 13, 0, 0},    /*!< IF_FREQ_9 */
	{0, 64, 0, 0, 1, 0, 0},     /*!< CHANN_OVERRIDE_AGC_GAIN */
	{0, 64, 1, 0, 4, 0, 7},     /*!< CHANN_AGC_GAIN */
	{0, 65, 0, 0, 7, 0, 0},     /*!< CORR0_DETECT_EN */
	{0, 66, 0, 0, 7, 0, 0},     /*!< CORR1_DETECT_EN */
	{0, 67, 0, 0, 7, 0, 0},     /*!< CORR2_DETECT_EN */
	{0, 68, 0, 0, 7, 0, 0},     /*!< CORR3_DETECT_EN */
	{0, 69, 0, 0, 7, 0, 0},     /*!< CORR4_DETECT_EN */
	{0, 70, 0, 0, 7, 0, 0},     /*!< CORR5_DETECT_EN */
	{0, 71, 0, 0, 7, 0, 0},     /*!< CORR6_DETECT_EN */
	{0, 72, 0, 0, 7, 0, 0},     /*!< CORR7_DETECT_EN */
	{0, 73, 0, 0, 1, 0, 0},     /*!< CORR_SAME_PEAKS_OPTION_SF6 */
	{0, 73, 1, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF7 */
	{0, 73, 2, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF8 */
	{0, 73, 3, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF9 */
	{0, 73, 4, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF10 */
	{0, 73, 5, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF11 */
	{0, 73, 6, 0, 1, 0, 1},     /*!< CORR_SAME_PEAKS_OPTION_SF12 */
	{0, 74, 0, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF6 */
	{0, 74, 4, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF7 */
	{0, 75, 0, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF8 */
	{0, 75, 4, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF9 */
	{0, 76, 0, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF10 */
	{0, 76, 4, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF11 */
	{0, 77, 0, 0, 4, 0, 4},     /*!< CORR_SIG_NOISE_RATIO_SF12 */
	{0, 78, 0, 0, 4, 0, 4},     /*!< CORR_NUM_SAME_PEAK */
	{0, 78, 4, 0, 3, 0, 5},     /*!< CORR_MAC_GAIN */
	{0, 81, 0, 0, 12, 0, 0},    /*!< ADJUST_MODEM_START_OFFSET_RDX4 */
	{0, 83, 0, 0, 12, 0, 4092}, /*!< ADJUST_MODEM_START_OFFSET_SF12_RDX4 */
	{0, 85, 0, 0, 8, 0, 7},     /*!< DBG_CORR_SELECT_SF */
	{0, 86, 0, 0, 8, 0, 0},     /*!< DBG_CORR_SELECT_CHANNEL */
	{0, 87, 0, 0, 8, 1, 0},     /*!< DBG_DETECT_CPT */
	{0, 88, 0, 0, 8, 1, 0},     /*!< DBG_SYMB_CPT */
	{0, 89, 0, 0, 1, 0, 1},     /*!< CHIRP_INVERT_RX */
	{0, 89, 1, 0, 1, 0, 1},     /*!< DC_NOTCH_EN */
	{0, 90, 0, 0, 1, 0, 0},     /*!< IMPLICIT_CRC_EN */
	{0, 90, 1, 0, 3, 0, 0},     /*!< IMPLICIT_CODING_RATE */
	{0, 91, 0, 0, 8, 0, 0},     /*!< IMPLICIT_PAYLOAD_LENGHT */
	{0, 92, 0, 0, 8, 0, 29},    /*!< FREQ_TO_TIME_INVERT */
	{0, 93, 0, 0, 6, 0, 9},     /*!< FREQ_TO_TIME_DRIFT */
	{0, 94, 0, 0, 2, 0, 2},     /*!< PAYLOAD_FINE_TIMING_GAIN */
	{0, 94, 2, 0, 2, 0, 1},     /*!< PREAMBLE_FINE_TIMING_GAIN */
	{0, 94, 4, 0, 2, 0, 0},     /*!< TRACKING_INTEGRAL */
	{0, 95, 0, 0, 4, 0, 1},     /*!< FRAME_SYNCH_PEAK1_POS */
	{0, 95, 4, 0, 4, 0, 2},     /*!< FRAME_SYNCH_PEAK2_POS */
	{0, 96, 0, 0, 16, 0, 10},   /*!< PREAMBLE_SYMB1_NB */
	{0, 98, 0, 0, 1, 0, 1},     /*!< FRAME_SYNCH_GAIN */
	{0, 98, 1, 0, 1, 0, 1},     /*!< SYNCH_DETECT_TH */
	{0, 99, 0, 0, 4, 0, 8},     /*!< LLR_SCALE */
	{0, 99, 4, 0, 2, 0, 2},     /*!< SNR_AVG_CST */
	{0, 100, 0, 0, 7, 0, 0},    /*!< PPM_OFFSET */
	{0, 101, 0, 0, 8, 0, 255},  /*!< MAX_PAYLOAD_LEN */
	{0, 102, 0, 0, 1, 0, 1},    /*!< ONLY_CRC_EN */
	{0, 103, 0, 0, 8, 0, 0},    /*!< ZERO_PAD */
	{0, 104, 0, 0, 4, 0, 8},    /*!< DEC_GAIN_OFFSET */
	{0, 104, 4, 0, 4, 0, 7},    /*!< CHAN_GAIN_OFFSET */
	{0, 105, 1, 0, 1, 0, 1},    /*!< FORCE_HOST_RADIO_CTRL */
	{0, 105, 2, 0, 1, 0, 1},    /*!< FORCE_HOST_FE_CTRL */
	{0, 105, 3, 0, 1, 0, 1},    /*!< FORCE_DEC_FILTER_GAIN */
	{0, 106, 0, 0, 1, 0, 1},    /*!< MCU_RST_0 */
	{0, 106, 1, 0, 1, 0, 1},    /*!< MCU_RST_1 */
	{0, 106, 2, 0, 1, 0, 0},    /*!< MCU_SELECT_MUX_0 */
	{0, 106, 3, 0, 1, 0, 0},    /*!< MCU_SELECT_MUX_1 */
	{0, 106, 4, 0, 1, 1, 0},    /*!< MCU_CORRUPTION_DETECTED_0 */
	{0, 106, 5, 0, 1, 1, 0},    /*!< MCU_CORRUPTION_DETECTED_1 */
	{0, 106, 6, 0, 1, 0, 0},    /*!< MCU_SELECT_EDGE_0 */
	{0, 106, 7, 0, 1, 0, 0},    /*!< MCU_SELECT_EDGE_1 */
	{0, 107, 0, 0, 8, 0, 1},    /*!< CHANN_SELECT_RSSI */
	{0, 108, 0, 0, 8, 0, 32},   /*!< RSSI_BB_DEFAULT_VALUE */
	{0, 109, 0, 0, 8, 0, 100},  /*!< RSSI_DEC_DEFAULT_VALUE */
	{0, 110, 0, 0, 8, 0, 100},  /*!< RSSI_CHANN_DEFAULT_VALUE */
	{0, 111, 0, 0, 5, 0, 7},    /*!< RSSI_BB_FILTER_ALPHA */
	{0, 112, 0, 0, 5, 0, 5},    /*!< RSSI_DEC_FILTER_ALPHA */
	{0, 113, 0, 0, 5, 0, 8},    /*!< RSSI_CHANN_FILTER_ALPHA */
	{0, 114, 0, 0, 6, 0, 0},    /*!< IQ_MISMATCH_A_AMP_COEFF */
	{0, 115, 0, 0, 6, 0, 0},    /*!< IQ_MISMATCH_A_PHI_COEFF */
	{0, 116, 0, 0, 6, 0, 0},    /*!< IQ_MISMATCH_B_AMP_COEFF */
	{0, 116, 6, 0, 1, 0, 0},    /*!< IQ_MISMATCH_B_SEL_I */
	{0, 117, 0, 0, 6, 0, 0},    /*!< IQ_MISMATCH_B_PHI_COEFF */
	{1, 33, 0, 0, 1, 0, 0},     /*!< TX_TRIG_IMMEDIATE */
	{1, 33, 1, 0, 1, 0, 0},     /*!< TX_TRIG_DELAYED */
	{1, 33, 2, 0, 1, 0, 0},     /*!< TX_TRIG_GPS */
	{1, 34, 0, 0, 16, 0, 0},    /*!< TX_START_DELAY */
	{1, 36, 0, 0, 4, 0, 1},     /*!< TX_FRAME_SYNCH_PEAK1_POS */
	{1, 36, 4, 0, 4, 0, 2},     /*!< TX_FRAME_SYNCH_PEAK2_POS */
	{1, 37, 0, 0, 3, 0, 0},     /*!< TX_RAMP_DURATION */
	{1, 39, 0, 1, 8, 0, 0},     /*!< TX_OFFSET_I */
	{1, 40, 0, 1, 8, 0, 0},     /*!< TX_OFFSET_Q */
	{1, 41, 0, 0, 1, 0, 0},     /*!< TX_MODE */
	{1, 41, 1, 0, 4, 0, 0},     /*!< TX_ZERO_PAD */
	{1, 41, 5, 0, 1, 0, 0},     /*!< TX_EDGE_SELECT */
	{1, 41, 6, 0, 1, 0, 0},     /*!< TX_EDGE_SELECT_TOP */
	{1, 42, 0, 0, 2, 0, 0},     /*!< TX_GAIN */
	{1, 42, 2, 0, 3, 0, 5},     /*!< TX_CHIRP_LOW_PASS */
	{1, 42, 5, 0, 2, 0, 0},     /*!< TX_FCC_WIDEBAND */
	{1, 42, 7, 0, 1, 0, 1},     /*!< TX_SWAP_IQ */
	{1, 43, 0, 0, 1, 0, 0},     /*!< MBWSSF_IMPLICIT_HEADER */
	{1, 43, 1, 0, 1, 0, 0},     /*!< MBWSSF_IMPLICIT_CRC_EN */
	{1, 43, 2, 0, 3, 0, 0},     /*!< MBWSSF_IMPLICIT_CODING_RATE */
	{1, 44, 0, 0, 8, 0, 0},     /*!< MBWSSF_IMPLICIT_PAYLOAD_LENGHT */
	{1, 45, 0, 0, 1, 0, 1},     /*!< MBWSSF_AGC_FREEZE_ON_DETECT */
	{1, 46, 0, 0, 4, 0, 1},     /*!< MBWSSF_FRAME_SYNCH_PEAK1_POS */
	{1, 46, 4, 0, 4, 0, 2},     /*!< MBWSSF_FRAME_SYNCH_PEAK2_POS */
	{1, 47, 0, 0, 16, 0, 10},   /*!< MBWSSF_PREAMBLE_SYMB1_NB */
	{1, 49, 0, 0, 1, 0, 1},     /*!< MBWSSF_FRAME_SYNCH_GAIN */
	{1, 49, 1, 0, 1, 0, 1},     /*!< MBWSSF_SYNCH_DETECT_TH */
	{1, 50, 0, 0, 8, 0, 10},    /*!< MBWSSF_DETECT_MIN_SINGLE_PEAK */
	{1, 51, 0, 0, 3, 0, 3},     /*!< MBWSSF_DETECT_TRIG_SAME_PEAK_NB */
	{1, 52, 0, 0, 8, 0, 29},    /*!< MBWSSF_FREQ_TO_TIME_INVERT */
	{1, 53, 0, 0, 6, 0, 36},    /*!< MBWSSF_FREQ_TO_TIME_DRIFT */
	{1, 54, 0, 0, 12, 0, 0},    /*!< MBWSSF_PPM_CORRECTION */
	{1, 56, 0, 0, 2, 0, 2},     /*!< MBWSSF_PAYLOAD_FINE_TIMING_GAIN */
	{1, 56, 2, 0, 2, 0, 1},     /*!< MBWSSF_PREAMBLE_FINE_TIMING_GAIN */
	{1, 56, 4, 0, 2, 0, 0},     /*!< MBWSSF_TRACKING_INTEGRAL */
	{1, 57, 0, 0, 8, 0, 0},     /*!< MBWSSF_ZERO_PAD */
	{1, 58, 0, 0, 2, 0, 0},     /*!< MBWSSF_MODEM_BW */
	{1, 58, 2, 0, 1, 0, 0},     /*!< MBWSSF_RADIO_SELECT */
	{1, 58, 3, 0, 1, 0, 1},     /*!< MBWSSF_RX_CHIRP_INVERT */
	{1, 59, 0, 0, 4, 0, 8},     /*!< MBWSSF_LLR_SCALE */
	{1, 59, 4, 0, 2, 0, 3},     /*!< MBWSSF_SNR_AVG_CST */
	{1, 59, 6, 0, 1, 0, 0},     /*!< MBWSSF_PPM_OFFSET */
	{1, 60, 0, 0, 4, 0, 7},     /*!< MBWSSF_RATE_SF */
	{1, 60, 4, 0, 1, 0, 1},     /*!< MBWSSF_ONLY_CRC_EN */
	{1, 61, 0, 0, 8, 0, 255},   /*!< MBWSSF_MAX_PAYLOAD_LEN */
	{1, 62, 0, 0, 8, 1, 128},   /*!< TX_STATUS */
	{1, 63, 0, 0, 3, 0, 0},     /*!< FSK_CH_BW_EXPO */
	{1, 63, 3, 0, 3, 0, 0},     /*!< FSK_RSSI_LENGTH */
	{1, 63, 6, 0, 1, 0, 0},     /*!< FSK_RX_INVERT */
	{1, 63, 7, 0, 1, 0, 0},     /*!< FSK_PKT_MODE */
	{1, 64, 0, 0, 3, 0, 0},     /*!< FSK_PSIZE */
	{1, 64, 3, 0, 1, 0, 0},     /*!< FSK_CRC_EN */
	{1, 64, 4, 0, 2, 0, 0},     /*!< FSK_DCFREE_ENC */
	{1, 64, 6, 0, 1, 0, 0},     /*!< FSK_CRC_IBM */
	{1, 65, 0, 0, 5, 0, 0},     /*!< FSK_ERROR_OSR_TOL */
	{1, 65, 7, 0, 1, 0, 0},     /*!< FSK_RADIO_SELECT */
	{1, 66, 0, 0, 16, 0, 0},    /*!< FSK_BR_RATIO */
	{1, 68, 0, 0, 32, 0, 0},    /*!< FSK_REF_PATTERN_LSB */
	{1, 72, 0, 0, 32, 0, 0},    /*!< FSK_REF_PATTERN_MSB */
	{1, 76, 0, 0, 8, 0, 0},     /*!< FSK_PKT_LENGTH */
	{1, 77, 0, 0, 1, 0, 1},     /*!< FSK_TX_GAUSSIAN_EN */
	{1, 77, 1, 0, 2, 0, 0},     /*!< FSK_TX_GAUSSIAN_SELECT_BT */
	{1, 77, 3, 0, 1, 0, 1},     /*!< FSK_TX_PATTERN_EN */
	{1, 77, 4, 0, 1, 0, 0},     /*!< FSK_TX_PREAMBLE_SEQ */
	{1, 77, 5, 0, 3, 0, 0},     /*!< FSK_TX_PSIZE */
	{1, 80, 0, 0, 8, 0, 0},     /*!< FSK_NODE_ADRS */
	{1, 81, 0, 0, 8, 0, 0},     /*!< FSK_BROADCAST */
	{1, 82, 0, 0, 1, 0, 1},     /*!< FSK_AUTO_AFC_ON */
	{1, 83, 0, 0, 10, 0, 0},    /*!< FSK_PATTERN_TIMEOUT_CFG */
	{2, 33, 0, 0, 8, 0, 0},     /*!< SPI_RADIO_A__DATA */
	{2, 34, 0, 0, 8, 1, 0},     /*!< SPI_RADIO_A__DATA_READBACK */
	{2, 35, 0, 0, 8, 0, 0},     /*!< SPI_RADIO_A__ADDR */
	{2, 37, 0, 0, 1, 0, 0},     /*!< SPI_RADIO_A__CS */
	{2, 38, 0, 0, 8, 0, 0},     /*!< SPI_RADIO_B__DATA */
	{2, 39, 0, 0, 8, 1, 0},     /*!< SPI_RADIO_B__DATA_READBACK */
	{2, 40, 0, 0, 8, 0, 0},     /*!< SPI_RADIO_B__ADDR */
	{2, 42, 0, 0, 1, 0, 0},     /*!< SPI_RADIO_B__CS */
	{2, 43, 0, 0, 1, 0, 0},     /*!< RADIO_A_EN */
	{2, 43, 1, 0, 1, 0, 0},     /*!< RADIO_B_EN */
	{2, 43, 2, 0, 1, 0, 1},     /*!< RADIO_RST */
	{2, 43, 3, 0, 1, 0, 0},     /*!< LNA_A_EN */
	{2, 43, 4, 0, 1, 0, 0},     /*!< PA_A_EN */
	{2, 43, 5, 0, 1, 0, 0},     /*!< LNA_B_EN */
	{2, 43, 6, 0, 1, 0, 0},     /*!< PA_B_EN */
	{2, 44, 0, 0, 2, 0, 0},     /*!< PA_GAIN */
	{2, 45, 0, 0, 4, 0, 2},     /*!< LNA_A_CTRL_LUT */
	{2, 45, 4, 0, 4, 0, 4},     /*!< PA_A_CTRL_LUT */
	{2, 46, 0, 0, 4, 0, 2},     /*!< LNA_B_CTRL_LUT */
	{2, 46, 4, 0, 4, 0, 4},     /*!< PA_B_CTRL_LUT */
	{2, 47, 0, 0, 5, 0, 0},     /*!< CAPTURE_SOURCE */
	{2, 47, 5, 0, 1, 0, 0},     /*!< CAPTURE_START */
	{2, 47, 6, 0, 1, 0, 0},     /*!< CAPTURE_FORCE_TRIGGER */
	{2, 47, 7, 0, 1, 0, 0},     /*!< CAPTURE_WRAP */
	{2, 48, 0, 0, 16, 0, 0},    /*!< CAPTURE_PERIOD */
	{2, 51, 0, 0, 8, 1, 0},     /*!< MODEM_STATUS */
	{2, 52, 0, 0, 8, 1, 0},     /*!< VALID_HEADER_COUNTER_0 */
	{2, 54, 0, 0, 8, 1, 0},     /*!< VALID_PACKET_COUNTER_0 */
	{2, 56, 0, 0, 8, 1, 0},     /*!< VALID_HEADER_COUNTER_MBWSSF */
	{2, 57, 0, 0, 8, 1, 0},     /*!< VALID_HEADER_COUNTER_FSK */
	{2, 58, 0, 0, 8, 1, 0},     /*!< VALID_PACKET_COUNTER_MBWSSF */
	{2, 59, 0, 0, 8, 1, 0},     /*!< VALID_PACKET_COUNTER_FSK */
	{2, 60, 0, 0, 8, 1, 0},     /*!< CHANN_RSSI */
	{2, 61, 0, 0, 8, 1, 0},     /*!< BB_RSSI */
	{2, 62, 0, 0, 8, 1, 0},     /*!< DEC_RSSI */
	{2, 63, 0, 0, 8, 1, 0},     /*!< DBG_MCU_DATA */
	{2, 64, 0, 0, 8, 1, 0},     /*!< DBG_ARB_MCU_RAM_DATA */
	{2, 65, 0, 0, 8, 1, 0},     /*!< DBG_AGC_MCU_RAM_DATA */
	{2, 66, 0, 0, 16, 1, 0},    /*!< NEXT_PACKET_CNT */
	{2, 68, 0, 0, 16, 1, 0},    /*!< ADDR_CAPTURE_COUNT */
	{2, 70, 0, 0, 32, 1, 0},    /*!< TIMESTAMP */
	{2, 74, 0, 0, 4, 1, 0},     /*!< DBG_CHANN0_GAIN */
	{2, 74, 4, 0, 4, 1, 0},     /*!< DBG_CHANN1_GAIN */
	{2, 75, 0, 0, 4, 1, 0},     /*!< DBG_CHANN2_GAIN */
	{2, 75, 4, 0, 4, 1, 0},     /*!< DBG_CHANN3_GAIN */
	{2, 76, 0, 0, 4, 1, 0},     /*!< DBG_CHANN4_GAIN */
	{2, 76, 4, 0, 4, 1, 0},     /*!< DBG_CHANN5_GAIN */
	{2, 77, 0, 0, 4, 1, 0},     /*!< DBG_CHANN6_GAIN */
	{2, 77, 4, 0, 4, 1, 0},     /*!< DBG_CHANN7_GAIN */
	{2, 78, 0, 0, 4, 1, 0},     /*!< DBG_DEC_FILT_GAIN */
	{2, 79, 0, 0, 3, 1, 0},     /*!< SPI_DATA_FIFO_PTR */
	{2, 79, 3, 0, 3, 1, 0},     /*!< PACKET_DATA_FIFO_PTR */
	{2, 80, 0, 0, 8, 0, 0},     /*!< DBG_ARB_MCU_RAM_ADDR */
	{2, 81, 0, 0, 8, 0, 0},     /*!< DBG_AGC_MCU_RAM_ADDR */
	{2, 82, 0, 0, 1, 0, 0},     /*!< SPI_MASTER_CHIP_SELECT_POLARITY */
	{2, 82, 1, 0, 1, 0, 0},     /*!< SPI_MASTER_CPOL */
	{2, 82, 2, 0, 1, 0, 0},     /*!< SPI_MASTER_CPHA */
	{2, 83, 0, 0, 1, 0, 0},     /*!< SIG_GEN_ANALYSER_MUX_SEL */
	{2, 84, 0, 0, 1, 0, 0},     /*!< SIG_GEN_EN */
	{2, 84, 1, 0, 1, 0, 0},     /*!< SIG_ANALYSER_EN */
	{2, 84, 2, 0, 2, 0, 0},     /*!< SIG_ANALYSER_AVG_LEN */
	{2, 84, 4, 0, 3, 0, 0},     /*!< SIG_ANALYSER_PRECISION */
	{2, 84, 7, 0, 1, 1, 0},     /*!< SIG_ANALYSER_VALID_OUT */
	{2, 85, 0, 0, 8, 0, 0},     /*!< SIG_GEN_FREQ */
	{2, 86, 0, 0, 8, 0, 0},     /*!< SIG_ANALYSER_FREQ */
	{2, 87, 0, 0, 8, 1, 0},     /*!< SIG_ANALYSER_I_OUT */
	{2, 88, 0, 0, 8, 1, 0},     /*!< SIG_ANALYSER_Q_OUT */
	{2, 89, 0, 0, 1, 0, 0},     /*!< GPS_EN */
	{2, 89, 1, 0, 1, 0, 1},     /*!< GPS_POL */
	{2, 90, 0, 1, 8, 0, 0},     /*!< SW_TEST_REG1 */
	{2, 91, 2, 1, 6, 0, 0},     /*!< SW_TEST_REG2 */
	{2, 92, 0, 1, 16, 0, 0},    /*!< SW_TEST_REG3 */
	{2, 94, 0, 0, 4, 1, 0},     /*!< DATA_MNGT_STATUS */
	{2, 95, 0, 0, 5, 1, 0},     /*!< DATA_MNGT_CPT_FRAME_ALLOCATED */
	{2, 96, 0, 0, 5, 1, 0},     /*!< DATA_MNGT_CPT_FRAME_FINISHED */
	{2, 97, 0, 0, 5, 1, 0},     /*!< DATA_MNGT_CPT_FRAME_READEN */
	{1, 33, 0, 0, 8, 0, 0}      /*!< TX_TRIG_ALL (alias) */
};

static const struct reg2 rreg[SX1301AR_RREG_NB] =
    {
	{0, 0, 8},    /*!< SPI_CONF */
	{0, 7, 1},    /*!< SPI_CONF__SOFT_RESET */
	{0, 6, 1},    /*!< SPI_CONF__WIRE3_SPI */
	{0, 5, 1},    /*!< SPI_CONF__LSB_FIRST */
	{0, 2, 1},    /*!< SPI_CONF___LSB_FIRST */
	{0, 1, 1},    /*!< SPI_CONF___WIRE3_SPI */
	{0, 0, 1},    /*!< SPI_CONF___SOFT_RESET */
	{1, 0, 8},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL */
	{1, 6, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__TX2_MONITOR_ENABLE */
	{1, 5, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__TX1_MONITOR_ENABLE */
	{1, 3, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_RF_ENABLE */
	{1, 2, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_BBPLL_ENABLE */
	{1, 1, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_DIGITAL_CLK_ENABLE */
	{1, 0, 1},    /*!< MULTICHIP_SYNC_AND_TX_MON_CTRL__MCS_BB_ENABLE */
	{2, 0, 8},    /*!< TX_ENABLE_FILTER_CTRL */
	{2, 3, 1},    /*!< TX_ENABLE_FILTER_CTRL__THB2_EN */
	{2, 2, 1},    /*!< TX_ENABLE_FILTER_CTRL__THB1_EN */
	{2, 6, 2},    /*!< TX_ENABLE_FILTER_CTRL__TX_CHANNEL_ENABLE */
	{2, 4, 2},    /*!< TX_ENABLE_FILTER_CTRL__THB3_ENABLE_INTERP */
	{2, 0, 2},    /*!< TX_ENABLE_FILTER_CTRL__TX_FIR_ENABLE_INTERPOLATION */
	{3, 0, 8},    /*!< RX_ENABLE_FILTER_CTRL */
	{3, 3, 1},    /*!< RX_ENABLE_FILTER_CTRL__RHB2_EN */
	{3, 2, 1},    /*!< RX_ENABLE_FILTER_CTRL__RHB1_EN */
	{3, 6, 2},    /*!< RX_ENABLE_FILTER_CTRL__RX_CHANNEL_ENABLE */
	{3, 4, 2},    /*!< RX_ENABLE_FILTER_CTRL__DEC3_ENABLE_DECIMATION */
	{3, 0, 2},    /*!< RX_ENABLE_FILTER_CTRL__RX_FIR_ENABLE_DECIMATION */
	{4, 0, 8},    /*!< INPUT_SELECT */
	{4, 6, 1},    /*!< INPUT_SELECT__TX_OUTPUT */
	{4, 0, 6},    /*!< INPUT_SELECT__RX_INPUT */
	{5, 0, 8},    /*!< RFPLL_DIVIDERS */
	{5, 4, 4},    /*!< RFPLL_DIVIDERS__TX_VCO_DIVIDER */
	{5, 0, 4},    /*!< RFPLL_DIVIDERS__RX_VCO_DIVIDER */
	{6, 0, 8},    /*!< RX_CLOCK_DATA_DELAY */
	{6, 4, 4},    /*!< RX_CLOCK_DATA_DELAY__DATA_CLK_DELAY */
	{6, 0, 4},    /*!< RX_CLOCK_DATA_DELAY__RX_DATA_DELAY */
	{7, 0, 8},    /*!< TX_CLOCK_DATA_DELAY */
	{7, 4, 4},    /*!< TX_CLOCK_DATA_DELAY__FB_CLK_DELAY */
	{7, 0, 4},    /*!< TX_CLOCK_DATA_DELAY__TX_DATA_DELAY */
	{9, 0, 8},    /*!< CLOCK_ENABLE */
	{9, 4, 1},    /*!< CLOCK_ENABLE__XO_BYPASS */
	{9, 2, 1},    /*!< CLOCK_ENABLE__DIGITAL_POWER_UP */
	{9, 1, 1},    /*!< CLOCK_ENABLE__CLOCK_ENABLE_DFLT */
	{9, 0, 1},    /*!< CLOCK_ENABLE__BBPLL_ENABLE */
	{10, 0, 8},   /*!< BBPLL */
	{10, 4, 1},   /*!< BBPLL__CLKOUT_ENABLE */
	{10, 3, 1},   /*!< BBPLL__DAC_CLK_DIV2 */
	{10, 5, 3},   /*!< BBPLL__CLKOUT_SELECT */
	{10, 0, 3},   /*!< BBPLL__BBPLL_DIVIDER */
	{11, 0, 8},   /*!< TEMP_OFFSET */
	{12, 0, 8},   /*!< START_TEMP_READING */
	{12, 0, 1},   /*!< START_TEMP_READING__START_TEMP_READING */
	{13, 0, 8},   /*!< TEMP_SENSE2 */
	{13, 0, 1},   /*!< TEMP_SENSE2__TEMP_SENSE_PERIODIC_ENABLE */
	{13, 1, 7},   /*!< TEMP_SENSE2__MEASUREMENT_TIME_INTERVAL */
	{14, 0, 8},   /*!< TEMPERATURE */
	{15, 0, 8},   /*!< TEMP_SENSOR_CONFIG */
	{15, 0, 3},   /*!< TEMP_SENSOR_CONFIG__TEMP_SENSOR_DECIMATION */
	{16, 0, 8},   /*!< PARALLEL_PORT_CONF_1 */
	{16, 7, 1},   /*!< PARALLEL_PORT_CONF_1__PP_TX_SWAP_IQ */
	{16, 6, 1},   /*!< PARALLEL_PORT_CONF_1__PP_RX_SWAP_IQ */
	{16, 5, 1},   /*!< PARALLEL_PORT_CONF_1__TX_CHANNEL_SWAP */
	{16, 4, 1},   /*!< PARALLEL_PORT_CONF_1__RX_CHANNEL_SWAP */
	{16, 3, 1},   /*!< PARALLEL_PORT_CONF_1__RX_FRAME_PULSE_MODE */
	{16, 2, 1},   /*!< PARALLEL_PORT_CONF_1__R2T2_TIMING */
	{16, 1, 1},   /*!< PARALLEL_PORT_CONF_1__INVERT_DATA_BUS */
	{16, 0, 1},   /*!< PARALLEL_PORT_CONF_1__INVERT_DATA_CLK */
	{17, 0, 8},   /*!< PARALLEL_PORT_CONF_2 */
	{17, 7, 1},   /*!< PARALLEL_PORT_CONF_2__FDD_ALT_WORD_ORDER */
	{17, 6, 1},   /*!< PARALLEL_PORT_CONF_2__INVERT_RX1 */
	{17, 5, 1},   /*!< PARALLEL_PORT_CONF_2__INVERT_RX2 */
	{17, 4, 1},   /*!< PARALLEL_PORT_CONF_2__INVERT_TX1 */
	{17, 3, 1},   /*!< PARALLEL_PORT_CONF_2__INVERT_TX2 */
	{17, 2, 1},   /*!< PARALLEL_PORT_CONF_2__INVERT_RX_FRAME */
	{17, 0, 2},   /*!< PARALLEL_PORT_CONF_2__DELAY_RX_DATA */
	{18, 0, 8},   /*!< PARALLEL_PORT_CONF_3 */
	{18, 7, 1},   /*!< PARALLEL_PORT_CONF_3__FDD_RX_RATE_2TX_RATE */
	{18, 6, 1},   /*!< PARALLEL_PORT_CONF_3__SWAP_PORTS */
	{18, 5, 1},   /*!< PARALLEL_PORT_CONF_3__SINGLE_DATA_RATE */
	{18, 4, 1},   /*!< PARALLEL_PORT_CONF_3__LVDS_MODE */
	{18, 3, 1},   /*!< PARALLEL_PORT_CONF_3__HALF_DUPLEX_MODE */
	{18, 2, 1},   /*!< PARALLEL_PORT_CONF_3__SINGLE_PORT_MODE */
	{18, 1, 1},   /*!< PARALLEL_PORT_CONF_3__FULL_PORT */
	{18, 0, 1},   /*!< PARALLEL_PORT_CONF_3__FULL_DUPLEX_SWAP_BITS */
	{19, 0, 8},   /*!< ENSM_MODE */
	{19, 0, 1},   /*!< ENSM_MODE__FDD_MODE */
	{20, 0, 8},   /*!< ENSM_CONFIG_1 */
	{20, 7, 1},   /*!< ENSM_CONFIG_1__ENABLE_RX_DATA_PORT_FOR_CAL */
	{20, 6, 1},   /*!< ENSM_CONFIG_1__FORCE_RX_ON */
	{20, 5, 1},   /*!< ENSM_CONFIG_1__FORCE_TX_ON */
	{20, 4, 1},   /*!< ENSM_CONFIG_1__ENABLE_ENSM_PIN_CTRL */
	{20, 3, 1},   /*!< ENSM_CONFIG_1__LEVEL_MODE */
	{20, 2, 1},   /*!< ENSM_CONFIG_1__FORCE_ALERT_STATE */
	{20, 1, 1},   /*!< ENSM_CONFIG_1__AUTO_GAIN_LOCK */
	{20, 0, 1},   /*!< ENSM_CONFIG_1__TO_ALERT */
	{21, 0, 8},   /*!< ENSM_CONFIG_2 */
	{21, 7, 1},   /*!< ENSM_CONFIG_2__FDD_EXTERNAL_CTRL_ENABLE */
	{21, 6, 1},   /*!< ENSM_CONFIG_2__POWER_DOWN_RX_SYNTH */
	{21, 5, 1},   /*!< ENSM_CONFIG_2__POWER_DOWN_TX_SYNTH */
	{21, 4, 1},   /*!< ENSM_CONFIG_2__TXNRX_SPI_CTRL */
	{21, 3, 1},   /*!< ENSM_CONFIG_2__SYNTH_ENABLE_PIN_CTRL_MODE */
	{21, 2, 1},   /*!< ENSM_CONFIG_2__DUAL_SYNTH_MODE */
	{21, 1, 1},   /*!< ENSM_CONFIG_2__RX_SYNTH_READY_MASK */
	{21, 0, 1},   /*!< ENSM_CONFIG_2__TX_SYNTH_READY_MASK */
	{22, 0, 8},   /*!< CALIBRATION_CTRL */
	{22, 7, 1},   /*!< CALIBRATION_CTRL__RX_BB_TUNE_CAL */
	{22, 6, 1},   /*!< CALIBRATION_CTRL__TX_BB_TUNE_CAL */
	{22, 5, 1},   /*!< CALIBRATION_CTRL__RX_QUAD_CAL */
	{22, 4, 1},   /*!< CALIBRATION_CTRL__TX_QUAD_CAL */
	{22, 3, 1},   /*!< CALIBRATION_CTRL__RX_GAIN_STEP_CAL */
	{22, 2, 1},   /*!< CALIBRATION_CTRL__TXMON_CAL */
	{22, 1, 1},   /*!< CALIBRATION_CTRL__RFDC_CAL */
	{22, 0, 1},   /*!< CALIBRATION_CTRL__BBDC_CAL */
	{23, 0, 8},   /*!< STATE */
	{23, 4, 4},   /*!< STATE__CALIBRATION_SEQUENCE_STATE */
	{23, 0, 4},   /*!< STATE__ENSM_STATE */
	{24, 0, 8},   /*!< AUXDAC_1_WORD */
	{25, 0, 8},   /*!< AUXDAC_2_WORD */
	{26, 0, 8},   /*!< AUXDAC_1_CONFIG */
	{26, 5, 1},   /*!< AUXDAC_1_CONFIG__COMP_CTRL_1 */
	{26, 4, 1},   /*!< AUXDAC_1_CONFIG__AUXDAC1_STP_FACTOR */
	{26, 2, 2},   /*!< AUXDAC_1_CONFIG__AUXDAC_1_VREF */
	{26, 0, 2},   /*!< AUXDAC_1_CONFIG__AUXDAC_1_WORD_LSB */
	{27, 0, 8},   /*!< AUXDAC_2_CONFIG */
	{27, 5, 1},   /*!< AUXDAC_2_CONFIG__COMP_CTRL_2 */
	{27, 4, 1},   /*!< AUXDAC_2_CONFIG__AUXDAC2_STP_FACTOR */
	{27, 2, 2},   /*!< AUXDAC_2_CONFIG__AUXDAC_2_VREF */
	{27, 0, 2},   /*!< AUXDAC_2_CONFIG__AUXDAC_2_WORD_LSB */
	{28, 0, 8},   /*!< AUXADC_CLOCK_DIVIDER */
	{28, 0, 6},   /*!< AUXADC_CLOCK_DIVIDER__AUXADC_CLOCK_DIVIDER */
	{29, 0, 8},   /*!< AUXADC_CONFIG */
	{29, 0, 1},   /*!< AUXADC_CONFIG__AUXADC_POWER_DOWN */
	{29, 1, 3},   /*!< AUXADC_CONFIG__AUX_ADC_DECIMATION */
	{30, 0, 8},   /*!< AUXADC_WORD_MSB */
	{31, 0, 8},   /*!< AUXADC_LSB */
	{31, 0, 4},   /*!< AUXADC_LSB__AUXADC_WORD_LSB */
	{32, 0, 8},   /*!< AUTO_GPO */
	{32, 4, 4},   /*!< AUTO_GPO__GPO_ENABLE_AUTO_RX */
	{32, 0, 4},   /*!< AUTO_GPO__GPO_ENABLE_AUTO_TX */
	{33, 0, 8},   /*!< AGC_GAIN_LOCK_DELAY */
	{34, 0, 8},   /*!< AGC_ATTACK_DELAY */
	{34, 6, 1},   /*!< AGC_ATTACK_DELAY__INVERT_BYPASSED_LNA_POLARITY */
	{34, 0, 6},   /*!< AGC_ATTACK_DELAY__AGC_ATTACK_DELAY */
	{35, 0, 8},   /*!< AUXDAC_ENABLE_CTRL */
	{35, 6, 2},   /*!< AUXDAC_ENABLE_CTRL__AUXDAC_MANUAL_BAR */
	{35, 4, 2},   /*!< AUXDAC_ENABLE_CTRL__AUXDAC_AUTO_TX_BAR */
	{35, 2, 2},   /*!< AUXDAC_ENABLE_CTRL__AUXDAC_AUTO_RX_BAR */
	{35, 0, 2},   /*!< AUXDAC_ENABLE_CTRL__AUXDAC_INIT_BAR */
	{36, 0, 8},   /*!< RX_LOAD_SYNTH_DELAY */
	{37, 0, 8},   /*!< TX_LOAD_SYNTH_DELAY */
	{38, 0, 8},   /*!< EXTERNAL_LNA_CTRL */
	{38, 7, 1},   /*!< EXTERNAL_LNA_CTRL__AUXDAC_MANUAL_SELECT */
	{38, 6, 1},   /*!< EXTERNAL_LNA_CTRL__EXTERNAL_LNA2_CTRL */
	{38, 5, 1},   /*!< EXTERNAL_LNA_CTRL__EXTERNAL_LNA1_CTRL */
	{38, 4, 1},   /*!< EXTERNAL_LNA_CTRL__GPO_MANUAL_SELECT */
	{38, 0, 4},   /*!< EXTERNAL_LNA_CTRL__OPEN */
	{39, 0, 8},   /*!< GPO_FORCE_AND_INIT */
	{39, 4, 4},   /*!< GPO_FORCE_AND_INIT__GPO_MANUAL_CTRL */
	{39, 0, 4},   /*!< GPO_FORCE_AND_INIT__GPO_INIT_STATE */
	{40, 0, 8},   /*!< GPO0_RX_DELAY */
	{41, 0, 8},   /*!< GPO1_RX_DELAY */
	{42, 0, 8},   /*!< GPO2_RX_DELAY */
	{43, 0, 8},   /*!< GPO3_RX_DELAY */
	{44, 0, 8},   /*!< GPO0_TX_DELAY */
	{45, 0, 8},   /*!< GPO1_TX_DELAY */
	{46, 0, 8},   /*!< GPO2_TX_DELAY */
	{47, 0, 8},   /*!< GPO3_TX_DELAY */
	{48, 0, 8},   /*!< AUXDAC1_RX_DELAY */
	{49, 0, 8},   /*!< AUXDAC1_TX_DELAY */
	{50, 0, 8},   /*!< AUXDAC2_RX_DELAY */
	{51, 0, 8},   /*!< AUXDAC2_TX_DELAY */
	{53, 0, 8},   /*!< CTRL_OUTPUT_POINTER */
	{54, 0, 8},   /*!< CTRL_OUTPUT_ENABLE */
	{54, 7, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL7 */
	{54, 6, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL6 */
	{54, 5, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL5 */
	{54, 4, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL4 */
	{54, 3, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL3 */
	{54, 2, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL2 */
	{54, 1, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL1 */
	{54, 0, 1},   /*!< CTRL_OUTPUT_ENABLE__EN_CTRL0 */
	{55, 0, 8},   /*!< PRODUCT_ID */
	{58, 0, 8},   /*!< REFERENCE_CLOCK_CYCLES */
	{58, 0, 7},   /*!< REFERENCE_CLOCK_CYCLES__REFERENCE_CLOCK_CYCLES_PER_US */
	{59, 0, 8},   /*!< DIGITAL_IO_CTRL */
	{59, 7, 1},   /*!< DIGITAL_IO_CTRL__CLK_OUT_DRIVE */
	{59, 6, 1},   /*!< DIGITAL_IO_CTRL__DATACLK_DRIVE */
	{59, 2, 1},   /*!< DIGITAL_IO_CTRL__DATA_PORT_DRIVE */
	{59, 4, 2},   /*!< DIGITAL_IO_CTRL__DATACLK_SLEW */
	{59, 0, 2},   /*!< DIGITAL_IO_CTRL__DATA_PORT_SLEW */
	{60, 0, 8},   /*!< LVDS_BIAS_CTRL */
	{60, 5, 1},   /*!< LVDS_BIAS_CTRL__RX_ON_CHIP_TERM */
	{60, 4, 1},   /*!< LVDS_BIAS_CTRL__LVDS_BYPASS_BIAS_R */
	{60, 3, 1},   /*!< LVDS_BIAS_CTRL__LVDS_TX_LO_VCM */
	{60, 6, 2},   /*!< LVDS_BIAS_CTRL__CLK_OUT_SLEW */
	{60, 0, 3},   /*!< LVDS_BIAS_CTRL__LVDS_BIAS */
	{61, 0, 8},   /*!< LVDS_INVERT_CTRL1 */
	{62, 0, 8},   /*!< LVDS_INVERT_CTRL2 */
	{63, 0, 8},   /*!< SDM_CTRL_1 */
	{63, 2, 1},   /*!< SDM_CTRL_1__INIT_BB_FO_CAL */
	{63, 0, 1},   /*!< SDM_CTRL_1__BBPLL_RESET_BAR */
	{65, 0, 8},   /*!< FRACT_BB_FREQ_WORD_1 */
	{66, 0, 8},   /*!< FRACT_BB_FREQ_WORD_2 */
	{67, 0, 8},   /*!< FRACT_BB_FREQ_WORD_3 */
	{68, 0, 8},   /*!< INTEGER_BB_FREQ_WORD */
	{69, 0, 8},   /*!< CLOCK_CTRL */
	{69, 0, 2},   /*!< CLOCK_CTRL__REF_FREQ_SCALER */
	{70, 0, 8},   /*!< CP_CURRENT */
	{70, 0, 6},   /*!< CP_CURRENT__CHARGE_PUMP_CURRENT */
	{71, 0, 8},   /*!< CP_BLEED_CURRENT */
	{71, 7, 1},   /*!< CP_BLEED_CURRENT__MCS_REFCLK_SCALE_EN */
	{72, 0, 8},   /*!< LOOP_FILTER_1 */
	{72, 5, 3},   /*!< LOOP_FILTER_1__C1_WORD */
	{72, 0, 5},   /*!< LOOP_FILTER_1__R1_WORD */
	{73, 0, 8},   /*!< LOOP_FILTER_2 */
	{73, 7, 1},   /*!< LOOP_FILTER_2__R2_WORD */
	{73, 2, 5},   /*!< LOOP_FILTER_2__C2_WORD */
	{73, 0, 2},   /*!< LOOP_FILTER_2__C1_WORD_LSB */
	{74, 0, 8},   /*!< LOOP_FILTER_3 */
	{74, 7, 1},   /*!< LOOP_FILTER_3__BYPASS_C3 */
	{74, 6, 1},   /*!< LOOP_FILTER_3__BYPASS_R2 */
	{74, 2, 4},   /*!< LOOP_FILTER_3__C3_WORD */
	{74, 0, 2},   /*!< LOOP_FILTER_3__R2_WORD_LSB */
	{75, 0, 8},   /*!< VCO_CTRL */
	{75, 7, 1},   /*!< VCO_CTRL__FREQ_CAL_ENABLE */
	{75, 4, 1},   /*!< VCO_CTRL__FREQ_CAL_RESET */
	{75, 5, 2},   /*!< VCO_CTRL__FREQ_CAL_COUNT_LENGTH */
	{76, 0, 8},   /*!< VCO_PROGRAM_1 */
	{77, 0, 8},   /*!< VCO_PROGRAM_2 */
	{78, 0, 8},   /*!< SDM_CTRL */
	{78, 4, 1},   /*!< SDM_CTRL__CAL_CLOCK_DIV_4 */
	{80, 0, 8},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE */
	{80, 4, 1},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE__RX_LO_POWER_DOWN */
	{80, 3, 1},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_ALC_POWER_DOWN */
	{80, 2, 1},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_PTAT_POWER_DOWN */
	{80, 1, 1},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_POWER_DOWN */
	{80, 0, 1},   /*!< RX_SYNTH_POWER_DOWN_OVERRIDE__RX_SYNTH_VCO_LDO_POWER_DOWN */
	{81, 0, 8},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE */
	{81, 4, 1},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE__TX_LO_POWER_DOWN */
	{81, 3, 1},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_ALC_POWER_DOWN */
	{81, 2, 1},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_PTAT_POWER_DOWN */
	{81, 1, 1},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_POWER_DOWN */
	{81, 0, 1},   /*!< TX_SYNTH_POWER_DOWN_OVERRIDE__TX_SYNTH_VCO_LDO_POWER_DOWN */
	{82, 0, 8},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_1 */
	{82, 6, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_OFFSET_DAC_CGIN_POWER_DOWN */
	{82, 4, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_LMT_OVERLOAD_POWER_DOWN */
	{82, 2, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_MIXER_GM_POWER_DOWN */
	{82, 0, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_1__RX_CGB_POWER_DOWN */
	{83, 0, 8},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_2 */
	{83, 6, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_BBF_POWER_DOWN */
	{83, 4, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_TIA_POWER_DOWN */
	{83, 2, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_MIXER_POWER_DOWN */
	{83, 0, 2},   /*!< RX_ANALOG_POWER_DOWN_OVERRIDE_2__RX_OFFSET_DAC_CGOUT_POWER_DOWN */
	{84, 0, 8},   /*!< RX1_ADC_POWER_DOWN_OVERRIDE */
	{85, 0, 8},   /*!< RX2_ADC_POWER_DOWN_OVERRIDE */
	{86, 0, 8},   /*!< TX_ANALOG_POWER_DOWN_OVERRIDE_1 */
	{86, 6, 2},   /*!< TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_SECONDARY_FILTER_POWER_DOWN */
	{86, 4, 2},   /*!< TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_BBF_POWER_DOWN */
	{86, 2, 2},   /*!< TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_DAC_POWER_DOWN */
	{86, 0, 2},   /*!< TX_ANALOG_POWER_DOWN_OVERRIDE_1__TX_DAC_BIAS_POWER_DOWN */
	{87, 0, 8},   /*!< ANALOG_POWER_DOWN_OVERRIDE */
	{87, 5, 1},   /*!< ANALOG_POWER_DOWN_OVERRIDE__RX_EXT_VCO_BUFFER_POWER_DOWN */
	{87, 4, 1},   /*!< ANALOG_POWER_DOWN_OVERRIDE__TX_EXT_VCO_BUFFER_POWER_DOWN */
	{87, 2, 2},   /*!< ANALOG_POWER_DOWN_OVERRIDE__TX_MONITOR_POWER_DOWN */
	{87, 0, 2},   /*!< ANALOG_POWER_DOWN_OVERRIDE__TX_UPCONVERTER_POWER_DOWN */
	{88, 0, 8},   /*!< MISC_POWER_DOWN_OVERRIDE */
	{88, 6, 1},   /*!< MISC_POWER_DOWN_OVERRIDE__RX_LNA_POWER_DOWN */
	{88, 1, 1},   /*!< MISC_POWER_DOWN_OVERRIDE__DCXO_POWER_DOWN */
	{88, 0, 1},   /*!< MISC_POWER_DOWN_OVERRIDE__MASTER_BIAS_POWER_DOWN */
	{88, 2, 2},   /*!< MISC_POWER_DOWN_OVERRIDE__RX_CALIBRATION_POWER_DOWN */
	{94, 0, 8},   /*!< CH_1_OVERFLOW */
	{94, 7, 1},   /*!< CH_1_OVERFLOW__BBPLL_LOCK */
	{94, 6, 1},   /*!< CH_1_OVERFLOW__CH_1_INT3 */
	{94, 5, 1},   /*!< CH_1_OVERFLOW__CH1_HB3 */
	{94, 4, 1},   /*!< CH_1_OVERFLOW__CH1_HB2 */
	{94, 3, 1},   /*!< CH_1_OVERFLOW__CH1_QEC */
	{94, 2, 1},   /*!< CH_1_OVERFLOW__CH1_HB1 */
	{94, 1, 1},   /*!< CH_1_OVERFLOW__CH1_TFIR */
	{94, 0, 1},   /*!< CH_1_OVERFLOW__CH1_RFIR */
	{95, 0, 8},   /*!< CH_2_OVERFLOW */
	{95, 6, 1},   /*!< CH_2_OVERFLOW__CH2_INT3 */
	{95, 5, 1},   /*!< CH_2_OVERFLOW__CH2_HB3 */
	{95, 4, 1},   /*!< CH_2_OVERFLOW__CH2_HB2 */
	{95, 3, 1},   /*!< CH_2_OVERFLOW__CH2_QEC */
	{95, 2, 1},   /*!< CH_2_OVERFLOW__CH2_HB1 */
	{95, 1, 1},   /*!< CH_2_OVERFLOW__CH2_TFIR */
	{95, 0, 1},   /*!< CH_2_OVERFLOW__CH2_RFIR */
	{96, 0, 8},   /*!< TX_FILTER_COEF_ADDR */
	{97, 0, 8},   /*!< TX_FILTER_COEF_WRITE_DATA_1 */
	{98, 0, 8},   /*!< TX_FILTER_COEF_WRITE_DATA_2 */
	{99, 0, 8},   /*!< TX_FILTER_COEF_READ_DATA_1 */
	{100, 0, 8},  /*!< TX_FILTER_COEF_READ_DATA_2 */
	{101, 0, 8},  /*!< TX_FILTER_CONF */
	{101, 0, 1},  /*!< TX_FILTER_CONF__TX_FIR_GAIN_6DB */
	{101, 1, 1},  /*!< TX_FILTER_CONF__FIR_START_CLK */
	{101, 2, 1},  /*!< TX_FILTER_CONF__FIR_WRITE */
	{101, 3, 2},  /*!< TX_FILTER_CONF__FIR_SELECT */
	{101, 5, 3},  /*!< TX_FILTER_CONF__FIR_NUM_TAPS */
	{103, 0, 8},  /*!< TX_MON_LOW_GAIN */
	{103, 5, 1},  /*!< TX_MON_LOW_GAIN__TX_MON_TRACK */
	{103, 0, 5},  /*!< TX_MON_LOW_GAIN__TX_MON_LOW_GAIN */
	{104, 0, 8},  /*!< TX_MON_HIGH_GAIN */
	{104, 0, 5},  /*!< TX_MON_HIGH_GAIN__TX_MON_HIGH_GAIN */
	{105, 0, 8},  /*!< TX_MON_DELAY */
	{106, 0, 8},  /*!< TX_LEVEL_THRESH */
	{106, 2, 6},  /*!< TX_LEVEL_THRESH__TX_LEVEL_THRESH */
	{106, 0, 2},  /*!< TX_LEVEL_THRESH__TX_MON_DELAY_COUNTER */
	{107, 0, 8},  /*!< TX_RSSI1 */
	{108, 0, 8},  /*!< TX_RSSI2 */
	{109, 0, 8},  /*!< TX_RSSI_LSB */
	{109, 1, 1},  /*!< TX_RSSI_LSB__TX_RSSI_2 */
	{109, 0, 1},  /*!< TX_RSSI_LSB__TX_RSSI_1 */
	{110, 0, 8},  /*!< TPM_MODE_ENABLE */
	{110, 6, 1},  /*!< TPM_MODE_ENABLE__ONE_SHOT_MODE */
	{110, 0, 4},  /*!< TPM_MODE_ENABLE__TX_MON_DURATION */
	{111, 0, 8},  /*!< TX_MON_TEMP_GAIN_COEF */
	{112, 0, 8},  /*!< TX_MON_1_CONFIG */
	{112, 2, 6},  /*!< TX_MON_1_CONFIG__TX_MON_1_LO_CM */
	{112, 0, 2},  /*!< TX_MON_1_CONFIG__TX_MON_1_GAIN */
	{113, 0, 8},  /*!< TX_MON_2_CONFIG */
	{113, 2, 6},  /*!< TX_MON_2_CONFIG__TX_MON_2_LO_CM */
	{113, 0, 2},  /*!< TX_MON_2_CONFIG__TX_MON_2_GAIN */
	{115, 0, 8},  /*!< TX1_ATTEN_0 */
	{116, 0, 8},  /*!< TX1_ATTEN_1 */
	{116, 0, 1},  /*!< TX1_ATTEN_1__TX_1_ATTEN */
	{117, 0, 8},  /*!< TX2_ATTEN_0 */
	{118, 0, 8},  /*!< TX2_ATTEN_1 */
	{118, 0, 1},  /*!< TX2_ATTEN_1__TX_2_ATTEN */
	{119, 0, 8},  /*!< TX_ATTEN_OFFSET */
	{119, 6, 1},  /*!< TX_ATTEN_OFFSET__MASK_CLR_ATTEN_UPDATE */
	{119, 0, 6},  /*!< TX_ATTEN_OFFSET__TX_ATTEN_OFFSET */
	{120, 0, 8},  /*!< TX_ATTEN_THRESH */
	{121, 0, 8},  /*!< TX1_DIG_ATTEN */
	{121, 6, 1},  /*!< TX1_DIG_ATTEN__SEL_TX1_TX2 */
	{124, 0, 8},  /*!< TX2_DIG_ATTEN */
	{124, 6, 1},  /*!< TX2_DIG_ATTEN__IMMEDIATELY_UPDATE_TPC_ATTEN */
	{127, 0, 8},  /*!< TX1_SYMBOL_ATTEN */
	{127, 0, 7},  /*!< TX1_SYMBOL_ATTEN__TX_1_SYMBOL_ATTEN */
	{128, 0, 8},  /*!< TX2_SYMBOL_ATTEN */
	{128, 0, 7},  /*!< TX2_SYMBOL_ATTEN__TX_2_SYMBOL_ATTEN */
	{129, 0, 8},  /*!< TX_SYMBOL_ATTEN_CONFIG */
	{129, 3, 1},  /*!< TX_SYMBOL_ATTEN_CONFIG__USE_TX1_PIN_SYMBOL_ATTEN */
	{129, 1, 1},  /*!< TX_SYMBOL_ATTEN_CONFIG__USE_CTRL_IN_FOR_SYMBOL_ATTEN */
	{129, 0, 1},  /*!< TX_SYMBOL_ATTEN_CONFIG__ENABLE_SYMBOL_ATTEN */
	{142, 0, 8},  /*!< TX1_OUT_1_PHASE_CORR */
	{143, 0, 8},  /*!< TX1_OUT_1_GAIN_CORR */
	{144, 0, 8},  /*!< TX2_OUT_1_PHASE_CORR */
	{145, 0, 8},  /*!< TX2_OUT_1_GAIN_CORR */
	{146, 0, 8},  /*!< TX1_OUT_1_OFFSET_I */
	{147, 0, 8},  /*!< TX1_OUT_1_OFFSET_Q */
	{148, 0, 8},  /*!< TX2_OUT_1_OFFSET_I */
	{149, 0, 8},  /*!< TX2_OUT_1_OFFSET_Q */
	{150, 0, 8},  /*!< TX1_OUT_2_PHASE_CORR */
	{151, 0, 8},  /*!< TX1_OUT_2_GAIN_CORR */
	{152, 0, 8},  /*!< TX2_OUT_2_PHASE_CORR */
	{153, 0, 8},  /*!< TX2_OUT_2_GAIN_CORR */
	{154, 0, 8},  /*!< TX1_OUT_2_OFFSET_I */
	{155, 0, 8},  /*!< TX1_OUT_2_OFFSET_Q */
	{156, 0, 8},  /*!< TX2_OUT_2_OFFSET_I */
	{157, 0, 8},  /*!< TX2_OUT_2_OFFSET_Q */
	{159, 0, 8},  /*!< TX_FORCE_BITS */
	{159, 7, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_2_TX2_OFFSET */
	{159, 6, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_2_TX1_OFFSET */
	{159, 5, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_2_TX2_PHASE_GAIN */
	{159, 4, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_2_TX1_PHASE_GAIN */
	{159, 3, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_1_TX2_OFFSET */
	{159, 2, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_1_TX1_OFFSET */
	{159, 1, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_1_TX2_PHASE_GAIN */
	{159, 0, 1},  /*!< TX_FORCE_BITS__FORCE_OUT_1_TX1_PHASE_GAIN */
	{160, 0, 8},  /*!< QUAD_CAL_NCO_FREQ_PHASE_OFFSET */
	{160, 5, 2},  /*!< QUAD_CAL_NCO_FREQ_PHASE_OFFSET__RX_NCO_FREQ */
	{160, 0, 5},  /*!< QUAD_CAL_NCO_FREQ_PHASE_OFFSET__RX_NCO_PHASE_OFFSET */
	{161, 0, 8},  /*!< QUAD_CAL_CTRL */
	{161, 7, 1},  /*!< QUAD_CAL_CTRL__FREE_RUN_ENABLE */
	{161, 6, 1},  /*!< QUAD_CAL_CTRL__SETTLE_MAIN_ENABLE */
	{161, 5, 1},  /*!< QUAD_CAL_CTRL__DC_OFFSET_ENABLE */
	{161, 4, 1},  /*!< QUAD_CAL_CTRL__GAIN_ENABLE */
	{161, 3, 1},  /*!< QUAD_CAL_CTRL__PHASE_ENABLE */
	{161, 2, 1},  /*!< QUAD_CAL_CTRL__QUAD_CAL_SOFT_RESET */
	{161, 0, 2},  /*!< QUAD_CAL_CTRL__M_DECIM */
	{162, 0, 8},  /*!< KEXP_1 */
	{162, 6, 2},  /*!< KEXP_1__KEXP_TX */
	{162, 4, 2},  /*!< KEXP_1__KEXP_TX_COMP */
	{162, 2, 2},  /*!< KEXP_1__KEXP_DC_I */
	{162, 0, 2},  /*!< KEXP_1__KEXP_DC_Q */
	{163, 0, 8},  /*!< KEXP_2 */
	{163, 5, 1},  /*!< KEXP_2__INVERT_I_DATA */
	{163, 4, 1},  /*!< KEXP_2__INVERT_Q_DATA */
	{163, 6, 2},  /*!< KEXP_2__TX_NCO_FREQ */
	{163, 2, 2},  /*!< KEXP_2__KEXP_PHASE */
	{163, 0, 2},  /*!< KEXP_2__KEXP_AMP */
	{164, 0, 8},  /*!< QUAD_SETTLE_COUNT */
	{165, 0, 8},  /*!< MAG_FTEST_THRESH */
	{166, 0, 8},  /*!< MAG_FTEST_THRESH_2 */
	{167, 0, 8},  /*!< QUAD_CAL_STATUS_TX1 */
	{167, 1, 1},  /*!< QUAD_CAL_STATUS_TX1__TX1_LO_CONV */
	{167, 0, 1},  /*!< QUAD_CAL_STATUS_TX1__TX1_SSB_CONV */
	{167, 2, 6},  /*!< QUAD_CAL_STATUS_TX1__TX1_CONVERGENCE_COUNT */
	{168, 0, 8},  /*!< QUAD_CAL_STATUS_TX2 */
	{168, 1, 1},  /*!< QUAD_CAL_STATUS_TX2__TX2_LO_CONV */
	{168, 0, 1},  /*!< QUAD_CAL_STATUS_TX2__TX2_SSB_CONV */
	{168, 2, 6},  /*!< QUAD_CAL_STATUS_TX2__TX2_CONVERGENCE_COUNT */
	{169, 0, 8},  /*!< QUAD_CAL_COUNT */
	{170, 0, 8},  /*!< TX_QUAD_FULL_LMT_GAIN */
	{170, 0, 7},  /*!< TX_QUAD_FULL_LMT_GAIN__RX_FULL_TABLELMT_TABLE_GAIN */
	{171, 0, 8},  /*!< SQUARER_CONFIG */
	{171, 5, 1},  /*!< SQUARER_CONFIG__GM_STAGE_TIME_CON_OVERRIDE */
	{171, 4, 1},  /*!< SQUARER_CONFIG__GM_STAGE_MV_HP_POLE */
	{171, 3, 1},  /*!< SQUARER_CONFIG__GM_STAGE_LOWER_CM */
	{171, 0, 1},  /*!< SQUARER_CONFIG__BYPASS_BIAS_R */
	{171, 1, 2},  /*!< SQUARER_CONFIG__VBIAS_CTRL */
	{172, 0, 8},  /*!< TX_QUAD_CAL_ATTEN */
	{173, 0, 8},  /*!< THRESH_ACCUM */
	{173, 0, 4},  /*!< THRESH_ACCUM__THRESH_ACCUMULATOR */
	{174, 0, 8},  /*!< TX_QUAD_LPF_GAIN */
	{174, 0, 5},  /*!< TX_QUAD_LPF_GAIN__RX_LPF_GAIN */
	{176, 0, 8},  /*!< TXDAC_VDS_I */
	{176, 0, 6},  /*!< TXDAC_VDS_I__TXDAC_VDS_I */
	{177, 0, 8},  /*!< TXDAC_VDS_Q */
	{177, 0, 6},  /*!< TXDAC_VDS_Q__TXDAC_VDS_Q */
	{178, 0, 8},  /*!< TXDAC_GN_I */
	{178, 0, 6},  /*!< TXDAC_GN_I__TXDAC_GN_I */
	{179, 0, 8},  /*!< TXDAC_GN_Q */
	{179, 0, 6},  /*!< TXDAC_GN_Q__TXDAC_GN_Q */
	{192, 0, 8},  /*!< TXBBF_OPAMP_A */
	{192, 5, 2},  /*!< TXBBF_OPAMP_A__OPAMPA_OUTPUT_BIAS */
	{192, 3, 2},  /*!< TXBBF_OPAMP_A__OPAMPA_RZ */
	{192, 0, 3},  /*!< TXBBF_OPAMP_A__OPAMP_A_CC */
	{193, 0, 8},  /*!< TXBBF_OPAMP_B */
	{193, 5, 2},  /*!< TXBBF_OPAMP_B__OPAMPB_OUTPUT_BIAS */
	{193, 3, 2},  /*!< TXBBF_OPAMP_B__OPAMPB_RZ */
	{193, 0, 3},  /*!< TXBBF_OPAMP_B__OPAMP_B_CC */
	{194, 0, 8},  /*!< TX_BBF_R1 */
	{194, 7, 1},  /*!< TX_BBF_R1__OVERRIDE_ENABLE */
	{194, 0, 5},  /*!< TX_BBF_R1__R1 */
	{195, 0, 8},  /*!< TX_BBF_R2 */
	{195, 0, 5},  /*!< TX_BBF_R2__R2 */
	{196, 0, 8},  /*!< TX_BBF_R3 */
	{196, 0, 5},  /*!< TX_BBF_R3__R3 */
	{197, 0, 8},  /*!< TX_BBF_R4 */
	{197, 0, 5},  /*!< TX_BBF_R4__R4 */
	{198, 0, 8},  /*!< TX_BBF_RP */
	{198, 0, 5},  /*!< TX_BBF_RP__RP */
	{199, 0, 8},  /*!< TX_BBF_C1 */
	{199, 0, 6},  /*!< TX_BBF_C1__C1 */
	{200, 0, 8},  /*!< TX_BBF_C2 */
	{200, 0, 6},  /*!< TX_BBF_C2__C2 */
	{201, 0, 8},  /*!< TX_BBF_CP */
	{201, 0, 6},  /*!< TX_BBF_CP__CP */
	{202, 0, 8},  /*!< TX_TUNE_CTRL */
	{202, 2, 1},  /*!< TX_TUNE_CTRL__PD_TUNE */
	{202, 1, 1},  /*!< TX_TUNE_CTRL__TUNER_RESAMPLE */
	{202, 0, 1},  /*!< TX_TUNE_CTRL__TUNER_RESAMPLE_PHASE */
	{202, 5, 2},  /*!< TX_TUNE_CTRL__TUNE_CTRL */
	{203, 0, 8},  /*!< TX_BBF_R2B */
	{203, 7, 1},  /*!< TX_BBF_R2B__TX_BBF_BYPASS_BIAS_R */
	{203, 5, 1},  /*!< TX_BBF_R2B__R2B_OVR */
	{203, 0, 5},  /*!< TX_BBF_R2B__R2B */
	{204, 0, 8},  /*!< TX_BBF_TUNE */
	{204, 3, 1},  /*!< TX_BBF_TUNE__BBF1_COMP_I */
	{204, 2, 1},  /*!< TX_BBF_TUNE__BBF1_COMP_Q */
	{204, 1, 1},  /*!< TX_BBF_TUNE__BBF2_COMP_I */
	{204, 0, 1},  /*!< TX_BBF_TUNE__BBF2_COMP_Q */
	{208, 0, 8},  /*!< CONFIG0 */
	{208, 6, 2},  /*!< CONFIG0__BIAS */
	{208, 4, 2},  /*!< CONFIG0__RGM */
	{208, 2, 2},  /*!< CONFIG0__CC */
	{208, 0, 2},  /*!< CONFIG0__AMPBIAS */
	{209, 0, 8},  /*!< RESISTOR */
	{209, 0, 4},  /*!< RESISTOR__RESISTOR */
	{210, 0, 8},  /*!< CAPACITOR */
	{210, 0, 6},  /*!< CAPACITOR__CAPACITOR */
	{211, 0, 8},  /*!< LO_CM */
	{211, 5, 2},  /*!< LO_CM__LO_COMMON_MODE */
	{214, 0, 8},  /*!< TX_BBF_TUNE_DIVIDER */
	{215, 0, 8},  /*!< TX_BBF_TUNE_MODE */
	{215, 4, 1},  /*!< TX_BBF_TUNE_MODE__EVALTIME */
	{215, 0, 1},  /*!< TX_BBF_TUNE_MODE__TX_BBF_TUNE_DIVIDER */
	{215, 5, 2},  /*!< TX_BBF_TUNE_MODE__TUNE_COMP_MASK */
	{215, 1, 3},  /*!< TX_BBF_TUNE_MODE__TUNER_MODE */
	{240, 0, 8},  /*!< RX_FILTER_COEF_ADDR */
	{241, 0, 8},  /*!< RX_FILTER_COEF_DATA_1 */
	{242, 0, 8},  /*!< RX_FILTER_COEF_DATA_2 */
	{243, 0, 8},  /*!< RX_FILTER_COEF_READ_DATA_1 */
	{244, 0, 8},  /*!< RX_FILTER_COEF_READ_DATA_2 */
	{245, 0, 8},  /*!< RX_FILTER_CONFIG */
	{245, 2, 1},  /*!< RX_FILTER_CONFIG__WRITE_RX */
	{245, 1, 1},  /*!< RX_FILTER_CONFIG__START_RX_CLOCK */
	{245, 5, 3},  /*!< RX_FILTER_CONFIG__NUMBER_OF_TAPS */
	{245, 3, 2},  /*!< RX_FILTER_CONFIG__SELECT_RX_CH */
	{246, 0, 8},  /*!< RX_FILTER_GAIN */
	{246, 0, 2},  /*!< RX_FILTER_GAIN__FILTER_GAIN */
	{250, 0, 8},  /*!< AGC_CONFIG_1 */
	{250, 7, 1},  /*!< AGC_CONFIG_1__DEC_PWR_FOR_LOW_PWR */
	{250, 6, 1},  /*!< AGC_CONFIG_1__DEC_PWR_FOR_LOCK_LEVEL */
	{250, 5, 1},  /*!< AGC_CONFIG_1__DEC_PWR_FOR_GAIN_LOCK_EXIT */
	{250, 4, 1},  /*!< AGC_CONFIG_1__SLOW_ATTACK_HYBRID_MODE */
	{250, 2, 2},  /*!< AGC_CONFIG_1__RX2_GAIN_CTRL_SETUP */
	{250, 0, 2},  /*!< AGC_CONFIG_1__RX1_GAIN_CTRL_SETUP */
	{251, 0, 8},  /*!< AGC_CONFIG_2 */
	{251, 7, 1},  /*!< AGC_CONFIG_2__AGC_SOFT_RESET */
	{251, 6, 1},  /*!< AGC_CONFIG_2__AGC_GAIN_UNLOCK_CTRL */
	{251, 3, 1},  /*!< AGC_CONFIG_2__AGC_USE_FULL_GAIN_TABLE */
	{251, 2, 1},  /*!< AGC_CONFIG_2__DIG_GAIN_EN */
	{251, 1, 1},  /*!< AGC_CONFIG_2__MAN_GAIN_CTRL_RX2 */
	{251, 0, 1},  /*!< AGC_CONFIG_2__MAN_GAIN_CTRL_RX1 */
	{252, 0, 8},  /*!< AGC_CONFIG_3 */
	{252, 4, 1},  /*!< AGC_CONFIG_3__INCDEC_LMT_GAIN */
	{252, 3, 1},  /*!< AGC_CONFIG_3__USE_AGC_FOR_LMTLPF_GAIN */
	{252, 5, 3},  /*!< AGC_CONFIG_3__MANUAL_INCR_STEP_SIZE */
	{252, 0, 3},  /*!< AGC_CONFIG_3__ADC_OVERRANGE_SAMPLE_SIZE */
	{253, 0, 8},  /*!< MAX_LMT_FULL_GAIN */
	{253, 0, 7},  /*!< MAX_LMT_FULL_GAIN__MAXIMUM_FULL_TABLELMT_TABLE_INDEX */
	{254, 0, 8},  /*!< PEAK_WAIT_TIME */
	{254, 5, 3},  /*!< PEAK_WAIT_TIME__MANUAL_CTRL_IN_DECR_GAIN_STP_SIZE */
	{254, 0, 5},  /*!< PEAK_WAIT_TIME__PEAK_OVERLOAD_WAIT_TIME */
	{256, 0, 8},  /*!< DIGITAL_GAIN */
	{256, 5, 3},  /*!< DIGITAL_GAIN__DIG_GAIN_STP_SIZE */
	{256, 0, 5},  /*!< DIGITAL_GAIN__MAXIMUM_DIGITAL_GAIN */
	{257, 0, 8},  /*!< AGC_LOCK_LEVEL */
	{257, 7, 1},  /*!< AGC_LOCK_LEVEL__ENABLE_DIG_SAT_OVRG */
	{257, 0, 7},  /*!< AGC_LOCK_LEVEL__AGC_LOCK_LEVEL_FAST_AGC_INNER_HIGH_THRESH_SLOW */
	{258, 0, 8},  /*!< ADC_NOISE_CORRECTION_FACTOR */
	{259, 0, 8},  /*!< GAIN_STP_CONFIG1 */
	{259, 5, 3},  /*!< GAIN_STP_CONFIG1__LMT_DETECTOR_SETTLING_TIME */
	{259, 2, 3},  /*!< GAIN_STP_CONFIG1__DEC_STP_SIZE_FOR_LARGE_LMT_OVERLOAD */
	{259, 0, 2},  /*!< GAIN_STP_CONFIG1__ADC_NOISE_CORRECTION_FACTOR */
	{260, 0, 8},  /*!< ADC_SMALL_OVERLOAD_THRESH */
	{261, 0, 8},  /*!< ADC_LARGE_OVERLOAD_THRESH */
	{262, 0, 8},  /*!< GAIN_STP_CONFIG_2 */
	{262, 4, 3},  /*!< GAIN_STP_CONFIG_2__DECREMENT_STP_SIZE_FOR_SMALL_LPF_GAIN_CHANGE */
	{262, 0, 4},  /*!< GAIN_STP_CONFIG_2__LARGE_LPF_GAIN_STEP */
	{263, 0, 8},  /*!< SMALL_LMT_OVERLOAD_THRESH */
	{263, 7, 1},  /*!< SMALL_LMT_OVERLOAD_THRESH__FORCE_PD_RESET_RX2 */
	{263, 6, 1},  /*!< SMALL_LMT_OVERLOAD_THRESH__FOR_PD_RESET_RX1 */
	{263, 0, 6},  /*!< SMALL_LMT_OVERLOAD_THRESH__SMALL_LMT_OVERLOAD_THRESH */
	{264, 0, 8},  /*!< LARGE_LMT_OVERLOAD_THRESH */
	{264, 0, 6},  /*!< LARGE_LMT_OVERLOAD_THRESH__LARGE_LMT_OVERLOAD_THRESH */
	{265, 0, 8},  /*!< RX1_MANUAL_LMT_FULL_GAIN */
	{265, 7, 1},  /*!< RX1_MANUAL_LMT_FULL_GAIN__POWER_MEAS_IN_STATE_5_MSB */
	{265, 0, 7},  /*!< RX1_MANUAL_LMT_FULL_GAIN__RX1_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX */
	{266, 0, 8},  /*!< RX1_MANUAL_LPF_GAIN */
	{266, 5, 3},  /*!< RX1_MANUAL_LPF_GAIN__POWER_MEAS_IN_STATE_5 */
	{266, 0, 5},  /*!< RX1_MANUAL_LPF_GAIN__RX1_MANUAL_LPF_GAIN */
	{267, 0, 8},  /*!< RX1_MANUAL_DIGITALFORCED_GAIN */
	{267, 5, 1},  /*!< RX1_MANUAL_DIGITALFORCED_GAIN__FORCE_RX1_DIGITAL_GAIN */
	{267, 0, 5},  /*!< RX1_MANUAL_DIGITALFORCED_GAIN__RX1_MANUALFORCED_DIGITAL_GAIN */
	{268, 0, 8},  /*!< RX2_MANUAL_LMT_FULL_GAIN */
	{268, 0, 7},  /*!< RX2_MANUAL_LMT_FULL_GAIN__RX2_MANUAL_FULL_TABLE_LMT_TABLE_GAIN_INDEX */
	{269, 0, 8},  /*!< RX2_MANUAL_LPF_GAIN */
	{269, 0, 5},  /*!< RX2_MANUAL_LPF_GAIN__RX2_MANUAL_LPF_GAIN */
	{270, 0, 8},  /*!< RX2_MANUAL_DIGITALFORCED_GAIN */
	{270, 5, 1},  /*!< RX2_MANUAL_DIGITALFORCED_GAIN__FORCE_RX2_DIGITAL_GAIN */
	{270, 0, 5},  /*!< RX2_MANUAL_DIGITALFORCED_GAIN__RX2_MANUALFORCED_DIGITAL_GAIN */
	{272, 0, 8},  /*!< FAST_CONFIG_1 */
	{272, 7, 1},  /*!< FAST_CONFIG_1__ENABLE_GAIN_INC_AFTER_GAIN_LOCK */
	{272, 6, 1},  /*!< FAST_CONFIG_1__GOTO_OPT_GAIN_IF_ENERGY_LOST_OR_EN_AGC_HIGH */
	{272, 5, 1},  /*!< FAST_CONFIG_1__GOTO_SET_GAIN_IF_EN_AGC_HIGH */
	{272, 4, 1},  /*!< FAST_CONFIG_1__GOTO_SET_GAIN_IF_EXIT_RX_STATE */
	{272, 3, 1},  /*!< FAST_CONFIG_1__DONT_UNLOCK_GAIN_IF_ENERGY_LOST */
	{272, 2, 1},  /*!< FAST_CONFIG_1__GOTO_OPTIMIZED_GAIN_IF_EXIT_RX_STATE */
	{272, 1, 1},  /*!< FAST_CONFIG_1__DONT_UNLOCK_GAIN_IF_LG_ADC_OR_LMT_OVRG */
	{272, 0, 1},  /*!< FAST_CONFIG_1__ENABLE_INCR_GAIN */
	{273, 0, 8},  /*!< FAST_CONFIG_2_SETTLING_DELAY */
	{273, 7, 1},  /*!< FAST_CONFIG_2_SETTLING_DELAY__USE_LAST_LOCK_LEVEL_FOR_SET_GAIN */
	{273, 6, 1},  /*!< FAST_CONFIG_2_SETTLING_DELAY__ENABLE_LMT_GAIN_INC_FOR_LOCK_LEVEL */
	{273, 5, 1},  /*!< FAST_CONFIG_2_SETTLING_DELAY__GOTO_MAX_GAIN_OR_OPT_GAIN_IF_EN_AGC_HIGH */
	{273, 0, 5},  /*!< FAST_CONFIG_2_SETTLING_DELAY__SETTLING_DELAY */
	{274, 0, 8},  /*!< FAST_ENERGY_LOST_THRESH */
	{274, 6, 2},  /*!< FAST_ENERGY_LOST_THRESH__POST_LOCK_LEVEL_STP_SIZE_FOR_LPF_TABLE_FULL_TABLE */
	{274, 0, 6},  /*!< FAST_ENERGY_LOST_THRESH__ENERGY_LOST_THRESH */
	{275, 0, 8},  /*!< FAST_STRONGER_SIGNAL_THRESH */
	{275, 6, 2},  /*!< FAST_STRONGER_SIGNAL_THRESH__POST_LOCK_LEVEL_STP_FOR_LMT_TABLE */
	{275, 0, 6},  /*!< FAST_STRONGER_SIGNAL_THRESH__STRONGER_SIGNAL_THRESH */
	{276, 0, 8},  /*!< FAST_LOW_POWER_THRESH */
	{276, 7, 1},  /*!< FAST_LOW_POWER_THRESH__DONT_UNLOCK_GAIN_IF_ADC_OVRG */
	{276, 0, 7},  /*!< FAST_LOW_POWER_THRESH__LOW_POWER_THRESH */
	{277, 0, 8},  /*!< FAST_STRONG_SIGNAL_FREEZE */
	{277, 7, 1},  /*!< FAST_STRONG_SIGNAL_FREEZE__DONT_UNLOCK_GAIN_IF_STRONGER_SIGNAL */
	{278, 0, 8},  /*!< FAST_FINAL_OVER_RANGE_AND_OPT_GAIN */
	{278, 5, 3},  /*!< FAST_FINAL_OVER_RANGE_AND_OPT_GAIN__FINAL_OVER_RANGE_COUNT */
	{278, 0, 4},  /*!< FAST_FINAL_OVER_RANGE_AND_OPT_GAIN__OPTIMIZE_GAIN_OFFSET */
	{279, 0, 8},  /*!< FAST_ENERGY_DETECT_COUNT */
	{279, 5, 3},  /*!< FAST_ENERGY_DETECT_COUNT__INCREMENT_GAIN_STP_LPFLMT */
	{279, 0, 5},  /*!< FAST_ENERGY_DETECT_COUNT__ENERGY_DETECT_COUNT */
	{280, 0, 8},  /*!< FAST_AGCLL_UPPER_LIMIT */
	{280, 0, 6},  /*!< FAST_AGCLL_UPPER_LIMIT__AGCLL_MAX_INCREASE */
	{281, 0, 8},  /*!< FAST_GAIN_LOCK_EXIT_COUNT */
	{281, 0, 6},  /*!< FAST_GAIN_LOCK_EXIT_COUNT__GAIN_LOCK_EXIT_COUNT */
	{282, 0, 8},  /*!< FAST_INITIAL_LMT_GAIN_LIMIT */
	{282, 0, 7},  /*!< FAST_INITIAL_LMT_GAIN_LIMIT__INITIAL_LMT_GAIN_LIMIT */
	{283, 0, 8},  /*!< FAST_INCREMENT_TIME */
	{288, 0, 8},  /*!< AGC_INNER_LOW_THRESH */
	{288, 7, 1},  /*!< AGC_INNER_LOW_THRESH__PREVENT_GAIN_INC */
	{288, 0, 7},  /*!< AGC_INNER_LOW_THRESH__AGC_INNER_LOW_THRESH */
	{289, 0, 8},  /*!< LMT_OVERLOAD_COUNTERS */
	{289, 4, 4},  /*!< LMT_OVERLOAD_COUNTERS__LARGE_LMT_OVERLOAD_EXED_COUNTER */
	{289, 0, 4},  /*!< LMT_OVERLOAD_COUNTERS__SMALL_LMT_OVERLOAD_EXED_COUNTER */
	{290, 0, 8},  /*!< ADC_OVERLOAD_COUNTERS */
	{290, 4, 4},  /*!< ADC_OVERLOAD_COUNTERS__LARGE_ADC_OVERLOAD_EXED_COUNTER */
	{290, 0, 4},  /*!< ADC_OVERLOAD_COUNTERS__SMALL_ADC_OVERLOAD_EXED_COUNTER */
	{291, 0, 8},  /*!< GAIN_STP1 */
	{291, 7, 1},  /*!< GAIN_STP1__IMMED_GAIN_CHANGE_IF_LG_LMT_OVERLOAD */
	{291, 3, 1},  /*!< GAIN_STP1__IMMED_GAIN_CHANGE_IF_LG_ADC_OVERLOAD */
	{291, 4, 3},  /*!< GAIN_STP1__AGC_INNER_HIGH_THRESH_EXED_STP_SIZE */
	{291, 0, 3},  /*!< GAIN_STP1__AGC_INNER_LOW_THRESH_EXED_STP_SIZE */
	{292, 0, 8},  /*!< GAIN_UPDATE_COUNTER1 */
	{293, 0, 8},  /*!< GAIN_UPDATE_COUNTER2 */
	{296, 0, 8},  /*!< DIGITAL_SAT_COUNTER */
	{296, 5, 1},  /*!< DIGITAL_SAT_COUNTER__DOUBLE_GAIN_COUNTER */
	{296, 4, 1},  /*!< DIGITAL_SAT_COUNTER__ENABLE_SYNC_FOR_GAIN_COUNTER */
	{296, 0, 4},  /*!< DIGITAL_SAT_COUNTER__DIG_SATURATION_EXED_COUNTER */
	{297, 0, 8},  /*!< OUTER_POWER_THRESHS */
	{297, 4, 4},  /*!< OUTER_POWER_THRESHS__AGC_OUTER_HIGH_THRESH */
	{297, 0, 4},  /*!< OUTER_POWER_THRESHS__AGC_OUTER_LOW_THRESH */
	{298, 0, 8},  /*!< GAIN_STP_2 */
	{298, 4, 4},  /*!< GAIN_STP_2__AGC_OUTER_HIGH_THRESH_EXED_STP_SIZE */
	{298, 0, 4},  /*!< GAIN_STP_2__AGC_OUTER_LOW_THRESH_EXED_STP_SIZE */
	{300, 0, 8},  /*!< EXT_LNA_HIGH_GAIN */
	{300, 0, 6},  /*!< EXT_LNA_HIGH_GAIN__EXT_LNA_HIGH_GAIN */
	{301, 0, 8},  /*!< EXT_LNA_LOW_GAIN */
	{301, 0, 6},  /*!< EXT_LNA_LOW_GAIN__EXT_LNA_LOW_GAIN */
	{304, 0, 8},  /*!< GAIN_TABLE_ADDRESS */
	{304, 0, 7},  /*!< GAIN_TABLE_ADDRESS__GAIN_TABLE_ADDRESS */
	{305, 0, 8},  /*!< GAIN_TABLE_WRITE_DATA1 */
	{305, 7, 1},  /*!< GAIN_TABLE_WRITE_DATA1__EXT_LNA_CTRL */
	{305, 5, 2},  /*!< GAIN_TABLE_WRITE_DATA1__LNA_GAIN */
	{305, 0, 5},  /*!< GAIN_TABLE_WRITE_DATA1__MIXER_GM_GAIN */
	{306, 0, 8},  /*!< GAIN_TABLE_WRITE_DATA2 */
	{306, 5, 1},  /*!< GAIN_TABLE_WRITE_DATA2__TIA_GAIN */
	{306, 0, 5},  /*!< GAIN_TABLE_WRITE_DATA2__LPF_GAIN */
	{307, 0, 8},  /*!< GAIN_TABLE_WRITE_DATA3 */
	{307, 5, 1},  /*!< GAIN_TABLE_WRITE_DATA3__RF_DC_CAL */
	{307, 0, 5},  /*!< GAIN_TABLE_WRITE_DATA3__DIGITAL_GAIN */
	{308, 0, 8},  /*!< GAIN_TABLE_READ_DATA1 */
	{309, 0, 8},  /*!< GAIN_TABLE_READ_DATA2 */
	{310, 0, 8},  /*!< GAIN_TABLE_READ_DATA3 */
	{311, 0, 8},  /*!< GAIN_TABLE_CONFIG */
	{311, 2, 1},  /*!< GAIN_TABLE_CONFIG__WRITE_GAIN_TABLE */
	{311, 1, 1},  /*!< GAIN_TABLE_CONFIG__START_GAIN_TABLE_CLOCK */
	{311, 3, 2},  /*!< GAIN_TABLE_CONFIG__RECEIVER_SELECT */
	{312, 0, 8},  /*!< GM_SUB_TABLE_ADDRESS */
	{313, 0, 8},  /*!< GM_SUB_TABLE_GAIN_WRITE */
	{313, 0, 7},  /*!< GM_SUB_TABLE_GAIN_WRITE__GM_SUB_TABLE_GAIN_WRITE */
	{314, 0, 8},  /*!< GM_SUB_TABLE_BIAS_WRITE */
	{314, 0, 5},  /*!< GM_SUB_TABLE_BIAS_WRITE__GM_SUB_TABLE_BIAS_WRITE */
	{315, 0, 8},  /*!< GM_SUB_TABLE_CTRL_WRITE */
	{315, 0, 6},  /*!< GM_SUB_TABLE_CTRL_WRITE__GM_SUB_TABLE_CTRL_WRITE */
	{316, 0, 8},  /*!< GM_SUB_TABLE_GAIN_READ */
	{316, 0, 7},  /*!< GM_SUB_TABLE_GAIN_READ__GM_SUB_TABLE_GAIN_READ */
	{317, 0, 8},  /*!< GM_SUB_TABLE_BIAS_READ */
	{317, 0, 5},  /*!< GM_SUB_TABLE_BIAS_READ__GM_SUB_TABLE_BIAS_READ */
	{318, 0, 8},  /*!< GM_SUB_TABLE_CTRL_READ */
	{318, 0, 6},  /*!< GM_SUB_TABLE_CTRL_READ__GM_SUB_TABLE_CTRL_READ */
	{319, 0, 8},  /*!< GM_SUB_TABLE_CONFIG */
	{319, 2, 1},  /*!< GM_SUB_TABLE_CONFIG__WRITE_GM_SUB_TABLE */
	{319, 1, 1},  /*!< GM_SUB_TABLE_CONFIG__START_GM_SUB_TABLE_CLOCK */
	{320, 0, 8},  /*!< WORD_ADDRESS */
	{321, 0, 8},  /*!< GAIN_DIFF_WORDERROR_WRITE */
	{321, 0, 6},  /*!< GAIN_DIFF_WORDERROR_WRITE__CALIB_TABLE_GAIN_DIFFERROR_WORD */
	{322, 0, 8},  /*!< GAIN_ERROR_READ */
	{322, 0, 5},  /*!< GAIN_ERROR_READ__CALIB_TABLE_GAIN_ERROR */
	{323, 0, 8},  /*!< CONFIG */
	{323, 4, 1},  /*!< CONFIG__READ_SELECT */
	{323, 3, 1},  /*!< CONFIG__WRITE_MIXER_ERROR_TABLE */
	{323, 2, 1},  /*!< CONFIG__WRITE_LNA_ERROR_TABLE */
	{323, 1, 1},  /*!< CONFIG__WRITE_LNA_GAIN_DIFF */
	{323, 0, 1},  /*!< CONFIG__START_CALIB_TABLE_CLOCK */
	{323, 5, 2},  /*!< CONFIG__CALIB_TABLE_SELECT */
	{324, 0, 8},  /*!< LNA_GAIN_DIFF_READ_BACK */
	{324, 0, 6},  /*!< LNA_GAIN_DIFF_READ_BACK__LNA_CALIB_TABLE_GAIN_DIFFERENCE_WORD */
	{325, 0, 8},  /*!< MAX_MIXER_CALIBRATION_GAIN_INDEX */
	{325, 0, 5},  /*!< MAX_MIXER_CALIBRATION_GAIN_INDEX__MAX_MIXER_CALIBRATION_GAIN_INDEX */
	{326, 0, 8},  /*!< TEMP_GAIN_COEF */
	{327, 0, 8},  /*!< SETTLE_TIME */
	{327, 7, 1},  /*!< SETTLE_TIME__ENABLE_DIG_GAIN_CORR */
	{327, 6, 1},  /*!< SETTLE_TIME__FORCE_TEMP_SENSOR_FOR_CAL */
	{327, 0, 6},  /*!< SETTLE_TIME__SETTLE_TIME */
	{328, 0, 8},  /*!< MEASURE_DURATION */
	{328, 0, 4},  /*!< MEASURE_DURATION__GAIN_CAL_MEAS_DURATION */
	{329, 0, 8},  /*!< CAL_TEMP_SENSOR_WORD */
	{336, 0, 8},  /*!< MEASURE_DURATION_01 */
	{336, 4, 4},  /*!< MEASURE_DURATION_01__MEASUREMENT_DURATION_1 */
	{336, 0, 4},  /*!< MEASURE_DURATION_01__MEASUREMENT_DURATION_0 */
	{337, 0, 8},  /*!< MEASURE_DURATION_23 */
	{337, 4, 4},  /*!< MEASURE_DURATION_23__MEASUREMENT_DURATION_3 */
	{337, 0, 4},  /*!< MEASURE_DURATION_23__MEASUREMENT_DURATION_2 */
	{338, 0, 8},  /*!< RSSI_WEIGHT_0 */
	{339, 0, 8},  /*!< RSSI_WEIGHT_1 */
	{340, 0, 8},  /*!< RSSI_WEIGHT_2 */
	{341, 0, 8},  /*!< RSSI_WEIGHT_3 */
	{342, 0, 8},  /*!< RSSI_DELAY */
	{343, 0, 8},  /*!< RSSI_WAIT_TIME */
	{344, 0, 8},  /*!< RSSI_CONFIG */
	{344, 5, 1},  /*!< RSSI_CONFIG__START_RSSI_MEAS */
	{344, 1, 1},  /*!< RSSI_CONFIG__ENABLE_ADC_POWER_MEAS */
	{344, 0, 1},  /*!< RSSI_CONFIG__DEFAULT_RSSI_MEAS_MODE */
	{344, 6, 2},  /*!< RSSI_CONFIG__RFIR_FOR_RSSI_MEASUREMENT */
	{344, 2, 3},  /*!< RSSI_CONFIG__RSSI_MODE_SELECT */
	{345, 0, 8},  /*!< ADC_MEASURE_DURATION_01 */
	{345, 4, 4},  /*!< ADC_MEASURE_DURATION_01__ADC_POWER_MEASUREMENT_DURATION_1 */
	{345, 0, 4},  /*!< ADC_MEASURE_DURATION_01__ADC_POWER_MEASUREMENT_DURATION_0 */
	{346, 0, 8},  /*!< ADC_WEIGHT_0 */
	{347, 0, 8},  /*!< ADC_WEIGHT_1 */
	{348, 0, 8},  /*!< DEC_POWER_MEASURE_DURATION_0 */
	{348, 7, 1},  /*!< DEC_POWER_MEASURE_DURATION_0__USE_HB3_OUT_FOR_ADC_PWR_MEAS */
	{348, 6, 1},  /*!< DEC_POWER_MEASURE_DURATION_0__USE_HB1_OUT_FOR_DEC_PWR_MEAS */
	{348, 5, 1},  /*!< DEC_POWER_MEASURE_DURATION_0__ENABLE_DEC_PWR_MEAS */
	{348, 4, 1},  /*!< DEC_POWER_MEASURE_DURATION_0__DEFAULT_MODE_ADC_POWER */
	{348, 0, 4},  /*!< DEC_POWER_MEASURE_DURATION_0__DEC_POWER_MEASUREMENT_DURATION */
	{349, 0, 8},  /*!< LNA_GAIN */
	{349, 0, 1},  /*!< LNA_GAIN__DB_GAIN_READBACK_CHANNEL */
	{349, 1, 7},  /*!< LNA_GAIN__MAX_LNA_GAIN */
	{352, 0, 8},  /*!< CH1_ADC_POWER */
	{353, 0, 8},  /*!< CH1_RX_FILTER_POWER */
	{354, 0, 8},  /*!< CH2_ADC_POWER */
	{355, 0, 8},  /*!< CH2_RX_FILTER_POWER */
	{360, 0, 8},  /*!< RX_QUAD_CAL_LEVEL */
	{360, 0, 4},  /*!< RX_QUAD_CAL_LEVEL__RX_QUAD_CAL_LEVEL */
	{361, 0, 8},  /*!< CALIBRATION_CONFIG_1 */
	{361, 7, 1},  /*!< CALIBRATION_CONFIG_1__ENABLE_PHASE_CORR */
	{361, 6, 1},  /*!< CALIBRATION_CONFIG_1__ENABLE_GAIN_CORR */
	{361, 5, 1},  /*!< CALIBRATION_CONFIG_1__USE_SETTLE_COUNT_FOR_DC_CAL_WAIT */
	{361, 4, 1},  /*!< CALIBRATION_CONFIG_1__FIXED_DC_CAL_WAIT_TIME */
	{361, 3, 1},  /*!< CALIBRATION_CONFIG_1__FREE_RUN_MODE */
	{361, 2, 1},  /*!< CALIBRATION_CONFIG_1__ENABLE_CORR_WORD_DECIMATION */
	{361, 1, 1},  /*!< CALIBRATION_CONFIG_1__ENABLE_TRACKING_MODE_CH2 */
	{361, 0, 1},  /*!< CALIBRATION_CONFIG_1__ENABLE_TRACKING_MODE_CH1 */
	{362, 0, 8},  /*!< CALIBRATION_CONFIG_2 */
	{362, 7, 1},  /*!< CALIBRATION_CONFIG_2__SOFT_RESET */
	{362, 0, 5},  /*!< CALIBRATION_CONFIG_2__K_EXP_PHASE */
	{363, 0, 8},  /*!< CALIBRATION_CONFIG_3 */
	{363, 7, 1},  /*!< CALIBRATION_CONFIG_3__PREVENT_POS_LOOP_GAIN */
	{363, 0, 5},  /*!< CALIBRATION_CONFIG_3__K_EXP_AMPLITUDE */
	{364, 0, 8},  /*!< CALIB_COUNT */
	{365, 0, 8},  /*!< SETTLE_COUNT */
	{366, 0, 8},  /*!< RX_QUAD_GAIN1 */
	{366, 0, 7},  /*!< RX_QUAD_GAIN1__RX_FULL_TABLELMT_TABLE_GAIN */
	{367, 0, 8},  /*!< RX_QUAD_GAIN2 */
	{367, 5, 3},  /*!< RX_QUAD_GAIN2__CORRECTION_WORD_DECIMATION_M */
	{367, 0, 5},  /*!< RX_QUAD_GAIN2__RX_LPF_GAIN */
	{368, 0, 8},  /*!< RX1_INPUT_A_PHASE_CORR */
	{369, 0, 8},  /*!< RX1_INPUT_A_GAIN_CORR */
	{370, 0, 8},  /*!< RX2_INPUT_A_PHASE_CORR */
	{371, 0, 8},  /*!< RX2_INPUT_A_GAIN_CORR */
	{372, 0, 8},  /*!< RX1_INPUT_A_Q_OFFSET */
	{373, 0, 8},  /*!< RX1_INPUT_A_OFFSETS */
	{373, 2, 6},  /*!< RX1_INPUT_A_OFFSETS__RX1_INPUT_A_I_DC_OFFSET_LSB */
	{373, 0, 2},  /*!< RX1_INPUT_A_OFFSETS__RX1_INPUT_A_Q_DC_OFFSET */
	{374, 0, 8},  /*!< INPUT_A_OFFSETS_1 */
	{374, 4, 4},  /*!< INPUT_A_OFFSETS_1__RX2_INPUT_A_Q_DC_OFFSET_LSB */
	{374, 0, 4},  /*!< INPUT_A_OFFSETS_1__RX1_INPUT_A_I_DC_OFFSET_MSB */
	{375, 0, 8},  /*!< RX2_INPUT_A_OFFSETS */
	{375, 6, 2},  /*!< RX2_INPUT_A_OFFSETS__RX2_INPUT_A_I_DC_OFFSET */
	{375, 0, 6},  /*!< RX2_INPUT_A_OFFSETS__RX2_INPUT_A_Q_DC_OFFSET_MSB */
	{376, 0, 8},  /*!< RX2_INPUT_A_I_OFFSET */
	{377, 0, 8},  /*!< RX1_INPUT_BC_PHASE_CORR */
	{378, 0, 8},  /*!< RX1_INPUT_BC_GAIN_CORR */
	{379, 0, 8},  /*!< RX2_INPUT_BC_PHASE_CORR */
	{380, 0, 8},  /*!< RX2_INPUT_BC_GAIN_CORR */
	{381, 0, 8},  /*!< RX1_INPUT_BC_Q_OFFSET */
	{382, 0, 8},  /*!< RX1_INPUT_BC_OFFSETS */
	{382, 2, 6},  /*!< RX1_INPUT_BC_OFFSETS__RX1_INPUT_BC_I_DC_OFFSET_LSB */
	{382, 0, 2},  /*!< RX1_INPUT_BC_OFFSETS__RX1_INPUT_BC_Q_DC_OFFSET */
	{383, 0, 8},  /*!< INPUT_BC_OFFSETS_1 */
	{383, 4, 4},  /*!< INPUT_BC_OFFSETS_1__RX2_INPUT_BC_Q_DC_OFFSET_LSB */
	{383, 0, 4},  /*!< INPUT_BC_OFFSETS_1__RX1_INPUT_BC_I_DC_OFFSET_MSB */
	{384, 0, 8},  /*!< RX2_INPUT_BC_OFFSETS */
	{384, 6, 2},  /*!< RX2_INPUT_BC_OFFSETS__RX2_INPUT_BC_I_DC_OFFSET */
	{384, 0, 6},  /*!< RX2_INPUT_BC_OFFSETS__RX2_INPUT_BC_Q_DC_OFFSET_MSB */
	{385, 0, 8},  /*!< RX2_INPUT_BC_I_OFFSET */
	{386, 0, 8},  /*!< FORCE_BITS */
	{386, 7, 1},  /*!< FORCE_BITS__RX2_INPUT_BC_FORCE_OFFSET */
	{386, 6, 1},  /*!< FORCE_BITS__RX1_INPUT_BC_FORCE_OFFSET */
	{386, 5, 1},  /*!< FORCE_BITS__RX2_INPUT_BC_FORCE_PHGAIN */
	{386, 4, 1},  /*!< FORCE_BITS__RX1_INPUT_BC_FORCE_PHGAIN */
	{386, 3, 1},  /*!< FORCE_BITS__RX2_INPUT_A_FORCE_OFFSET */
	{386, 2, 1},  /*!< FORCE_BITS__RX1_INPUT_A_FORCE_OFFSET */
	{386, 1, 1},  /*!< FORCE_BITS__RX2_INPUT_A_FORCE_PHGAIN */
	{386, 0, 1},  /*!< FORCE_BITS__RX1_INPUT_A_FORCE_PHGAIN */
	{389, 0, 8},  /*!< WAIT_COUNT */
	{390, 0, 8},  /*!< RF_DC_OFFSET_COUNT */
	{391, 0, 8},  /*!< RF_DC_OFFSET_CONFIG_1 */
	{391, 4, 2},  /*!< RF_DC_OFFSET_CONFIG_1__DAC_FS */
	{391, 0, 4},  /*!< RF_DC_OFFSET_CONFIG_1__RF_DC_CALIBRATION_COUNT */
	{392, 0, 8},  /*!< RF_DC_OFFSET_ATTEN */
	{392, 5, 3},  /*!< RF_DC_OFFSET_ATTEN__RF_DC_OFFSET_TABLE_UPDATE_COUNT */
	{392, 0, 5},  /*!< RF_DC_OFFSET_ATTEN__RF_DC_OFFSET_ATTEN */
	{393, 0, 8},  /*!< INVERT_BITS */
	{393, 7, 1},  /*!< INVERT_BITS__INVERT_RX2_RF_DC_CGIN_WORD */
	{393, 6, 1},  /*!< INVERT_BITS__INVERT_RX1_RF_DC_CGIN_WORD */
	{393, 5, 1},  /*!< INVERT_BITS__INVERT_RX2_RF_DC_CGOUT_WORD */
	{393, 4, 1},  /*!< INVERT_BITS__INVERT_RX1_RF_DC_CGOUT_WORD */
	{395, 0, 8},  /*!< DC_OFFSET_CONFIG2 */
	{395, 7, 1},  /*!< DC_OFFSET_CONFIG2__USE_WAIT_COUNTER_FOR_RF_DC_INIT_CAL */
	{395, 6, 1},  /*!< DC_OFFSET_CONFIG2__ENABLE_FAST_SETTLE_MODE */
	{395, 5, 1},  /*!< DC_OFFSET_CONFIG2__ENABLE_BB_DC_OFFSET_TRACKING */
	{395, 4, 1},  /*!< DC_OFFSET_CONFIG2__RESET_ACC_ON_GAIN_CHANGE */
	{395, 3, 1},  /*!< DC_OFFSET_CONFIG2__ENABLE_RF_OFFSET_TRACKING */
	{395, 0, 3},  /*!< DC_OFFSET_CONFIG2__DC_OFFSET_UPDATE */
	{396, 0, 8},  /*!< RF_CAL_GAIN_INDEX */
	{396, 0, 7},  /*!< RF_CAL_GAIN_INDEX__RF_MINIMUM_CALIBRATION_GAIN_INDEX */
	{397, 0, 8},  /*!< SOI_THRESH */
	{397, 0, 7},  /*!< SOI_THRESH__RF_SOI_THRESH */
	{400, 0, 8},  /*!< BB_DC_OFFSET_SHIFT */
	{400, 7, 1},  /*!< BB_DC_OFFSET_SHIFT__INCREASE_COUNT_DURATION */
	{400, 5, 2},  /*!< BB_DC_OFFSET_SHIFT__BB_TRACKING_DECIMATE */
	{400, 0, 5},  /*!< BB_DC_OFFSET_SHIFT__BB_DC_M_SHIFT */
	{401, 0, 8},  /*!< BB_DC_OFFSET_FAST_SETTLE_SHIFT */
	{401, 7, 1},  /*!< BB_DC_OFFSET_FAST_SETTLE_SHIFT__READ_BACK_CH_SEL */
	{401, 6, 1},  /*!< BB_DC_OFFSET_FAST_SETTLE_SHIFT__UPDATE_TRACKING_WORD */
	{401, 5, 1},  /*!< BB_DC_OFFSET_FAST_SETTLE_SHIFT__FORCE_RX_NULL */
	{401, 0, 5},  /*!< BB_DC_OFFSET_FAST_SETTLE_SHIFT__BB_DC_TRACKING_FAST_SETTLE_M_SHIFT */
	{402, 0, 8},  /*!< BB_FAST_SETTLE_DUR */
	{403, 0, 8},  /*!< BB_DC_OFFSET_COUNT */
	{404, 0, 8},  /*!< BB_DC_OFFSET_ATTEN */
	{404, 0, 4},  /*!< BB_DC_OFFSET_ATTEN__BB_DC_OFFSET_ATTEN */
	{410, 0, 8},  /*!< RX1_BB_DC_WORD_I_MSB */
	{410, 0, 7},  /*!< RX1_BB_DC_WORD_I_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_I */
	{411, 0, 8},  /*!< RX1_BB_DC_WORD_I_LSB */
	{412, 0, 8},  /*!< RX1_BB_DC_WORD_Q_MSB */
	{412, 0, 7},  /*!< RX1_BB_DC_WORD_Q_MSB__RX1_BB_DC_OFFSET_CORRECTION_WORD_Q */
	{413, 0, 8},  /*!< RX1_BB_DC_WORD_Q_LSB */
	{414, 0, 8},  /*!< RX2_BB_DC_WORD_I_MSB */
	{414, 0, 7},  /*!< RX2_BB_DC_WORD_I_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_I */
	{415, 0, 8},  /*!< RX2_BB_DC_WORD_I_LSB */
	{416, 0, 8},  /*!< RX2_BB_DC_WORD_Q_MSB */
	{416, 0, 7},  /*!< RX2_BB_DC_WORD_Q_MSB__RX2_BB_DC_OFFSET_CORRECTION_WORD_Q */
	{417, 0, 8},  /*!< RX2_BB_DC_WORD_Q_LSB */
	{418, 0, 8},  /*!< BB_TRACK_CORR_WORD_I_MSB */
	{418, 0, 7},  /*!< BB_TRACK_CORR_WORD_I_MSB__RX1RX2_BB_DC_OFFSET_TRACKING_CORRECTION_WORD_I */
	{419, 0, 8},  /*!< BB_TRACK_CORR_WORD_I_LSB */
	{420, 0, 8},  /*!< BB_TRACK_CORR_WORD_Q_MSB */
	{420, 0, 7},  /*!< BB_TRACK_CORR_WORD_Q_MSB__RX1RX2_BB_DC_OFFSET_TRACKING_CORRECTION_WORD_Q */
	{421, 0, 8},  /*!< BB_TRACK_CORR_WORD_Q_LSB */
	{423, 0, 8},  /*!< RX1_RSSI_SYMBOL */
	{424, 0, 8},  /*!< RX1_RSSI_PREAMBLE */
	{425, 0, 8},  /*!< RX2_RSSI_SYMBOL */
	{426, 0, 8},  /*!< RX2_RSSI_PREAMBLE */
	{427, 0, 8},  /*!< SYMBOL_LSB */
	{427, 1, 1},  /*!< SYMBOL_LSB__RX2_RSSI_SYMBOL */
	{427, 0, 1},  /*!< SYMBOL_LSB__RX1_RSSI_SYMBOL */
	{428, 0, 8},  /*!< PREAMBLE_LSB */
	{428, 1, 1},  /*!< PREAMBLE_LSB__RX2_RSSI_PREAMBLE */
	{428, 0, 1},  /*!< PREAMBLE_LSB__RX1_RSSI_PREAMBLE */
	{429, 0, 8},  /*!< RX_PATH_GAIN_MSB */
	{430, 0, 8},  /*!< RX_PATH_GAIN_LSB */
	{430, 0, 1},  /*!< RX_PATH_GAIN_LSB__RX_PATH_GAIN */
	{432, 0, 8},  /*!< RX_DIFF_LNA_FORCE */
	{432, 7, 1},  /*!< RX_DIFF_LNA_FORCE__FORCE_RX2_LNA_GAIN */
	{432, 6, 1},  /*!< RX_DIFF_LNA_FORCE__RX2_LNA_BYPASS */
	{432, 3, 1},  /*!< RX_DIFF_LNA_FORCE__FORCE_RX1_LNA_GAIN */
	{432, 2, 1},  /*!< RX_DIFF_LNA_FORCE__RX1_LNA_BYPASS */
	{432, 4, 2},  /*!< RX_DIFF_LNA_FORCE__RX2_LNA_GAIN */
	{432, 0, 2},  /*!< RX_DIFF_LNA_FORCE__RX1_LNA_GAIN */
	{433, 0, 8},  /*!< RX_LNA_BIAS_COARSE */
	{433, 0, 4},  /*!< RX_LNA_BIAS_COARSE__RX_LNA_BIAS_COARSE */
	{434, 0, 8},  /*!< RX_LNA_BIAS_FINE_0 */
	{434, 5, 3},  /*!< RX_LNA_BIAS_FINE_0__RX_LNA_PCASCODE_BIAS */
	{434, 0, 5},  /*!< RX_LNA_BIAS_FINE_0__RX_LNA_BIAS */
	{435, 0, 8},  /*!< RX_LNA_BIAS_FINE_1 */
	{435, 0, 2},  /*!< RX_LNA_BIAS_FINE_1__RX_LNA_P_CASCODE_BIAS_FINE */
	{448, 0, 8},  /*!< RX_MIX_GM_CONFIG */
	{448, 5, 3},  /*!< RX_MIX_GM_CONFIG__RX_MIX_GM_CM_OUT */
	{448, 0, 2},  /*!< RX_MIX_GM_CONFIG__RX_MIX_GM_PLOAD */
	{449, 0, 8},  /*!< RX1_MIX_GM_FORCE */
	{449, 6, 1},  /*!< RX1_MIX_GM_FORCE__FORCE_RX1_MIX_GM */
	{449, 0, 6},  /*!< RX1_MIX_GM_FORCE__RX1_MIX_GM_GAIN */
	{450, 0, 8},  /*!< RX1_MIX_GM_BIAS_FORCE */
	{450, 0, 5},  /*!< RX1_MIX_GM_BIAS_FORCE__RX1_MIX_GM_BIAS */
	{451, 0, 8},  /*!< RX2_MIX_GM_FORCE */
	{451, 6, 1},  /*!< RX2_MIX_GM_FORCE__FORCE_RX2_MIX_GM */
	{451, 0, 6},  /*!< RX2_MIX_GM_FORCE__RX2_MIX_GM_GAIN */
	{452, 0, 8},  /*!< RX2_MIX_GM_BIAS_FORCE */
	{452, 0, 5},  /*!< RX2_MIX_GM_BIAS_FORCE__RX2_MIX_GM_BIAS */
	{456, 0, 8},  /*!< INPUT_A_MSBS */
	{456, 6, 2},  /*!< INPUT_A_MSBS__INPUT_A_RX1_Q */
	{456, 4, 2},  /*!< INPUT_A_MSBS__INPUT_A_RX1_I */
	{456, 2, 2},  /*!< INPUT_A_MSBS__INPUT_A_RX2_I */
	{456, 0, 2},  /*!< INPUT_A_MSBS__INPUT_A_RX2_Q */
	{457, 0, 8},  /*!< INPUT_A_RX1_I */
	{458, 0, 8},  /*!< INPUT_A_RX1_Q */
	{459, 0, 8},  /*!< INPUT_A_RX2_I */
	{460, 0, 8},  /*!< INPUT_A_RX2_Q */
	{461, 0, 8},  /*!< INPUTS_BC_RX1_I */
	{462, 0, 8},  /*!< BAND1_RX1_Q */
	{463, 0, 8},  /*!< INPUTS_BC_RX2_I */
	{464, 0, 8},  /*!< INPUTS_BC_RX2_Q */
	{465, 0, 8},  /*!< INPUTS_BC_MSBS */
	{465, 6, 2},  /*!< INPUTS_BC_MSBS__INPUTS_BC_RX1_Q */
	{465, 4, 2},  /*!< INPUTS_BC_MSBS__INPUTS_BC_RX1_I */
	{465, 2, 2},  /*!< INPUTS_BC_MSBS__INPUTS_BC_RX2_I */
	{465, 0, 2},  /*!< INPUTS_BC_MSBS__INPUTS_BC_RX2_Q */
	{466, 0, 8},  /*!< FORCE_OS_DAC */
	{466, 2, 1},  /*!< FORCE_OS_DAC__FORCE_CGIN_DAC */
	{469, 0, 8},  /*!< RX_MIX_LO_CM */
	{469, 0, 6},  /*!< RX_MIX_LO_CM__RX_MIX_LO_CM */
	{470, 0, 8},  /*!< RX_CGB_SEG_ENABLE */
	{470, 0, 6},  /*!< RX_CGB_SEG_ENABLE__RX_CGB_SEG_ENABLE */
	{471, 0, 8},  /*!< RX_MIX_INPUTBIAS */
	{471, 4, 2},  /*!< RX_MIX_INPUTBIAS__RX_CGB_INPUT_CM_SEL */
	{471, 0, 4},  /*!< RX_MIX_INPUTBIAS__RX_CGB_BIAS */
	{475, 0, 8},  /*!< RX_TIA_CONFIG */
	{475, 3, 1},  /*!< RX_TIA_CONFIG__TIA2_OVERRIDE_C */
	{475, 2, 1},  /*!< RX_TIA_CONFIG__TIA2_OVERRIDE_R */
	{475, 1, 1},  /*!< RX_TIA_CONFIG__TIA1_OVERRIDE_C */
	{475, 0, 1},  /*!< RX_TIA_CONFIG__TIA1_OVERRIDE_R */
	{475, 5, 3},  /*!< RX_TIA_CONFIG__TIA_SEL_CC */
	{476, 0, 8},  /*!< TIA1_C_LSB */
	{476, 6, 2},  /*!< TIA1_C_LSB__TIA1_RF */
	{476, 0, 6},  /*!< TIA1_C_LSB__TIA1_C_LSB */
	{477, 0, 8},  /*!< TIA1_C_MSB */
	{477, 0, 7},  /*!< TIA1_C_MSB__TIA1_C_MSB */
	{478, 0, 8},  /*!< TIA2_C_LSB */
	{478, 6, 2},  /*!< TIA2_C_LSB__TIA2_RF */
	{478, 0, 6},  /*!< TIA2_C_LSB__TIA2_C_LSB */
	{479, 0, 8},  /*!< TIA2_C_MSB */
	{479, 0, 7},  /*!< TIA2_C_MSB__TIA2_C_MSB */
	{480, 0, 8},  /*!< RX1_BBF_R1A */
	{480, 7, 1},  /*!< RX1_BBF_R1A__FORCE_RX1_RESISTORS */
	{480, 0, 6},  /*!< RX1_BBF_R1A__RX1_BBF_R1A */
	{481, 0, 8},  /*!< RX2_BBF_R1A */
	{481, 7, 1},  /*!< RX2_BBF_R1A__FORCE_RX2_RESISTORS */
	{481, 0, 6},  /*!< RX2_BBF_R1A__RX2_BBF_R1A */
	{482, 0, 8},  /*!< RX1_TUNE_CTRL */
	{482, 2, 1},  /*!< RX1_TUNE_CTRL__RX1_TUNE_RESAMPLE_PHASE */
	{482, 1, 1},  /*!< RX1_TUNE_CTRL__RX1_TUNE_RESAMPLE */
	{482, 0, 1},  /*!< RX1_TUNE_CTRL__RX1_PD_TUNE */
	{483, 0, 8},  /*!< RX2_TUNE_CTRL */
	{483, 2, 1},  /*!< RX2_TUNE_CTRL__RX2_TUNE_RESAMPLE_PHASE */
	{483, 1, 1},  /*!< RX2_TUNE_CTRL__RX2_TUNE_RESAMPLE */
	{483, 0, 1},  /*!< RX2_TUNE_CTRL__RX2_PD_TUNE */
	{484, 0, 8},  /*!< RX1_BBF_R5 */
	{485, 0, 8},  /*!< RX2_BBF_R5 */
	{486, 0, 8},  /*!< RX_BBF_R2346 */
	{486, 7, 1},  /*!< RX_BBF_R2346__TUNE_OVERRIDE */
	{486, 0, 3},  /*!< RX_BBF_R2346__RX_BBF_R2346 */
	{487, 0, 8},  /*!< RX_BBF_C1_MSB */
	{487, 0, 6},  /*!< RX_BBF_C1_MSB__RX_BBF_C1_MSB */
	{488, 0, 8},  /*!< RX_BBF_C1_LSB */
	{488, 0, 7},  /*!< RX_BBF_C1_LSB__RX_BBF_C1_LSB */
	{489, 0, 8},  /*!< RX_BBF_C2_MSB */
	{489, 0, 6},  /*!< RX_BBF_C2_MSB__RX_BBF_C2_MSB */
	{490, 0, 8},  /*!< RX_BBF_C2_LSB */
	{490, 0, 7},  /*!< RX_BBF_C2_LSB__RX_BBF_C2_LSB */
	{491, 0, 8},  /*!< RX_BBF_C3_MSB */
	{491, 0, 6},  /*!< RX_BBF_C3_MSB__RX_BBF_C3_MSB */
	{492, 0, 8},  /*!< RX_BBF_C3_LSB */
	{492, 0, 7},  /*!< RX_BBF_C3_LSB__RX_BBF_C3_LSB */
	{493, 0, 8},  /*!< RX_BBF_CC1_CTR */
	{493, 0, 7},  /*!< RX_BBF_CC1_CTR__RX_BBF_CC1_CTR */
	{494, 0, 8},  /*!< RX_BBF_POW_RZ_BYTE0 */
	{494, 7, 1},  /*!< RX_BBF_POW_RZ_BYTE0__MUST_BE_ZERO */
	{494, 5, 2},  /*!< RX_BBF_POW_RZ_BYTE0__RX1_BBF_POW_CTR */
	{494, 3, 2},  /*!< RX_BBF_POW_RZ_BYTE0__RX_BBF_RZ1_CTR */
	{495, 0, 8},  /*!< RX_BBF_CC2_CTR */
	{495, 0, 7},  /*!< RX_BBF_CC2_CTR__RX_BBF_CC2_CTR */
	{496, 0, 8},  /*!< RX_BBF_POW_RZ_BYTE1 */
	{496, 6, 2},  /*!< RX_BBF_POW_RZ_BYTE1__RX_BBF_POW3_CTR */
	{496, 4, 2},  /*!< RX_BBF_POW_RZ_BYTE1__RX_BBF_RZ3_CTR */
	{496, 2, 2},  /*!< RX_BBF_POW_RZ_BYTE1__RX_BBF_POW2_CTR */
	{496, 0, 2},  /*!< RX_BBF_POW_RZ_BYTE1__RX_BBF_RZ2_CTR */
	{497, 0, 8},  /*!< RX_BBF_CC3_CTR */
	{497, 0, 7},  /*!< RX_BBF_CC3_CTR__RX_BBF_CC3_CTR */
	{498, 0, 8},  /*!< RX_BBF_R5_TUNE */
	{499, 0, 8},  /*!< RX_BBF_TUNE */
	{499, 7, 1},  /*!< RX_BBF_TUNE__RXBBF_BYPASS_BIAS_R */
	{499, 4, 1},  /*!< RX_BBF_TUNE__RX_BBF_R5_TUNE */
	{499, 3, 1},  /*!< RX_BBF_TUNE__RX1_BBF_TUNE_COMP_I */
	{499, 2, 1},  /*!< RX_BBF_TUNE__RX1_BBF_TUNE_COMP_Q */
	{499, 1, 1},  /*!< RX_BBF_TUNE__RX2_BBF_TUNE_COMP_I */
	{499, 0, 1},  /*!< RX_BBF_TUNE__RX2_BBF_TUNE_COMP_Q */
	{499, 5, 2},  /*!< RX_BBF_TUNE__RX_BBF_TUNE_CTR */
	{500, 0, 8},  /*!< RX1_BBF_MAN_GAIN */
	{500, 5, 1},  /*!< RX1_BBF_MAN_GAIN__RX1_BBF_FORCE_GAIN */
	{500, 3, 2},  /*!< RX1_BBF_MAN_GAIN__RX1_BBF_BQ_GAIN */
	{500, 0, 3},  /*!< RX1_BBF_MAN_GAIN__RX1_BBF_POLE_GAIN */
	{501, 0, 8},  /*!< RX2_BBF_MAN_GAIN */
	{501, 5, 1},  /*!< RX2_BBF_MAN_GAIN__RX2_BBF_FORCE_GAIN */
	{501, 3, 2},  /*!< RX2_BBF_MAN_GAIN__RX2_BBF_BQ_GAIN */
	{501, 0, 3},  /*!< RX2_BBF_MAN_GAIN__RX2_BBF_POLE_GAIN */
	{504, 0, 8},  /*!< RX_BBF_TUNE_DIVIDE */
	{505, 0, 8},  /*!< RX_BBF_TUNE_CONFIG */
	{505, 4, 1},  /*!< RX_BBF_TUNE_CONFIG__RX_TUNE_EVALTIME */
	{505, 0, 1},  /*!< RX_BBF_TUNE_CONFIG__RX_BBF_TUNE_DIVIDE */
	{505, 5, 2},  /*!< RX_BBF_TUNE_CONFIG__TUNE_COMP_MASK */
	{505, 1, 3},  /*!< RX_BBF_TUNE_CONFIG__RX_TUNE_MODE */
	{506, 0, 8},  /*!< POLE_GAIN */
	{506, 0, 2},  /*!< POLE_GAIN__POLE_GAIN_TUNE */
	{507, 0, 8},  /*!< RX_BBBW_MHZ */
	{507, 0, 5},  /*!< RX_BBBW_MHZ__RX_TUNE_BBBW_MHZ */
	{508, 0, 8},  /*!< RX_BBBW_KHZ */
	{508, 0, 7},  /*!< RX_BBBW_KHZ__RX_TUNE_BBBW_KHZ */
	{513, 0, 8},  /*!< FB_DAC_CLK_DELAY1 */
	{514, 0, 8},  /*!< FB_DAC_CLK_DELAY2 */
	{515, 0, 8},  /*!< FLASH_SAMPLE_CLK_DELAY_3P */
	{516, 0, 8},  /*!< FLASH_SAMPLE_CLK_DELAY_3N */
	{517, 0, 8},  /*!< TEST_MUX_2I */
	{518, 0, 8},  /*!< TEST_MUX_2Q */
	{519, 0, 8},  /*!< INTEGRATOR_1_RESISTANCE */
	{520, 0, 8},  /*!< INTEGRATOR_1_CAPACITANCE */
	{521, 0, 8},  /*!< INTEGRATOR_23_RESISTANCE */
	{522, 0, 8},  /*!< INTEGRATOR_2_RESISTANCE */
	{523, 0, 8},  /*!< INTEGRATOR_2_CAPACITANCE */
	{524, 0, 8},  /*!< INTEGRATOR_3_RESISTANCE */
	{525, 0, 8},  /*!< INTEGRATOR_3_CAPACITANCE */
	{526, 0, 8},  /*!< INTEGRATOR_AMP_CC */
	{527, 0, 8},  /*!< INT_1_FB_DAC_NMOS_CURRENT_SOURCE */
	{528, 0, 8},  /*!< INT_1_FB_DAC_NMOS_CASOADE_BIAS_CURRENT */
	{529, 0, 8},  /*!< INT_1_FB_DAC_PMOS_CURRENT_SOURCE */
	{530, 0, 8},  /*!< INT_2_FB_DAC_NMOS_CURRENT_SOURCE */
	{531, 0, 8},  /*!< INT_2_FB_DAC_NMOS_CASCODE_BIAS_CURRENT */
	{532, 0, 8},  /*!< INT_2_FB_DAC_PMOS_CURRENT_SOURCE */
	{533, 0, 8},  /*!< INT_3_FB_DAC_NMOS_CURRENT_SOURCE */
	{534, 0, 8},  /*!< INT_3_FB_DAC_NMOS_CASCODE_BIAS_CURRENT */
	{535, 0, 8},  /*!< INT_3_FB_DAC_PMOS_CURRENT_SOURCE */
	{536, 0, 8},  /*!< FB_DAC_BIAS_CURRENT */
	{537, 0, 8},  /*!< INT_1_1ST_STAGE_CURRENT */
	{538, 0, 8},  /*!< INT_1_1ST_STAGE_CASCODE_CURRENT */
	{539, 0, 8},  /*!< INT_1_2ND_STAGE_CURRENT */
	{540, 0, 8},  /*!< INTEGRATOR_2_1ST_STAGE_CURRENT */
	{541, 0, 8},  /*!< INT_2_1ST_STAGE_CASCODE_CURRENT */
	{542, 0, 8},  /*!< INT_2_2ND_STAGE_CURRENT */
	{543, 0, 8},  /*!< INT_3_1ST_STAGE_CURRENT */
	{544, 0, 8},  /*!< INT_3_1ST_STAGE_CASCODE_CURRENT */
	{545, 0, 8},  /*!< INT_3_2ND_STAGE_CURRENT */
	{546, 0, 8},  /*!< FLASH_BIAS_CURRENT */
	{547, 0, 8},  /*!< FLASH_LADDER_BIAS */
	{548, 0, 8},  /*!< FLASH_LADDER_CASCODE_CURRENT */
	{549, 0, 8},  /*!< FLASH_LADDER_BIAS2 */
	{550, 0, 8},  /*!< RESET */
	{560, 0, 8},  /*!< RX_PFD_CONFIG */
	{560, 0, 1},  /*!< RX_PFD_CONFIG__BYPASS_LD_SYNTH */
	{561, 0, 8},  /*!< RX_INTEGER_BYTE_0 */
	{562, 0, 8},  /*!< RX_INTEGER_BYTE_1 */
	{562, 0, 3},  /*!< RX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD */
	{563, 0, 8},  /*!< RX_FRACT_BYTE_0 */
	{564, 0, 8},  /*!< RX_FRACT_BYTE_1 */
	{565, 0, 8},  /*!< RX_FRACT_BYTE_2 */
	{565, 0, 7},  /*!< RX_FRACT_BYTE_2__SYNTH_FRACT_WORD */
	{566, 0, 8},  /*!< RX_FORCE_ALC */
	{567, 0, 8},  /*!< RX_FORCE_VCO_TUNE_0 */
	{568, 0, 8},  /*!< RX_FORCE_VCO_TUNE_1 */
	{568, 3, 4},  /*!< RX_FORCE_VCO_TUNE_1__VCO_CAL_OFFSET */
	{569, 0, 8},  /*!< RX_ALC_VARACTOR */
	{569, 4, 4},  /*!< RX_ALC_VARACTOR__INIT_ALC_VALUE */
	{569, 0, 4},  /*!< RX_ALC_VARACTOR__VCO_VARACTOR */
	{570, 0, 8},  /*!< RX_VCO_OUTPUT */
	{570, 6, 1},  /*!< RX_VCO_OUTPUT__PORB_VCO_LOGIC */
	{570, 0, 4},  /*!< RX_VCO_OUTPUT__VCO_OUTPUT_LEVEL */
	{571, 0, 8},  /*!< RX_CP_CURRENT */
	{571, 0, 6},  /*!< RX_CP_CURRENT__CHARGE_PUMP_CURRENT */
	{572, 0, 8},  /*!< RX_CP_OFFSET */
	{572, 7, 1},  /*!< RX_CP_OFFSET__SYNTH_RECAL */
	{573, 0, 8},  /*!< RX_CP_CONFIG */
	{573, 7, 1},  /*!< RX_CP_CONFIG__HALF_VCO_CAL_CLK */
	{573, 3, 1},  /*!< RX_CP_CONFIG__F_CPCAL */
	{573, 2, 1},  /*!< RX_CP_CONFIG__CP_CAL_ENABLE */
	{574, 0, 8},  /*!< RX_LOOP_FILTER_1 */
	{574, 4, 4},  /*!< RX_LOOP_FILTER_1__LOOP_FILTER_C2 */
	{574, 0, 4},  /*!< RX_LOOP_FILTER_1__LOOP_FILTER_C1 */
	{575, 0, 8},  /*!< RX_LOOP_FILTER_2 */
	{575, 4, 4},  /*!< RX_LOOP_FILTER_2__LOOP_FILTER_R1 */
	{575, 0, 4},  /*!< RX_LOOP_FILTER_2__LOOP_FILTER_C3 */
	{576, 0, 8},  /*!< RX_LOOP_FILTER_3 */
	{576, 7, 1},  /*!< RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R3 */
	{576, 6, 1},  /*!< RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R1 */
	{576, 5, 1},  /*!< RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C2 */
	{576, 4, 1},  /*!< RX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C1 */
	{576, 0, 4},  /*!< RX_LOOP_FILTER_3__LOOP_FILTER_R3 */
	{577, 0, 8},  /*!< RX_DITHERCP_CAL */
	{577, 0, 4},  /*!< RX_DITHERCP_CAL__FORCED_CP_CAL_WORD */
	{578, 0, 8},  /*!< RX_VCO_BIAS_1 */
	{578, 3, 2},  /*!< RX_VCO_BIAS_1__VCO_BIAS_TCF */
	{578, 0, 3},  /*!< RX_VCO_BIAS_1__VCO_BIAS_REF */
	{580, 0, 8},  /*!< RX_CAL_STATUS */
	{580, 7, 1},  /*!< RX_CAL_STATUS__CP_CAL_VALID */
	{580, 5, 1},  /*!< RX_CAL_STATUS__CP_CAL_DONE */
	{580, 4, 1},  /*!< RX_CAL_STATUS__VCO_CAL_BUSY */
	{580, 0, 4},  /*!< RX_CAL_STATUS__CP_CAL_WORD */
	{581, 0, 8},  /*!< RX_VCO_CAL_REF */
	{581, 0, 3},  /*!< RX_VCO_CAL_REF__VCO_CAL_REF_TCF */
	{582, 0, 8},  /*!< RX_VCO_PD_OVERRIDES */
	{582, 3, 1},  /*!< RX_VCO_PD_OVERRIDES__POWER_DOWN_VARACTOR_REF */
	{582, 2, 1},  /*!< RX_VCO_PD_OVERRIDES__PWR_DOWN_VARACT_REF_TCF */
	{582, 1, 1},  /*!< RX_VCO_PD_OVERRIDES__POWER_DOWN_CAL_TCF */
	{582, 0, 1},  /*!< RX_VCO_PD_OVERRIDES__POWER_DOWN_VCO_BUFFFER */
	{583, 0, 8},  /*!< RX_CP_OVERRANGE_VCO_LOCK */
	{583, 7, 1},  /*!< RX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_HIGH */
	{583, 6, 1},  /*!< RX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_LOW */
	{583, 1, 1},  /*!< RX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK */
	{584, 0, 8},  /*!< RX_VCO_LDO */
	{584, 7, 1},  /*!< RX_VCO_LDO__VCO_LDO_BYPASS */
	{584, 5, 2},  /*!< RX_VCO_LDO__VCO_LDO_INRUSH */
	{584, 2, 3},  /*!< RX_VCO_LDO__VCO_LDO_SEL */
	{584, 0, 2},  /*!< RX_VCO_LDO__VCO_LDO_VDROP_SEL */
	{585, 0, 8},  /*!< RX_VCO_CAL */
	{585, 7, 1},  /*!< RX_VCO_CAL__VCO_CAL_EN */
	{585, 4, 3},  /*!< RX_VCO_CAL__VCO_CAL_ALC_WAIT */
	{585, 2, 2},  /*!< RX_VCO_CAL__VCO_CAL_COUNT */
	{586, 0, 8},  /*!< RX_LOCK_DETECT_CONFIG */
	{586, 2, 2},  /*!< RX_LOCK_DETECT_CONFIG__LOCK_DETECT_COUNT */
	{586, 0, 2},  /*!< RX_LOCK_DETECT_CONFIG__LOCK_DETECT_MODE */
	{587, 0, 8},  /*!< RX_CP_LEVEL_DETECT */
	{587, 6, 1},  /*!< RX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_POWER_DOWN */
	{587, 3, 3},  /*!< RX_CP_LEVEL_DETECT__CP_LEVEL_THRESH_LOW */
	{587, 0, 3},  /*!< RX_CP_LEVEL_DETECT__CP_LEVEL_THRESH_HIGH */
	{588, 0, 8},  /*!< RX_DSM_SETUP_0 */
	{588, 0, 4},  /*!< RX_DSM_SETUP_0__DSM_PROG */
	{589, 0, 8},  /*!< RX_DSM_SETUP_1 */
	{589, 6, 1},  /*!< RX_DSM_SETUP_1__SIF_CLOCK */
	{589, 5, 1},  /*!< RX_DSM_SETUP_1__SIF_RESET_BAR */
	{589, 0, 5},  /*!< RX_DSM_SETUP_1__SIF_ADDR */
	{590, 0, 8},  /*!< RX_CORRECTION_WORD0 */
	{590, 7, 1},  /*!< RX_CORRECTION_WORD0__UPDATE_FREQ_WORD */
	{590, 5, 1},  /*!< RX_CORRECTION_WORD0__READ_EFFECTIVE_TUNING_WORD */
	{590, 0, 5},  /*!< RX_CORRECTION_WORD0__FREQ_CORRECTION_WORD_MSB */
	{591, 0, 8},  /*!< RX_CORRECTION_WORD1 */
	{591, 7, 1},  /*!< RX_CORRECTION_WORD1__UPDATE_FREQ_WORD */
	{591, 0, 7},  /*!< RX_CORRECTION_WORD1__FREQ_CORRECTION_WORD_LSB */
	{592, 0, 8},  /*!< RX_VCO_VARACTOR_CTRL_0 */
	{592, 4, 3},  /*!< RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_REFERENCE_TCF */
	{592, 0, 4},  /*!< RX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_OFFSET */
	{593, 0, 8},  /*!< RX_VCO_VARACTOR_CTRL_1 */
	{593, 0, 4},  /*!< RX_VCO_VARACTOR_CTRL_1__VCO_VARACTOR_REFERENCE */
	{602, 0, 8},  /*!< RX_FAST_LOCK_SETUP */
	{602, 3, 1},  /*!< RX_FAST_LOCK_SETUP__RX_FAST_LOCK_LOAD_SYNTH */
	{602, 2, 1},  /*!< RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE_INIT */
	{602, 1, 1},  /*!< RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE_PIN_SELECT */
	{602, 0, 1},  /*!< RX_FAST_LOCK_SETUP__RX_FAST_LOCK_MODE_ENABLE */
	{602, 5, 3},  /*!< RX_FAST_LOCK_SETUP__RX_FAST_LOCK_PROFILE */
	{603, 0, 8},  /*!< RX_FAST_LOCK_SETUP_INIT_DELAY */
	{604, 0, 8},  /*!< RX_FAST_LOCK_PROGRAM_ADDR */
	{604, 4, 3},  /*!< RX_FAST_LOCK_PROGRAM_ADDR__RX_FAST_LOCK_PROFILE_ADDR */
	{604, 0, 4},  /*!< RX_FAST_LOCK_PROGRAM_ADDR__RX_FAST_LOCK_PROFILE_WORD */
	{605, 0, 8},  /*!< RX_FAST_LOCK_PROGRAM_DATA */
	{606, 0, 8},  /*!< RX_FAST_LOCK_PROGRAM_READ */
	{607, 0, 8},  /*!< RX_FAST_LOCK_PROGRAM_CTRL */
	{607, 1, 1},  /*!< RX_FAST_LOCK_PROGRAM_CTRL__RX_FAST_LOCK_PROGRAM_WRITE */
	{607, 0, 1},  /*!< RX_FAST_LOCK_PROGRAM_CTRL__RX_FAST_LOCK_PROGRAM_CLOCK_ENABLE */
	{609, 0, 8},  /*!< RX_LO_GEN_POWER_MODE */
	{609, 4, 2},  /*!< RX_LO_GEN_POWER_MODE__RX_LO_GEN_POWER_MODE */
	{624, 0, 8},  /*!< TX_PFD_CONFIG */
	{624, 5, 1},  /*!< TX_PFD_CONFIG__DIV_TEST_EN */
	{624, 1, 1},  /*!< TX_PFD_CONFIG__PFD_CLK_EDGE */
	{624, 0, 1},  /*!< TX_PFD_CONFIG__BYPASS_LD_SYNTH */
	{624, 2, 2},  /*!< TX_PFD_CONFIG__PFD_WIDTH */
	{625, 0, 8},  /*!< TX_INTEGER_BYTE_0 */
	{626, 0, 8},  /*!< TX_INTEGER_BYTE_1 */
	{626, 7, 1},  /*!< TX_INTEGER_BYTE_1__SDM_BYPASS */
	{626, 6, 1},  /*!< TX_INTEGER_BYTE_1__SDM_POWER_DOWN */
	{626, 0, 3},  /*!< TX_INTEGER_BYTE_1__SYNTH_INTEGER_WORD */
	{627, 0, 8},  /*!< TX_FRACT_BYTE_0 */
	{628, 0, 8},  /*!< TX_FRACT_BYTE_1 */
	{629, 0, 8},  /*!< TX_FRACT_BYTE_2 */
	{629, 0, 7},  /*!< TX_FRACT_BYTE_2__SYNTH_FRACT_WORD */
	{630, 0, 8},  /*!< TX_FORCE_ALC */
	{630, 7, 1},  /*!< TX_FORCE_ALC__FORCE_ALC_ENABLE */
	{630, 0, 7},  /*!< TX_FORCE_ALC__FORCE_ALC_WORD */
	{631, 0, 8},  /*!< TX_FORCE_VCO_TUNE_0 */
	{632, 0, 8},  /*!< TX_FORCE_VCO_TUNE_1 */
	{632, 7, 1},  /*!< TX_FORCE_VCO_TUNE_1__BYPASS_LOAD_DELAY */
	{632, 1, 1},  /*!< TX_FORCE_VCO_TUNE_1__FORCE_VCO_TUNE_ENABLE */
	{632, 0, 1},  /*!< TX_FORCE_VCO_TUNE_1__FORCE_VCO_TUNE */
	{632, 3, 4},  /*!< TX_FORCE_VCO_TUNE_1__VCO_CAL_OFFSET */
	{633, 0, 8},  /*!< TX_ALCVARACT_OR */
	{633, 4, 4},  /*!< TX_ALCVARACT_OR__INIT_ALC_VALUE */
	{633, 0, 4},  /*!< TX_ALCVARACT_OR__VCO_VARACTOR */
	{634, 0, 8},  /*!< TX_VCO_OUTPUT */
	{634, 6, 1},  /*!< TX_VCO_OUTPUT__PORB_VCO_LOGIC */
	{634, 0, 4},  /*!< TX_VCO_OUTPUT__VCO_OUTPUT_LEVEL */
	{635, 0, 8},  /*!< TX_CP_CURRENT */
	{635, 7, 1},  /*!< TX_CP_CURRENT__TX_CP_CURRENT_DFLT */
	{635, 6, 1},  /*!< TX_CP_CURRENT__VTUNE_FORCE */
	{635, 0, 6},  /*!< TX_CP_CURRENT__CHARGE_PUMP_CURRENT */
	{636, 0, 8},  /*!< TX_CP_OFFSET */
	{636, 7, 1},  /*!< TX_CP_OFFSET__SYNTH_RECAL */
	{636, 0, 6},  /*!< TX_CP_OFFSET__CHARGE_PUMP_OFFSET */
	{637, 0, 8},  /*!< TX_CP_CONFIG */
	{637, 7, 1},  /*!< TX_CP_CONFIG__HALF_VCO_CAL_CLK */
	{637, 6, 1},  /*!< TX_CP_CONFIG__DITHER_MODE */
	{637, 4, 1},  /*!< TX_CP_CONFIG__CP_OFFSET_OFF */
	{637, 3, 1},  /*!< TX_CP_CONFIG__F_CPCAL */
	{637, 2, 1},  /*!< TX_CP_CONFIG__CP_CAL_ENABLE */
	{637, 0, 2},  /*!< TX_CP_CONFIG__CP_TEST */
	{638, 0, 8},  /*!< TX_LOOP_FILTER_1 */
	{638, 4, 4},  /*!< TX_LOOP_FILTER_1__LOOP_FILTER_C2 */
	{638, 0, 4},  /*!< TX_LOOP_FILTER_1__LOOP_FILTER_C1 */
	{639, 0, 8},  /*!< TX_LOOP_FILTER_2 */
	{639, 4, 4},  /*!< TX_LOOP_FILTER_2__LOOP_FILTER_R1 */
	{639, 0, 4},  /*!< TX_LOOP_FILTER_2__LOOP_FILTER_C3 */
	{640, 0, 8},  /*!< TX_LOOP_FILTER_3 */
	{640, 7, 1},  /*!< TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R3 */
	{640, 6, 1},  /*!< TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_R1 */
	{640, 5, 1},  /*!< TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C2 */
	{640, 4, 1},  /*!< TX_LOOP_FILTER_3__LOOP_FILTER_BYPASS_C1 */
	{640, 0, 4},  /*!< TX_LOOP_FILTER_3__LOOP_FILTER_R3 */
	{641, 0, 8},  /*!< TX_DITHERCP_CAL */
	{641, 4, 4},  /*!< TX_DITHERCP_CAL__NUMBER_SDM_DITHER_BITS */
	{641, 0, 4},  /*!< TX_DITHERCP_CAL__FORCED_CP_CAL_WORD */
	{642, 0, 8},  /*!< TX_VCO_BIAS_1 */
	{642, 5, 2},  /*!< TX_VCO_BIAS_1__MUST_BE_ZEROS */
	{642, 3, 2},  /*!< TX_VCO_BIAS_1__VCO_BIAS_TCF */
	{642, 0, 3},  /*!< TX_VCO_BIAS_1__VCO_BIAS_REF */
	{643, 0, 8},  /*!< TX_VCO_BIAS_2 */
	{643, 7, 1},  /*!< TX_VCO_BIAS_2__VCO_BYPASS_BIAS_DAC_R */
	{643, 4, 1},  /*!< TX_VCO_BIAS_2__VCO_COMP_BYPASS_BIAS_R */
	{643, 3, 1},  /*!< TX_VCO_BIAS_2__BYPASS_PRESCALE_R */
	{643, 2, 1},  /*!< TX_VCO_BIAS_2__LAST_ALC_ENABLE */
	{643, 0, 2},  /*!< TX_VCO_BIAS_2__PRESCALE_BIAS */
	{644, 0, 8},  /*!< TX_CAL_STATUS */
	{644, 7, 1},  /*!< TX_CAL_STATUS__CP_CAL_VALID */
	{644, 6, 1},  /*!< TX_CAL_STATUS__COMP_OUT */
	{644, 5, 1},  /*!< TX_CAL_STATUS__CP_CAL_DONE */
	{644, 4, 1},  /*!< TX_CAL_STATUS__VCO_CAL_BUSY */
	{644, 0, 4},  /*!< TX_CAL_STATUS__CP_CAL_WORD */
	{645, 0, 8},  /*!< TX_VCO_CAL_REF */
	{645, 3, 1},  /*!< TX_VCO_CAL_REF__VCO_CAL_REF_MONITOR */
	{645, 0, 3},  /*!< TX_VCO_CAL_REF__VCO_CAL_REF_TCF */
	{646, 0, 8},  /*!< TX_VCO_PD_OVERRIDES */
	{646, 3, 1},  /*!< TX_VCO_PD_OVERRIDES__POWER_DOWN_VARACTOR_REF */
	{646, 2, 1},  /*!< TX_VCO_PD_OVERRIDES__POWER_DOWN_VARACT_REF_TCF */
	{646, 1, 1},  /*!< TX_VCO_PD_OVERRIDES__POWER_DOWN_CAL_TCF */
	{646, 0, 1},  /*!< TX_VCO_PD_OVERRIDES__POWER_DOWN_VCO_BUFFFER */
	{647, 0, 8},  /*!< TX_CP_OVERRANGE_VCO_LOCK */
	{647, 7, 1},  /*!< TX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_HIGH */
	{647, 6, 1},  /*!< TX_CP_OVERRANGE_VCO_LOCK__CP_OVRG_LOW */
	{647, 1, 1},  /*!< TX_CP_OVERRANGE_VCO_LOCK__VCO_LOCK */
	{648, 0, 8},  /*!< TX_VCO_LDO */
	{648, 7, 1},  /*!< TX_VCO_LDO__VCO_LDO_BYPASS */
	{648, 5, 2},  /*!< TX_VCO_LDO__VCO_LDO_INRUSH */
	{648, 2, 3},  /*!< TX_VCO_LDO__VCO_LDO_VOUT_SEL */
	{648, 0, 2},  /*!< TX_VCO_LDO__VCO_LDO_VDROP_SEL */
	{649, 0, 8},  /*!< TX_VCO_CAL */
	{649, 7, 1},  /*!< TX_VCO_CAL__VCO_CAL_EN */
	{649, 4, 3},  /*!< TX_VCO_CAL__VCO_CAL_ALC_WAIT */
	{649, 2, 2},  /*!< TX_VCO_CAL__VCO_CAL_COUNT */
	{649, 0, 2},  /*!< TX_VCO_CAL__FB_CLOCK_ADV */
	{650, 0, 8},  /*!< TX_LOCK_DETECT_CONFIG */
	{650, 2, 2},  /*!< TX_LOCK_DETECT_CONFIG__LOCK_DETECT_COUNT */
	{650, 0, 2},  /*!< TX_LOCK_DETECT_CONFIG__LOCK_DETECT_MODE */
	{651, 0, 8},  /*!< TX_CP_LEVEL_DETECT */
	{651, 6, 1},  /*!< TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_POWER_DOWN */
	{651, 3, 3},  /*!< TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_THRESH_LOW */
	{651, 0, 3},  /*!< TX_CP_LEVEL_DETECT__CP_LEVEL_DETECT_THRESH_HIGH */
	{652, 0, 8},  /*!< TX_DSM_SETUP_0 */
	{652, 0, 4},  /*!< TX_DSM_SETUP_0__DSM_PROG */
	{653, 0, 8},  /*!< TX_DSM_SETUP_1 */
	{653, 6, 1},  /*!< TX_DSM_SETUP_1__SIF_CLOCK */
	{653, 5, 1},  /*!< TX_DSM_SETUP_1__SIF_RESET_BAR */
	{653, 0, 5},  /*!< TX_DSM_SETUP_1__SIF_ADDR */
	{654, 0, 8},  /*!< TX_CORRECTION_WORD0 */
	{654, 7, 1},  /*!< TX_CORRECTION_WORD0__UPDATE_FREQ_WORD */
	{654, 5, 1},  /*!< TX_CORRECTION_WORD0__READ_EFFECTIVE_TUNING_WORD */
	{654, 0, 5},  /*!< TX_CORRECTION_WORD0__FREQ_CORRECTION_WORD_MSB */
	{655, 0, 8},  /*!< TX_CORRECTION_WORD1 */
	{655, 7, 1},  /*!< TX_CORRECTION_WORD1__UPDATE_FREQ_WORD */
	{655, 0, 7},  /*!< TX_CORRECTION_WORD1__FREQ_CORRECTION_WORD_LSB */
	{656, 0, 8},  /*!< TX_VCO_VARACTOR_CTRL_0 */
	{656, 4, 3},  /*!< TX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_REFERENCE_TCF */
	{656, 0, 4},  /*!< TX_VCO_VARACTOR_CTRL_0__VCO_VARACTOR_OFFSET */
	{657, 0, 8},  /*!< TX_VCO_VARACTOR_CTRL_1 */
	{657, 0, 4},  /*!< TX_VCO_VARACTOR_CTRL_1__VCO_VARACTOR_REFERENCE */
	{658, 0, 8},  /*!< DCXO_COARSE_TUNE */
	{658, 0, 6},  /*!< DCXO_COARSE_TUNE__DCXO_TUNE_COARSE */
	{659, 0, 8},  /*!< DCXO_FINE_TUNE_HIGH */
	{660, 0, 8},  /*!< DCXO_FINE_TUNE_LOW */
	{660, 3, 5},  /*!< DCXO_FINE_TUNE_LOW__DCXO_TUNE_FINE_LOW */
	{661, 0, 8},  /*!< DCXO_CONFIG */
	{661, 7, 1},  /*!< DCXO_CONFIG__MUST_BE_ZERO */
	{661, 4, 3},  /*!< DCXO_CONFIG__DCXO_RTAIL */
	{661, 2, 2},  /*!< DCXO_CONFIG__DCXO_RD */
	{662, 0, 8},  /*!< DCXO_TEMPCO_WRITE */
	{663, 0, 8},  /*!< DCXO_TEMPCO_READ */
	{664, 0, 8},  /*!< DCXO_TEMPCO_ADDR */
	{664, 7, 1},  /*!< DCXO_TEMPCO_ADDR__DCXO_TEMPCO_EN */
	{664, 6, 1},  /*!< DCXO_TEMPCO_ADDR__DCXO_TEMPCO_CLK */
	{664, 0, 6},  /*!< DCXO_TEMPCO_ADDR__DCXO_TEMPERATURE_COEF_ADDRESS */
	{665, 0, 8},  /*!< DELTA_T_READ */
	{666, 0, 8},  /*!< TX_FAST_LOCK_SETUP */
	{666, 3, 1},  /*!< TX_FAST_LOCK_SETUP__TX_FAST_LOCK_LOAD_SYNTH */
	{666, 2, 1},  /*!< TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE_INIT */
	{666, 1, 1},  /*!< TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE_PIN_SELECT */
	{666, 0, 1},  /*!< TX_FAST_LOCK_SETUP__TX_FAST_LOCK_MODE_ENABLE */
	{666, 5, 3},  /*!< TX_FAST_LOCK_SETUP__TX_FAST_LOCK_PROFILE */
	{667, 0, 8},  /*!< TX_FAST_LOCK_SETUP_INIT_DELAY */
	{668, 0, 8},  /*!< TX_FAST_LOCK_PROGRAM_ADDR */
	{669, 0, 8},  /*!< TX_FAST_LOCK_PROGRAM_DATA */
	{670, 0, 8},  /*!< TX_FAST_LOCK_PROGRAM_READ */
	{671, 0, 8},  /*!< TX_FAST_LOCK_PROGRAM_CTRL */
	{671, 1, 1},  /*!< TX_FAST_LOCK_PROGRAM_CTRL__TX_FAST_LOCK_PROGRAM_WRITE */
	{671, 0, 1},  /*!< TX_FAST_LOCK_PROGRAM_CTRL__TX_FAST_LOCK_PROGRAM_CLOCK_ENABLE */
	{673, 0, 8},  /*!< TX_LO_GEN_POWER_MODE */
	{673, 4, 4},  /*!< TX_LO_GEN_POWER_MODE__TX_LO_GEN_POWER_MODE */
	{678, 0, 8},  /*!< BANDGAP_CONFIG0 */
	{678, 7, 1},  /*!< BANDGAP_CONFIG0__POWER_DOWN_BANDGAP_REF */
	{678, 6, 1},  /*!< BANDGAP_CONFIG0__MASTER_BIAS_FILTER_BYPASS */
	{678, 5, 1},  /*!< BANDGAP_CONFIG0__MASTER_BIAS_REF_SEL */
	{678, 0, 5},  /*!< BANDGAP_CONFIG0__MASTER_BIAS_TRIM */
	{680, 0, 8},  /*!< BANDGAP_CONFIG1 */
	{680, 7, 1},  /*!< BANDGAP_CONFIG1__VCO_LDO_FILTER_BYPASS */
	{680, 6, 1},  /*!< BANDGAP_CONFIG1__VCO_LDO_REF_SEL */
	{680, 5, 1},  /*!< BANDGAP_CONFIG1__BANDGAP_REF_RESET */
	{680, 0, 5},  /*!< BANDGAP_CONFIG1__BANDGAP_TEMP_TRIM */
	{683, 0, 8},  /*!< REF_DIVIDE_CONFIG_1 */
	{683, 2, 1},  /*!< REF_DIVIDE_CONFIG_1__REF_DIVIDE_CONFIG_1_DFLT */
	{683, 1, 1},  /*!< REF_DIVIDE_CONFIG_1__RX_REF_RESET_BAR */
	{683, 0, 1},  /*!< REF_DIVIDE_CONFIG_1__RX_REF_DIVIDER_MSB */
	{684, 0, 8},  /*!< REF_DIVIDE_CONFIG_2 */
	{684, 7, 1},  /*!< REF_DIVIDE_CONFIG_2__RX_REF_DIVIDER_LSB */
	{684, 4, 1},  /*!< REF_DIVIDE_CONFIG_2__TX_REF_RESET_BAR */
	{684, 5, 2},  /*!< REF_DIVIDE_CONFIG_2__RX_REF_DOUBLER_FB_DELAY */
	{684, 2, 2},  /*!< REF_DIVIDE_CONFIG_2__TX_REF_DIVIDER */
	{684, 0, 2},  /*!< REF_DIVIDE_CONFIG_2__TX_REF_DOUBLER_FB_DELAY */
	{688, 0, 8},  /*!< GAIN_RX1 */
	{688, 0, 7},  /*!< GAIN_RX1__FULL_TABLE_GAIN_INDEX */
	{689, 0, 8},  /*!< LPF_GAIN_RX1 */
	{689, 0, 5},  /*!< LPF_GAIN_RX1__LPF_GAIN_RX */
	{690, 0, 8},  /*!< DIG_GAIN_RX1 */
	{690, 0, 5},  /*!< DIG_GAIN_RX1__DIGITAL_GAIN_RX */
	{691, 0, 8},  /*!< FAST_ATTACK_STATE */
	{691, 4, 3},  /*!< FAST_ATTACK_STATE__FAST_ATTACK_STATE_RX2 */
	{691, 0, 3},  /*!< FAST_ATTACK_STATE__FAST_ATTACK_STATE_RX1 */
	{692, 0, 8},  /*!< SLOW_LOOP_STATE */
	{692, 4, 3},  /*!< SLOW_LOOP_STATE__SLOW_LOOP_STATE_RX2 */
	{692, 0, 3},  /*!< SLOW_LOOP_STATE__SLOW_LOOP_STATE_RX1 */
	{693, 0, 8},  /*!< GAIN_RX2 */
	{693, 0, 7},  /*!< GAIN_RX2__FULL_TABLE_GAIN_INDEX */
	{694, 0, 8},  /*!< LPF_GAIN_RX2 */
	{694, 0, 5},  /*!< LPF_GAIN_RX2__LPF_GAIN_RX */
	{695, 0, 8},  /*!< DIG_GAIN_RX2 */
	{695, 0, 5},  /*!< DIG_GAIN_RX2__DIGITAL_GAIN_RX */
	{696, 0, 8},  /*!< OVRG_SIGS_RX1 */
	{696, 6, 1},  /*!< OVRG_SIGS_RX1__GAIN_LOCK_1 */
	{696, 5, 1},  /*!< OVRG_SIGS_RX1__LOW_POWER_1 */
	{696, 4, 1},  /*!< OVRG_SIGS_RX1__LARGE_LMT_OL */
	{696, 3, 1},  /*!< OVRG_SIGS_RX1__SMALL_LMT_OL */
	{696, 2, 1},  /*!< OVRG_SIGS_RX1__LARGE_ADC_OL */
	{696, 1, 1},  /*!< OVRG_SIGS_RX1__SMALL_ADC_OL */
	{696, 0, 1},  /*!< OVRG_SIGS_RX1__DIG_SAT */
	{697, 0, 8},  /*!< OVRG_SIGS_RX2 */
	{697, 6, 1},  /*!< OVRG_SIGS_RX2__GAIN_LOCK_1 */
	{697, 5, 1},  /*!< OVRG_SIGS_RX2__LOW_POWER_1 */
	{697, 4, 1},  /*!< OVRG_SIGS_RX2__LARGE_LMT_OL */
	{697, 3, 1},  /*!< OVRG_SIGS_RX2__SMALL_LMT_OL */
	{697, 2, 1},  /*!< OVRG_SIGS_RX2__LARGE_ADC_OL */
	{697, 1, 1},  /*!< OVRG_SIGS_RX2__SMALL_ADC_OL */
	{697, 0, 1},  /*!< OVRG_SIGS_RX2__DIG_SAT */
	{991, 0, 8},  /*!< CTRL */
	{991, 0, 1},  /*!< CTRL__CTRL_ENABLE */
	{1012, 0, 8}, /*!< BIST_CONFIG */
	{1012, 1, 1}, /*!< BIST_CONFIG__TONE_PRBS */
	{1012, 0, 1}, /*!< BIST_CONFIG__BIST_ENABLE */
	{1012, 6, 2}, /*!< BIST_CONFIG__TONE_FREQ */
	{1012, 4, 2}, /*!< BIST_CONFIG__TONE_LEVEL */
	{1012, 2, 2}, /*!< BIST_CONFIG__BIST_CTRL_POINT */
	{1013, 0, 8}, /*!< OBSERVE_CONFIG */
	{1013, 7, 1}, /*!< OBSERVE_CONFIG__DATA_PORT_SP_HD_LOOP_TEST_OE */
	{1013, 6, 1}, /*!< OBSERVE_CONFIG__RX_MASK */
	{1013, 5, 1}, /*!< OBSERVE_CONFIG__CHANNEL */
	{1013, 0, 1}, /*!< OBSERVE_CONFIG__DATA_PORT_LOOP_TEST_ENABLE */
	{1013, 1, 4}, /*!< OBSERVE_CONFIG__OBSERVATION_POINT */
	{1014, 0, 8}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG */
	{1014, 5, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_2_Q_DATA */
	{1014, 4, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_2_I_DATA */
	{1014, 3, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_1_Q_DATA */
	{1014, 2, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__BIST_MASK_CHANNEL_1_I_DATA */
	{1014, 1, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__DATA_PORT_HILOW */
	{1014, 0, 1}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__USE_DATA_PORT */
	{1014, 6, 2}, /*!< BIST_AND_DATA_PORT_TEST_CONFIG__TEMP_SENSE_VBE_TEST */
	{1020, 0, 8}, /*!< DAC_TEST_0 */
	{1021, 0, 8}, /*!< DAC_TEST_1 */
	{1022, 0, 8}, /*!< DAC_TEST_2 */
	{1022, 7, 1}, /*!< DAC_TEST_2__DAC_TEST_ENABLE */
	{1022, 0, 7}, /*!< DAC_TEST_2__DAC_TEST_WORD */
};

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

/**
@brief Flexible register read function (muxing/pagination handled externally)
@param brd  Target board number
@param r    Pointer to register parameters
@param val  Pointer to a variable that will receive the register value
@return 0 if register was read successfully, -1 else
*/
static int reg_read(uint8_t brd, uint8_t spi_header, const struct reg *r, int32_t *val);

/**
@brief Flexible register write function (muxing/pagination/write protection handled externally)
@param brd  Target board number
@param r    Pointer to register parameters
@param val  Value to write to the register
@return 0 if register was written successfully, -1 else
*/
static int reg_write(uint8_t brd, uint8_t spi_header, const struct reg *r, int32_t val);

/**
@brief Switch SX1301 page register if it's not already at the target page
@param brd   Target board number
@param chip  Target chip number
@param page  Target page number
@return 0 if SX1301 page register is set correctly, -1 else
*/
static int chip_page_switch(uint8_t brd, uint8_t chip, int8_t page);

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

int sx1301ar_reg_setup(uint8_t brd, uint8_t brd_type, spi_r_t spi_read, spi_w_t spi_write)
{
	unsigned i; /* loop variable */
	sx1301ar_breg_t idx = BREG_;
	struct reg cur_breg;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_type != BRD_MASTER) && (brd_type != BRD_SLAVE)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_TYPE);
	}
	if ((spi_read == NULL) || (spi_write == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Store the function pointers for this board */
	brd_stat_priv[brd].spi_r = spi_read;
	brd_stat_priv[brd].spi_w = spi_write;

	if (brd_type == BRD_MASTER) {
		brd_stat_priv[brd].header_reg_fpga = SPI_HDR_REG_FPGA;
		brd_stat_priv[brd].header_reg_chip = SPI_HDR_SX1301_0; /* first chip of the board */
	} else {
		brd_stat_priv[brd].header_reg_fpga = SPI_HDR_REG_FPGA_SLAVE;
		brd_stat_priv[brd].header_reg_chip = SPI_HDR_SX1301_2; /* first chip of the board */
	}

	/* Initialize the register tables */
	memset(&breg_table[brd][0], -1, SX1301AR_BREG_NB * sizeof(struct reg));
	i = 0;
	do {
		if (brd_type == BRD_MASTER) {
			idx = breg_master[i].breg_idx;
			cur_breg = breg_master[i].breg;
		} else {
			idx = breg_slave[i].breg_idx;
			cur_breg = breg_slave[i].breg;
		}

		if (idx != BREG_) {
			memcpy(&breg_table[brd][idx], &cur_breg, sizeof(struct reg));
		}
		i++;
	} while ((idx != -1) && (i < SX1301AR_BREG_NB));

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_init(uint8_t brd, uint8_t nb_chip, int16_t *fpga_version)
{
	int x;
	unsigned i; /* loop variable */
	int32_t val;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		debug("spi_r/w=NULL");
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}

	/* Soft-reset the whole board */
	x = sx1301ar_reg_reset(brd);
	x += sx1301ar_reg_reset(brd); // Second soft-reset, because first SPI transaction is discarded on PCDuino
	SX1301AR_ERR_PROPAGATE(x);

	/* Check FPGA version register */
	x = sx1301ar_reg_brd_r(brd, BREG_VERSION, &val);
	SX1301AR_ERR_PROPAGATE(x);
	if (val == 0) {
		debug("[BRD%u] FPGA version = %i", brd, val); // DEBUG
		SX1301AR_ERR_RETURN(ERR_FPGA_VZERO);
	} else if (val != breg_table[brd][BREG_VERSION].dflt) {
		debug("[BRD%u] FPGA version = %i", brd, val); // DEBUG
		SX1301AR_ERR_RETURN(ERR_FPGA_VERSION);
	}
	if (fpga_version != NULL) {
		*fpga_version = val;
	}

	/* Check version register and chip ID of each SX1301 */
	for (i = 0; i < nb_chip; ++i) {
		x = sx1301ar_reg_chip_r(brd, i, CREG_VERSION, &val);
		SX1301AR_ERR_PROPAGATE(x);
		if (val == 0) {
			SX1301AR_ERR_RETURN(ERR_CHIP_VZERO);
		} else if (val != creg[CREG_VERSION].dflt) {
			debug("[BRD%u] SX1301 #%u version = %i", brd, i, val); // DEBUG
			SX1301AR_ERR_RETURN(ERR_CHIP_VERSION);
		}
		x = sx1301ar_reg_chip_r(brd, i, CREG_CHIP_ID, &val);
		SX1301AR_ERR_PROPAGATE(x);
		if (val != creg[CREG_CHIP_ID].dflt) {
			debug("SX1301 #%u chip ID = %i", i, val); // DEBUG
			SX1301AR_ERR_RETURN(ERR_CHIP_ID);
		}
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_reset(uint8_t brd)
{
	int i, j; /* loop & return variable */
	uint8_t data;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/*
    This function only writes to the board, no reads.
    That sequence must put the board in a known state (mux & pages).
    */

	/* Soft-reset SX1301 registers */
	/*** Not needed, since SX1301 reset pin is controlled by FPGA ***/
	for (i = 0; i < SX1301AR_BOARD_CHIPS_NB; ++i) {
		//data = i;
		//j = (*(brd_stat_priv[brd].spi_w))( MUX_ADDRESS, &data, 1, NULL );
		//RETURN_SPI_W_ERR( j );

		//data = 1 << creg[CREG_SOFT_RESET].offs;
		//j = (*(brd_stat_priv[brd].spi_w))( creg[CREG_SOFT_RESET].addr, &data, 1, NULL );
		//RETURN_SPI_W_ERR( j );
		brd_stat_priv[brd].chip_page_state[i] = creg[CREG_PAGE_REG].dflt;
	}

	/* Reset radio (reset active low) */
	/*** Should not be here ***/
	//data = 0;
	//j = (*(brd_stat_priv[brd].spi_w))( breg[BREG_AD9361_RESETB].addr, &data, 1, NULL );
	//RETURN_SPI_W_ERR( j );

	/* Soft-reset FPGA registers (will bring radio out of reset) */
	data = 1 << breg_table[brd][BREG_SOFT_RESET].offs;
	j = (*(brd_stat_priv[brd].spi_w))(brd_stat_priv[brd].header_reg_fpga, breg_table[brd][BREG_SOFT_RESET].addr, &data, 1, NULL);
	RETURN_SPI_W_ERR(j);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_brd_r(uint8_t brd, sx1301ar_breg_t idx, int32_t *val)
{
	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((idx >= SX1301AR_BREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}
	if (val == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (breg_table[brd][idx].addr == (uint16_t)(-1)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}

	/* Read register */
	return reg_read(brd, brd_stat_priv[brd].header_reg_fpga, &breg_table[brd][idx], val);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_brd_w(uint8_t brd, sx1301ar_breg_t idx, int32_t val)
{
	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((idx >= SX1301AR_BREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}
	if (breg_table[brd][idx].addr == (uint16_t)(-1)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}

	/* Protect read-only registers */
	if (breg_table[brd][idx].rdon == true) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Write register */
	return reg_write(brd, brd_stat_priv[brd].header_reg_fpga, &breg_table[brd][idx], val);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_brd_rb(uint8_t brd, sx1301ar_breg_t idx, uint8_t *data, uint32_t size)
{
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((idx >= SX1301AR_BREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}
	if (data == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (breg_table[brd][idx].addr == (uint16_t)(-1)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}

	/* Read register */
	i = (*(brd_stat_priv[brd].spi_r))(brd_stat_priv[brd].header_reg_fpga, breg_table[brd][idx].addr, data, size, NULL);
	RETURN_SPI_R_ERR(i);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_brd_wb(uint8_t brd, sx1301ar_breg_t idx, const uint8_t *data, uint32_t size)
{
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if ((idx >= SX1301AR_BREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}
	if (data == NULL) {
		SX1301AR_ERR_RETURN(ERR_NULL_POINTER);
	}
	if (breg_table[brd][idx].addr == (uint16_t)(-1)) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_REG);
	}

	/* Protect read-only registers */
	if (breg_table[brd][idx].rdon == true) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Write register */
	i = (*(brd_stat_priv[brd].spi_w))(brd_stat_priv[brd].header_reg_fpga, breg_table[brd][idx].addr, data, size, NULL);
	RETURN_SPI_W_ERR(i);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_chip_r(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, int32_t *val)
{
	int x;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((idx >= SX1301AR_CREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if (val == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Switch register page (if needed) */
	x = chip_page_switch(brd, chip, creg[idx].page);
	SX1301AR_ERR_PROPAGATE(x);

	/* Read register */
	return reg_read(brd, (sx1301ar_spi_header_t)(brd_stat_priv[brd].header_reg_chip + chip), &creg[idx], val);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_chip_w(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, int32_t val)
{
	int x;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((idx >= SX1301AR_CREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Protect read-only registers */
	if (creg[idx].rdon == true) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Switch register page (if needed) */
	x = chip_page_switch(brd, chip, creg[idx].page);
	SX1301AR_ERR_PROPAGATE(x);

	/* Write register */
	return reg_write(brd, (sx1301ar_spi_header_t)(brd_stat_priv[brd].header_reg_chip + chip), &creg[idx], val);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_chip_rb(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, uint8_t *data, uint32_t size)
{
	int x;
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((idx >= SX1301AR_CREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if (data == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Switch register page (if needed) */
	x = chip_page_switch(brd, chip, creg[idx].page);
	SX1301AR_ERR_PROPAGATE(x);

	/* Read register */
	i = (*(brd_stat_priv[brd].spi_r))((sx1301ar_spi_header_t)(brd_stat_priv[brd].header_reg_chip + chip), creg[idx].addr, data, size, NULL);
	RETURN_SPI_R_ERR(i);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_reg_chip_wb(uint8_t brd, uint8_t chip, sx1301ar_creg_t idx, const uint8_t *data, uint32_t size)
{
	int x;
	int i;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if (chip >= SX1301AR_BOARD_CHIPS_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_CHIP_NB);
	}
	if ((brd_stat_priv[brd].spi_r == NULL) || (brd_stat_priv[brd].spi_w == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((idx >= SX1301AR_CREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if (data == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Protect read-only registers */
	if (creg[idx].rdon == true) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Switch register page (if needed) */
	x = chip_page_switch(brd, chip, creg[idx].page);
	SX1301AR_ERR_PROPAGATE(x);

	/* Write register */
	i = (*(brd_stat_priv[brd].spi_w))((sx1301ar_spi_header_t)(brd_stat_priv[brd].header_reg_chip + chip), creg[idx].addr, data, size, NULL);
	RETURN_SPI_W_ERR(i);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_ad9361_r(uint8_t brd, sx1301ar_rreg_t idx, uint8_t *val)
{
	int x;
	uint8_t buf[2] = {0, 0};
	uint8_t buf2[3] = {0, 0, 0};
	const struct reg2 *r;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((idx >= SX1301AR_RREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if (val == NULL) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	r = &rreg[idx];

	if (r->leng + r->offs <= 8) /* 1 to 8 bit reg, with eventual offset */
	{
		/* Read one byte */
		buf2[0] = (uint8_t)(r->addr >> 8);
		buf2[1] = (uint8_t)(r->addr & 0xFF);
		x = sx1301ar_spibdg_w_r(brd, buf2, 2, &buf[0], 1);
		SX1301AR_ERR_PROPAGATE(x);

		/* Shift to dump unwanted bits and extend sign if needed */
		if (r->leng < 8) {
			buf[1] = buf[0] << (8 - r->leng - r->offs); /* left-align the data */
			buf[0] = buf[1] >> (8 - r->leng);	   /* right align the data, no sign extension */
		}

		*val = buf[0];
	} else /* unsupported register type */
	{
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_ad9361_w(uint8_t brd, sx1301ar_rreg_t idx, uint8_t val)
{
	int x;
	uint8_t buf[4] = {0, 0, 0, 0};
	uint8_t buf2[3] = {0, 0, 0};
	const struct reg2 *r;

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((idx >= SX1301AR_RREG_NB) || (idx < 0)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	r = &rreg[idx];

	if (r->leng + r->offs <= 8) /* single byte read-modify-write */
	{
		/* Read one byte */
		buf2[0] = (uint8_t)(r->addr >> 8);
		buf2[1] = (uint8_t)(r->addr & 0xFF);
		x = sx1301ar_spibdg_w_r(brd, buf2, 2, &buf[3], 1);
		SX1301AR_ERR_PROPAGATE(x);

		buf[2] = val << r->offs;			 /* new data, at the right position */
		buf[1] = ((1 << r->leng) - 1) << r->offs;	/* bit mask */
		buf[0] = (~buf[1] & buf[3]) | (buf[1] & buf[2]); /* mixing old & new data */
		/* SPI write will only write buf[0] */
	} else /* unsupported register type */
	{
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Write one byte */
	buf2[0] = (1 << 7) | (uint8_t)(r->addr >> 8);
	buf2[1] = (uint8_t)(r->addr & 0xFF);
	buf2[2] = buf[0];
	return sx1301ar_spibdg_w_r(brd, buf2, 3, NULL, 0);
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

int sx1301ar_spibdg_w_r(uint8_t brd, const uint8_t *tx_data, uint8_t tx_size, uint8_t *rx_data, uint8_t rx_size)
{
	int x;
	int i;
	int32_t val;
	uint8_t buff[SPI_BG_FIFO_SIZE];

	/* Check input parameters */
	if (brd >= SX1301AR_MAX_BOARD_NB) {
		SX1301AR_ERR_RETURN(ERR_BAD_BOARD_NB);
	}
	if ((tx_size > 0) && (tx_data == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if ((rx_size > 0) && (rx_data == NULL)) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}
	if (rx_size + tx_size > SPI_BG_FIFO_SIZE) {
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	if (rx_size == 0) /* zero-copy write-only */
	{
		/* Put outbound data into SPI FIFO */
		x = sx1301ar_reg_brd_wb(brd, BREG_AD9361_SPI_DATA_OUT, tx_data, tx_size);
		SX1301AR_ERR_PROPAGATE(x);
	} else {
		/* Copy outbound data in a bigger buffer and send it (zero padded) */
		memset((void *)buff, 0, sizeof buff);
		memcpy((void *)buff, (void *)tx_data, tx_size);
		x = sx1301ar_reg_brd_wb(brd, BREG_AD9361_SPI_DATA_OUT, buff, rx_size + tx_size);
		SX1301AR_ERR_PROPAGATE(x);
	}

	/* Trigger SPI exchange and check status */
	x = sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_SEL, 1);
	SX1301AR_ERR_PROPAGATE(x);
	val = -1; /* 1 if SPI is buzzy, 0 when finished */
	for (i = 0; (i < SPI_BG_TIMEOUT) && (val != 0); ++i) {
		x = sx1301ar_reg_brd_r(brd, BREG_AD9361_SPI_SEL, &val);
		SX1301AR_ERR_PROPAGATE(x);
	}
	if (val != 0) /* time-out */
	{
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	x = sx1301ar_reg_brd_w(brd, BREG_AD9361_SPI_SEL, 0);
	SX1301AR_ERR_PROPAGATE(x);

	if (rx_size != 0) {
		/* Get inbound data and copy to rx buffer */
		x = sx1301ar_reg_brd_rb(brd, BREG_AD9361_SPI_DATA_IN, buff, rx_size + tx_size);
		SX1301AR_ERR_PROPAGATE(x);
		memcpy((void *)rx_data, (void *)(buff + tx_size), rx_size);
	}

	return 0;
}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

static int reg_read(uint8_t brd, uint8_t spi_header, const struct reg *r, int32_t *val)
{
	int i;
	uint8_t size_byte;
	uint8_t buf_u[4] = {0, 0, 0, 0};
	int8_t *buf_s = (int8_t *)buf_u;
	uint32_t u;

	/* /!\ Input parameters assumed correct /!\ */

	/* Read one to four bytes */
	size_byte = (r->leng + 7) / 8; /* add a byte if it's not an exact multiple of 8 */
	i = (*(brd_stat_priv[brd].spi_r))(spi_header, r->addr, buf_u, size_byte, NULL);
	RETURN_SPI_R_ERR(i);

	if (r->leng + r->offs <= 8) /* 1 to 8 bit reg, with eventual offset */
	{
		/* Shift to dump unwanted bits and extend sign if needed */
		if (r->leng < 8) {
			buf_u[1] = buf_u[0] << (8 - r->leng - r->offs); /* left-align the data */
			if (r->sign == true) {
				buf_s[0] = buf_s[1] >> (8 - r->leng); /* right align the data with sign extension (ARITHMETIC right shift) */
			} else {
				buf_u[0] = buf_u[1] >> (8 - r->leng); /* right align the data, no sign extension */
			}
		}

		/* Extend 8 bits to 32 bits */
		if (r->sign == true) {
			*val = (int32_t)buf_s[0]; /* signed pointer -> 32b sign extension */
		} else {
			*val = (int32_t)buf_u[0]; /* unsigned pointer -> no sign extension */
		}
	} else if ((r->leng > 8) && (r->leng <= 32) && (r->offs == 0)) /* 9+ bits, no offset */
	{
		/* Transform a little endian byte array into a 32 bit word */
		u = 0;
		for (i = 0; i < size_byte; ++i) {
			u += (uint32_t)buf_u[i] << (8 * i);
		}

		/* Extend sign bit if needed */
		if (r->sign == true) {
			u = u << (32 - r->leng);	     /* left-align the data */
			*val = (int32_t)u >> (32 - r->leng); /* right-align the data with sign extension (ARITHMETIC right shift) */
		} else {
			*val = (int32_t)u; /* unsigned value -> return 'as is' */
		}
	} else /* unsupported register type */
	{
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int reg_write(uint8_t brd, uint8_t spi_header, const struct reg *r, int32_t val)
{
	int i;
	uint8_t size_byte;
	uint8_t buf[4] = {0, 0, 0, 0};

	/* /!\ Input parameters assumed correct /!\ */

	if ((r->offs == 0) && (r->leng >= 8) && (r->leng <= 32)) /* 1 to 4 bytes direct write */
	{
		/* Transform a 32 bit word into a little endian byte array */
		size_byte = (r->leng + 7) / 8; /* add a byte if it's not an exact multiple of 8 */
		for (i = 0; i < size_byte; ++i) {
			buf[i] = (uint8_t)(0xFF & (val >> (8 * i)));
		}
	} else if (r->leng + r->offs <= 8) /* single byte read-modify-write */
	{
		size_byte = 1;

		/* Read one byte */
		i = (*(brd_stat_priv[brd].spi_r))(spi_header, r->addr, &buf[3], 1, NULL);
		RETURN_SPI_R_ERR(i);

		buf[2] = (uint8_t)(0xFF & val) << r->offs;       /* new data, at the right position */
		buf[1] = ((1 << r->leng) - 1) << r->offs;	/* bit mask */
		buf[0] = (~buf[1] & buf[3]) | (buf[1] & buf[2]); /* mixing old & new data */
		size_byte = 1;					 /* SPI write will only write buf[0] */
	} else							 /* unsupported register type */
	{
		SX1301AR_ERR_RETURN(ERR_UNSPECIFIED); // TODO: define specific code
	}

	/* Write one to four bytes */
	i = (*(brd_stat_priv[brd].spi_w))(spi_header, r->addr, buf, size_byte, NULL);
	RETURN_SPI_W_ERR(i);

	return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int chip_page_switch(uint8_t brd, uint8_t chip, int8_t page)
{
	int i;
	uint8_t data;

	if ((page != -1) && (brd_stat_priv[brd].chip_page_state[chip] != page)) {
		/* Need to switch */
		data = page & 0x03; /* 4 pages max */
		i = (*(brd_stat_priv[brd].spi_w))((sx1301ar_spi_header_t)(brd_stat_priv[brd].header_reg_chip + chip), creg[CREG_PAGE_REG].addr, &data, 1, NULL);
		RETURN_SPI_W_ERR(i);
		brd_stat_priv[brd].chip_page_state[chip] = page;
	}
	return 0;
}

/* --- EOF ------------------------------------------------------------------ */
